/*-----------------------------------------------------------------------
 * File: PARSEBER.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "ber_der.h"
#include "ber_internal.h"

#define AUTO_BER_ITEMS 30
/*
 * parse complex BER object, given a DER_COMPLEX_TYPE structure for it.
 * return parse tree
 */
DER_NODE_PTR 
BER_ParseTypedObject(         /* return parse tree */
    BER_HANDLE AppHandle,
    BER_MEMORY_FUNCS_PTR MemoryFunctions, 
    DER_COMPLEX_TYPE_PTR Type,  /* type of object */
    BER_PARSED_ITEM_PTR  Item)  /* BER object broken into tag, length, content */
{
    sint32 i, j, nItems=0;
    BER_PARSED_ITEM *items = 0, autoItems[AUTO_BER_ITEMS];
    DER_NODE *node, staticNode;
    DER_COMPLEX_TYPE *childType;

    if (!Type)
        return 0;
    node = 0;
    if (Item->Tag[0] & BER_CONSTRUCTED) {
        node = &staticNode;
        node->Tag = Item->Tag;
        node->Length = Item->ContentLength; 
        node->OriginalTag = Item->Tag;
        node->Type = Type;
        if (Type->TypeTag && (Type->TypeTag[0] & 31) == BER_SET) {
            nItems = BER_CountItems(Item->Content, Item->ContentLength) + Type->NumberOfDefaults;
            if (nItems > AUTO_BER_ITEMS) 
                items = BERDER_MALLOC_WRAPPER(MemoryFunctions->malloc_func, AppHandle, nItems * sizeof(BER_PARSED_ITEM));
            else 
                items = autoItems;
            node->Count = BER_ExpandSet(Item->Content, 
                Item->ContentLength,
                nItems,
                Type->NumberOfDefaults,
                Type->Defaults, 
                items);
        } else if (Type->TypeTag && (Type->TypeTag[0] & 31) == BER_SEQUENCE) {
            nItems = Type->NumberOfDefaults;
            if (nItems > AUTO_BER_ITEMS) 
                items = BERDER_MALLOC_WRAPPER(MemoryFunctions->malloc_func, AppHandle, nItems * sizeof(BER_PARSED_ITEM));
            else 
                items = autoItems;
            node->Count = BER_ExpandSequence(
                Item->Content,
                Item->ContentLength, 
                Type->NumberOfDefaults,
                Type->Choices,
                Type->ContextTags,
                Type->OriginalTags,
                Type->Defaults,     
                items);
            if (node->Count < Type->NumberOfDefaults) {
                if (nItems > AUTO_BER_ITEMS) 
                    BERDER_FREE_WRAPPER(MemoryFunctions->free_func, AppHandle, items);
                return 0;    /* syntax error */
            }
        } 
        /* allocate the proper size node, now that we know how many BER components it has */
        node = DER_AllocateNode(AppHandle, MemoryFunctions, node->Count);
        if (!node) {
            if (nItems > AUTO_BER_ITEMS) 
                BERDER_FREE_WRAPPER(MemoryFunctions->free_func, AppHandle, items);
            return 0;
        }
        node->Tag = staticNode.Tag;
        node->Length = staticNode.Length;   
        node->OriginalTag = staticNode.OriginalTag;
        node->Type = staticNode.Type;
        /* expand all constructed children, copy the rest */
        for (i = 0; i < node->Count; i++) {
            node->Child[i].IsLeaf =  (uint8) (items[i].Tag == 0 || (items[i].Tag[0] & BER_CONSTRUCTED) == 0);
            node->Child[i].X.Input = items[i];
            node->Child[i].X.Node = 0;
            if (!node->Child[i].IsLeaf) {
                childType = 0;
                /* check for implicit or explicit choices */
                if (Type->Choices && Type->Choices[i]) {
                    DER_EXPLICIT_CHOICES_PTR choice;
                    choice = Type->Choices[i];
                    if (choice) {
                        for (j = 0; j < choice->NumberOfChoices; j++) {
                            if (BER_TagCompare(choice->Oids[j]->Tag, node->Child[choice->ParentField].X.Input.Tag))
                                continue;
                            if (choice->ParentField != i) {  /* the tag is enough if same field */
                                if (choice->Oids[j]->ContentLength != node->Child[choice->ParentField].X.Input.ContentLength) 
                                    continue;
                                if (cssm_memcmp(choice->Oids[j]->Content, node->Child[choice->ParentField].X.Input.Content, 
                                    choice->Oids[j]->ContentLength)) continue;
                            }
                            childType = choice->Subtypes[j];
                            break;
                        }
                    }
                } else if (Type->Subtypes) {     /* try simple implicit tagging */
                    switch (Type->TypeTag[0] & 31) {
                    case BER_SET:
                        for (j = 0; j < Type->NumberOfDefaults; j++) {
                            if (!BER_TagCompare(Type->ContextTags[j], node->Child[i].X.Input.Tag)) {
                                childType = Type->Subtypes[j];
                                break;
                            }
                        }
                        if (j==Type->NumberOfDefaults)
                            childType = Type->Subtypes[j]; /* If no defaults are identified for the set, use the subtype itself */
                        break;
                    default:
                        childType = Type->Subtypes[i];
                        break;
                    }
                }
                if (childType) { /* don't parse subtree unless typed */
                    node->Child[i].X.Node = BER_ParseTypedObject(AppHandle, MemoryFunctions, childType, &items[i]);
                    if (!node->Child[i].X.Node) {   /* an error has occurred--propagate */
                        if (nItems > AUTO_BER_ITEMS) 
                            BERDER_FREE_WRAPPER(MemoryFunctions->free_func, AppHandle, items);
                        return 0;
                    }
                } else {    /* make untyped subtrees leaves */
                    node->Child[i].X.Node = 0;
                    node->Child[i].IsLeaf = 1;
                }
            }
        }
        if (nItems > AUTO_BER_ITEMS) 
                BERDER_FREE_WRAPPER(MemoryFunctions->free_func, AppHandle, items);
    }
    return node;
}
