/*-----------------------------------------------------------------------
 * File: CLDERLL.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 

#include "cssmport.h"
#include "guids.h"
#include "clderll.h"
#include "internal.h"
#include <cssmMemTrack.h>


/*-----------------------------------------------------------------------------
 * Name: cl_der_FindLeafNode
 *
 * Description:
 * Finds the Leaf node with the input handle
 *
 * Parameters: 
 * Handle (input) : The handle of the leaf node 
 *
 * result_val (output): A pointer to the leaf node
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_FindLeafNode(CSSM_CLUTIL_CONTEXT *pContext,
                                CL_DER_NODE_HANDLE Handle,
                                CL_DER_LL_LEAF_NODE_PTR *result_val)
{
    CL_DER_LL_LEAF_NODE_PTR end_node = pContext->LeafCurrentPtr;
    
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }    

    *result_val = NULL;
    /* If the handle is invalid or the list is empty, error */
    if (Handle == 0 || pContext->LeafCurrentPtr == NULL)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* If there is only 1 node in the list, check the value & return */
    if (pContext->LeafCurrentPtr->NextNode == pContext->LeafCurrentPtr)
    {
        if (pContext->LeafCurrentPtr->Handle == Handle)
        {
            *result_val = pContext->LeafCurrentPtr;
            return CSSM_OK;
        }
        else 
        {
            /* This handle is not in the list */
            return CSSMERR_CL_INVALID_DATA;
        }
    }

    /* Find the node */
    do
    {
        if (pContext->LeafCurrentPtr->Handle == Handle)
        {
            break;
        }
        pContext->LeafCurrentPtr = pContext->LeafCurrentPtr->NextNode;
    }
    while (pContext->LeafCurrentPtr != end_node);

    if (pContext->LeafCurrentPtr->Handle == Handle)
    {
        *result_val = pContext->LeafCurrentPtr;
        return CSSM_OK;
    }

    return CSSMERR_CL_INVALID_DATA;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_GetLeafNode
 *
 * Description:
 * This functions return a pointer to the BER parsed item stored in the leaf node
 *
 * Parameters: 
 * Handle (input) : The handle of the node from which to 
 *                  retrieve the BER parsed item
 *
 * result_val (output): A pointer to the BER parsed item in the node
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_GetLeafNode( CSSM_CLUTIL_CONTEXT *pContext,
                                CL_DER_NODE_HANDLE Handle,
                                BER_PARSED_ITEM_PTR *result_val)
{
    CL_DER_LL_LEAF_NODE_PTR leaf_node_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }    

    *result_val = NULL;

    if (Handle == 0 || pContext->LeafCurrentPtr == NULL)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

    ret = cl_der_FindLeafNode(pContext, Handle, &leaf_node_ptr);

    if (CSSM_OK != ret)
    {
        return ret;
    }
    
    *result_val = leaf_node_ptr->BerItemPtr;
    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_FindParentNode
 *
 * Description:
 * Finds the parent node with the input handle
 *
 * Parameters: 
 * Handle (input) : The handle of the parent node 
 *
 * result_val (output) : A pointer to the parent node
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_FindParentNode(CSSM_CLUTIL_CONTEXT *pContext,
                                  CL_DER_NODE_HANDLE Handle,
                                  CL_DER_LL_PARENT_NODE_PTR *result_val)
{
    CL_DER_LL_PARENT_NODE_PTR end_node = pContext->ParentCurrentPtr;

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }    

    *result_val = NULL;

    if (Handle == 0 || pContext->ParentCurrentPtr == NULL)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* If there is only 1 node in the list, check the value & return */
    if (pContext->ParentCurrentPtr->NextNode == pContext->ParentCurrentPtr)
    {
        if (pContext->ParentCurrentPtr->Handle == Handle)
        {
            *result_val = pContext->ParentCurrentPtr;
            return CSSM_OK;
        }
        else 
        {
            /* This handle is not in the list */
            return CSSMERR_CL_INVALID_DATA;
        }
    }

    /* If there is more than 1 node in the list.... */
    do
    {
        if (pContext->ParentCurrentPtr->Handle == Handle)
        {
            break;
        }
        pContext->ParentCurrentPtr = pContext->ParentCurrentPtr->NextNode;
    }
    while (pContext->ParentCurrentPtr != end_node);

    if (pContext->ParentCurrentPtr->Handle == Handle)
    {
        *result_val = pContext->ParentCurrentPtr;
        return CSSM_OK;
    }

    return CSSMERR_CL_INVALID_DATA;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_GetParentNode
 *
 * Description:
 * Finds the parent DER node with the input handle
 *
 * Parameters: 
 * Handle (input) : The handle of the parent node 
 *
 * result_val (putput): A pointer to the DER node
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_GetParentNode(CSSM_CLUTIL_CONTEXT *pContext,
                                    CL_DER_NODE_HANDLE Handle,
                                    DER_NODE_PTR *result_val)
{
    CL_DER_LL_PARENT_NODE_PTR parent_node_ptr =NULL;
    CSSM_RETURN ret = CSSM_OK;

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }    

    *result_val = NULL;

    if (Handle == 0 || pContext->ParentCurrentPtr == NULL)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

    ret = cl_der_FindParentNode(pContext, Handle, &parent_node_ptr);

    if (CSSM_OK != ret)
    {
        return ret;
    }

    *result_val = parent_node_ptr->DerNodePtr;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_AddLeafNode
 *
 * Description:
 * This functions creates a leaf node containing the input data 
 * and then adds it to the linked list of leaf nodes
 *
 * Parameters: 
 * NodePtr (input) : The data to be added to the list of leaf nodes
 *
 * results_handle (output): A handle identifying the leaf node 
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_AddLeafNode(CSSM_CLUTIL_CONTEXT *pContext,
                               BER_PARSED_ITEM_PTR NodePtr,
                               CL_DER_NODE_HANDLE *results_handle)
{
    CSSM_RETURN ret = CSSM_OK;
    CL_DER_LL_LEAF_NODE_PTR new_leaf_ptr = NULL;

    if (!results_handle)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }    

    *results_handle = 0;
    /* Check input */

    if (!NodePtr || port_IsBadReadPtr(NodePtr, sizeof(BER_PARSED_ITEM)))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
       
    }
    if (!NodePtr->Tag || 
        ((NodePtr->Content != NULL) && port_IsBadReadPtr(NodePtr->Content, NodePtr->ContentLength)))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;      
    }

    /* Create a new leaf node */
    new_leaf_ptr = CSSM_MALLOC_WRAPPER(internal_malloc, sizeof(CL_DER_LL_LEAF_NODE), NULL);
    if (!new_leaf_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;       
    }

    /* Fill in the new leaf node */
    new_leaf_ptr->Handle = clutil_GetHandle(pContext);
    new_leaf_ptr->BerItemPtr = NodePtr;

    /* Add it to the list */
    /* If this is the first item in the list */
    if (pContext->LeafCurrentPtr == NULL)
    {
        new_leaf_ptr->NextNode = new_leaf_ptr;
        new_leaf_ptr->PreviousNode = new_leaf_ptr;
        pContext->LeafCurrentPtr = new_leaf_ptr;
    } 
    else
    {
        new_leaf_ptr->NextNode = pContext->LeafCurrentPtr;
        new_leaf_ptr->PreviousNode = pContext->LeafCurrentPtr->PreviousNode;

        pContext->LeafCurrentPtr->PreviousNode->NextNode = new_leaf_ptr;
        pContext->LeafCurrentPtr->PreviousNode = new_leaf_ptr;

        pContext->LeafCurrentPtr = new_leaf_ptr;
    }

    /* Return the node handle */
    *results_handle = new_leaf_ptr->Handle;
    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_AddParentNode
 *
 * Description:
 * This functions creates a parent node containing the input data 
 * and then adds it to the linked list of parent nodes
 *
 * Parameters: 
 * NodePtr (input) : The data to be added to the list of parent nodes
 * result_val(output): A handle identifying the parent node 
 *---------------------------------------------------------------------------*/
CSSM_RETURN  cl_der_AddParentNode(CSSM_CLUTIL_CONTEXT *pContext,
                                  DER_NODE_PTR NodePtr,
                                  CL_DER_NODE_HANDLE *result_val)
{
    CL_DER_LL_PARENT_NODE_PTR new_parent_ptr = NULL;

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }    

    *result_val = 0;
    /* Check input */
    if (!NodePtr || port_IsBadReadPtr(NodePtr, sizeof(DER_NODE)))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
        
    }
    if (!NodePtr->Tag)
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
        
    }

    /* Create a new parent node */
    new_parent_ptr = CSSM_MALLOC_WRAPPER(internal_malloc, sizeof(CL_DER_LL_PARENT_NODE), NULL);
    if (!new_parent_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Fill in the new parent node */
    new_parent_ptr->Handle = clutil_GetHandle(pContext);
    new_parent_ptr->DerNodePtr = NodePtr;

    /* Add it to the front of the list */
    /* If this is the first item in the list */
    if (pContext->ParentCurrentPtr == NULL)
    {
        new_parent_ptr->NextNode = new_parent_ptr;
        new_parent_ptr->PreviousNode = new_parent_ptr;
        pContext->ParentCurrentPtr = new_parent_ptr;
    } 
    else 
    {
        new_parent_ptr->NextNode = pContext->ParentCurrentPtr;
        new_parent_ptr->PreviousNode = pContext->ParentCurrentPtr->PreviousNode;

        pContext->ParentCurrentPtr->PreviousNode->NextNode = new_parent_ptr;
        pContext->ParentCurrentPtr->PreviousNode = new_parent_ptr;

        pContext->ParentCurrentPtr = new_parent_ptr;
    }

    /* Return the node handle */
    *result_val = new_parent_ptr->Handle;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_RemoveLeafNode
 *
 * Description:
 * Remove the node from the linked list
 * Free the node and the BER parsed item, but not the BER parsed item
 * tag or the BER parsed item content.
 *
 * Parameters: 
 * Handle (input) : The handle of the node to be removed
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_RemoveLeafNode(CSSM_CLUTIL_CONTEXT *pContext,
                                  CL_DER_NODE_HANDLE Handle)
{
    CL_DER_LL_LEAF_NODE_PTR leaf_node_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;

    /* If the handle is invalid or the list is empty, error */
    if (Handle == 0 || pContext->LeafCurrentPtr == NULL)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

     ret = cl_der_FindLeafNode(pContext, Handle, &leaf_node_ptr);

    /* If the leaf was not found, return */
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* Remove the Node from the list */
    /* If it is the only node, set the current pointer to NULL */
    if (leaf_node_ptr->NextNode == leaf_node_ptr)
    {
        pContext->LeafCurrentPtr = NULL;
    } 
    else 
    {   /* Otherwise, fix the list pointers */
        leaf_node_ptr->PreviousNode->NextNode = leaf_node_ptr->NextNode;
        leaf_node_ptr->NextNode->PreviousNode = leaf_node_ptr->PreviousNode;
        pContext->LeafCurrentPtr = leaf_node_ptr->NextNode;
    }

    /* Delete the node */
    CSSM_FREE_WRAPPER(internal_free, leaf_node_ptr->BerItemPtr, NULL);
    CSSM_FREE_WRAPPER(internal_free, leaf_node_ptr, NULL);

    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: cl_der_RemoveParentNode
 *
 * Description:
 * Remove the node from the linked list
 * Free the linked list node, but not the DER node it contains
 *
 * Parameters: 
 * Handle (input) : The handle of the node to be removed
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_RemoveParentNode(CSSM_CLUTIL_CONTEXT *pContext,
                                    CL_DER_NODE_HANDLE Handle)
{
    CL_DER_LL_PARENT_NODE_PTR parent_node_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;

    if (Handle == 0 || pContext->ParentCurrentPtr == NULL)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

    ret = cl_der_FindParentNode(pContext, Handle, &parent_node_ptr);

    /* If the node was not found, return */
    if (CSSM_OK != ret)
    {
        return ret;
        
    }

    /* Remove the Node from the list */
    /* If it is the only node, set the pointers to NULL */
    if (parent_node_ptr->NextNode == parent_node_ptr)
    {
        pContext->ParentCurrentPtr = NULL;
    }
    else 
    {   
        /* Otherwise, fix the list pointers */
        parent_node_ptr->PreviousNode->NextNode = parent_node_ptr->NextNode;
        parent_node_ptr->NextNode->PreviousNode = parent_node_ptr->PreviousNode;
        pContext->ParentCurrentPtr = parent_node_ptr->NextNode;
    }

    /* Delete the node */
    CSSM_FREE_WRAPPER(internal_free, parent_node_ptr, NULL);

    return CSSM_OK;
}



/*-----------------------------------------------------------------------------
 * Name: cl_der_DeleteAllLeafNodes
 *
 * Description: 
 * This functions frees all of the memory associated with the linked list
 * of leaf nodes.
 *
 * Parameters: 
 *---------------------------------------------------------------------------*/
void cl_der_DeleteAllLeafNodes(CSSM_CLUTIL_CONTEXT *pContext)
{
    if (pContext->LeafCurrentPtr == NULL)
    {
        return;
    }
    /* Delete all the nodes */
    pContext->LeafCurrentPtr->PreviousNode->NextNode = NULL;
    while (pContext->LeafCurrentPtr != NULL) 
    {
        CSSM_FREE_WRAPPER(internal_free, (void*)pContext->LeafCurrentPtr->BerItemPtr->Tag, NULL);
        CSSM_FREE_WRAPPER(internal_free, (void*)pContext->LeafCurrentPtr->BerItemPtr->Content, NULL);
        CSSM_FREE_WRAPPER(internal_free, (void*)pContext->LeafCurrentPtr->BerItemPtr, NULL);
        pContext->LeafCurrentPtr->BerItemPtr = NULL;
        
        if (pContext->LeafCurrentPtr->NextNode == NULL) /* If this is the last node */
        {
            CSSM_FREE_WRAPPER(internal_free, pContext->LeafCurrentPtr, NULL);
            pContext->LeafCurrentPtr = NULL;
        }
        else
        {
            pContext->LeafCurrentPtr = pContext->LeafCurrentPtr->NextNode;
            CSSM_FREE_WRAPPER(internal_free, pContext->LeafCurrentPtr->PreviousNode, NULL);
        }
    } 

    return;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_FreeDerNodeTree
 *
 * Description: 
 * This functions frees all of the structures in the DER NODE tree
 * and then frees the tree structure itself.
 *
 * Parameters: 
 *
 * Return value:
 *
 * Error Codes:
 *---------------------------------------------------------------------------*/
void cl_der_FreeDerNodeTree(const CSSM_CLUTIL_CONTEXT *pContext,
                            DER_NODE_PTR DerNodePtr)
{
    sint16 i=0;

    if (!DerNodePtr)
    {
        return;
    }

    /* Free the Tag and Original tag */
    CSSM_FREE_WRAPPER(internal_free, (void*)DerNodePtr->Tag, NULL);
    CSSM_FREE_WRAPPER(internal_free, (void*)DerNodePtr->OriginalTag, NULL);

    /* Free the data in the Children array */
    for (i=0; i<DerNodePtr->Count; i++)
    {
        if (DerNodePtr->Child[i].IsLeaf)
        {
            CSSM_FREE_WRAPPER(internal_free, (void*)DerNodePtr->Child[i].X.Input.Tag, NULL);
            CSSM_FREE_WRAPPER(internal_free, (void*)DerNodePtr->Child[i].X.Input.Content, NULL);
        }
        else 
        {
            cl_der_FreeDerNodeTree(pContext,DerNodePtr->Child[i].X.Node);
        }
    }
    CSSM_FREE_WRAPPER(internal_free, DerNodePtr->Child, NULL);

    /* Free the Type information */
    if (DerNodePtr->Type)
    {
        cl_der_InternalDeleteType(pContext,CSSM_FALSE, 0, DerNodePtr->Type);
    }

    CSSM_FREE_WRAPPER(internal_free, DerNodePtr, NULL);

    return;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_DeleteAllParentNodes
 *
 * Description:
 * This functions frees all of the memory associated with the linked list
 * of parent nodes.
 *
 * Parameters: 
 *
 * Return value:
 *
 * Error Codes:
 *---------------------------------------------------------------------------*/
void cl_der_DeleteAllParentNodes(CSSM_CLUTIL_CONTEXT *pContext)
{
    if (pContext->ParentCurrentPtr == NULL)
    {
        return;
    }

    /* Delete all the nodes */
    pContext->ParentCurrentPtr->PreviousNode->NextNode = NULL;
    while (pContext->ParentCurrentPtr != NULL)
    {
        cl_der_FreeDerNodeTree(pContext,pContext->ParentCurrentPtr->DerNodePtr);

        if (pContext->ParentCurrentPtr->NextNode == NULL) /* If this is the last node */
        {
            CSSM_FREE_WRAPPER(internal_free, pContext->ParentCurrentPtr, NULL);
            pContext->ParentCurrentPtr = NULL;
        }
        else 
        {
            pContext->ParentCurrentPtr = pContext->ParentCurrentPtr->NextNode;
            CSSM_FREE_WRAPPER(internal_free, pContext->ParentCurrentPtr->PreviousNode, NULL);
        }
    }

    return;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_DeleteAllNodes
 *
 * Description:
 * This functions frees all of the memory associated with the linked list
 * of parent nodes and all  of the memory associated with the linked list
 * of leaf nodes.
 *
 * Parameters: 
 *
 * Return value:
 *
 * Error Codes:
 *---------------------------------------------------------------------------*/
void cl_der_DeleteAllNodes(CSSM_CLUTIL_CONTEXT *pContext)
{
    cl_der_DeleteAllLeafNodes(pContext);
    cl_der_DeleteAllParentNodes(pContext);
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_DeleteLeafNode
 *
 * Description: 
 * This functions frees all of the memory associated with the input leaf node.
 *
 * Parameters: 
 * Handle (input) : The handle of the node to be freed 
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_DeleteLeafNode(CSSM_CLUTIL_CONTEXT *pContext,
                                  CL_DER_NODE_HANDLE Handle)
{
    CL_DER_LL_LEAF_NODE_PTR leaf_node_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;

    ret = cl_der_FindLeafNode(pContext, Handle, &leaf_node_ptr);

    /* If the leaf was not found, return */
    if (CSSM_OK != ret)
    {
        return ret;
        
    }

    /* Remove the Node from the list */
    /* If it is the only node, set the pointers to NULL */
    if (leaf_node_ptr->NextNode == leaf_node_ptr)
    {
        pContext->LeafCurrentPtr = NULL;
    }
    else 
    {   /* Otherwise, fix the list pointers */
        leaf_node_ptr->PreviousNode->NextNode = leaf_node_ptr->NextNode;
        leaf_node_ptr->NextNode->PreviousNode = leaf_node_ptr->PreviousNode;
        pContext->LeafCurrentPtr = leaf_node_ptr->NextNode;
    }

    /* Delete the node */
    CSSM_FREE_WRAPPER(internal_free, (void*)leaf_node_ptr->BerItemPtr->Tag, NULL);
    CSSM_FREE_WRAPPER(internal_free, (void*)leaf_node_ptr->BerItemPtr->Content, NULL);
    CSSM_FREE_WRAPPER(internal_free, (void*)leaf_node_ptr->BerItemPtr, NULL);
    CSSM_FREE_WRAPPER(internal_free, (void*)leaf_node_ptr, NULL);

    return CSSM_OK;;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_DeleteParentNode
 *
 * Description:
 * Remove the linked list node and free its contents
 *
 * Parameters: 
 * Handle (input) : The handle of the node to be deleted
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_DeleteParentNode(CSSM_CLUTIL_CONTEXT *pContext,
                                    CL_DER_NODE_HANDLE Handle)
{
    CL_DER_LL_PARENT_NODE_PTR parent_node_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;

    if (Handle == 0 || pContext->ParentCurrentPtr == NULL)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

    ret = cl_der_FindParentNode(pContext, Handle, &parent_node_ptr);

    /* If the node was not found, return */
    if (CSSM_OK != ret)
    {
        return CSSMERR_CL_INVALID_DATA;
    }

    /* Remove the Node from the list */
    /* If it is the only node, set the pointers to NULL */
    if (parent_node_ptr->NextNode == parent_node_ptr)
    {
        pContext->ParentCurrentPtr = NULL;
    }
    else 
    {   /* Otherwise, fix the list pointers */
        parent_node_ptr->PreviousNode->NextNode = parent_node_ptr->NextNode;
        parent_node_ptr->NextNode->PreviousNode = parent_node_ptr->PreviousNode;
        pContext->ParentCurrentPtr = parent_node_ptr->NextNode;
    }

    /* Delete the node */
    cl_der_FreeDerNodeTree(pContext,parent_node_ptr->DerNodePtr);
    CSSM_FREE_WRAPPER(internal_free, parent_node_ptr, NULL);

    return CSSM_OK;
}

