/*-----------------------------------------------------------------------
 * File: CRLLIST.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 

/*  This file defines the routines for managing the CRL revoked certificates
    list.
*/


#include <assert.h>       /* For assertions */
#include "cssm.h"         /* CSSM Include */
#include "guids.h"        /* GUIDs */
#include "x509defs.h"     /* List Data Structures */
#include "clutil.h"
#include "x_qtable.h"      /* Query Table Functions */
#include "crllist.h"      /* Traversal-specific structures/functions */
#include "internal.h"
#include <cssmMemTrack.h>

/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_RevokedListCreate
*
* Description:
* This function creates a new, empty revoked certificates list.
*
* Parameters:
*
* Return value:
* X509_REVOKED_CERT_LIST_PTR: The newly created list.
*
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_RevokedListCreate(const CSSM_CLUTIL_CONTEXT *pContext,
                                               CSSM_HANDLE AppHandle,
                                               X509_REVOKED_CERT_LIST_PTR *results_val)
{
    X509_REVOKED_CERT_LIST_PTR return_val;

    if (!results_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    /* Allocate Memory */
    return_val = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(X509_REVOKED_CERT_LIST));
    if (!return_val)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Initialize list */
   
   
    return_val->head = NULL;
    return_val->tail = NULL;
    return_val->numNodes = 0;
    *results_val = return_val;

    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_RevokedListNumNodes
*
* Description:
* This function returns the number of nodes in the the revoked cert list.
* It simply accesses the numNodes field of the list, and does NO updating
* of that field - that is left to the programmer in the insert and delete
* node functions.
*
* Parameters: 
* RCList:   The list to be queried.
*
* Return value:
* uint32:  The number of nodes in the list.  If an error occurs, 0 is returned and
*           and error code is set.
*
*---------------------------------------------------------------------------*/
uint32 CSSM_CLUTIL_X509_RevokedListNumNodes(const CSSM_CLUTIL_CONTEXT *pContext,
                                            X509_REVOKED_CERT_LIST_PTR RCList)
{
    if (RCList)
    {
        return RCList->numNodes;
    }
    else
    {
        return 0;
    }
}


/*-----------------------------------------------------------------------------
* Name:  CSSM_CLUTIL_X509_RevokedListTraverseInitialize
*
* Description:
* Initializes a query table to prepare for calls to CSSM_CLUTIL_X509_RevokedListTraverseNext,
* to traverse the CRL revoked certificates list. 
* 
* This function's implementation is dependent on the type of list being used.  
* The present implementation is for a singly-linked unordered list.
*
* Note:  Since the list may be large, it is unreasonable to copy the entire list
*        for traversal.  Therefore, the user must not add or remove nodes from the
*        list between calls to Initialize and Abort the Traversal, or the traversal
*        may become corrupted.  Currently, there are no safeguards to prevent the
*        user from making such changes during a traversal.
*
* Parameters: 
* RCList:           The revoked certificates list that will be traversed.
* result_val(output)The handle to be used for further traversal operations.  0 if 
*                   an error occurred.
*
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_RevokedListTraverseInitialize(CSSM_CLUTIL_CONTEXT *pContext,
                                                           X509_REVOKED_CERT_LIST_PTR RCList, 
                                                           CSSM_HANDLE *result_val)
{
    REVOKED_TRAVERSE_INFO traverse_info;
    CSSM_DATA             query_entry;
    CSSM_HANDLE           return_val = 0;
    CSSM_RETURN           ret = CSSM_OK;

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = 0;
    /* Get initial traversal information */
    traverse_info.nextNode = RCList->head;

    /* Save initialization info in query table, get results_handle */
    query_entry.Length = sizeof(REVOKED_TRAVERSE_INFO);
    query_entry.Data = CSSM_MALLOC_WRAPPER(internal_malloc, query_entry.Length, NULL);
    if (!query_entry.Data)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    port_memcpy(query_entry.Data, &traverse_info, query_entry.Length);

    ret = cl_QT_AddResults(pContext,&query_entry, 1, &return_val);

    CSSM_FREE_WRAPPER(internal_free, query_entry.Data, NULL);

    *result_val = return_val;
    return ret;
}



/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_RevokedListTraverseNext 
*
* Description:
* This function traverses to the next entry in the revoked certificates list,
* (corresponding to the results_handle, which is retrieved by the user via a call to 
* CSSM_CLUTIL_X509_RevokedListIntialize).
* Successive calls to this function will return successive entries in the list.
* The entries are returned in the same order in which they were added by the
* CSSM_CLUTIL_X509_InsertRevokedListEntry function.  This requirement may be inefficient
* with some implementations, but is necessary due to the fact that the ASN.1
* encoding of a revoked certs list must be ordered.  This preserves that order
* between encoding and decoding.
* 
* This function's implementation is dependent on the type of list being used.  
* The present implementation is for a singly-linked unordered list.
*
* Note:  Nodes should not be added or removed from the list during traversal,
*        as this may result in corruption of the traversal.  Currently, there
*        are no safeguards to prevent the user from doing this.
*
* Parameters: 
* RCList:       The revoked certificates list being traversed.
* results_handle: The traversal handle, as received via a call to 
*                 CSSM_CLUTIL_X509_RevokedListIntialize.
*
* result_val (Output):
*                     The next entry in the list traversal (note- not
*                     the next node, but the entry that node contains.
*                     NULL is returned and an error code set if an error
*                     occurs.  NULL is returned and error code cleared if
*                     there are no more entries in the list (i.e., at
*                     end of list).
*
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_RevokedListTraverseNext(CSSM_CLUTIL_CONTEXT *pContext,
                                                     CSSM_HANDLE AppHandle, 
                                                     X509_REVOKED_CERT_LIST_PTR RCList,
                                                     CSSM_HANDLE results_handle,
                                                     CSSM_X509_REVOKED_CERT_ENTRY_PTR *result_val)
{
    CSSM_DATA_PTR               query_entry;
    REVOKED_TRAVERSE_INFO_PTR   traverse_info;
    CSSM_X509_REVOKED_CERT_ENTRY_PTR return_val;
    CSSM_RETURN ret = CSSM_OK;
    
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *result_val = NULL;
    /* Get the traversal data from the query table */
    ret = cl_QT_GetNextResult(pContext,AppHandle, results_handle, &query_entry);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    if (query_entry->Length != sizeof(REVOKED_TRAVERSE_INFO_PTR))
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, query_entry->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, query_entry);
        return CSSMERR_CL_INVALID_RESULTS_HANDLE;
        
    }
    /* use traverse_info to make access to query_entry cleaner */
    traverse_info = (REVOKED_TRAVERSE_INFO_PTR)query_entry->Data;
    
    /* Get next node from traverse_info */
    if (!traverse_info->nextNode)
    {
        /* At end of list */
        /* Prepare for next call to this function, clean up, return false */
        cl_QT_ReplaceResults(pContext,results_handle, query_entry,1);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, query_entry->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, query_entry);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, traverse_info);
        return CSSM_OK;
    }
    return_val = CSSM_CLUTIL_X509_RevokedEntryCopy(pContext,AppHandle, traverse_info->nextNode->revokedCertEntry);
    if (!return_val)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Prepare for next call to this function */
    /* Can reuse traverse_info and query_entry */
    assert(traverse_info == (REVOKED_TRAVERSE_INFO_PTR)query_entry->Data);
    traverse_info->nextNode = traverse_info->nextNode->nextNode;
    ret = cl_QT_ReplaceResults(pContext,results_handle, query_entry, 1);
    if (CSSM_OK != ret)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,query_entry->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,query_entry);
        return ret;
    }

    /* Clean up */
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,traverse_info);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,query_entry);

    *result_val = return_val;
    return ret;
}



/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_RevokedListTraverseAbort
*
* Description:
* This function removes the traversal data related to the results_handle (obtained
* by the user via a call to CSSM_CLUTIL_X509_RevokedListInitialize) from the query list.
*
* Parameters: 
* results_handle: The handle for which related traversal data is to be removed.
*
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_RevokedListTraverseAbort(CSSM_CLUTIL_CONTEXT *pContext,
                                                    CSSM_HANDLE results_handle)
{
    
    /* Remove the results_handle info */
    return cl_QT_DeleteResults(pContext,results_handle);
    
}


/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_GetLastRevokedEntryPtr
*
* Description:
* This function returns a pointer to the last entry that was inserted in to the input list.
* It is meant to be used internally only, as it (purposefully) allows the entry to
* be modified in place.  Note that this trusts the programmer not to modify the serialnumber
* of the entry, since the entry is keyed by serial number in the list.
*
* This function's implementation is dependent on the type of list being used.  
* The present implementation is for a singly-linked unordered list.
*
* Parameters: 
* RCList: The list to look at.
*
* Return value:
* CSSM_X509_REVOKED_CERT_ENTRY_PTR: Pointer to the last entry that was inserted in the list.
*           Note that this is not a pointer to a copy of the entry, but to the actual
*           entry itself.  NULL is returned if there is an error or if the list is empty.
*
*---------------------------------------------------------------------------*/
CSSM_X509_REVOKED_CERT_ENTRY_PTR CSSM_CLUTIL_X509_GetLastRevokedEntryPtr(X509_REVOKED_CERT_LIST_PTR RCList)
{
    if (!RCList || !RCList->tail)
    {
        return NULL;
    }

    return RCList->tail->revokedCertEntry;
}
    




/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_FindRevokedCertEntry
*
* Description:
* This function will search the revoked certificate list for an entry whose
* serial number matches that specified in the input.
*
* This function's implementation is dependent on the type of list being used.  
* The present implementation is for a singly-linked unordered list.
*
* Parameters: 
* RCList:   The revoked certificates list to be searched.
* serial_num:   The serial number to search for.
*
* result_val (output):If the search was successful, a pointer to a copy
*                     of the found entry.  NULL is returned otherwise,
*                     with an error code set if there was an error, or
*                     no error code set if no error occurred (i.e., the
*                     requested entry was not found).
*
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_FindRevokedCertEntry(const CSSM_CLUTIL_CONTEXT *pContext,
                                                  CSSM_HANDLE AppHandle, 
                                                  X509_REVOKED_CERT_LIST_PTR RCList,
                                                  CSSM_DATA_PTR serial_num,
                                                  CSSM_X509_REVOKED_CERT_ENTRY_PTR *result_val)
{
    CSSM_X509_REVOKED_CERT_ENTRY_PTR return_val = NULL;
    X509_REVOKED_LIST_NODE_PTR  node_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    
    *result_val = NULL;
    /* Search for the serial_num - for this implementation, it is just a sequential search */
    node_ptr = RCList->head;
    while (node_ptr && 
           (serial_num->Length != node_ptr->revokedCertEntry->certificateSerialNumber.Length ||
            port_memcmp(serial_num->Data,
                        node_ptr->revokedCertEntry->certificateSerialNumber.Data,
                        serial_num->Length)))
    {
        node_ptr = node_ptr->nextNode;
    }

    /* node_ptr is now pointing to the found node, or is NULL if a match wasn't made */
    /* Allocate the return value, and copy the found entry into it (if one exists) */
    if (node_ptr)
    {
        return_val = CSSM_CLUTIL_X509_RevokedEntryCopy(pContext,AppHandle, node_ptr->revokedCertEntry);
        if (!return_val)
        {
            return CSSMERR_CL_MEMORY_ERROR;
            
        }

    }
    else
    {
        return_val = NULL;
    }

    

    *result_val = return_val;
    return ret;
}



/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_InsertRevokedCertEntry
*
* Description:
* Inserts the input RCEntry into the RCList, keyed on the certificateSerialNumber
* field of the RCEntry.  This function must be designed in conjunction with the
* traversal functions to allow traversal of the list in the same order that list
* elements are inserted.  This requirement is a result of the CRL revoked list
* entries being encoded as an ASN.1 sequence, which has order.
*
* This function's implementation is dependent on the type of list being used.  
* The present implementation is for a singly-linked unordered list.
*
* Parameters: 
* RCList (input/output):  The list in which the entry is to be inserted.  The list passed
*                   is directly manipulated by this function.
* RCEntry (input):  The entry to be inserted.  It must at least have the
*                   certificateSerialNumber field filled with valid data, and may not
*                   have the same serial number as any other entry already in the list.
*
*
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_InsertRevokedCertEntry(const CSSM_CLUTIL_CONTEXT *pContext,
                                                       CSSM_HANDLE AppHandle, 
                                                  X509_REVOKED_CERT_LIST *RCList,
                                                  CSSM_X509_REVOKED_CERT_ENTRY_PTR RCEntry)
{
    X509_REVOKED_LIST_NODE_PTR  new_node = NULL;
    
    CSSM_X509_REVOKED_CERT_ENTRY_PTR find_rtn = NULL;
    CSSM_RETURN ret = CSSM_OK;
    

    /* Error check inputs */
    /* Note that RCEntry must have a serial number, and may not be a duplicate number */
    ret = CSSM_CLUTIL_X509_FindRevokedCertEntry(pContext,AppHandle, RCList, &RCEntry->certificateSerialNumber,&find_rtn);
    
    if (find_rtn || (!find_rtn && ret != CSSM_OK))
    {
        if (find_rtn)
        {
            CSSM_CLUTIL_X509_FreeRevokedCertEntry(pContext,AppHandle, find_rtn);
        }
        
        if (CSSM_OK != ret)
        {
            return ret;
        }
        else
        {
            return CSSMERR_CL_INVALID_DATA;
        }
        
    }
    
    /* Allocate a new node */
    new_node = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(X509_REVOKED_LIST_NODE));
    if (!new_node)
    {
        
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    
    /* Copy data to new node */
    new_node->revokedCertEntry = CSSM_CLUTIL_X509_RevokedEntryCopy(pContext,AppHandle, RCEntry);
    if (!new_node->revokedCertEntry)
    {
        
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Add new node to end of list */
    if (!RCList->numNodes)
    {
        /* First node in list */
        RCList->head = new_node;
        RCList->tail = new_node;
        new_node->nextNode = NULL;
    }
    else
    {
        RCList->tail->nextNode = new_node;
        RCList->tail = new_node;
        new_node->nextNode = NULL;
    }
    (RCList->numNodes)++;
    
    
    return ret;
}



/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_FreeRevokedCertList
*
* Description:
* This function frees all memory from a revoked certificate list.
*
* This function's implementation is dependent on the type of list being used.  
* The present implementation is for a singly-linked unordered list.
*
* Parameters: 
* RCList:   Pointer to the list to be freed.  All memory associated with this
*           pointer will be freed.
*
* Return value:
* CSSM_BOOL:    CSSM_TRUE if the operation is successful, CSSM_FALSE otherwise.
*
*---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_X509_FreeRevokedCertList(const CSSM_CLUTIL_CONTEXT *pContext,
                                               CSSM_HANDLE AppHandle, 
                                               X509_REVOKED_CERT_LIST_PTR RCList)
{
    X509_REVOKED_LIST_NODE_PTR  current_node = NULL,
                                next_node = NULL;
    /* Check Inputs */
    if (!RCList)
    {
        return CSSM_TRUE;  /* Nothing needs to be freed */
    }
    /* Traverse list, freeing each node */
    next_node = RCList->head;
    while (next_node)
    {
        current_node = next_node;
        next_node = current_node->nextNode;
        /* Free entry information */
        if (CSSM_FALSE ==
            CSSM_CLUTIL_X509_FreeRevokedCertEntry(pContext,AppHandle, current_node->revokedCertEntry))
        {
            return CSSM_FALSE;
        }
        /* Free node */
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, current_node);
    }

    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, RCList);

    return CSSM_TRUE;
}



/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_RemoveRevokedCertEntry
*
* Description:
* This function finds the list entry corresponding to the input serial_num,
* removes that entry from the list and frees the associated memory.
*
* This function's implementation is dependent on the type of list being used.  
* The present implementation is for a singly-linked unordered list.
*
* Parameters: 
* RCList (input/output):    The list to be queried.  It will be modified in place.
* serial_num (input):       The serial number of the list entry to be removed.
*
*
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_RemoveRevokedCertEntry(const CSSM_CLUTIL_CONTEXT *pContext,
                                                       CSSM_HANDLE AppHandle, 
                                                  X509_REVOKED_CERT_LIST *RCList,
                                                  CSSM_DATA_PTR serial_num)
{
    X509_REVOKED_LIST_NODE_PTR  next = NULL,
                                prev = NULL;
   
    /* Search for the serial_num */
    next = RCList->head;
    prev = NULL;
    while (next &&
           (serial_num->Length != next->revokedCertEntry->certificateSerialNumber.Length ||
            port_memcmp(serial_num->Data,
                        next->revokedCertEntry->certificateSerialNumber.Data,
                        serial_num->Length)))
    {
        prev = next;
        next = next->nextNode;
    }

    if (!next)
    {
        /* A match wasn't found */
        return CSSMERR_CL_INVALID_DATA;
    }

    /* if a match was found, free the node */
    /* reassign list pointers */
    if (next == RCList->tail)
    {
        RCList->tail = prev;
    }
    if (next == RCList->head)
    {
        RCList->head = next->nextNode;
    }
    if (prev)
    {
        prev->nextNode = next->nextNode;
    }

    /* Took out one node */
    (RCList->numNodes)--;

    /* Free entry information */
    if (CSSM_FALSE == CSSM_CLUTIL_X509_FreeRevokedCertEntry(pContext,AppHandle, next->revokedCertEntry))
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    /* Free node */
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, next);

    return CSSM_OK;
}
    
