/*-----------------------------------------------------------------------
 * File: IS_BAD.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/* 
 * This file performs error checking - 
 * it looks for bad pointers in input data structures
*/
#include "cssmtype.h"
#include "x509defs.h"
#include "clutil.h"
#include "internal.h"
#include "oidsx509.h" /* CSSMOID_* definitions */
#include <cssmMemTrack.h>

#ifndef VMS
#pragma warning(disable:4057)
#endif
/* To remove unsigned char/char warnings in str functions related to CSSM_X509_TIME */

/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_IsBadReadPtr
 *
 * Description:
 * This function checks whether the input length number of bytes 
 * can be read from the input pointer.  This function is needed because 
 * port_IsBadReadPtr is sensitive to NULLs.
 *
 * Parameters: 
 * ptr (input)    : The pointer to be checked
 * length (input) : The number of bytes that we'd like to read from the input ptr
 *
 * Return value:
 * A TRUE/FALSE indicator of whether length bytes can be read from ptr
 * 
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_X509_IsBadReadPtr(const void *ptr, uint32 length)
{
    if ((length != 0) && (ptr == NULL))
    {
        return CSSM_TRUE;
    }
    if (length == 0)
    {
        return CSSM_FALSE;
    }
    return port_IsBadReadPtr(ptr, length);
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_IsBadWritePtr
 *
 * Description:
 * This function checks whether the input length number of bytes 
 * can be written to the input pointer.  This function is needed because 
 * port_IsBadWritePtr is sensitive to NULLs.
 *
 * Parameters: 
 * ptr (input)    : The pointer to be checked
 * length (input) : The number of bytes that we'd like to write into the input ptr
 *
 * Return value:
 * A TRUE/FALSE indicator of whether length bytes can be written to ptr
 * 
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_X509_IsBadWritePtr(void *ptr, uint32 length)
{
    if ((length != 0) && (ptr == NULL))
    {
        return CSSM_TRUE;
    }
    if (length == 0)
    {
        return CSSM_FALSE;
    }
    return port_IsBadWritePtr(ptr, length);
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_IsBadCssmDataPtr
 *
 * Description:
 * This function checks whether there are any bad pointers in the input CSSM_DATA struct
 *
 * Parameters: 
 * CssmData (input) : The CssmData to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether CssmData contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_IsBadCssmDataPtr(const CSSM_DATA *CssmData)
{
    if (CssmData == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(CssmData, sizeof(CSSM_DATA))) 
    {
        return CSSM_TRUE;
    }
    if (port_IsBadReadPtr(CssmData->Data, CssmData->Length)) 
    {
        return CSSM_TRUE;
    }
    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_IsBadAlgIdPtr
 *
 * Description:
 * This function checks whether there are any bad pointers in the input algid struct
 *
 * Parameters: 
 * AlgId (input) : The AlgId to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether AlgId contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_X509_IsBadAlgIdPtr(const CSSM_X509_ALGORITHM_IDENTIFIER *AlgId)
{
    if (AlgId == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(AlgId, sizeof(CSSM_X509_ALGORITHM_IDENTIFIER))) 
    {
        return CSSM_TRUE;
    }
    if ((AlgId->algorithm.Data != NULL) &&
        CSSM_CLUTIL_X509_IsBadReadPtr(AlgId->algorithm.Data, AlgId->algorithm.Length))
    {
        return CSSM_TRUE;
    }
    if ((AlgId->parameters.Data != NULL) && 
        CSSM_CLUTIL_X509_IsBadReadPtr(AlgId->parameters.Data, AlgId->parameters.Length)) 
    {
        return CSSM_TRUE;
    }
    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_IsBadNamePtr
 *
 * Description:
 * This function checks whether there are any bad pointers in the input name struct
 *
 * Parameters: 
 * Name (input) : The Name to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether Name contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_X509_IsBadNamePtr(const CSSM_X509_NAME *Name)
{
    CSSM_X509_TYPE_VALUE_PAIR_PTR tv_array;
    uint32 i=0, j=0;

    if (Name == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(Name, sizeof(CSSM_X509_NAME))) 
    {
        return CSSM_TRUE;
    }
    if ((Name->RelativeDistinguishedName == NULL) && (Name->numberOfRDNs == 0)) 
    {
        return CSSM_FALSE;
    }
    if ((Name->RelativeDistinguishedName == NULL) || (Name->numberOfRDNs == 0)) 
    {
        return CSSM_TRUE;
    }
    if (port_IsBadReadPtr(Name->RelativeDistinguishedName, Name->numberOfRDNs * sizeof(CSSM_X509_RDN))) 
    {
        return CSSM_TRUE;
    }

    for (i=0; i < Name->numberOfRDNs; i++)
    {
#ifndef UNIX
    /*
     * This causes a different error to be returned on UNIX because malloc is does
     * not work the same as on Windows.
     */
        if ((Name->RelativeDistinguishedName[i].AttributeTypeAndValue == NULL)
            && (Name->RelativeDistinguishedName[i].numberOfPairs == 0)) 
        {
            continue;
        }
#endif
        if ((Name->RelativeDistinguishedName[i].AttributeTypeAndValue == NULL)
            || (Name->RelativeDistinguishedName[i].numberOfPairs == 0)) 
        {
            return CSSM_TRUE;
        }
        if (port_IsBadReadPtr(Name->RelativeDistinguishedName[i].AttributeTypeAndValue, 
                              Name->RelativeDistinguishedName[i].numberOfPairs * sizeof(CSSM_X509_TYPE_VALUE_PAIR))) 
        {
            return CSSM_TRUE;
        }
        tv_array = Name->RelativeDistinguishedName[i].AttributeTypeAndValue;
        for(j=0; j < Name->RelativeDistinguishedName[i].numberOfPairs; j++)
        {
            if ((tv_array[j].type.Data != NULL) 
                && port_IsBadReadPtr(tv_array[j].type.Data, tv_array[j].type.Length))
            {
                return CSSM_TRUE;
            }
            if ((tv_array[j].value.Data != NULL)
                && port_IsBadReadPtr(tv_array[j].value.Data, tv_array[j].value.Length))
            {
                return CSSM_TRUE;
            }
        }
    }
    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_IsBadSPKIPtr
 *
 * Description:
 * This function checks whether there are any bad pointers in the input SPKI struct
 *
 * Parameters: 
 * SPKI (input) : The SPKI to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether SPKI contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_X509_IsBadSpkiPtr(const CSSM_X509_SUBJECT_PUBLIC_KEY_INFO *SPKI)
{
    if (SPKI == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(SPKI, sizeof(CSSM_X509_SUBJECT_PUBLIC_KEY_INFO))) 
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadAlgIdPtr(&SPKI->algorithm))
    {
        return CSSM_TRUE;
    }
    if ((SPKI->subjectPublicKey.Data != NULL) 
        && port_IsBadReadPtr(SPKI->subjectPublicKey.Data, SPKI->subjectPublicKey.Length)) 
    {
        return CSSM_TRUE;
    }
    return CSSM_FALSE;
}



/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_IsBadExtensionPtr
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input extension struct array
 *
 * Parameters: 
 * ptr (input)            : The Extension array to be checked
 * num_extensions (input) : The number of extensions in the array
 *
 * Return value:
 * A TRUE/FALSE indicator of whether ptr contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
static CSSM_BOOL cssm_clutil_x509_IsBadExtensionValue(const CSSM_X509_EXTENSION *ptr)
{

    if (CSSM_CLUTIL_X509_IsEqualOid(&ptr->extnId, &CSSMOID_X509ExtBasicConstraints) == CSSM_TRUE) 
    {
        return CSSM_FALSE;
    }

    /* add new extension checking here */

    /* default */
    if (ptr->value.tagAndValue != NULL && (ptr->value.tagAndValue)->value.Data != NULL)
    {
        return port_IsBadReadPtr((ptr->value.tagAndValue)->value.Data, (ptr->value.tagAndValue)->value.Length);
    }
    return CSSM_FALSE;
}

CSSM_BOOL CSSM_CLUTIL_X509_IsBadExtensionPtr(const CSSM_X509_EXTENSION *ptr, uint32 num_extensions)
{
    uint32 i=0;

    if ((ptr == NULL) && (num_extensions == 0)) 
    {
        return CSSM_FALSE;
    }
    if ((ptr == NULL) || (num_extensions == 0))
    {
        return CSSM_TRUE;
    }
    if (port_IsBadReadPtr(ptr, num_extensions*sizeof(CSSM_X509_EXTENSION))) 
    {
        return CSSM_TRUE;
    }

    for (i=0; i<num_extensions; i++)
    {
        if (CSSM_CLUTIL_X509_IsBadReadPtr(ptr[i].extnId.Data, ptr[i].extnId.Length))
        {
            return CSSM_TRUE;
        }
        if (cssm_clutil_x509_IsBadExtensionValue(&ptr[i]))
        {
            return CSSM_TRUE;
        }
    }

    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_IsBadSignaturePtr
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input signature struct 
 *
 * Parameters: 
 * ptr (input) : The signature struct to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether ptr contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_X509_IsBadSignaturePtr(const CSSM_X509_SIGNATURE *ptr)
{
    if (ptr == NULL)  
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(ptr, sizeof(CSSM_X509_SIGNATURE))) 
    {
        return CSSM_TRUE;
    }
    if ((ptr->encrypted.Data != NULL) 
        && port_IsBadReadPtr(ptr->encrypted.Data, ptr->encrypted.Length))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadAlgIdPtr(&ptr->algorithmIdentifier))
    {
        return CSSM_TRUE;
    }
    return CSSM_FALSE;
}

/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_IsBadTbsCertPtr
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input ToBeSigned certificate struct 
 *
 * Parameters: 
 * TbsCert (input) : The ToBeSigned certificate struct to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether TbsCert contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_X509_IsBadTbsCertPtr(const CSSM_X509_TBS_CERTIFICATE *TbsCert)
{
    if (TbsCert == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(TbsCert, sizeof(CSSM_X509_TBS_CERTIFICATE)))
    {
        return CSSM_TRUE;
    }
    if ((TbsCert->version.Data != NULL) 
        && port_IsBadReadPtr(TbsCert->version.Data, TbsCert->version.Length))
    {
        return CSSM_TRUE;
    }
    if ((TbsCert->serialNumber.Data != NULL) 
        && port_IsBadReadPtr(TbsCert->serialNumber.Data, TbsCert->serialNumber.Length))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadAlgIdPtr(&TbsCert->signature))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadNamePtr(&TbsCert->issuer))
    {
        return CSSM_TRUE;
    }
    if ((TbsCert->validity.notBefore.time.Data != NULL) 
        && port_IsBadStrPtr((const char *)TbsCert->validity.notBefore.time.Data,           /* ifdef VMS, apply cast */ 
                             strlen((const char *)TbsCert->validity.notBefore.time.Data))) /* ifdef VMS, apply cast */ 
    {
        return CSSM_TRUE;
    }
    if ((TbsCert->validity.notAfter.time.Data != NULL)
        && port_IsBadStrPtr((const char *)TbsCert->validity.notAfter.time.Data,            /* ifdef VMS, apply cast */
                            strlen((const char *)TbsCert->validity.notAfter.time.Data)))   /* ifdef VMS, apply cast */
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadNamePtr(&TbsCert->subject))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadAlgIdPtr(&TbsCert->subjectPublicKeyInfo.algorithm))
    {
        return CSSM_TRUE;
    }
    if ((TbsCert->subjectPublicKeyInfo.subjectPublicKey.Data != NULL)
        && port_IsBadReadPtr(TbsCert->subjectPublicKeyInfo.subjectPublicKey.Data, 
                           TbsCert->subjectPublicKeyInfo.subjectPublicKey.Length))
    {
        return CSSM_TRUE;
    }
    if ((TbsCert->issuerUniqueIdentifier.Data != NULL) 
        && port_IsBadReadPtr(TbsCert->issuerUniqueIdentifier.Data, 
                TbsCert->issuerUniqueIdentifier.Length))
    {
        return CSSM_TRUE;
    }
    if ((TbsCert->subjectUniqueIdentifier.Data != NULL)
        && port_IsBadReadPtr(TbsCert->subjectUniqueIdentifier.Data, 
                TbsCert->subjectUniqueIdentifier.Length) )
    {
         return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadExtensionPtr(TbsCert->extensions.extensions, 
            TbsCert->extensions.numberOfExtensions))
    {
        return CSSM_TRUE;
    }
    return CSSM_FALSE;
}

/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_IsBadCertPtr
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input certificate struct 
 *
 * Parameters: 
 * Cert (input) : The certificate struct to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether Cert contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_X509_IsBadCertPtr(const CSSM_X509_SIGNED_CERTIFICATE *Cert)
{
    if (Cert == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(Cert, sizeof(CSSM_X509_SIGNED_CERTIFICATE)))
    {
         return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadTbsCertPtr(&Cert->certificate))
    { 
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadSignaturePtr(&Cert->signature))
    {
        return CSSM_TRUE;
    }
    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: cssm_clutil_IsBadCssmDataArray
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input array of CSSM_DATA structs 
 *
 * Parameters: 
 * DataArray (input) : The array of CSSM_DATA structs to be checked
 * NumberOfElements (input) : The number of elements in the array
 *
 * Return value:
 * A TRUE/FALSE indicator of whether DataArray contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL cssm_clutil_IsBadCssmDataArray(CSSM_DATA_PTR DataArray, uint32 NumberOfElements)
{
    uint32 i=0;

    if ((DataArray == NULL) && (NumberOfElements == 0)) 
    {
        return CSSM_FALSE;
    }
    if ((DataArray == NULL) || (NumberOfElements == 0)) 
    {
        return CSSM_TRUE;
    }
    if (port_IsBadReadPtr(DataArray, NumberOfElements * sizeof(CSSM_DATA))) 
    {
        return CSSM_TRUE;
    }
    for (i=0; i < NumberOfElements; i++)
    {
        if (DataArray[i].Data == NULL)
        {
            continue;
        }
        if (port_IsBadReadPtr(DataArray[i].Data, DataArray[i].Length)) 
        {
            return CSSM_TRUE;
        }
    }
    return CSSM_FALSE;
}

/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_IsBadCertGroup
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input certificate group struct 
 *
 * Parameters: 
 * CertGroupPtr (input) : The certificate group struct to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether CertGroupPtr contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_IsBadCertGroup(const CSSM_CERTGROUP *CertGroupPtr)
{
    if (CertGroupPtr == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(CertGroupPtr, sizeof(CSSM_CERTGROUP))) 
    {
        return CSSM_TRUE;
    }
    return cssm_clutil_IsBadCssmDataArray(CertGroupPtr->GroupList.CertList, CertGroupPtr->NumCerts);
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_IsBadCrlRevokedCertEntryPtr
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input revoked cert entry struct 
 *
 * Parameters: 
 * entry (input) : The revoked cert entrystruct to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether entry contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_X509_IsBadCrlRevokedCertEntryPtr(const CSSM_X509_REVOKED_CERT_ENTRY *entry)
{
    if (entry == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(entry, sizeof(CSSM_X509_REVOKED_CERT_ENTRY)))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadExtensionPtr(entry->extensions.extensions, 
            entry->extensions.numberOfExtensions))
    {
        return CSSM_TRUE;
    }
    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_IsBadRevokedCertList
*
* Description:
* This function checks a revoked certificate list to see if it is valid.
*
* Parameters: 
* RCList:   The list to be checked.
*
* Return value:
* CSSM_BOOL:    CSSM_TRUE if the list is found to be invalid, CSSM_FALSE otherwise.
*
*---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_X509_IsBadRevokedCertList(const X509_REVOKED_CERT_LIST *RCList)
{
    if (RCList == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(RCList, sizeof(X509_REVOKED_CERT_LIST)))
    {
        return CSSM_TRUE;
    }
    if ((RCList->head == NULL) && (RCList->numNodes == 0)) 
    {
        return CSSM_FALSE;
    }
    if ((RCList->head == NULL) || (RCList->numNodes == 0))
    {
        return CSSM_TRUE;
    }
    if (RCList->tail == NULL)


    {
        return CSSM_TRUE;
    }
    /* Would take too long to check the existence of every node.  Just check the
       first and last nodes to make sure they are not corrupt */
    if (CSSM_CLUTIL_X509_IsBadReadPtr(RCList->head, sizeof(X509_REVOKED_LIST_NODE)))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadReadPtr(RCList->head->revokedCertEntry, sizeof(CSSM_X509_REVOKED_CERT_ENTRY)))
    {
        return CSSM_TRUE;
    }
    /* Check first node */
    if ((RCList->numNodes == 1) && (RCList->head->nextNode != NULL))
    {
        return CSSM_TRUE;
    }
    if ((RCList->numNodes == 1) && (RCList->head != RCList->tail))
    {
        return CSSM_TRUE;
    }
    if ((RCList->numNodes > 1) 
        && CSSM_CLUTIL_X509_IsBadReadPtr(RCList->head->nextNode, sizeof(X509_REVOKED_LIST_NODE)))
    {
        return CSSM_TRUE;
    }
    /* Check last node */
    if (CSSM_CLUTIL_X509_IsBadReadPtr(RCList->tail, sizeof(X509_REVOKED_LIST_NODE)))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadReadPtr(RCList->tail->revokedCertEntry, sizeof(CSSM_X509_REVOKED_CERT_ENTRY)))
    {
        return CSSM_TRUE;
    }
    if (RCList->tail->nextNode != NULL)
    {
        return CSSM_TRUE;
    }
    return CSSM_FALSE;
}




/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_IsBadTbsCertListPtr
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input tbsCertList struct 
 *
 * Parameters: 
 * tbsCertList (input) : The tbsCertList struct to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether tbsCertList contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_X509_IsBadTbsCertListPtr  (const X509_TBS_CERTLIST *tbsCertList)
{
    if (tbsCertList == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(tbsCertList, sizeof(X509_TBS_CERTLIST))) 
    {
        return CSSM_TRUE;
    }
    if ((tbsCertList->version.Data != NULL) 
        && port_IsBadReadPtr(tbsCertList->version.Data, tbsCertList->version.Length))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadAlgIdPtr(&tbsCertList->signature))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadNamePtr(&tbsCertList->issuer))
    {
        return CSSM_TRUE;
    }
    
    if ((tbsCertList->thisUpdate.time.Data != NULL) 
        && port_IsBadStrPtr((const char*)tbsCertList->thisUpdate.time.Data,           /* ifdef VMS, apply cast */
                            strlen((const char*)tbsCertList->thisUpdate.time.Data)))  /* ifdef VMS, apply cast */
    {
        return CSSM_TRUE;
    }
    if ((tbsCertList->nextUpdate.time.Data != NULL) 
        && port_IsBadStrPtr((const char*)tbsCertList->nextUpdate.time.Data,           /* ifdef VMS, apply cast */
                            strlen((const char*)tbsCertList->nextUpdate.time.Data)))  /* ifdef VMS, apply cast */
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadRevokedCertList(tbsCertList->revokedCertificates))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadExtensionPtr(tbsCertList->extensions.extensions, 
                                            tbsCertList->extensions.numberOfExtensions))
    {
        return CSSM_TRUE;
    }
    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_IsBadCrlPtr
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input revoked crl struct 
 *
 * Parameters: 
 * crl (input) : The crl struct to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether crl contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_X509_IsBadCrlPtr(const X509_SIGNED_CRL *crl)
{
    if (crl == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(crl, sizeof(X509_SIGNED_CRL)))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadTbsCertListPtr(&crl->tbsCertList))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadSignaturePtr(&crl->signature))
    {
        return CSSM_TRUE;
    }
    return CSSM_FALSE;
}
            


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_IsBadCrlGroup
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input crl group struct 
 *
 * Parameters: 
 * CrlGroupPtr (input) : The crl group struct to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether CrlGroupPtr contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_IsBadCrlGroup(const CSSM_CRLGROUP *CrlGroupPtr)
{
    if (CrlGroupPtr == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(CrlGroupPtr, sizeof(CSSM_CRLGROUP)))
    {
        return CSSM_TRUE;
    }
    return cssm_clutil_IsBadCssmDataArray(CrlGroupPtr->GroupList.CrlList, CrlGroupPtr->NumberOfCrls);
}



/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X501_IsBadAttributePtr
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input Attribute struct 
 *
 * Parameters: 
 * Attribute (input) : The X501 Attribute struct to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether Attribute contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_X501_IsBadAttributePtr(const CSSM_X501_ATTRIBUTE *Attribute)
{
    uint32 i=0;

    if (Attribute == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(Attribute, sizeof(CSSM_X501_ATTRIBUTE))) 
    {
        return CSSM_TRUE;
    }
    if ((Attribute->type.Data != NULL)
        && (port_IsBadReadPtr(Attribute->type.Data, Attribute->type.Length)))
    {
        return CSSM_TRUE;
    }
    if ((Attribute->values == NULL) && (Attribute->numberOfValues == 0)) 
    {
        return CSSM_FALSE;
    }
    if ((Attribute->values == NULL) || (Attribute->numberOfValues == 0)) 
    {
        return CSSM_TRUE;
    }
    if (port_IsBadReadPtr(Attribute->values, Attribute->numberOfValues * sizeof(CSSM_X501_ANY_VALUE))) 
    {
        return CSSM_TRUE;
    }
    for (i=0; i < Attribute->numberOfValues; i++)
    {
        if ((Attribute->values[i].value.Data != NULL)
            && port_IsBadReadPtr(Attribute->values[i].value.Data, Attribute->values[i].value.Length)) 
        {
            return CSSM_TRUE;
        }
    }
    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X501_IsBadAttributesPtr
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input Attributes struct 
 *
 * Parameters: 
 * Attributes (input) : The X501 Attributes struct to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether Attributes contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_X501_IsBadAttributesPtr (const CSSM_X501_ATTRIBUTES *Attributes)
{
    uint32 i=0;

    if (Attributes == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(Attributes, sizeof(CSSM_X501_ATTRIBUTES))) 
    {
        return CSSM_TRUE;
    }
    if ((Attributes->attributes == NULL) && (Attributes->numberOfAttributes == 0)) 
    {
        return CSSM_FALSE;
    }
    if ((Attributes->attributes == NULL) || (Attributes->numberOfAttributes == 0)) 
    {
        return CSSM_TRUE;
    }
    if (port_IsBadReadPtr(Attributes->attributes, 
                            Attributes->numberOfAttributes * sizeof(CSSM_X501_ATTRIBUTE))) 
    {
        return CSSM_TRUE;
    }
    for (i=0; i < Attributes->numberOfAttributes; i++)
    {
        if (CSSM_CLUTIL_X501_IsBadAttributePtr(&Attributes->attributes[i])) 
        {
            return CSSM_TRUE;
        }
    }
    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X501_IsBadAlgIdSet
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input set of AlgId structures 
 *
 * Parameters: 
 * AlgIds (input) : The set of AlgId structures to be checked
 * NumberOfAlgIds (input) : The number of elements in the set
 *
 * Return value:
 * A TRUE/FALSE indicator of whether AlgIds contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL cssm_clutil_x509_IsBadAlgIdSet(CSSM_X509_ALGORITHM_IDENTIFIER_PTR AlgIds, uint32 NumberOfAlgIds)
{
    uint32 i=0;

    if ((AlgIds == NULL) && (NumberOfAlgIds == 0)) 
    {
        return CSSM_FALSE;
    }
    if ((AlgIds == NULL) || (NumberOfAlgIds == 0)) 
    {
        return CSSM_TRUE;
    }
    if (port_IsBadReadPtr(AlgIds, NumberOfAlgIds * sizeof(CSSM_X509_ALGORITHM_IDENTIFIER))) 
    {
        return CSSM_TRUE;
    }
    for (i=0; i < NumberOfAlgIds; i++)
    {
        if (CSSM_CLUTIL_X509_IsBadAlgIdPtr(&AlgIds[i])) 
        {
            return CSSM_TRUE;
        }
    }
    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: cssm_clutil_pkcs7_IsBadSignerInfoPtr
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input SignerInfo structure
 *
 * Parameters: 
 * SignerInfoPtr (input) : The signer info struct to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether SignerInfoPtr contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL cssm_clutil_pkcs7_IsBadSignerInfoPtr(cssm_PKCS7_SIGNER_INFO_PTR SignerInfoPtr)
{
    if (SignerInfoPtr == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(SignerInfoPtr, sizeof(cssm_PKCS7_SIGNER_INFO))) 
    {
        return CSSM_TRUE;
    }
    if ((SignerInfoPtr->version.Data != NULL)
        && port_IsBadReadPtr(SignerInfoPtr->version.Data, SignerInfoPtr->version.Length))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadNamePtr(&SignerInfoPtr->issuerAndSerialNumber.issuer))
    {
        return CSSM_TRUE;
    }
    if ((SignerInfoPtr->issuerAndSerialNumber.serialNumber.Data != NULL)
        && port_IsBadReadPtr(SignerInfoPtr->issuerAndSerialNumber.serialNumber.Data, 
                                SignerInfoPtr->issuerAndSerialNumber.serialNumber.Length))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadAlgIdPtr(&SignerInfoPtr->digestAlgId))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X501_IsBadAttributesPtr(&SignerInfoPtr->authenticatedAttrs))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadAlgIdPtr(&SignerInfoPtr->digestEncryptionAlgId))
    {
        return CSSM_TRUE;
    }
    if ((SignerInfoPtr->encryptedDigest.Data != NULL) 
        && port_IsBadReadPtr(SignerInfoPtr->encryptedDigest.Data, 
                             SignerInfoPtr->encryptedDigest.Length))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X501_IsBadAttributesPtr(&SignerInfoPtr->unauthenticatedAttrs))
    {
        return CSSM_TRUE;
    }
    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: cssm_clutil_pkcs7_IsBadSignerInfoSet
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input set of SignerInfo structures 
 *
 * Parameters: 
 * SignerInfoArray (input)     : The set of SignerInfo structures to be checked
 * NumberOfSignerInfos (input) : The number of elements in the set
 *
 * Return value:
 * A TRUE/FALSE indicator of whether SignerInfoArray contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL cssm_clutil_pkcs7_IsBadSignerInfoSet(cssm_PKCS7_SIGNER_INFO_PTR SignerInfoArray, 
                                               uint32 NumberOfSignerInfos)
{
    uint32 i=0;

    if ((SignerInfoArray == NULL) && (NumberOfSignerInfos == 0)) 
    {
        return CSSM_FALSE;
    }
    if ((SignerInfoArray == NULL) || (NumberOfSignerInfos == 0))
    {
        return CSSM_TRUE;
    }
    if (port_IsBadReadPtr(SignerInfoArray, NumberOfSignerInfos * sizeof(cssm_PKCS7_SIGNER_INFO))) 
    {
        return CSSM_TRUE;
    }
    for (i=0; i < NumberOfSignerInfos; i++)
    {
        if (cssm_clutil_pkcs7_IsBadSignerInfoPtr(&SignerInfoArray[i])) 
        {
            return CSSM_TRUE;
        }
    }
    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_PKCS7_IsBadSignedDataPtr
 *
 * Description:
 * This function checks whether there are any bad pointers in 
 * the input PKCS #7 signed data structure
 *
 * Parameters: 
 * Pkcs7SignedDataPtr (input) : The PKCS #7 SignedData struct to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether Pkcs7SignedDataPtr contains any bad pointers
 * 
 *
 * NOTE: A NULL input is not considered bad
 *---------------------------------------------------------------------------*/
CSSM_BOOL CSSM_CLUTIL_PKCS7_IsBadSignedDataPtr(const cssm_PKCS7_SIGNED_DATA *Pkcs7SignedDataPtr)
{
    if (Pkcs7SignedDataPtr == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(Pkcs7SignedDataPtr, sizeof(cssm_PKCS7_SIGNED_DATA)))
    {
        return CSSM_TRUE;
    }
    if ((Pkcs7SignedDataPtr->version.Data != NULL)
        && port_IsBadReadPtr(Pkcs7SignedDataPtr->version.Data, Pkcs7SignedDataPtr->version.Length))
    {
        return CSSM_TRUE;
    }
    if (cssm_clutil_x509_IsBadAlgIdSet(Pkcs7SignedDataPtr->digestAlgIds, 
                                       Pkcs7SignedDataPtr->numberOfDigestAlgIds))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_X509_IsBadAlgIdPtr(&Pkcs7SignedDataPtr->contentInfo))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_IsBadCertGroup(&Pkcs7SignedDataPtr->certificates))
    {
        return CSSM_TRUE;
    }
    if (CSSM_CLUTIL_IsBadCrlGroup(&Pkcs7SignedDataPtr->crls))
    {
        return CSSM_TRUE;
    }
    if (cssm_clutil_pkcs7_IsBadSignerInfoSet(Pkcs7SignedDataPtr->signerInfos, 
                                             Pkcs7SignedDataPtr->numberOfSignerInfos))
    {
        return CSSM_TRUE;
    }
    return CSSM_FALSE;
}




/*-----------------------------------------------------------------------------
 * 
 *             Error checking on Parse Tree Structures 
 *
 *---------------------------------------------------------------------------*/



/*-----------------------------------------------------------------------------
 * Name: x509_cl_IsBadAlgorithmParseTree
 *
 * Description:
 * This function checks whether this is a valid AlgId parse tree
 *
 * Parameters: 
 * ParentNode (input) : The top node in the parse tree to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether this is a valid AlgId parse tree
 * 
 *---------------------------------------------------------------------------*/
CSSM_BOOL x509_cl_IsBadAlgorithmParseTree(CL_DER_PARENT_NODE_PTR ParentNode)
{
    if (ParentNode == NULL) 
    {
        return CSSM_FALSE;
    }
    if ((ParentNode->NumberOfChildren == 1) 
         && (port_IsBadReadPtr(ParentNode->ChildrenArray, sizeof(CL_DER_NODE)))) 
    {
        return CSSM_TRUE;
    }
    if ((ParentNode->NumberOfChildren == 2)
        && (port_IsBadReadPtr(ParentNode->ChildrenArray, 2*sizeof(CL_DER_NODE))))
    {
        return CSSM_TRUE;
    }
    if ((ParentNode->NumberOfChildren != 1) && (ParentNode->NumberOfChildren != 2))
    {
        return CSSM_TRUE;
    }
    if (ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Tag != DER_TAG_OID)
    {
        return CSSM_TRUE;
    }
    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: x509_cl_IsBadNameParseTree
 *
 * Description:
 * This function checks whether this is a valid Name parse tree
 *
 * Parameters: 
 * ParentNode (input) : The top node in the parse tree to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether this is a valid Name parse tree
 * 
 *---------------------------------------------------------------------------*/
CSSM_BOOL x509_cl_IsBadNameParseTree(CL_DER_PARENT_NODE_PTR ParentNode)
{
    CL_DER_PARENT_NODE_PTR current_rdn = NULL,
                           current_type_value = NULL;
    uint32 i=0, j=0;

    /* Parent is a SEQUENCE */
    if (ParentNode == NULL)
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(ParentNode, sizeof(CL_DER_PARENT_NODE))) 
    {
        return CSSM_TRUE;
    }
    if (ParentNode->Tag != DER_TAG_SEQUENCE) 
    {
        return CSSM_TRUE;
    }
    
    if ((ParentNode->NumberOfChildren != 0) &&
        port_IsBadReadPtr(ParentNode->ChildrenArray, ParentNode->NumberOfChildren * sizeof(CL_DER_NODE)))
    {
        return CSSM_TRUE;
    }

    /* Each RDN is a SET */
    for (i=0; i < ParentNode->NumberOfChildren; i++)
    {
        if (ParentNode->ChildrenArray[i].IsLeaf) 
        {
            return CSSM_TRUE;
        }
        if (port_IsBadReadPtr(ParentNode->ChildrenArray[i].DerNodeType.ParentNode, 
            sizeof(CL_DER_PARENT_NODE))) 
        {
            return CSSM_TRUE;
        }
        current_rdn = ParentNode->ChildrenArray[i].DerNodeType.ParentNode;
        if (current_rdn->Tag != DER_TAG_SET) 
        {
            return CSSM_TRUE;
        }
        if (port_IsBadReadPtr(current_rdn->ChildrenArray, current_rdn->NumberOfChildren * sizeof(CL_DER_NODE))) 
        {
            return CSSM_TRUE;
        }

        /* Each type/value pair is a SEQ */
        for (j=0; j < current_rdn->NumberOfChildren; j++)
        {
            if (current_rdn->ChildrenArray[j].IsLeaf) 
            {
                return CSSM_TRUE;
            }
            if (port_IsBadReadPtr(current_rdn->ChildrenArray[j].DerNodeType.ParentNode, 
                                  sizeof(CL_DER_PARENT_NODE))) 
            {
                return CSSM_TRUE;
            }
            current_type_value = current_rdn->ChildrenArray[j].DerNodeType.ParentNode;
            if (current_type_value->Tag != DER_TAG_SEQUENCE) 
            {
                return CSSM_TRUE;
            }
            if (port_IsBadReadPtr(current_type_value->ChildrenArray, 
                                  current_type_value->NumberOfChildren * sizeof(CL_DER_NODE))) 
            {
                return CSSM_TRUE;
            }
            if (current_type_value->NumberOfChildren != 2) 
            {
                return CSSM_TRUE;
            }
            if (!current_type_value->ChildrenArray[0].IsLeaf)
            {
                return CSSM_TRUE;
            }
            if (port_IsBadReadPtr(current_type_value->ChildrenArray[0].DerNodeType.LeafNode, 
                                  sizeof(CL_DER_LEAF_NODE)))
            {
                return CSSM_TRUE;
            }
            if (current_type_value->ChildrenArray[0].DerNodeType.LeafNode->Tag != DER_TAG_OID)
            {
                return CSSM_TRUE;
            }
            if (port_IsBadReadPtr(current_type_value->ChildrenArray[0].DerNodeType.LeafNode->Data.Data, 
                                  current_type_value->ChildrenArray[0].DerNodeType.LeafNode->Data.Length))
            {
                return CSSM_TRUE;
            }
        }
    }
    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: x509_cl_IsBadTbsCertParseTree
 *
 * Description:
 * This function checks whether this is a valid TbsCert parse tree
 *
 * Parameters: 
 * ParentNode (input) : The top node in the parse tree to be checked
 * Existance (output) : An array indicating the presence or absence of
 *                      each field
 *
 * Return value:
 * A TRUE/FALSE indicator of whether this is a valid TbsCert parse tree
 * 
 *---------------------------------------------------------------------------*/
CSSM_BOOL x509_cl_IsBadTbsCertParseTree(CL_DER_PARENT_NODE_PTR ParentNode, CSSM_BOOL *Existence)
{
    CL_DER_PARENT_NODE_PTR  tree_validity = NULL,
                            tree_key_info = NULL;
    uint32                  possibly_missing[4] = {0, 7, 8, 9};

    if (ParentNode == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(ParentNode, sizeof(CL_DER_PARENT_NODE)))
    {
        return CSSM_TRUE;
    }
    /* Error checking on first level of the tree */
    if ((ParentNode->NumberOfChildren != 10)
        || port_IsBadReadPtr(ParentNode->ChildrenArray, 
                             ParentNode->NumberOfChildren*sizeof(CL_DER_NODE)) )
    {
        return CSSM_TRUE;
    }
    if (cssm_clutil_DetermineFieldExistence(ParentNode->NumberOfChildren, 
                                            ParentNode->ChildrenArray, 
                                            4, 
                                            possibly_missing, 
                                            Existence) != CSSM_OK)
    {
        return CSSM_TRUE;
    }
    if (ParentNode->ChildrenArray[1].DerNodeType.LeafNode->Tag   != DER_TAG_INTEGER  ||
        ParentNode->ChildrenArray[2].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE ||
        ParentNode->ChildrenArray[3].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE ||
        ParentNode->ChildrenArray[4].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE ||
        ParentNode->ChildrenArray[5].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE ||
        ParentNode->ChildrenArray[6].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE )
    {
        return CSSM_TRUE;
    }
    if ((Existence[0] && ParentNode->ChildrenArray[0].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE) ||
        (Existence[7] && ParentNode->ChildrenArray[7].DerNodeType.LeafNode->Tag != DER_TAG_BIT_STRING) ||
        (Existence[8] && ParentNode->ChildrenArray[8].DerNodeType.LeafNode->Tag != DER_TAG_BIT_STRING) ||
        (Existence[9] && ParentNode->ChildrenArray[9].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE) )
    {
        return CSSM_TRUE;
    }

    /* Error checking on the signature algorithm subtree */
    if (x509_cl_IsBadAlgorithmParseTree(ParentNode->ChildrenArray[2].DerNodeType.ParentNode))
    {
        return CSSM_TRUE;
    }

    /* Error checking on the issuer and subject subtrees */
    if (x509_cl_IsBadNameParseTree(ParentNode->ChildrenArray[3].DerNodeType.ParentNode) ||
        x509_cl_IsBadNameParseTree(ParentNode->ChildrenArray[5].DerNodeType.ParentNode))
    {
        return CSSM_TRUE;
    }

    /* Error checking on the validity subtree */
    tree_validity = ParentNode->ChildrenArray[4].DerNodeType.ParentNode;
    if ((tree_validity->NumberOfChildren != 2) 
        || port_IsBadReadPtr(tree_validity->ChildrenArray, 2*sizeof(CL_DER_NODE)) )
    {
        return CSSM_TRUE;
    }
    if (((tree_validity->ChildrenArray[0].DerNodeType.LeafNode->Tag != DER_TAG_UTC_TIME) &&
         (tree_validity->ChildrenArray[0].DerNodeType.LeafNode->Tag != DER_TAG_GENERALIZED_TIME))
        ||
        ((tree_validity->ChildrenArray[1].DerNodeType.LeafNode->Tag != DER_TAG_UTC_TIME) &&
         (tree_validity->ChildrenArray[1].DerNodeType.LeafNode->Tag != DER_TAG_GENERALIZED_TIME)))
    {
        return CSSM_TRUE;
    }

    /* Error checking on subjectPublicKeyInfo subtree */
    tree_key_info = ParentNode->ChildrenArray[6].DerNodeType.ParentNode;
    if ((tree_key_info->NumberOfChildren != 2) || 
        port_IsBadReadPtr(tree_key_info->ChildrenArray, 2*sizeof(CL_DER_NODE)) )
    {
        return CSSM_TRUE;
    }
    if (x509_cl_IsBadAlgorithmParseTree(tree_key_info->ChildrenArray[0].DerNodeType.ParentNode) ||
        (tree_key_info->ChildrenArray[1].DerNodeType.LeafNode->Tag != DER_TAG_BIT_STRING) )
    {
        return CSSM_TRUE;
    }

    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: x509_cl_IsBadCertificateParseTree
 *
 * Description:
 * This function checks whether this is a valid Cert parse tree
 *
 * Parameters: 
 * ParentNode (input) : The top node in the parse tree to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether this is a valid Cert parse tree
 * 
 *---------------------------------------------------------------------------*/
CSSM_BOOL x509_cl_IsBadCertificateParseTree(CL_DER_PARENT_NODE_PTR ParentNode)
{
    CSSM_BOOL existence[10];
    uint32    num_bytes = 0;

    if (ParentNode == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(ParentNode, sizeof(CL_DER_PARENT_NODE))) 
    {
        return CSSM_TRUE;
    }

    if ((ParentNode->NumberOfChildren != 3) || 
        port_IsBadReadPtr(ParentNode->ChildrenArray, ParentNode->NumberOfChildren*sizeof(CL_DER_NODE)) )
    {
        return CSSM_TRUE;
    }

    if ((ParentNode->ChildrenArray[0].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE) ||
        (ParentNode->ChildrenArray[1].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE) ||
        (ParentNode->ChildrenArray[2].DerNodeType.LeafNode->Tag != DER_TAG_BIT_STRING) )
    {
        return CSSM_TRUE;
    }

    if (x509_cl_IsBadTbsCertParseTree(ParentNode->ChildrenArray[0].DerNodeType.ParentNode, 
                                      existence)) 
    {
        return CSSM_TRUE;
    }
    if (x509_cl_IsBadAlgorithmParseTree(ParentNode->ChildrenArray[1].DerNodeType.ParentNode)) 
    {
        return CSSM_TRUE;
    }

    if ((ParentNode->ChildrenArray[2].DerNodeType.LeafNode->Data.Length % 8) == 0)
    {
        num_bytes = ParentNode->ChildrenArray[2].DerNodeType.LeafNode->Data.Length/8;
    }
    else
    {
        num_bytes = ParentNode->ChildrenArray[2].DerNodeType.LeafNode->Data.Length/8 +1;
    }
    if ((ParentNode->ChildrenArray[2].DerNodeType.LeafNode->Data.Data != NULL) && 
        port_IsBadReadPtr(ParentNode->ChildrenArray[2].DerNodeType.LeafNode->Data.Data, num_bytes)) 
    {
        return CSSM_TRUE;
    }

    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: x509_cl_IsBadTbsCertListParseTree
 *
 * Description:
 * This function checks whether this is a valid TbsCertList parse tree
 *
 * Parameters: 
 * ParentNode (input) : The top node in the parse tree to be checked
 * Existance (output) : An array indicating the presence or absence of
 *                      each field
 *
 * Return value:
 * A TRUE/FALSE indicator of whether this is a valid TbsCertList parse tree
 * 
 *---------------------------------------------------------------------------*/
CSSM_BOOL x509_cl_IsBadTbsCertListParseTree(CL_DER_PARENT_NODE_PTR ParentNode, CSSM_BOOL *Existence)
{
    uint32 possibly_missing[3] = {0, 5, 6};

    if (ParentNode == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(ParentNode, sizeof(CL_DER_PARENT_NODE))) 
    {
        return CSSM_TRUE;
    }

    /* Error Checking on first level of tree */
    if ((ParentNode->NumberOfChildren != 7) ||
        CSSM_CLUTIL_X509_IsBadReadPtr(ParentNode->ChildrenArray, ParentNode->NumberOfChildren*sizeof(CL_DER_NODE)))
    {
        return CSSM_TRUE;
    }

    /* find out which optional fields are missing, and error check parse tree tags */
    if ( cssm_clutil_DetermineFieldExistence(ParentNode->NumberOfChildren, 
                                             ParentNode->ChildrenArray, 
                                             3, 
                                             possibly_missing,
                                             Existence) != CSSM_OK)
    {
       return CSSM_TRUE;
    }

    /* These fields should always be present (non-optional) */
    if ((ParentNode->ChildrenArray[1].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE) ||
        (ParentNode->ChildrenArray[2].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE) ||
        ((ParentNode->ChildrenArray[3].DerNodeType.ParentNode->Tag != DER_TAG_UTC_TIME) &&
         (ParentNode->ChildrenArray[3].DerNodeType.ParentNode->Tag != DER_TAG_GENERALIZED_TIME))||
        ((ParentNode->ChildrenArray[4].DerNodeType.ParentNode->Tag != DER_TAG_UTC_TIME) &&
         (ParentNode->ChildrenArray[4].DerNodeType.ParentNode->Tag != DER_TAG_GENERALIZED_TIME)))
    {
        return CSSM_TRUE;
    }

    /* These fields are optional */
    if ((Existence[0] && (ParentNode->ChildrenArray[0].DerNodeType.ParentNode->Tag != DER_TAG_INTEGER)) ||
        (Existence[5] && (ParentNode->ChildrenArray[5].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE)) ||
        (Existence[6] && (ParentNode->ChildrenArray[6].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE)))
    {
        return CSSM_TRUE;
    }

    /* Error Checking on the signature algorithm subtree */
    if (x509_cl_IsBadAlgorithmParseTree(ParentNode->ChildrenArray[1].DerNodeType.ParentNode))
    {
        return CSSM_TRUE;
    }

    /* Error Checking on the issuer subtree */
    if (x509_cl_IsBadNameParseTree(ParentNode->ChildrenArray[2].DerNodeType.ParentNode))
    {
        return CSSM_TRUE;
    }

    return CSSM_FALSE;
}



/*-----------------------------------------------------------------------------
 * Name: x509_cl_IsBadCrlParseTree
 *
 * Description:
 * This function checks whether this is a valid Crl parse tree
 *
 * Parameters: 
 * ParentNode (input) : The top node in the parse tree to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether this is a valid Crl parse tree
 * 
 *---------------------------------------------------------------------------*/
CSSM_BOOL x509_cl_IsBadCrlParseTree(CL_DER_PARENT_NODE_PTR ParentNode)
{
    CSSM_BOOL existence[7];

    if (ParentNode == NULL) 
    {
        return CSSM_FALSE;
    }

    if (port_IsBadReadPtr(ParentNode, sizeof(CL_DER_PARENT_NODE))) 
    {
        return CSSM_TRUE;
    }

    if ((ParentNode->NumberOfChildren != 3) ||
        CSSM_CLUTIL_X509_IsBadReadPtr(ParentNode->ChildrenArray, 3*sizeof(CL_DER_NODE)))
    {
        return CSSM_TRUE;
    }

    if ((ParentNode->ChildrenArray[0].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE) ||
        (ParentNode->ChildrenArray[1].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE) ||
        (ParentNode->ChildrenArray[2].DerNodeType.LeafNode->Tag != DER_TAG_BIT_STRING))
    {
        return CSSM_TRUE;
    }

    if (x509_cl_IsBadTbsCertListParseTree(ParentNode->ChildrenArray[0].DerNodeType.ParentNode, 
                                          existence)) 
    {
        return CSSM_TRUE;
    }
    if (x509_cl_IsBadAlgorithmParseTree(ParentNode->ChildrenArray[1].DerNodeType.ParentNode)) 
    {
        return CSSM_TRUE;
    }
    return CSSM_FALSE;
}



/*-----------------------------------------------------------------------------
 * Name: x509_cl_IsBadSignerInfoParseTree
 *
 * Description:
 * This function checks whether this is a valid SignerInfo parse tree
 *
 * Parameters: 
 * ParentNode (input) : The top node in the parse tree to be checked
 * Existance (output) : An array indicating the presence or absence of
 *                      each field
 *
 * Return value:
 * A TRUE/FALSE indicator of whether this is a valid SignerInfo parse tree
 * 
 *---------------------------------------------------------------------------*/
CSSM_BOOL x509_cl_IsBadSignerInfoParseTree(CL_DER_PARENT_NODE_PTR ParentNode, CSSM_BOOL* Existence)
{
    CL_DER_PARENT_NODE_PTR  tree_iname_and_snum = NULL;
    uint32 possibly_missing[2] = {3, 6};

    if (ParentNode == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(ParentNode, sizeof(CL_DER_PARENT_NODE))) 
    {
        return CSSM_TRUE;
    }

    if (ParentNode->NumberOfChildren == 0) 
    {
        return CSSM_FALSE;  /* Its an empty SET OF */
    }

    /* Error checking on first level of the tree */
    if ((ParentNode->NumberOfChildren != 7) ||
        port_IsBadReadPtr(ParentNode->ChildrenArray, 
                          ParentNode->NumberOfChildren*sizeof(CL_DER_NODE)) )
    {
        return CSSM_TRUE;
    }

    if ( cssm_clutil_DetermineFieldExistence(ParentNode->NumberOfChildren, 
                                             ParentNode->ChildrenArray, 
                                             2, 
                                             possibly_missing, 
                                             Existence) != CSSM_OK)
    {
        return CSSM_TRUE;
    }

    if ((ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Tag   != DER_TAG_INTEGER)  ||
        (ParentNode->ChildrenArray[1].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE) ||
        (ParentNode->ChildrenArray[2].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE) ||
        (ParentNode->ChildrenArray[4].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE) ||
        (ParentNode->ChildrenArray[5].DerNodeType.ParentNode->Tag != DER_TAG_OCTET_STRING) )
    {
        return CSSM_TRUE;
    }

    if ((Existence[3] && (ParentNode->ChildrenArray[3].DerNodeType.ParentNode->Tag != DER_TAG_SET)) ||
        (Existence[6] && (ParentNode->ChildrenArray[6].DerNodeType.ParentNode->Tag != DER_TAG_SET)) )
    {
        return CSSM_TRUE;
    }

    /* Error checking on the version, serial number, and encrypted digest */
    tree_iname_and_snum = ParentNode->ChildrenArray[1].DerNodeType.ParentNode; 
    if ((ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Data.Data != NULL) &&
        port_IsBadReadPtr(ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Data.Data, 
                          ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Data.Length))
    {
        return CSSM_TRUE;
    }
    if ((tree_iname_and_snum->ChildrenArray[1].DerNodeType.LeafNode->Data.Data != NULL) &&
        port_IsBadReadPtr(tree_iname_and_snum->ChildrenArray[1].DerNodeType.LeafNode->Data.Data, 
                          tree_iname_and_snum->ChildrenArray[1].DerNodeType.LeafNode->Data.Length))
    {
        return CSSM_TRUE;
    }
    if ((ParentNode->ChildrenArray[5].DerNodeType.LeafNode->Data.Data != NULL) &&
        port_IsBadReadPtr(ParentNode->ChildrenArray[5].DerNodeType.LeafNode->Data.Data, ParentNode->ChildrenArray[5].DerNodeType.LeafNode->Data.Length))
    {
        return CSSM_TRUE;
    }

    /* Error checking on the issuer subtree */
    if (x509_cl_IsBadNameParseTree(tree_iname_and_snum->ChildrenArray[0].DerNodeType.ParentNode) )
    {
        return CSSM_TRUE;
    }

    /* Error checking on the digest and digest encryption algorithm id subtrees */
    if (x509_cl_IsBadAlgorithmParseTree(ParentNode->ChildrenArray[2].DerNodeType.ParentNode) ||
        x509_cl_IsBadAlgorithmParseTree(ParentNode->ChildrenArray[4].DerNodeType.ParentNode) )
    {
        return CSSM_TRUE;
    }

    /* Error checking on the attributes subtrees */
    if (x509_cl_IsBadAttributesParseTree(ParentNode->ChildrenArray[3].DerNodeType.ParentNode) ||
        x509_cl_IsBadAttributesParseTree(ParentNode->ChildrenArray[6].DerNodeType.ParentNode) )
    {
        return CSSM_TRUE;
    }

    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: x509_cl_IsBadAttributesParseTree
 *
 * Description:
 * This function checks whether this is a valid Attributes parse tree
 *
 * Parameters: 
 * ParentNode (input) : The top node in the parse tree to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether this is a valid Attributes parse tree
 * 
 *---------------------------------------------------------------------------*/
CSSM_BOOL x509_cl_IsBadAttributesParseTree (CL_DER_PARENT_NODE_PTR ParentNode)
{

    uint32 i=0;

    if (ParentNode == NULL) 
    {
        return CSSM_FALSE;
    }
    if (port_IsBadReadPtr(ParentNode, sizeof(CL_DER_PARENT_NODE))) 
    {
        return CSSM_TRUE;
    }

    /* For each attribute */
    for (i=0; i < ParentNode->NumberOfChildren; i++)
    {
        if ((ParentNode->ChildrenArray[i].DerNodeType.ParentNode->NumberOfChildren != 2) ||
            port_IsBadReadPtr(ParentNode->ChildrenArray[i].DerNodeType.ParentNode->ChildrenArray, 
                              ParentNode->ChildrenArray[i].DerNodeType.ParentNode->NumberOfChildren
                                *sizeof(CL_DER_NODE)) )
        {
            return CSSM_TRUE;
        }
        if ((ParentNode->ChildrenArray[i].DerNodeType.ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Tag   != DER_TAG_OID)  ||
            (ParentNode->ChildrenArray[i].DerNodeType.ParentNode->ChildrenArray[1].DerNodeType.ParentNode->Tag != DER_TAG_SET) )
        {
            return CSSM_TRUE;
        }
        if ((ParentNode->ChildrenArray[i].DerNodeType.ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Data.Data != NULL) &&
            port_IsBadReadPtr(ParentNode->ChildrenArray[i].DerNodeType.ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Data.Data, 
                              ParentNode->ChildrenArray[i].DerNodeType.ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Data.Length))
        {
            return CSSM_TRUE;
        }
    }

    return CSSM_FALSE;
}
