/*-----------------------------------------------------------------------
 * File: MISC_DEC.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
 * DER decoding of complex security objects besides certs and crls
 */
#include "cssmport.h"
#include "ber_der.h"
#include "oidsalg.h"
#include "pkcs7oid.h"
#include "clutil.h"
#include "internal.h"
#include <cssmMemTrack.h>

/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_DerDecodeSignedObject
 *
 * Description:
 * This function unpacks a signed object into its components of an object, 
 * an algorithm identifier and a signature
 *
 * Parameters: 
 * SignedObject (input) : the signed object to be decoded
 * Object      (output) : the blob that was signed
 * AlgId       (output) : the algorithm used for signing
 * Signature   (output) : the signature on the object
 *
 * Return value:
 * A success/fail indicator
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_DerDecodeSignedObject (const CSSM_CLUTIL_CONTEXT *pContext,
                                               CSSM_HANDLE AppHandle, 
                                               const CSSM_DATA *SignedObject,
                                               CSSM_DATA_PTR   *Object,
                                               CSSM_ALGORITHMS *AlgId,
                                               CSSM_DATA_PTR   *Signature)
{
    
    BER_PARSED_ITEM s_obj_ber_item = {NULL, 0, NULL},
                    obj_ber_item = {NULL, 0, NULL},
                    alg_ber_item = {NULL, 0, NULL},
                    sig_ber_item = {NULL, 0, NULL};
    const uint8 *block_start = NULL;
    uint32 max_block_length =0;
    CL_DER_TAG_TYPE alg_tag, sig_tag;
    CSSM_DATA temp_data = {0, NULL};
    CSSM_X509_ALGORITHM_IDENTIFIER_PTR alg_id_struct = NULL;
    CSSM_DATA_PTR   blob = NULL, sig = NULL;
    CSSM_ALGORITHMS alg_id = 0;
    CSSM_RETURN     ret = CSSM_OK;

    /* Find the tag, length and data starting points for the object, alg_oid and sig */
    BER_ExpandItem(SignedObject->Data, SignedObject->Length, &s_obj_ber_item);
    if ( s_obj_ber_item.Tag != SignedObject->Data || 
         s_obj_ber_item.ContentLength > SignedObject->Length ||
         s_obj_ber_item.Content >= SignedObject->Data +  SignedObject->Length ||
         s_obj_ber_item.Content != SignedObject->Data + BER_LengthOfTag(s_obj_ber_item.Tag) + BER_LengthOfLength(s_obj_ber_item.ContentLength))
    {
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    BER_ExpandItem(s_obj_ber_item.Content, s_obj_ber_item.ContentLength, &obj_ber_item);
    if ( obj_ber_item.Tag != s_obj_ber_item.Content || 
         obj_ber_item.ContentLength > s_obj_ber_item.ContentLength ||
         obj_ber_item.Content >= s_obj_ber_item.Content + s_obj_ber_item.ContentLength ||
         obj_ber_item.Content != s_obj_ber_item.Content + BER_LengthOfTag(obj_ber_item.Tag) + BER_LengthOfLength(obj_ber_item.ContentLength))
    {
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    block_start = &obj_ber_item.Content[obj_ber_item.ContentLength];
    max_block_length = s_obj_ber_item.ContentLength - (block_start - s_obj_ber_item.Content);
    BER_ExpandItem(block_start, max_block_length,  &alg_ber_item);
    if ( alg_ber_item.Tag != block_start || 
         alg_ber_item.ContentLength > max_block_length ||
         alg_ber_item.Content >= block_start + max_block_length ||
         alg_ber_item.Content != block_start + BER_LengthOfTag(alg_ber_item.Tag) + BER_LengthOfLength(alg_ber_item.ContentLength))
    {
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    block_start = &alg_ber_item.Content[alg_ber_item.ContentLength];
    max_block_length = s_obj_ber_item.ContentLength - (block_start - s_obj_ber_item.Content);
    BER_ExpandItem(block_start, max_block_length, &sig_ber_item);
    if ( sig_ber_item.Tag != block_start || 
         sig_ber_item.ContentLength > max_block_length ||
         sig_ber_item.Content >= block_start + max_block_length ||
         sig_ber_item.Content != block_start + BER_LengthOfTag(sig_ber_item.Tag) + BER_LengthOfLength(sig_ber_item.ContentLength))
    {
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    /* Checks  */
    alg_tag = cl_der_Uint8TagToCLDERTag((uint8*)alg_ber_item.Tag);
    sig_tag = cl_der_Uint8TagToCLDERTag((uint8*)sig_ber_item.Tag);
    if (alg_tag != DER_TAG_SEQUENCE || sig_tag != DER_TAG_BIT_STRING)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    /* Get the algid */
    temp_data.Length = BER_LengthOfTag(alg_ber_item.Tag) + BER_LengthOfLength(alg_ber_item.ContentLength) + alg_ber_item.ContentLength;
    temp_data.Data   = (uint8*)alg_ber_item.Tag;
    ret = CSSM_CLUTIL_X509_DerDecodeAlgorithmId(pContext,AppHandle, &temp_data, &alg_id_struct);
    if (ret != CSSM_OK)
    {
        return ret;
        
    }
    alg_id = CSSM_CLUTIL_X509_AlgorithmOidToAlgId(&alg_id_struct->algorithm);
    CSSM_CLUTIL_X509_FreeAlgorithmId(pContext,AppHandle, alg_id_struct);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, alg_id_struct);

    /* Copy them to the output */
    /* Allocate object and signature */
    blob = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(CSSM_DATA));
    sig  = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(CSSM_DATA));
    if (!blob || !sig)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, blob);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, sig);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    blob->Length = BER_LengthOfTag(obj_ber_item.Tag) + BER_LengthOfLength(obj_ber_item.ContentLength) + obj_ber_item.ContentLength;
    blob->Data = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, blob->Length);
    sig->Length = sig_ber_item.ContentLength -1;
    sig->Data  = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sig->Length);
    if (!blob->Data || !sig->Data)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, blob->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, blob);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, sig->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, sig);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    memcpy(blob->Data, obj_ber_item.Tag, blob->Length);
    memcpy(sig->Data,  sig_ber_item.Content +1, sig->Length);

    *Object = blob;
    *AlgId = alg_id;
    *Signature = sig;

    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_DerGetFirstLevelTagPtrs
 *
 * Description:
 * This function returns an array of pointers to the first level objects
 * inside this DER-encoded object.  Specifically, each element of the array
 * points to the first byte of the Tag of an object contained inside of the
 * top level object 
 *
 * Parameters: 
 * DerEncodedObject (input) : A Der encoded container object, such as a set or sequence
 * EmbeddedObjects (output) : An array of uint8 pointers which point to the first
 *                            Tag byte of each element of this set or sequence
 * NumberOfEmbeddedObjects (output) : The length of the output array
 *
 * Return value:
 * A success/fail indicator
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_DerGetFirstLevelTagPtrs(const CSSM_CLUTIL_CONTEXT *pContext,
                                                CSSM_HANDLE AppHandle, 
                                                const CSSM_DATA *DerEncodedObject,
                                                uint8* **EmbeddedObjects,
                                                uint32 *NumberOfEmbeddedObjects)
{
    BER_PARSED_ITEM obj_ber_item = {NULL, 0, NULL};
    const uint8           *block_start = NULL;
    sint32          max_block_length =0;
    BER_PARSED_ITEM current_ber_item = {NULL, 0, NULL};
    uint32          num_blocks = 0;
    const uint8           **output_array = NULL;
    CSSM_BOOL       bIsIndefinite = CSSM_FALSE;
  
    /* Expand the container */
    BER_ExpandItem(DerEncodedObject->Data, DerEncodedObject->Length, &obj_ber_item);

    /* If the object is not constructed, return failure */
    if ((*obj_ber_item.Tag & BER_CONSTRUCTED) == 0) 
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }
    /* If the object is of constructed indefinite length, subtract 2 from the ContentLength */
    if (cl_der_IsConstructedIndefiniteLength(pContext,DerEncodedObject, &bIsIndefinite) != CSSM_OK)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }
    if (bIsIndefinite)
    {
        obj_ber_item.ContentLength -= 2;
    }

    /* Count the number of nodes */
    current_ber_item = obj_ber_item;
    block_start = obj_ber_item.Content;
    max_block_length = obj_ber_item.ContentLength;

    while (block_start < DerEncodedObject->Data + DerEncodedObject->Length && 
           max_block_length > 0)
    {
        BER_ExpandItem(block_start, max_block_length, &current_ber_item);
        if ( current_ber_item.Tag != block_start || 
             current_ber_item.ContentLength > (uint32) max_block_length ||
             current_ber_item.Content > block_start + max_block_length )
        {
            return CSSMERR_CL_UNKNOWN_FORMAT;
            
        }
        block_start = &current_ber_item.Content[current_ber_item.ContentLength];
        max_block_length = max_block_length - (block_start - current_ber_item.Tag);
        num_blocks++;
    }

    /* Allocate the output array */
    output_array = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, num_blocks, sizeof(uint8 *));
    if (!output_array)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Get the elements of each array */
    current_ber_item = obj_ber_item;
    block_start = obj_ber_item.Content;
    max_block_length = obj_ber_item.ContentLength;
    num_blocks = 0;
    while (block_start < DerEncodedObject->Data + DerEncodedObject->Length && 
           max_block_length > 0)
    {
        BER_ExpandItem(block_start, max_block_length, &current_ber_item);

        output_array[num_blocks] = block_start;
        num_blocks++;

        block_start = &current_ber_item.Content[current_ber_item.ContentLength];
        max_block_length = max_block_length - (block_start - current_ber_item.Tag);
    }

    *NumberOfEmbeddedObjects = num_blocks;
    *EmbeddedObjects = (uint8**)output_array;

    return CSSM_OK;
}
                                          



/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_DerDecodeOneLevel
 *
 * Description:
 * This function returns an array containing the first level objects
 * inside this DER-encoded object.  
 *
 * Parameters: 
 * DerEncodedObject (input) : A Der encoded container object, such as a set or sequence
 * EmbeddedObjects (output) : An array of pointers to the Tag and Data of  
 *                            each element of this set or sequence
 * NumberOfEmbeddedObjects (output) : The length of the output array
 *
 * Return value:
 * A success/fail indicator
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_DerDecodeOneLevel(const CSSM_CLUTIL_CONTEXT *pContext,
                                          CSSM_HANDLE AppHandle, 
                                          const CSSM_DATA *DerEncodedObject,
                                          CL_DER_LEAF_NODE_PTR *EmbeddedObjects,
                                          uint32 *NumberOfEmbeddedObjects)
{
    BER_PARSED_ITEM obj_ber_item = {NULL, 0, NULL};
    const uint8 *block_start = NULL;
    sint32 max_block_length =0;
    BER_PARSED_ITEM current_ber_item = {NULL, 0, NULL};
    uint32 num_blocks = 0;
    CL_DER_LEAF_NODE_PTR output_array = NULL;
    CSSM_BOOL bIsIndefinite = CSSM_FALSE;
    uint32 i=0;

    /* Expand the container */
    BER_ExpandItem(DerEncodedObject->Data, DerEncodedObject->Length, &obj_ber_item);

    /* If the object is not constructed, return failure */
    if ((*obj_ber_item.Tag & BER_CONSTRUCTED) == 0) 
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }
    /* If the object is of constructed indefinite length, subtract 2 from the ContentLength */
    if (cl_der_IsConstructedIndefiniteLength(pContext,DerEncodedObject, &bIsIndefinite) != CSSM_OK)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }
    if (bIsIndefinite)
    {
        obj_ber_item.ContentLength -= 2;
    }
    /* Count the number of nodes */
    current_ber_item = obj_ber_item;
    block_start = obj_ber_item.Content;
    max_block_length = obj_ber_item.ContentLength;

    while (block_start < DerEncodedObject->Data + DerEncodedObject->Length && 
           max_block_length > 0)
    {
        BER_ExpandItem(block_start, max_block_length, &current_ber_item);
        if ( current_ber_item.Tag != block_start || 
             current_ber_item.ContentLength > (uint32) max_block_length ||
             current_ber_item.Content > block_start + max_block_length ||
             current_ber_item.Content != block_start + BER_LengthOfTag(current_ber_item.Tag) + BER_LengthOfLength(current_ber_item.ContentLength))
        {
            return CSSMERR_CL_UNKNOWN_FORMAT;
            
        }
        block_start = &current_ber_item.Content[current_ber_item.ContentLength];
        max_block_length = max_block_length - (block_start - current_ber_item.Tag);
        num_blocks++;
    }

    /* Allocate the output array */
    output_array = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, num_blocks, sizeof(CL_DER_LEAF_NODE));
    if (!output_array)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Get the elements of each array */
    current_ber_item = obj_ber_item;
    block_start = obj_ber_item.Content;
    max_block_length = obj_ber_item.ContentLength;
    num_blocks = 0;
    while (block_start < DerEncodedObject->Data + DerEncodedObject->Length && 
           max_block_length > 0)
    {
        BER_ExpandItem(block_start, max_block_length, &current_ber_item);

        output_array[num_blocks].Tag = cl_der_Uint8TagToCLDERTag((uint8*)current_ber_item.Tag);
        output_array[num_blocks].Data.Length = current_ber_item.ContentLength;
        output_array[num_blocks].Data.Data = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, output_array[num_blocks].Data.Length);
        if (!output_array[num_blocks].Data.Data)
        {
            for (i=0; i < num_blocks; i++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, output_array[i].Data.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, output_array);
            return CSSMERR_CL_MEMORY_ERROR;
        } 
        memcpy(output_array[num_blocks].Data.Data, current_ber_item.Content, output_array[num_blocks].Data.Length);
        num_blocks++;

        block_start = &current_ber_item.Content[current_ber_item.ContentLength];
        max_block_length = max_block_length - (block_start - current_ber_item.Tag);
    }

    *NumberOfEmbeddedObjects = num_blocks;
    *EmbeddedObjects = output_array;

    return CSSM_OK;
}
                                          

/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_PKCS10_DerDecodeCertReq
 *
 * Description:
 * This function gets the Name, SPKI, and Attributes from an input 
 * PKCS#10 CertRequestInfo structure
 *
 * Parameters: 
 * DerEncodedPKCS10 (input)      : the Der encoded PKCS#10 CertRequestInfo structure
 * DerEncodedName  (output)      : A copy of the Der encoded name from inside the PKCS #10 CRI
 * DerEncodedSPKI  (output)      : A copy of the Der encoded SPKI from inside the PKCS #10 CRI
 * DerEncodedAttributes (output) : A copy of the Der encoded attributes from inside the PKCS #10 CRI
 *
 * Return value:
 * A success/fail indicator
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_PKCS10_DerDecodeCertReq (const CSSM_CLUTIL_CONTEXT *pContext,
                                                 CSSM_HANDLE AppHandle, 
                                                 const CSSM_DATA *DerEncodedPKCS10,
                                                 CSSM_DATA_PTR *DerEncodedName,
                                                 CSSM_DATA_PTR *DerEncodedSPKI,
                                                 CSSM_DATA_PTR *DerEncodedAttributes)
{
    CL_DER_LEAF_NODE_PTR pkcs10_leaf_nodes = NULL;
    uint32               num_leaf_nodes = 0;
    CSSM_RETURN          ret = CSSM_OK;
    CSSM_DATA_PTR        temp_name = NULL,
                         temp_spki = NULL,
                         temp_attr = NULL;
    uint32 i=0;

   
    if ( CSSM_CLUTIL_DerDecodeOneLevel(pContext,AppHandle, DerEncodedPKCS10, 
                                       &pkcs10_leaf_nodes, &num_leaf_nodes) != CSSM_OK)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    ret = cl_der_EncodeObject(pContext,AppHandle, pkcs10_leaf_nodes[1].Tag, &pkcs10_leaf_nodes[1].Data, &temp_name);
    if (!temp_name)
    {
        for (i=0; i < num_leaf_nodes; i++)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, pkcs10_leaf_nodes[i].Data.Data);
        }
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, pkcs10_leaf_nodes);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    ret = cl_der_EncodeObject(pContext,AppHandle, pkcs10_leaf_nodes[2].Tag, &pkcs10_leaf_nodes[2].Data, &temp_spki);
    if (!temp_spki)
    {
        for (i=0; i < num_leaf_nodes; i++)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, pkcs10_leaf_nodes[i].Data.Data);
        }
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, pkcs10_leaf_nodes);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_name->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_name);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }
    
    if (num_leaf_nodes == 4)
    {
        if (pkcs10_leaf_nodes[3].Tag == DER_TAG_UNKNOWN) pkcs10_leaf_nodes[3].Tag = DER_TAG_SET;
        ret = cl_der_EncodeObject(pContext,AppHandle, pkcs10_leaf_nodes[3].Tag, &pkcs10_leaf_nodes[3].Data, &temp_attr);
        if (!temp_attr)
        {
            for (i=0; i < num_leaf_nodes; i++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, pkcs10_leaf_nodes[i].Data.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, pkcs10_leaf_nodes);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_name->Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_name);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_spki->Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_spki);
            return CSSMERR_CL_INTERNAL_ERROR;
            
        }
    }

    for (i=0; i < num_leaf_nodes; i++)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, pkcs10_leaf_nodes[i].Data.Data);
    }
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, pkcs10_leaf_nodes);

    *DerEncodedName = temp_name;
    *DerEncodedSPKI = temp_spki;
    *DerEncodedAttributes = temp_attr;

    return CSSM_OK;
}



/*-----------------------------------------------------------------------------
 * Name: cssm_clutil_PKCS7_GetContentPtrFromContentInfo
 *
 * Description:
 * This function obtains a pointer to the content data structure
 * from a DER encoded ContentInfo 
 *
 * Parameters: 
 * ContentInfo (input) : The DER encoded PKCS #7 contentInfo structure
 * ExpectedContentType (input) : The OID of the expected type of PKCS #7 content 
 * Content (output) : a pointer to the underlying content
 *
 * Return value:
 * A CSSM_RETURN value.  If CSSM_OK, the ContentType equaled the Expected ContentType
 * and the Content pointer was obtained
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN cssm_clutil_PKCS7_GetContentPtrFromContentInfo (const CSSM_CLUTIL_CONTEXT *pContext,
                                                            CSSM_HANDLE AppHandle, 
                                                            const CSSM_DATA *ContentInfo,
                                                            const CSSM_OID *ExpectedContentType,
                                                            CSSM_DATA *Content  )
{
    uint8           **ci_obj_ptr_array = NULL, wrap_tag = 0xA0;
    uint32          num_ci_obj = 0;
    CSSM_DATA       content_type;
    BER_PARSED_ITEM ber_item_c_type, ber_item_wrap_content;

    /* Obtain pointers to the ContentType OID and Content */
    if ( CSSM_CLUTIL_DerGetFirstLevelTagPtrs (pContext,AppHandle, ContentInfo, &ci_obj_ptr_array, &num_ci_obj) != CSSM_OK)
    {
        return CSSMERR_CL_INVALID_DATA;
    }
    if (num_ci_obj != 2)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, ci_obj_ptr_array);
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* If the ContentType != ExpectedContentType OID, fail */
    BER_ExpandItem(ci_obj_ptr_array[0], BER_SizeofObject(ci_obj_ptr_array[0]), &ber_item_c_type);
    content_type.Length = ber_item_c_type.ContentLength;
    content_type.Data   = (uint8*)ber_item_c_type.Content;
    if (cl_der_Uint8TagToCLDERTag((uint8*)ber_item_c_type.Tag) != DER_TAG_OID ||
        !cssm_clutil_IsEquivalentCssmData(&content_type, ExpectedContentType))
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, ci_obj_ptr_array);
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* Remove the explicit wrapping */
    BER_ExpandItem(ci_obj_ptr_array[1], BER_SizeofObject(ci_obj_ptr_array[1]), &ber_item_wrap_content);
    if (memcmp(ber_item_wrap_content.Tag, &wrap_tag, sizeof(uint8)) != 0 )
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, ci_obj_ptr_array);
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* Assign the output content pointer */
    (*Content).Length = ber_item_wrap_content.ContentLength;
    (*Content).Data   = (uint8*)ber_item_wrap_content.Content;
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, ci_obj_ptr_array);

    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_PKCS7_DerDecodeSignedData
 *
 * Description:
 * This function decodes a PKCS #7 signed data structure
 *
 * Parameters: 
 * DerEncodedPkcs7SignedData (input) : The DER encoded PKCS #7 signed data structure
 *
 * Return value:
 * The C data structure equivalent of the input DER-encoded PKCS #7 signed data blob 
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_PKCS7_DerDecodeSignedData (const CSSM_CLUTIL_CONTEXT *pContext,
                                                   CSSM_HANDLE AppHandle, 
                                                   const CSSM_DATA *DerEncodedPkcs7SignedData,
                                                   cssm_PKCS7_SIGNED_DATA_PTR *result_val)
{
    cssm_PKCS7_SIGNED_DATA_PTR   results_ptr = NULL;
    CL_DER_NODE_PTR         parsed_tree = NULL;
    DER_COMPLEX_TYPE_PTR    der_type_ptr = NULL;
    CSSM_DATA               pkcs7_data = {0, NULL};
    CSSM_RETURN             ret = CSSM_OK;
  
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *result_val = NULL;
    if (cssm_clutil_PKCS7_GetContentPtrFromContentInfo(pContext,AppHandle, DerEncodedPkcs7SignedData, 
                                                       &CSSMOID_Pkcs7SignedData, 
                                                       &pkcs7_data) != CSSM_OK)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* Unpack the certificate data */
    ret = clutil_pkcs7_CreateSignedDataDerType(pContext,AppHandle, &pkcs7_data, &der_type_ptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    ret = cl_der_ParseDerObject (pContext,AppHandle, &pkcs7_data, der_type_ptr, &parsed_tree);
    der_type_ptr->Subtypes[5] = 0;  /* part of fix for bug #634 - SignerInfo subtype is static */

    cl_der_DeleteType(pContext,AppHandle, der_type_ptr);
    if (!parsed_tree)
    {
        return ret;
    }

    /* Get the PKS #7 Signed Data structure */
    ret = clutil_pkcs7_SignedDataParseTreeToStruct(pContext,AppHandle, parsed_tree, &pkcs7_data, &results_ptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* Cleanup */
    ret = cl_der_DeleteParsedDerObject (pContext,AppHandle, parsed_tree);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, parsed_tree);

    /* Return the output */
    
    *result_val = results_ptr;
    return CSSM_OK;

}




/*-----------------------------------------------------------------------------
 * Name: clutil_pkcs7_GetNumCertsAndCrlsFromPkcs7Blob
 *
 * Description:
 * This function returns pointers to the beginning of the CertGroup and CrlGroup
 * encodings, as well as a count of the number of certs and the number of CRLs
 * in the encoded structure
 * 
 * Parameters: 
 * Pkcs7SignedData  (input) : The DER-encoded PKCS #7 signed data structure 
 *                            containing 0 or more certs and crls
 * NumCerts        (output) : The number of certs in the PKCS #7 struct
 * EncCertGroupPtr (output) : A pointer to the DerEncoded Cert set inside the 
 *                            der encoded PKCS#7 signed data blob
 * NumCrls         (output) : The number of CRLs in the PKCS #7 struct
 * EncCrlGroupPtr  (output) : A pointer to the DerEncoded Crl set inside the 
 *                            der encoded PKCS#7 signed data blob
 *
 * Return value:
 * A success/failure indicator
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN clutil_pkcs7_GetCertAndCrlInfoFromPkcs7Blob(const CSSM_CLUTIL_CONTEXT *pContext,
                                                        CSSM_HANDLE AppHandle, 
                                                        CSSM_DATA_PTR Pkcs7SignedData, 
                                                        uint32 *NumCerts, 
                                                        uint8* *EncCertGroupPtr,
                                                        uint32 *NumCrls,
                                                        uint8* *EncCrlGroupPtr)
{
    uint8  **pkcs7_obj_array = NULL,
           **obj_array = NULL;
    uint32 num_pkcs7_obj = 0,
           num_obj = 0;
    CSSM_DATA temp_data = {0, NULL};

    if( CSSM_CLUTIL_DerGetFirstLevelTagPtrs(pContext,AppHandle, Pkcs7SignedData, &pkcs7_obj_array, &num_pkcs7_obj) != CSSM_OK)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

    if (num_pkcs7_obj < 4 || num_pkcs7_obj > 6)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, pkcs7_obj_array);
        return CSSMERR_CL_INVALID_DATA;
        
    }
    if (*pkcs7_obj_array[0] != 0x02 || *pkcs7_obj_array[1] != 0x31 ||
        *pkcs7_obj_array[2] != 0x30 || *pkcs7_obj_array[num_pkcs7_obj-1] != 0x31 )
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, pkcs7_obj_array);
        return CSSMERR_CL_INVALID_DATA;
        
    }

    if (num_pkcs7_obj == 4) /* There are no CRLs or Certs */
    {
        *NumCerts = 0;
        *EncCertGroupPtr = NULL;
        *NumCrls = 0;
        *EncCrlGroupPtr = NULL;
        return CSSM_OK;
    }

    /* More error checking */
    if (*pkcs7_obj_array[3] != 0xA0 && *pkcs7_obj_array[3] != 0xA1)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, pkcs7_obj_array);
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* Get the number of objects in the third set */
    temp_data.Length = BER_SizeofObject(pkcs7_obj_array[3]);
    temp_data.Data = pkcs7_obj_array[3];
    if( CSSM_CLUTIL_DerGetFirstLevelTagPtrs(pContext,AppHandle, &temp_data, &obj_array, &num_obj) != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, pkcs7_obj_array);
        return CSSMERR_CL_INVALID_DATA;
        
    }
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, obj_array);

    if (num_pkcs7_obj == 5)
    {
        if (*pkcs7_obj_array[3] == 0xA0) /* This PKCS#7 contains a certificate set, but no CRL set */
        {
            *NumCerts = num_obj;
            *EncCertGroupPtr = pkcs7_obj_array[3];
            *NumCrls = 0;
            *EncCrlGroupPtr = NULL;
        } 
        else  
        {   /* This PKCS#7 contains a CRL set, but no certificate set */
            *NumCerts = 0;
            *EncCertGroupPtr = NULL;
            *NumCrls = num_obj;
            *EncCrlGroupPtr = pkcs7_obj_array[3];
        }
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, pkcs7_obj_array);
        return CSSM_OK;
    }

    /* This contained a cert set and a crl set */
    *NumCerts = num_obj;  /* From the third position */
    *EncCertGroupPtr = pkcs7_obj_array[3];

    /* Get the number of objects in the CRL set */
    temp_data.Length = BER_SizeofObject(pkcs7_obj_array[4]);
    temp_data.Data = pkcs7_obj_array[4];
    if( CSSM_CLUTIL_DerGetFirstLevelTagPtrs(pContext,AppHandle, &temp_data, &obj_array, &num_obj) != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, pkcs7_obj_array);
        return CSSMERR_CL_INVALID_DATA;
        
    }
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, obj_array);
    *NumCrls = num_obj;
    *EncCrlGroupPtr = pkcs7_obj_array[4];

    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, pkcs7_obj_array);
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: clutil_pkcs7_CreateSignedDataDerType
 *
 * Description:
 * This function creates the DER type structure required to properly decode
 * the DER-encoded pkcs #7 Signed data structure
 *
 * Parameters: 
 * Pkcs7SignedData (input) : The DER-encoded PKCS #7 signed data structure 
 *
 * Return value:
 * The DER type structure corresponding to a PKCS #7 signed data structure
 * 
 *---------------------------------------------------------------------------*/

/* 
 * The following static data is part of the fix to bug 634.
 *
 * The rest of the CL is using dynamic memory for analogous structures.
 * Do not attempt to free the static structures.
 */
static unsigned char constructed_context_specific_0[]   = 
                            { BER_CONSTRUCTED | BER_CLASS_CONTEXTSPECIFIC | 0 };
static unsigned char constructed_context_specific_1[]   = 
                            { BER_CONSTRUCTED | BER_CLASS_CONTEXTSPECIFIC | 1 };
static unsigned char objectIDTag[] =            { BER_OBJECT_IDENTIFIER };
static unsigned char octetStringTag[] =         { BER_OCTET_STRING };
static unsigned char integerTag[] =             { BER_INTEGER };
static unsigned char constructedSequence[] =    { BER_SEQUENCE | BER_CONSTRUCTED };
static unsigned char constructedSet[] =         { BER_SET | BER_CONSTRUCTED };
static unsigned char printableStringTag[] =     { BER_PRINTABLE_STRING };

/* signerInfoType */
static unsigned char *signerInfoContextTags[7] = {  /* context-specific signedData Tags */
    integerTag,             /* version */
    constructedSequence,    /* issuer and serial number */
    constructedSequence,    /* digest algorithm identifier (and parameters) */
    constructed_context_specific_0, /* authenticated attributes */
    constructedSequence,    /* signature algorithm (and parameters) */
    octetStringTag,         /* encrypted digest */
    constructed_context_specific_1  /* unauthenticated attributes */
};
static unsigned char *signerInfoOriginalTags[7] = {  /* original signedData Tags */
    integerTag,             /* version */
    constructedSequence,    /* issuer and serial number */
    constructedSequence,    /* digest algorithm identifier (and parameters) */
    constructedSet,         /* authenticated attributes */
    constructedSequence,    /* signature algorithm (and parameters) */
    octetStringTag,         /* encrypted digest */
    constructedSet          /* unauthenticated attributes */
};
static DER_COMPLEX_TYPE signerInfoType = {
    constructedSequence,
    7,
    0,                      /* all implicit */
    (const uint8 **)signerInfoContextTags,        /* ifdef VMS, apply cast */		
    (const uint8 **)signerInfoOriginalTags,       /* ifdef VMS, apply cast */
    0,                      /* no defaults */
    0,                      /* ignore subTypes for now */
    "signerInfoType"
};                      

/* set of signerInfo Type */
static unsigned char *setOfSignerInfoContextTags[] = {constructedSequence}; 
static DER_COMPLEX_TYPE *setOfSignerInfoSubTypes = {&signerInfoType};
static DER_COMPLEX_TYPE setOfSignerInfoType = {
    constructedSet,
    1,
    0,                      /* all implicit */
    (const uint8 **)setOfSignerInfoContextTags,                     /* ifdef VMS, apply cast */
    (const uint8 **)setOfSignerInfoContextTags, /* original Tags */ /* ifdef VMS, apply cast */
    0,                      /* no defaults */
    &setOfSignerInfoSubTypes,
    "setOfSignerInfoType"
};
#define signerInfos_index 5

 
CSSM_RETURN  clutil_pkcs7_CreateSignedDataDerType(const CSSM_CLUTIL_CONTEXT *pContext,
                                                       CSSM_HANDLE AppHandle, CSSM_DATA_PTR Pkcs7SignedData,
                                                       DER_COMPLEX_TYPE_PTR *result_val)
{
    DER_COMPLEX_TYPE_PTR subtype_ptr = NULL;
    sint16          num_types = 6;
    CL_DER_TAG_TYPE tag = 0;
    sint16          i=0;
    uint8           *enc_certgroup_ptr = NULL, 
                    *enc_crlgroup_ptr = NULL;
    uint32          num_certs = 0, num_crls = 0;
    CSSM_RETURN     ret = CSSM_OK;

    *result_val = NULL;
    /* Get num_certs and num_crls */
    ret = clutil_pkcs7_GetCertAndCrlInfoFromPkcs7Blob(pContext,AppHandle, 
                  Pkcs7SignedData, 
                  &num_certs, &enc_certgroup_ptr,
                  &num_crls,  &enc_crlgroup_ptr);
    if (CSSM_OK != ret)
    {
        return ret;
        
    }

    /* Allocate a type structure for the unsigned certificate */
    subtype_ptr = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(DER_COMPLEX_TYPE));
    if (!subtype_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    subtype_ptr->TypeTag      = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(CL_DER_TAG_TYPE));
    subtype_ptr->OriginalTags = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, num_types, sizeof(CL_DER_TAG_TYPE *));
    subtype_ptr->ContextTags  = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, num_types, sizeof(CL_DER_TAG_TYPE *));
    subtype_ptr->Subtypes     = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, num_types, sizeof(DER_COMPLEX_TYPE_PTR));
    if (!subtype_ptr->TypeTag      || 
        !subtype_ptr->OriginalTags || !subtype_ptr->ContextTags || !subtype_ptr->Subtypes)
    {
        cl_der_DeleteType(pContext,AppHandle, subtype_ptr);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Fill in the type tag and numberOfDefaults */
    tag = BER_CONSTRUCTED_SEQUENCE;
    port_memcpy((void*)subtype_ptr->TypeTag, &tag, sizeof(CL_DER_TAG_TYPE));
    subtype_ptr->NumberOfDefaults = num_types;

    /* Fill in the certs subtype */
    if (num_certs)
    {
        ret = x509_cl_CreateSetofSignedCertDerType(pContext,AppHandle, num_certs, &subtype_ptr->Subtypes[3]);
        if (ret != CSSM_OK)
        {
            cl_der_DeleteType(pContext,AppHandle, subtype_ptr);
            return ret;
            
        }
    }

    /* Fill in the crls subtype */
    if (num_crls)
    {
        x509_cl_CreateSetofSignedCrlDerType(pContext,AppHandle, num_crls, enc_crlgroup_ptr, &(subtype_ptr->Subtypes[4]));
        if (!subtype_ptr->Subtypes[4])
        {
            cl_der_DeleteType(pContext,AppHandle, subtype_ptr);
            return CSSMERR_CL_INTERNAL_ERROR;
            
        }
    }

    /* Allocate the original and context specific tags for the data portion of a certificate */
    for (i=0; i < num_types; i++)
    {
        subtype_ptr->OriginalTags[i] = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(CL_DER_TAG_TYPE));
        subtype_ptr->ContextTags[i]  = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(CL_DER_TAG_TYPE));
        if (!subtype_ptr->OriginalTags[i] || !subtype_ptr->ContextTags[i])
        {
            cl_der_DeleteType(pContext,AppHandle, subtype_ptr);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
    }

    /* Fill in the original tags  */
    tag = DER_TAG_INTEGER;
    port_memcpy((void*)subtype_ptr->OriginalTags[0], &tag, sizeof(CL_DER_TAG_TYPE)); 
    tag = BER_CONSTRUCTED_SEQUENCE;
    port_memcpy((void*)subtype_ptr->OriginalTags[2], &tag, sizeof(CL_DER_TAG_TYPE));  
    tag = BER_CONSTRUCTED_SET;
    port_memcpy((void*)subtype_ptr->OriginalTags[1], &tag, sizeof(CL_DER_TAG_TYPE)); 
    port_memcpy((void*)subtype_ptr->OriginalTags[3], &tag, sizeof(CL_DER_TAG_TYPE));  
    port_memcpy((void*)subtype_ptr->OriginalTags[4], &tag, sizeof(CL_DER_TAG_TYPE));  
    port_memcpy((void*)subtype_ptr->OriginalTags[5], &tag, sizeof(CL_DER_TAG_TYPE));  

    /* Fill in the context-specific tags */
    tag = DER_TAG_INTEGER;
    port_memcpy((void*)subtype_ptr->ContextTags[0], &tag, sizeof(CL_DER_TAG_TYPE));  
    tag = BER_CONSTRUCTED_SEQUENCE;
    port_memcpy((void*)subtype_ptr->ContextTags[2], &tag, sizeof(CL_DER_TAG_TYPE));  
    tag = BER_CONSTRUCTED_SET;
    port_memcpy((void*)subtype_ptr->ContextTags[1], &tag, sizeof(CL_DER_TAG_TYPE));  
    port_memcpy((void*)subtype_ptr->ContextTags[5], &tag, sizeof(CL_DER_TAG_TYPE));  
    tag = 0xA0;
    port_memcpy((void*)subtype_ptr->ContextTags[3], &tag, sizeof(CL_DER_TAG_TYPE));  
    tag = 0xA1;
    port_memcpy((void*)subtype_ptr->ContextTags[4], &tag, sizeof(CL_DER_TAG_TYPE)); 

    /* Fill in the SignerInfos subtype */
    subtype_ptr->Subtypes[signerInfos_index] = &setOfSignerInfoType;

    *result_val = subtype_ptr;
    return CSSM_OK;
}



/*-----------------------------------------------------------------------------
 * Name: x509_cl_GetNumRecordsFromCrlParseTree
 *
 * Description:
 * This function returns the number of Revoked Cert entries based
 * on the number of children in the Crl Parse tree
 *
 * Parameters: 
 * CrlRoot (input) : The root of the Crl parse tree
 *
 * Return value:
 * The number of revoked entries
 * 
 *---------------------------------------------------------------------------*/
uint32 x509_cl_GetNumRecordsFromCrlParseTree(CL_DER_PARENT_NODE_PTR CrlRoot)
{
    CL_DER_PARENT_NODE_PTR tbsCertList = NULL;
    CSSM_BOOL              existence[7];
    uint32                 possibly_missing[3] = {0, 5, 6};

    tbsCertList = CrlRoot->ChildrenArray[0].DerNodeType.ParentNode;
    if ( cssm_clutil_DetermineFieldExistence (tbsCertList->NumberOfChildren, tbsCertList->ChildrenArray, 
                                              3, possibly_missing,
                                              existence) != CSSM_OK)
    {
       return 0;
    }

    if (existence[5])
    {
        return tbsCertList->ChildrenArray[5].DerNodeType.ParentNode->NumberOfChildren;
    }

    return 0;
}


/*-----------------------------------------------------------------------------
 * Name: clutil_pkcs7_SetOfCertsParseTreeToStruct
 *
 * Description:
 * This function translates the input Der encoded certificate set into 
 * a CSSM_CERTGROUP
 *
 * Parameters: 
 * EncodedCertGroup (input) : A pointer to the beginning of the DER encoded certificate set
 *
 * Return value:
 * A CSSM_CERTGROUP containing copies of the certs in the input EncodedCertGroup
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN clutil_pkcs7_SetOfCertsParseTreeToStruct(const CSSM_CLUTIL_CONTEXT *pContext,
                                                     CSSM_HANDLE AppHandle, 
                                                     uint8 * EncodedCertGroup,
                                                     CSSM_CERTGROUP_PTR *result_val)
{
    CSSM_CERTGROUP_PTR   results_ptr = NULL;
    CSSM_DATA            temp_data = {0, NULL};
    uint8**              enc_cert_ptrs = NULL;
    uint32               num_certs = 0;
    uint32               i=0, j=0;
    *result_val = NULL;
    /* Get a list of the embedded certs */
    temp_data.Length = BER_SizeofObject(EncodedCertGroup);
    temp_data.Data = EncodedCertGroup;
    if ( CSSM_CLUTIL_DerGetFirstLevelTagPtrs(pContext,
                                             AppHandle,
                                             &temp_data,
                                             &enc_cert_ptrs,
                                             &num_certs) != CSSM_OK)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }

    /* Allocate the result */
    results_ptr = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(CSSM_CERTGROUP));
    if (!results_ptr)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_cert_ptrs);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    results_ptr->NumCerts = num_certs;
    results_ptr->GroupList.CertList = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, results_ptr->NumCerts, sizeof(CSSM_DATA));
    if (!results_ptr->GroupList.CertList)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_cert_ptrs);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, results_ptr);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    for (i=0; i < results_ptr->NumCerts; i++)
    {
        results_ptr->GroupList.CertList[i].Length = BER_SizeofObject(enc_cert_ptrs[i]);
        if (results_ptr->GroupList.CertList[i].Length)
        {
            results_ptr->GroupList.CertList[i].Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, results_ptr->GroupList.CertList[i].Length);
            if (!results_ptr->GroupList.CertList[i].Data)
            {
                for (j=0; j < i; j++)
                {
                    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, results_ptr->GroupList.CertList[j].Data);
                }
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_cert_ptrs);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, results_ptr);
                return CSSMERR_CL_MEMORY_ERROR;
                
            }
            memcpy(results_ptr->GroupList.CertList[i].Data, enc_cert_ptrs[i], results_ptr->GroupList.CertList[i].Length);
        }
    }
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_cert_ptrs);

    *result_val = results_ptr;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: clutil_pkcs7_SetOfCrlsParseTreeToStruct
 *
 * Description:
 * This function translates the input Der encoded CRL set into 
 * a CSSM_CRLGROUP
 *
 * Parameters: 
 * EncodedCrlGroup (input) : A pointer to the beginning of the DER encoded CRL set
 *
 * Return value:
 * A CSSM_CRLGROUP containing copies of the certs in the input EncodedCrlGroup
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN clutil_pkcs7_SetOfCrlsParseTreeToStruct (const CSSM_CLUTIL_CONTEXT *pContext,
                                                     CSSM_HANDLE AppHandle, 
                                                     uint8 * EncodedCrlGroup,
                                                     CSSM_CRLGROUP_PTR *result_val)
{
    CSSM_CRLGROUP_PTR    results_ptr = NULL;
    CSSM_DATA            temp_data = {0, NULL};
    uint8**              enc_crl_ptrs = NULL;
    uint32               num_crls = 0;
    uint32               i=0, j=0;
    CSSM_RETURN          ret = CSSM_OK;

    /* Get a list of the embedded crls */
    temp_data.Length = BER_SizeofObject(EncodedCrlGroup);
    temp_data.Data = EncodedCrlGroup;
    ret = CSSM_CLUTIL_DerGetFirstLevelTagPtrs(pContext,AppHandle, &temp_data, &enc_crl_ptrs, &num_crls);
    if (CSSM_OK != ret)
    {
        return ret;
        
    }

    /* Allocate the result */
    results_ptr = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(CSSM_CRLGROUP));
    if (!results_ptr)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_crl_ptrs);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    results_ptr->NumberOfCrls = num_crls;
    results_ptr->GroupList.CrlList = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, results_ptr->NumberOfCrls, sizeof(CSSM_DATA));
    if (!results_ptr->GroupList.CrlList)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_crl_ptrs);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, results_ptr);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    for (i=0; i < results_ptr->NumberOfCrls; i++)
    {
        results_ptr->GroupList.CrlList[i].Length = BER_SizeofObject(enc_crl_ptrs[i]);
        if (results_ptr->GroupList.CrlList[i].Length)
        {
            results_ptr->GroupList.CrlList[i].Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, results_ptr->GroupList.CrlList[i].Length);
            if (!results_ptr->GroupList.CrlList[i].Data)
            {
                for (j=0; j < i; j++)
                {
                    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, results_ptr->GroupList.CrlList[j].Data);
                }
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_crl_ptrs);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, results_ptr);
                return CSSMERR_CL_MEMORY_ERROR;
                
            }
            memcpy(results_ptr->GroupList.CrlList[i].Data, enc_crl_ptrs[i], results_ptr->GroupList.CrlList[i].Length);
        }
    }
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_crl_ptrs);

    *result_val = results_ptr;
    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: clutil_x509_IsEquivalentDName
 *
 * Description:
 * This function compares the fields of the input certs to determine if they are equivalent
 *
 * Parameters: 
 * Name1 (input) : An X501 distinguished name as a C data structure
 * Name2 (input) : An X501 distinguished name as a C data structure
 *
 * Return value:
 * CSSM_TRUE if the input distinguished names are equivalent
 * CSSM_FALSE otherwise
 *---------------------------------------------------------------------------*/
CSSM_BOOL clutil_x509_IsEquivalentDName(CSSM_X509_NAME_PTR Name1, CSSM_X509_NAME_PTR Name2)
{
    uint32 i=0, j=0;

    if (Name1->numberOfRDNs != Name2->numberOfRDNs)
    {
        return CSSM_FALSE;
    }
    for (i=0; i< Name1->numberOfRDNs; i++)
    { 
        if (Name1->RelativeDistinguishedName[i].numberOfPairs != Name2->RelativeDistinguishedName[i].numberOfPairs)
        {
            return CSSM_FALSE;
        }

        for (j=0; j < Name1->RelativeDistinguishedName[i].numberOfPairs; j++)
        {
            if (Name1->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Length != Name2->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Length ||
                memcmp(Name1->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Data, Name2->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Data, Name1->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Length) ||
                Name1->RelativeDistinguishedName[i].AttributeTypeAndValue[j].valueType != Name2->RelativeDistinguishedName[i].AttributeTypeAndValue[j].valueType ||
                memcmp(Name1->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Data, Name2->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Data,Name1->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Length) )
            {
                return CSSM_FALSE;
            }
        }
    }

    return CSSM_TRUE;

}

/*-----------------------------------------------------------------------------
 * Name: clutil_pkcs7_SignerInfosCertFirst
 *
 * Description:
 * This function places the certificate associated with the first SignerInfos
 * in the PKCS #7 Signed Data struct first in the certificate list
 *
 * Parameters: 
 * pkcs7_ptr (input/output) : the PKCS7 signed data struct
 *
 * Return value:
 * none
 * 
 *---------------------------------------------------------------------------*/

void clutil_pkcs7_SignerInfosCertFirst(const CSSM_CLUTIL_CONTEXT *pContext,
                                       CSSM_HANDLE AppHandle, 
                                      cssm_PKCS7_SIGNED_DATA_PTR pkcs7_ptr)
{
    CSSM_RETURN   ret = CSSM_OK;
    CSSM_X509_NAME_PTR FirstSignerInfosIssuerNamep;
    CSSM_DATA_PTR FirstSignerInfosSerialNumberp;
    CSSM_X509_NAME_PTR CertNamep;
    CSSM_DATA_PTR CertSerialNumberp;
    unsigned int i;
    CSSM_X509_SIGNED_CERTIFICATE_PTR SignedCertificatep;
    CSSM_DATA TempCert;

    /* a SignerInfos exists */
    if (pkcs7_ptr->numberOfSignerInfos == 0) {
        return; /* no SignerInfos */
    }

    /* Initialize variables */
    FirstSignerInfosIssuerNamep = NULL;
    FirstSignerInfosSerialNumberp = NULL;
    CertNamep = NULL;
    CertSerialNumberp = NULL;
    SignedCertificatep = NULL;

    /* Get IssuerAndSerialNumber from first SignerInfos */
    FirstSignerInfosIssuerNamep = &(pkcs7_ptr->signerInfos->issuerAndSerialNumber.issuer);
    FirstSignerInfosSerialNumberp = &(pkcs7_ptr->signerInfos->issuerAndSerialNumber.serialNumber);

    /* Find matching certificate */
    for (i=0; i < pkcs7_ptr->certificates.NumCerts; i++) 
    {
        /* Decode cert */
        ret = CSSM_CLUTIL_X509_DerDecodeSignedCertificate(pContext,AppHandle, 
                                    &(pkcs7_ptr->certificates.GroupList.CertList[i]), &SignedCertificatep);
        if (SignedCertificatep == NULL)
        {
            return;  /* decode failed */
        }

        /* Check serial number */
        if (FirstSignerInfosSerialNumberp->Length != SignedCertificatep->certificate.serialNumber.Length)
        {
            CSSM_CLUTIL_X509_FreeX509V3SignedCert(pContext,AppHandle, SignedCertificatep);
            continue; /* no match */
        }
        if (memcmp(FirstSignerInfosSerialNumberp->Data, SignedCertificatep->certificate.serialNumber.Data, 
                FirstSignerInfosSerialNumberp->Length) != 0)
        {
            CSSM_CLUTIL_X509_FreeX509V3SignedCert(pContext,AppHandle, SignedCertificatep);
            continue; /* no match */
        }

        /* Check issuer name */
        if (clutil_x509_IsEquivalentDName(FirstSignerInfosIssuerNamep, &(SignedCertificatep->certificate.issuer))
                == CSSM_TRUE)
        {
            CSSM_CLUTIL_X509_FreeX509V3SignedCert(pContext,AppHandle, SignedCertificatep);
            break; /* IssuerAndSerialNumber match! */
        }

        CSSM_CLUTIL_X509_FreeX509V3SignedCert(pContext,AppHandle, SignedCertificatep);
    }

    /*  Check for a match */
    if (i != pkcs7_ptr->certificates.NumCerts)
    {
        /* if the matching certificate's index is zero, its already first in the certificate list */
        if (i != 0)
        {
            /* Order matching certificate first in CSSM_CERTGROUP */
            TempCert.Length = pkcs7_ptr->certificates.GroupList.CertList[0].Length;
            TempCert.Data = pkcs7_ptr->certificates.GroupList.CertList[0].Data;

            pkcs7_ptr->certificates.GroupList.CertList[0].Length = pkcs7_ptr->certificates.GroupList.CertList[i].Length;
            pkcs7_ptr->certificates.GroupList.CertList[0].Data = pkcs7_ptr->certificates.GroupList.CertList[i].Data;

            pkcs7_ptr->certificates.GroupList.CertList[i].Length = TempCert.Length;
            pkcs7_ptr->certificates.GroupList.CertList[i].Data = TempCert.Data;
        }
    }

    AppHandle;  /* for MSVC warning level 4 */

    return;
}

/*-----------------------------------------------------------------------------
 * Name: clutil_pkcs7_SignedDataParseTreeToStruct
 *
 * Description:
 * This function translates a CL parse tree into a PKCS #7 Signed Data struct
 *
 * Parameters: 
 * parsed_tree (input) : The CL parse tree
 *
 * result_val (output): A pointer to a PKCS #7 Signed Data struct
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN clutil_pkcs7_SignedDataParseTreeToStruct (const CSSM_CLUTIL_CONTEXT *pContext,
                                                      CSSM_HANDLE AppHandle, 
                                                      CL_DER_NODE_PTR parsed_tree, 
                                                      CSSM_DATA_PTR EncPkcs7SignedData,
                                                      cssm_PKCS7_SIGNED_DATA_PTR *result_val)
{
    cssm_PKCS7_SIGNED_DATA_PTR   results_ptr = NULL;
    CL_DER_PARENT_NODE_PTR  tree_alg_ids = NULL,
                            tree_content_info = NULL,
                            tree_certs = NULL,
                            tree_crls = NULL,
                            tree_signer_infos = NULL;
    CSSM_X509_ALGORITHM_IDENTIFIER_PTR  algorithm_ptr = NULL;
    uint32                  possibly_missing[2] = {3, 4};
    CSSM_BOOL               existence[6], 
                            siginfo_exist[7];
    CSSM_CERTGROUP_PTR      cert_group_ptr = NULL;
    CSSM_RETURN             ret = CSSM_OK;
    CSSM_CRLGROUP_PTR       crl_group_ptr = NULL;
    
    cssm_PKCS7_SIGNER_INFO_PTR   sig_info_ptr = NULL;
    
    uint32 i=0;
    uint8           *enc_certgroup_ptr = NULL, 
                    *enc_crlgroup_ptr = NULL;
    uint32          num_certs = 0, num_crls = 0;

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    /* Error checking on first level of the tree */
    if (parsed_tree->DerNodeType.ParentNode->NumberOfChildren != 6 ||
        port_IsBadReadPtr(parsed_tree->DerNodeType.ParentNode->ChildrenArray, parsed_tree->DerNodeType.ParentNode->NumberOfChildren*sizeof(CL_DER_NODE)) )
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }
    if (cssm_clutil_DetermineFieldExistence(parsed_tree->DerNodeType.ParentNode->NumberOfChildren, parsed_tree->DerNodeType.ParentNode->ChildrenArray, 2, possibly_missing, existence) != CSSM_OK)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }
    if (parsed_tree->DerNodeType.ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Tag   != DER_TAG_INTEGER  ||
        parsed_tree->DerNodeType.ParentNode->ChildrenArray[1].DerNodeType.ParentNode->Tag != DER_TAG_SET ||
        parsed_tree->DerNodeType.ParentNode->ChildrenArray[2].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE ||
        parsed_tree->DerNodeType.ParentNode->ChildrenArray[5].DerNodeType.ParentNode->Tag != DER_TAG_SET )
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }
    if ((existence[3] && parsed_tree->DerNodeType.ParentNode->ChildrenArray[3].DerNodeType.ParentNode->Tag != DER_TAG_SET) ||
        (existence[4] && parsed_tree->DerNodeType.ParentNode->ChildrenArray[4].DerNodeType.ParentNode->Tag != DER_TAG_SET) )
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }

    /* Error checking on the digest algorithm ids subtree */
    tree_alg_ids = parsed_tree->DerNodeType.ParentNode->ChildrenArray[1].DerNodeType.ParentNode;
    for (i=0; i < tree_alg_ids->NumberOfChildren; i++)
    {
        if (x509_cl_IsBadAlgorithmParseTree(tree_alg_ids->ChildrenArray[i].DerNodeType.ParentNode))
        {
            return CSSMERR_CL_UNKNOWN_FORMAT;
            
        }
    }

    /* Error checking on the content info subtree */
    tree_content_info = parsed_tree->DerNodeType.ParentNode->ChildrenArray[2].DerNodeType.ParentNode;
    if (x509_cl_IsBadAlgorithmParseTree(tree_content_info))
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }

    /* Error checking on the certificate subtree */
    if (existence[3])
    {
        tree_certs = parsed_tree->DerNodeType.ParentNode->ChildrenArray[3].DerNodeType.ParentNode;
    }

    /* Error checking on the CRL subtree */
    if (existence[4])
    {
        tree_crls = parsed_tree->DerNodeType.ParentNode->ChildrenArray[4].DerNodeType.ParentNode;
        for (i=0; i < tree_crls->NumberOfChildren; i++)
        {
            if (x509_cl_IsBadCrlParseTree(tree_crls->ChildrenArray[i].DerNodeType.ParentNode) )
            {
                return CSSMERR_CL_UNKNOWN_FORMAT;
                
            }
        }
    }

    /* Error checking on the signer info subtree */
    tree_signer_infos = parsed_tree->DerNodeType.ParentNode->ChildrenArray[5].DerNodeType.ParentNode;
    for (i=0; i < tree_signer_infos->NumberOfChildren; i++)
    {
        if (x509_cl_IsBadSignerInfoParseTree(tree_signer_infos->ChildrenArray[i].DerNodeType.ParentNode, siginfo_exist))
        {
            return CSSMERR_CL_UNKNOWN_FORMAT;
            
        }
    }

    /* Get pointers to the encoded certs and crls */
    if (clutil_pkcs7_GetCertAndCrlInfoFromPkcs7Blob(pContext,AppHandle, 
                  EncPkcs7SignedData, 
                  &num_certs, &enc_certgroup_ptr,
                  &num_crls,  &enc_crlgroup_ptr) != CSSM_OK)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }


    /* Allocate memory for the new pkcs #7 structure */
    results_ptr = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(cssm_PKCS7_SIGNED_DATA));
    if (!results_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Fill in all data values */
    /* Fill in the version */
    results_ptr->version.Length = parsed_tree->DerNodeType.ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Data.Length;
    results_ptr->version.Data = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, results_ptr->version.Length);
    if(!results_ptr->version.Data)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, results_ptr);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    port_memcpy(results_ptr->version.Data, parsed_tree->DerNodeType.ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Data.Data, results_ptr->version.Length);

    /* Fill in the digest algids info */
    if (tree_alg_ids->NumberOfChildren)
    {
        results_ptr->numberOfDigestAlgIds = tree_alg_ids->NumberOfChildren;
        results_ptr->digestAlgIds = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, results_ptr->numberOfDigestAlgIds, sizeof(CSSM_X509_ALGORITHM_IDENTIFIER));
        if (!results_ptr->digestAlgIds)
        {
            CSSM_CLUTIL_PKCS7_FreeSignedData(pContext,AppHandle, results_ptr);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
        for (i=0; i < results_ptr->numberOfDigestAlgIds; i++)
        {
            ret = x509_cl_AlgorithmParseTreeToStruct(pContext,AppHandle, &tree_alg_ids->ChildrenArray[i], &algorithm_ptr);
            if (!algorithm_ptr)
            {
                CSSM_CLUTIL_PKCS7_FreeSignedData(pContext,AppHandle, results_ptr);
                return CSSMERR_CL_UNKNOWN_FORMAT;
                
            }
            results_ptr->digestAlgIds[i] = *algorithm_ptr;
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, algorithm_ptr);
        }
    }

    /* Fill in the content info */
    ret = x509_cl_AlgorithmParseTreeToStruct(pContext,AppHandle, &parsed_tree->DerNodeType.ParentNode->ChildrenArray[2], &algorithm_ptr);
    if (!algorithm_ptr)
    {
        CSSM_CLUTIL_PKCS7_FreeSignedData(pContext,AppHandle, results_ptr);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    results_ptr->contentInfo = *algorithm_ptr;
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, algorithm_ptr);

    /* Fill in the certificates */
    if (existence[3] && tree_certs->NumberOfChildren)
    {
        ret = clutil_pkcs7_SetOfCertsParseTreeToStruct(pContext,AppHandle, enc_certgroup_ptr, &cert_group_ptr);
        if (CSSM_OK != ret)
        {
            CSSM_CLUTIL_PKCS7_FreeSignedData(pContext,AppHandle, results_ptr);
            return ret; 
            
        }
        results_ptr->certificates = *cert_group_ptr;
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, cert_group_ptr);
    }

    /* Fill in the CRLs */
    if (existence[4] && tree_crls->NumberOfChildren)
    {
        ret = clutil_pkcs7_SetOfCrlsParseTreeToStruct(pContext,AppHandle, enc_crlgroup_ptr, &crl_group_ptr);
        if (CSSM_OK != ret)
        {
            CSSM_CLUTIL_PKCS7_FreeSignedData(pContext,AppHandle, results_ptr);
            return ret; 
            
        }
        results_ptr->crls = *crl_group_ptr;
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, crl_group_ptr);
    }

    /* Fill in the Signer Infos */
    if (tree_signer_infos->NumberOfChildren)  /* There is a SignerInfos */
    {
        if (tree_signer_infos->ChildrenArray[0].DerNodeType.ParentNode->NumberOfChildren != 0) /* empty SET OF */
        {
            results_ptr->numberOfSignerInfos = tree_signer_infos->NumberOfChildren;
            results_ptr->signerInfos = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, results_ptr->numberOfSignerInfos, sizeof(cssm_PKCS7_SIGNER_INFO));
            if (!results_ptr->signerInfos)
            {
                CSSM_CLUTIL_PKCS7_FreeSignedData(pContext,AppHandle, results_ptr);
                return CSSMERR_CL_MEMORY_ERROR;
                
            }
            for (i=0; i < results_ptr->numberOfSignerInfos; i++)
            {
                clutil_pkcs7_SignerInfoParseTreeToStruct(pContext,AppHandle, &tree_signer_infos->ChildrenArray[i], &sig_info_ptr);
                if (!sig_info_ptr)
                {
                    CSSM_CLUTIL_PKCS7_FreeSignedData(pContext,AppHandle, results_ptr);
                    return CSSMERR_CL_UNKNOWN_FORMAT;
                    
                }
                results_ptr->signerInfos[i] = *sig_info_ptr;
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, sig_info_ptr);
            }
            /* if there is a SignerInfos, 
               order the certificate corresponding to the first SignerInfos (IssuerAndSerialNumber)
               first in the list of certificates ...
            */
            clutil_pkcs7_SignerInfosCertFirst(pContext,AppHandle, results_ptr);
        }
    }

    /* Return the struct */
    
    *result_val = results_ptr;
    return CSSM_OK;
}



/*-----------------------------------------------------------------------------
 * Name: clutil_pkcs7_SignerInfoParseTreeToStruct
 *
 * Description:
 * This function translates a CL parse tree into a PKCS #7 Signer Info struct
 *
 * Parameters: 
 * parsed_tree (input) : the CL parse tree
 *
 * result_val (output): A pointer to a PKCS #7 Signer Info struct
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN clutil_pkcs7_SignerInfoParseTreeToStruct(const CSSM_CLUTIL_CONTEXT *pContext,
                                                     CSSM_HANDLE AppHandle, 
                                                     CL_DER_NODE_PTR parsed_tree,
                                                     cssm_PKCS7_SIGNER_INFO_PTR *result_val)
{
    cssm_PKCS7_SIGNER_INFO_PTR   results_ptr = NULL;
    CL_DER_PARENT_NODE_PTR  tree_iname_and_snum = NULL,
                            tree_dig_algid = NULL,
                            tree_auth_attr = NULL,
                            tree_enc_algid = NULL,
                            tree_unauth_attr = NULL;
    CSSM_X509_NAME_PTR           temp_name_ptr = NULL;
    CSSM_X509_ALGORITHM_IDENTIFIER_PTR  algorithm_ptr = NULL;
    CSSM_X501_ATTRIBUTES_PTR     attrs_ptr = NULL;
   
    CSSM_BOOL               existence[7];
    CSSM_RETURN             ret = CSSM_OK;
 
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    if ( x509_cl_IsBadSignerInfoParseTree (parsed_tree->DerNodeType.ParentNode, existence)) 
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }

    tree_iname_and_snum = parsed_tree->DerNodeType.ParentNode->ChildrenArray[1].DerNodeType.ParentNode; 
    tree_dig_algid = parsed_tree->DerNodeType.ParentNode->ChildrenArray[2].DerNodeType.ParentNode;
    tree_enc_algid = parsed_tree->DerNodeType.ParentNode->ChildrenArray[4].DerNodeType.ParentNode;
    tree_auth_attr = parsed_tree->DerNodeType.ParentNode->ChildrenArray[3].DerNodeType.ParentNode;
    tree_unauth_attr = parsed_tree->DerNodeType.ParentNode->ChildrenArray[6].DerNodeType.ParentNode;


    /* Allocate memory for the new pkcs #7 signer info structure */
    results_ptr = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(cssm_PKCS7_SIGNER_INFO));
    if (!results_ptr)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, existence);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Fill in all data values */
    /* Fill in the version */
    results_ptr->version.Length = parsed_tree->DerNodeType.ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Data.Length;
    results_ptr->version.Data = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, results_ptr->version.Length);
    if(!results_ptr->version.Data)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, results_ptr);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, existence);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    port_memcpy(results_ptr->version.Data, parsed_tree->DerNodeType.ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Data.Data, results_ptr->version.Length);

    /* Fill in the issuer name */
    ret = x509_cl_X509NameParseTreeToStruct(pContext,AppHandle, &tree_iname_and_snum->ChildrenArray[0], &temp_name_ptr);
    if (!temp_name_ptr)
    {
        CSSM_CLUTIL_PKCS7_FreeSignerInfo(pContext,AppHandle, results_ptr);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, existence);
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }
    results_ptr->issuerAndSerialNumber.issuer = *temp_name_ptr;
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_name_ptr);

    /* Fill in the serial number */
    results_ptr->issuerAndSerialNumber.serialNumber.Length = tree_iname_and_snum->ChildrenArray[1].DerNodeType.LeafNode->Data.Length;
    results_ptr->issuerAndSerialNumber.serialNumber.Data = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, results_ptr->issuerAndSerialNumber.serialNumber.Length);
    if(!results_ptr->issuerAndSerialNumber.serialNumber.Data)
    {
        CSSM_CLUTIL_PKCS7_FreeSignerInfo(pContext,AppHandle, results_ptr);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, existence);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    port_memcpy(results_ptr->issuerAndSerialNumber.serialNumber.Data, 
        tree_iname_and_snum->ChildrenArray[1].DerNodeType.LeafNode->Data.Data, 
        results_ptr->issuerAndSerialNumber.serialNumber.Length);

    /* Fill in the digest algid info */
    ret = x509_cl_AlgorithmParseTreeToStruct(pContext,AppHandle, &parsed_tree->DerNodeType.ParentNode->ChildrenArray[2], &algorithm_ptr);
    if (!algorithm_ptr)
    {
        CSSM_CLUTIL_PKCS7_FreeSignerInfo(pContext,AppHandle, results_ptr);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, existence);
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }
    results_ptr->digestAlgId = *algorithm_ptr;
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, algorithm_ptr);

    /* Fill in the auth attr info */
    if (existence[3])
    {
        clutil_x501_AttributesParseTreeToStruct(pContext,AppHandle, &parsed_tree->DerNodeType.ParentNode->ChildrenArray[3], &attrs_ptr);
        if (!attrs_ptr)
        {
            CSSM_CLUTIL_PKCS7_FreeSignerInfo(pContext,AppHandle, results_ptr);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, existence);
            return CSSMERR_CL_UNKNOWN_FORMAT;
            
        }
        results_ptr->authenticatedAttrs = *attrs_ptr;
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attrs_ptr);
    }

    /* Fill in the digest encryption algid info */
    ret = x509_cl_AlgorithmParseTreeToStruct(pContext,AppHandle, &parsed_tree->DerNodeType.ParentNode->ChildrenArray[4], &algorithm_ptr);
    if (!algorithm_ptr)
    {
        CSSM_CLUTIL_PKCS7_FreeSignerInfo(pContext,AppHandle, results_ptr);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, existence);
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }
    results_ptr->digestEncryptionAlgId = *algorithm_ptr;
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, algorithm_ptr);

    /* Fill in the encrypted digest */
    results_ptr->encryptedDigest.Length = parsed_tree->DerNodeType.ParentNode->ChildrenArray[5].DerNodeType.LeafNode->Data.Length;
    results_ptr->encryptedDigest.Data = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, results_ptr->encryptedDigest.Length);
    if(!results_ptr->encryptedDigest.Data)
    {
        CSSM_CLUTIL_PKCS7_FreeSignerInfo(pContext,AppHandle, results_ptr);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, existence);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    port_memcpy(results_ptr->encryptedDigest.Data, 
                parsed_tree->DerNodeType.ParentNode->ChildrenArray[5].DerNodeType.LeafNode->Data.Data, 
                results_ptr->encryptedDigest.Length);

    /* Fill in the unauth attr info */
    if (existence[6])
    {
        clutil_x501_AttributesParseTreeToStruct(pContext,AppHandle, &parsed_tree->DerNodeType.ParentNode->ChildrenArray[6], &attrs_ptr);
        if (!attrs_ptr)
        {
            CSSM_CLUTIL_PKCS7_FreeSignerInfo(pContext,AppHandle, results_ptr);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, existence);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
        results_ptr->unauthenticatedAttrs = *attrs_ptr;
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attrs_ptr);
    }

    /* Return the struct */ 
    
    *result_val = results_ptr;
    return CSSM_OK;
}



/*-----------------------------------------------------------------------------
 * Name: clutil_pkcs7_AttributesParseTreeToStruct
 *
 * Description:
 * This function translates a CL parse tree into an X501 Attributes struct
 *
 * Parameters: 
 * parsed_tree (input) : the CL parse tree
 *
 * result_val (output) : A pointer to an X501 Attributes struct
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN clutil_x501_AttributesParseTreeToStruct(const CSSM_CLUTIL_CONTEXT *pContext,
                                                    CSSM_HANDLE AppHandle, 
                                                    CL_DER_NODE_PTR parsed_tree,
                                                    CSSM_X501_ATTRIBUTES_PTR *result_val)
{
    CSSM_X501_ATTRIBUTES_PTR     results_ptr = NULL;
    CL_DER_PARENT_NODE_PTR  tree_attr = NULL,
                            tree_values = NULL;
    CSSM_RETURN             ret = CSSM_OK;
    CSSM_DATA_PTR           temp_data_ptr = NULL;
    CL_DER_LEAF_NODE_PTR    temp_leaf = NULL;
    uint32 i=0, j=0;

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    if (x509_cl_IsBadAttributesParseTree(parsed_tree->DerNodeType.ParentNode) )
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
        
    }

    /* Allocate the attributes structure */
    results_ptr = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(CSSM_X501_ATTRIBUTES));
    if (!results_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Fill in the attributes structure */
    /* Create the attributes array */
    results_ptr->numberOfAttributes = parsed_tree->DerNodeType.ParentNode->NumberOfChildren;
    if (results_ptr->numberOfAttributes) 
    {
        results_ptr->attributes = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
            results_ptr->numberOfAttributes, sizeof(CSSM_X501_ATTRIBUTE));
        if (!results_ptr->attributes)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, results_ptr);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
    } 
    else
    {
        results_ptr->attributes = NULL;
    }
    /* For each attributes array, create and fill in the values array */
    for (i=0; i < results_ptr->numberOfAttributes; i++)
    {
        tree_attr = parsed_tree->DerNodeType.ParentNode->ChildrenArray[i].DerNodeType.ParentNode;

        /* Fill in the type */
        results_ptr->attributes[i].type.Length = tree_attr->ChildrenArray[0].DerNodeType.LeafNode->Data.Length;
        if (results_ptr->attributes[i].type.Length)
        {
            results_ptr->attributes[i].type.Data  = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, results_ptr->attributes[i].type.Length);
            if (!results_ptr->attributes[i].type.Data)
            {
                CSSM_CLUTIL_X501_FreeAttributes(pContext,AppHandle, results_ptr);
                return CSSMERR_CL_MEMORY_ERROR;
                
            }
            port_memcpy(results_ptr->attributes[i].type.Data, 
                tree_attr->ChildrenArray[0].DerNodeType.LeafNode->Data.Data, 
                results_ptr->attributes[i].type.Length);
        }
        else
        {
            results_ptr->attributes[i].type.Data = NULL;
        }

        /* Create the values array */
        tree_values = tree_attr->ChildrenArray[1].DerNodeType.ParentNode;
        results_ptr->attributes[i].numberOfValues = tree_values->NumberOfChildren;
        if (results_ptr->attributes[i].numberOfValues)
        {
            results_ptr->attributes[i].values =
                CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, results_ptr->attributes[i].numberOfValues, sizeof(CSSM_X501_ANY_VALUE));
            if (!results_ptr->attributes[i].values)
            {
                CSSM_CLUTIL_X501_FreeAttributes(pContext,AppHandle, results_ptr);
                return CSSMERR_CL_MEMORY_ERROR;
                
            }
        }
        else
        {
            results_ptr->attributes[i].values = NULL;
        }
        /* Fill in the values array */
        for (j=0; j < results_ptr->attributes[i].numberOfValues; j++)
        {

            /* Reconstruct the value */
            if (tree_values->ChildrenArray[j].IsLeaf)
            {
                ret = cl_der_EncodeObject(pContext,AppHandle, tree_values->ChildrenArray[j].DerNodeType.LeafNode->Tag, &tree_values->ChildrenArray[j].DerNodeType.LeafNode->Data, &temp_data_ptr);
            }
            else
            {
                ret = cl_der_PackParseTree (pContext,AppHandle, tree_values->ChildrenArray[j].DerNodeType.ParentNode, NULL, &temp_data_ptr);
            }
            if (!temp_data_ptr)
            {
                CSSM_CLUTIL_X501_FreeAttributes(pContext,AppHandle, results_ptr);
                return CSSMERR_CL_MEMORY_ERROR;
                
            }

            ret =  cl_der_DecodeObject(pContext,AppHandle, temp_data_ptr, &temp_leaf);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_data_ptr->Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_data_ptr);
            if (!temp_leaf)
            {
                CSSM_CLUTIL_X501_FreeAttributes(pContext,AppHandle, results_ptr);
                return CSSMERR_CL_MEMORY_ERROR;
                
            }

            /* Fill in the value Tag */
            results_ptr->attributes[i].values[j].valueTag = temp_leaf->Tag;
            results_ptr->attributes[i].values[j].value    = temp_leaf->Data;
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_leaf);
        }
    }

    /* Return */
    
    *result_val = results_ptr;
    return CSSM_OK;

}


