/*-----------------------------------------------------------------------
 * File: MISC_ENC.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
#include "cssmport.h"
#include "ber_der.h"
#include "oidsalg.h"
#include "pkcs7oid.h"
#include "clutil.h"
#include "internal.h"
#include <cssmMemTrack.h>

/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_DerEncodeSignedObject
 *
 * Description:
 * Many objects consist of a data blob followed by a DER signature..
 * the signature algid and a bit string signature.  This routine accepts
 * the data blob, signing algid and signature and encodes them into a 
 * single blob
 *
 * Parameters:
 * Object    (input) : The DER encoded objet which was signed
 * AlgId     (input) : The algorithm used for signing
 * Signature (input) : The signature 
 *
 * result_val (output): The DER encoding of the signed object
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_DerEncodeSignedObject(CSSM_CLUTIL_CONTEXT *pContext,
                                              CSSM_HANDLE AppHandle, 
                                              const CSSM_DATA *Object,
                                              CSSM_ALGORITHMS AlgId,
                                              const CSSM_DATA *Signature,
                                              CSSM_DATA_PTR *result_val)
{
    /* Initializations */
    CSSM_DATA_PTR        results_ptr = NULL;
    CSSM_RETURN          ret = CSSM_OK, retVal = CSSM_OK;
    CSSM_OID_PTR         algoid = NULL;  
    CSSM_X509_ALGORITHM_IDENTIFIER x509algid = {0, NULL, 0, NULL};
    CSSM_DATA_PTR        enc_sigalg_ptr = NULL;

    uint32               num_types  = 3;
    CL_DER_CHILD_INFO    object_array[3];
    CL_DER_NODE_HANDLE   root = 0;

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    /* Initializations */
    port_memset(object_array, 0, num_types*sizeof(CL_DER_CHILD_INFO));
    
    /*  Create the encoded algorithm identifier */
    algoid = CSSM_CLUTIL_X509_AlgIdToAlgorithmOid(AlgId);
    if (algoid == NULL)
    {
        return CSSMERR_CL_INVALID_DATA;
    }
    x509algid.algorithm = *algoid; 

    ret =  CSSM_CLUTIL_X509_DerEncodeAlgorithmId(pContext,AppHandle, &x509algid, &enc_sigalg_ptr); 
    if (ret != CSSM_OK)
    {
        return ret;
        
    }
        
    /*  Create the DER encoded signed object */
    ret = cl_der_CreateLeafNode(pContext,
            (CL_DER_TAG_TYPE) (Object->Data[0] & 0x1f),
            cl_der_LengthOfContent(Object),
            &Object->Data[cl_der_LengthOfTag(Object) + cl_der_LengthOfLength(Object)], &(object_array[0].ChildHandle));
    
    if (ret != CSSM_OK)
    {
        return ret;
        
    }

    ret = cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE,
            cl_der_LengthOfContent(enc_sigalg_ptr),
            &enc_sigalg_ptr->Data[cl_der_LengthOfTag(enc_sigalg_ptr) + cl_der_LengthOfLength(enc_sigalg_ptr)], &(object_array[1].ChildHandle));

     if (ret != CSSM_OK)
     {
         retVal = cl_der_DeleteLeafNode(pContext,object_array[0].ChildHandle);
         CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_sigalg_ptr->Data);
         CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_sigalg_ptr);
         return ret;
         
     }
    
    
    ret = cl_der_CreateLeafNode(pContext,DER_TAG_BIT_STRING,
                                                      8*Signature->Length, Signature->Data, &(object_array[2].ChildHandle));
    

    if (ret != CSSM_OK)
    {
         retVal = cl_der_DeleteLeafNode(pContext,object_array[0].ChildHandle);
         retVal = cl_der_DeleteLeafNode(pContext,object_array[1].ChildHandle);
         CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_sigalg_ptr->Data);
         CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_sigalg_ptr);
         return ret;
         
    }

    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_sigalg_ptr->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_sigalg_ptr);

    if (!object_array[0].ChildHandle || !object_array[1].ChildHandle || !object_array[2].ChildHandle)
    {
        retVal = cl_der_DeleteLeafNode(pContext,object_array[0].ChildHandle);
        retVal = cl_der_DeleteLeafNode(pContext,object_array[1].ChildHandle);
        retVal = cl_der_DeleteLeafNode(pContext,object_array[2].ChildHandle);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    ret =  cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_types, object_array, &root); 

    if(ret != CSSM_OK)
    {
         cl_der_DeleteTree (pContext,root);
         return ret;
    }

    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, &results_ptr);
    cl_der_DeleteTree (pContext,root);

    
    *result_val = results_ptr;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X501_DerEncodeAttributes
 *
 * Description:
 * This function translates an X501 attributes structure into a DER-encoded blob
 *
 * Parameters: 
 * Attributes (input) : The C representation of the Attributes to be encoded
 *
 * result_val(output): The DER encoding of the input Attributes
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X501_DerEncodeAttributes (CSSM_CLUTIL_CONTEXT *pContext,
                                                  CSSM_HANDLE AppHandle, 
                                                  const CSSM_X501_ATTRIBUTES *Attributes,
                                                  CSSM_DATA_PTR *result_val)
{
    CSSM_DATA_PTR           results_ptr = NULL;
    CL_DER_CHILD_INFO_PTR   attrs_array = NULL, value_array = NULL;
    uint32                  num_attr_types = 2;
    CL_DER_CHILD_INFO       attr_struct[2];
    CL_DER_NODE_HANDLE      root = 0;
    uint32 i=0, j=0, k=0;
    CSSM_RETURN             ret1 = CSSM_OK, ret2 = CSSM_OK,
                            ret = CSSM_OK, retVal = CSSM_OK;
    /* Initializations */
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    attrs_array = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, Attributes->numberOfAttributes, sizeof(CL_DER_CHILD_INFO));
    if (!attrs_array)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    port_memset(attr_struct, 0, num_attr_types * sizeof(CL_DER_CHILD_INFO));

    /* For each Attribute, create its values array */
    for (i=0; i < Attributes->numberOfAttributes; i++)
    {
        /* Allocate the values array */
        value_array = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 
            Attributes->attributes[i].numberOfValues, sizeof(CL_DER_CHILD_INFO));
        if (!value_array)
        {
            for (k=0; k<i; k++)
            {
                cl_der_DeleteParentNode(pContext,attrs_array[k].ChildHandle);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attrs_array);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }

        /* Create a leaf for each value */
        for (j=0; j < Attributes->attributes[i].numberOfValues; j++)
        {
            ret =  
                cl_der_CreateLeafNode  (pContext,Attributes->attributes[i].values[j].valueTag, 
                Attributes->attributes[i].values[j].value.Length,
                Attributes->attributes[i].values[j].value.Data, &(value_array[j].ChildHandle));

            if (ret != CSSM_OK)
            {
                for (k=0; k<j; k++)
                {
                    retVal = cl_der_DeleteLeafNode(pContext,value_array[k].ChildHandle);
                }
                for (k=0; k<i; k++)
                {
                    cl_der_DeleteParentNode(pContext,attrs_array[k].ChildHandle);
                }
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, value_array);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attrs_array);
                return ret;
                
            }

        }

        /* Create the leaves for the attr type and value array */
        ret1 = cl_der_CreateLeafNode  (pContext,DER_TAG_OID, 
                Attributes->attributes[i].type.Length,
                Attributes->attributes[i].type.Data, &(attr_struct[0].ChildHandle));

        ret2 = cl_der_CreateParentNode(pContext,DER_TAG_SET, 
            Attributes->attributes[i].numberOfValues, 
            value_array, &(attr_struct[1].ChildHandle));
        if ((ret1 != CSSM_OK) || (ret2 != CSSM_OK))
        {
            if (!attr_struct[1].ChildHandle) 
            {
                for (k=0; k<Attributes->attributes[i].numberOfValues; k++)
                {
                    retVal = cl_der_DeleteLeafNode(pContext,value_array[k].ChildHandle);
                }
            }
            for (k=0; k<i; k++)
            {
                cl_der_DeleteParentNode(pContext,attrs_array[k].ChildHandle);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, value_array);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attrs_array);
            return CSSMERR_CL_INTERNAL_ERROR;
            
        }
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, value_array);

        /* Create the parent node for this Attribute */
        ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_attr_types, attr_struct, &(attrs_array[i].ChildHandle));
        if (ret != CSSM_OK)
        {
            retVal = cl_der_DeleteLeafNode(pContext,attr_struct[0].ChildHandle);
            cl_der_DeleteParentNode(pContext,attr_struct[1].ChildHandle);
            for (k=0; k<i; k++)
            {
                cl_der_DeleteParentNode(pContext,attrs_array[k].ChildHandle);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attrs_array);
            return ret;
            
        }
        port_memset(attr_struct, 0, num_attr_types * sizeof(CL_DER_CHILD_INFO));
    }

    /* Create the parent node for this name struct */
    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, Attributes->numberOfAttributes, attrs_array, &root);

    if(ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attrs_array);
        cl_der_DeleteTree (pContext,root);
        return ret;
    }

    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attrs_array);

    /* Pack the key info struct into a DER encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, &results_ptr);

    /* Delete the parse tree */
    cl_der_DeleteTree (pContext,root);

    /* Return the result */
    
    *result_val = results_ptr;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_PKCS10_DerEncodeCertReq
 *
 * Description:
 * This function combines the input Name, SPKI, and Attributes structures
 * into a Der encoded PKCS #10 v1 Certificate Request Info 
 *
 * Parameters: 
 * DerEncodedName       (input) : The Der encoded Name
 * DerEncodedSPKI       (input) : The Der encoded SPKI
 * DerEncodedAttributes (input) : The Der encoded Attributes
 *
 * result_val (output): A Der encoded PKCS #10 certificate request structure 
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_PKCS10_DerEncodeCertReq (CSSM_CLUTIL_CONTEXT *pContext,
                                                 CSSM_HANDLE AppHandle, 
                                                 const CSSM_DATA *DerEncodedName,
                                                 const CSSM_DATA *DerEncodedSPKI,
                                                 const CSSM_DATA *DerEncodedAttributes,
                                                 CSSM_DATA_PTR *result_val)
{
    /* Initializations */
    CSSM_DATA_PTR        results_ptr = NULL;
    uint32               num_certreq  = 4;
    CL_DER_CHILD_INFO    certreq_array[4];
    CL_DER_NODE_HANDLE   root = 0;
    uint8                version[1] = {0};
    CSSM_RETURN          ret = CSSM_OK, retVal = CSSM_OK;
    port_memset(certreq_array, 0, num_certreq*sizeof(CL_DER_CHILD_INFO));
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    /*  Create the leaf nodes for the PKCS #10 message. */
    ret = cl_der_CreateLeafNode(pContext,DER_TAG_INTEGER,
                                    sizeof(uint8),
                                    version,
                                    &(certreq_array[0].ChildHandle));

    if(ret != CSSM_OK)
    {
        return ret;
        
    }
    
    ret = cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE,
            cl_der_LengthOfContent(DerEncodedName),
            &DerEncodedName->Data[cl_der_LengthOfTag(DerEncodedName) + cl_der_LengthOfLength(DerEncodedName)], &(certreq_array[1].ChildHandle));        

    if(ret != CSSM_OK)
    {
        retVal = cl_der_DeleteLeafNode(pContext,certreq_array[0].ChildHandle);
        return ret;
        
    }
    
    ret = cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE,
            cl_der_LengthOfContent(DerEncodedSPKI),
            &DerEncodedSPKI->Data[cl_der_LengthOfTag(DerEncodedSPKI) + cl_der_LengthOfLength(DerEncodedSPKI)], &(certreq_array[2].ChildHandle));

    if(ret != CSSM_OK)
    {
        retVal =cl_der_DeleteLeafNode(pContext,certreq_array[0].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,certreq_array[1].ChildHandle);
        return ret;
        
    }
    certreq_array[3].IsOptional = CSSM_TRUE;
    certreq_array[3].HasCtxSpecificTag = CSSM_TRUE;
    ret = cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE,
        cl_der_LengthOfContent(DerEncodedAttributes),
        &DerEncodedAttributes->Data[cl_der_LengthOfTag(DerEncodedAttributes) + cl_der_LengthOfLength(DerEncodedAttributes)], &(certreq_array[3].ChildHandle));

    if(ret != CSSM_OK)
    {
        retVal =cl_der_DeleteLeafNode(pContext,certreq_array[0].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,certreq_array[1].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,certreq_array[2].ChildHandle);
        return  ret;
        
    }
    
    if (!certreq_array[0].ChildHandle || !certreq_array[1].ChildHandle ||
        !certreq_array[2].ChildHandle || !certreq_array[3].ChildHandle )
    {
        retVal =cl_der_DeleteLeafNode(pContext,certreq_array[0].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,certreq_array[1].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,certreq_array[2].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,certreq_array[3].ChildHandle);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    } 

    ret =  cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE,
                                    num_certreq,
                                    certreq_array, &root);

    if(ret != CSSM_OK)
    {
        cl_der_DeleteTree (pContext,root);
        return ret;
        
    }
    
    
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, &results_ptr);

    if(ret != CSSM_OK)
    {
        cl_der_DeleteTree (pContext,root);
        return ret;
        
    }

    
    cl_der_DeleteTree (pContext,root);
    
    *result_val = results_ptr;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: cssm_clutil_PKCS7_DerEncodeContentInfo
 *
 * Description:
 * This function returns the DER encoding of the input content with the input OID
 *
 * Parameters: 
 * ContentType (input) : The OID identifying the encapsulated content
 * Content     (input) : The data to be encapsulated
 *
 * result_val (output): A DER encoded ContentInfo structure
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN cssm_clutil_PKCS7_DerEncodeContentInfo (CSSM_CLUTIL_CONTEXT *pContext,
                                                      CSSM_HANDLE AppHandle, 
                                                      const CSSM_OID *ContentType,
                                                      const CSSM_DATA *Content,
                                                      CSSM_DATA_PTR *result_val)
{
    CSSM_DATA_PTR       results_ptr = NULL;
    uint32              num_ci_types = 2;
    CL_DER_CHILD_INFO   ci_struct_array[2];
    CL_DER_NODE_HANDLE  root = 0;
    CSSM_RETURN         ret = CSSM_OK, retVal = CSSM_OK;
    /* Initializations */
    port_memset(ci_struct_array, 0, num_ci_types*sizeof(CL_DER_CHILD_INFO));
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    /* Create the DER encoded PKCS #7 structure */
    ret =  
        cl_der_CreateLeafNode (pContext,DER_TAG_OID, ContentType->Length, ContentType->Data, &(ci_struct_array[0].ChildHandle));

    if(ret != CSSM_OK)
    {
        return ret;
        
    }

    ci_struct_array[1].IsOptional = CSSM_TRUE;
    ci_struct_array[1].HasCtxSpecificTag = CSSM_TRUE;
    ci_struct_array[1].IsExplicit = CSSM_TRUE;
    ret =  
        cl_der_CreateLeafNode (pContext,cl_der_Uint8TagToCLDERTag(Content->Data), 
            cl_der_LengthOfContent(Content),
            &Content->Data[ cl_der_LengthOfTag(Content) + cl_der_LengthOfLength(Content)], &(ci_struct_array[1].ChildHandle));
    
    if(ret != CSSM_OK)
    {
        retVal =cl_der_DeleteLeafNode(pContext,ci_struct_array[0].ChildHandle);
        return ret;
        
    }

    if (!ci_struct_array[0].ChildHandle || !ci_struct_array[1].ChildHandle)
    {
        retVal =cl_der_DeleteLeafNode(pContext,ci_struct_array[0].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,ci_struct_array[1].ChildHandle);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_ci_types, ci_struct_array, &root);

  
    if(ret != CSSM_OK)
    {
        retVal =cl_der_DeleteLeafNode(pContext,ci_struct_array[0].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,ci_struct_array[1].ChildHandle);
        cl_der_DeleteTree (pContext,root);
        return ret;
        
    }

    /* Pack the signature struct into a DER encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, &results_ptr);

    if(ret != CSSM_OK)
    {
        retVal = cl_der_DeleteLeafNode(pContext,ci_struct_array[0].ChildHandle);
        retVal = cl_der_DeleteLeafNode(pContext,ci_struct_array[1].ChildHandle);
        cl_der_DeleteTree (pContext,root);
        return ret;
        
    }

    /* Delete the parse tree */
    cl_der_DeleteTree (pContext,root);

    /* Return the result */
    
    *result_val = results_ptr;
    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_DerEncodePkcs7SignedData
 *
 * Description:
 * This function returns the DER encoding of the input PKCS #7 Signed Data
 * C structure
 *
 * Parameters: 
 * pkcs7SignedData (input) : The C representation of the PKCS #7 signed data
 *
 * result_val (output): A DER encoded PKCS #7 signed data structure
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_PKCS7_DerEncodeSignedData (CSSM_CLUTIL_CONTEXT *pContext,
                                                   CSSM_HANDLE AppHandle, 
                                                   const cssm_PKCS7_SIGNED_DATA *pkcs7SignedData,
                                                   CSSM_DATA_PTR *result_val)
{
    CSSM_DATA_PTR       enc_pkcs7_ptr = NULL,
                        results_ptr = NULL;
    uint32              num_pkcs7_types = 6;
    CL_DER_CHILD_INFO   pkcs7_struct_array[6];
    uint8               pkcs7_version_data = 1;
    CSSM_DATA           pkcs7_version;
    CSSM_DATA_PTR       enc_algs_ptr = NULL,
                        enc_content_info_ptr = NULL,
                        enc_certs_ptr = NULL,
                        enc_crls_ptr = NULL,
                        enc_signer_infos = NULL;
    CL_DER_NODE_HANDLE  root = 0;
    CSSM_RETURN         ret = CSSM_OK,
                        ret1 = CSSM_OK, ret2 = CSSM_OK, 
                        ret3 = CSSM_OK, ret4 = CSSM_OK, 
                        ret5 = CSSM_OK, ret6 = CSSM_OK,
                        retVal = CSSM_OK;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    if ( CSSM_CLUTIL_PKCS7_IsBadSignedDataPtr(pkcs7SignedData)) 
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }
    if ( !pkcs7SignedData )
    {
        
        return cssm_clutil_EncodeNullSequence(pContext,AppHandle, result_val);
    }

    /* Initializations */
    port_memset(pkcs7_struct_array, 0, num_pkcs7_types*sizeof(CL_DER_CHILD_INFO));
    pkcs7_version.Length = sizeof(uint8);
    pkcs7_version.Data = &pkcs7_version_data;

    /* DER encode the pkcs7 components */
    cssm_clutil_pkcs7_DerEncodeAlgIdSet(pContext,AppHandle, 
                                        pkcs7SignedData->digestAlgIds, 
                                        pkcs7SignedData->numberOfDigestAlgIds,
                                        &enc_algs_ptr);

    ret = cssm_clutil_pkcs7_DerEncodeContentInfo(pContext,AppHandle, &pkcs7SignedData->contentInfo, &enc_content_info_ptr);
 
    if(ret != CSSM_OK)
    {
        cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_algs_ptr);
        return ret;
        
    }

    CSSM_CLUTIL_DerEncodeCertGroup(pContext,AppHandle, &pkcs7SignedData->certificates, &enc_certs_ptr);
    ret = CSSM_CLUTIL_DerEncodeCrlGroup(pContext,AppHandle, &pkcs7SignedData->crls, &enc_crls_ptr);
    cssm_clutil_pkcs7_DerEncodeSignerInfoSet(pContext,AppHandle, pkcs7SignedData->signerInfos, pkcs7SignedData->numberOfSignerInfos, &enc_signer_infos);
    if (!enc_algs_ptr || !enc_content_info_ptr || !enc_certs_ptr || !enc_crls_ptr || !enc_signer_infos) 
    {
        cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_algs_ptr);
        cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_content_info_ptr);
        cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_certs_ptr);
        cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_crls_ptr);
        cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_signer_infos);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    /* Create the DER encoded PKCS #7 structure */
    ret1 =  (pkcs7_version.Length != 0) ?
        cl_der_CreateLeafNode (pContext,DER_TAG_INTEGER, pkcs7_version.Length, pkcs7_version.Data, &(pkcs7_struct_array[0].ChildHandle)) :
        cl_der_CreateLeafNode (pContext,DER_TAG_INTEGER, 0, NULL, &(pkcs7_struct_array[0].ChildHandle));


    ret2 = (cl_der_LengthOfContent(enc_algs_ptr) != 0) ?
        cl_der_CreateLeafNode (pContext,DER_TAG_SET, 
            cl_der_LengthOfContent(enc_algs_ptr), 
            &enc_algs_ptr->Data[ cl_der_LengthOfTag(enc_algs_ptr) + cl_der_LengthOfLength(enc_algs_ptr)], &(pkcs7_struct_array[1].ChildHandle)) :
        cl_der_CreateLeafNode (pContext,DER_TAG_SET, 0, NULL, &(pkcs7_struct_array[1].ChildHandle));

    ret3 =  (cl_der_LengthOfContent(enc_content_info_ptr) != 0) ?
        cl_der_CreateLeafNode (pContext,DER_TAG_SEQUENCE, 
            cl_der_LengthOfContent(enc_content_info_ptr),
            &enc_content_info_ptr->Data[cl_der_LengthOfTag(enc_content_info_ptr) 
              + cl_der_LengthOfLength(enc_content_info_ptr)], &(pkcs7_struct_array[2].ChildHandle)) :
        cl_der_CreateLeafNode (pContext,DER_TAG_SEQUENCE, 0, NULL, &(pkcs7_struct_array[2].ChildHandle));

    pkcs7_struct_array[3].IsOptional = CSSM_TRUE;
    pkcs7_struct_array[3].HasCtxSpecificTag = CSSM_TRUE;
    ret4 =  (cl_der_LengthOfContent(enc_certs_ptr) != 0) ?
        cl_der_CreateLeafNode (pContext,DER_TAG_SET, 
            cl_der_LengthOfContent(enc_certs_ptr),
            &enc_certs_ptr->Data[ cl_der_LengthOfTag(enc_certs_ptr) + cl_der_LengthOfLength(enc_certs_ptr)], &(pkcs7_struct_array[3].ChildHandle)) :
        cl_der_CreateLeafNode (pContext,DER_TAG_SET, 0, NULL, &(pkcs7_struct_array[3].ChildHandle));

    pkcs7_struct_array[4].IsOptional = CSSM_TRUE;
    pkcs7_struct_array[4].HasCtxSpecificTag = CSSM_TRUE;
    ret5 =  (cl_der_LengthOfContent(enc_crls_ptr) != 0) ?
        cl_der_CreateLeafNode (pContext,DER_TAG_SET, 
            cl_der_LengthOfContent(enc_crls_ptr),
            &enc_crls_ptr->Data[ cl_der_LengthOfTag(enc_crls_ptr) + cl_der_LengthOfLength(enc_crls_ptr)], &(pkcs7_struct_array[4].ChildHandle)) :
        cl_der_CreateLeafNode (pContext,DER_TAG_SET, 0, NULL, &(pkcs7_struct_array[4].ChildHandle));

    ret6 =  (cl_der_LengthOfContent(enc_signer_infos) != 0) ?
        cl_der_CreateLeafNode (pContext,DER_TAG_SET, 
            cl_der_LengthOfContent(enc_signer_infos),
            &enc_signer_infos->Data[ cl_der_LengthOfTag(enc_signer_infos) 
              + cl_der_LengthOfLength(enc_signer_infos)], &(pkcs7_struct_array[5].ChildHandle)) :
        cl_der_CreateLeafNode (pContext,DER_TAG_SET, 0, NULL, &(pkcs7_struct_array[5].ChildHandle));

    cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_algs_ptr);
    cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_content_info_ptr);
    cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_certs_ptr);
    cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_crls_ptr);
    cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_signer_infos);
    
    if ((ret1 != CSSM_OK) || (ret2 != CSSM_OK) ||
        (ret3 != CSSM_OK) || (ret4 != CSSM_OK) ||
        (ret5 != CSSM_OK) || (ret6 != CSSM_OK))
    {
        retVal = cl_der_DeleteLeafNode(pContext,pkcs7_struct_array[0].ChildHandle);
        retVal = cl_der_DeleteLeafNode(pContext,pkcs7_struct_array[1].ChildHandle);
        retVal = cl_der_DeleteLeafNode(pContext,pkcs7_struct_array[2].ChildHandle);
        retVal = cl_der_DeleteLeafNode(pContext,pkcs7_struct_array[3].ChildHandle);
        retVal = cl_der_DeleteLeafNode(pContext,pkcs7_struct_array[4].ChildHandle);
        retVal = cl_der_DeleteLeafNode(pContext,pkcs7_struct_array[5].ChildHandle);

        /* set proper error code */
        if(CSSM_OK != ret1)
        {
            return ret1;
        }
        if(CSSM_OK != ret2)
        {
            return ret2;
        }
        if(CSSM_OK != ret3)
        {
            return ret3;
        }
        if(CSSM_OK != ret4)
        {
            return ret4;
        }
        if(CSSM_OK != ret5)
        {
            return ret5;
        }
        if(CSSM_OK != ret6)
        {
            return ret6;
        }
        
    }

    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_pkcs7_types, pkcs7_struct_array, &root);

    /* Pack the signature struct into a DER encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, &enc_pkcs7_ptr);

    /* Delete the parse tree */
    cl_der_DeleteTree (pContext,root);

    /* Return the result */

    ret = cssm_clutil_PKCS7_DerEncodeContentInfo(pContext,AppHandle, &CSSMOID_Pkcs7SignedData, enc_pkcs7_ptr, &results_ptr);
    cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_pkcs7_ptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    *result_val = results_ptr;
    return CSSM_OK;

}


/*-----------------------------------------------------------------------------
 * Name: cssm_clutil_ConvertEndian
 *
 * Description:
 * This function returns a buffer containing the reverse byte ordering 
 * from the input buffer
 *
 * Parameters: 
 * original   (input) : A valid buffer 
 * byteLength (input) : The length of the input buffer, also the length of the output buffer
 *
 * result_val (output):
 * A buffer of the same length as the original input, 
 * containing the reverse byte ordering
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN cssm_clutil_ConvertEndian(const CSSM_CLUTIL_CONTEXT *pContext,
                                  CSSM_HANDLE AppHandle, uint8 *original, uint32 byteLength,
                                  uint8 **result_val)
{
    uint8 *new_buf = NULL;
    uint8 *i = NULL, *j = NULL;

    /* allocate enough space for the new byte array */
    new_buf = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, byteLength);
    if (new_buf == NULL) 
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Copy the reverse contents of the original buffer into the new buffer */
    for (i = new_buf, j = &original[byteLength-1]; i < new_buf + byteLength; i++, j--) 
    {
        *i = *j;
    }

    *result_val = new_buf;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_MicrosoftRSAKeyToDerRSAKey
 *
 * Description:
 * The byte ordering of the RSA keys generated by the Microsoft base CSP is
 * that of the system the key was generated on.  DER requires that the
 * byte ordering be Big Endian, which means that the Microsoft RSA keys need
 * to be converted if they are generated on IA platforms.
 *
 * Parameters: 
 * MicrosoftRSAPublicKey (input) : The CSSM_KEY representation of a Microsoft RSA key
 *
 * result_val (output): A CSSM_KEY whose key header is the same as the input, 
 *                      but whose key data has the reverse byte ordering from 
 *                      the input key data
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_MicrosoftRSAKeyToDerRSAKey(CSSM_CLUTIL_CONTEXT *pContext,
                                                    CSSM_HANDLE  AppHandle,
                                                    CSSM_KEY_PTR MicrosoftRSAPublicKey,
                                                    CSSM_KEY_PTR *result_val)
{
    CSSM_KEY_PTR         new_key = NULL;
    CSSM_DATA_PTR        key_data_ptr = NULL;
    CSSM_RETURN          ret = CSSM_OK, ret1 = CSSM_OK, 
                         ret2 = CSSM_OK, retVal = CSSM_OK;
    CL_DER_LEAF_NODE_PTR obj_array = NULL;
    uint32               num_obj = 0;
    CSSM_DATA            new_mod = {0, NULL},
                         new_exp = {0, NULL};

   
    uint32              num_types = 2;
    CL_DER_CHILD_INFO   key_info_struct[2];
    CL_DER_NODE_HANDLE  root = 0;

    uint32 i=0;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    /* Get pointers to the Der-encoded modulus and exponent */
    ret = CSSM_CLUTIL_DerDecodeOneLevel(pContext,AppHandle, &MicrosoftRSAPublicKey->KeyData, &obj_array, &num_obj);
    if (CSSM_OK != ret)
    {
        return ret;
        
    }
    if (num_obj != 2 ||
        obj_array[0].Tag != DER_TAG_INTEGER ||
        obj_array[1].Tag != DER_TAG_INTEGER )
    {
        for (i=0; i < num_obj; i++)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, obj_array[i].Data.Data);
        }
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, obj_array);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    /* Determine the modulus length without the leading zeros */
    for (i=0; i < obj_array[0].Data.Length; i++)
    {
        if (obj_array[0].Data.Data[i] != 0x00) 
        {
            break;
        }
    }
    /* For the modulus, reverse the byte ordering, excluding the leading zeros */
    new_mod.Length = obj_array[0].Data.Length - i;
    cssm_clutil_ConvertEndian(pContext,AppHandle, &obj_array[0].Data.Data[i], new_mod.Length, &(new_mod.Data));
    if (!new_mod.Data)
    {
        for (i=0; i < num_obj; i++)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, obj_array[i].Data.Data);
        }
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, obj_array);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    /* For the exponent, reverse the byte ordering */
    new_exp.Length = obj_array[1].Data.Length;
    cssm_clutil_ConvertEndian(pContext,AppHandle, obj_array[1].Data.Data, new_exp.Length, &(new_exp.Data));
    for (i=0; i < num_obj; i++)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, obj_array[i].Data.Data);
    }
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, obj_array);
    if (!new_exp.Data)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, new_mod.Data);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    /* Re-encode the RSA key */
    port_memset(key_info_struct, 0, num_types*sizeof(CL_DER_CHILD_INFO));
    ret1 =  
        cl_der_CreateLeafNode (pContext,DER_TAG_INTEGER, new_mod.Length, new_mod.Data, &(key_info_struct[0].ChildHandle));

    
    ret2 =  
        cl_der_CreateLeafNode (pContext,DER_TAG_INTEGER, new_exp.Length, new_exp.Data, &(key_info_struct[1].ChildHandle));

    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, new_mod.Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, new_exp.Data);

    if ((ret1 != CSSM_OK) || (ret2 != CSSM_OK) )
    {
        retVal = cl_der_DeleteLeafNode(pContext,key_info_struct[0].ChildHandle);
        retVal = cl_der_DeleteLeafNode(pContext,key_info_struct[1].ChildHandle);
        if(ret1 != CSSM_OK)
        {
            return ret1;
        }
        if(ret2 != CSSM_OK)
        {
            return ret2;
        }
        
    }
    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_types, key_info_struct, &root);
    
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, &key_data_ptr);
    cl_der_DeleteTree (pContext,root);

    /* Allocate the new key and copy the old KeyHeader & new key data into it */
    new_key = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(CSSM_KEY));
    if (!new_key)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, key_data_ptr->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, key_data_ptr);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    port_memcpy(&new_key->KeyHeader, &MicrosoftRSAPublicKey->KeyHeader, sizeof(CSSM_KEYHEADER));
    new_key->KeyData = *key_data_ptr;
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, key_data_ptr);

    
    *result_val = new_key;
    return CSSM_OK;

    /* Failed attempts to decode Microsoft RSA key format
    Leave key and signature alone
    Leave key, flip signature
    Flip  key, leave signature (hangs)
    flip key, flip signature (hangs)
    Removing leading zero on modulus, leave signature alone
    Removing leading zero on modulus, flip signature
    Removing leading zero on modulus, flip key, flip signature
    */
    /* Successful attempt
    Removing leading zero on modulus, flip key, leave signature alone
    */
}



/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_DerRSAKeyToMicrosoftRSAKey
 *
 * Description:
 * The byte ordering of the RSA keys generated by the Microsoft base CSP is
 * that of the system the key was generated on.  DER requires that the
 * byte ordering be Big Endian, which means that the DER keys need
 * to be converted if they are to be used by Microsoft CSPs.
 *
 * Parameters: 
 * DerEncodedRSAPublicKey (input) : The CSSM_KEY representation of a 
 *                                  properly DER encoded RSA key
 *
 * result_val (output):
 * A CSSM_KEY whose key header is the same as the input, but whose key data has
 * the reverse byte ordering from the input key data
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_DerRSAKeyToMicrosoftRSAKey (CSSM_CLUTIL_CONTEXT *pContext,
                                                    CSSM_HANDLE  AppHandle,
                                                    CSSM_KEY_PTR DerEncodedRSAPublicKey,
                                                    CSSM_KEY_PTR *result_val)
{
    CSSM_KEY_PTR         new_key = NULL;
    CSSM_DATA_PTR        key_data_ptr = NULL;
    CSSM_RETURN         ret = CSSM_OK, ret1 = CSSM_OK, 
                        ret2 = CSSM_OK, retVal = CSSM_OK;
    CL_DER_LEAF_NODE_PTR obj_array = NULL;
    uint32               num_obj = 0;
    CSSM_DATA            new_mod = {0, NULL},
                         new_exp = {0, NULL};
    uint8               *temp_data = NULL;

    
    uint32              num_types = 2;
    CL_DER_CHILD_INFO   key_info_struct[2];
    CL_DER_NODE_HANDLE  root = 0;

    uint32 i=0;
    *result_val = NULL;
    /* Get pointers to the Der-encoded modulus and exponent */
    if (CSSM_CLUTIL_DerDecodeOneLevel(pContext,AppHandle, &DerEncodedRSAPublicKey->KeyData, &obj_array, &num_obj) != CSSM_OK)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }
    if (num_obj != 2 ||
        obj_array[0].Tag != DER_TAG_INTEGER ||
        obj_array[1].Tag != DER_TAG_INTEGER )
    {
        for (i=0; i < num_obj; i++)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, obj_array[i].Data.Data);
        }
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, obj_array);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    /* For the modulus, reverse the byte ordering, and add one leading zero byte */
    new_mod.Length = obj_array[0].Data.Length +1;
    new_mod.Data   = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, new_mod.Length);
    cssm_clutil_ConvertEndian(pContext,AppHandle, obj_array[0].Data.Data, obj_array[0].Data.Length, &temp_data);
    if (!temp_data || !new_mod.Data)
    {
        for (i=0; i < num_obj; i++)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, obj_array[i].Data.Data);
        }
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, obj_array);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, new_mod.Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_data);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }
    port_memcpy(&new_mod.Data[1], temp_data, obj_array[0].Data.Length);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_data);


    /* For the exponent, reverse the byte ordering */
    new_exp.Length = obj_array[1].Data.Length;
    cssm_clutil_ConvertEndian(pContext,AppHandle, obj_array[1].Data.Data, new_exp.Length, &(new_exp.Data));
    for (i=0; i < num_obj; i++)
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, obj_array[i].Data.Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, obj_array);
    if (!new_exp.Data)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, new_mod.Data);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    /* Re-encode the RSA key */
    port_memset(key_info_struct, 0, num_types*sizeof(CL_DER_CHILD_INFO));
    ret1 =  
        cl_der_CreateLeafNode (pContext,DER_TAG_INTEGER, new_mod.Length, new_mod.Data, &(key_info_struct[0].ChildHandle));
    ret2 =  
        cl_der_CreateLeafNode (pContext,DER_TAG_INTEGER, new_exp.Length, new_exp.Data, &(key_info_struct[1].ChildHandle));

    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, new_mod.Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, new_exp.Data);

    if ((ret1 != CSSM_OK) || (ret2 != CSSM_OK) )
    {
        retVal = cl_der_DeleteLeafNode(pContext,key_info_struct[0].ChildHandle);
        retVal = cl_der_DeleteLeafNode(pContext,key_info_struct[1].ChildHandle);
        if(ret1 != CSSM_OK)
        {
            return ret1;
        }
        if(ret2 != CSSM_OK)
        {
            return ret2;
        }
        
    }
    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_types, key_info_struct, &root);
 
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, &key_data_ptr);

    if(ret != CSSM_OK)
    {
        return ret;
        
    }

    cl_der_DeleteTree (pContext,root);

    /* Allocate the new key and copy the old KeyHeader & new key data into it */
    new_key = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(CSSM_KEY));
    if (!new_key)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, key_data_ptr->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, key_data_ptr);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    port_memcpy(&new_key->KeyHeader, &DerEncodedRSAPublicKey->KeyHeader, sizeof(CSSM_KEYHEADER));
    new_key->KeyData = *key_data_ptr;
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, key_data_ptr);

    
    *result_val = new_key;
    return CSSM_OK;

}



/*-----------------------------------------------------------------------------
 * Name: cssm_clutil_pkcs7_DerEncodeAlgIdSet
 *
 * Description:
 * This function DER encodes the input array of AlgIds
 *
 * Parameters: 
 * AlgIds         (input) : The array of AlgIds to be DEr encoded
 * NumberOfAlgIds (input) : The number of elements in the AlgIds array
 *
 * result_val(output):
 * The DER encoding of the input AlgId array
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN cssm_clutil_pkcs7_DerEncodeAlgIdSet(CSSM_CLUTIL_CONTEXT *pContext,
                                                CSSM_HANDLE AppHandle, 
                                                CSSM_X509_ALGORITHM_IDENTIFIER_PTR AlgIds, 
                                                uint32 NumberOfAlgIds,
                                                CSSM_DATA_PTR *result_val)
{
    CSSM_DATA_PTR       results_ptr = NULL;
    CSSM_DATA_PTR       *enc_algid_array = NULL;
    uint32              i=0, j=0;
    CSSM_RETURN         ret = CSSM_OK;

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    /* If there are no algIds, return the encoding for a NULL SET */
    if (!AlgIds && !NumberOfAlgIds)
    {
        return cssm_clutil_EncodeNullSet(pContext,AppHandle, result_val);
    }
    if (!AlgIds || !NumberOfAlgIds)
    {
        return CSSMERR_CL_INVALID_DATA;   
    }

    /* Initializations */
    enc_algid_array = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, NumberOfAlgIds, sizeof(CSSM_DATA_PTR) );
    if (!enc_algid_array)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* For each algid */
    for (i=0; i < NumberOfAlgIds; i++)
    {
        /* Der encode the individual algid */
        ret = CSSM_CLUTIL_X509_DerEncodeAlgorithmId(pContext,AppHandle, &AlgIds[i], &enc_algid_array[i]);

        if (!enc_algid_array[i])
        {
            for (j=0; j<i; j++)
            {
                cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_algid_array[j]);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_algid_array[j]);
            return ret;
        }
    }

    /* Der encode the array of algids */
    ret = cssm_clutil_DerEncodeArrayAsSetOrSeq(pContext,
                                                AppHandle, 
                                                enc_algid_array, 
                                                NumberOfAlgIds, 
                                                DER_TAG_SET,
                                                &results_ptr);

    for (i=0; i < NumberOfAlgIds; i++) 
    {
        cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_algid_array[i]);
    }
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_algid_array);

    /* Return the result */
    if (CSSM_OK != ret)
    {
        return ret;
    }
    *result_val = results_ptr;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: cssm_clutil_pkcs7_DerEncodeContentInfo
 *
 * Description:
 * This function DER encodes t3he input content info
 *
 * Parameters: 
 * ContentInfoPtr (input) : The C structure representation of 
 *                          the content info to be encoded
 * 
 * results_ptr(output):
 * The DER encoding of the input Content Info
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN cssm_clutil_pkcs7_DerEncodeContentInfo(CSSM_CLUTIL_CONTEXT *pContext,
                                                     CSSM_HANDLE AppHandle, 
                                                     const cssm_PKSC7_CONTENT_INFO *ContentInfoPtr,
                                                     CSSM_DATA_PTR *results_ptr)
{

    if (!results_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *results_ptr = NULL;
    return CSSM_CLUTIL_X509_DerEncodeAlgorithmId(pContext,AppHandle, ContentInfoPtr, results_ptr);
}


/*-----------------------------------------------------------------------------
 * Name: cssm_clutil_pkcs7_DerEncodeSignerInfoSet
 *
 * Description:
 * This function DER encodes the input array of SignerInfo C structures
 *
 * Parameters: 
 * SignerInfos         (input) : The SignerInfo C structures to be encoded
 * NumberOfSignerInfos (input) : The number of element in the SignerInfos array
 *
 * result_val (output):
 * The DER encoding of the input SignerInfo array
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN cssm_clutil_pkcs7_DerEncodeSignerInfoSet(CSSM_CLUTIL_CONTEXT *pContext,
                                                     CSSM_HANDLE AppHandle, 
                                                     cssm_PKCS7_SIGNER_INFO_PTR SignerInfos, 
                                                     uint32 NumberOfSignerInfos,
                                                     CSSM_DATA_PTR *result_val)
{
    CSSM_DATA_PTR       results_ptr = NULL;
    CSSM_DATA_PTR       *enc_siginfo_array = NULL;
    uint32              i=0, j=0;
    CSSM_RETURN         ret = CSSM_OK;

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    /* If there are no certs, return the encoding for a NULL SET */
    if (!SignerInfos && !NumberOfSignerInfos)
    {
        
        return cssm_clutil_EncodeNullSet(pContext,AppHandle, result_val);
    }
    if (!SignerInfos || !NumberOfSignerInfos)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* Initializations */
    enc_siginfo_array = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, NumberOfSignerInfos, sizeof(CSSM_DATA_PTR) );
    if (!enc_siginfo_array)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* For each algid */
    for (i=0; i < NumberOfSignerInfos; i++)
    {
        /* Der encode the individual algid */
        ret = cssm_clutil_pkcs7_DerEncodeSignerInfo(pContext,AppHandle, &SignerInfos[i], &(enc_siginfo_array[i]));

        if (CSSM_OK != ret)
        {
            for (j=0; j<i; j++)
            {
                cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_siginfo_array[j]);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_siginfo_array[j]);
            return ret;
        }
    }

    /* Der encode the array of SignerInfos */
    ret = cssm_clutil_DerEncodeArrayAsSetOrSeq(pContext,
                                                AppHandle, 
                                                enc_siginfo_array, 
                                                NumberOfSignerInfos, 
                                                DER_TAG_SET,
                                                &results_ptr);

    for (i=0; i < NumberOfSignerInfos; i++)
    {
        cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_siginfo_array[i]);
    }
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_siginfo_array[i]);

    /* Return the result */
    if (CSSM_OK != ret)
    {
        return ret;
    }

    *result_val = results_ptr;
    return CSSM_OK;
}



/*-----------------------------------------------------------------------------
 * Name: cssm_clutil_pkcs7_DerEncodeSignerInfo
 *
 * Description:
 * This function DER encodes the input SignerInfo C structures
 *
 * Parameters: 
 * SignerInfo (input) : The SignerInfo C structure to be encoded
 *
 * result_val(output):
 * The DER encoding of the input SignerInfo struct
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN cssm_clutil_pkcs7_DerEncodeSignerInfo (CSSM_CLUTIL_CONTEXT *pContext,
                                                   CSSM_HANDLE AppHandle, 
                                                   cssm_PKCS7_SIGNER_INFO_PTR SignerInfo,
                                                   CSSM_DATA_PTR *result_val)
{
    CSSM_DATA_PTR       results_ptr = NULL;
    uint32              num_types = 7;
    CL_DER_CHILD_INFO   siginfo_struct_array[7];
    uint8               version_data;
    CSSM_DATA           version;
    CSSM_DATA_PTR       enc_iss_and_snum, 
                        enc_digest_algid, enc_encr_algid,
                        enc_auth_attrs, enc_unauth_attrs;
    CL_DER_NODE_HANDLE  root;
    uint32              i;
    CSSM_RETURN         ret = CSSM_OK, ret1 = CSSM_OK, 
                        ret2 = CSSM_OK, ret3 = CSSM_OK, 
                        ret4 = CSSM_OK,ret5 = CSSM_OK, 
                        ret6 = CSSM_OK, ret7 = CSSM_OK,
                        retVal = CSSM_OK;
    /* Initializations */
    port_memset(siginfo_struct_array, 0, num_types*sizeof(CL_DER_CHILD_INFO));
    version_data = 0x01;
    version.Length = 1;
    version.Data   = &version_data;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    /* Create the DER encoded structs */
    cssm_clutil_pkcs7_DerEncodeIssuerAndSNum(pContext,AppHandle, &SignerInfo->issuerAndSerialNumber, &enc_iss_and_snum);
    ret = CSSM_CLUTIL_X509_DerEncodeAlgorithmId(pContext,AppHandle, &SignerInfo->digestAlgId, &enc_digest_algid); 

    if(ret != CSSM_OK)
    {
         cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_iss_and_snum);
         return ret;
         
    }
    CSSM_CLUTIL_X501_DerEncodeAttributes(pContext,AppHandle, &SignerInfo->authenticatedAttrs, &enc_auth_attrs); 

    ret   = CSSM_CLUTIL_X509_DerEncodeAlgorithmId(pContext,AppHandle, &SignerInfo->digestEncryptionAlgId, &enc_encr_algid); 

    if(ret != CSSM_OK)
    {
         cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_iss_and_snum);
         cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_auth_attrs);
         cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_digest_algid);
         return ret;
         
    }
    
    CSSM_CLUTIL_X501_DerEncodeAttributes(pContext,AppHandle, &SignerInfo->unauthenticatedAttrs, &enc_unauth_attrs);

    if (!enc_iss_and_snum || !enc_digest_algid || !enc_auth_attrs || !enc_encr_algid || !enc_unauth_attrs)
    {
        cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_iss_and_snum);
        cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_digest_algid);
        cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_auth_attrs);
        cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_encr_algid);
        cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_unauth_attrs);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }


    /* Create the DER encoded X509 structure */
    ret1 =  
        cl_der_CreateLeafNode (pContext,DER_TAG_INTEGER, version.Length, version.Data, &(siginfo_struct_array[0].ChildHandle));
    ret2 =  
        cl_der_CreateLeafNode (pContext,DER_TAG_SEQUENCE, 
            cl_der_LengthOfContent(enc_iss_and_snum),
            &enc_iss_and_snum->Data[cl_der_LengthOfTag(enc_iss_and_snum) + cl_der_LengthOfLength(enc_iss_and_snum)], &(siginfo_struct_array[1].ChildHandle));        
    ret3 =  
        cl_der_CreateLeafNode (pContext,DER_TAG_SEQUENCE, 
            cl_der_LengthOfContent(enc_digest_algid),
            &enc_digest_algid->Data[cl_der_LengthOfTag(enc_digest_algid) + cl_der_LengthOfLength(enc_digest_algid)], &(siginfo_struct_array[2].ChildHandle));        

    siginfo_struct_array[3].IsOptional = CSSM_TRUE;
    siginfo_struct_array[3].HasCtxSpecificTag = CSSM_TRUE;
    ret4 =  
       cl_der_CreateLeafNode (pContext,DER_TAG_SET, 
            cl_der_LengthOfContent(enc_auth_attrs),
            &enc_auth_attrs->Data[ cl_der_LengthOfTag(enc_auth_attrs) + cl_der_LengthOfLength(enc_auth_attrs)], &(siginfo_struct_array[3].ChildHandle));
    ret5 =  
        cl_der_CreateLeafNode (pContext,DER_TAG_SEQUENCE, 
            cl_der_LengthOfContent(enc_encr_algid),
            &enc_encr_algid->Data[ cl_der_LengthOfTag(enc_encr_algid) + cl_der_LengthOfLength(enc_encr_algid)], &(siginfo_struct_array[4].ChildHandle));
    ret6 =  
        cl_der_CreateLeafNode(pContext,DER_TAG_OCTET_STRING, 
            SignerInfo->encryptedDigest.Length, SignerInfo->encryptedDigest.Data, &(siginfo_struct_array[5].ChildHandle));
    siginfo_struct_array[6].IsOptional = CSSM_TRUE;
    siginfo_struct_array[6].HasCtxSpecificTag = CSSM_TRUE;
    ret7 =  
        cl_der_CreateLeafNode (pContext,DER_TAG_SET, 
            cl_der_LengthOfContent(enc_unauth_attrs),
            &enc_unauth_attrs->Data[ cl_der_LengthOfTag(enc_unauth_attrs) + cl_der_LengthOfLength(enc_unauth_attrs)], &(siginfo_struct_array[6].ChildHandle));

    cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_iss_and_snum);
    cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_digest_algid);
    cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_auth_attrs);
    cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_encr_algid);
    cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_unauth_attrs);

    if ((ret1 != CSSM_OK) || (ret2 != CSSM_OK) ||
        (ret3 != CSSM_OK) || (ret4 != CSSM_OK) ||
        (ret5 != CSSM_OK) || (ret6 != CSSM_OK) ||
        (ret7 != CSSM_OK) )
    {
        for (i=0; i<=num_types; i++)
        {
            retVal = cl_der_DeleteLeafNode(pContext,siginfo_struct_array[i].ChildHandle);
        }
        if(CSSM_OK != ret1)
        {
            return ret1;
        }
        if(CSSM_OK != ret2)
        {
            return ret2;
        }
        if(CSSM_OK != ret3)
        {
            return ret3;
        }
        if(CSSM_OK != ret4)
        {
            return ret4;
        }
        if(CSSM_OK != ret5)
        {
            return ret5;
        }
        if(CSSM_OK != ret6)
        {
            return ret6;
        }
        if(CSSM_OK != ret7)
        {
            return ret7;
        }

    }

    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_types, siginfo_struct_array, &root);


    /* Pack the signature struct into a DER encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, &results_ptr);

    if(ret != CSSM_OK)
    {
        cl_der_DeleteTree (pContext,root);
        return ret;
        
    }

    /* Delete the parse tree */
    cl_der_DeleteTree (pContext,root);

    /* Return the result */
    
    *result_val = results_ptr;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: cssm_clutil_pkcs7_DerEncodeIssuerAndSNum
 *
 * Description:
 * This function DER encodes the Issuer and Serial Number C structure
 *
 * Parameters: 
 * IssuerAndSNumPtr (input) The issuer and serial number C structure
 * 
 * result_val(output):
 * The DER encoded Issuer and Serial Number C structure
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN cssm_clutil_pkcs7_DerEncodeIssuerAndSNum(CSSM_CLUTIL_CONTEXT *pContext,
                                                     CSSM_HANDLE AppHandle, 
                                                     cssm_PKCS7_ISSUER_AND_SERIAL_NUMBER_PTR IssuerAndSNumPtr,
                                                     CSSM_DATA_PTR *result_val)
{
    CSSM_DATA_PTR       results_ptr = NULL;
    uint32              num_types = 2;
    CL_DER_CHILD_INFO   iname_snum_struct_array[3];
    CSSM_DATA_PTR       enc_iname_ptr = NULL;
    CL_DER_NODE_HANDLE  root = 0;
    CSSM_RETURN         ret = CSSM_OK, ret1 = CSSM_OK, ret2 = CSSM_OK, retVal = CSSM_OK;
    /* Initializations */
    port_memset(iname_snum_struct_array, 0, num_types*sizeof(CL_DER_CHILD_INFO));
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    /* DER encode the components */
    ret = CSSM_CLUTIL_X509_DerEncodeName(pContext,AppHandle, &IssuerAndSNumPtr->issuer, &enc_iname_ptr);
    
    if (ret != CSSM_OK)
    {
        return ret;
        
    }
    /* Create the DER encoded object */
    ret1 =  
        cl_der_CreateLeafNode (pContext,DER_TAG_SEQUENCE, 
            cl_der_LengthOfContent(enc_iname_ptr),
            &enc_iname_ptr->Data[cl_der_LengthOfTag(enc_iname_ptr) + cl_der_LengthOfLength(enc_iname_ptr)], &(iname_snum_struct_array[0].ChildHandle));
    ret2 =  
        cl_der_CreateLeafNode (pContext,DER_TAG_INTEGER, 
            IssuerAndSNumPtr->serialNumber.Length, 
            IssuerAndSNumPtr->serialNumber.Data, &(iname_snum_struct_array[1].ChildHandle));

    cssm_clutil_freeCssmDataPtr(pContext,AppHandle, enc_iname_ptr);
    
    if ((ret1 != CSSM_OK) || (ret2 != CSSM_OK) )
    {
        retVal = cl_der_DeleteLeafNode(pContext,iname_snum_struct_array[0].ChildHandle);
        retVal = cl_der_DeleteLeafNode(pContext,iname_snum_struct_array[1].ChildHandle);
        if(ret1 != CSSM_OK)
        {
            return ret1;
        }
        if(ret2 != CSSM_OK)
        {
            return ret2;
        }
        
    }

    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_types, iname_snum_struct_array, &root);


    /* Pack the signature struct into a DER encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, &results_ptr);

    if(ret != CSSM_OK)
    {
        cl_der_DeleteTree (pContext,root);
        return ret;
        
    }

    /* Delete the parse tree */
    cl_der_DeleteTree (pContext,root);

    /* Return the result */
    *result_val = results_ptr;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: cssm_clutil_EncodeNullSequence
 *
 * Description:
 * This function returns the DER encoding of a sequence with no contents
 *
 * Parameters: 
 * 
 * 
 * result_val (output):
 * 0x30 0x00
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN   cssm_clutil_EncodeNullSequence  (const CSSM_CLUTIL_CONTEXT *pContext,
                                               CSSM_HANDLE AppHandle,
                                               CSSM_DATA_PTR *result_val)
{
    uint8      enc_null_seq_data[2];
    CSSM_DATA  enc_null_seq;

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    enc_null_seq_data[0] = 0x30;
    enc_null_seq_data[1] = 0x00;
    enc_null_seq.Length = 2;
    enc_null_seq.Data   = enc_null_seq_data;

    return cssm_clutil_AllocAndCopyCssmData(pContext,AppHandle, &enc_null_seq, result_val);
}


/*-----------------------------------------------------------------------------
 * Name: cssm_clutil_EncodeNullSet
 *
 * Description:
 * This function returns the DER encoding of a set with no contents
 *
 * Parameters: 
 * 
 * 
 * result_val (output):
 * 0x31 0x00
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN   cssm_clutil_EncodeNullSet (const CSSM_CLUTIL_CONTEXT *pContext,
                                         CSSM_HANDLE AppHandle,
                                         CSSM_DATA_PTR *result_val)
{
    uint8      enc_null_set_data[2];
    CSSM_DATA  enc_null_set;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    enc_null_set_data[0] = 0x31;
    enc_null_set_data[1] = 0x00;
    enc_null_set.Length = 2;
    enc_null_set.Data   = enc_null_set_data;

    return cssm_clutil_AllocAndCopyCssmData(pContext,AppHandle, &enc_null_set, result_val);
}
