/*-----------------------------------------------------------------------
 * File: CRTGETKE.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */


#include "maf_include.h"
    /* Windows & CSSM includes */
#ifndef VMS
#pragma warning (disable:4201 4214 4115)
#include <stdio.h>
#pragma warning (default:4201 4214 4115)
#else
#include <stdio.h>
#endif
#include "cssm.h"
#include "cssmspi.h"
#include "guids.h"
#include "cssmport.h"
#include "cssmutil.h"

    /* Certificate-specific includes */
#include "x509defs.h"
#include "x_fndefs.h"

#include "oidsx509.h"
#include "oidsalg.h"
#include "x509hlpr.h"
#include "clutil.h"
#include "x_qtable.h"
#include "internal.h"

#include "api_shim.h"
#include "maf_include.h"

# include "clberder.h"

#include <cssmMemTrack.h>

#define X509_FORMAT_VERSION 2

extern MEMORY_POOL_FUNCS memory_pool_funcs;
extern const CSSM_API_MEMORY_FUNCS Addin_APIMemFuncs;

CSSM_BOOL CSSMAPI
X509_IsEqualOid(const CSSM_OID *Oid1, const CSSM_OID *Oid2);
/*-----------------------------------------------------------------------------
 * Name: CL_CertGetKeyInfo
 *
 * Description:
 * This function obtains information about the certificates public key. 
 * This information comprises the key fields required to 
 * create a cryptographic context.
 *
 * Parameters: 
 * Cert (input)           : The certificate from which to extract 
 *                          the public key information. 
 *
 * result_val (output): The public key information
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSMCLI CL_CertGetKeyInfo (CSSM_CL_HANDLE CLHandle,
                                        const CSSM_DATA *Cert,
                                        CSSM_KEY_PTR *result_val)
{
    /* Initializations */
    CSSM_X509_SIGNED_CERTIFICATE_PTR cert_ptr = NULL;
    CSSM_KEY_PTR        key_ptr = NULL;

    CSSM_RETURN         ret = CSSM_OK;
    MEMORY_POOL pool;
    CSSM_DATA temp_params;

    /* Check inputs */
    /* CLHandle is checked by CSSM */
    /* Check that Cert is a valid data pointer */
    
    
    if ((Cert == NULL) || 
        port_IsBadReadPtr (Cert, sizeof(CSSM_DATA)) ) 
    {
        ret = CSSMERR_CL_INVALID_CERT_POINTER;
        return ret;
    }
    if (port_IsBadReadPtr (Cert->Data, Cert->Length) ||
        !Cert->Data || 
        !Cert->Length) 
    {
        ret = CSSMERR_CL_INVALID_CERT_POINTER;
        return ret;
    }
    if (port_IsBadWritePtr (result_val, sizeof(CSSM_KEY_PTR)))
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *result_val = NULL;

    /* Setup the Memory pool */
    ret = (*memory_pool_funcs.init)((CSSM_MEMORY_FUNCS_PTR)&Addin_APIMemFuncs, NULL, &pool);
    if (CSSM_OK != ret)
    {
        return ret;
    }
  
    /* Unpack the Cert */
    ret = MP_DerDecodeSignedCertificate(&pool, Cert, &cert_ptr);

    if(ret != CSSM_OK)
    {
        if (CSSMERR_CL_MEMORY_ERROR != ret)
        {
            ret = CSSMERR_CL_UNKNOWN_FORMAT;
        }
        (*memory_pool_funcs.recycle)(&pool);
        return ret;
    }
    
    
    /* Check that Cert is a CSSM certificate */
    if (CSSM_CLUTIL_X509_IsBadCertPtr(cert_ptr))
    {
        (*memory_pool_funcs.recycle)(&pool);
        return  CSSMERR_CL_UNKNOWN_FORMAT;
        
    }

    /* Check that it has a key */
    if (!cert_ptr->certificate.subjectPublicKeyInfo.subjectPublicKey.Data)
    {
        (*memory_pool_funcs.recycle)(&pool);
        return CSSMERR_CL_FUNCTION_FAILED;
        
    }

    if (X509_IsEqualOid(&cert_ptr->certificate.subjectPublicKeyInfo.algorithm.algorithm,
                        &CSSMOID_X9CM_DSA) &&
        ((!cert_ptr->certificate.subjectPublicKeyInfo.algorithm.parameters.Data ||
            !cert_ptr->certificate.subjectPublicKeyInfo.algorithm.parameters.Data)) &&
        (X509_IsEqualOid(&cert_ptr->signature.algorithmIdentifier.algorithm,
                        &CSSMOID_SHA1WithDSA) ||
            X509_IsEqualOid(&cert_ptr->signature.algorithmIdentifier.algorithm,
                        &CSSMOID_X9CM_DSAWithSHA1))
                        )
    {
        /* If DA key and NO params and Issuer used DSA, use Issuer's params instead*/
        temp_params = cert_ptr->certificate.subjectPublicKeyInfo.algorithm.parameters;
        cert_ptr->certificate.subjectPublicKeyInfo.algorithm.parameters = 
            cert_ptr->signature.algorithmIdentifier.parameters;
        ret = MP_SpkiToCSSMKey(&pool, &cert_ptr->certificate.subjectPublicKeyInfo, &key_ptr);
        cert_ptr->certificate.subjectPublicKeyInfo.algorithm.parameters = temp_params;
    }
    else
    {

        ret = MP_SpkiToCSSMKey(&pool, &cert_ptr->certificate.subjectPublicKeyInfo, &key_ptr);
    }
   
    /* Copy and Return the key value */
    if (CSSM_OK != ret)
    {
        (*memory_pool_funcs.recycle)(&pool);
        return ret;
    }

    *result_val = App_Malloc(CLHandle, sizeof(CSSM_KEY));
    if (!*result_val)
    {
        (*memory_pool_funcs.recycle)(&pool);
        return CSSMERR_CL_MEMORY_ERROR;
    }

    port_memcpy( *result_val, key_ptr, sizeof(CSSM_KEY) );
    (*result_val)->KeyData.Length = key_ptr->KeyData.Length;
    if (!(*result_val)->KeyData.Length)
    {
        (*result_val)->KeyData.Data = NULL;
    }
    else
    {
        (*result_val)->KeyData.Data = App_Malloc(CLHandle, (*result_val)->KeyData.Length);
        if (!(*result_val)->KeyData.Data)
        {
            (*memory_pool_funcs.recycle)(&pool);
            App_Free(CLHandle, *result_val);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        port_memcpy((*result_val)->KeyData.Data, key_ptr->KeyData.Data, (*result_val)->KeyData.Length);
    }

    (*memory_pool_funcs.recycle)(&pool);
    return CSSM_OK;
} /* ifdef VMS remove extraneous ';' */
