/*-----------------------------------------------------------------------
 * File: CRTVERFY.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 

#include "maf_include.h"
/* Windows & CSSM includes */
#ifndef VMS
#pragma warning (disable:4201 4214 4115)
#include <stdio.h>
#pragma warning (default:4201 4214 4115)
#else
#include <stdio.h>
#endif
#include "cssm.h"
#include "cssmspi.h"
#include "guids.h"
#include "cssmport.h"
#include "cssmutil.h"

/* Certificate-specific includes */
#include "x509defs.h"
#include "x_fndefs.h"

#include "oidsx509.h"
#include "oidsalg.h"
#include "x509hlpr.h"
#include "clutil.h"
#include "x_qtable.h"
#include "internal.h"

#include "api_shim.h"
#include "maf_util.h"

#include <cssmMemTrack.h>

#include "clberder.h"
#include "algoid.h"

extern const CSSM_API_MEMORY_FUNCS Addin_APIMemFuncs;
extern MEMORY_POOL_FUNCS memory_pool_funcs;

#define X509_FORMAT_VERSION 2


extern CSSM_API_MEMORY_FUNCS CLApiMemFuncs;
CSSM_ALGORITHMS CSSMAPI X509_AlgorithmOidToAlgId (CSSM_OID_PTR Oid);


CSSM_RETURN 
GetCCHandle(MEMORY_POOL_PTR poolptr,
            CSSM_CL_HANDLE CLHandle, 
            CSSM_CC_HANDLE CCHandle, 
            const CSSM_DATA *SignerCert,
            CSSM_ALGORITHMS verifyAlg,
            CSSM_CSP_HANDLE *NewCspHandle,
            CSSM_SUBSERVICE_UID_PTR *SsUidp,
            CSSM_CC_HANDLE *result_val);

void
FreeCCHandle(CSSM_CL_HANDLE CLHandle,
                CSSM_CC_HANDLE CCHandle, 
                CSSM_CC_HANDLE NewCCHandle, 
                CSSM_CSP_HANDLE NewCspHandle,
                CSSM_SUBSERVICE_UID *SsUidp);

/*-----------------------------------------------------------------------------
 * Name: CL_CertVerify
 *
 * Description:
 * This function verifies the signature on the input certificate.
 * 
 * Parameters: 
 * CCHandle (input)       : Handle identifying the Cryptographic Context 
 *                          to use for verification
 * SubjectCert (input)    : The cert to be verified
 * SignerCert (input)     : The cert which was used for signing
 * VerifyScope (input)    : This input is not used by this CL module
 * ScopeSize (input)      : This input is not used by this CL module
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSMCLI CL_CertVerify(CSSM_CL_HANDLE CLHandle,
                                CSSM_CC_HANDLE CCHandle, 
                                const CSSM_DATA *SubjectCert, 
                                const CSSM_DATA *SignerCert, 
                                const CSSM_FIELD *VerifyScope,
                                uint32 ScopeSize)
{
    /* Initializations */
    CSSM_X509_SIGNED_CERTIFICATE_PTR subj_cert = NULL;
    CSSM_RETURN             ret = CSSM_OK;   
    CSSM_CC_HANDLE          useCCHandle = 0;
    CSSM_CSP_HANDLE         newCspHandle = 0;
    CSSM_SUBSERVICE_UID_PTR ssUidp = NULL;
    CSSM_ALGORITHMS     verifyAlg = 0;
    CSSM_DATA               data_to_verify = {0, NULL};
    MEMORY_POOL pool;

    /* Check inputs */
    /* CLHandle and CCHandle are checked by CSSM */
    if (!SubjectCert || 
        CL_IsBadReadPtr(SubjectCert, sizeof(CSSM_DATA)))
    {
        
        return CSSMERR_CL_INVALID_CERT_POINTER;
    }
    if (!SubjectCert->Data   || 
        !SubjectCert->Length ||
        CL_IsBadReadPtr(SubjectCert->Data, SubjectCert->Length))
    {
        
        return CSSMERR_CL_INVALID_CERT_POINTER;
    }

    /* Signer Cert is optional. If it exists, it must be a valid readable pointer */
    if (SignerCert  &&
        CL_IsBadReadPtr(SignerCert,  sizeof(CSSM_DATA)) )
    {
        
        return CSSMERR_CL_INVALID_CERT_POINTER;
    }
    if (SignerCert &&
        (!SignerCert->Data    || 
        !SignerCert->Length  ||
        CL_IsBadReadPtr(SignerCert->Data,  SignerCert->Length)))
    {
        
        return CSSMERR_CL_INVALID_CERT_POINTER;
    }

    
    /* VerifyScope is not supported in this release. */
    /* Flag a warning if VerifyScope is passed to this routine */
    if (VerifyScope || ScopeSize)
    {
    
        return CSSMERR_CL_SCOPE_NOT_SUPPORTED;
    }

    /* Setup the Memory pool */
    ret = (*memory_pool_funcs.init)((CSSM_MEMORY_FUNCS_PTR)&Addin_APIMemFuncs, NULL, &pool);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    ret   = MP_DerDecodeSignedCertificate(&pool, SubjectCert, &subj_cert);
    if( CSSM_OK != ret )
    {
        if (ret != CSSMERR_CL_MEMORY_ERROR)
        {
            ret = CSSMERR_CL_UNKNOWN_FORMAT;
        }
        (*memory_pool_funcs.recycle)(&pool);
        return ret;
    }
    
    if ((subj_cert->signature.encrypted.Data == NULL) 
        || (subj_cert->signature.encrypted.Length == 0))
    {

		ret = CSSMERR_CL_VERIFICATION_FAILURE; 
        (*memory_pool_funcs.recycle)(&pool);
        return ret;
    }

    verifyAlg = X509_AlgorithmOidToAlgId(&subj_cert->signature.algorithmIdentifier.algorithm);

    /* Use SignerCert for verification if specified. */
    if (SignerCert)
    {
        GetCCHandle(&pool, CLHandle, CCHandle, SignerCert, verifyAlg, &newCspHandle, &ssUidp, &useCCHandle);
    }
    else /* else use CCHandle */
    {
        useCCHandle = CCHandle;
    }

    

    
    /* Verify the signature */
    x509_getToBeSignedCertificate(SubjectCert, &data_to_verify);
    ret = addin_VerifyData (useCCHandle, &data_to_verify, 1, CSSM_ALGID_NONE, &subj_cert->signature.encrypted);

    FreeCCHandle(CLHandle, CCHandle, useCCHandle, newCspHandle, ssUidp);
    (*memory_pool_funcs.recycle)(&pool);

    return ret;
} 

/*-----------------------------------------------------------------------------
 * Name: CL_CertVerifyWithKey
 *
 * Description:
 * This function verifies the signature on the input certificate.
 * 
 * Parameters: 
 * CCHandle (input)       : Handle identifying the Cryptographic Context 
 *                          to use for verification
 * SubjectCert (input)    : The cert to be verified
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSMCLI CL_CertVerifyWithKey(CSSM_CL_HANDLE CLHandle,
                                CSSM_CC_HANDLE CCHandle, 
                                const CSSM_DATA *SubjectCert)
{
    /* Initializations */
    CSSM_X509_SIGNED_CERTIFICATE_PTR subj_cert = NULL;
    CSSM_DATA               data_to_verify = {0, NULL};
    
    CSSM_RETURN             ret = CSSM_OK;   


    MEMORY_POOL pool;

    /* Check inputs */
    /* CLHandle and CCHandle are checked by CSSM */
    if (!SubjectCert || 
        CL_IsBadReadPtr(SubjectCert, sizeof(CSSM_DATA)))
    {
        
        return CSSMERR_CL_INVALID_CERT_POINTER;
    }
    if (!SubjectCert->Data   || 
        !SubjectCert->Length ||
        CL_IsBadReadPtr(SubjectCert->Data, SubjectCert->Length))
    {
        
        return CSSMERR_CL_INVALID_CERT_POINTER;
    }
    
    if (CCHandle == 0)
    {
        return CSSMERR_CL_INVALID_CONTEXT_HANDLE;
    }
    
    /* Setup the Memory pool */
    ret = (*memory_pool_funcs.init)((CSSM_MEMORY_FUNCS_PTR)&Addin_APIMemFuncs, NULL, &pool);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* Obtain the SignerCert in CSSM certificate format */
    ret   = MP_DerDecodeSignedCertificate(&pool, SubjectCert, &subj_cert);
    if( CSSM_OK != ret )
    {
        if (ret != CSSMERR_CL_MEMORY_ERROR)
        {
            ret = CSSMERR_CL_UNKNOWN_FORMAT;
        }
        (*memory_pool_funcs.recycle)(&pool);
        return ret;
    }

    if ((subj_cert->signature.encrypted.Data == NULL) 
        || (subj_cert->signature.encrypted.Length == 0))
    {
        ret = CSSMERR_CL_VERIFICATION_FAILURE ;
        (*memory_pool_funcs.recycle)(&pool);
        return ret;
    }

    /* Verify the signature */
    x509_getToBeSignedCertificate(SubjectCert, &data_to_verify);
    ret = addin_VerifyData (CCHandle, &data_to_verify, 1, CSSM_ALGID_NONE, &subj_cert->signature.encrypted);

    /* Cleanup & return */
    (*memory_pool_funcs.recycle)(&pool);    
    
    return ret;
} 
