/*-----------------------------------------------------------------------
 * File: FREEFLD.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 

#include "maf_include.h"
/* Windows & CSSM includes */
#ifndef VMS
#pragma warning (disable:4201 4214 4115)
#include <stdio.h>
#pragma warning (default:4201 4214 4115)
#else
#include <stdio.h>
#endif
#include "cssm.h"
#include "cssmspi.h"
#include "guids.h"
#include "cssmport.h"
#include "cssmutil.h"

/* Certificate-specific includes */
#include "x509defs.h"
#include "x_fndefs.h"

#include "oidsx509.h"
#include "oidsalg.h"
#include "x509hlpr.h"
#include "clutil.h"
#include "x_qtable.h"
#include "internal.h"

#include "api_shim.h"
#include "maf_util.h"

#include <cssmMemTrack.h>

#define X509_FORMAT_VERSION 2


/*---------------------------------------------------------------
 *
 *Name: CL_FreeFieldValue
 *
 *Description:
 *   Deallocates memory associated with the specified FieldValue 
 *   data element
 *
 *Parameters: 
 *   CLHandle - handle of CL module
 *   CertOrCrlField - pointer to a CSSM_FIELD structure
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMCLI CL_FreeFieldValue (CSSM_CL_HANDLE CLHandle,
                                       const CSSM_OID *CertOrCrlOid,
                                       CSSM_DATA_PTR Value)
{
    CSSM_RETURN retval = CSSM_OK;

    if (!CertOrCrlOid || !Value || !Value->Data)
    {
        return CSSMERR_CL_INVALID_FIELD_POINTER;
        
    }       

    switch ( x509_cl_OidToInt(*CertOrCrlOid) ) {

        case INT_SignedCertificateCStruct :
            retval = free_signedcert(CLHandle, 
                        (CSSM_X509_SIGNED_CERTIFICATE_PTR) Value->Data);
            break;

        case INT_CertificateCStruct :
            retval = free_certvalues(CLHandle, 
                        (CSSM_X509_TBS_CERTIFICATE_PTR) Value->Data);
            if (retval == CSSM_OK)
            {
                App_Free(CLHandle, Value->Data);
                Value->Data = NULL;
            }
            break;

        case INT_CRLAllExtensionsCStruct:
        case INT_CRLRevokedEntryAllExtensionsCStruct:
        case INT_CertificateExtensionsCStruct :
            retval = free_extensions(CLHandle,
                        ((CSSM_X509_EXTENSIONS_PTR) Value->Data));
            if (retval == CSSM_OK)
            {
                App_Free(CLHandle, Value->Data);
                Value->Data = NULL;
            }
            break;

        case INT_CSSMKeyStruct :
            {
                CSSM_KEY_PTR   key_ptr = NULL;

                key_ptr = (CSSM_KEY_PTR) Value->Data;
                if (key_ptr)
                {
                    if (key_ptr->KeyData.Data != NULL)
                    {
                        App_Free(CLHandle, key_ptr->KeyData.Data);
                        key_ptr->KeyData.Data = NULL;
                    }
                    App_Free(CLHandle, key_ptr);
                    key_ptr = NULL;
                }
                retval = CSSM_OK;
            }
            break;

        case INT_IssuerNameCStruct :
        case INT_SubjectNameCStruct :
        case INT_CRLIssuerNameCStruct :
            {
                CSSM_X509_NAME_PTR  name_ptr = NULL;

                name_ptr = (CSSM_X509_NAME_PTR) Value->Data;
                retval = free_name(CLHandle, name_ptr);
                if (retval == CSSM_OK)
                {
                    if (name_ptr != NULL)
                    {
                        App_Free(CLHandle, name_ptr);
                        name_ptr = NULL;
                    }
                }
            }
            break;

        case INT_SubjectPublicKeyCStruct :
        case INT_SignatureCStruct :
        case INT_CRLSignatureCStruct :
            {
                CSSM_X509_SUBJECT_PUBLIC_KEY_INFO_PTR spki_ptr = NULL;

                spki_ptr = (CSSM_X509_SUBJECT_PUBLIC_KEY_INFO_PTR) Value->Data;
                retval = free_subjectPublicKeyInfo(CLHandle, spki_ptr);
                if (retval == CSSM_OK)
                {
                    if (spki_ptr != NULL)
                    {
                        App_Free(CLHandle, spki_ptr);
                        spki_ptr = NULL;
                    }   
                }
            }
            break;

        case INT_ExtBasicConstraints :
            break;

        case INT_CRLSignedCrlCStruct:
            retval = free_signedCRL_array(CLHandle, 
                        (CSSM_X509_SIGNED_CRL_PTR)Value->Data); 
            /* free_signedCRL frees X509_SIGNED_CRL_PTR and 
               free_signedCRL_array frees CSSM_X509_SIGNED_CRL_PTR */
            break;

        case INT_CRLTbsCertListCStruct:
            /* free_certListvalues frees X509_TBS_CERTLIST_PTR and 
               free_certListvalues_array frees CSSM_X509_TBS_CERTLIST_PTR */
            retval = free_certListvalues_array(CLHandle, 
                        (CSSM_X509_TBS_CERTLIST_PTR)Value->Data);
            /* Do not free the top level structure in this case.
             * Freeing the data pointer may free an entire signed CRL structure. 
             */
            break;

        case INT_CRLRevokedCertificatesCStruct:
            /* free_revokedCertificates frees X509_REVOKED_CERT_LIST_PTR and 
               free_revokedCertificates_array frees CSSM_X509_REVOKED_CERT_LIST_PTR */
            retval = free_revokedCertificates_array(CLHandle,
                        (CSSM_X509_REVOKED_CERT_LIST_PTR)Value->Data);
            if (retval == CSSM_OK)
            {
                if (Value->Data != NULL)
                {
                    App_Free(CLHandle, Value->Data);
                    Value->Data = NULL;
                }
            }
            break;
        case INT_CRLRevokedEntryCStruct:
            retval = free_revokedCertEntry(CLHandle, 
                (CSSM_X509_REVOKED_CERT_ENTRY_PTR)Value->Data);
            if (retval == CSSM_OK)
            {
                if (Value->Data != NULL)
                {
                    App_Free(CLHandle, Value->Data);
                    Value->Data = NULL;
                }
            }
            break;

        case INT_NoMatch:
            return CSSMERR_CL_UNKNOWN_TAG;
            

        default :
            retval = CSSM_OK;
            App_Free(CLHandle, Value->Data);
            break;
    }

    if (retval == CSSM_OK)
    {
        Value->Length = 0;
        Value->Data = NULL;
    } 

    return retval;
}

/*---------------------------------------------------------------
 *
 *Name: CL_FreeFields
 *
 *Description:
 *   This function frees the fields in the FieldArray by freeing the 
 *   Data pointers for both the FieldOid and FieldValue fields. 
 *   It also frees the top level FieldArray pointer.
 *
 *Parameters: 
 *   CLHandle (input)
 *      The handle that describes the add-in certificate library module 
 *      used to perform this function.
 *  NumberOfFields (input)
 *      The length of the array of fields in FieldArray
 *  FieldArray (input)
 *      A pointer to an array of CSSM_FIELD structures that need to be 
 *      deallocated.
 *
 *Returns:
 *  A CSSM_RETURN value indicating success or specifying a particular 
 *  error condition. The value CSSM_OK indicates success. All other 
 *  values represent an error condition.
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMCLI CL_FreeFields (CSSM_CL_HANDLE CLHandle,
                                   uint32 NumberOfFields,
                                   CSSM_FIELD_PTR FieldArray)
{
    CSSM_RETURN retval = CSSM_OK;
    uint32 i=0;
    

    if (FieldArray == NULL) 
    {
        return CSSMERR_CL_INVALID_FIELD_POINTER;
    }

    if (CL_IsBadReadPtr (FieldArray, NumberOfFields*sizeof(CSSM_FIELD)) ) 
    {
        return CSSMERR_CL_INVALID_FIELD_POINTER;
        
    }

    for (i=0; i < NumberOfFields; i++)
    {
        retval = CL_FreeFieldValue(CLHandle, 
                    &FieldArray[i].FieldOid, 
                    &FieldArray[i].FieldValue);
        if (CSSM_OK != retval)
        {
            return retval;
        }

        if (FieldArray[i].FieldOid.Data != NULL)
        {
            App_Free(CLHandle, FieldArray[i].FieldOid.Data);
            FieldArray[i].FieldOid.Data = NULL;
        }
    }

    App_Free(CLHandle, FieldArray);
    return CSSM_OK;
}

