/*-----------------------------------------------------------------------
 * File: X509MAIN.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
 * This file implements all of the global module functions
 */

#include <assert.h>

/* Windows & CSSM includes */
#include "cssm.h"       /* CSSM types, errors, and APIs */
#include "cssmspi.h"    /* CSSM service provider interfaces */
#include "cssmcli.h"    /* CSSM certificate library interfaces */
#include "cssmport.h"

/* Intel includes */
#define CSSM_INITGUID
#include "guids.h"      /* guid definitions */
#include "mds_util_api.h"

#ifdef ISL_INCLUDED
#include "eislapi.h"    /* verification includes */
#endif

/* Certificate-specific includes */
#include "x_fndefs.h"
#include "x_qtable.h"
#include "clutil.h"
#include "internal.h"

#include "api_shim.h"   /* Shim for calls to CSSM addins */
#include "clm_cspi.h"   /* CL SPI functions */
#include "maf_include.h"
#define ALLOCREF (NULL) 




/* CL internal memory functions */
#ifndef CSSM_MEMTRACK_ON
static void * CSSMAPI CL_Malloc(CSSM_MODULE_HANDLE AddinHandle,
                                uint32 size)
{
    AddinHandle;
    return internal_malloc(size, ALLOCREF);
}

static void * CSSMAPI CL_Realloc(CSSM_MODULE_HANDLE AddinHandle,
                                 void *Ptr,
                                 uint32 size)
{
    AddinHandle;
    return internal_realloc(Ptr, size, ALLOCREF);
}

static void CSSMAPI CL_Free(CSSM_MODULE_HANDLE AddinHandle, void *Ptr)
{
    AddinHandle;
    internal_free(Ptr, ALLOCREF);
}

static void * CSSMAPI CL_Calloc(CSSM_MODULE_HANDLE AddinHandle,
                                uint32 size,
                                uint32 count)
{
    AddinHandle;
    return internal_calloc(size, count, ALLOCREF);
}

#else /* CSSM_MEMTRACK_ON*/

static void * CSSMAPI CL_Malloc(CSSM_MODULE_HANDLE AddinHandle,
                                uint32 size,
                                const char * szFilename,
                                uint32 u32LineNumber)
{
    AddinHandle;
    return internal_malloc(size, ALLOCREF, szFilename, u32LineNumber);
}

static void * CSSMAPI CL_Realloc(CSSM_MODULE_HANDLE AddinHandle,
                                 void *Ptr,
                                 uint32 size,
                                 const char * szFilename,
                                 uint32 u32LineNumber)
{
    AddinHandle;
    return internal_realloc(Ptr, size, ALLOCREF, szFilename, u32LineNumber);
}

static void CSSMAPI CL_Free(CSSM_MODULE_HANDLE AddinHandle, void *Ptr,const char * szFilename, uint32 u32LineNumber)
{
    AddinHandle;
    internal_free(Ptr, ALLOCREF, szFilename, u32LineNumber);
}

static void * CSSMAPI CL_Calloc(CSSM_MODULE_HANDLE AddinHandle,
                                uint32 size,
                                uint32 count,
                                const char * szFilename,
                                uint32 u32LineNumber)
{
    AddinHandle;
    return internal_calloc(size, count, ALLOCREF, szFilename, u32LineNumber);
}
#endif

#include <cssmMemTrack.h>

CSSM_UPCALLS CLMemFuncs =   { CL_Malloc,
                                       CL_Free,
                                       CL_Realloc,
                                       CL_Calloc };
CSSM_API_MEMORY_FUNCS CLApiMemFuncs ={ internal_malloc,
                                       internal_free,
                                       internal_realloc,
                                       internal_calloc };

static CSSM_MODULE_FUNCS_PTR s_pCLModuleFuncs = NULL;
static CSSM_SPI_CL_FUNCS_PTR s_pCLSPIFuncs = NULL;

PORT_TLS_INDEX TLS_MemoryPool;

CSSM_RETURN 
Addin_callout_Initialize(void)
{
    

    return port_AllocateTlsIndex(&TLS_MemoryPool, NULL);
}

CSSM_RETURN
Addin_callout_Terminate(void)
{
    
    CSSM_FREE_WRAPPER(CL_Free, 0x1234, s_pCLModuleFuncs);
    return port_DeleteTlsIndex(TLS_MemoryPool);
}

CSSM_RETURN
Addin_callout_LoadDataConstruct(ADDIN_MODULE_HANDLE CssmHandle, 
                                ADDIN_LOAD_DATA * pAddinLoadData)
{




    /* setup clutil library - must export to it mem (and previously err) func */


    assert( pAddinLoadData != NULL );

    /* Clear the structure */
    port_memset( pAddinLoadData, 0, sizeof(*pAddinLoadData) );

    MDSUTIL_Init();

    


    /* obtain needed CSSM entry points */
    /* CSSM API */
    if( Addin_GetProcAddress( CssmHandle,
                                "CSSM_GetAPIMemoryFunctions", 
                                (CSSM_PROC_ADDR*)&pAddinLoadData->GetAPIMemoryFunctions)
        != CSSM_OK )
        goto FAIL;
    if( Addin_GetProcAddress( CssmHandle, 
                                "CSSM_GenerateKeyPair", 
                                (CSSM_PROC_ADDR*)&pAddinLoadData->GenerateKeyPair)
        != CSSM_OK )
        goto FAIL;  
    if( Addin_GetProcAddress(   CssmHandle, 
                                "CSSM_QueryKeySizeInBits",
                                (CSSM_PROC_ADDR*)&pAddinLoadData->QueryKeySizeInBits)
        != CSSM_OK )
        goto FAIL;
    if( Addin_GetProcAddress( CssmHandle,
#ifdef VMS
                                "CSSM_CSP_ObtainPrivateKeyFromPu",
#else
                                "CSSM_CSP_ObtainPrivateKeyFromPublicKey",
#endif
                                (CSSM_PROC_ADDR*)&pAddinLoadData->csp_ObtainPrivateKeyFromPublicKey)
        != CSSM_OK )
        goto FAIL;
    if( Addin_GetProcAddress( CssmHandle,
                                "CSSM_SignData",
                                (CSSM_PROC_ADDR*)&pAddinLoadData->SignData)
        != CSSM_OK )
        goto FAIL;
    if( Addin_GetProcAddress( CssmHandle,
                                "CSSM_VerifyData",
                                (CSSM_PROC_ADDR*)&pAddinLoadData->VerifyData)
        != CSSM_OK )
        goto FAIL;  
    if( Addin_GetProcAddress( CssmHandle,
                                "CSSM_CSP_CreateSignatureContext", 
                                (CSSM_PROC_ADDR*)&pAddinLoadData->csp_CreateSignatureContext)
        != CSSM_OK )
        goto FAIL;
    if( Addin_GetProcAddress( CssmHandle,
                                "CSSM_CSP_CreateKeyGenContext", 
                                (CSSM_PROC_ADDR*)&pAddinLoadData->csp_CreateKeyGenContext)
        != CSSM_OK )
        goto FAIL;
    if( Addin_GetProcAddress( CssmHandle,
                                "CSSM_DeleteContext",
                                (CSSM_PROC_ADDR*)&pAddinLoadData->DeleteContext)
        != CSSM_OK )
        goto FAIL;   
    if( Addin_GetProcAddress( CssmHandle,
                                "CSSM_GetContext",
                                (CSSM_PROC_ADDR*)&pAddinLoadData->GetContext)
        != CSSM_OK )
        goto FAIL;   
    if( Addin_GetProcAddress( CssmHandle,
                                "CSSM_FreeContext",
                                (CSSM_PROC_ADDR*)&pAddinLoadData->FreeContext)
        != CSSM_OK )
        goto FAIL;   
     
    if( Addin_GetProcAddress( CssmHandle,
                                "CSSM_ModuleLoad",
                                (CSSM_PROC_ADDR*)&pAddinLoadData->ModuleLoad)
        != CSSM_OK )
        goto FAIL;   
    if( Addin_GetProcAddress( CssmHandle,
                                "CSSM_ModuleUnload",
                                (CSSM_PROC_ADDR*)&pAddinLoadData->ModuleUnload)
        != CSSM_OK )
        goto FAIL;
    if( Addin_GetProcAddress( CssmHandle,
                                "CSSM_ModuleAttach",
                                (CSSM_PROC_ADDR*)&pAddinLoadData->ModuleAttach)
        != CSSM_OK )
        goto FAIL;   
    if( Addin_GetProcAddress( CssmHandle,
                                "CSSM_ModuleDetach",
                                (CSSM_PROC_ADDR*)&pAddinLoadData->ModuleDetach)
        != CSSM_OK )
        goto FAIL;  
    if( Addin_GetProcAddress( CssmHandle,
                                "CSSM_GetContextAttribute",
                                (CSSM_PROC_ADDR*)&pAddinLoadData->GetContextAttribute)
        != CSSM_OK )
        goto FAIL;  
    if( Addin_GetProcAddress( CssmHandle,
                                "CSSM_DeleteContextAttributes",
                                (CSSM_PROC_ADDR*)&pAddinLoadData->DeleteContextAttributes)
        != CSSM_OK )
        goto FAIL;  

    return CSSM_OK;

FAIL:
    /* Clear the structure, with a known bad value (for debugging) */
    port_memset( pAddinLoadData, 0x33, sizeof(*pAddinLoadData) );

    return CSSMERR_CL_INTERNAL_ERROR;
}


void
Addin_callout_LoadDataDestroy( ADDIN_LOAD_DATA * pAddinLoadData )
{


    assert( pAddinLoadData != NULL );

    /* Clear the structure */
    port_memset( pAddinLoadData, 0, sizeof(*pAddinLoadData) );
    
    MDSUTIL_Term();
    return;
}


CSSM_RETURN
Addin_callout_AttachDataConstruct(
                                ADDIN_ATTACH_DATA * pAddinAttachData)
{


    CSSM_RETURN Rtn = CSSM_OK;

    /* setup clutil library - must export to it mem (and previously err) func */

    CSSM_UPCALLS MemFuncs = { _App_Malloc,
                              _App_Free,
                              _App_Realloc,
                              _App_Calloc };

    assert( pAddinAttachData != NULL );

    /* Clear the structure */
    port_memset( pAddinAttachData, 0, sizeof(*pAddinAttachData) );

    

    Rtn = CSSM_CLUTIL_Initialize( &pAddinAttachData->clContext,
                                 &MemFuncs);
                                  
    if( Rtn != CSSM_OK )
    {
        return( Rtn );
    }

    /* Initialize Query Table */
    Rtn = cl_QT_Initialize(&pAddinAttachData->clContext);
    if (Rtn != CSSM_OK)
    {
        CSSM_CLUTIL_Uninitialize( &pAddinAttachData->clContext );
        return Rtn;
    }

    



    return CSSM_OK;

}


void
Addin_callout_AttachDataDestroy( ADDIN_ATTACH_DATA * pAddinAttachData )
{
    CSSM_RETURN qtRtn = CSSM_OK;

    assert( pAddinAttachData != NULL );

    /* uninitialize query table */
    qtRtn = cl_QT_Uninitialize(&pAddinAttachData->clContext);

    assert(qtRtn == CSSM_OK);

    /* Shutdown the CL context */
    CSSM_CLUTIL_Uninitialize( &pAddinAttachData->clContext );

    /* Clear the structure */
    port_memset( pAddinAttachData, 0, sizeof(*pAddinAttachData) );

    return;
}

#define CL_FUNCTBL_LEN (sizeof(CSSM_SPI_CL_FUNCS)+sizeof(CSSM_MODULE_FUNCS)-sizeof(CSSM_PROC_ADDR))

CSSM_RETURN Addin_callout_ModuleAttach(
    MAF_MODULE_LOAD_TRACKER_PTR pLoadTracker, 
    MAF_MODULE_ATTACH_TRACKER_PTR pAttachTracker,
    CSSM_MODULE_FUNCS_PTR *FuncTbl)
{
    /* The size of the buffer is the size of the header + the size needed to store the csp functions */
    
    static CSSM_MODULE_FUNCS    FunctionList;
    static CSSM_SPI_CL_FUNCS    CLFunctions;

    /* Reference unused parameters to quiet the compiler */
    pLoadTracker;
    pAttachTracker;

    

    /* Assign the pointers */
    
    
    *FuncTbl = &FunctionList;
    

    /* Fill out the header */
    /* pFunctionList->TblSize = sizeof(CLFuncTblBuffer);
        Removed in CDSA spec 2.0 */
    FunctionList.ServiceType = CSSM_SERVICE_CL;
    FunctionList.NumberOfServiceFuncs = sizeof(CSSM_SPI_CL_FUNCS) / sizeof(CSSM_PROC_ADDR);
    FunctionList.ServiceFuncs =  (CSSM_PROC_ADDR*)&CLFunctions;
    /* Fill in the CL SPI functions */
    CLFunctions.CertCreateTemplate      = CLM_CertCreateTemplate;
    CLFunctions.CertGetAllTemplateFields = CLM_CertGetAllTemplateFields;
    CLFunctions.CertSign                = CLM_CertSign;

    CLFunctions.CertVerify              = CLM_CertVerify;

    CLFunctions.CertVerifyWithKey       = CLM_CertVerifyWithKey;

    CLFunctions.CertGetFirstFieldValue  = CLM_CertGetFirstFieldValue;
    CLFunctions.CertGetNextFieldValue   = CLM_CertGetNextFieldValue;
    CLFunctions.CertAbortQuery          = CLM_CertAbortQuery;
    CLFunctions.CertGetKeyInfo          = CLM_CertGetKeyInfo;
    CLFunctions.CertGetAllFields        = CLM_CertGetAllFields; 
    CLFunctions.FreeFields              = CLM_FreeFields;
    CLFunctions.FreeFieldValue          = CLM_FreeFieldValue;

    CLFunctions.CertCache                 = CLM_CertCache;
    CLFunctions.CertGetFirstCachedFieldValue = CLM_CertGetFirstCachedFieldValue;
    CLFunctions.CertGetNextCachedFieldValue = CLM_CertGetNextCachedFieldValue;
    CLFunctions.CertAbortCache          = CLM_CertAbortCache;

    CLFunctions.CertGroupToSignedBundle = CLM_CertGroupToSignedBundle;     
    CLFunctions.CertGroupFromVerifiedBundle = CLM_CertGroupFromVerifiedBundle;     
    CLFunctions.CertDescribeFormat      = CLM_CertDescribeFormat;

    CLFunctions.CrlCreateTemplate       = CLM_CrlCreateTemplate; 
    CLFunctions.CrlSetFields            = CLM_CrlSetFields;
    CLFunctions.CrlAddCert              = CLM_CrlAddCert;     
    CLFunctions.CrlRemoveCert           = CLM_CrlRemoveCert;  
    CLFunctions.CrlSign                 = CLM_CrlSign;        
    CLFunctions.CrlVerify               = CLM_CrlVerify; 
    
    CLFunctions.CrlVerifyWithKey        = CLM_CrlVerifyWithKey;

    CLFunctions.IsCertInCrl             = CLM_IsCertInCrl;   
    CLFunctions.CrlGetFirstFieldValue   = CLM_CrlGetFirstFieldValue;  
    CLFunctions.CrlGetNextFieldValue    = CLM_CrlGetNextFieldValue;   
    CLFunctions.CrlAbortQuery           = CLM_CrlAbortQuery;

    CLFunctions.CrlGetAllFields         = CLM_CrlGetAllFields;
    CLFunctions.CrlCache                = CLM_CrlCache;
    CLFunctions.IsCertInCachedCrl       = CLM_IsCertInCachedCrl;   
    CLFunctions.CrlGetFirstCachedFieldValue = CLM_CrlGetFirstCachedFieldValue;  
    CLFunctions.CrlGetNextCachedFieldValue = CLM_CrlGetNextCachedFieldValue;   
    CLFunctions.CrlGetAllCachedRecordFields = CLM_CrlGetAllCachedRecordFields;   
    CLFunctions.CrlAbortCache           = CLM_CrlAbortCache;

    CLFunctions.CrlDescribeFormat       = CLM_CrlDescribeFormat;  
    CLFunctions.PassThrough             = CLM_PassThrough;

    return CSSM_OK;
}

CSSM_RETURN Addin_callout_ModuleLoad(
    MAF_MODULE_LOAD_TRACKER_PTR pLoadTracker,
    CSSM_SPI_ModuleEventHandler CssmNotifyCallback,
    const void* CssmNotifyCallbackCtx)
{
    /* Reference unused parameters to quiet the compiler */
    pLoadTracker;
    
    /* Trigger a callback into the CSSM to simulate an insert */
    if ( CssmNotifyCallback != NULL )
    {
        CssmNotifyCallback( &intel_x509v3_clm_guid,
                            (void *)CssmNotifyCallbackCtx,
                            0,
                            CSSM_SERVICE_CL,
                            CSSM_NOTIFY_INSERT );
        return CSSM_OK;
    }

    return CSSMERR_CL_INTERNAL_ERROR;
}

CSSM_RETURN Addin_callout_ModuleUnload(
    MAF_MODULE_LOAD_TRACKER_PTR pLoadTracker,
    CSSM_SPI_ModuleEventHandler CssmNotifyCallback,
    const void* CssmNotifyCallbackCtx)
{
    /* Reference unused parameters to quiet the compiler */
    pLoadTracker;
    CssmNotifyCallback;
    CssmNotifyCallbackCtx;

    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: cl_CheckVersion
 *
 * Description:
 * This function check whether the input major & minor version is compatible
 * with this version of the CL.
 * 
 * Parameters: 
 * VerMajor (input)       : The major version number to be compared for compatiblity
 * VerMinor (input)       : The minor version number to be compared for compatiblity
 *
 * Return value:
 * An indicator of compatibility
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_CheckVersion (uint32 VerMajor, uint32 VerMinor)
{
    if (VerMajor == INTEL_X509V3_CLM_MAJOR_VER && VerMinor == INTEL_X509V3_CLM_MINOR_VER)
    {
        return CSSM_OK;
    }

    return CSSMERR_CSSM_INCOMPATIBLE_VERSION;
}

CSSM_UPCALLS *Get_App_UpcallFunctions (void)
{
    const MAF_MODULE_ATTACH_TRACKER * pTracker = Addin_GetAttachTracker();
    
    if (pTracker == NULL)
    {
        /* Did someone forget to call Addin_SPIBegin? */
        return NULL;
    }

    return (CSSM_UPCALLS *) &pTracker->AppUpcalls;
}

CSSM_RETURN CSSMAPI CL_CertGetAllTemplateFields(
                    CSSM_CL_HANDLE CLHandle,
                    const CSSM_DATA *CertTemplate,
                    uint32 *NumberOfFields,
                    CSSM_FIELD_PTR *CertFields )
{
    CSSM_RETURN ret = CSSMERR_CL_FUNCTION_NOT_IMPLEMENTED;

    return ret;
}


CSSM_RETURN CSSMAPI CL_CertCache(
                    CSSM_CL_HANDLE CLHandle,
                    const CSSM_DATA *Cert,
                    CSSM_HANDLE_PTR CertHandle )
{
    CSSM_RETURN ret = CSSMERR_CL_FUNCTION_NOT_IMPLEMENTED;

    return ret;
}

CSSM_RETURN CSSMAPI CL_CertGetFirstCachedFieldValue(
                    CSSM_CL_HANDLE CLHandle,
                    CSSM_HANDLE CertHandle,
                    const CSSM_OID *CertField,
                    CSSM_HANDLE_PTR ResultsHandle,
                    uint32 *NumberOfMatchedFields,
                    CSSM_DATA_PTR *FieldValue )
{
    CSSM_RETURN ret = CSSMERR_CL_FUNCTION_NOT_IMPLEMENTED;

    return ret;
}

CSSM_RETURN CSSMAPI CL_CertGetNextCachedFieldValue(
                    CSSM_CL_HANDLE CLHandle,
                    CSSM_HANDLE ResultsHandle,
                    CSSM_DATA_PTR *FieldValue )
{
    CSSM_RETURN ret = CSSMERR_CL_FUNCTION_NOT_IMPLEMENTED;

    return ret;
}

CSSM_RETURN CSSMAPI CL_CertAbortCache(
                    CSSM_CL_HANDLE CLHandle,
                    CSSM_HANDLE CertHandle )
{
    CSSM_RETURN ret = CSSMERR_CL_FUNCTION_NOT_IMPLEMENTED;

    return ret;
}


CSSM_RETURN CSSMAPI CL_CrlGetAllFields(
                    CSSM_CL_HANDLE CLHandle,
                    const CSSM_DATA *Crl,
                    uint32 *NumberOfCrlFields,
                    CSSM_FIELD_PTR *CrlFields )
{
    CSSM_RETURN ret = CSSMERR_CL_FUNCTION_NOT_IMPLEMENTED;

    return ret;
}

CSSM_RETURN CSSMAPI CL_CrlCache(
                    CSSM_CL_HANDLE CLHandle,
                    const CSSM_DATA *Crl,
                    CSSM_HANDLE_PTR CrlHandle )
{
    CSSM_RETURN ret = CSSMERR_CL_FUNCTION_NOT_IMPLEMENTED;

    return ret;
}

CSSM_RETURN CSSMAPI CL_IsCertInCachedCrl(
                    CSSM_CL_HANDLE CLHandle,
                    const CSSM_DATA *Cert,
                    CSSM_HANDLE CrlHandle,
                    CSSM_BOOL *CertFound,
                    CSSM_DATA_PTR CrlRecordIndex )
{
    CSSM_RETURN ret = CSSMERR_CL_FUNCTION_NOT_IMPLEMENTED;

    return ret;
}

CSSM_RETURN CSSMAPI CL_CrlGetFirstCachedFieldValue(
                    CSSM_CL_HANDLE CLHandle,
                    CSSM_HANDLE CrlHandle,
                    const CSSM_DATA *CrlRecordIndex,
                    const CSSM_OID *CrlField,
                    CSSM_HANDLE_PTR ResultsHandle,
                    uint32 *NumberOfMatchedFields,
                    CSSM_DATA_PTR *FieldValue )
{
    CSSM_RETURN ret = CSSMERR_CL_FUNCTION_NOT_IMPLEMENTED;

    return ret;
}

CSSM_RETURN CSSMAPI CL_CrlGetNextCachedFieldValue(
                    CSSM_CL_HANDLE CLHandle,
                    CSSM_HANDLE ResultsHandle,
                    CSSM_DATA_PTR *FieldValue )
{
    CSSM_RETURN ret = CSSMERR_CL_FUNCTION_NOT_IMPLEMENTED;

    return ret;
}

CSSM_RETURN CSSMAPI CL_CrlGetAllCachedRecordFields(
                    CSSM_CL_HANDLE CLHandle,
                    CSSM_HANDLE CrlHandle,
                    const CSSM_DATA *CrlRecordIndex,
                    uint32 *NumberOfFields,
                    CSSM_FIELD_PTR *Fields )
{
    CSSM_RETURN ret = CSSMERR_CL_FUNCTION_NOT_IMPLEMENTED;

    return ret;
}

CSSM_RETURN CSSMAPI CL_CrlAbortCache(
                    CSSM_CL_HANDLE CLHandle,
                    CSSM_HANDLE CrlHandle )
{
    CSSM_RETURN ret = CSSMERR_CL_FUNCTION_NOT_IMPLEMENTED;

    return ret;
}
