/*-----------------------------------------------------------------------
 * File: CSP_DGST.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "cspinc.h"

static CSSM_RETURN __ConstructHash(
                        CSSM_ALGORITHMS Algorithm,
                        EVP_MD_CTX *pSrcState,
                        EVP_MD_CTX *pDestState )
{
    CSSM_RETURN rv = CSSM_OK;

    assert( pDestState );

    if ( pSrcState != NULL )
    {
        memcpy( pDestState, pSrcState, sizeof(*pDestState) );
    }
    else
    {
        memset( pDestState, 0, sizeof(*pDestState) );
        switch ( Algorithm )
        {
        case CSSM_ALGID_MD2:
            {
                EVP_DigestInit( pDestState, EVP_md2() );
                break;
            }
        case CSSM_ALGID_MD5:
            {
                EVP_DigestInit( pDestState, EVP_md5() );
                break;
            }
        case CSSM_ALGID_SHA1:
            {
                EVP_DigestInit( pDestState, EVP_sha1() );
                break;
            }
        default:
            {
                ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
            }
        } /* switch Algorithm */
    } /* pSrcState */

    return rv;
}

static CSSM_RETURN __DigestInit( CSM_SESSION *pSession )
{
    CSSM_RETURN rv = CSSM_OK;
    EVP_MD_CTX *pState = NULL;

    /* Allocate a place to store the hash state */
    if ( ( rv = CSM_AllocateSessionState(
                                    pSession,
                                    sizeof(EVP_MD_CTX) ) ) == CSSM_OK )
    {
        pState = (EVP_MD_CTX*)(pSession->pCryptoState);

        /* Create the hasing object */
        rv = __ConstructHash( pSession->pContext->AlgorithmType,
                              NULL,
                              pState );
    }

    return rv;
}

static CSSM_RETURN __DigestUpdate(
                        CSM_SESSION *pSession,
                        const CSSM_DATA *DataBufs,
                        uint32 DataBufCount )
{
    CSSM_RETURN rv = CSSM_OK;
    EVP_MD_CTX *pState = NULL;
    uint32 uCount;

    /* Fetch the hash object from the session */
    assert( pSession->pCryptoState &&
            ( pSession->uCryptoStateSize == sizeof(EVP_MD_CTX) ) );
    pState = (EVP_MD_CTX*)pSession->pCryptoState;
    assert( pState );

    /* Add the data to the hash */
    for ( uCount = 0; uCount < DataBufCount; uCount++ )
    {
        /* Only call update if there is really data */
        if ( DataBufs[uCount].Length != 0 )
        {
            EVP_DigestUpdate( pState, DataBufs[uCount].Data, DataBufs[uCount].Length );
        }
    }

    return rv;
}

static CSSM_RETURN __DigestFinal(
                        CSM_SESSION *pSession,
                        CSSM_DATA *Digest )
{
    CSSM_RETURN rv = CSSM_OK;
    uint32 DigestLength = 0;
    CSSM_BOOL bAllocatedDigest = CSSM_FALSE;
    EVP_MD_CTX *pState = NULL;

    /* Fetch the hash object from the session */
    assert( pSession->pCryptoState &&
            ( pSession->uCryptoStateSize == sizeof(EVP_MD_CTX) ) );
    pState = (EVP_MD_CTX*)pSession->pCryptoState;
    assert( pState );

    /* Get the length of the output */
    DigestLength = EVP_MD_CTX_size( pState );

    /* Make sure there is enough space to return the digest */
    if ( ( rv = MAF_CreateOutputBuffer( pSession->hCsp,
                                        DigestLength, 
                                        Digest,
                                        &bAllocatedDigest ) ) == CSSM_OK )
    {
        /* Complete the digest */
        EVP_DigestFinal( pState, Digest->Data, (unsigned int *)&Digest->Length );
    }
    else
    {
        MAF_DeleteOutputBuffer( pSession->hCsp,
                                Digest,
                                bAllocatedDigest );
    }

    return rv;
}

void ADDIN_CSM_DigestDataError( CSM_SESSION *pSession )
{
    EVP_MD_CTX *pState;

    /* Just make sure the state is correct */
    assert( pSession->pCryptoState &&
            ( pSession->uCryptoStateSize == sizeof(EVP_MD_CTX) ) );

    /* Clear the context */
    pState = (EVP_MD_CTX*)pSession->pCryptoState;
    memset( pState, 0, sizeof(*pState) );
}

CSSM_RETURN CSSMAPI ADDIN_CSM_DigestData(
                        CSM_SESSION *pSession,
                        const CSSM_DATA *DataBufs,
                        uint32 DataBufCount,
                        CSSM_DATA_PTR Digest )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Start the digest operation */
    if ( ( rv = __DigestInit( pSession ) ) == CSSM_OK )
    {
        /* Perform the digesting operation */
        if ( ( rv = __DigestUpdate( pSession,
                                    DataBufs,
                                    DataBufCount ) ) == CSSM_OK )
        {
            /* Finalize the digest */
            rv = __DigestFinal( pSession,
                                Digest );
        }

        /* Make sure the session is cleared of the digest operation */
        ADDIN_CSM_DigestDataError( pSession );
    }

    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_DigestDataInit( CSM_SESSION *pSession )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Initialize the digest operation */
    rv = __DigestInit( pSession );

    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_DigestDataUpdate( CSM_SESSION *pSession,
                                       const CSSM_DATA *DataBufs,
                                       uint32 DataBufCount )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Feed data buffers into the digest one at a time */
    rv = __DigestUpdate( pSession,
                         DataBufs,
                         DataBufCount );

    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_DigestDataFinal( CSM_SESSION *pSession,
                                      CSSM_DATA *Digest )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Finalize the digest operation */
    rv = __DigestFinal( pSession,
                        Digest );

    /* Clear the hash state */
    ADDIN_CSM_DigestDataError( pSession );

    return rv;
}

CSSM_RETURN ADDIN_CSM_DuplicateDigestState(
                                    const CSSM_CONTEXT *pContext,
                                    const void *pSrcState,
                                    uint32 uSrcStateSize,
                                    void *pDestState )
{
    CSSM_RETURN rv = CSSM_OK;
    /* Assert that the MAF-CSM appears to be working correctly */
    assert( pSrcState &&
            pDestState && 
            ( uSrcStateSize == sizeof(EVP_MD_CTX) ) );

    /* Construct a duplicate hash */
    rv = __ConstructHash( pContext->AlgorithmType,
                          (EVP_MD_CTX*)(pSrcState),
                          (EVP_MD_CTX*)(pDestState) );

    return rv;
}
