/*-----------------------------------------------------------------------
 * File: KMG_CALLOUTS.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "cspinc.h"

#define KMGR_ALG_SPEC               EVP_des_ede3_cbc()
#define KMGR_KEY_LENGTH_BYTES       ( KMGR_ALG_SPEC->key_len )
#define KMGR_IV_LENGTH_BYTES        ( KMGR_ALG_SPEC->iv_len )
#define KMGR_ENCRYPT_ALG            CSSM_ALGID_3DES_3KEY
#define KMGR_ENCRYPT_MODE           CSSM_ALGMODE_CBCPadIV8
#define KMGR_ENCRYPT_BLOCK_SIZE     ( KMGR_ALG_SPEC->block_size )

CSSM_RETURN CSSMAPI CSM_KMCallout_GetWrapParams(
                        uint32 *pKeyLengthBytes,
                        uint32 *pIVLengthBytes,
                        CSSM_ALGORITHMS *pAlgId,
                        CSSM_ENCRYPT_MODE *pAlgMode )
{
    *pKeyLengthBytes = KMGR_KEY_LENGTH_BYTES;
    *pIVLengthBytes = KMGR_IV_LENGTH_BYTES;
    *pAlgId = KMGR_ENCRYPT_ALG;
    *pAlgMode = KMGR_ENCRYPT_MODE;

    return CSSM_OK;
}

CSSM_RETURN CSSMAPI CSM_KMCallout_GetRandomBytes(
                        uint32 Length,
                        uint8 *RandomBytes )
{
    RAND_bytes( RandomBytes, Length );

    return CSSM_OK;
}

CSSM_RETURN CSSMAPI CSM_KMCallout_EncryptData(
                        uint8 *EncryptKey,
                        uint8 *IV,
                        const CSSM_DATA *Input,
                        CSSM_DATA_PTR Output )
{
    CSSM_RETURN rv = CSSM_OK;
    EVP_CIPHER_CTX CipherCtx;
    uint32 OutputLength;
    uint32 RemainingBytes;
    int ResultBytes;
    uint8 *DataPtr;
    CSSM_BOOL bAllocated = CSSM_FALSE;
    
    /* Calculate the output size */
    /* OutputLength = NumberOfEncryptedBlocks * BlockSize */
    OutputLength = ( ( Input->Length / KMGR_ENCRYPT_BLOCK_SIZE ) + 1 ) *
                                                    KMGR_ENCRYPT_BLOCK_SIZE;

    /* Allocate the output buffer memory */
    Output->Data = NULL;
    Output->Length = 0;
    if ( ( rv = MAF_CreateOutputBuffer( MAF_ALLOCATE_LOCAL,
                                        OutputLength,
                                        Output,
                                        &bAllocated ) ) == CSSM_OK )
    {
        /* Clear the context */
        memset( &CipherCtx, 0, sizeof(CipherCtx) );

        /* Initialize the encryption */
        EVP_EncryptInit( &CipherCtx, 
                         KMGR_ALG_SPEC,
                         EncryptKey,
                         IV );

        /* Feed in the data to encrypt */
        DataPtr = Output->Data;
        RemainingBytes = Output->Length;
        ResultBytes = 0;
        EVP_EncryptUpdate( &CipherCtx,
                           DataPtr,
                           &ResultBytes,
                           Input->Data,
                           Input->Length );
        DataPtr = &Output->Data[ResultBytes];
        Output->Length = ResultBytes;

        /* Finalize the encryption */
        EVP_EncryptFinal( &CipherCtx,
                          DataPtr,
                          &ResultBytes );
        Output->Length += ResultBytes;

        memset( &CipherCtx, 0, sizeof(CipherCtx) );
    } /* MAF_CreateOutputBuffer */

    return rv;
}

CSSM_RETURN CSSMAPI CSM_KMCallout_DecryptData(
                        uint8 *DecryptKey,
                        uint8 *IV,
                        const CSSM_DATA *Input,
                        CSSM_DATA_PTR Output )
{
    CSSM_RETURN rv = CSSM_OK;
    EVP_CIPHER_CTX CipherCtx;
    int ResultBytes;
    uint8 *DataPtr;
    CSSM_BOOL bAllocated = CSSM_FALSE;
    
    /* Allocate the output buffer memory */
    Output->Data = NULL;
    Output->Length = 0;
    if ( ( rv = MAF_CreateOutputBuffer( MAF_ALLOCATE_LOCAL,
                                        Input->Length,
                                        Output,
                                        &bAllocated ) ) == CSSM_OK )
    {
        /* Clear the context */
        memset( &CipherCtx, 0, sizeof(CipherCtx) );

        /* Initialize the encryption */
        EVP_DecryptInit( &CipherCtx, 
                         KMGR_ALG_SPEC,
                         DecryptKey,
                         IV );

        /* Feed in the data to encrypt */
        DataPtr = Output->Data;
        ResultBytes = 0;
        EVP_DecryptUpdate( &CipherCtx,
                           DataPtr,
                           &ResultBytes,
                           Input->Data,
                           Input->Length );
        DataPtr = &Output->Data[ResultBytes];
        Output->Length = ResultBytes;

        /* Finalize the encryption */
        if ( !EVP_DecryptFinal( &CipherCtx,
                                DataPtr,
                                &ResultBytes ) )
        {
            ERR( rv = CSSMERR_CSP_INVALID_ACCESS_CREDENTIALS );
        }
        else
        {
            Output->Length += ResultBytes;
        }

        memset( &CipherCtx, 0, sizeof(CipherCtx) );
    } /* MAF_CreateOutputBuffer */

    return rv;
}

    

