/*-----------------------------------------------------------------------
 * File: CSP_RAND.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "cspinc.h"

/* Function to gather random information from the environment.
 */
#define EAYCSP_GET_ENV_RANDOM_MAX   ( 64 )
#define EAYCSP_MAX_SEED_LENGTH      ( 2048 )

CSSM_RETURN EAYCSP_GetEnvironmentRandom(
                        uint8 *pRandomBuf,
                        uint32 *pnRandomLen)
{
    uint8       RanBuf[EAYCSP_GET_ENV_RANDOM_MAX];
    uint32      RanCount=0;
    sint32      nLen=0;
    uint32      dwVal;
    uint32      wVal;
    uint8       bVal;
    
#if defined (WIN32)
    /* RSA Labs and DoD recommendations for seed generation */

    /* Clock tick count */
    {
        dwVal = GetTickCount();
        memcpy(RanBuf + RanCount, (char *)&dwVal, sizeof(dwVal));
        RanCount += sizeof(dwVal);
    }

    /* Mouse cursor coordinates */
    {
        POINT       point;

        GetCursorPos(&point);

        bVal = (uint8)(LOBYTE(point.x) ^ HIBYTE(point.x));
        memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
        RanCount += sizeof(bVal);

        bVal = (uint8)(LOBYTE(point.y) ^ HIBYTE(point.y));
        memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
        RanCount += sizeof(bVal);
    }
    
    /* System time in milliseconds */
    {
        SYSTEMTIME  sysTime;

        GetLocalTime(&sysTime);

        memcpy(RanBuf + RanCount, (char *)&sysTime.wMilliseconds, 
            sizeof(sysTime.wMilliseconds));
        RanCount += sizeof(sysTime.wMilliseconds);

        bVal = (uint8)(LOBYTE(sysTime.wSecond) ^ LOBYTE(sysTime.wMinute));
        memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
        RanCount += sizeof(bVal);
    }
        
    /* Queries the processor's high performance counter */
    {
        LARGE_INTEGER   liPerfCount;

        if (QueryPerformanceCounter(&liPerfCount))
        {
            bVal = (uint8)(LOBYTE(LOWORD(liPerfCount.LowPart)) 
                          ^ LOBYTE(LOWORD(liPerfCount.HighPart)));
            memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
            RanCount += sizeof(bVal);

            bVal = (uint8)(HIBYTE(LOWORD(liPerfCount.LowPart)) 
                          ^ LOBYTE(LOWORD(liPerfCount.HighPart)));
            memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
            RanCount += sizeof(bVal);

            bVal = (uint8)(LOBYTE(HIWORD(liPerfCount.LowPart)) 
                           ^ LOBYTE(LOWORD(liPerfCount.HighPart)));
            memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
            RanCount += sizeof(bVal);

            bVal = (uint8)(HIBYTE(HIWORD(liPerfCount.LowPart)) 
                          ^ LOBYTE(LOWORD(liPerfCount.HighPart)));
            memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
            RanCount += sizeof(bVal);
        }
    }

    /* Memory usage statistics */
    {
        MEMORYSTATUS    mstMemStat;

        mstMemStat.dwLength = sizeof(MEMORYSTATUS);
        GlobalMemoryStatus(&mstMemStat);

        wVal = HIWORD(mstMemStat.dwAvailPhys);
        memcpy(RanBuf + RanCount, (char *)&wVal, sizeof(wVal));
        RanCount += sizeof(wVal);

        wVal = HIWORD(mstMemStat.dwAvailPageFile);
        memcpy(RanBuf + RanCount, (char *)&wVal, sizeof(wVal));
        RanCount += sizeof(wVal);

        bVal = LOBYTE(HIWORD(mstMemStat.dwAvailVirtual));
        memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
        RanCount += sizeof(bVal);
    }

    /* Copy the environment random number to out buffer */
    nLen = (*pnRandomLen > RanCount)? RanCount: *pnRandomLen;
    memcpy(pRandomBuf, RanBuf, nLen);
    *pnRandomLen = nLen;

#elif defined (LINUX) || defined (VMS)
    
    srandom((unsigned int)time(NULL));
    dwVal = random();
    memcpy(RanBuf, (char *)&dwVal, sizeof(dwVal));
    RanCount = sizeof(dwVal);
    nLen = (*pnRandomLen > RanCount)? RanCount: *pnRandomLen;
    memcpy(pRandomBuf, RanBuf, nLen);
    *pnRandomLen = nLen;

#endif

    return CSSM_OK;
}

void ADDIN_CSM_GenerateRandomError( CSM_SESSION *pSession )
{
    /* Nothing required for the existing PRNGs */
}

CSSM_RETURN CSSMAPI ADDIN_CSM_GenerateRandom(
                        CSM_SESSION *pSession,
                        CSSM_DATA_PTR RandomNumber )
{
    CSSM_RETURN rv = CSSM_OK;
    const CSSM_CONTEXT_ATTRIBUTE *pAttr;
    uint32 NumBytes;
    CSSM_BOOL bAllocatedRandom = CSSM_FALSE;
    uint8 SeedBuf[EAYCSP_MAX_SEED_LENGTH];
    CSSM_DATA Seed;
    CSSM_BOOL bAllocatedSeed = CSSM_FALSE;
    uint32 SeedNeeded;
    uint32 SeedFetched;

    /* Make sure it's a supported algorithm */
    if ( pSession->pContext->AlgorithmType != CSSM_ALGID_MD5Random )
    {
        ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
    }
    else
    {
        /* Fetch the output length attribute early to keep from doing
         * extra work.
         */
        pAttr = CSM_GetContextAttribute( pSession->pContext,
                                         CSSM_ATTRIBUTE_OUTPUT_SIZE );
        if ( pAttr == NULL )
        {
            ERR( rv = CSSMERR_CSP_INVALID_ATTR_OUTPUT_SIZE );
        }
        else
        {
            NumBytes = pAttr->Attribute.Uint32;

            /* Add initial seed data to the state */
            pAttr = CSM_GetContextAttribute( pSession->pContext,
                                             CSSM_ATTRIBUTE_SEED );
            if ( pAttr != NULL )
            {

                Seed.Data = SeedBuf;
                Seed.Length = sizeof(SeedBuf);
                if ( ( rv = MAF_ResolveCryptoData( 
                                            pAttr->Attribute.CryptoData,
                                            &Seed,
                                            Seed.Length,
                                            &bAllocatedSeed ) ) == CSSM_OK )
                {
                    RAND_seed( Seed.Data, Seed.Length );
                }
                else if ( rv == CSSMERR_CSP_INVALID_CRYPTO_DATA )
                {
                    ERR( rv = CSSMERR_CSP_INVALID_ATTR_SEED );
                }
            }

            /* Add a few extra bytes of randomness to the PRNG */
            SeedNeeded = sizeof(SeedBuf);
            while ( SeedNeeded )
            {
                SeedFetched = SeedNeeded;
                EAYCSP_GetEnvironmentRandom( 
                                &SeedBuf[EAYCSP_MAX_SEED_LENGTH - SeedNeeded],
                                &SeedFetched );
                SeedNeeded -= SeedFetched;
            }
            RAND_seed( SeedBuf, sizeof(SeedBuf) );

            /* Allocate the output buffer and fetch the random data */
            if ( rv == CSSM_OK )
            {
                if ( ( rv = MAF_CreateOutputBuffer(
                                        pSession->hCsp,
                                        NumBytes,
                                        RandomNumber,
                                        &bAllocatedRandom ) ) == CSSM_OK )
                {
                    RandomNumber->Length = NumBytes;
                    RAND_bytes( RandomNumber->Data, RandomNumber->Length );
                }
            }
        } /* Attributes present */
    } /* Supported algorithm */

    return rv;
}

