/*-----------------------------------------------------------------------
 * File: CSP_VRFY.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "cspinc.h"

typedef struct _sig_state
{
    EVP_MD_CTX MdCtx;
    EVP_PKEY *pSigKey;
}
EAYCSP_SIG_STATE;

static CSSM_RETURN __ValidateSignature(
						CSSM_ALGORITHMS AlgId,
						const CSSM_DATA *pSig )
{
	CSSM_RETURN rv = CSSM_OK;
	const uint8 *pSigData = NULL;
    BER_PARSED_ITEM BerData[4];
    sint32 SetItemCount;
	sint32 i;

	if ( !pSig ||
		 ( pSig->Length == 0 ) )
	{
		ERR( rv = CSSMERR_CSP_INVALID_SIGNATURE );
	}
	else
	{
		pSigData = pSig->Data;

		switch ( AlgId )
		{
		case CSSM_ALGID_DSA:
		case CSSM_ALGID_SHA1WithDSA:
			{
				/* Do a quick expansion of the signature to make sure it's
				 * the right general format.
				 */
				pSigData = BER_ExpandItem( pSigData, pSig->Length, BerData );
				if ( ( pSigData != NULL) &&
					 ( *BerData[0].Tag == BER_CONSTRUCTED_SEQUENCE ) )
				{
					SetItemCount = BER_ExpandSet( BerData[0].Content,
												  BerData[0].ContentLength,
												  ( sizeof(BerData) / 
													sizeof(BER_PARSED_ITEM) ),
												  0,
												  NULL,
												  BerData );
					if ( SetItemCount != 2 )
					{
						ERR( rv = CSSMERR_CSP_INVALID_SIGNATURE );

						for ( i = 0; i < SetItemCount; i++ )
						{
							/* Check to make sure the lengths are reasonable */
							if ( ( BerData[i].ContentLength != 20 ) &&
								 ( ( BerData[i].ContentLength != 21 ) &&
								   ( BerData[i].Content != 0 ) ) )
							{
								ERR( rv = CSSMERR_CSP_INVALID_SIGNATURE );
							}
						}
					}
				}
				else
				{
					ERR( rv = CSSMERR_CSP_INVALID_SIGNATURE );
				}
				break;
			}
		default:
			{
				ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
			}
		}
	}

	return rv;
}

static CSSM_RETURN __ConstructVerifyHash(
                        CSSM_ALGORITHMS Algorithm,
                        EVP_MD_CTX *pMdCtx )
{
    CSSM_RETURN rv = CSSM_OK;

    assert( pMdCtx );

    memset( pMdCtx, 0, sizeof(*pMdCtx) );
    switch ( Algorithm )
    {
    case CSSM_ALGID_SHA1WithDSA:
        {
            EVP_VerifyInit( pMdCtx, EVP_dss1() );
            break;
        }
    default:
        {
            ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
        }
    } /* switch Algorithm */

    return rv;
}

static CSSM_RETURN __VerifyDataInit(
                        CSM_SESSION *pSession )
{
    CSSM_RETURN rv = CSSM_OK;
    EAYCSP_SIG_STATE *pState = NULL;

    /* Allocate a place to store the hash state */
    if ( ( rv = CSM_AllocateSessionState(
                                    pSession,
                                    sizeof(EAYCSP_SIG_STATE) ) ) == CSSM_OK )
    {
        pState = (EAYCSP_SIG_STATE*)(pSession->pCryptoState);

        /* Fetch the signing key */
        if ( ( rv = EAYCSP_FetchPublicKey( pSession,
                                           CSSM_KEYUSE_VERIFY,
                                           CSSM_TRUE,
                                           &pState->pSigKey ) ) == CSSM_OK )
        {
            /* Create the hashing object */
            rv = __ConstructVerifyHash( pSession->pContext->AlgorithmType,
                                        &pState->MdCtx );

			if ( rv != CSSM_OK )
			{
				ADDIN_CSM_VerifyDataError( pSession );
			}
		}
    }

    return rv;
}

static CSSM_RETURN __VerifyDataUpdate(
                        CSM_SESSION *pSession,
                        const CSSM_DATA *DataBufs,
                        uint32 DataBufCount )
{
    CSSM_RETURN rv = CSSM_OK;
    EAYCSP_SIG_STATE *pState = NULL;
    uint32 uCount;

    /* Fetch the hash object from the session */
    assert( pSession->pCryptoState &&
            ( pSession->uCryptoStateSize == sizeof(EAYCSP_SIG_STATE) ) );
    pState = (EAYCSP_SIG_STATE*)pSession->pCryptoState;
    assert( pState );

    /* Add the data to the hash */
    for ( uCount = 0; uCount < DataBufCount; uCount++ )
    {
        /* Only call update if there is really data */
        if ( DataBufs[uCount].Length != 0 )
        {
            EVP_VerifyUpdate( &pState->MdCtx,
                              DataBufs[uCount].Data,
                              DataBufs[uCount].Length );
        }
    }

    return rv;
}

static CSSM_RETURN __VerifyDataFinal(
                        CSM_SESSION *pSession,
                        const CSSM_DATA *Signature )
{
    CSSM_RETURN rv = CSSM_OK;
    EAYCSP_SIG_STATE *pState = NULL;
    int err;

    /* Fetch the hash object from the session */
    assert( pSession->pCryptoState &&
            ( pSession->uCryptoStateSize == sizeof(EAYCSP_SIG_STATE) ) );
    pState = (EAYCSP_SIG_STATE*)pSession->pCryptoState;
    assert( pState );

    /* The crypto library appears to leak memory if a non-NULL signature
     * pointer is passed with a zero length.
     */
    if ( ( rv = __ValidateSignature( pSession->pContext->AlgorithmType,
									 Signature ) ) == CSSM_OK )
    {
        /* Complete the digest and verify */
        err = EVP_VerifyFinal( &pState->MdCtx,
                               Signature->Data,
                               Signature->Length,
                               pState->pSigKey );

        if (err <= 0)
        {
            ERR_clear_error();

            if (err == 0)
            {
                ERR( rv = CSSMERR_CSP_VERIFY_FAILED );
            }
            else
            {
                ERR( rv = CSSMERR_CSP_INVALID_SIGNATURE );
            }
        }
    }

    return rv;
}

void ADDIN_CSM_VerifyDataError( CSM_SESSION *pSession )
{
    EAYCSP_SIG_STATE *pState;

    /* Just make sure the state is correct */
    assert( pSession->pCryptoState &&
            ( pSession->uCryptoStateSize == sizeof(EAYCSP_SIG_STATE) ) );

    /* Clear the context */
    pState = (EAYCSP_SIG_STATE*)pSession->pCryptoState;
    EVP_PKEY_free( pState->pSigKey );
    memset( pState, 0, sizeof(*pState) );
}

CSSM_RETURN CSSMAPI ADDIN_CSM_VerifyData(
                        CSM_SESSION *pSession,
                        const CSSM_DATA *DataBufs,
                        uint32 DataBufCount,
                        CSSM_ALGORITHMS DigestAlgorithm,
                        const CSSM_DATA *Signature )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Make sure that only combo-ops are requested */
    if ( DigestAlgorithm != CSSM_ALGID_NONE )
    {
        ERR( rv = CSSMERR_CSP_INVALID_DIGEST_ALGORITHM );
    }
    else
    {
        /* Start the digest operation */
        if ( ( rv = __VerifyDataInit( pSession ) ) == CSSM_OK )
        {
            /* Perform the digesting operation */
            if ( ( rv = __VerifyDataUpdate( pSession,
                                            DataBufs,
                                            DataBufCount ) ) == CSSM_OK )
            {
                /* Finalize the digest */
                rv = __VerifyDataFinal( pSession,
                                        Signature );
            } /* Update */

            /* Make sure the session is cleared of the digest operation */
            ADDIN_CSM_VerifyDataError( pSession );
        } /* Init */
    }

    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_VerifyDataInit(
                        CSM_SESSION *pSession )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Initialize the digest operation */
    rv = __VerifyDataInit( pSession );

    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_VerifyDataUpdate(
                        CSM_SESSION *pSession,
                        const CSSM_DATA *DataBufs,
                        uint32 DataBufCount )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Feed data buffers into the digest one at a time */
    rv = __VerifyDataUpdate( pSession,
                             DataBufs,
                             DataBufCount );

    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_VerifyDataFinal(
                        CSM_SESSION *pSession,
                        const CSSM_DATA *Digest )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Finalize the digest operation */
    rv = __VerifyDataFinal( pSession,
                            Digest );

    /* Clean up the signature verification state */
    ADDIN_CSM_VerifyDataError( pSession );

    return rv;
}
