/*-----------------------------------------------------------------------
 * File: CSP_DERV.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
 * File Description: Contains derive key operations for the CSP.
 */

#include "geninc.h"

#define CHECKBITS( value, mask ) ( ( value & mask ) == mask )

#define MAX_HASH_LENGTH         (20)
#define PKCS5_PBKD1_SALT_LEN    (8)
#define PKCS5_PBKD1_IV_LEN      (8)

extern B_ALGORITHM_METHOD *DIGEST_MD5_CHOOSER[];
extern B_ALGORITHM_METHOD *DIGEST_MD2_CHOOSER[];
extern B_ALGORITHM_METHOD *DIGEST_SHA_CHOOSER[];

/* BSafe Diffie Hellman Param Generation Algorithm Chooser */
B_ALGORITHM_METHOD *DH_PARAM_CHOOSER[] = {
  &AM_DH_PARAM_GEN,
  (B_ALGORITHM_METHOD *)NULL_PTR
};

typedef struct _valid_key_params
{
    CSSM_ALGORITHMS AlgId;
    uint32 MinLength;
    uint32 MaxLength;
    uint32 LengthMultiple;
}
_VALID_KEY_PARAMS;

static const _VALID_KEY_PARAMS _KeyCheckParams[] =
{
    { CSSM_ALGID_DES, 64, 64, 1 },
    { CSSM_ALGID_DESX, 64, 64, 1 },
    { CSSM_ALGID_3DES_3KEY_EDE, 192, 192, 1 },
    { CSSM_ALGID_3DES_2KEY_EDE, 128, 128, 1 },
    { CSSM_ALGID_3DES_1KEY_EEE, 64, 64, 1 },
    { CSSM_ALGID_RC2, 40, 1024, 8 },
    { CSSM_ALGID_RC5, 40, 1024, 8 },
    { CSSM_ALGID_RC4, 40, 128, 8 },
    { CSSM_ALGID_RSA, 512, 2048, 256 },
    { CSSM_ALGID_DSA, 512, 2048, 64 },
    { CSSM_ALGID_GenericSecret, 8, 2048, 8 },
    { CSSM_ALGID_3DES, 192, 192, 1 }
};
static const _NumKeyCheckParams = sizeof(_KeyCheckParams) / 
                                    sizeof(_VALID_KEY_PARAMS);

CSSM_RETURN bsafe_CheckKeyParams(
                        CSSM_ALGORITHMS AlgId,
                        uint32 KeyLength )
{
    CSSM_RETURN rv = CSSM_OK;
    const _VALID_KEY_PARAMS *pParams = NULL;

    /* Find the key parameters for the specified algorithm */
    {
        uint32 Top = 0;
        uint32 Bottom = _NumKeyCheckParams;
        uint32 Cursor;
        
        Cursor = ( Top + Bottom ) / 2;
        while ( ( _KeyCheckParams[Cursor].AlgId != AlgId ) &&
                ( Top <= Bottom ) )
        {
            if ( AlgId < _KeyCheckParams[Cursor].AlgId )
            {
                Bottom = Cursor - 1;
            }
            else
            {
                Top = Cursor + 1;
            }
            Cursor = ( Top + Bottom ) / 2;
        }
        if ( ( Top <= Bottom ) &&
             ( _KeyCheckParams[Cursor].AlgId == AlgId ) )
        {
            pParams = &_KeyCheckParams[Cursor];
        }
    }

    /* Set the proper error codes based on the search result and parameters */
    if ( pParams == NULL )
    {
        rv = CSSMERR_CSP_INVALID_ALGORITHM;
    }
    else if ( ( KeyLength < pParams->MinLength ) ||
              ( KeyLength > pParams->MaxLength ) ||
              ( ( KeyLength % pParams->LengthMultiple ) != 0 ) )
    {
        rv = CSSMERR_CSP_UNSUPPORTED_KEY_SIZE;
    }

    return rv;
}

CSSM_RETURN bsafe_PKCS5_PBKD1(
                        CSM_SESSION *pSession,
                        const CSSM_DATA *pPassphrase,
                        const CSSM_DATA *pSalt,
                        uint32 IterationCount,
                        uint32 KeyLength,
                        CSSM_DATA *pKeyData,
                        CSSM_DATA *pIV )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_BOOL bAllocatedIV = CSSM_FALSE;
    CSSM_BOOL bAllocatedKeyData = CSSM_FALSE;
    uint8 *BufPtr;
    CSSM_DATA InBuf;
    uint8 OutBufArray[MAX_HASH_LENGTH];
    CSSM_DATA OutBuf;
    CSSM_DATA TempData;
    uint32 i;

    int bsRv = 0;
    B_ALGORITHM_METHOD **bsAlgChooser = NULL;
    B_ALGORITHM_OBJ bsAlgObject;
    B_INFO_TYPE bsAlgInfoType;

    /* Allocate memory for the IV and key output */
    if ( ( ( rv = MAF_CreateOutputBuffer(
                                    pSession->hCsp,
                                    PKCS5_PBKD1_IV_LEN,
                                    pIV,
                                    &bAllocatedIV ) ) == CSSM_OK ) &&
         ( ( rv = MAF_CreateOutputBuffer(
                                    MAF_ALLOCATE_LOCAL,
                                    ( KeyLength / 8 ),
                                    pKeyData,
                                    &bAllocatedKeyData ) ) == CSSM_OK ) )
    {
        /* Allocate a temporary buffer to hold the output of the iterations */
        BufPtr = Addin_malloc( max( MAX_HASH_LENGTH,
                                    pPassphrase->Length + pSalt->Length ),
                               NULL );
        if ( BufPtr == NULL )
        {
            rv = CSSMERR_CSP_MEMORY_ERROR;
        }
        else
        {
            OutBuf.Data = OutBufArray;
            OutBuf.Length = sizeof(OutBufArray);

            /* Copy the passphrase and the salt into the input buffer */
            memcpy( BufPtr,
                    pPassphrase->Data,
                    pPassphrase->Length );
            memcpy( &BufPtr[pPassphrase->Length],
                    pSalt->Data, 
                    pSalt->Length );
            InBuf.Data = BufPtr;
            InBuf.Length = pPassphrase->Length + pSalt->Length;

            /* Determine the algorithm type and create the alg object */
            switch ( pSession->pContext->AlgorithmType )
            {
            case CSSM_ALGID_PKCS5_PBKDF1_MD2:
                {
                    bsAlgChooser = DIGEST_MD2_CHOOSER;
                    bsAlgInfoType = AI_MD2;
                    break;
                }
            case CSSM_ALGID_PKCS5_PBKDF1_MD5:
                {
                    bsAlgChooser = DIGEST_MD5_CHOOSER;
                    bsAlgInfoType = AI_MD5;
                    break;
                }
            case CSSM_ALGID_PKCS5_PBKDF1_SHA1:
                {
                    bsAlgChooser = DIGEST_SHA_CHOOSER;
                    bsAlgInfoType = AI_SHA1;
                    break;
                }
            default:
                {
                    bsAlgChooser = NULL;
                    bsAlgInfoType = NULL;
                    assert( 0 );
                }
            }
            if ( ( bsRv = B_CreateAlgorithmObject( &bsAlgObject ) ) == 0 )
            {
                if ( ( bsRv = B_SetAlgorithmInfo( bsAlgObject,
                                                  bsAlgInfoType,
                                                  NULL_PTR ) ) == 0 )
                {
                    /* Perform the derivation iterations */
                    for ( i = 0; ( i < IterationCount) && ( bsRv == 0 ); i++ )
                    {
                        if ( ( bsRv = B_DigestInit(
                                        bsAlgObject,
                                        NULL_PTR,
                                        bsAlgChooser,
                                        (A_SURRENDER_CTX*)NULL_PTR ) ) == 0 )
                        {
                            if ( ( bsRv = B_DigestUpdate(
                                        bsAlgObject,
                                        InBuf.Data,
                                        InBuf.Length,
                                        (A_SURRENDER_CTX*)NULL_PTR ) ) == 0 )
                            {
                                bsRv = B_DigestFinal(
                                                bsAlgObject,
                                                OutBuf.Data,
                                                (unsigned int *)&OutBuf.Length, /* ifdef VMS apply cast */
                                                MAX_HASH_LENGTH,
                                                (A_SURRENDER_CTX*)NULL_PTR );
                            } /* digest update */             
                        } /* digest init */

                        /* Quit if something failed */
                        if ( bsRv != 0 )
                        {
                            break;
                        }

                        /* Swap the input and output buffers */
                        TempData = InBuf;
                        InBuf = OutBuf;
                        OutBuf = InBuf;
                    } /* iterations */

                    if ( bsRv == 0 )
                    {
                        /* Extract the key and IV from the hash result */
                        /* The result is in InBuf because of the last swap */

                        /* Check that the requested key length is workable */
                        if ( ( InBuf.Length - PKCS5_PBKD1_IV_LEN ) >= 
                                                        ( KeyLength / 8 ) )
                        {
                            /* Copy the IV */
                            memcpy( pIV->Data, 
                                    &InBuf.Data[InBuf.Length - 
                                                        PKCS5_PBKD1_IV_LEN],
                                    PKCS5_PBKD1_IV_LEN );

                            /* Copy the key */
                            memcpy( pKeyData->Data,
                                    InBuf.Data,
                                    ( KeyLength / 8 ) );
                        }
                        else
                        {
                            rv = CSSMERR_CSP_INVALID_ATTR_KEY_LENGTH;
                        }
                    }
                } /* Set alg info */

                B_DestroyAlgorithmObject( &bsAlgObject );
            } /* Create alg object */
            if ( bsRv != 0 )
            {
                rv = ConvertBsafeError( bsRv );
            }

            /* Free temporary buffer */
            memset( BufPtr, 0, PKCS5_PBKD1_IV_LEN );
            memset( OutBufArray, 0, PKCS5_PBKD1_IV_LEN );
            Addin_free( BufPtr, NULL );
        } /* Allocate temporary buffer */

        /* Clean up if key derivation failed */
        if ( rv != CSSM_OK )
        {
            MAF_DeleteOutputBuffer(
                            MAF_ALLOCATE_LOCAL,
                            pKeyData,
                            bAllocatedKeyData );
            MAF_DeleteOutputBuffer( 
                            pSession->hCsp,
                            pIV,
                            bAllocatedIV );
        }
    } /* Allocate init vector */

    return rv;
}

CSSM_RETURN bsafe_PKCS5_PBKD1_WithValidation( 
                        CSM_SESSION *pSession,
                        CSSM_DATA_PTR pParam,
                        CSSM_KEY_PTR pResult )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_PKCS5_PBKDF1_PARAMS *pPbeParams = NULL;
    const CSSM_CONTEXT_ATTRIBUTE *pAttr;
    const CSSM_CONTEXT_ATTRIBUTE *pKeyTypeAttr;
    const CSSM_CONTEXT_ATTRIBUTE *pKeyLenAttr;
    CSSM_DATA Salt;
    uint32 IterationCount = 1000;
    CSSM_ALGORITHMS KeyType;
    uint32 KeyLength;

    assert( pSession && pResult );

    /* Validate the PBD parameters */
    if ( port_IsBadReadPtr( pParam, sizeof(CSSM_DATA) ) ||
         ( pParam->Length != sizeof(CSSM_PKCS5_PBKDF1_PARAMS) ) ||
         port_IsBadReadPtr( pParam->Data, sizeof(CSSM_PKCS5_PBKDF1_PARAMS) ) )
    {
        rv = CSSMERR_CSP_INVALID_INPUT_POINTER;
    }
    else 
    {
        pPbeParams = (CSSM_PKCS5_PBKDF1_PARAMS*)pParam->Data;

        if ( ( ( rv = CSM_ValidateInputBuffers( &pPbeParams->Passphrase,
                                                1,
                                                CSSM_FALSE ) ) == CSSM_OK ) &&
             ( ( rv = CSM_ValidateOutputBuffers( &pPbeParams->InitVector,
                                                 1,
                                                 CSSM_FALSE ) ) == CSSM_OK ) )
        {
            /* Fetch and validate the PBD parameters */
            pAttr = CSM_GetContextAttribute( 
                                        pSession->pContext,
                                        CSSM_ATTRIBUTE_ITERATION_COUNT );
            if ( pAttr )
            {
                IterationCount = pAttr->Attribute.Uint32;
            }
            pAttr = CSM_GetContextAttribute( pSession->pContext,
                                             CSSM_ATTRIBUTE_SALT );
            if ( !pAttr )
            {
                rv = CSSMERR_CSP_INVALID_ATTR_SALT;
            }
            else
            {
                Salt = *pAttr->Attribute.Data;
                if ( ( Salt.Length != 8 ) ||
                     port_IsBadReadPtr( Salt.Data, Salt.Length ) )
                {
                    rv = CSSMERR_CSP_INVALID_ATTR_SALT;
                }
                else
                {
                    /* Fetch and validate the result key parameters */
                    pKeyTypeAttr = CSM_GetContextAttribute(
                                                pSession->pContext,
                                                CSSM_ATTRIBUTE_KEY_TYPE );
                    pKeyLenAttr = CSM_GetContextAttribute( 
                                                pSession->pContext,
                                                CSSM_ATTRIBUTE_KEY_LENGTH );
                    if ( pKeyTypeAttr == NULL )
                    {
                        rv = CSSMERR_CSP_INVALID_ATTR_KEY_TYPE;
                    }
                    else if ( pKeyLenAttr == NULL )
                    {
                        rv = CSSMERR_CSP_INVALID_ATTR_KEY_LENGTH;
                    }
                    else
                    {   
                        KeyType = pKeyTypeAttr->Attribute.Uint32;
                        KeyLength = pKeyLenAttr->Attribute.Uint32;

                        if ( ( rv = bsafe_CheckKeyParams( 
                                                KeyType,
                                                KeyLength ) ) == CSSM_OK )
                        {
                            pResult->KeyData.Data = NULL;
                            pResult->KeyData.Length = 0;
                    
                            if ( ( rv = bsafe_PKCS5_PBKD1(
                                            pSession,
                                            &pPbeParams->Passphrase,
                                            &Salt,
                                            IterationCount,
                                            KeyLength,
                                            &pResult->KeyData,
                                            &pPbeParams->InitVector ) )
                                                                == CSSM_OK )
                            {
                                /* Fill in the key header */
                                csp_KeyHeader_Fill( &pResult->KeyHeader,
                                                    CSSM_KEYBLOB_RAW,
                                                    CSSM_KEYBLOB_RAW_FORMAT_OTHER,
                                                    KeyType,
                                                    CSSM_KEYCLASS_SESSION_KEY,
                                                    KeyLength,
                                                    0,
                                                    0,
                                                    NULL,
                                                    NULL,
                                                    CSSM_ALGID_NONE,
                                                    CSSM_ALGMODE_NONE,
                                                    0 );
                            }
                        } /* Fetch output key params 2 */
                    } /* Fetch output key params 1 */
                } /* Fetch PBD Params 2 */
            } /* Fetch PBD Params 1 */
        } /* Validate PBD Parameters 2 */
    } /* Validate PBD Parameters 1 */

    return rv;
}

CSSM_RETURN bsafe_DH_DeriveKey(CSM_SESSION *pSession,
                               CSSM_DATA_PTR Param,
                               uint32 KeyUsage,
                               uint32 KeyAttr,
                               CSSM_KEY_PTR pDerivedKey)
{
    CSSM_RETURN                     rv = CSSM_OK;
    uint32                          status;
    const ADDIN_LOAD_DATA           *loadData = NULL;
    CSM_KM_FORMAT_CONTEXT           kmFormat;
    B_ALGORITHM_OBJ                 algObject;
	CSSM_BOOL						bRecycleAlgObject = CSSM_FALSE;
    A_DH_KEY_AGREE_PARAMS           *getParams;
    const CSSM_CONTEXT_ATTRIBUTE    *pKeyAttr, *pKeyTypeAttr;
    const CSSM_CONTEXT_ATTRIBUTE    *pKeyLenAttr, *pCredAttr;
    const CSSM_ACCESS_CREDENTIALS   *pAccessCred = NULL;
    CSSM_KEY                        *pKey = NULL, BaseKey;
    CSSM_BOOL                       bAllocatedKeyData = CSSM_FALSE;
    CSSM_ALGORITHMS                 KeyType;
    uint32                          KeyLength, KeyLengthInBytes;
    CSSM_DATA                       tempData;

    /* Initialize */
    tempData.Data = NULL;
    tempData.Length = 0;

	memset(&algObject, 0, sizeof(B_ALGORITHM_OBJ));

    /* Fetch and validate the result key parameters */
    pKeyAttr = CSM_GetContextAttribute(
                                pSession->pContext,
                                CSSM_ATTRIBUTE_KEY );
    pKeyTypeAttr = CSM_GetContextAttribute(
                                pSession->pContext,
                                CSSM_ATTRIBUTE_KEY_TYPE );
    pKeyLenAttr = CSM_GetContextAttribute( 
                                pSession->pContext,
                                CSSM_ATTRIBUTE_KEY_LENGTH );

    pCredAttr = CSM_GetContextAttribute( 
                                pSession->pContext,
                                CSSM_ATTRIBUTE_ACCESS_CREDENTIALS );

    if ((pKeyAttr == NULL) || (pKeyAttr->Attribute.Key == NULL))
    {
        rv = CSSMERR_CSP_INVALID_ATTR_KEY;
        goto FAIL;
    }
    pKey = pKeyAttr->Attribute.Key;
    if ((pKey->KeyData.Data == NULL) || (pKey->KeyData.Length == 0))
    {
        rv = CSSMERR_CSP_INVALID_ATTR_KEY;
        goto FAIL;
    }

    if ( pCredAttr == NULL)
    {
        pAccessCred = NULL;
    }
    else
    {
        pAccessCred = pCredAttr->Attribute.AccessCredentials;
    }

    if ( pKeyTypeAttr == NULL )
    {
        rv = CSSMERR_CSP_INVALID_ATTR_KEY_TYPE;
    }
    else if ( pKeyLenAttr == NULL )
    {
        rv = CSSMERR_CSP_INVALID_ATTR_KEY_LENGTH;
    }
    else
    {   
        KeyType = pKeyTypeAttr->Attribute.Uint32;
        KeyLength = pKeyLenAttr->Attribute.Uint32;
        
        if (KeyType == CSSM_ALGID_GenericSecret)
        {
            if ( ( rv = bsafe_CheckKeyParams(KeyType, KeyLength ) ) == CSSM_OK )
            {
                /* Get the key from the key manager */
                loadData = Addin_GetLoadData();

                if (loadData == NULL)
                {
                    rv = CSSMERR_CSP_INTERNAL_ERROR;
                    goto FAIL;
                }

                kmFormat.Format = CSSM_KEYBLOB_RAW_FORMAT_OTHER;
                kmFormat.SigAlg = CSSM_ALGID_NONE;

                if ( ( pKey->KeyData.Length != sizeof(CSM_KM_KEY_HANDLE) ) ||
                                        port_IsBadReadPtr( pKey->KeyData.Data,
                                        sizeof(CSM_KM_KEY_HANDLE) ) )
                {
                    rv = CSSMERR_CSP_INVALID_KEY_REFERENCE;
                    goto FAIL;                      
                }
                rv = CSM_KMFindKeyByHandle(loadData->hKeyFile,
                                            (CSM_KM_KEY_HANDLE)*pKey->KeyData.Data,
                                            &kmFormat,
                                            pAccessCred,
                                            &BaseKey);

                if (rv != CSSM_OK)
                {
                    goto FAIL;
                }

                memcpy(&algObject, BaseKey.KeyData.Data, sizeof(B_ALGORITHM_OBJ));

				Addin_free (BaseKey.KeyData.Data, NULL);

                /* Determine how much space to allocate for the value */
                if ((status = B_GetAlgorithmInfo((POINTER *)&getParams, 
                            algObject, AI_DHKeyAgree)) != 0)
                {
                    rv = ConvertBsafeError(status);
                    goto FAIL;
                }

				bRecycleAlgObject = CSSM_TRUE;

                /* Make sure that the requested key length is not greater than the 
                 * size that is being returned from BSafe.
                 */
                if (KeyLength > (getParams->prime.len * 8))
                {
                    rv = CSSMERR_CSP_INVALID_ATTR_KEY_LENGTH;
                    goto FAIL;
                }

                tempData.Data = Addin_malloc(getParams->prime.len, NULL);
                if (tempData.Data == NULL)
                {
                    rv = CSSMERR_CSP_MEMORY_ERROR;
                    goto FAIL;
                }

                KeyLengthInBytes = KeyLength/8;
                if ((KeyLength % 8) != 0)
                {
                    KeyLengthInBytes++;
                }
                pDerivedKey->KeyData.Length = KeyLengthInBytes;

                if ( (rv = MAF_CreateOutputBuffer(MAF_ALLOCATE_LOCAL,
                                                  KeyLengthInBytes,
                                                  &(pDerivedKey->KeyData),
                                                  &bAllocatedKeyData)) != CSSM_OK)
                {
                    goto FAIL;
                }

                memset(pDerivedKey->KeyData.Data, 0, pDerivedKey->KeyData.Length);

                if ((status = B_KeyAgreePhase2(algObject, tempData.Data,
                                    (unsigned int *)&(tempData.Length), getParams->prime.len,  /* ifdef VMS apply cast */
                                    Param->Data, Param->Length,
                                    (A_SURRENDER_CTX *)NULL_PTR)) != 0)
                {
                    rv = ConvertBsafeError(status);                    
                    goto FAIL;
                }

                memcpy(pDerivedKey->KeyData.Data, tempData.Data, KeyLengthInBytes);
                pDerivedKey->KeyData.Length = KeyLengthInBytes;

                csp_KeyHeader_Fill(&pDerivedKey->KeyHeader,
                                   CSSM_KEYBLOB_RAW,
                                   CSSM_KEYBLOB_RAW_FORMAT_NONE,
                                   CSSM_ALGID_GenericSecret,
                                   CSSM_KEYCLASS_SECRET_PART,
                                   KeyLength,
                                   KeyAttr,
                                   KeyUsage,
                                   NULL/*&StartDate*/,
                                   NULL/*&EndDate*/,
                                   CSSM_ALGID_NONE,
                                   0,
                                   0);
                
            }
        }
        else
        {
            rv = CSSMERR_CSP_INVALID_ATTR_KEY_TYPE;
        }
    }

FAIL:

	 /* Remove the key from the key manager */

	if (bRecycleAlgObject == CSSM_TRUE) 
	{
                /* VMS removed this call 23-jan-2001 re: Bill Priestly's bug report
		CSM_KMDeleteKey(loadData->hKeyFile,
						(CSM_KM_KEY_HANDLE)*pKey->KeyData.Data,
						pAccessCred);
                */
		B_DestroyAlgorithmObject(&algObject);
	}	
    if (tempData.Data != NULL)
    {
        Addin_free(tempData.Data, NULL);
    }
    if (rv != CSSM_OK)
    {
        MAF_DeleteOutputBuffer(pSession->hCsp,
                               &(pDerivedKey->KeyData),
                               bAllocatedKeyData);
    }
    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_DeriveKey( 
                        CSM_SESSION *pSession,
                        CSSM_DATA_PTR Param,
                        uint32 KeyUsage,
                        uint32 KeyAttr,
                        const CSSM_DATA *KeyLabel,
                        const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
                        CSSM_KEY_PTR DerivedKey )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_KEY TempKey;
    CSM_KM_KEY_HANDLE KeyHandle = CSSM_INVALID_HANDLE;
    const ADDIN_LOAD_DATA *pLoadData = NULL;
    uint32 KeyReturnType;

    /* Clear the temporary key structure */
    memset( &TempKey, 0, sizeof(TempKey) );

    /* Validate the key attributes parameter */
    KeyReturnType = KeyAttr & ( CSSM_KEYATTR_RETURN_DATA |
                                CSSM_KEYATTR_RETURN_REF |
                                CSSM_KEYATTR_RETURN_NONE );
    if ( KeyReturnType == CSSM_KEYATTR_RETURN_DEFAULT )
    {
        KeyReturnType = CSSM_KEYATTR_RETURN_DATA;
    }
    switch ( KeyReturnType )
    {
    case CSSM_KEYATTR_RETURN_DATA:
        {
            /* The key must be extractable to be returned as data.
             * If the key is extractable but sensitive, then the key can
             * not be returned raw. It must be extracted wrapped.
             */
            if ( !CHECKBITS( KeyAttr, CSSM_KEYATTR_EXTRACTABLE ) ||
                 CHECKBITS( KeyAttr, CSSM_KEYATTR_SENSITIVE ) )
            {
                rv = CSSMERR_CSP_INVALID_KEYATTR_MASK;
            }
            break;
        }
    case CSSM_KEYATTR_RETURN_REF:
        {
            /* Any set of attributes is fine */
            break;
        }
    case CSSM_KEYATTR_RETURN_NONE:
        {
            /* The key must be permanent and named with a label. */
            if ( !CHECKBITS( KeyAttr, CSSM_KEYATTR_PERMANENT ) )
            {
                rv = CSSMERR_CSP_INVALID_KEYATTR_MASK;
            }
            else if ( KeyLabel == NULL )
            {
                rv = CSSMERR_CSP_INVALID_KEY_LABEL;
            }
            break;
        }
    default:
        {
            /* This code should never execute because MAF-CSM checks to make
             * sure the return type is valid.
             */
            assert( 0 );
            break;
        }
    }

    if (rv == CSSM_OK)
    {
        /* Derive the key according the algorithm ID */
        switch ( pSession->pContext->AlgorithmType )
        {
        case CSSM_ALGID_PKCS5_PBKDF1_MD2:
        case CSSM_ALGID_PKCS5_PBKDF1_MD5:
        case CSSM_ALGID_PKCS5_PBKDF1_SHA1:
            {
                rv = bsafe_PKCS5_PBKD1_WithValidation( pSession,
                                                       Param,
                                                       &TempKey );
                break;
            }
        case CSSM_ALGID_DH:
            {
                rv = bsafe_DH_DeriveKey(pSession, Param, KeyUsage, KeyAttr, &TempKey);

                break;
            }
        default:
            {
                rv = CSSMERR_CSP_INVALID_ALGORITHM;
            }
        }
    }

    /* "cook" the raw key into a returnable format */
    if ( rv == CSSM_OK )
    {
        /* Complete the key header */
        TempKey.KeyHeader.KeyAttr = ( KeyAttr & ~0xF0000000 );
        TempKey.KeyHeader.KeyUsage = KeyUsage;

        /* Allocate memory for the returned key data, and copy it if
         * possible. In the case of references, the value must be copied
         * after it is inserted into the key manager.
         */
        if ( KeyReturnType == CSSM_KEYATTR_RETURN_DATA )
        {
            DerivedKey->KeyData.Data = App_Malloc( pSession->hCsp, TempKey.KeyData.Length );
            if ( DerivedKey->KeyData.Data == NULL )
            {
                rv = CSSMERR_CSP_MEMORY_ERROR;
            }
            else
            {
                memcpy( DerivedKey->KeyData.Data, TempKey.KeyData.Data, TempKey.KeyData.Length );
                DerivedKey->KeyData.Length = TempKey.KeyData.Length;
            }
        }
        else if ( KeyReturnType == CSSM_KEYATTR_RETURN_REF )
        {
            DerivedKey->KeyData.Data = App_Malloc( pSession->hCsp, sizeof(CSM_KM_KEY_HANDLE) );
            if ( DerivedKey->KeyData.Data == NULL )
            {
                rv = CSSMERR_CSP_MEMORY_ERROR;
            }
        }

        /* Insert the key into the key manager */
        if ( ( rv == CSSM_OK ) &&
             ( CHECKBITS( KeyAttr, CSSM_KEYATTR_PERMANENT ) ||
               ( KeyReturnType == CSSM_KEYATTR_RETURN_REF ) ) )
        {
            pLoadData = Addin_GetLoadData();
            if ( pLoadData == NULL )
            {
                rv = CSSMERR_CSP_INTERNAL_ERROR;
            }
            else
            {
                rv = CSM_KMInsertKey( pLoadData->hKeyFile,
                                      &TempKey,
                                      NULL,
                                      KeyLabel,
                                      CredAndAclEntry,
                                      &KeyHandle );
            }
        }

        /* Nothing that can fail should be placed after this line because
         * the key has already been inserted into the key manager file.
         * Clean-up in the case of an error means deleting the key from the
         * key manager, and since that can fail and leave the file in a
         * different state than when we started nothing can fail from now
         * on. A key that is present in the file that should not exist is
         * a security risk.
         */

        if ( rv == CSSM_OK )
        {
            /* Copy the value for reference keys and adjust the format types */
            if ( KeyReturnType == CSSM_KEYATTR_RETURN_REF )
            {
                memcpy( DerivedKey->KeyData.Data, (void*)&KeyHandle, sizeof(KeyHandle) );
                DerivedKey->KeyData.Length = sizeof(KeyHandle);

                TempKey.KeyHeader.BlobType = CSSM_KEYBLOB_REFERENCE;
                TempKey.KeyHeader.Format = CSSM_KEYBLOB_REF_FORMAT_INTEGER;
            }

            /* Copy the key header over to the result key */
            if ( ( rv == CSSM_OK ) &&
                 ( KeyReturnType != CSSM_KEYATTR_RETURN_NONE ) )
            {
                memcpy( &DerivedKey->KeyHeader,
                        &TempKey.KeyHeader,
                        sizeof(DerivedKey->KeyHeader) );
            }
        }
        else
        {
            /* Wipe the output key buffer and release it */
            memset( DerivedKey->KeyData.Data, 0, DerivedKey->KeyData.Length );
            App_Free( pSession->hCsp, DerivedKey->KeyData.Data );
        }

        /* Free the temporary key data buffer */
        memset( TempKey.KeyData.Data, 0, TempKey.KeyData.Length );
        Addin_free( TempKey.KeyData.Data, NULL );
    }

    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_GenerateAlgorithmParams( 
                        CSM_SESSION *pSession,
                        uint32 ParamBits,
                        CSSM_DATA_PTR Param,
                        uint32* NumberOfUpdatedAttributes,
                        CSSM_CONTEXT_ATTRIBUTE_PTR* UpdatedAttributes )
{
    CSSM_RETURN                     rtn = CSSM_OK;
    CSSM_CONTEXT                    *pContext = pSession->pContext;
    CSSM_CONTEXT_ATTRIBUTE          *pAlgParams;
    const CSSM_CONTEXT_ATTRIBUTE    *pAttribute = NULL;
    uint32                          status;
    B_ALGORITHM_OBJ                 randomNumGen = (B_ALGORITHM_OBJ)NULL_PTR;
    B_ALGORITHM_OBJ                 dhParamGenerator = (B_ALGORITHM_OBJ)NULL_PTR;
    A_DH_PARAM_GEN_PARAMS           dhParams;
    B_ALGORITHM_OBJ                 dhParametersObj = (B_ALGORITHM_OBJ)NULL_PTR;
    ITEM                            *dhParametersBER;

    /* Allocate the parameter structure */
    pAlgParams = App_Malloc( pSession->hCsp, sizeof(CSSM_CONTEXT_ATTRIBUTE) );
    if ( pAlgParams == NULL )
    {
        return CSSMERR_CSP_MEMORY_ERROR;
    }
    memset( pAlgParams, 0, sizeof(CSSM_CONTEXT_ATTRIBUTE) );

    pAlgParams->AttributeType = CSSM_ATTRIBUTE_ALG_PARAMS;
    pAlgParams->AttributeLength = sizeof(CSSM_DATA);
    pAlgParams->Attribute.Data = App_Malloc( pSession->hCsp, sizeof(CSSM_DATA) );
    if ( pAlgParams->Attribute.Data == NULL )
    {
        Addin_free( pAlgParams, NULL );
        return CSSMERR_CSP_MEMORY_ERROR;
    }

    if (pContext->ContextType == CSSM_ALGCLASS_KEYGEN)
    {
        if (pContext->AlgorithmType == CSSM_ALGID_DH)
        {
            /* This format is specified in PKCS #3 with the addition of
             * the dhKeyAgreement object identifier (also defined in
             * PKCS #3.
             *
             * The data has the following ASN.1 format:
             * Parameters ::= SEQUENCE {
             *  oid OBJECT IDENTIFIER, -- dhKeyAgreement
             *  params DHParameter
             * }
             * DHParameter ::= SEQUENCE {
             *  prime INTEGER, -- p
             *  base INTEGER, -- g
             *  privateValueLength INTEGER OPTIONAL
             * }
             */
            pAttribute = CSM_GetContextAttribute(pContext, 
                                                CSSM_ATTRIBUTE_KEY_LENGTH );
            if ( pAttribute == NULL )
            {
                rtn = CSSMERR_CSP_INVALID_ATTR_KEY_LENGTH;
                goto FAIL;
            }

#if BSAFE_VERSION > 3
            rtn = csp_InitRandom(CSSM_ALGID_SHARandom, &randomNumGen);
#else
            rtn = csp_InitRandom(CSSM_ALGID_MD5Random, &randomNumGen);
#endif
            if (rtn != CSSM_OK)
            {
                goto FAIL;
            }

            /* Create algorithm object */
            if ((status = B_CreateAlgorithmObject (&dhParamGenerator)) != 0)
            {
                rtn = ConvertBsafeError(status);
                goto FAIL;
            }
             
            /* Set algorithm object to AI_DHParamGen */
            dhParams.primeBits = pAttribute->Attribute.Uint32;
            if (ParamBits != 0)
            {
                dhParams.exponentBits = ParamBits;
            } else {
                /* Set the exponentBits to be one byte less than the prime */
#if defined (COMPAQ_FIX) || defined (VMS)
                dhParams.exponentBits = dhParams.primeBits - 8;
#else  /* COMPAQ_FIX */
                dhParams.exponentBits = ParamBits - 8;
#endif /* COMPAQ_FIX */
            }
                
            if ((status = B_SetAlgorithmInfo(dhParamGenerator, AI_DHParamGen,
                                                    (POINTER)&dhParams)) != 0)
            {
                rtn = ConvertBsafeError(status);
                goto FAIL;
            }
            
            /* Init */
            if ((status = B_GenerateInit (dhParamGenerator, DH_PARAM_CHOOSER,
                  (A_SURRENDER_CTX *)NULL_PTR)) != 0)
            {
                rtn = ConvertBsafeError(status);
                goto FAIL;
            }

            /* Generate DH parameters */
            if ((status = B_CreateAlgorithmObject (&dhParametersObj)) != 0)
            {
                rtn = ConvertBsafeError(status);
                goto FAIL;
            }

            if ((status = B_GenerateParameters (dhParamGenerator, 
                        dhParametersObj, randomNumGen,
                        (A_SURRENDER_CTX *)NULL_PTR)) != 0)
            {
                rtn = ConvertBsafeError(status);
                goto FAIL;
            }

            /* Get the DH parameters. */
            if ((status = B_GetAlgorithmInfo((POINTER *)&dhParametersBER, 
                        dhParametersObj, AI_DHKeyAgreeBER)) != 0)
            {
                rtn = ConvertBsafeError(status);
                goto FAIL;
            }

            pAlgParams->Attribute.Data->Data = App_Malloc(pSession->hCsp, dhParametersBER->len);
            if (pAlgParams->Attribute.Data->Data == NULL)
            {
                rtn = CSSMERR_CSP_MEMORY_ERROR;
                goto FAIL;
            }
            pAlgParams->Attribute.Data->Length = dhParametersBER->len;
            memcpy(pAlgParams->Attribute.Data->Data, dhParametersBER->data,
                        dhParametersBER->len);
        }
        else
        {
            rtn = CSSMERR_CSP_INVALID_ALGORITHM;
            goto FAIL;
        }
    }
    else
    {
        rtn = CSSMERR_CSP_INVALID_ALGORITHM;
        goto FAIL;
    }

    B_DestroyAlgorithmObject(&randomNumGen);
    B_DestroyAlgorithmObject(&dhParamGenerator);
    B_DestroyAlgorithmObject(&dhParametersObj);

    *UpdatedAttributes = pAlgParams;
    *NumberOfUpdatedAttributes = 1;

    return rtn;

FAIL:
    if (pAlgParams)
    {
        if (pAlgParams->Attribute.Data)
        {
            App_Free(pSession->hCsp, pAlgParams->Attribute.Data);
        }
        App_Free(pSession->hCsp, pAlgParams);
    }
    if (randomNumGen)
        B_DestroyAlgorithmObject(&randomNumGen);
    if (dhParamGenerator)
        B_DestroyAlgorithmObject(&dhParamGenerator);
    if (dhParametersObj)
        B_DestroyAlgorithmObject(&dhParametersObj);
    return rtn;
}
