/*-----------------------------------------------------------------------
 * File: CSP_ENC.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
 * File Description: Contains encryption operations for the CSP.
 */

#include "geninc.h"

/* BSafe Encrypting Algorithm Choosers */
B_ALGORITHM_METHOD  *ENCRYPT_RSA_CHOOSER[] = {
    &AM_RSA_ENCRYPT,
    &AM_RSA_CRT_ENCRYPT,
    (B_ALGORITHM_METHOD*)NULL_PTR
};
B_ALGORITHM_METHOD  *ENCRYPT_RC5_CHOOSER[] = {
    &AM_RC5_CBC_ENCRYPT,
    (B_ALGORITHM_METHOD*)NULL_PTR
};
B_ALGORITHM_METHOD  *ENCRYPT_RC4_CHOOSER[] = {
    &AM_RC4_ENCRYPT,
    (B_ALGORITHM_METHOD*)NULL_PTR
};
B_ALGORITHM_METHOD  *ENCRYPT_RC2_CHOOSER[] = {
    &AM_RC2_CBC_ENCRYPT,
    (B_ALGORITHM_METHOD*)NULL_PTR
};
B_ALGORITHM_METHOD  *ENCRYPT_DES_CHOOSER[] = {
    &AM_DES_CBC_ENCRYPT,
    (B_ALGORITHM_METHOD*)NULL_PTR
};
B_ALGORITHM_METHOD  *ENCRYPT_3DES_CHOOSER[] = {
    &AM_DES_EDE3_CBC_ENCRYPT,
    (B_ALGORITHM_METHOD*)NULL_PTR
};
B_ALGORITHM_METHOD  *ENCRYPT_DESX_CHOOSER[] = {
    &AM_DESX_CBC_ENCRYPT,
    (B_ALGORITHM_METHOD*)NULL_PTR
};

static CSSM_ALGORITHMS ValidAlgorithms[] = 
									{   CSSM_ALGID_RSA,
										CSSM_ALGID_DES,
										CSSM_ALGID_3DES_3KEY,
										CSSM_ALGID_DESX,
										CSSM_ALGID_RC2,
										CSSM_ALGID_RC4,
										CSSM_ALGID_RC5,
										CSSM_ALGID_NONE};


static const uint32 NumValidAlgorithms = sizeof(ValidAlgorithms) / 
                                            sizeof(CSSM_ALGORITHMS);

CSSM_RETURN bsafe_ValidateEncDecAlg (CSSM_ALGORITHMS Alg)

{
	uint32 i;

	for (i = 0; i < NumValidAlgorithms; i++)
	{
		if (Alg == ValidAlgorithms[i])
			break;
	}

	if (i == NumValidAlgorithms)	              
		return CSSMERR_CSP_INVALID_ALGORITHM;

	return CSSM_OK;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_EncryptDataInit( CSM_SESSION *pSession,
                                               CSSM_PRIVILEGE Privilege )
{
    CSSM_RETURN                     rtn = CSSM_OK;
    int                             status;
    CSM_KM_FORMAT_CONTEXT           kmFormat;
    const ADDIN_LOAD_DATA           *loadData;
    BSAFECSP_STATE_PTR              pState;
    const CSSM_CONTEXT_ATTRIBUTE    *pKeyAttr = NULL, *pModeAttr = NULL,
                                    *pIVAttr = NULL, *pEffectiveBitsAttr = NULL,
                                    *pRoundsAttr = NULL, *pCredAttr = NULL, *pPaddingAttr = NULL;
    const CSSM_ACCESS_CREDENTIALS   *pAccessCred = NULL;
    CSSM_KEY                        *pKey = NULL, priKey;
    ITEM                            bsafeKey;
    A_DESX_KEY                      DESXKey;
    A_RC5_CBC_PARAMS                RC5Params;
    A_RC2_CBC_PARAMS                RC2Params;
    uint8                           keybuf[MAX_KEYBUF_LEN];
    B_INFO_TYPE                     keyType, algType;
    unsigned char                   *initVector = NULL;
    B_ALGORITHM_METHOD              **algChooser;
	CSSM_PADDING					Padding = CSSM_PADDING_NONE;
	uint32							KeySize;


    /* Allocate space to store the algorithm map pointer */
    rtn = CSM_AllocateSessionState( pSession, sizeof(BSAFECSP_STATE) );
    
    if ( rtn != CSSM_OK )
    {
        rtn = CSSMERR_CSP_INTERNAL_ERROR;
    }
    else
    {
        pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
        memset(pState, 0, sizeof(BSAFECSP_STATE));

        /* Create the encryption algorithm object */
        if ((status = B_CreateAlgorithmObject(&pState->algObject)) != 0)
        {
            rtn = ConvertBsafeError(status);
        }
        else
        {
            if ((status = B_CreateKeyObject(&pState->keyObject)) != 0)
            {
                rtn = ConvertBsafeError(status);
                goto FAIL_INIT;
            }

            /* Get all of the context attributes */

			if ((rtn = bsafe_ValidateEncDecAlg (pSession->pContext->AlgorithmType)) != CSSM_OK)
			{                
				rtn = CSSMERR_CSP_INVALID_ALGORITHM;
                goto FAIL_INIT;
            }

            /* Get the key from the context */
            pKeyAttr = CSM_GetContextAttribute(pSession->pContext,
                                        CSSM_ATTRIBUTE_KEY);

            if ((pKeyAttr == NULL) || (pKeyAttr->Attribute.Key == NULL))
            {
                rtn = CSSMERR_CSP_MISSING_ATTR_KEY;
                goto FAIL_INIT;
            }
            pKey = pKeyAttr->Attribute.Key;
			if ( pKey == NULL )
            {
                rtn = CSSMERR_CSP_INVALID_ATTR_KEY;
                goto FAIL_INIT;
            }
            else
            {
				if ((pKey->KeyData.Data == NULL) || (pKey->KeyData.Length == 0))
				{
					rtn = CSSMERR_CSP_INVALID_KEY;
					goto FAIL_INIT;
				}
			}			
						
            /* Check that the key matches the requested operation */
            if (pKey->KeyHeader.AlgorithmId != pSession->pContext->AlgorithmType)
            {
                rtn = CSSMERR_CSP_ALGID_MISMATCH;
                goto FAIL_INIT;
            }

			if (!CHECKBITS(pKey->KeyHeader.KeyUsage,
                                  CSSM_KEYUSE_ENCRYPT ) &&
                      !CHECKBITS(pKey->KeyHeader.KeyUsage,
                                  CSSM_KEYUSE_ANY ) &&
					  !CHECKBITS(pKey->KeyHeader.KeyUsage,
                                  CSSM_KEYUSE_WRAP ) )								  
            {                
				rtn = CSSMERR_CSP_KEY_USAGE_INCORRECT;
                goto FAIL_INIT;
            }

			KeySize = pKey->KeyHeader.LogicalKeySizeInBits;

			if ((rtn = bsafe_CheckKeyParams(pKey->KeyHeader.AlgorithmId, KeySize)) != CSSM_OK)
				goto FAIL_INIT;

            /* Get the mode attribute.  The check for a good value will be 
             * done when the mode is actually used */
            pModeAttr = CSM_GetContextAttribute(pSession->pContext,
                                            CSSM_ATTRIBUTE_MODE);

			/* Get the Padding from the context if it is there */
			pPaddingAttr = CSM_GetContextAttribute( pSession->pContext,
                                                CSSM_ATTRIBUTE_PADDING );
			if ( pPaddingAttr != NULL )
			{
				Padding = pPaddingAttr->Attribute.Uint32;
			}

            /* Get the IV from the context if it is there */
            pIVAttr = CSM_GetContextAttribute(pSession->pContext, 
                                            CSSM_ATTRIBUTE_INIT_VECTOR);

            if ((pSession->pContext->AlgorithmType != CSSM_ALGID_RSA) &&
                (pSession->pContext->AlgorithmType != CSSM_ALGID_RC4))
            {
                if (pIVAttr == NULL)
                {
                    rtn = CSSMERR_CSP_MISSING_ATTR_INIT_VECTOR;
                    goto FAIL_INIT;
                }
                else
                {
                    if (pIVAttr->Attribute.Data)
                    {
                        initVector = pIVAttr->Attribute.Data->Data;
                    }
                    else
                    {
                        rtn = CSSMERR_CSP_INVALID_ATTR_INIT_VECTOR;
                        goto FAIL_INIT;
                    }
                }
            }

            /* Get the Effective Bits attribute */
            pEffectiveBitsAttr = CSM_GetContextAttribute(pSession->pContext,
                                            CSSM_ATTRIBUTE_EFFECTIVE_BITS);

            /* Get the Rounds attribute */
            pRoundsAttr = CSM_GetContextAttribute(pSession->pContext,
                                            CSSM_ATTRIBUTE_ROUNDS);

            switch (pSession->pContext->AlgorithmType)
            {
            case CSSM_ALGID_RSA:

                algChooser = ENCRYPT_RSA_CHOOSER;

                /* Setup the key information */
                bsafeKey.data = keybuf;
                bsafeKey.len = sizeof(keybuf);
                if (pKey->KeyHeader.KeyClass == CSSM_KEYCLASS_PUBLIC_KEY)
                {
                    rtn = Cryp_CSSMRSAPublicKey_to_BSafeRSAPublicKey(pKey, 
                                        bsafeKey.data, (uint32 *)&bsafeKey.len);
                    if (rtn == CSSM_OK)
                    {
                        keyType = KI_RSAPublicBER;
                        algType = AI_PKCS_RSAPublic;
                    }
                    else
                    {
                        goto FAIL_INIT;
                    }
                }
                else if (pKey->KeyHeader.KeyClass == CSSM_KEYCLASS_PRIVATE_KEY)
                {
                    /* Get the key from the key manager */
                    loadData = Addin_GetLoadData();

                    if (loadData == NULL)
                    {
                        rtn = CSSMERR_CSP_INTERNAL_ERROR;
                        goto FAIL_INIT;
                    }

                    /* Fetch the access credentials for the key */
                    pCredAttr = CSM_GetContextAttribute( pSession->pContext,
                                                         CSSM_ATTRIBUTE_ACCESS_CREDENTIALS );
#if defined (COMPAQ_FIX) || defined (VMS)
                    /*
                     * If access credentials were not provided, leave it to
                     * later to decided whether they were needed.
                     */
                    if ( ( pCredAttr == NULL ) ||
                         ( pCredAttr->Attribute.AccessCredentials == NULL ) )
                    {
                      pAccessCred = NULL;
                    }
                    else
                    {
                      pAccessCred = pCredAttr->Attribute.AccessCredentials;
                    }
#else

                    if ( ( pCredAttr == NULL ) ||
                         ( pCredAttr->Attribute.AccessCredentials == NULL ) )
                    {
                        rtn = CSSMERR_CSP_INVALID_ACCESS_CREDENTIALS;
                        goto FAIL_INIT;
                    }
                    pAccessCred = pCredAttr->Attribute.AccessCredentials;
#endif
                    if ( ( pKey->KeyData.Length != sizeof(CSM_KM_KEY_HANDLE) ) ||
                                            port_IsBadReadPtr( pKey->KeyData.Data,
                                            sizeof(CSM_KM_KEY_HANDLE) ) )
                    {
                        rtn = CSSMERR_CSP_INVALID_KEY_REFERENCE;
                        goto FAIL_INIT;                     
                    }
#if defined (COMPAQ_FIX) || defined (VMS)
                    /* Setup key format to be the default for this algorithm */
                    kmFormat.Format = CSSM_KEYBLOB_RAW_FORMAT_PKCS1;
                    kmFormat.SigAlg = CSSM_ALGID_NONE;
#endif
                    rtn = CSM_KMFindKeyByHandle(loadData->hKeyFile,
                                               *(CSM_KM_KEY_HANDLE*)pKey->KeyData.Data,
                                               &kmFormat,
                                               pAccessCred,
                                               &priKey);

                    if (rtn == CSSM_OK)
                    {
                        rtn = Cryp_CSSMRSAPrivateKey_to_BSafeRSAPrivateKey(&priKey,
                                            bsafeKey.data, (uint32 *)&bsafeKey.len);
                        if (rtn == CSSM_OK)
                        {
                            keyType = KI_PKCS_RSAPrivateBER;
                            algType = AI_PKCS_RSAPrivate;
                        }
                        else
                        {
                            goto FAIL_INIT;
                        }
                    }
                    else
                    {
                        goto FAIL_INIT;
                    }
                }
                else
                {
                    rtn = CSSMERR_CSP_INVALID_KEY_CLASS;
                    goto FAIL_INIT;
                }

                /* Set the key info */
                if ((status = B_SetKeyInfo(pState->keyObject, keyType,
                                                (POINTER)&bsafeKey)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL_INIT;
                }
                /* Set the algorithm info */
                if ((status = B_SetAlgorithmInfo(pState->algObject, algType, 
                                                NULL_PTR)) != 0) 
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL_INIT;
                }

                break;

            case CSSM_ALGID_DES:

                algChooser = ENCRYPT_DES_CHOOSER;

                bsafeKey.data = pKey->KeyData.Data;
                bsafeKey.len = pKey->KeyData.Length;

                /* Set the key information */
                if ((status = B_SetKeyInfo(pState->keyObject, KI_DES8, 
                                                (POINTER)bsafeKey.data)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL_INIT;
                }

                if (pModeAttr == NULL)
                {
                    rtn = CSSMERR_CSP_INVALID_ATTR_MODE;
                    goto FAIL_INIT;
                }

                /* Set the mode and IV */
                switch (pModeAttr->Attribute.Uint32)
                {
                case CSSM_ALGMODE_CBCPadIV8:
                    algType = AI_DES_CBCPadIV8;
					/* Make sure the padding mode is acceptable */
					if ( ( Padding != CSSM_PADDING_PKCS5 ) &&
						 ( Padding != CSSM_PADDING_PKCS7 ) )
					{
						rtn = CSSMERR_CSP_INVALID_ATTR_PADDING;
						goto FAIL_INIT;						
					}
                    break;

                case CSSM_ALGMODE_CBC:
                case CSSM_ALGMODE_CBC_IV8:
                    algType = AI_DES_CBC_IV8;
                    break;

                default:
                    rtn = CSSMERR_CSP_INVALID_ATTR_MODE;
                    goto FAIL_INIT;
                }

                if ((status = B_SetAlgorithmInfo(pState->algObject, algType,
                                        initVector)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL_INIT;
                }
                break;

            case CSSM_ALGID_3DES_3KEY:

                algChooser = ENCRYPT_3DES_CHOOSER;

                bsafeKey.data = pKey->KeyData.Data;
                bsafeKey.len = pKey->KeyData.Length;

                /* Set the key information */
                if ((status = B_SetKeyInfo(pState->keyObject, KI_DES24Strong,
                                        (POINTER)bsafeKey.data)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL_INIT;
                }

                if (pModeAttr == NULL)
                {
                    rtn = CSSMERR_CSP_INVALID_ATTR_MODE;
                    goto FAIL_INIT;
                }

                /* Set the mode and IV */
                switch (pModeAttr->Attribute.Uint32)
                {
                case CSSM_ALGMODE_CBCPadIV8:
                    algType = AI_DES_EDE3_CBCPadIV8;
					/* Make sure the padding mode is acceptable */
					if ( ( Padding != CSSM_PADDING_PKCS5 ) &&
						 ( Padding != CSSM_PADDING_PKCS7 ) )
					{
						rtn = CSSMERR_CSP_INVALID_ATTR_PADDING;
						goto FAIL_INIT;						
					}
                    break;
                case CSSM_ALGMODE_CBC:
                case CSSM_ALGMODE_CBC_IV8:
                    algType = AI_DES_EDE3_CBC_IV8;
                    break;
                default:
                    rtn = CSSMERR_CSP_INVALID_ATTR_MODE;
                    goto FAIL_INIT;
                }

                if ((status = B_SetAlgorithmInfo(pState->algObject, algType,
                        initVector)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL_INIT;
                }
                break;

            case CSSM_ALGID_DESX:

                algChooser = ENCRYPT_DESX_CHOOSER;

                /* Set the key information */
                DESXKey.key = pKey->KeyData.Data;
                DESXKey.inputWhitener = pKey->KeyData.Data + CSP_DES_BLOCK;
                DESXKey.outputWhitener = (unsigned char*)NULL_PTR;
                if ((status = B_SetKeyInfo(pState->keyObject, KI_DESX, (POINTER)&DESXKey)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL_INIT;
                }

                if (pModeAttr == NULL)
                {
                    rtn = CSSMERR_CSP_INVALID_ATTR_MODE;
                    goto FAIL_INIT;
                }

                /* Set the mode and IV */
                switch (pModeAttr->Attribute.Uint32)
                {
                case CSSM_ALGMODE_CBCPadIV8:
                    algType = AI_DESX_CBCPadIV8;
					/* Make sure the padding mode is acceptable */
					if ( ( Padding != CSSM_PADDING_PKCS5 ) &&
						 ( Padding != CSSM_PADDING_PKCS7 ) )
					{
						rtn = CSSMERR_CSP_INVALID_ATTR_PADDING;
						goto FAIL_INIT;						
					}
                    break;

                case CSSM_ALGMODE_CBC:
                case CSSM_ALGMODE_CBC_IV8:
                    algType = AI_DESX_CBC_IV8;
                    break;

                default:
                    rtn = ConvertBsafeError(status);
                    goto FAIL_INIT;
                }

                if ((status = B_SetAlgorithmInfo(pState->algObject, algType,
                        initVector)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL_INIT;
                }

                break;

            case CSSM_ALGID_RC2:

                algChooser = ENCRYPT_RC2_CHOOSER;

                /* Set the key information */
                bsafeKey.data = pKey->KeyData.Data;
                bsafeKey.len = pKey->KeyData.Length;

                if ((status = B_SetKeyInfo(pState->keyObject, KI_Item,
                                            (POINTER)&bsafeKey)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL_INIT;
                }

                /* Set up the RC2 Parameters */
                if ((pEffectiveBitsAttr) && 
                            (pEffectiveBitsAttr->Attribute.Uint32 != 0))
                {
                    RC2Params.effectiveKeyBits = pEffectiveBitsAttr->Attribute.Uint32;
                }
                else
                {
                    RC2Params.effectiveKeyBits = 1024;
                }
                RC2Params.iv = initVector;
                
                if (pModeAttr == NULL)
                {
                    rtn = CSSMERR_CSP_INVALID_ATTR_MODE;
                    goto FAIL_INIT;
                }

                /* Set the mode and IV */
                switch (pModeAttr->Attribute.Uint32)
                {
                case CSSM_ALGMODE_CBCPadIV8:
                    algType = AI_RC2_CBCPad;
					/* Make sure the padding mode is acceptable */
					if ( ( Padding != CSSM_PADDING_PKCS5 ) &&
						 ( Padding != CSSM_PADDING_PKCS7 ) )
					{
						rtn = CSSMERR_CSP_INVALID_ATTR_PADDING;
						goto FAIL_INIT;						
					}
                    break;
                case CSSM_ALGMODE_CBC:
                case CSSM_ALGMODE_CBC_IV8:
                    algType = AI_RC2_CBC;
                    break;
                default:
                    rtn = CSSMERR_CSP_INVALID_ATTR_MODE;
                    goto FAIL_INIT;
                }
                if ((status = B_SetAlgorithmInfo(pState->algObject, algType, 
                                            (POINTER)&RC2Params)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL_INIT;
                }

                break;

            case CSSM_ALGID_RC4:

                algChooser = ENCRYPT_RC4_CHOOSER;

                /* Set the key information */
                bsafeKey.data = pKey->KeyData.Data;
                bsafeKey.len = pKey->KeyData.Length;

                if ((status = B_SetKeyInfo(pState->keyObject, KI_Item,
                                            (POINTER)&bsafeKey)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL_INIT;
                }

                /* Set the algorithm information */
                if ((status = B_SetAlgorithmInfo(pState->algObject, AI_RC4,
                                            NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL_INIT;
                }
                break;

            case CSSM_ALGID_RC5:
                algChooser = ENCRYPT_RC5_CHOOSER;

                /* Set the key information */
                bsafeKey.data = pKey->KeyData.Data;
                bsafeKey.len = pKey->KeyData.Length;

                if ((status = B_SetKeyInfo(pState->keyObject, KI_Item, 
                                            (POINTER)&bsafeKey)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL_INIT;
                }

                RC5Params.version = 0x10;
                if ((pRoundsAttr) && (pRoundsAttr->Attribute.Uint32 != 0))
                {
                    RC5Params.rounds = pRoundsAttr->Attribute.Uint32;
                }
                else
                {
                    RC5Params.rounds = 16;
                }
                RC5Params.wordSizeInBits = 32;
                RC5Params.iv = initVector;

                if (pModeAttr == NULL)
                {
                    rtn = CSSMERR_CSP_INVALID_ATTR_MODE;
                    goto FAIL_INIT;
                }

                /* Set the mode and IV */
                switch (pModeAttr->Attribute.Uint32)
                {
                case CSSM_ALGMODE_CBCPadIV8:
                    algType = AI_RC5_CBCPad;
					/* Make sure the padding mode is acceptable */
					if ( ( Padding != CSSM_PADDING_PKCS5 ) &&
						 ( Padding != CSSM_PADDING_PKCS7 ) )
					{
						rtn = CSSMERR_CSP_INVALID_ATTR_PADDING;
						goto FAIL_INIT;						
					}
                    break;
                case CSSM_ALGMODE_CBC:
                case CSSM_ALGMODE_CBC_IV8:
                    algType = AI_RC5_CBC;
                    break;
                default:
                    rtn = CSSMERR_CSP_INVALID_ATTR_MODE;
                    goto FAIL_INIT;
                }

                if ((status = B_SetAlgorithmInfo(pState->algObject, algType,
                                            (POINTER)&RC5Params)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL_INIT;
                }

                break;

            default:
                rtn = CSSMERR_CSP_INVALID_ALGORITHM;
                goto FAIL_INIT;
            }

            /* Initialize the encryption */
            if ((status = B_EncryptInit(pState->algObject, 
                            pState->keyObject, algChooser,
                            (A_SURRENDER_CTX*)NULL_PTR)) != 0)
            {
                rtn = ConvertBsafeError(status);
                goto FAIL_INIT;
            }

        }
    }

    return rtn;

FAIL_INIT:
    CleanState( pState );
    return rtn;
}

static CSSM_RETURN __ADDIN_CSM_EncryptDataUpdate( CSM_SESSION *pSession,
                                                 const CSSM_DATA *ClearBufs,
                                                 uint32 ClearBufCount,
                                                 CSSM_DATA_PTR CipherBufs,
                                                 uint32 CipherBufCount,
                                                 uint32 *bytesEncrypted,
                                                 CSSM_DATA_PTR RemData,
                                                 CSSM_BOOL bStaged )
{
    CSSM_RETURN rv = CSSM_OK;
    int status;
    BSAFECSP_STATE_PTR  pState;
    CSM_VECTOR InputVector;
    CSM_VECTOR OutputVector;
    CSSM_DATA OutputData = { 0, NULL };
    CSSM_DATA InputData = { 0, NULL };
    uint32 OutputLength = 0;


    /* Fetch the cipher object from the session */
    assert( pSession->pCryptoState &&
            ( pSession->uCryptoStateSize == sizeof(BSAFECSP_STATE) ) );
    pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
    assert( pState );

    /* Create the input and output vectors */
    if ( ( rv = CSM_CreateVector( pSession->hCsp,
                                  &InputVector,
                                  CSSM_TRUE,
                                  ClearBufs,
                                  ClearBufCount,
                                  NULL ) ) == CSSM_OK )
    {
        if ( ( rv = CSM_CreateVector( pSession->hCsp,
                                      &OutputVector,
                                      CSSM_FALSE,
                                      CipherBufs,
                                      CipherBufCount,
                                      RemData ) ) == CSSM_OK )
        {
			CSSM_QUERY_SIZE_DATA DataBlock;

			/* Calculate the output length and reserve the output space */
			CSM_GetVectorLength( &InputVector, &OutputLength );

			DataBlock.SizeInputBlock = OutputLength;
			rv = ADDIN_CSM_QuerySize (pSession, CSSM_TRUE, 1, &DataBlock);
			OutputLength = DataBlock.SizeOutputBlock;

			switch ( pSession->pContext->AlgorithmType )
			{
				case CSSM_ALGID_DES:
				case CSSM_ALGID_DESX:
				case CSSM_ALGID_3DES_3KEY:	
				case CSSM_ALGID_RC5:
				case CSSM_ALGID_RC4:
				case CSSM_ALGID_RC2:
					OutputLength += ( ( bStaged ) ? 0 : CSP_DES_BLOCK ); 			
				default:
					break;
			}
			
			            
            if ( ( rv = CSM_GuaranteeVectorSpace(
                                            &OutputVector,
                                            OutputLength ) ) == CSSM_OK )
            {
                /* Allocate a temporary buffer to hold the output */
                OutputData.Data = Addin_malloc( OutputLength, NULL );
                if ( OutputData.Data == NULL )
                {
                    ERR( rv = CSSMERR_CSP_MEMORY_ERROR );
                }
                else
                {
                    OutputData.Length = OutputLength;

                    /* Fetch the data to encrypt */
                    rv = CSM_GetBlockLengthVectorBuffer(
                                                &InputVector,
                                                1,
                                                &InputData );
                    if ( ( rv == CSSM_OK ) ||
                         ( rv == CSMR_NO_MORE_DATA ) )
                    {
                        /* Encrypt the data */
                        if ((status = B_EncryptUpdate(pState->algObject,
                                                      OutputData.Data,
                                                      (unsigned int *)&OutputData.Length,
                                                      OutputData.Length,
                                                      InputData.Data,
                                                      InputData.Length,
                                                      g_RandomNumGen,
                                                      (A_SURRENDER_CTX*)NULL_PTR)) == 0)
                        {
                            /* Add the output data to the vector */
                            CSM_PopulateVector( &OutputVector, &OutputData );

                            /* Is this the a single staged operation that needs
                             * to be finalized?
                             */
                            if ( !bStaged )
                            {
                                /* Reset the buffer size to its true length */
                                OutputData.Length = OutputLength;

                                /* Finalize the encryption */
                                if ( ( rv = ADDIN_CSM_EncryptDataFinal( 
                                                    pSession,
                                                    &OutputData ) ) == CSSM_OK )
                                {
                                    CSM_PopulateVector( &OutputVector,
                                                        &OutputData );
                                }
                            }

                            if ( rv == CSSM_OK )
                            {
                                CSM_GetVectorIndex( &OutputVector, bytesEncrypted );
                            }
                        }
                        else
                        {
                            rv = ConvertBsafeError(status);
                        }
                    } /* Fetch contiguous buffer */

                    /* Release the temporary buffer */
                    memset( OutputData.Data, 0, OutputLength );
                    Addin_free( OutputData.Data, NULL );
                } /* Allocate temp buffer */
            } /* Allocate output space */

            /* Shutdown the output vector */
            CSM_DeleteVector( &OutputVector, ( rv != CSSM_OK ) );
        } /* Create output vector */

        /* Shutdown the input vector */
        CSM_DeleteVector( &InputVector, CSSM_FALSE );
    } /* Create input vector */

    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_EncryptDataUpdate( CSM_SESSION *pSession,
                                                 const CSSM_DATA *ClearBufs,
                                                 uint32 ClearBufCount,
                                                 CSSM_DATA_PTR CipherBufs,
                                                 uint32 CipherBufCount,
                                                 uint32 *bytesEncrypted )
{
    return __ADDIN_CSM_EncryptDataUpdate( pSession,
                                          ClearBufs,
                                          ClearBufCount,
                                          CipherBufs,
                                          CipherBufCount,
                                          bytesEncrypted,
                                          NULL,
                                          CSSM_TRUE );
}

#if 0
    int                 status;
    CSSM_RETURN         rtn = CSSM_OK;
    CSM_VECTOR          InputVector, OutputVector;
    CSSM_DATA           InBuffer, OutBuffer;
    BSAFECSP_STATE_PTR  pState;

    pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
    
    if (!pState)
        return CSSMERR_CSP_INTERNAL_ERROR;

    /* Validate the buffers */
    rtn = CSM_ValidateInputBuffers(ClearBufs, ClearBufCount);
    
    if (rtn != CSSM_OK)
    {
        goto FAIL_UPDATE;
    }

    rtn = CSM_ValidateOutputBuffers(CipherBufs, CipherBufCount);
    if (rtn != CSSM_OK)
    {
        goto FAIL_UPDATE;
    }

    /* Create the vectors to work with */
    rtn = CSM_CreateVector(pSession->hCsp,
                           &InputVector,
                           CSSM_TRUE,
                           NULL,
                           ClearBufs,
                           ClearBufCount);
    if (rtn != CSSM_OK)
    {
        goto FAIL_UPDATE;
    }

    rtn = CSM_CreateVector(pSession->hCsp,
                           &OutputVector,
                           CSSM_FALSE,
                           NULL,
                           CipherBufs,
                           CipherBufCount);
    if (rtn != CSSM_OK)
    {
        CSM_DeleteVector( &InputVector, CSSM_TRUE );
        goto FAIL_UPDATE;
    }

    /* Fetch the complete input buffer */
    if ( (rtn = CSM_GetBlockLengthVectorBuffer(&InputVector,
                                               1,
                                               &InBuffer)) != CSSM_OK )
    {
        /* Take care of the 0 length buffer case */
        if (rtn == CSMR_NO_MORE_DATA)
        {
            rtn = CSSM_OK;
            InBuffer.Data = NULL;
            InBuffer.Length = 0;
        }
        else
        {
            CSM_DeleteVector( &InputVector, CSSM_TRUE );
            CSM_DeleteVector( &OutputVector, CSSM_TRUE );
            rtn =  CSSMERR_CSP_INPUT_LENGTH_ERROR;
            goto FAIL_UPDATE;
        }
    }
    
    *bytesEncrypted = 0;

    /* Setup the temp output, add an extra 8 bytes for a possible pad */
    OutBuffer.Length = InBuffer.Length + 8;
    OutBuffer.Data = Addin_calloc( (uint32)OutBuffer.Length, 1, NULL );
    if ( OutBuffer.Data == NULL )
    {
        CSM_DeleteVector( &InputVector, CSSM_TRUE );
        CSM_DeleteVector( &OutputVector, CSSM_TRUE );
        rtn = CSSMERR_CSP_MEMORY_ERROR;
        goto FAIL_UPDATE;
    }

    if ((status = B_EncryptUpdate(pState->algObject,
                                  OutBuffer.Data,
                                  (unsigned int *)&OutBuffer.Length,
                                  OutBuffer.Length,
                                  InBuffer.Data,
                                  InBuffer.Length,
                                  g_RandomNumGen,
                                  (A_SURRENDER_CTX*)NULL_PTR)) != 0)
    {
        Addin_free( OutBuffer.Data, NULL );
        CSM_DeleteVector( &InputVector, CSSM_TRUE );
        CSM_DeleteVector( &OutputVector, CSSM_TRUE );
        rtn = ConvertBsafeError(status);
        goto FAIL_UPDATE;
    }

    /* Insure that there is enough output space */
    rtn = CSM_GuaranteeVectorSpace( &OutputVector, OutBuffer.Length);
    if ( rtn != CSSM_OK )
    {
        Addin_free( OutBuffer.Data, NULL );
        CSM_DeleteVector( &InputVector, CSSM_TRUE );
        CSM_DeleteVector( &OutputVector, CSSM_TRUE );
        goto FAIL_UPDATE;
    }

    /* Populate the output buffer */
    rtn = CSM_PopulateVector( &OutputVector, &OutBuffer );
    if ( rtn != CSSM_OK )
    {
        Addin_free( OutBuffer.Data, NULL );
        CSM_DeleteVector( &InputVector, CSSM_TRUE );
        CSM_DeleteVector( &OutputVector, CSSM_TRUE );

        goto FAIL_UPDATE;
    }

    *bytesEncrypted = OutBuffer.Length;

    /* Clean up */
    Addin_free( OutBuffer.Data, NULL );

    CSM_DeleteVector( &InputVector, CSSM_TRUE );
    CSM_DeleteVector( &OutputVector, CSSM_FALSE );

    return rtn;

FAIL_UPDATE:
    CleanState( (BSAFECSP_STATE_PTR)pSession->pCryptoState );
    return rtn;
}
#endif

CSSM_RETURN CSSMAPI ADDIN_CSM_EncryptDataFinal( CSM_SESSION *pSession,
                                                CSSM_DATA *RemData )
{
    int                 status;
    uint32              length = RemData->Length;
    CSSM_RETURN         rtn = CSSM_OK;
    BSAFECSP_STATE_PTR  pState;

    pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
    
    if (!pState)
    {
        return CSSMERR_CSP_INTERNAL_ERROR;
    }

    /* Finalize encryption */
    if ((status = B_EncryptFinal(pState->algObject,
                                 RemData->Data,
                                 (unsigned int *)&length,  /* ifdef VMS apply cast */
                                 RemData->Length,
                                 g_RandomNumGen,
                                 (A_SURRENDER_CTX*)NULL_PTR)) != 0)
    {
        /* check to see we need to allocate space for the data */
        if (status == BE_OUTPUT_LEN)
        {
            if ((RemData->Data == NULL) && (RemData->Length == 0))
            {
                RemData->Data = App_Calloc(pSession->hCsp, length, 1);
                if (RemData->Data == NULL)
                {
                    return CSSMERR_CSP_MEMORY_ERROR;
                }

                /* Retry the call with allocated structure */
                status = B_EncryptFinal(pState->algObject,
                                        RemData->Data,
                                        (unsigned int *)&length,  /* ifdef VMS apply cast */
                                        length,
                                        g_RandomNumGen,
                                        (A_SURRENDER_CTX*)NULL_PTR);
                if (status != 0)
                {
                    App_Free(pSession->hCsp, RemData->Data);
                }
            }
        }

        if (status != 0)
        {
            rtn = ConvertBsafeError(status);
        }
    }

    if (status == 0)
    {
        RemData->Length = length;
    }


    CleanState(pState);

    return rtn;
}

void ADDIN_CSM_EncryptDataError( CSM_SESSION *pSession )
{
    BSAFECSP_STATE_PTR  pState = NULL;

    /* Get the pState->algObject object from the algorithm state */
    if (pSession)
    {
        pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
        if (pState)
        {
            CleanState(pState);
        }
    }
}

CSSM_RETURN CSSMAPI ADDIN_CSM_EncryptData( CSM_SESSION *pSession,
                                           const CSSM_DATA *ClearBufs,
                                           uint32 ClearBufCount,
                                           CSSM_DATA_PTR CipherBufs,
                                           uint32 CipherBufCount,
                                           uint32 *bytesEncrypted,
                                           CSSM_DATA_PTR RemData,
                                           CSSM_PRIVILEGE Privilege )
{
    CSSM_RETURN     rtn = CSSM_OK;

    rtn = ADDIN_CSM_EncryptDataInit(pSession, Privilege);

    if (rtn == CSSM_OK)
    {
        rtn = __ADDIN_CSM_EncryptDataUpdate(pSession, ClearBufs, ClearBufCount,
                                          CipherBufs, CipherBufCount,
                                          bytesEncrypted,
                                          RemData,
                                          CSSM_FALSE );
    }

    CleanState( (BSAFECSP_STATE_PTR)pSession->pCryptoState );

    return rtn;
}
