/*-----------------------------------------------------------------------
 * File: CSP_SIGN.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */
/*
** OpenVMS History
 *
 *    01-Nov-1999	Carol Kelly
 *		o Merge in changes from Unix code base
 *
 */

/*
 * File Description: Contains signature operations for the CSP.
 */

#include "geninc.h"

extern B_ALGORITHM_OBJ  g_RandomNumGen;

/* BSafe Signing Algorithm Choosers */
B_ALGORITHM_METHOD  *SIGN_DSA_CHOOSER[] = {
    &AM_SHA,
    &AM_DSA_SIGN,
    (B_ALGORITHM_METHOD*)NULL_PTR
};
B_ALGORITHM_METHOD  *SIGN_MD5WithRSA_CHOOSER[] = {
    &AM_MD5,
    &AM_RSA_CRT_ENCRYPT,
    (B_ALGORITHM_METHOD*)NULL_PTR
};
B_ALGORITHM_METHOD  *SIGN_MD2WithRSA_CHOOSER[] = {
    &AM_MD2,
    &AM_RSA_CRT_ENCRYPT,
    (B_ALGORITHM_METHOD*)NULL_PTR
};
B_ALGORITHM_METHOD  *SIGN_SHA1WithRSA_CHOOSER[] = {
    &AM_SHA,
    &AM_RSA_CRT_ENCRYPT,
    (B_ALGORITHM_METHOD*)NULL_PTR
};

static CSSM_ALGORITHMS ValidAlgorithms[] = 
									{   CSSM_ALGID_SHA1WithDSA,
										CSSM_ALGID_MD2WithRSA,
										CSSM_ALGID_MD5WithRSA,
										CSSM_ALGID_SHA1WithRSA};										


static const uint32 NumValidAlgorithms = sizeof(ValidAlgorithms) / 
                                            sizeof(CSSM_ALGORITHMS);

CSSM_RETURN bsafe_ValidateSignVrfyAlg (CSSM_ALGORITHMS Alg)

{
	uint32 i;

	for (i = 0; i < NumValidAlgorithms; i++)
	{
		if (Alg == ValidAlgorithms[i])
			break;
	}

	if (i == NumValidAlgorithms)	              
		return CSSMERR_CSP_INVALID_ALGORITHM;

	return CSSM_OK;
}


CSSM_RETURN CSSMAPI ADDIN_CSM_SignData( CSM_SESSION *pSession,
                                        const CSSM_DATA *DataBufs,
                                        uint32 DataBufCount,
                                        CSSM_ALGORITHMS DigestAlgorithm,
                                        CSSM_DATA_PTR Signature )
{
    CSSM_RETURN  rtn = CSSM_OK;

#ifndef VMS /* COMPAQ_FIX */
	/* Make sure that only combo-ops are requested */
    if ( DigestAlgorithm != CSSM_ALGID_NONE )
    {
        rtn = CSSMERR_CSP_INVALID_DIGEST_ALGORITHM;
    }
#else
    /* Make sure that digest algorithm are not requested with combo-ops */
    if ((pSession->pContext->AlgorithmType != CSSM_ALGID_RSA) &&
	(DigestAlgorithm != CSSM_ALGID_NONE))
    {
        rtn = CSSMERR_CSP_INVALID_DIGEST_ALGORITHM;
    }
    else
    {       
      const CSSM_CONTEXT_ATTRIBUTE *pModeAttr = NULL;
      uint32                       encrypt_only = 0;
      CSSM_DATA                    ClearBufs;
      CSSM_DATA                    remData;
      CSSM_DATA                    CipherBufs;
      uint32                       bytesEncrypted;
 
      /*
       * Check for special case. If the signature algorithm is RSA and the
       * mode attribute is NONE, then digest argument is ignored and the data
       * is just encrypted.
       */
      if ((pSession->pContext->AlgorithmType == CSSM_ALGID_RSA))
      {
        pModeAttr = CSM_GetContextAttribute(pSession->pContext,
                                            CSSM_ATTRIBUTE_MODE);
        if (pModeAttr != NULL &&
            pModeAttr->Attribute.Uint32 == CSSM_ALGMODE_NONE)
        {
          encrypt_only = 1;
        }
      }

      if (encrypt_only == 1)
      {
        /* Only one buffer allowed for encryption only */
        if (DataBufCount != 1)
        {
	  rtn = CSSMERR_CSP_VECTOR_OF_BUFS_UNSUPPORTED;
        }
        else
        {
	  ClearBufs.Data = DataBufs->Data;
	  ClearBufs.Length = DataBufs->Length;
	  CipherBufs.Data = Signature->Data;
	  CipherBufs.Length = Signature->Length;
	  remData.Data = NULL;
	  remData.Length = 0;
	  rtn = ADDIN_CSM_EncryptData(pSession,
				      &ClearBufs,
				      1,
				      &CipherBufs,
				      1,
				      &bytesEncrypted,
				      &remData,
				      (unsigned __int64)NULL); /* ifdef VMS apply cast */
	  if (rtn == CSSM_OK && remData.Length !=0)
	  {
	    /* We should not have a remainder */
	    rtn = CSSMERR_CSP_OUTPUT_LENGTH_ERROR;

	    /* Free any allocated buffers */
	    if (remData.Data != NULL)
            {
	      App_Free(pSession->hCsp, remData.Data);
	    }
	    if (Signature->Data == NULL && CipherBufs.Data != NULL)
            {
	      App_Free(pSession->hCsp, CipherBufs.Data);
	    }
	  }
	  /* If automatic allocation was specified then setup the return */
	  if (Signature->Data == NULL)
          {
	    Signature->Data = CipherBufs.Data;
	    Signature->Length = CipherBufs.Length;
	  }
        }
      }
      else
      {
#endif
		rtn = ADDIN_CSM_SignDataInit(pSession);

		if (rtn == CSSM_OK)
		{
			rtn = ADDIN_CSM_SignDataUpdate(pSession, DataBufs, DataBufCount);
			if (rtn == CSSM_OK)
			{
				rtn = ADDIN_CSM_SignDataFinal(pSession, Signature);
			}
		}

		CleanState( (BSAFECSP_STATE_PTR)pSession->pCryptoState );

#ifdef VMS /* COMPAQ_FIX */
      }
#endif
	}
    
    return rtn;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_SignDataInit( CSM_SESSION *pSession )
{
    int                             status;
    uint8                           prikeybuf[MAX_KEYBUF_LEN];
    CSSM_KEY                        *pKey, priKey;
    ITEM                            priKeyBER;
    BSAFECSP_STATE_PTR              pState = NULL;
    CSSM_RETURN                     rtn = CSSM_OK;
    const CSSM_CONTEXT_ATTRIBUTE    *pKeyAttr = NULL;
    CSM_KM_FORMAT_CONTEXT           kmFormat;
    const ADDIN_LOAD_DATA           *loadData;
    const CSSM_CONTEXT_ATTRIBUTE    *pCredAttr = NULL;
    const CSSM_ACCESS_CREDENTIALS   *pAccessCred = NULL;

    /* Initialize */
    priKeyBER.data = prikeybuf;
    priKeyBER.len = sizeof(prikeybuf);

    priKey.KeyData.Length = 0;
    priKey.KeyData.Data = NULL;

    /* Fetch the signature key */
    pKeyAttr = CSM_GetContextAttribute(pSession->pContext, CSSM_ATTRIBUTE_KEY);
    if ((pKeyAttr == NULL) || (pKeyAttr->Attribute.Key == NULL))
    {
        rtn = CSSMERR_CSP_MISSING_ATTR_KEY;
        goto FAIL;
    }

    pKey = pKeyAttr->Attribute.Key;

	if ( pKey == NULL )
    {
        rtn = CSSMERR_CSP_INVALID_ATTR_KEY;
        goto FAIL;
    }
    else
    {
		if ((pKey->KeyData.Data == NULL) || (pKey->KeyData.Length == 0))
		{
			rtn = CSSMERR_CSP_INVALID_KEY;
			goto FAIL;
		}
	}

    /* Verify the correct key was passed in */
    if (pKey->KeyHeader.KeyClass != CSSM_KEYCLASS_PRIVATE_KEY)
    {
        rtn = CSSMERR_CSP_INVALID_KEY_CLASS;
        goto FAIL;
	}

	if ((rtn = bsafe_ValidateSignVrfyAlg (pSession->pContext->AlgorithmType)) != CSSM_OK)
	{                
		rtn = CSSMERR_CSP_INVALID_ALGORITHM;
		goto FAIL;
	}

    if (pKey->KeyHeader.AlgorithmId == CSSM_ALGID_RSA)
    {
        if ((pSession->pContext->AlgorithmType != CSSM_ALGID_MD2WithRSA) &&
            (pSession->pContext->AlgorithmType != CSSM_ALGID_MD5WithRSA) &&
            (pSession->pContext->AlgorithmType != CSSM_ALGID_SHA1WithRSA))
        {
            rtn = CSSMERR_CSP_ALGID_MISMATCH;
            goto FAIL;
        }
    }
    else if (pKey->KeyHeader.AlgorithmId == CSSM_ALGID_DSA)
    {
        if (pSession->pContext->AlgorithmType != CSSM_ALGID_SHA1WithDSA)
        {
            rtn = CSSMERR_CSP_ALGID_MISMATCH;
            goto FAIL;
        }
    }

	if (!CHECKBITS(pKey->KeyHeader.KeyUsage,
							  CSSM_KEYUSE_SIGN ) &&
				  !CHECKBITS(pKey->KeyHeader.KeyUsage,
						  CSSM_KEYUSE_ANY ) )
	{                
		rtn = CSSMERR_CSP_KEY_USAGE_INCORRECT;
		goto FAIL;
	}

    /* Fetch the access credentials for the key */
    pCredAttr = CSM_GetContextAttribute( pSession->pContext,
                                         CSSM_ATTRIBUTE_ACCESS_CREDENTIALS );
#ifdef VMS  /* COMPAQ_FIX */
    /*
     * If access credentials were not provided, leave it to
     * later to decided whether they were needed.
     */
    if ( ( pCredAttr == NULL ) ||
	 ( pCredAttr->Attribute.AccessCredentials == NULL ) )
    {
      pAccessCred = NULL;
    }
    else
    {
      pAccessCred = pCredAttr->Attribute.AccessCredentials;
    }
#else
    if ( ( pCredAttr == NULL ) || ( pCredAttr->Attribute.AccessCredentials == NULL ) )
    {
        rtn = CSSMERR_CSP_INVALID_ACCESS_CREDENTIALS;
        goto FAIL;
    }
    pAccessCred = pCredAttr->Attribute.AccessCredentials;
#endif

    /* Choose the correct key format */
    kmFormat.SigAlg = CSSM_ALGID_NONE;
    switch ( pSession->pContext->AlgorithmType )
    {
    case CSSM_ALGID_SHA1WithDSA:
        {
            kmFormat.Format = CSSM_KEYBLOB_RAW_FORMAT_FIPS186;
            break;
        }
    case CSSM_ALGID_MD2WithRSA:
    case CSSM_ALGID_MD5WithRSA:
    case CSSM_ALGID_SHA1WithRSA:
        {
            kmFormat.Format = CSSM_KEYBLOB_RAW_FORMAT_PKCS1;
            break;
        }
    default:
        {
            rtn = CSSMERR_CSP_INVALID_ALGORITHM;
            goto FAIL;
        }
    }

    /* Find the key in the key file */
    loadData = Addin_GetLoadData();

    if (loadData == NULL)
    {
        rtn = CSSMERR_CSP_INTERNAL_ERROR;
        goto FAIL;
    }

    if ( ( pKey->KeyData.Length != sizeof(CSM_KM_KEY_HANDLE) ) ||
                                        port_IsBadReadPtr( pKey->KeyData.Data,
                                        sizeof(CSM_KM_KEY_HANDLE) ) )
    {
        rtn = CSSMERR_CSP_INVALID_KEY_REFERENCE;
        goto FAIL;                      
    }
    rtn = CSM_KMFindKeyByHandle( loadData->hKeyFile,
                                 *(CSM_KM_KEY_HANDLE*)pKey->KeyData.Data,
                                 &kmFormat,
                                 pAccessCred,
                                 &priKey);
    
    if (rtn == CSSM_OK)
    {
        /* Allocate space to store the algorithm map pointer */
        rtn = CSM_AllocateSessionState( pSession, sizeof(BSAFECSP_STATE) );
    }
    
    if ( rtn != CSSM_OK )
    {
        rtn = CSSMERR_CSP_INTERNAL_ERROR;
    }
    else
    {
        pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
        memset(pState, 0, sizeof(BSAFECSP_STATE));

        /* Create the digest algorithm object */
        if ((status = B_CreateAlgorithmObject(&pState->algObject)) != 0)
        {
            rtn = ConvertBsafeError(status);
        }
        else
        {
            if ((status = B_CreateKeyObject(&pState->keyObject)) != 0)
            {
                rtn = ConvertBsafeError(status);
                goto FAIL;
            }

            /* Set and Initialize the signing algorithm */
            switch (pSession->pContext->AlgorithmType) {
            case CSSM_ALGID_SHA1WithDSA:
                /* Convert key to BSafe Key Blob Format */
                if (CSSM_OK != Cryp_CSSMDSAPrivateKey_to_BSafeDSAPrivateKey(&priKey,
                                    priKeyBER.data, (uint32 *)&priKeyBER.len))
                {
                    rtn = CSSMERR_CSP_INVALID_KEY;
                    goto FAIL;
                }

                if ((status = B_SetKeyInfo(pState->keyObject, KI_DSAPrivateBER,
                                    (POINTER)&priKeyBER)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_SetAlgorithmInfo(pState->algObject,
                                    AI_DSAWithSHA1, NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_SignInit(pState->algObject, pState->keyObject,
                                    SIGN_DSA_CHOOSER,
                                    (A_SURRENDER_CTX*)NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                break;
            case CSSM_ALGID_MD2WithRSA:
                /* Convert key to BSafe Key Blob Format */
                if (CSSM_OK != Cryp_CSSMRSAPrivateKey_to_BSafeRSAPrivateKey(&priKey,
                                    priKeyBER.data, (uint32 *)&priKeyBER.len))
                {
                    rtn = CSSMERR_CSP_INVALID_KEY;
                    goto FAIL;
                }

                if ((status = B_SetKeyInfo(pState->keyObject,
                                    KI_PKCS_RSAPrivateBER,
                                    (POINTER)&priKeyBER)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_SetAlgorithmInfo(pState->algObject,
                                    AI_MD2WithRSAEncryption, NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_SignInit(pState->algObject, pState->keyObject,
                                    SIGN_MD2WithRSA_CHOOSER,
                                    (A_SURRENDER_CTX*)NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                break;
            case CSSM_ALGID_MD5WithRSA:
                /* Convert key to BSafe Key Blob Format */
                if (CSSM_OK != Cryp_CSSMRSAPrivateKey_to_BSafeRSAPrivateKey(&priKey,
                                    priKeyBER.data, (uint32 *)&priKeyBER.len))
                {
                    rtn = CSSMERR_CSP_INVALID_KEY;
                    goto FAIL;
                }

                if ((status = B_SetKeyInfo(pState->keyObject,
                                    KI_PKCS_RSAPrivateBER,
                                    (POINTER)&priKeyBER)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_SetAlgorithmInfo(pState->algObject,
                                    AI_MD5WithRSAEncryption, NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_SignInit(pState->algObject, pState->keyObject,
                                    SIGN_MD5WithRSA_CHOOSER,
                                    (A_SURRENDER_CTX*)NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                break;
            case CSSM_ALGID_SHA1WithRSA:
                /* Convert key to BSafe Key Blob Format */
                if (CSSM_OK != Cryp_CSSMRSAPrivateKey_to_BSafeRSAPrivateKey(&priKey,
                                    priKeyBER.data, (uint32 *)&priKeyBER.len))
                {
                    rtn = CSSMERR_CSP_INVALID_KEY;
                    goto FAIL;
                }

                if ((status = B_SetKeyInfo(pState->keyObject,
                                    KI_PKCS_RSAPrivateBER,
                                    (POINTER)&priKeyBER)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_SetAlgorithmInfo(pState->algObject,
                                    AI_SHA1WithRSAEncryption, NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_SignInit(pState->algObject, pState->keyObject,
                                    SIGN_SHA1WithRSA_CHOOSER,
                                    (A_SURRENDER_CTX*)NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                break;
            default:
                rtn = CSSMERR_CSP_INVALID_ALGORITHM;
                goto FAIL;
            }
        }

        /* Clean up the private key */
        if (priKey.KeyData.Data != NULL)
        {
            /* Clean up the private key */
            memset(priKey.KeyData.Data, 0, priKey.KeyData.Length);
            Addin_free( priKey.KeyData.Data, NULL );
        }
        return rtn;
    }

FAIL:

    /* Clean up the private key */
    if (priKey.KeyData.Data != NULL)
    {
        /* Clean up the private key */
        memset(priKey.KeyData.Data, 0, priKey.KeyData.Length);
        MAF_DeleteOutputBuffer(pSession->hCsp, &(priKey.KeyData), CSSM_TRUE);
    }
    CleanState(pState);
    return rtn;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_SignDataUpdate( CSM_SESSION *pSession,
                                              const CSSM_DATA *DataBufs,
                                              uint32 DataBufCount )
{
    int                 status;
    uint32              i = 0;
    BSAFECSP_STATE_PTR  pState = NULL;
    CSSM_RETURN         rtn = CSSM_OK;

    /* Get the digital signer object from the algorithm state */
    pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
    if (!pState)
    {
        rtn = CSSMERR_CSP_INTERNAL_ERROR;
    }
    else
    {
        /* Update */
        for (i = 0; i < DataBufCount; i++)
        {
            if ((status = B_SignUpdate(pState->algObject, DataBufs[i].Data,
                                DataBufs[i].Length,
                                (A_SURRENDER_CTX*)NULL_PTR)) != 0)
            {
                rtn = ConvertBsafeError(status);
                CleanState(pState);
            }
        }
    }

    return rtn;

}

CSSM_RETURN CSSMAPI ADDIN_CSM_SignDataFinal( CSM_SESSION *pSession,
                                             CSSM_DATA *Signature )
{
    int                             status;
    BSAFECSP_STATE_PTR              pState = NULL;
    CSSM_KEY_PTR                    pKey;
    const CSSM_CONTEXT_ATTRIBUTE    *pKeyAttr = NULL;
    uint32                          nModulusBytes=0, nModulusBits=0;
    CSSM_BOOL                       allocate = CSSM_FALSE;
    CSSM_RETURN                     rtn = CSSM_OK;

    /* Get the digital signer object from the algorithm state */
    pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
    if (!pState)
    {
        rtn = CSSMERR_CSP_INTERNAL_ERROR;
    }
    else
    {
        /* Signature buffer should be as big as the key modulus in bytes. Do
           not know how to get key modulus from the BSafe KeyBlob. Temporary
           solution use the SizeInBits in the key header */

        /* Get the raw key from the context */
        pKeyAttr = CSM_GetContextAttribute(pSession->pContext, CSSM_ATTRIBUTE_KEY);

        if (pKeyAttr == NULL)
        {
            rtn = CSSMERR_CSP_INVALID_ATTR_KEY;
        }
        else
        {
            pKey = pKeyAttr->Attribute.Key;

            nModulusBits = pKey->KeyHeader.LogicalKeySizeInBits;
            nModulusBytes = nModulusBits/8;

            rtn = MAF_CreateOutputBuffer(pSession->hCsp, nModulusBytes, 
                                      Signature, &allocate);


            if (rtn == CSSM_OK)
            {
                if ((status = B_SignFinal(pState->algObject, Signature->Data,
                                    (unsigned int *)&Signature->Length,
                                    Signature->Length,
                                    g_RandomNumGen,
                                    (A_SURRENDER_CTX*)NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                }
            }
        }
    }

    /* Destroy the signature object */
    CleanState(pState);

    return rtn;
}

void ADDIN_CSM_SignDataError( CSM_SESSION *pSession )
{
    BSAFECSP_STATE_PTR  pState = NULL;

    if (pSession)
    {
        /* Get the pState->algObject object from the algorithm state */
        pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
        if (pState)
        {
            CleanState(pState);
        }
    }
}
