/*-----------------------------------------------------------------------
 * File: KMGRCALL.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "geninc.h"

#define KMGR_KEY_LENGTH_BYTES       (24)
#define KMGR_IV_LENGTH_BYTES        (8)
#define KMGR_ENCRYPT_ALG            CSSM_ALGID_3DES_3KEY
#define KMGR_ENCRYPT_MODE           CSSM_ALGMODE_CBCPadIV8
#define KMGR_ENCRYPT_BLOCK_SIZE     (8)

extern B_ALGORITHM_METHOD *ENCRYPT_3DES_CHOOSER[];
extern B_ALGORITHM_METHOD *DECRYPT_3DES_CHOOSER[];

CSSM_RETURN CSSMAPI CSM_KMCallout_GetWrapParams(
                        uint32 *pKeyLengthBytes,
                        uint32 *pIVLengthBytes,
                        CSSM_ALGORITHMS *pAlgId,
                        CSSM_ENCRYPT_MODE *pAlgMode )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Fill in parameters for Triple-DES-CBC */
    *pKeyLengthBytes = KMGR_KEY_LENGTH_BYTES;
    *pIVLengthBytes  = KMGR_IV_LENGTH_BYTES;
    *pAlgId          = KMGR_ENCRYPT_ALG;
    *pAlgMode        = KMGR_ENCRYPT_MODE;

    return rv;
}

CSSM_RETURN CSSMAPI CSM_KMCallout_GetRandomBytes(
                        uint32 Length,
                        uint8 *RandomBytes )
{
    CSSM_RETURN rv = CSSM_OK;

    rv = csp_GenerateRandomBytes( Length, RandomBytes );

    return rv;
}

CSSM_RETURN CSSMAPI CSM_KMCallout_EncryptData(
                        uint8 *EncryptKey,
                        uint8 *IV,
                        const CSSM_DATA *Input,
                        CSSM_DATA_PTR Output )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_BOOL bAllocated = CSSM_TRUE;
    uint32 OutputLength;
    uint32 RemainingBytes;
    uint32 ResultBytes;
    uint8 *DataPtr;

    int             bsRv;
    B_ALGORITHM_OBJ bsAlgObject;
    B_KEY_OBJ       bsKeyObject;

    /* Calculate the output size */
    /* OutputLength = NumberOfEncryptedBlocks * BlockSize */
    OutputLength = ( ( Input->Length / KMGR_ENCRYPT_BLOCK_SIZE ) + 1 ) *
                                                    KMGR_ENCRYPT_BLOCK_SIZE;

    /* Allocate the output buffer memory */
    Output->Data = NULL;
    Output->Length = 0;
    if ( ( rv = MAF_CreateOutputBuffer( MAF_ALLOCATE_LOCAL,
                                        OutputLength,
                                        Output,
                                        &bAllocated ) ) == CSSM_OK )
    {
        /* Create the encryption algorithm object */
        if ( ( ( bsRv = B_CreateAlgorithmObject( &bsAlgObject ) ) == 0 ) &&
             ( ( bsRv = B_CreateKeyObject( &bsKeyObject ) ) == 0 ) &&
             ( ( bsRv = B_SetKeyInfo( bsKeyObject,
                                      KI_DES24Strong,
                                      EncryptKey ) ) == 0 ) &&
             ( ( bsRv = B_SetAlgorithmInfo( bsAlgObject,
                                            AI_DES_EDE3_CBCPadIV8,
                                            IV ) ) == 0 ) )
        {
            /* Initialize the encryption */
            if ( ( bsRv = B_EncryptInit( bsAlgObject, 
                                         bsKeyObject,
                                         ENCRYPT_3DES_CHOOSER,
                                         (A_SURRENDER_CTX*)NULL_PTR ) ) == 0 )
            {
                DataPtr = Output->Data;
                RemainingBytes = Output->Length;
                ResultBytes = 0;

                /* Feed in the data to encrypt */
                if ( ( bsRv = B_EncryptUpdate(
                                        bsAlgObject,
                                        DataPtr,
                                        (unsigned int *)&ResultBytes,  /* ifdef VMS apply cast */
                                        RemainingBytes,
                                        Input->Data,
                                        Input->Length,
                                        g_RandomNumGen,
                                        (A_SURRENDER_CTX*)NULL_PTR ) ) == 0 )
                {
                    DataPtr = &Output->Data[ResultBytes];
                    RemainingBytes -= ResultBytes;

                    bsRv = B_EncryptFinal( bsAlgObject,
                                           DataPtr,
                                           (unsigned int *)&ResultBytes,  /* ifdef VMS apply cast */
                                           RemainingBytes,
                                           g_RandomNumGen,
                                           (A_SURRENDER_CTX*)NULL_PTR );
                } /* B_EncryptUpdate */
            } /* B_EncryptInit */

            B_DestroyAlgorithmObject( &bsAlgObject );
            B_DestroyKeyObject( &bsKeyObject );
        } /* Algorithm setup */

        if ( bsRv != 0 )
        {
            rv = ConvertBsafeError( bsRv );
        }
    } /* MAF_CreateOutputBuffer */


    return rv;
}

CSSM_RETURN CSSMAPI CSM_KMCallout_DecryptData(
                        uint8 *DecryptKey,
                        uint8 *IV,
                        const CSSM_DATA *Input,
                        CSSM_DATA_PTR Output )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_BOOL bAllocated = CSSM_FALSE;
    uint32 RemainingBytes;
    uint32 ResultBytes;
    uint8 *DataPtr;

    int             bsRv;
    B_ALGORITHM_OBJ bsAlgObject;
    B_KEY_OBJ       bsKeyObject;
#ifndef VMS
#pragma message( "warning: Implement real decryption for key manager callout" )
#endif
    /* Allocate the output buffer memory */
    Output->Data = NULL;
    Output->Length = 0;
    if ( ( rv = MAF_CreateOutputBuffer( MAF_ALLOCATE_LOCAL,
                                        Input->Length,
                                        Output,
                                        &bAllocated ) ) == CSSM_OK )
    {
        /* Create the encryption algorithm object */
        if ( ( ( bsRv = B_CreateAlgorithmObject( &bsAlgObject ) ) == 0 ) &&
             ( ( bsRv = B_CreateKeyObject( &bsKeyObject ) ) == 0 ) &&
             ( ( bsRv = B_SetKeyInfo( bsKeyObject,
                                      KI_DES24Strong,
                                      DecryptKey ) ) == 0 ) &&
             ( ( bsRv = B_SetAlgorithmInfo( bsAlgObject,
                                            AI_DES_EDE3_CBCPadIV8,
                                            IV ) ) == 0 ) )
        {
            /* Initialize the encryption */
            if ( ( bsRv = B_DecryptInit( bsAlgObject, 
                                         bsKeyObject,
                                         DECRYPT_3DES_CHOOSER,
                                         (A_SURRENDER_CTX*)NULL_PTR ) ) == 0 )
            {
                DataPtr = Output->Data;
                RemainingBytes = Output->Length;
                ResultBytes = 0;

                /* Feed in the data to encrypt */
                if ( ( bsRv = B_DecryptUpdate(
                                        bsAlgObject,
                                        DataPtr,
                                        (unsigned int *)&ResultBytes,   /* ifdef VMS apply cast */
                                        RemainingBytes,
                                        Input->Data,
                                        Input->Length,
                                        g_RandomNumGen,
                                        (A_SURRENDER_CTX*)NULL_PTR ) ) == 0 )
                {
                    DataPtr = &Output->Data[ResultBytes];
                    RemainingBytes -= ResultBytes;
                    Output->Length = ResultBytes;

                    if ( ( bsRv = B_DecryptFinal( bsAlgObject,
                                                  DataPtr,
                                                  (unsigned int *)&ResultBytes,   /* ifdef VMS apply cast */
                                                  RemainingBytes,
                                                  g_RandomNumGen,
                                                  (A_SURRENDER_CTX*)NULL_PTR ) ) == 0 )
                    {
                        Output->Length += ResultBytes;
                    }
                } /* B_EncryptUpdate */
            } /* B_EncryptInit */

            B_DestroyAlgorithmObject( &bsAlgObject );
            B_DestroyKeyObject( &bsKeyObject );
        } /* Algorithm setup */

        if ( bsRv != 0 )
        {
            rv = ConvertBsafeError( bsRv );
        }
    } /* MAF_CreateOutputBuffer */

    return rv;
}
