/*-----------------------------------------------------------------------
 * File: MOD_INSTALL.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */
/*
** OpenVMS History
**
 *    01-Nov-1999	Carol Kelly
 *		o Merge in changes from Unix code base
 *
 */

#include <sys/types.h>
#include <sys/stat.h>

#include "geninc.h"

#include "mds.h"
#include "mds_schema.h"
#include "mds_util.h"

#include "intel/installdefs.h"

/* Number of MDS update operations that will be remembered by mds_util for
 * error clean-up during installation.
 */
#define MAF_BSAFE_MDS_COMMIT_SIZE   150

/* Capability group code markers for run-time processor function __CalcGCodes */
#define NEW_CAP     ( 1 )
#define CAP_ATTR    ( 0 )

/* Generic attribute declaration macros */
/* Use to indicate the start of a new capability */
#define DECLARE_CAP( _usee_, _cctype_, _algid_, _attr_, _val_, _desc_ ) \
    { ADDIN_GUID_INITIALIZER, \
      ADDIN_CSP_SUBSERVICE_ID, \
      CSSM_USEE_##_usee_, \
      CSSM_ALGCLASS_##_cctype_, \
      CSSM_ALGID_##_algid_, \
      NEW_CAP, \
      CSSM_ATTRIBUTE_##_attr_, \
      _val_, _desc_ }
/* Use to indicate addition of another attribute to the current capability */
#define DECLARE_ATR( _usee_, _cctype_, _algid_, _attr_, _val_, _desc_ ) \
    { ADDIN_GUID_INITIALIZER, \
      ADDIN_CSP_SUBSERVICE_ID, \
      CSSM_USEE_##_usee_, \
      CSSM_ALGCLASS_##_cctype_, \
      CSSM_ALGID_##_algid_, \
      CAP_ATTR, \
      CSSM_ATTRIBUTE_##_attr_, \
      _val_, _desc_ }
/* Macro to create the initializer for a single value attribute */
#define SINGLE_VAL( _val_ ) \
    { _val_, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, 1
/* Macro to create the initializer for a triple value attribute */
#define TRIPLE_VAL( _val0_, _val1_, _val2_ ) \
    { _val0_, _val1_, _val2_, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, 3
/* Macro to create the initilizer for a range value attribute */
#define RANGE_VAL( _min_, _max_ ) \
    { _min_, _max_, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, 2
/* Macro to create the initilizer for an algorithm mode value attribute */
#define MODE_VAL( _mode_ ) \
    { CSSM_ALGMODE_##_mode_, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, 1
/* Macro to create the initilizer for an algorithm mode value attribute */
#define PADDING_VAL( _mode_ ) \
    { CSSM_PADDING_##_mode_, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, 1

/* Static capability table listing the attributes that need to be inserted
 * into the MDS.
 */
/*
 * The function ADDIN_CSM_PassThrough is tightly coupled with this table. If
 * a change is made to this table this function should be review.
 */
static MDSU_CDSA_CSP_CAPABILITY_SCHEMA CapabilityTable[] =
{
    /*           USEE       AlgClass    AlgID       Attribute       Attribute Value     Description */
    /* MD2 Random */
    DECLARE_CAP( NONE,      RANDOMGEN,  MD2Random,  NONE,           SINGLE_VAL( 0 ),        "MD2 Random" ),
    /* MD5 Random */
    DECLARE_CAP( NONE,      RANDOMGEN,  MD5Random,  NONE,           SINGLE_VAL( 0 ),        "MD5 Random" ),
    /* SHA Random */
    DECLARE_CAP( NONE,      RANDOMGEN,  SHARandom,  NONE,           SINGLE_VAL( 0 ),        "SHA Random" ),

    /* MD2 Digest */
    DECLARE_CAP( NONE,      DIGEST,     MD2,        OUTPUT_SIZE,    SINGLE_VAL( 16 ),       "MD2 Digest" ),
    /* MD5 Digest */
    DECLARE_CAP( NONE,      DIGEST,     MD5,        OUTPUT_SIZE,    SINGLE_VAL( 16 ),       "MD5 Digest" ),
    /* SHA-1 Digest */
    DECLARE_CAP( NONE,      DIGEST,     SHA1,       OUTPUT_SIZE,    SINGLE_VAL( 20 ),       "SHA-1 Digest" ),
    
    /* RSA Key Generation */
    DECLARE_CAP( NONE,      KEYGEN,     RSA,        KEY_LENGTH_RANGE,   RANGE_VAL( 512, 2048 ),     "RSA Key Generation" ),

    /* DSA Key Generation */
    DECLARE_CAP( NONE,      KEYGEN,     DSA,        KEY_LENGTH,     TRIPLE_VAL( 512, 768, 1024 ),   "DSA Key Generation" ),
    
    /* MD5WithRSA Signature */
    DECLARE_CAP( NONE,      SIGNATURE,  MD5WithRSA,     KEY_LENGTH_RANGE,   RANGE_VAL( 512, 2048 ),     "MD5 With RSA Signature" ),
    /* MD2WithRSA Signature */
    DECLARE_CAP( NONE,      SIGNATURE,  MD2WithRSA,     KEY_LENGTH_RANGE,   RANGE_VAL( 512, 2048 ),     "MD2 With RSA Signature" ),
    /* SHA1WithRSA Signature */
    DECLARE_CAP( NONE,      SIGNATURE,  SHA1WithRSA,    KEY_LENGTH_RANGE,   RANGE_VAL( 512, 2048 ),     "SHA1 With RSA Signature" ),

    /* SHA1WithDSA Signature */
    DECLARE_CAP( NONE,      SIGNATURE,  SHA1WithDSA,    KEY_LENGTH,     TRIPLE_VAL( 512, 768, 1024 ),       "SHA1 With DSA Signature" ),
    DECLARE_ATR( NONE,      SIGNATURE,  SHA1WithDSA,    OUTPUT_SIZE,    SINGLE_VAL( 40 ),           "SHA1 With DSA Signature" ),

    /* DES Key Generation */
    DECLARE_CAP( NONE,      KEYGEN,     DES,        KEY_LENGTH,     SINGLE_VAL( 64 ),       "DES Key Generation" ),

    /* 3DES Key Generation */
    DECLARE_CAP( NONE,      KEYGEN,     3DES_3KEY,  KEY_LENGTH,     SINGLE_VAL( 192 ),       "3DES 3Key Generation" ),

    /* DESX Key Generation */
    DECLARE_CAP( NONE,      KEYGEN,     DESX,       KEY_LENGTH,     SINGLE_VAL( 128 ),       "DESX Key Generation" ),

    /* RC2 Key Generation */
    DECLARE_CAP( NONE,      KEYGEN,     RC2,        KEY_LENGTH_RANGE,   RANGE_VAL( 8, 1024 ),    "RC2 Key Generation" ),

    /* RC4 Key Generation */
    DECLARE_CAP( NONE,      KEYGEN,     RC4,        KEY_LENGTH_RANGE,   RANGE_VAL( 8, 128 ),    "RC4 Key Generation" ),

    /* RC5 Key Generation */
    DECLARE_CAP( NONE,      KEYGEN,     RC5,        KEY_LENGTH_RANGE,   RANGE_VAL( 8, 128 ),    "RC5 Key Generation" ),

    /* RSA Encryption */
    DECLARE_CAP( NONE,      ASYMMETRIC, RSA,        KEY_LENGTH_RANGE,   RANGE_VAL( 512, 2048 ),    "RSA Asymmetric" ),

    /* DES CBC Padding IV8 */
    DECLARE_CAP( NONE,      SYMMETRIC,  DES,        KEY_LENGTH,     SINGLE_VAL( 64 ),           "DES CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DES,        MODE,           MODE_VAL( CBCPadIV8 ),      "DES CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DES,        IV_SIZE,        SINGLE_VAL( 8 ),            "DES CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DES,        BLOCK_SIZE,     SINGLE_VAL( 8 ),            "DES CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DES,        PADDING,        PADDING_VAL( PKCS5 ),       "DES CBC Padding IV8" ),

    /* DES CBC IV8 */
    DECLARE_CAP( NONE,      SYMMETRIC,  DES,        KEY_LENGTH,     SINGLE_VAL( 64 ),           "DES CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DES,        MODE,           MODE_VAL( CBC_IV8 ),        "DES CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DES,        IV_SIZE,        SINGLE_VAL( 8 ),            "DES CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DES,        BLOCK_SIZE,     SINGLE_VAL( 8 ),            "DES CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DES,        PADDING,        PADDING_VAL( NONE ),        "DES CBC IV8" ),

    /* 3DES 3KEY CBC Padding IV8 */
    DECLARE_CAP( NONE,      SYMMETRIC,  3DES_3KEY,  KEY_LENGTH,     SINGLE_VAL( 192 ),          "3DES 3KEY CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  3DES_3KEY,  MODE,           MODE_VAL( CBCPadIV8 ),      "3DES 3KEY CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  3DES_3KEY,  IV_SIZE,        SINGLE_VAL( 8 ),            "3DES 3KEY CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  3DES_3KEY,  BLOCK_SIZE,     SINGLE_VAL( 8 ),            "3DES 3KEY CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  3DES_3KEY,  PADDING,        PADDING_VAL( PKCS5 ),       "3DES 3KEY CBC Padding IV8" ),

    /* 3DES 3KEY CBC IV8 */
    DECLARE_CAP( NONE,      SYMMETRIC,  3DES_3KEY,  KEY_LENGTH,     SINGLE_VAL( 192 ),          "3DES 3KEY CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  3DES_3KEY,  MODE,           MODE_VAL( CBC_IV8 ),        "3DES 3KEY CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  3DES_3KEY,  IV_SIZE,        SINGLE_VAL( 8 ),            "3DES 3KEY CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  3DES_3KEY,  BLOCK_SIZE,     SINGLE_VAL( 8 ),            "3DES 3KEY CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  3DES_3KEY,  PADDING,        PADDING_VAL( NONE ),        "3DES 3KEY CBC IV8" ),

    /* DESX CBC Padding IV8 */
    DECLARE_CAP( NONE,      SYMMETRIC,  DESX,       KEY_LENGTH,     SINGLE_VAL( 128 ),          "DESX CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DESX,       MODE,           MODE_VAL( CBCPadIV8 ),      "DESX CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DESX,       IV_SIZE,        SINGLE_VAL( 8 ),            "DESX CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DESX,       BLOCK_SIZE,     SINGLE_VAL( 8 ),            "DESX CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DESX,       PADDING,        PADDING_VAL( PKCS5 ),       "DESX CBC Padding IV8" ),

    /* DESX CBC IV8 */
    DECLARE_CAP( NONE,      SYMMETRIC,  DESX,       KEY_LENGTH,     SINGLE_VAL( 128 ),          "DESX CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DESX,       MODE,           MODE_VAL( CBC_IV8 ),        "DESX CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DESX,       IV_SIZE,        SINGLE_VAL( 8 ),            "DESX CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DESX,       BLOCK_SIZE,     SINGLE_VAL( 8 ),            "DESX CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  DESX,       PADDING,        PADDING_VAL( NONE ),        "DESX CBC IV8" ),

    /* RC2 CBC Padding IV8 */
    DECLARE_CAP( NONE,      SYMMETRIC,  RC2,        KEY_LENGTH_RANGE,   RANGE_VAL( 8, 128 ),    "RC2 CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC2,        MODE,           MODE_VAL( CBCPadIV8 ),      "RC2 CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC2,        IV_SIZE,        SINGLE_VAL( 8 ),            "RC2 CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC2,        EFFECTIVE_BITS, RANGE_VAL( 1, 1024 ),       "RC2 CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC2,        PADDING,        PADDING_VAL( PKCS5 ),       "RC2 CBC Padding IV8" ),

    /* RC2 CBC IV8 */
    DECLARE_CAP( NONE,      SYMMETRIC,  RC2,        KEY_LENGTH_RANGE,   RANGE_VAL( 8, 128 ),    "RC2 CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC2,        MODE,           MODE_VAL( CBC_IV8 ),        "RC2 CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC2,        IV_SIZE,        SINGLE_VAL( 8 ),            "RC2 CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC2,        EFFECTIVE_BITS, RANGE_VAL( 1, 1024 ),       "RC2 CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC2,        PADDING,        PADDING_VAL( NONE ),        "RC2 CBC IV8" ),

    /* RC4 */
    DECLARE_CAP( NONE,      SYMMETRIC,  RC4,        KEY_LENGTH_RANGE,   RANGE_VAL( 8, 128 ),    "RC4" ),

    /* RC5 CBC Padding IV8 */
    DECLARE_CAP( NONE,      SYMMETRIC,  RC5,        KEY_LENGTH_RANGE,   RANGE_VAL( 8, 128 ),    "RC5 CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC5,        MODE,           MODE_VAL( CBCPadIV8 ),      "RC5 CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC5,        IV_SIZE,        SINGLE_VAL( 8 ),            "RC5 CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC5,        ROUNDS_RANGE,   RANGE_VAL( 1, 16 ),         "RC5 CBC Padding IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC5,        PADDING,        PADDING_VAL( PKCS5 ),       "RC5 CBC Padding IV8" ),

    /* RC5 CBC IV8 */
    DECLARE_CAP( NONE,      SYMMETRIC,  RC5,        KEY_LENGTH_RANGE,   RANGE_VAL( 8, 128 ),    "RC5 CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC5,        MODE,           MODE_VAL( CBC_IV8 ),        "RC5 CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC5,        IV_SIZE,        SINGLE_VAL( 8 ),            "RC5 CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC5,        ROUNDS_RANGE,   RANGE_VAL( 1, 16 ),         "RC5 CBC IV8" ),
    DECLARE_ATR( NONE,      SYMMETRIC,  RC5,        PADDING,        PADDING_VAL( NONE ),        "RC5 CBC IV8" ),

    /* PKCS #5 v1.5 PBKD1 w/MD2 */
    DECLARE_CAP( NONE,      DERIVEKEY,  PKCS5_PBKDF1_MD2,KEY_LENGTH,SINGLE_VAL( 64 ),           "PKCS #5 v1.5 PBKD1 w/MD2" ),

    /* PKCS #5 v1.5 PBKD1 w/MD5 */
    DECLARE_CAP( NONE,      DERIVEKEY,  PKCS5_PBKDF1_MD5,KEY_LENGTH,SINGLE_VAL( 64 ),           "PKCS #5 v1.5 PBKD1 w/MD5" ),

    /* PKCS #5 v1.5 PBKD1 w/SHA-1 */
    DECLARE_CAP( NONE,      DERIVEKEY,  PKCS5_PBKDF1_SHA1,KEY_LENGTH,SINGLE_VAL( 64 ),          "PKCS #5 v1.5 PBKD1 w/SHA-1" ),

};
static const uint32 CapabilityTableSize = 
        sizeof(CapabilityTable) / sizeof(MDSU_CDSA_CSP_CAPABILITY_SCHEMA);

/* Perform run-time calculation of group codes. */
static void __CalcGCodes()
{
    uint32 uGCode = (uint32)-1; /* Cause first increment to rollover to zero */
    uint32 i;

    /* Iterate through the capability array and set the appropriate group codes. */
    for ( i = 0; i < CapabilityTableSize; i++ )
    {
        if ( CapabilityTable[i].GroupId == NEW_CAP )
        {
            uGCode++;
        }
        CapabilityTable[i].GroupId = uGCode;
    }
}

static CSSM_RETURN __CleanMDS( MDSU_CONTEXT *pContext )
{
    CSSM_RETURN rv = CSSM_OK;
    MDSU_CDSA_COMMON_SCHEMA CommonTemplate;
    MDSU_CDSA_COMMON_SCHEMA CommonData;
    MDSU_CDSA_CSP_PRIMARY_SCHEMA CSPPrimaryTemplate;
    MDSU_CDSA_CSP_PRIMARY_SCHEMA CSPPrimaryData;
    MDSU_CDSA_CSP_ENC_PRODUCT_SCHEMA CSPEncProdTemplate;
    MDSU_CDSA_CSP_ENC_PRODUCT_SCHEMA CSPEncProdData;
    MDSU_CDSA_CSP_CAPABILITY_SCHEMA CSPCapabilityTemplate;
    MDSU_CDSA_CSP_CAPABILITY_SCHEMA CSPCapabilityData;

    /* Execute a search on every CSP record type to get a handle of all
     * records and then delete them.
     */
    /* Common Schema */
    CommonTemplate.ModuleID = ADDIN_GUID;
    rv = MDSU_FindFirst( pContext,
                         &IfiMdsuCdsadirCommonSchema,
                         &CommonTemplate,
                         MDSU_COM_MODULE_ID,
                         &CommonData,
                         NULL );
    while ( rv == CSSM_OK )
    {
        if ( CommonData.Manifest.Data )
        {
            MDSU_free( pContext, CommonData.Manifest.Data );
        }

        rv = MDSU_FindNext( pContext,
                            &CommonData,
                            NULL );
    }
    if ( rv == CSSMERR_DL_ENDOFDATA )
    {
        /* CSP Primary Schema */
        CSPPrimaryTemplate.ModuleID = ADDIN_GUID;
        rv = MDSU_FindFirst( pContext,
                             &IfiMdsuCdsadirCSPPrimarySchema,
                             &CSPPrimaryTemplate,
                             MDSU_CSPP_MODULE_ID,
                             &CSPPrimaryData,
                             NULL );
        while ( rv == CSSM_OK )
        {
            if ( CSPPrimaryData.Manifest.Data )
            {
                MDSU_free( pContext, CSPPrimaryData.Manifest.Data );
            }

            rv = MDSU_FindNext( pContext,
                                &CSPPrimaryData,
                                NULL );
        }
        if ( rv == CSSMERR_DL_ENDOFDATA )
        {
            /* CSP Encapsulated Product Schema */
            CSPEncProdTemplate.ModuleID = ADDIN_GUID;
            rv = MDSU_FindFirst( pContext,
                                 &IfiMdsuCdsadirCSPEncProductSchema,
                                 &CSPEncProdTemplate,
                                 MDSU_CSPEP_MODULE_ID,
                                 &CSPEncProdData,
                                 NULL );
            while ( rv == CSSM_OK )
            {
                rv = MDSU_FindNext( pContext,
                                    &CSPEncProdData,
                                    NULL );
            }
            if ( rv == CSSMERR_DL_ENDOFDATA )
            {
                /* CSP Capability Schema */
                CSPCapabilityTemplate.ModuleID = ADDIN_GUID;
                rv = MDSU_FindFirst( pContext,
                                     &IfiMdsuCdsadirCSPCapabilitySchema,
                                     &CSPCapabilityTemplate,
                                     MDSU_CSPCAP_MODULE_ID,
                                     &CSPCapabilityData,
                                     NULL );
                while ( rv == CSSM_OK )
                {
                    rv = MDSU_FindNext( pContext,
                                        &CSPCapabilityData,
                                        NULL );
                }
                if ( rv == CSSMERR_DL_ENDOFDATA )
                {
                    /* Flush everything from the MDS */
                    rv = MDSU_Revert( pContext );
                } /* CSP Capability removed */
            } /* CSP Encapsulated Product removed */
        } /* CSP Primary removed */
    } /* Common removed */

    /* Make sure nothing gets removed if there is an error */
    if ( ( rv != CSSMERR_DL_ENDOFDATA ) &&
         ( rv != CSSM_OK ) ) 
    {
        MDSU_Commit( pContext );
    }

    return rv;
}

CSSM_RETURN CSSMAPI INSTALL_ENTRY_NAME(
                        const char *szModuleName,
                        const char *szSearchPath,
                        CSSM_DATA *pCredential,
                        INSTALL_ACTION Action,
                        INSTALL_ERROR *pError )
{
    CSSM_RETURN rv = CSSM_OK;
    MDSU_CONTEXT MDSUContext;
    MDSU_CDSA_COMMON_SCHEMA Common;
    MDSU_CDSA_CSP_PRIMARY_SCHEMA CSPPrimary;
    MDSU_CDSA_CSP_ENC_PRODUCT_SCHEMA CSPEncProduct;

    uint32 i;

    /* Calculate the group codes for the capability table */
    __CalcGCodes();

    /* Initialize the MDS update */
    if ( ( rv = MDSU_Init( &MDSUContext, 
                           &Addin_APIMemFuncs,
                           &ADDIN_GUID,
                           MDS_CDSA_DIRECTORY,
                           ( CSSM_DB_ACCESS_READ | 
                             CSSM_DB_ACCESS_WRITE ),
                           MAF_BSAFE_MDS_COMMIT_SIZE ) ) != CSSM_OK )
    {
        sprintf( pError->ErrorString, "MDS Error (Init): %X\n", rv );
        pError->ErrorCode = rv;
        return rv;
    }

    /* Place code to remove existing entries here */
    if ( ( rv = __CleanMDS( &MDSUContext ) ) != CSSM_OK )
    {
        sprintf( pError->ErrorString, "MDS Error (Clean): %X\n", rv );
        pError->ErrorCode = rv;
        MDSU_Term( &MDSUContext );
        return rv;
    }
    if ( Action == INSTALL_ACTION_UNINSTALL )
    {
        MDSU_Term( &MDSUContext );
        return rv;
    }

    /* Fill in the "easy" common schema structure */
    memcpy( &Common.ModuleID, &ADDIN_GUID, sizeof(ADDIN_GUID) );
    Common.Manifest = *pCredential;
    strcpy( Common.ModuleName, szModuleName );
    strcpy( Common.Path, szSearchPath );
    Common.CDSAVersion.Major = CSSM_MAJOR;
    Common.CDSAVersion.Minor = CSSM_MINOR;
    strcpy( Common.Desc, ADDIN_PRINT_NAME );
    Common.DynamicFlag = CSSM_FALSE;
    Common.MultiThreadFlag = CSSM_FALSE;
    Common.ServiceMask = CSSM_SERVICE_CSP;

    /* Insert the common schema record into the MDS */
    rv = MDSU_UpdateSchema( &MDSUContext,
                            &IfiMdsuCdsadirCommonSchema,
                            &Common,
                            NULL );
    if ( rv != CSSM_OK )
    {
        sprintf( pError->ErrorString, "MDS Error (common): %X\n", rv );
        pError->ErrorCode = rv;
        MDSU_Term( &MDSUContext );
        return rv;
    }

    /* Fill in the CSP primary schema data */
    memcpy( &CSPPrimary.ModuleID, &ADDIN_GUID, sizeof(ADDIN_GUID) );
    CSPPrimary.SSID = ADDIN_CSP_SUBSERVICE_ID;
    CSPPrimary.Manifest.Data = NULL;
    CSPPrimary.Manifest.Length = 0;
    strcpy( CSPPrimary.ModuleName, ADDIN_PRINT_NAME );
    CSPPrimary.ProductVersion.Major = ADDIN_PRODUCT_VERSION_MAJOR;
    CSPPrimary.ProductVersion.Minor = ADDIN_PRODUCT_VERSION_MINOR;
    strcpy( CSPPrimary.Vendor, ADDIN_VENDOR_NAME );
    CSPPrimary.CspType = CSSM_CSP_SOFTWARE;
    CSPPrimary.CspFlags = 0;
    CSPPrimary.CspCustomFlags = 0;
    CSPPrimary.UseeTags[0] = CSSM_USEE_NONE;
    CSPPrimary.NumUseeTags = 1;
    CSPPrimary.NumSampleTypes = 0;
    CSPPrimary.NumAclSubjectTypes = 0;
    CSPPrimary.NumAclAuthTags = 0;

    /* Write the data to MDS */
    rv = MDSU_UpdateSchema( &MDSUContext,
                            &IfiMdsuCdsadirCSPPrimarySchema,
                            &CSPPrimary,
                            NULL );
    if ( rv != CSSM_OK )
    {
        sprintf( pError->ErrorString, "MDS Error (CSP Primary): %X\n", rv );
        pError->ErrorCode = rv;
        MDSU_Revert( &MDSUContext );
        MDSU_Term( &MDSUContext );
        return rv;
    }

    /* Fill in the CSP encapsulated product schema data */
    memcpy( &CSPEncProduct.ModuleID, &ADDIN_GUID, sizeof(ADDIN_GUID) );
    CSPEncProduct.SSID = ADDIN_CSP_SUBSERVICE_ID;
    strcpy( CSPEncProduct.ProductVendor, ADDIN_VENDOR_NAME );
    CSPEncProduct.ProductVersion.Major = ADDIN_PRODUCT_VERSION_MAJOR;
    CSPEncProduct.ProductVersion.Minor = ADDIN_PRODUCT_VERSION_MINOR;
    strcpy( CSPEncProduct.ProductDesc, ADDIN_PRODUCT_DESCRIPTION );
    CSPEncProduct.ProductFlags = 0;
    CSPEncProduct.CustomFlags = 0;
    strcpy( CSPEncProduct.StandardDesc, ADDIN_STANDARD_DESCRIPTION );
    CSPEncProduct.StandardVersion.Major = ADDIN_STANDARD_VERSION_MAJOR;
    CSPEncProduct.StandardVersion.Minor = ADDIN_STANDARD_VERSION_MINOR;
    strcpy( CSPEncProduct.ReaderDesc, "" );
    strcpy( CSPEncProduct.ReaderVendor, "" );
    CSPEncProduct.ReaderVersion.Major = 0;
    CSPEncProduct.ReaderVersion.Minor = 0;
    CSPEncProduct.ReaderFirmwareVersion.Major = 0;
    CSPEncProduct.ReaderFirmwareVersion.Minor = 0;
    CSPEncProduct.ReaderFlags = 0;
    CSPEncProduct.ReaderCustomFlags = 0;
    strcpy( CSPEncProduct.ReaderSerialNumber, "" );

    /* Write the data to MDS */
    rv = MDSU_UpdateSchema( &MDSUContext,
                            &IfiMdsuCdsadirCSPEncProductSchema,
                            &CSPEncProduct,
                            NULL );
    if ( rv != CSSM_OK )
    {
        sprintf( pError->ErrorString, "MDS Error (CSP Enc): %X\n", rv );
        pError->ErrorCode = rv;
        MDSU_Revert( &MDSUContext );
        MDSU_Term( &MDSUContext );
        return rv;
    }

    /* Write the CSP capabilities table */
    for ( i = 0; i < CapabilityTableSize; i++ )
    {
        rv = MDSU_UpdateSchema( &MDSUContext,
                                &IfiMdsuCdsadirCSPCapabilitySchema,
                                (void*)&CapabilityTable[i],
                                NULL );
        if ( rv != CSSM_OK )
        {
            sprintf( pError->ErrorString, "MDS Error (CSP Cap): %X\n", rv );
            pError->ErrorCode = rv;
            MDSU_Revert( &MDSUContext );
            MDSU_Term( &MDSUContext );
            return rv;
        }
    }

    /* Commit the records */
    MDSU_Commit( &MDSUContext );

    /* Shut down the MDS connection */
    MDSU_Term( &MDSUContext );

    return CSSM_OK;
}

#ifdef VMS /* COMPAQ_FIX */
/*
 * This provides a pointer to the capability table for the passthrough
 * routine.
*/
CSSM_RETURN return_capability_table(MDSU_CDSA_CSP_CAPABILITY_SCHEMA **pointer,
				    uint32 *sizeoftable)
{
  *pointer = CapabilityTable;
  *sizeoftable = CapabilityTableSize;
  return CSSM_OK;
}
#endif
