/*-----------------------------------------------------------------------
 * File: CRTGRPVR.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 

/* CSSM INCLUDES */
#include "cssm.h"
#include "cssmtpi.h"
#include "cssmspi.h"
#include "cssmMemTrack.h"

/* X.509 and ALG INCLUDES */
#include "x509defs.h"
#include "oidscert.h"
#include "oidsalg.h"            
#include "algoid.h"

/* LOCAL INCLUDES */
#include "inx509tp.h"           /* func and data definitions specific to this module */
#include "Chn_util.h"
#include "api_shim.h"
#include "maf_include.h"

/*-----------------------------------------------------------------------------
 * Name: TP_CertGroupVerify
 *
 * Description:
 * This function constructs and verifies that an ordered cert chain is valid.  
 * It checks the chain's linkage and verifies the validity of each link (cert), 
 * including revocation checks.
 *
 * Parameters: 
 * TPHandle (input):           handle to the TP module being called
 * CLHandle (input):           handle to the CL module to be used
 * CSPHandle (input):          handle to the CSP module to use for 
 *                             verifications
 * DBList (input):             list of databases to be used to construct
 *                             the chain of certificates to be verified
 * CertGroupToBeVerified:      certificate group contianing the cert to 
 *                             be verified first 
 * VerifyContext
 *   PolicyIdentifiers (input):   [unused - must be NULL]
 *   NumberOfPolicyIdentifiers (input): [unused - must be zero]
 *   VerificationAbortOn (input): condition on which to stop verification
 *                                [must be specified as CSSM_TP_STOP_ON_POLICY]
 *   UserAuthentication (input):  authentication information to be used in 
 *                                association with this request
 *   AnchorCerts (input):         [unused - must be NULL]
 *   NumberOfAnchorCerts (input): [unused - must be 0]
 *   VerifyScope (input):         [unused - must be NULL]
 *   ScopeSize (input):           [unused - must be 0]
 *   Action (input):              [unused - must be 0]
 *   CallbackWithVerifiedCert (input): ?
 *   ActionData (input):          [unused - must be NULL]
 *   Evidence (output):           used to return the pointer to the constructed 
 *                                chain, which was verified
 *   NumberOfEvidences (output):  number of CSSM_DATA structures returned in 
 *                                Evidence
 *
 * Return values:
 *   CSSM_TRUE, if verify succeeded
 *   CSSM_FALSE, if verify failed
 *
 * Error Codes:
 *   -TBD-
 *
 *---------------------------------------------------------------------------*/

CSSM_RETURN CSSMAPI 
TP_CertGroupVerify(CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_CERTGROUP *CertGroupToBeVerified,
                        const CSSM_TP_VERIFY_CONTEXT *VerifyContext,
                        CSSM_TP_VERIFY_CONTEXT_RESULT_PTR VerifyContextResult )
{
  CSSM_CERTGROUP_PTR pConstructedGroup = NULL;
  CSSM_CC_HANDLE hCC = 0;
  CSSM_DATA_PTR pSubjectCert = NULL;
  CSSM_DATA_PTR pIssuerCert = NULL; 
  CSSM_DATA_PTR pSubjectCertName = NULL;
  CSSM_BOOL bRootFound = CSSM_FALSE;
  CSSM_KEY_PTR pSubjectCertPublicKey = NULL;
  CSSM_KEY_PTR pIssuerCertPublicKey = NULL;
  CSSM_DATA_PTR pTempData = NULL;
  CSSM_API_MEMORY_FUNCS CLMemFuncs = {NULL, NULL, NULL, NULL, NULL};
  CSSM_HANDLE ResultsHandle;
  uint32 NumberOfFields;
  uint32 SignatureAlgId = 0;
  uint32 counter1 = 0;
  uint32 counter2 = 0;
  uint32 num_certs_in_grp = 0;
  CSSM_RETURN RetCode = CSSM_OK;

  const ADDIN_LOAD_DATA *pLoadData = NULL;
  CSSM_DATA_PTR pAnchorIssuerName = NULL;
  CSSM_DATA_PTR pCurrentIssuerName = NULL;
  CSSM_KEY_PTR pAnchorPubKey = NULL; 

  CSSM_DATA_PTR pCurrentCert = NULL;
  CSSM_BOOL bVerifyResult = CSSM_FALSE;
  uint32 num_certs_processed = 0;   /* used for basic constraints extension processing */
  CSSM_X509_EXTENSION_PTR extnp = NULL; /* used for basic constraints extension processing */

  CSSM_RETURN Rtn = CSSM_OK;

  pCurrentCert;
  extnp;
  bVerifyResult;
  num_certs_processed;

  /* Do we know our execution context? */
  pLoadData = Addin_GetLoadData();
  if ( pLoadData == NULL )
  {
      return CSSMERR_TP_INTERNAL_ERROR;
  }

/* --CHECK INPUTS-- */
  /* check params */
  if( TPHandle == 0 )
  {
    return( CSSMERR_TP_INTERNAL_ERROR );
  }
  if( CLHandle == 0 )
  {
    return( CSSMERR_TP_INVALID_CL_HANDLE );
  }
  if( CertGroupToBeVerified == NULL )
  {
    /* nothing to verify! */
    return( CSSMERR_TP_INVALID_CERTGROUP_POINTER );
  }
  if( (CertGroupToBeVerified->NumCerts == 0) || 
      (CertGroupToBeVerified->GroupList.CertList == NULL) )
  {
    /* nothing to verify! */
    return( CSSMERR_TP_INVALID_CERTGROUP_POINTER );
  }

    /* check all unsupported params */
    if(VerifyContext != NULL )
    {
        if(NULL != VerifyContext->Cred)
        {
            if ((VerifyContext->Cred->Policy.PolicyIds ) || 
                (VerifyContext->Cred->Policy.NumberOfPolicyIds) ) 
            {
                return( CSSMERR_TP_INVALID_POLICY_IDENTIFIERS );
            }
            if( VerifyContext->Cred->VerificationAbortOn != CSSM_TP_STOP_ON_POLICY )
            {
                return( CSSMERR_TP_INVALID_STOP_ON_POLICY );
            }
            if( (VerifyContext->Action != CSSM_TP_ACTION_DEFAULT) || 
                (VerifyContext->ActionData.Data != NULL) ||
                (VerifyContext->ActionData.Length != 0) )
            {
                return( CSSMERR_TP_INVALID_ACTION );
            }
        }
        else
        {
            return(ERR(CSSMERR_TP_INVALID_CALLERAUTH_CONTEXT_POINTER));
        }
    }
    else
    {
        return(ERR(CSSMERR_TP_INVALID_CALLERAUTH_CONTEXT_POINTER));
    }

/* --CONSTRUCT CERT GROUP-- */
    Rtn = tp_CertGroupConstruct(TPHandle, CLHandle, CSPHandle, VerifyContext->Cred->DBList, 
                                            CertGroupToBeVerified, 
                                            CSSM_FALSE, &pConstructedGroup);
    if( Rtn != CSSM_OK )
    {
        return( Rtn );
    }

    if ( (VerifyContextResult != NULL)
         && (VerifyContextResult->Evidence != NULL) 
       )
    {
        /* if evidence ptrs provided, set evidence ptrs to built cert group */
        (*(VerifyContextResult->Evidence)).EvidenceForm = CSSM_EVIDENCE_FORM_CERT;
        (*(VerifyContextResult->Evidence)).Evidence = pConstructedGroup->GroupList.CertList;
        VerifyContextResult->NumberOfEvidences = pConstructedGroup->NumCerts;
    }

/* --GET CL API MEMORY FUNCTIONS-- */
    RetCode = addin_GetAPIMemoryFunctions( CLHandle, &CLMemFuncs );  
    if( RetCode != CSSM_OK )
    {
        free_certgroup(pConstructedGroup);
        /* cssm should have set an error */
        return( CSSMERR_TP_INTERNAL_ERROR );
    }

/* --INITIALIZE VARS FOR LOOP-- */
  pSubjectCert = pConstructedGroup->GroupList.CertList;
  if( pSubjectCert == NULL )
  {
    free_certgroup(pConstructedGroup);
    /* nothing to verify! */
    return( CSSMERR_TP_INVALID_CERTGROUP_POINTER );
  }

#if 0 
  num_certs_in_grp = 
      ((CSSM_TP_X509V3_GRP_INFO_PTR)pConstructedGroup->Reserved)->IndexToRootCert + 1;

  if( num_certs_in_grp > pConstructedGroup->NumCerts )
  {
#endif /* mhs */
    /* try to continue - assume all certs in group */
    num_certs_in_grp = pConstructedGroup->NumCerts;
#if 0
  }
#endif /* mhs */

/* --ITERATE THROUGH EACH CERT IN CONSTRUCTED CERT GROUP UNTIL ROOT IS FOUND-- */
  for( counter1=0; counter1 < num_certs_in_grp; counter1++)
  {
/* --GET SUBJ CERT POSITIVE ID-- */
    Rtn = addin_cl_CertGetKeyInfo( CLHandle, pSubjectCert, &pSubjectCertPublicKey );
    if( Rtn != CSSM_OK )
    {
      free_certgroup(pConstructedGroup);      
      return( Rtn );
    }
    /* Memory allocated for key at pSubjectCertPublicKey */

    Rtn = tp_CertGetSubjectX509Name( CLHandle, pSubjectCert, &pSubjectCertName );
    if( pSubjectCertName == NULL )
    {
      /* clean-up memory */
      free_certgroup(pConstructedGroup);
      /* Free key Memory */
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey->KeyData.Data, CLMemFuncs.AllocRef );
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey, CLMemFuncs.AllocRef );

      /* error should have been set by CL already */
      return( Rtn );
    }
    /* Memory allocated for subject name at pSubjectCertName */

/* --GET ISSUER CERT FOR SUBJECT-- */
    if(VerifyContext->Cred->AnchorCerts != NULL) /* Are there AnchorCerts? */
    {
        for( counter2 = 0; 
             counter2 < VerifyContext->Cred->NumberOfAnchorCerts;
             counter2++ )
        {
            Rtn = tp_CertGetIssuerX509Name(CLHandle, 
                                            &(VerifyContext->Cred->AnchorCerts[counter2]),
                                            &pAnchorIssuerName);
            Rtn = tp_CertGetIssuerX509Name(CLHandle, pSubjectCert, &pCurrentIssuerName);
            Rtn = addin_cl_CertGetKeyInfo(CLHandle, 
                                          &(VerifyContext->Cred->AnchorCerts[counter2]),
                                          &pAnchorPubKey);
            if( Rtn == CSSM_OK )
            {
                
                if( tp_IsSameX509Name( pAnchorIssuerName, pCurrentIssuerName ) 
                    && !memcmp(pAnchorPubKey->KeyData.Data, 
                               pSubjectCertPublicKey->KeyData.Data, 
                               pAnchorPubKey->KeyData.Length) )
                {
                    if( tp_IsSelfSignedCert( TPHandle, CLHandle, pSubjectCert ) == CSSM_TRUE ) 
                    {
                        /* issuer cert equals subject cert */
                        pIssuerCert = pSubjectCert;
                    }
                    else
                    {
                        pIssuerCert = pSubjectCert + (sizeof( *pSubjectCert )/8);
                    }
                    bRootFound = CSSM_TRUE;
                }
                /* Free key Memory */
                CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pAnchorPubKey->KeyData.Data, CLMemFuncs.AllocRef );
                CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pAnchorPubKey, CLMemFuncs.AllocRef );
            }
            if( pAnchorIssuerName )
            {
                App_Free( TPHandle, pAnchorIssuerName->Data );
                App_Free( TPHandle, pAnchorIssuerName );
            }
            if( pCurrentIssuerName )
            {
                App_Free( TPHandle, pCurrentIssuerName->Data );
                App_Free( TPHandle, pCurrentIssuerName );
            }
        }
    }
    else
    {
        /* determine if cert is self-signed or just a link */
        if( tp_IsSelfSignedCert( TPHandle, CLHandle, pSubjectCert ) == CSSM_TRUE ) 
        {
          /* issuer cert equals subject cert */
          pIssuerCert = pSubjectCert;
          bRootFound = CSSM_TRUE;
        }
    }
    
    /* link cert */
    if( bRootFound != CSSM_TRUE )
    {
      /* issuer is next link in chain */
      pIssuerCert = pSubjectCert + (sizeof( *pSubjectCert )/8);
    }

    /* make sure that an issuer cert was retrieved */
    if( pIssuerCert == NULL)
    {
      /* clean-up memory */
      free_certgroup(pConstructedGroup);
      /* Free key Memory */
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey->KeyData.Data, CLMemFuncs.AllocRef );
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey, CLMemFuncs.AllocRef );
      /* Free subject name Memory */
      App_Free( TPHandle, pSubjectCertName->Data );
      App_Free( TPHandle, pSubjectCertName );

      /* set error - broken chain */
      return( CSSMERR_TP_CERTGROUP_INCOMPLETE );
    }

/* --GET ISSUER KEY FOR VERIFY SIGNATURE CONTEXT-- */
    /* get issuer public key */
    Rtn = addin_cl_CertGetKeyInfo( CLHandle, 
                                   pIssuerCert, 
                                   &pIssuerCertPublicKey );
    if( Rtn != CSSM_OK )
    {
      /* clean-up memory */
      free_certgroup(pConstructedGroup);
      /* Free key memory */
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey->KeyData.Data, CLMemFuncs.AllocRef );
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey, CLMemFuncs.AllocRef );
      /* Free subject name Memory */
      App_Free( TPHandle, pSubjectCertName->Data );
      App_Free( TPHandle, pSubjectCertName );

      return( Rtn );
    }
    /* Memory allocated for issuer public key at pIssuerCertPublicKey */

/* --VERIFY SUBJECT CERT-- */
    /* setup signature verify context */
    Rtn = addin_cl_CertGetFirstFieldValue( CLHandle, 
                                                pSubjectCert, 
                                                &CSSMOID_X509V1SignatureAlgorithm,
                                                &ResultsHandle, 
                                                &NumberOfFields, 
                                                &pTempData );
    if( Rtn != CSSM_OK )
    {
      /* clean-up memory */
      free_certgroup(pConstructedGroup);
      /* Free key Memory */
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey->KeyData.Data, CLMemFuncs.AllocRef );
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey, CLMemFuncs.AllocRef );
      /* Free subject name Memory */
      App_Free( TPHandle, pSubjectCertName->Data );
      App_Free( TPHandle, pSubjectCertName );
      /* Free issuer public key Memory */
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pIssuerCertPublicKey->KeyData.Data, CLMemFuncs.AllocRef );
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pIssuerCertPublicKey, CLMemFuncs.AllocRef );

      return( Rtn );
    }
    /* Memory allocated for field at pTempData */

    SignatureAlgId = X509_AlgorithmOidToAlgId( (CSSM_OID_PTR)pTempData );
    if( SignatureAlgId == 0 )
    {
      /* clean-up memory */
      free_certgroup(pConstructedGroup);
      /* Free key Memory */
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey->KeyData.Data, CLMemFuncs.AllocRef );
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey, CLMemFuncs.AllocRef );
      /* Free subject name Memory */
      App_Free( TPHandle, pSubjectCertName->Data );
      App_Free( TPHandle, pSubjectCertName );
      /* Free issuer public key Memory */
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pIssuerCertPublicKey->KeyData.Data, CLMemFuncs.AllocRef );
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pIssuerCertPublicKey, CLMemFuncs.AllocRef );
      /* Free field Memory */
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pTempData->Data, CLMemFuncs.AllocRef );
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pTempData, CLMemFuncs.AllocRef );

      /* restore error code */
      return( CSSMERR_TP_INVALID_SIGNATURE );
    }

    hCC = 0;
    if (CSPHandle != 0)
    {
        Rtn = addin_csp_CreateSignatureContext( CSPHandle, 
                                                 SignatureAlgId, 
                                                 NULL, 
                                                 pIssuerCertPublicKey, &hCC );
        /* cleanup unneeded memory */
        /* Free field Memory */
        CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pTempData->Data, CLMemFuncs.AllocRef );
        CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pTempData, CLMemFuncs.AllocRef );
        /* Free issuer public key Memory */
        CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pIssuerCertPublicKey->KeyData.Data, CLMemFuncs.AllocRef );
        CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pIssuerCertPublicKey, CLMemFuncs.AllocRef );
        /* Memory remains allocated for key and subject name */
        
        /* check context handle */
        if( Rtn != CSSM_OK)
        {
            /* clean-up memory */
            free_certgroup(pConstructedGroup);
            /* Free key Memory */
            CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey->KeyData.Data, CLMemFuncs.AllocRef );
            CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey, CLMemFuncs.AllocRef );
            /* Free subject name Memory */
            App_Free( TPHandle, pSubjectCertName->Data );
            App_Free( TPHandle, pSubjectCertName );

            return( Rtn );
        }

        /* Created hCC Context */
    }
    else
    {
        /* cleanup unneeded memory */
        /* Free field Memory */
        CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pTempData->Data, CLMemFuncs.AllocRef );
        CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pTempData, CLMemFuncs.AllocRef );
        /* Free issuer public key Memory */
        CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pIssuerCertPublicKey->KeyData.Data, CLMemFuncs.AllocRef );
        CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pIssuerCertPublicKey, CLMemFuncs.AllocRef );
        /* Memory remains allocated for key and subject name */
    }

    /* --CHECK VALIDITY PERIOD-- */
    Rtn = tp_CheckCertValidityPeriodValid(CLHandle, pSubjectCert);
    if ( Rtn != CSSM_OK)
    {
      free_certgroup(pConstructedGroup);
      /* Free hCC Context */
      if (hCC != 0)
      {
        addin_DeleteContext( hCC );
      }
      /* Free key Memory */
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey->KeyData.Data, CLMemFuncs.AllocRef );
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey, CLMemFuncs.AllocRef );
      /* Free subject name Memory */
      App_Free( TPHandle, pSubjectCertName->Data );
      App_Free( TPHandle, pSubjectCertName );

      /* fail */
      return( Rtn );
    }

    {   
        /* verify the subject's cert has a valid signature */
        Rtn = addin_cl_CertVerify( CLHandle, hCC, pSubjectCert,
                                              pIssuerCert, NULL, 0 );
    }

    if( Rtn != CSSM_OK )    /* subject's credential is bad */
    {
      /* clean-up memory */
      free_certgroup(pConstructedGroup);
      /* Free hCC Context */
      if (hCC != 0)
      {
        addin_DeleteContext( hCC );
      }
      /* Free key Memory */
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey->KeyData.Data, CLMemFuncs.AllocRef );
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey, CLMemFuncs.AllocRef );
      /* Free subject name Memory */
      App_Free( TPHandle, pSubjectCertName->Data );
      App_Free( TPHandle, pSubjectCertName );

      /* set error - invalid cert in chain */
      return( Rtn );
    }


/* --CHECK FOR REVOCATION OF SUBJECT CERT-- */
    {

        if (tp_IsCertRevoked( TPHandle, CLHandle, 
                VerifyContext->Cred->DBList, pSubjectCert ) == CSSM_TRUE)
        {
            free_certgroup(pConstructedGroup);
            /* Free hCC Context */
            if (hCC != 0)
            {
                addin_DeleteContext( hCC );
            }
            /* Free key Memory */
            CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey->KeyData.Data, 
                                CLMemFuncs.AllocRef );
            CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey, 
                                CLMemFuncs.AllocRef );
            /* Free subject name Memory */
            App_Free( TPHandle, pSubjectCertName->Data );
            App_Free( TPHandle, pSubjectCertName );

            /* fail */
            return( CSSMERR_TP_CERT_REVOKED );
        }
    }

#ifdef MHS_TURNONEXT

/* --CHECK BASIC CONSTRAINTS-- */
    /* Get the Basic Constraints X509 Extension Value if there is one */
    pTempData = addin_cl_CertGetFirstFieldValue( CLHandle, pSubjectCert, 
                                                 &CSSMOID_X509ExtBasicConstraints,
                                                 &ResultsHandle, &NumberOfFields );
    if ( pTempData != NULL ) 
    { /* there is a Basic Constraints X509 Extension */
      extnp = (X509_EXTENSION_PTR) pTempData->Data;
      if ( extnp->critical == CSSM_TRUE )
      { /* it is a critical Extension */
        if ( extnp->value.basicConstraints.cA == CSSM_TRUE )
        { /* the CA flag is set */
          /* should there be a check to verify this is a self-signed cert? */
          if ( num_certs_processed > extnp->value.basicConstraints.pathLenConstraint )
          { /* num_certs_processed exceeds pathLenConstraint */
            free_certgroup(pConstructedGroup);
            if (hCC != 0)
            {
              addin_DeleteContext( hCC );  /* Free hCC Context */
            }
            /* Free key Memory */
            CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey->KeyData.Data, CLMemFuncs.AllocRef );
            CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey, CLMemFuncs.AllocRef );
            /* Free subject name Memory */
            App_Free( TPHandle, pSubjectCertName->Data );
            App_Free( TPHandle, pSubjectCertName );
            /* fail */
            return( CSSMERR_TP_CERT_VERIFY_FAIL );
          } /* pathLenConstraint */
        }  /* cA flag */
      }  /* critical extension */
    }  /* Basic Constraints Extension exists */

    num_certs_processed++;

#endif

/* --INSERT ADDITIONAL VALIDITY CHECKS HERE-- */

/* --INCREMENT TO NEXT LINK OR RETURN-- */
    if( bRootFound != CSSM_TRUE )
    {
      /* move to next chain link */
      pSubjectCert = pIssuerCert;
    }
    else    
    {
      /* chain is verified! return CSSM_OK */
      free_certgroup(pConstructedGroup);
      /* Free hCC Context */
      if (hCC != 0)
      {
        addin_DeleteContext( hCC );
      }
      /* Free key Memory */
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey->KeyData.Data, CLMemFuncs.AllocRef );
      CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey, CLMemFuncs.AllocRef );
      /* Free subject name Memory */
      App_Free( TPHandle, pSubjectCertName->Data );
      App_Free( TPHandle, pSubjectCertName );

      return( CSSM_OK );
    }

/* --CLEAN-UP MEMORY FOR THIS ITERATION-- */
    /* Free hCC Context */
    if (hCC != 0)
    {
      addin_DeleteContext( hCC );
    }
    /* Free key Memory */
    CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey->KeyData.Data, CLMemFuncs.AllocRef );
    CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  pSubjectCertPublicKey, CLMemFuncs.AllocRef );
    /* Free subject name Memory */
    App_Free( TPHandle, pSubjectCertName->Data );
    App_Free( TPHandle, pSubjectCertName );
  }

/* --CHAIN TERMINATION NOT FOUND, FAIL!-- */
  free_certgroup(pConstructedGroup);
  return( CSSMERR_TP_CERTGROUP_INCOMPLETE );
}


