/*-----------------------------------------------------------------------
 * File: MEF_INSTALL.C
 *
 * Copyright (c) 1999-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "precomp.h"



/* Number of MDS update operations that will be remembered by mds_util for
 * error clean-up during installation.
 */
#define EMM_MDS_COMMIT_SIZE 20



static CSSM_RETURN
__CleanMDS( MDSU_CONTEXT *pContext )
{
    CSSM_RETURN rv = CSSM_OK;
    MDSU_CDSA_EMM_SCHEMA EMM;
    MDSU_CDSA_EMM_SCHEMA EMMData;

    /* Execute a search on every EMM record type to get a handle of all
     * records and then delete them.
     */

    /* Common Schema */
    EMM.ModuleID = EMM_GUID;

    ERR(rv = MDSU_FindFirst(pContext,
                            &IfiMdsuCdsadirEMMSchema,
                            &EMM,
                            MDSU_EMM_MODULE_ID,
                            &EMMData,
                            NULL));

    while(CSSM_OK == rv)
    {
        if(NULL != EMMData.Manifest.Data)
        {
            MDSU_free(pContext,
                      EMMData.Manifest.Data);
        }

        ERR(rv = MDSU_FindNext(pContext,
                               &EMMData,
                               NULL));
    }
    /* Common removed */

    /* Make sure nothing gets removed if there is an error */
    if((CSSMERR_DL_ENDOFDATA != rv) &&
       (CSSM_OK != rv))
    {
        /* the logic of commit NOT doing an uninstall seems odd, but
           this does work
        */
        MDSU_Commit(pContext);
    }
    else
    {
        /* Flush everything from the MDS */
        ERR(rv = MDSU_Revert(pContext));
    }

    return rv;
}



CSSM_RETURN CSSMAPI
INSTALL_ENTRY_NAME(const char *szModuleName,
                   const char *szSearchPath,
                   CSSM_DATA *pCredential,
                   INSTALL_ACTION Action,
                   INSTALL_ERROR *pError )
{
    CSSM_RETURN rv = CSSM_OK;
    MDSU_CONTEXT MDSUContext;
    MDSU_CDSA_EMM_SCHEMA EMM;


    /* Initialize MDS */
    if (ERR(rv = MDSU_Init(&MDSUContext, 
                           &APIMemFuncs,
                           &EMM_GUID,
                           MDS_CDSA_DIRECTORY,
                           (CSSM_DB_ACCESS_READ | CSSM_DB_ACCESS_WRITE),
                           EMM_MDS_COMMIT_SIZE)) == CSSM_OK )
    {
        if(ERR(rv = __CleanMDS(&MDSUContext)) != CSSM_OK)
        {
            sprintf( pError->ErrorString, "MDS Error (Clean): %X\n", rv );
            pError->ErrorCode = rv;
        }

        if((CSSM_OK == rv) &&
           (INSTALL_ACTION_UNINSTALL != Action))
        {
            /* Fill in the EMM schema structure */
            memcpy( &EMM.ModuleID, &EMM_GUID, sizeof(CSSM_GUID) );
            EMM.Manifest = *pCredential;
            strcpy( EMM.ModuleName, szModuleName );
            strcpy( EMM.Path, szSearchPath );
            EMM.CDSAVersion.Major = CDSA_VERSION_MAJOR;
            EMM.CDSAVersion.Minor = CDSA_VERSION_MINOR;
            EMM.EmmSpecVersion.Major = EMM_VERSION_MAJOR;
            EMM.EmmSpecVersion.Minor = EMM_VERSION_MINOR;
            strcpy( EMM.Desc, EMM_NAME);
            EMM.EmmVersion.Major = PRODUCT_VERSION_MAJOR;
            EMM.EmmVersion.Minor = PRODUCT_VERSION_MINOR;
            strcpy( EMM.EmmVendor, EMM_VENDOR_NAME);
            EMM.EmmType = EMM_TYPE;

            /* this section needs to be changed by any EMM that uses the
               PolicyStmt field in MDS
            */
            EMM.PolicyStmt.Data = NULL;
            EMM.PolicyStmt.Length = 0;


            /* Insert the EMM schema record into the MDS */
            ERR(rv = MDSU_UpdateSchema(&MDSUContext,
                                       &IfiMdsuCdsadirEMMSchema,
                                       &EMM,
                                       NULL));

            /* if there hasn't been an error, commit the new record */
            if(CSSM_OK == rv)
            {
                MDSU_Commit(&MDSUContext);

                /* clear error for return so EXISTING_RECORD error isn't
                   returned from a successful operation
                */
                rv = CSSM_OK;
            }
            else
            {
                sprintf(pError->ErrorString,
                        "MDS Error (common-DL): %X\n",
                        rv);
                pError->ErrorCode = rv;
            }
        }

        /* Shut down the MDS connection */
        MDSU_Term( &MDSUContext );
    }
    else
    {
        sprintf(pError->ErrorString,
                "MDS Error (Init): %X\n",
                rv);
        pError->ErrorCode = rv;
    }

    return rv;
}
