/*-----------------------------------------------------------------------
 * File: CSM_KEYMGR.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

/* MAF Includes */
#include "maf_include.h"

/* MAF-CSM Headers */
#include "csm_keymgr.h"
#include "csm_keymgr_db.h"
#include "csm_keymgr_internal.h"

/*
 * This stores the key file handles and key handles.
 * Yes, they are static variables.
 */
static CSSM_BOOL csm_kmIsInitialized = CSSM_FALSE;
static MAF_LIST_COLLECTION csm_kmKeyFileHandles;
static CSM_KM_HANDLES discardedKeyFileHandles;
static CSM_KM_HANDLE csm_kmNextKeyFileHandle = 1;

/*
 * Name: CSM_KMInit
 *
 * Description:
 *     Initializes the key manager.  This must be called before any other
 *     key manager function.
 *
 * Parameters:
 *     none
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     CSSMERR_CSP_INTERNAL_ERROR              - Could not initialize.
 */
CSSM_RETURN CSSMAPI CSM_KMInit( void )
{
    CSSM_RETURN ret = CSSM_OK;

    /* csm_kmIsInitialized probably needs to be locked down */
    if ( !csm_kmIsInitialized )
    {
        if ( MLC_Init( &csm_kmKeyFileHandles, 
                       csm_kmTearDownKeyFileHandle ) != CSSM_OK )
        {
            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            if ( csm_kmInitHandles( &discardedKeyFileHandles ) != CSSM_OK )
            {
                ret = CSSMERR_CSP_INTERNAL_ERROR;
            }
            else
            {
                csm_kmIsInitialized = CSSM_TRUE;
            }
        }
    }

    return ret;
}

/*
 * Name: CSM_KMTerm
 *
 * Description:
 *     Terminates the key manager.  This must be called for the key manager
 *     to terminate properly.
 *
 * Parameters:
 *     none
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     CSSMERR_CSP_INTERNAL_ERROR              - Could not terminate.
 */
CSSM_RETURN CSSMAPI CSM_KMTerm( void )
{
    CSSM_RETURN ret = CSSM_OK;

    if ( csm_kmIsInitialized )
    {
        csm_kmIsInitialized = CSSM_FALSE;

        if ( csm_kmTermHandles( &discardedKeyFileHandles ) != CSSM_OK )
        {
            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            if ( MLC_Term( &csm_kmKeyFileHandles ) != CSSM_OK )
            {
                ret = CSSMERR_CSP_INTERNAL_ERROR;
            }
        }
    }

    return ret;
}

/*
 * Name: CSM_KMCreateKeyFile
 * 
 * Description:
 *     Creates a key file with the name and location specified by pFilename
 *     The filename should not include an extension.
 *
 *     The initial ACL entry for the file as a whole can be specified in
 *     pInitialAclEntry. If pInitialAclEntry is NULL, then the default
 *     "all access" ACL will be applied. This ACL controls the ability
 *     of clients to add new keys, delete existing keys, and other general
 *     file manipulations. This ACL does not apply to individual keys. They
 *     each have their own ACL.
 *
 * Parameters:
 *     const char *                           pFilename
 *     const CSSM_ACL_ENTRY_INPUT *           pInitialAclEntry
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     KMERR_NOT_INITIALIZED                 - The key manager wasn't initialized.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 *     CSSM_CSP_FILE_CREATE_FAILED          - Could not create the file.
 *     CSSM_CSP_FILE_WRITE_FAILED           - Could not write the file.
 *     CSSM_CSP_FILE_CLOSE_FAILED           - Could not close the file.
 *     CSSM_CSP_FILE_DELETE_FAILED          - Could not delete the file.
 *     CSSM_ACL_INVALID_NEW_ACL_ENTRY       - The specified ACL can not be
 *                                            supported by the key manager, or
 *                                            the ACL structure is constructed
 *                                            incorrectly.
 */
CSSM_RETURN CSSMAPI CSM_KMCreateKeyFile(
                        const char *pFilename,
                        const CSSM_ACL_ENTRY_INPUT *pInitialAclEntry )
{
    CSSM_RETURN ret = CSSM_OK;

    if ( csm_kmIsInitialized != CSSM_TRUE )
    {
        ret = KMERR_NOT_INITIALIZED;
    }
    else if ( pInitialAclEntry /* Disabled for now. */ )
    {
        ret = KMERR_NOT_SUPPORTED;
    }
    else
    {
        ret = csm_kmCreateDb( pFilename );
    }

    return ret;
}

/*
 * Name: CSM_KMOpenKeyFile
 *
 * Description:
 *     Logically opens the file pFilename. The implementation may or may not
 *     actually open the file, but it must at least check the ability to open,
 *     and read, and write the file. Write permission is not checked in order
 *     to allow high value key files to be made read-only.  It pFilename is NULL,
 *     then a memory-only database will be created.  A memory-only database has
 *     a handle just as a normal file database does.
 *
 *     A handle is returned in phKeyFile that must be used by the caller to
 *     access the key file. The handle remains valid until CSM_KMCloseKeyFile is
 *     called.
 *
 * Parameters:
 *     const char *                            pFilename
 *     CSM_KM_HANDLE *                         phKeyFile
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     KMERR_NOT_INITIALIZED                 - The key manager wasn't initialized.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 *     KMERR_HANDLES_EXHAUSTED               - No more key file handles.
 *     KMERR_FILE_DOES_NOT_EXIST             - The file could not be found.
 *     CSSM_CSP_FILE_OPEN_FAILED            - The file could not be opened.
 *     CSSM_CSP_FILE_READ_FAILED            - The file could not be read.
 *     CSSM_CSP_FILE_FORMAT_ERROR           - The file isn't a valid database.
 *     CSSM_CSP_FILE_CLOSE_FAILED           - Could not close the file.
 *     CSSMERR_CSP_MEMORY_ERROR                - Couldn't allocate memory.
 *     CSSMERR_CSP_INTERNAL_ERROR              - Something failed that shouldn't.
 */
CSSM_RETURN CSSMAPI CSM_KMOpenKeyFile(
                        const char *pFilename,
                        CSM_KM_HANDLE *phKeyFile )
{
    CSSM_RETURN ret = CSSM_OK;
    CSSM_BOOL noHandles = CSSM_FALSE;
    int fd = -1;
    CSM_KM_KEY_FILE_HANDLES_NODE * pKeyFileHandlesNode = NULL;
    MLC_LOCK_REF mlc_lock;

    if ( csm_kmIsInitialized != CSSM_TRUE )
    {
        ret = KMERR_NOT_INITIALIZED;
    }
    else if ( !phKeyFile )
    {
        /* No phKeyFile pointer. */
        ret = KMERR_INVALID_HANDLE_POINTER;
    }
    else
    {
        /*
         * If a key file handle gets allocated, then this will get set to
         * that value.  So, in order to detect if a handle gets allocated,
         * this value, if it can be set, must initially be 0.
         */
        *phKeyFile = CSSM_INVALID_HANDLE;

        /* Check to see if there are no more key file handles to allocate. */
        if ( ( ret = csm_kmNoHandles( &discardedKeyFileHandles, 
                               &noHandles ) ) == CSSM_OK )
        {
            if ( noHandles &&
                 !csm_kmNextKeyFileHandle )
            {
                /* No more key file handles. */
                ret = KMERR_HANDLES_EXHAUSTED;
            }
            else
            {
                /*
                 * If pFilename is NULL, then the keys will be kept in memory only.
                 */
                /* Attempt to open the file. */
                if ( pFilename )
                {
                    ret = csm_kmOpenDb( pFilename, &fd );
                }
                if ( ret == CSSM_OK )
                {
                    /* Allocate the key file node, and the keys node. */
                    pKeyFileHandlesNode = 
                        (CSM_KM_KEY_FILE_HANDLES_NODE *)Addin_calloc(
                                        sizeof(CSM_KM_KEY_FILE_HANDLES_NODE),
                                        1,
                                        NULL );
                    if ( !pKeyFileHandlesNode )
                    {
                        ERR( ret = CSSMERR_CSP_MEMORY_ERROR );
                    }
                    else
                    {
                        /*
                         * Get a discarded key file handle.  If there is none,
                         * get a fresh one.
                         */
                        if ( ( ret = csm_kmGetHandle(
                                        &discardedKeyFileHandles,
                                        &pKeyFileHandlesNode->hKeyFile ) )
                                                                == CSSM_OK)
                        {
                            if ( !pKeyFileHandlesNode->hKeyFile )
                            {
                                pKeyFileHandlesNode->hKeyFile = 
                                                csm_kmNextKeyFileHandle++;
                            }

                            /* Fill in the key node and initialize key list */
                            *phKeyFile = pKeyFileHandlesNode->hKeyFile;
                            pKeyFileHandlesNode->fd = fd;
                            pKeyFileHandlesNode->keysNextHandle = 1;
                            if ( MLC_Init( &pKeyFileHandlesNode->keys,
                                           csm_kmTearDownKeys ) != CSSM_OK )
                            {
                                ERR( ret = CSSMERR_CSP_INTERNAL_ERROR );
                            }
                            else
                            {
                                if ( ( ret = csm_kmInitHandles(
                                          &pKeyFileHandlesNode->
                                                    discardedKeyHandles ) ) 
                                                                == CSSM_OK )
                                {
                                    if ( ( ret = MLC_AddItem( 
                                                    &csm_kmKeyFileHandles,
                                                    (void *)pKeyFileHandlesNode,
                                                    MLC_NO_LOCK,
                                                    &mlc_lock ) ) != CSSM_OK )
                                    {
                                        ret = ( ret == CSSM_ERRCODE_MEMORY_ERROR )
                                                ? CSSMERR_CSP_MEMORY_ERROR
                                                : CSSMERR_CSP_INTERNAL_ERROR;                                   
                                    } /* Add file handle node to the list */
                                } /* Initialize discarded key handles list */
                            } /* Initialize key list */
                        } /* Get a key file handle */
                    } /* Allocate the file tracking node */
                } /* Open DB file */

                /* Clean up if something went wrong */
                if ( ret != CSSM_OK )
                {
                    if ( fd != -1 )
                    {
                        csm_kmCloseDb( fd );
                    }

                    if ( pKeyFileHandlesNode )
                    {
                        MLC_Term( &pKeyFileHandlesNode->keys );
                        csm_kmTermHandles( 
                                &pKeyFileHandlesNode->discardedKeyHandles );

                        Addin_free( (void *)pKeyFileHandlesNode, NULL );
                    }

                    if ( phKeyFile && *phKeyFile )
                    {
                        csm_kmInsertHandle( &discardedKeyFileHandles,
                                            *phKeyFile );
                        *phKeyFile = CSSM_INVALID_HANDLE;
                    }
                } /* Clean up if something went wrong */
            } /* Check for exhausted handles */
        } /* Check for discarded file handles */
    } /* Parameter check */

    return ret;
}

/*
 * Name: CSM_KMCloseKeyFile
 *
 * Description:
 *     Closes the opened file referenced by hKeyFile, and invalidates the
 *     handle.  hKeyFile must have been obtained by calling CSM_KMOpenKeyFile.
 *
 * Paramters:
 *     CSM_KM_HANDLE                          hKeyFile
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     KMERR_NOT_INITIALIZED                 - The key manager wasn't initialized.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 *     CSSM_CSP_FILE_NOT_OPEN               - The file handle is not valid.
 *     CSSMERR_CSP_INTERNAL_ERROR              - Something failed that shouldn't.
 *     CSSM_CSP_FILE_CLOSE_FAILED           - Could not close the file.
 */
CSSM_RETURN CSSMAPI CSM_KMCloseKeyFile(
                        CSM_KM_HANDLE hKeyFile )
{
    CSSM_RETURN ret = CSSM_OK;
    CSM_KM_KEY_FILE_HANDLES_NODE * pKeyFileHandlesNode = NULL;

    if ( csm_kmIsInitialized != CSSM_TRUE )
    {
        ret = KMERR_NOT_INITIALIZED;
    }
    else if ( !hKeyFile )
    {
        ret = KMERR_INVALID_HANDLE;
    }
    else
    {
        /* First delete the file handle node. */
        if ( ( ret = MLC_DeleteItem(
                                &csm_kmKeyFileHandles,
                                csm_kmFindKeyFileHandle,
                                (void *)&hKeyFile,
                                (void **)&pKeyFileHandlesNode ) ) != CSSM_OK )
        {
            ret = ( ret == MLC_ERR_NOT_FOUND )
                            ? KMERR_FILE_NOT_OPEN
                            : CSSMERR_CSP_INTERNAL_ERROR;
        }
        else if ( !pKeyFileHandlesNode )
        {
            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            /*
             * If any dirty caching happened, now is the
             * time to flush to the file.
             *
             * Currently, however, the cache is write-through, so there are
             * no keys in memory which should be in the file, but aren't.
             */

            /*
             * Reclaim the key file handle.
             */
            ret = csm_kmInsertHandle( &discardedKeyFileHandles, hKeyFile );

            /*
             * If this is a memory only key list,
             * there is no need to close a file.
             */
            if ( pKeyFileHandlesNode->fd != -1 )
            {
                /* Close the file stream. */
                csm_kmCloseDb( pKeyFileHandlesNode->fd );
            }

            /* Destroy the keys. */
            if ( MLC_Term( &pKeyFileHandlesNode->keys ) != CSSM_OK )
            {
                ret = CSSMERR_CSP_INTERNAL_ERROR;
            }

            /* Destroy the discarded handles. */
            ret = csm_kmTermHandles( 
                            &pKeyFileHandlesNode->discardedKeyHandles );

            /* Destroy the node. */
            Addin_free( (void *)pKeyFileHandlesNode, NULL );
        } /* Remove the key file handle from the list */
    } /* Parameter check */

    return ret;
}

/*
 * Name: CSM_KMInsertKey
 *
 * Description:
 *     Inserts pKey into the file referenced by hKeyFile if
 *     pCredAndAclEntry->AccessCred contains the proper authorizations. The
 *     ACL for the new key is contained in pCredAndAclEntry->InitialAclEntry.
 *     If pKey is a public key, then pCredAndAclEntry must be NULL.
 *
 *     pKey must have pKey->KeyHeader.BlobType equal to CSSM_KEYBLOB_RAW.
 *     If pRefKey is present, it must also have the same
 *     pKey->KeyHeader.BlobType value.
 *
 *     If pKey is a private key, then pRefKey will be used to generate a
 *     reference hash associated with pKey. The reference hash is used to
 *     search for the key using CSM_KMFindKeyByHash. If pKey is not a private
 *     key, then pRefKey must be NULL.
 *
 *     A reference hash of the public key is stored along with all public
 *     and private keys. No reference hash is stored with symmetric keys, so
 *     CSM_KMFindKeyByHash is not available for those keys. The reference hash
 *     is calculated by taking the SHA-1 message digest of the unique key
 *     value. For RSA keys, this is the public modulus. For DSA and most other
 *     asymmetric key types, this is the public value. If pKey is a public key,
 *     then the reference hash is calculated directly from that key. If it is
 *     a private key, the hash is calculated from the value of pRefKey. 
 *
 *     pName is stored with a key so that it can be found using
 *     CSM_KMFindKeyByName. This value is optional for all key types, unless
 *     the key is a permanent key.  In the case of a permanent key, the value
 *     is required if and only if the key is not a public or private key (i.e.
 *     a symmetric key).
 *
 *     After successful completion of the function, and if phKey is not NULL,
 *     phKey will contain a unique reference that can be used to access the key
 *     with CSM_KMDeleteKey, CSM_KMGetKeyAcl, and CSM_KMFindKeyByHandle. The
 *     handle becomes invalid when either CSM_KMDeleteKey or CSM_KMCloseKeyFile
 *     is called with the same hKeyFile value that was used to perform the
 *     insertion.  If phKey is NULL and the key has the permanent attribute set,
 *     then the key is not stored in memory.  However, if it is a memory-only
 *     database, then it is an error.  This can be used to preserve memory when
 *     the key won't be needed soon.
 *
 *     If the pKey->KeyHeader.KeyAttr field does not have the
 *     CSSM_KEYATTR_PERMANENT attribute set, the key is stored in memory only.
 *     The key will be wiped and deleted when CSM_KMCloseKeyFile is called with
 *     hKeyFile.
 *
 * Parameters:
 *     CSM_KM_HANDLE                          hKeyFile
 *     const CSSM_KEY *                       pKey
 *     const CSSM_KEY *                       pRefKey
 *     const CSSM_DATA *                      pName
 *     const CSSM_RESOURCE_CONTROL_CONTEXT *  pCredAndAclEntry
 *     CSM_KM_KEY_HANDLE *                    phKey
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     KMERR_NOT_INITIALIZED                 - The key manager wasn't initialized.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 *     CSSMERR_CSP_MEMORY_ERROR                - Couldn't allocate memory.
 *     CSSM_CSP_INPUT_LENGTH_OVERSIZE       - The input was too big.
 *     CSSM_CSP_OPERATION_UNSUPPORTED       - Could not hash this key.
 *     CSSMERR_CSP_INTERNAL_ERROR              - Something failed that shouldn't.
 *     KMERR_FILE_DOES_NOT_EXIST             - The file referenced by hKeyFile
 *                                            was not found. Most likely it
 *                                            was deleted by another process.
 *                                            Use CSM_KMCreateKeyFile to
 *                                            create the file.
 *     CSSM_CSP_PRIKEY_ALREADY_EXIST        - An identical key already exists
 *                                            in the file.
 *     CSSM_CSP_FILE_READ_FAILED            - The file could not be read.
 *     CSSM_CSP_FILE_FORMAT_ERROR           - The file isn't a valid database.
 *     CSSM_CSP_FILE_WRITE_FAILED           - Could not write the file.
 *     KMERR_HANDLES_EXHAUSTED               - No more key handles for hKeyFile.
 */
CSSM_RETURN CSSMAPI CSM_KMInsertKey(
                        CSM_KM_HANDLE hKeyFile,
                        const CSSM_KEY *pKey,
                        const CSSM_KEY *pRefKey,
                        const CSSM_DATA *pName,
                        const CSSM_RESOURCE_CONTROL_CONTEXT *pCredAndAclEntry,
                        CSM_KM_KEY_HANDLE *phKey )
{
    CSSM_RETURN ret = CSSM_OK;
    CSSM_BOOL storeKeyInFile = CSSM_TRUE;
    CSSM_BOOL storeKeyInMemory = CSSM_TRUE;
    CSM_KM_DB_RECORD * pRecord = NULL;
    CSM_KM_KEY_FILE_HANDLES_NODE * pKeyFileHandlesNode = NULL;
    MLC_LOCK_REF mlcFileHandlesLock = NULL;
    CSSM_BOOL mlcFileHandlesLockIsLocked = CSSM_FALSE;
    CSM_KM_KEYS_FIND_DUPLICATE findDuplicate;
    CSM_KM_KEYS_NODE * pKeysNode = NULL;
    MLC_LOCK_REF mlcKeysLock = NULL;
    unsigned int numQueries;
    CSM_KM_DB_QUERY_ITEM queries[2];
    CSSM_BOOL foundInFile = CSSM_FALSE;
    long dbIndex = 0;
    CSM_KM_HANDLE newHandle = 0;
    CSSM_BOOL noHandles = CSSM_FALSE;
    CSM_KM_KEYS_NODE * pNewKeysNode = NULL;

    /*
     * Verify the input data to a point where this sort of checking
     * later on is not required.
     */
    if ( csm_kmIsInitialized != CSSM_TRUE )
    {
        ret = KMERR_NOT_INITIALIZED;
    }
    else if ( !hKeyFile )
    {
        ret = KMERR_INVALID_HANDLE;
    }
    else if ( !pKey )
    {
        ret = CSSMERR_CSP_INVALID_KEY;
    }
    else if ( ( pKey->KeyHeader.KeyClass == CSSM_KEYCLASS_PUBLIC_KEY ) &&
              pCredAndAclEntry )
    {
        ret = CSSMERR_CSP_OBJECT_ACL_NOT_SUPPORTED;
    }
    else if ( !( pKey->KeyHeader.BlobType == CSSM_KEYBLOB_RAW ) )
    {
        ret = KMERR_CAN_NOT_INTERPRET_KEY_DATA;
    }
    else if ( pRefKey && 
              ( ( pRefKey->KeyHeader.BlobType != CSSM_KEYBLOB_RAW ) ||
                ( pRefKey->KeyHeader.KeyClass != CSSM_KEYCLASS_PUBLIC_KEY ) ) )
    {
        ret = CSSMERR_CSP_INVALID_KEY;
    }
    else if ( ( pKey->KeyHeader.KeyClass != CSSM_KEYCLASS_PRIVATE_KEY ) &&
              pRefKey )
    {
        ret = CSSMERR_CSP_INVALID_KEY;      
    }
    else
    {
        /*
         * If the key is marked as permanent, then it will be stored on disk.
         * For a permanent key, if phKey is NULL it will not be stored in
         * memory.
         *
         * For non-permanent keys, it will be stored in memory only.
         */

        if ( pKey->KeyHeader.KeyAttr & CSSM_KEYATTR_PERMANENT )
        {
            storeKeyInFile = CSSM_TRUE;
            storeKeyInMemory = phKey ? CSSM_TRUE : CSSM_FALSE;
        }
        else
        {
            storeKeyInFile = CSSM_FALSE;
            storeKeyInMemory = CSSM_TRUE;
        }
    }

    ret = csm_kmBuildNewDBRecord( pKey,
                                  pRefKey,
                                  pName,
                                  pCredAndAclEntry,
                                  &pRecord );

    if (ret == CSSM_OK)
    {
        /* Find the file handle node. */

        if ((ret = MLC_FindItem(&csm_kmKeyFileHandles,
                                csm_kmFindKeyFileHandle,
                                (void *)&hKeyFile,
                                MLC_WRITE_LOCK,
                                &mlcFileHandlesLock,
                                (void **)&pKeyFileHandlesNode)) != CSSM_OK)
        {
            ret = (ret == MLC_ERR_NOT_FOUND) ? KMERR_FILE_DOES_NOT_EXIST :
                                               CSSMERR_CSP_INTERNAL_ERROR;
        }
        else if (!pKeyFileHandlesNode)
        {
            MLC_ReleaseItem(MLC_WRITE_LOCK, mlcFileHandlesLock);
            mlcFileHandlesLockIsLocked = CSSM_FALSE;

            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            mlcFileHandlesLockIsLocked = CSSM_TRUE;

            /*
             * Don't check the cache if neither the hash nor the name is
             * available.
             */
            if ((pRecord->keyRefHashValid == CSSM_TRUE) ||
                (pName && pName->Data && (pName->Length > 0)))
            {
                /* Look and see if the key exists in memory. */

                if (pRecord->keyRefHashValid == CSSM_TRUE)
                {
                    findDuplicate.pHash = pRecord->keyRefHash;
                }
                else
                {
                    findDuplicate.pHash = NULL;
                }

                if ( pName && pName->Data && (pName->Length > 0))
                {
                    findDuplicate.pName = pName;
                }
                else
                {
                    findDuplicate.pName = NULL;
                }

                if ((ret = MLC_FindItem(&(pKeyFileHandlesNode->keys),
                                        csm_kmFindKeyDuplicate,
                                        (void *)&findDuplicate,
                                        MLC_NO_LOCK,
                                        &mlcKeysLock,
                                        (void **)&pKeysNode)) == CSSM_OK)
                {
                    ERR( ret = CSSMERR_CSP_PRIVATE_KEY_ALREADY_EXISTS );
                }
                else
                {
                    ret = (ret == MLC_ERR_NOT_FOUND) ? CSSM_OK :
                                                       CSSMERR_CSP_INTERNAL_ERROR;
                }
            }

            /* How about in the file? */
            if ((ret == CSSM_OK) &&
                (pKeyFileHandlesNode->fd != -1) &&
                (storeKeyInFile == CSSM_TRUE))
            {
                /*
                 * Create the query.  If the neither the hash was created
                 * nor a name given, then an error occured.  It should have
                 * been caught up at the top of the function, though.
                 */

                numQueries = 0;

                if (pRecord->keyRefHashValid == CSSM_TRUE)
                {
                    queries[numQueries].type = CSM_KM_DB_QUERY_HASH;
                    queries[numQueries].value = (void *)(pRecord->keyRefHash);
                    numQueries++;
                }

                if (pName && pName->Data && (pName->Length > 0))
                {
                    queries[numQueries].type = CSM_KM_DB_QUERY_KEY_NAME;
                    queries[numQueries].value = (void*)pRecord->keyName;
                    numQueries++;
                }


                if (numQueries)
                {
                    if ((ret = csm_kmQueryDb( pKeyFileHandlesNode->fd,
                                              numQueries,
                                              queries,
                                              0, /* Don't need anything back, just checking */
                                              NULL,
                                              0, /* No lock required */
                                              NULL,
                                              &foundInFile ) ) == CSSM_OK )
                    {
                        if (foundInFile == CSSM_TRUE)
                        {
                            ERR( ret = CSSMERR_CSP_PRIVATE_KEY_ALREADY_EXISTS );
                        }
                    }
                }
            }
        }
    }

    if ((ret == CSSM_OK) && (storeKeyInFile == CSSM_TRUE))
    {
        ret = csm_kmInsertRecord(pKeyFileHandlesNode->fd, pRecord, &dbIndex );
    }

    if ((ret == CSSM_OK) && (storeKeyInMemory == CSSM_TRUE))
    {
        /* Check to see if there are any available key handles. */

        ret = csm_kmNoHandles(&(pKeyFileHandlesNode->discardedKeyHandles),
                              &noHandles);
        if (ret == CSSM_OK)
        {
            if ((noHandles == CSSM_TRUE) &&
                !(pKeyFileHandlesNode->keysNextHandle))
            {
                /* Out of key handles for this open file. */
                ret = KMERR_HANDLES_EXHAUSTED;
            }
        }

        if (ret == CSSM_OK)
        {
            pNewKeysNode = (CSM_KM_KEYS_NODE *)
                           Addin_calloc(sizeof(CSM_KM_KEYS_NODE), 1, NULL);
            if (!pNewKeysNode)
            {
                ret = CSSMERR_CSP_MEMORY_ERROR;
            }
            else
            {
                ret = csm_kmGetHandle(&(pKeyFileHandlesNode->discardedKeyHandles),
                                      &newHandle);
                if (ret == CSSM_OK)
                {
                    if (newHandle)
                    {
                        pNewKeysNode->hKey = newHandle;
                    }
                    else
                    {
                        pNewKeysNode->hKey = pKeyFileHandlesNode->keysNextHandle++;
                    }

                    if (phKey)
                    {
                        *phKey = pNewKeysNode->hKey;
                    }

                    pNewKeysNode->dbIndex = dbIndex;
                    pNewKeysNode->pRecord = pRecord;

                    /* Insert it into the list. */
                    if (MLC_AddItem(&(pKeyFileHandlesNode->keys),
                                    (void *)pNewKeysNode,
                                    MLC_NO_LOCK,
                                    &mlcKeysLock)
                        != CSSM_OK)
                    {
                        /* Reclaim the handle. */
                        csm_kmInsertHandle(
                            &(pKeyFileHandlesNode->discardedKeyHandles),
                            pNewKeysNode->hKey);

                        ret = (ret == CSSM_ERRCODE_MEMORY_ERROR) 
                                                ? CSSMERR_CSP_MEMORY_ERROR
                                                : CSSMERR_CSP_INTERNAL_ERROR;
                    }
                    else
                    {
                        /* So they aren't freed. */
                        pRecord = NULL;
                        pNewKeysNode = NULL;
                    }
                }
            }
        }
    }

    if ((ret != CSSM_OK) && phKey)
    {
        *phKey = 0;
    }

    if (pRecord)
    {
        Addin_free((void *)pRecord, NULL);
    }

    if (pNewKeysNode)
    {
        Addin_free((void *)pNewKeysNode, NULL);
    }

    if (mlcFileHandlesLockIsLocked == CSSM_TRUE)
    {
        if (MLC_ReleaseItem(MLC_WRITE_LOCK, mlcFileHandlesLock) != CSSM_OK)
        {
            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
    }

    return ret;
}

/*
 * Name: CSM_KMDeleteKey
 *
 * Description:
 *     Removes the key referenced by hKey from the file referenced by hKeyFile
 *     if pAccessCred contains the proper authorizations. If hKey is a public
 *     key, then pAccessCred must be NULL.
 *
 *     hKey can be obtained using the CSM_KMInsertKey function, or any of the
 *     key searching functions.
 *
 *     After successful completion of the function, hKey will no longer contain
 *     a valid handle.
 *
 * Parameters:
 *     CSM_KM_HANDLE                          hKeyFile
 *     CSM_KM_KEY_HANDLE                      hKey
 *     const CSSM_ACCESS_CREDENTIALS *        pAccessCred
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     KMERR_NOT_INITIALIZED                 - The key manager wasn't initialized.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 *     KMERR_FILE_DOES_NOT_EXIST             - The file referenced by hKeyFile
 *                                            was not found. Most likely it
 *                                            was deleted by another process.
 *                                            Use CSM_KMCreateKeyFile to
 *                                            create the file.
 *     CSSMERR_CSP_INTERNAL_ERROR              - Something failed that shouldn't.
 *     CSSMERR_CSP_PRIVATE_KEY_NOT_FOUND            - The key couldn't be found.
 *     CSSM_CSP_FILE_READ_FAILED            - The file could not be read.
 *     CSSM_CSP_FILE_FORMAT_ERROR           - The file isn't a valid database.
 *     CSSM_CSP_FILE_WRITE_FAILED           - Could not write the file.
 */
CSSM_RETURN CSSMAPI CSM_KMDeleteKey(
                        CSM_KM_HANDLE hKeyFile,
                        CSM_KM_KEY_HANDLE hKey,
                        const CSSM_ACCESS_CREDENTIALS *pAccessCred )
{
    CSSM_RETURN ret = CSSM_OK;
    MLC_LOCK_REF mlcFileHandlesLock = NULL;
    MLC_LOCK_REF mlcLock = NULL;
    CSSM_BOOL mlcFileHandlesLockIsLocked = CSSM_FALSE;
    CSM_KM_KEY_FILE_HANDLES_NODE * pKeyFileHandlesNode = NULL;
    CSM_KM_KEYS_NODE * pKeysNodeAcl = NULL;
    CSM_KM_KEYS_NODE * pKeysNode = NULL;
    CSM_KM_DB_QUERY_ITEM indexQuery;
    CSSM_KEY wrappedKey;
    CSSM_KEY unWrappedKey;

    if (csm_kmIsInitialized != CSSM_TRUE)
    {
        ret = KMERR_NOT_INITIALIZED;
    }
    else if (!hKeyFile || !hKey)
    {
        ret = KMERR_INVALID_HANDLE;
    }
    else
    {
        /* Find the file handle node. */

        if ((ret = MLC_FindItem(&csm_kmKeyFileHandles,
                                csm_kmFindKeyFileHandle,
                                (void *)&hKeyFile,
                                MLC_WRITE_LOCK,
                                &mlcFileHandlesLock,
                                (void **)&pKeyFileHandlesNode)) != CSSM_OK)
        {
            ret = (ret == MLC_ERR_NOT_FOUND) ? KMERR_FILE_DOES_NOT_EXIST
                                             : CSSMERR_CSP_INTERNAL_ERROR;
        }
        else if (!pKeyFileHandlesNode)
        {
            mlcFileHandlesLockIsLocked = CSSM_TRUE;

            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            mlcFileHandlesLockIsLocked = CSSM_TRUE;
        }
    }

    /* Find the key in the cache and check the ACL */
    if ( ret == CSSM_OK )
    {
        if ( ( ret = MLC_FindItem( &pKeyFileHandlesNode->keys,
                                   csm_kmFindKeyByHandle,
                                   (void*)&hKey,
                                   MLC_READ_LOCK,
                                   &mlcLock,
                                   (void**)&pKeysNodeAcl ) ) != CSSM_OK )
        {
            ret = (ret == MLC_ERR_NOT_FOUND) ? CSSMERR_CSP_PRIVATE_KEY_NOT_FOUND :
                                               CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            if ( ( pKeysNodeAcl->pRecord->acl.aclType != 0 ) &&
                 ( pAccessCred == NULL ) )
            {
                ERR( ret = CSSMERR_CSP_MISSING_ATTR_ACCESS_CREDENTIALS );
            }
            else if ( pKeysNodeAcl->pRecord->acl.aclType != 0 )
            {
                CSSM_DATA PbeSalt;
                CSSM_DATA PbeIV;

                memcpy( &wrappedKey.KeyHeader,
                        &pKeysNodeAcl->pRecord->keyHeader,
                        sizeof(pKeysNodeAcl->pRecord->keyHeader) );

                wrappedKey.KeyData.Data = pKeysNodeAcl->pRecord->key;
                wrappedKey.KeyData.Length = pKeysNodeAcl->pRecord->keyLength;
                unWrappedKey.KeyData.Data = NULL;
                unWrappedKey.KeyData.Length = 0;

                /* Attempt to unwrap the key. */
                PbeSalt.Data = pKeysNodeAcl->pRecord->WrapPBESalt;
                PbeSalt.Length = sizeof(pKeysNodeAcl->pRecord->WrapPBESalt);
                PbeIV.Data = pKeysNodeAcl->pRecord->WrapIV;
                PbeIV.Length = pKeysNodeAcl->pRecord->WrapIVLength;
                ret = csm_kmUnwrapKey( MAF_ALLOCATE_LOCAL,
                                       &wrappedKey,
                                       &unWrappedKey,
                                       pAccessCred,
                                       CSSM_SAMPLE_TYPE_PASSWORD,
                                       NULL,
                                       &PbeSalt,
                                       &PbeIV );
                if ( unWrappedKey.KeyData.Data )
                {
                    memset( unWrappedKey.KeyData.Data,
                            0,
                            unWrappedKey.KeyData.Length );
                    Addin_free( unWrappedKey.KeyData.Data, NULL );
                }
            } /* Check access credentials */

            MLC_ReleaseItem( MLC_READ_LOCK, mlcLock );
        } /* Key found in cache */
    }

    if (ret == CSSM_OK)
    {
        /* Delete the key node. */

        if ((ret = MLC_DeleteItem(&(pKeyFileHandlesNode->keys),
                                  csm_kmFindKeyByHandle,
                                  (void *)&hKey,
                                  (void **)&pKeysNode))
            != CSSM_OK)
        {
            ret = (ret == MLC_ERR_NOT_FOUND) ? CSSMERR_CSP_PRIVATE_KEY_NOT_FOUND :
                                               CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            if (!pKeysNode || (pKeysNode->hKey != hKey) || !(pKeysNode->pRecord))
            {
                ret = CSSMERR_CSP_INTERNAL_ERROR;
            }
        }
    }

    if ((ret == CSSM_OK) &&
        ((pKeysNode->pRecord->keyHeader.KeyAttr) & CSSM_KEYATTR_PERMANENT))
    {
        /*
         * The key node has been removed from the list and
         * this is a key also in the file.
         */

        /* If there is a file open, delete the record from the file. */
        if ((pKeyFileHandlesNode->fd) != -1)
        {
            indexQuery.type = CSM_KM_DB_QUERY_INDEX;
            indexQuery.value = (void *)&(pKeysNode->dbIndex);

            ret = csm_kmDeleteRecord(pKeyFileHandlesNode->fd, 1,
                                     &indexQuery);
        }
    }

    if (ret == CSSM_OK)
    {
        /* Recycle the key handle. */
        ret = csm_kmInsertHandle(&(pKeyFileHandlesNode->discardedKeyHandles),
                                 pKeysNode->hKey);
    }

    if (pKeysNode)
    {
        if (pKeysNode->pRecord)
        {
            memset((void *)(pKeysNode->pRecord), 0, sizeof(CSM_KM_DB_RECORD));
            Addin_free((void *)(pKeysNode->pRecord), NULL);
        }

        Addin_free((void *)pKeysNode, NULL);
    }

    if (mlcFileHandlesLockIsLocked == CSSM_TRUE)
    {
        if (MLC_ReleaseItem(MLC_WRITE_LOCK, mlcFileHandlesLock) != CSSM_OK)
        {
            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
    }

    return ret;
}

/*
 * Name: CSM_KMFindKeyByHash
 *
 * Description:
 *     Search the key database for existence of a key by hash. If the requested
 *     key exists, it is returned in pResultKey with the format specified in
 *     pFormatContext. A handle for manipulation of the key is also returned in
 *     phKey.
 *
 *     pKeyHash must be calculated as described for CSM_KMInsertKey.
 *
 *     Setting pResultKey to NULL will cause only the key handle to be returned.
 *     pAccessCred may be NULL in this situation.
 *
 *     When a private or symmetric key value is returned, it is in an unwrapped
 *     state. pAccessCred must contain the appropriate credentials to authorize
 *     unwrapping the key. The key manager will make any callbacks necessary to
 *     acquire authentication samples.
 *
 *     Key values can be returned in any format supported by CSM_CreateKeyBlob.
 *     pFormatContext is used to indicate the desired format.
 *     pFormatContext->Format must be one of the raw key blob format constants
 *     (CSSM_KEYBLOB_RAW_*). If the format requested is CSSM_KEYBLOB_RAW_SPKI,
 *     the pFormatContext->SigAlg field is used to generate the signature
 *     algorithm tag in the s-expression.  
 *
 *     After successful completion of the function, phKey will contain a unique
 *     reference that can be used to access the key with CSM_KMDeleteKey,
 *     CSM_KMGetKeyAcl, and CSM_KMFindKeyByHandle. The handle becomes invalid
 *     when either CSM_KMDeleteKey or CSM_KMCloseKeyFile is called with the same
 *     hKeyFile value that was used to perform the insertion.
 *
 * Parameters:
 *     CSM_KM_HANDLE                          hKeyFile
 *     const CSSM_DATA *                      pKeyHash
 *     const CSM_KM_FORMAT_CONTEXT *          pFormatContext
 *     const CSSM_ACCESS_CREDENTIALS *        pAccessCred
 *     CSSM_KEY *                             pResultKey
 *     CSM_KM_KEY_HANDLE *                    phKey
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     KMERR_NOT_INITIALIZED                 - The key manager wasn't initialized.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 *     KMERR_FILE_DOES_NOT_EXIST             - The file referenced by hKeyFile
 *                                            was not found. Most likely it
 *                                            was deleted by another process.
 *                                            Use CSM_KMCreateKeyFile to
 *                                            create the file.
 *     CSSMERR_CSP_INTERNAL_ERROR              - Something failed that shouldn't.
 *     CSSMERR_CSP_PRIVATE_KEY_NOT_FOUND            - The key couldn't be found.
 *     CSSM_CSP_FILE_READ_FAILED            - The file could not be read.
 *     CSSM_CSP_FILE_FORMAT_ERROR           - The file isn't a valid database.
 *     CSSMERR_CSP_MEMORY_ERROR                - Couldn't allocate memory.
 *     KMERR_HANDLES_EXHAUSTED               - No more key handles for hKeyFile.
 */
CSSM_RETURN CSSMAPI CSM_KMFindKeyByHash(
                        CSM_KM_HANDLE hKeyFile,
                        const CSSM_DATA *pKeyHash,
                        const CSM_KM_FORMAT_CONTEXT *pFormatContext,
                        const CSSM_ACCESS_CREDENTIALS *pAccessCred,
                        CSSM_KEY *pResultKey,
                        CSM_KM_KEY_HANDLE *phKey )
{
    CSSM_RETURN ret = CSSM_OK;
    const MAF_MODULE_ATTACH_TRACKER * pAttachTracker = NULL;
    MLC_LOCK_REF mlcFileHandlesLock = NULL;
    CSSM_BOOL mlcFileHandlesLockIsLocked = CSSM_FALSE;
    CSM_KM_KEY_FILE_HANDLES_NODE * pKeyFileHandlesNode = NULL;
    MLC_LOCK_REF mlcKeysLock = NULL;
    CSSM_BOOL mlcKeysLockIsLocked = CSSM_FALSE;
    CSM_KM_KEYS_NODE * pKeysNode = NULL;
    CSM_KM_DB_RECORD * pRecord = NULL;
    CSM_KM_DB_QUERY_ITEM query;
    CSM_KM_DB_QUERY_ITEM requests[2];
    CSSM_BOOL found = CSSM_FALSE;
    CSM_KM_KEYS_NODE * pNewKeysNode = NULL;
    CSM_KM_HANDLE newHandle = 0;
    long index = 0;
    CSSM_KEY wrappedKey;
    CSSM_KEY unWrappedKey;

    pAttachTracker = Addin_GetAttachTracker();

    if (csm_kmIsInitialized != CSSM_TRUE)
    {
        ret = KMERR_NOT_INITIALIZED;
    }
    else if (!hKeyFile )
    {
        ret = KMERR_INVALID_HANDLE;
    }
    else if ( !pKeyHash || 
              (pKeyHash->Length != CSM_KM_DB_HASH_SIZE) || 
              !(pKeyHash->Data) )
    {
        ret = CSSMERR_CSP_INVALID_DATA;
    }
    /* When access credentials exist: (pResultKey && !pAccessCred) || */
    else if ( !pResultKey && !phKey )
    {
        ret = KMERR_INVALID_HANDLE_POINTER;
    }
    /* || (pResultKey && !pFormatContext) */
    else if (!pAttachTracker)
    {
        ret = CSSMERR_CSP_INTERNAL_ERROR;
    }
    else
    {
        /* Find the file handle node. */

        if ((ret = MLC_FindItem(&csm_kmKeyFileHandles,
                                csm_kmFindKeyFileHandle,
                                (void *)&hKeyFile,
                                MLC_WRITE_LOCK,
                                &mlcFileHandlesLock,
                                (void **)&pKeyFileHandlesNode)) != CSSM_OK)
        {
            ret = (ret == MLC_ERR_NOT_FOUND) ? KMERR_FILE_DOES_NOT_EXIST
                                             : CSSMERR_CSP_INTERNAL_ERROR;
        }
        else if (!pKeyFileHandlesNode)
        {
            mlcFileHandlesLockIsLocked = CSSM_TRUE;

            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            mlcFileHandlesLockIsLocked = CSSM_TRUE;
        }
    }

    if (ret == CSSM_OK)
    {
        /* Find the key node. */

        if ((ret = MLC_FindItem(&(pKeyFileHandlesNode->keys),
                                csm_kmFindKeyByHash,
                                (void *)pKeyHash,
                                MLC_READ_LOCK,
                                &mlcKeysLock,
                                (void **)&pKeysNode)) != CSSM_OK)
        {
            ret = (ret == MLC_ERR_NOT_FOUND) ? CSSMERR_CSP_PRIVATE_KEY_NOT_FOUND :
                                               CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            mlcKeysLockIsLocked = CSSM_TRUE;

            if (phKey)
            {
                *phKey = pKeysNode->hKey;
            }

            if (!pKeysNode || !(pKeysNode->pRecord))
            {
                ret = CSSMERR_CSP_INTERNAL_ERROR;
            }
        }
    }

    if ((ret == CSSMERR_CSP_PRIVATE_KEY_NOT_FOUND) &&
        (pKeyFileHandlesNode->fd != -1))
    {
        /* The key wasn't in memory.  Check the file. */

        /* Allocate a record. */
        pRecord = (CSM_KM_DB_RECORD *)
                  Addin_calloc(sizeof(CSM_KM_DB_RECORD), 1, NULL);
        if (!pRecord)
        {
            ret = CSSMERR_CSP_MEMORY_ERROR;
        }
        else
        {
            query.type = CSM_KM_DB_QUERY_HASH;
            query.value = (void *)(pKeyHash->Data);

            requests[0].type = CSM_KM_DB_QUERY_RECORD;
            requests[0].value = (void *)pRecord;
            requests[1].type = CSM_KM_DB_QUERY_INDEX;
            requests[1].value = (void *)&index;

            if ((ret = csm_kmQueryDb( pKeyFileHandlesNode->fd,
                                      1,
                                      &query,
                                      2,
                                      requests,
                                      0, /* No lock required */
                                      NULL,
                                      &found ) ) == CSSM_OK)
            {
                if (found == CSSM_FALSE)
                {
                    ret = CSSMERR_CSP_PRIVATE_KEY_NOT_FOUND;
                }
            }
        }

        if (ret == CSSM_OK)
        {
            /* The record was recovered.  Create the node. */
            pNewKeysNode = (CSM_KM_KEYS_NODE *)
                           Addin_calloc(sizeof(CSM_KM_KEYS_NODE), 1, NULL);
            if (!pNewKeysNode)
            {
                ret = CSSMERR_CSP_MEMORY_ERROR;
            }
            else
            {
                if ((ret = csm_kmGetHandle(
                               &(pKeyFileHandlesNode->discardedKeyHandles),
                               &newHandle)) == CSSM_OK)
                {
                    if (newHandle)
                    {
                        pNewKeysNode->hKey = newHandle;
                    }
                    else if (!(pKeyFileHandlesNode->keysNextHandle))
                    {
                        ret = KMERR_HANDLES_EXHAUSTED;
                    }
                    else
                    {
                        pNewKeysNode->hKey =
                            pKeyFileHandlesNode->keysNextHandle++;
                        newHandle = pNewKeysNode->hKey;
                    }
                }

                if (ret == CSSM_OK)
                {
                    pNewKeysNode->dbIndex = index;
                    pNewKeysNode->pRecord = pRecord;

                    if (phKey)
                    {
                        *phKey = pNewKeysNode->hKey;
                    }
                }
            }
        }

        if (ret == CSSM_OK)
        {
            /* The key node has been created, add it to the list. */

            if (MLC_AddItem(&(pKeyFileHandlesNode->keys),
                            (void *)pNewKeysNode,
                            MLC_READ_LOCK,
                            &mlcKeysLock) != CSSM_OK)
            {
                /* Reclaim the key handle. */
                csm_kmInsertHandle(&(pKeyFileHandlesNode->discardedKeyHandles),
                                   pNewKeysNode->hKey);

                ret = (ret == CSSM_ERRCODE_MEMORY_ERROR)
                                            ? CSSMERR_CSP_MEMORY_ERROR
                                            : CSSMERR_CSP_INTERNAL_ERROR;
            }
            else
            {
                mlcKeysLockIsLocked = CSSM_TRUE;

                /* So we can get some info from the node. */
                pKeysNode = pNewKeysNode;

                /* So they won't get freed. */
                pRecord = NULL;
                pNewKeysNode = NULL;
            }
        }
    }

    if ((ret == CSSM_OK) && (!pKeysNode || !(pKeysNode->pRecord)))
    {
        /*
         * pKeysNode should point to either a node already in memory, or
         * a new node created by loading a record from the file.
         */

        ret = CSSMERR_CSP_INTERNAL_ERROR;
    }

    if ((ret == CSSM_OK) && pResultKey)
    {
        /*
         * The key has been retrieved.  Copy the header, format the key,
         * and set the file handle.
         */

        memset((void *)&unWrappedKey, 0, sizeof(CSSM_KEY));

        if ( ( pKeysNode->pRecord->acl.aclType != 0 ) &&
             ( pAccessCred == NULL ) )
        {
            ERR( ret = CSSMERR_CSP_INVALID_ACCESS_CREDENTIALS );
        }
        else if ( pKeysNode->pRecord->acl.aclType != 0 )
        {
            CSSM_DATA PbeSalt;
            CSSM_DATA PbeIV;

            memcpy((void *)&(wrappedKey.KeyHeader),
                   (const void *)&(pKeysNode->pRecord->keyHeader),
                   sizeof(CSSM_KEYHEADER));
            wrappedKey.KeyData.Data = pKeysNode->pRecord->key;
            wrappedKey.KeyData.Length = pKeysNode->pRecord->keyLength;

            /* Attempt to unwrap the key. */
            PbeSalt.Data = pKeysNode->pRecord->WrapPBESalt;
            PbeSalt.Length = sizeof(pKeysNode->pRecord->WrapPBESalt);
            PbeIV.Data = pKeysNode->pRecord->WrapIV;
            PbeIV.Length = pKeysNode->pRecord->WrapIVLength;
            ret = csm_kmUnwrapKey( MAF_ALLOCATE_LOCAL,
                                   &wrappedKey,
                                   &unWrappedKey,
                                   pAccessCred,
                                   CSSM_SAMPLE_TYPE_PASSWORD,
                                   NULL,
                                   &PbeSalt,
                                   &PbeIV );
            if ((ret != CSSM_OK) && unWrappedKey.KeyData.Data)
            {
                Addin_free((void *)unWrappedKey.KeyData.Data, NULL);
                unWrappedKey.KeyData.Data = NULL;
                unWrappedKey.KeyData.Length = 0;
            }
        }
        else
        {
            memcpy((void *)&(unWrappedKey.KeyHeader),
                   (const void *)&(pKeysNode->pRecord->keyHeader),
                   sizeof(CSSM_KEYHEADER));

            unWrappedKey.KeyData.Data = (uint8 *)
                Addin_calloc(1, pKeysNode->pRecord->keyLength, NULL);
            if (!(unWrappedKey.KeyData.Data))
            {
                ret = CSSMERR_CSP_MEMORY_ERROR;
            }

            unWrappedKey.KeyData.Length = pKeysNode->pRecord->keyLength;
        }

        /* For now, just copy the key. */

        if (ret == CSSM_OK)
        {
            /* Reformat the key to match the requested format */
            if ((ret = CSM_KMReformatKey( MAF_ALLOCATE_LOCAL,
                                          &unWrappedKey,
                                          pFormatContext )) == CSSM_OK )
            {
                memcpy( &pResultKey->KeyHeader,
                        &unWrappedKey.KeyHeader,
                        sizeof(unWrappedKey.KeyHeader) );
                pResultKey->KeyData = unWrappedKey.KeyData;
            }
        }

        if ( ret != CSSM_OK )
        {
            if ( unWrappedKey.KeyData.Data != NULL )
            {
                Addin_free((void *)unWrappedKey.KeyData.Data, NULL);
                unWrappedKey.KeyData.Data = NULL;
                unWrappedKey.KeyData.Length = 0;
            }
        }
    }

    if ((ret != CSSM_OK) && phKey)
    {
        *phKey = 0;
    }

    if (pRecord)
    {
        Addin_free((void *)pRecord, NULL);
    }

    if (pNewKeysNode)
    {
        Addin_free((void *)pNewKeysNode, NULL);
    }

    if (mlcKeysLockIsLocked == CSSM_TRUE)
    {
        if (MLC_ReleaseItem(MLC_READ_LOCK, mlcKeysLock) != CSSM_OK)
        {
            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
    }

    if (mlcFileHandlesLockIsLocked == CSSM_TRUE)
    {
        if (MLC_ReleaseItem(MLC_WRITE_LOCK, mlcFileHandlesLock) != CSSM_OK)
        {
            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
    }

    return ret;
}

/*
 * Name: CSM_KMFindKeyByName
 *
 * Description:
 *     Search the key database for existence of a key by name. If the requested
 *     key exists, it is returned in pResultKey with the format specified in
 *     pFormatContext. A handle for manipulation of the key is also returned in
 *     phKey.
 *
 *     pName will match the key name given by CSM_KMInsertKey.
 *
 *     Setting pResultKey to NULL will cause only the key handle to be returned.
 *     pAccessCred may be NULL in this situation.
 *
 *     When a private or symmetric key value is returned, it is in an unwrapped
 *     state. pAccessCred must contain the appropriate credentials to authorize
 *     unwrapping the key. The key manager will make any callbacks necessary to
 *     acquire authentication samples.
 *
 *     Key values can be returned in any format supported by CSM_CreateKeyBlob.
 *     pFormatContext is used to indicate the desired format.
 *     pFormatContext->Format must be one of the raw key blob format constants
 *     (CSSM_KEYBLOB_RAW_*). If the format requested is CSSM_KEYBLOB_RAW_SPKI,
 *     the pFormatContext->SigAlg field is used to generate the signature
 *     algorithm tag in the s-expression.  
 *
 *     After successful completion of the function, phKey will contain a unique
 *     reference that can be used to access the key with CSM_KMDeleteKey,
 *     CSM_KMGetKeyAcl, and CSM_KMFindKeyByHandle. The handle becomes invalid
 *     when either CSM_KMDeleteKey or CSM_KMCloseKeyFile is called with the same
 *     hKeyFile value that was used to perform the insertion.
 *
 * Parameters:
 *     CSM_KM_HANDLE                          hKeyFile
 *     const CSSM_DATA *                      pName
 *     const CSM_KM_FORMAT_CONTEXT *          pFormatContext
 *     const CSSM_ACCESS_CREDENTIALS *        pAccessCred
 *     CSSM_KEY *                             pResultKey
 *     CSM_KM_KEY_HANDLE *                    phKey
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     KMERR_NOT_INITIALIZED                 - The key manager wasn't initialized.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 *     KMERR_FILE_DOES_NOT_EXIST             - The file referenced by hKeyFile
 *                                            was not found. Most likely it
 *                                            was deleted by another process.
 *                                            Use CSM_KMCreateKeyFile to
 *                                            create the file.
 *     CSSMERR_CSP_INTERNAL_ERROR              - Something failed that shouldn't.
 *     CSSMERR_CSP_PRIVATE_KEY_NOT_FOUND            - The key couldn't be found.
 *     CSSM_CSP_FILE_READ_FAILED            - The file could not be read.
 *     CSSM_CSP_FILE_FORMAT_ERROR           - The file isn't a valid database.
 *     CSSMERR_CSP_MEMORY_ERROR                - Couldn't allocate memory.
 *     KMERR_HANDLES_EXHAUSTED               - No more key handles for hKeyFile.
 */
CSSM_RETURN CSSMAPI CSM_KMFindKeyByName(
                        CSM_KM_HANDLE hKeyFile,
                        const CSSM_DATA *pName, 
                        const CSM_KM_FORMAT_CONTEXT *pFormatContext,
                        const CSSM_ACCESS_CREDENTIALS *pAccessCred,
                        CSSM_KEY *pResultKey,
                        CSM_KM_KEY_HANDLE *phKey )
{
    CSSM_RETURN ret = CSSM_OK;
    const MAF_MODULE_ATTACH_TRACKER * pAttachTracker = NULL;
    MLC_LOCK_REF mlcFileHandlesLock = NULL;
    CSSM_BOOL mlcFileHandlesLockIsLocked = CSSM_FALSE;
    CSM_KM_KEY_FILE_HANDLES_NODE * pKeyFileHandlesNode = NULL;
    MLC_LOCK_REF mlcKeysLock = NULL;
    CSSM_BOOL mlcKeysLockIsLocked = CSSM_FALSE;
    CSM_KM_KEYS_NODE * pKeysNode = NULL;
    CSM_KM_DB_RECORD * pRecord = NULL;
    CSSM_STRING keyName;
    CSM_KM_DB_QUERY_ITEM query;
    CSM_KM_DB_QUERY_ITEM requests[2];
    CSSM_BOOL found = CSSM_FALSE;
    CSM_KM_KEYS_NODE * pNewKeysNode = NULL;
    CSM_KM_HANDLE newHandle = 0;
    long index = 0;
    CSSM_KEY wrappedKey;
    CSSM_KEY unWrappedKey;

    pAttachTracker = Addin_GetAttachTracker();

    if (csm_kmIsInitialized != CSSM_TRUE)
    {
        ret = KMERR_NOT_INITIALIZED;
    }
    else if ( !hKeyFile )
    {
        ret = KMERR_INVALID_HANDLE;
    }
    else if ( !pName ||
              !(pName->Length) ||
              ( pName->Length > CSSM_MODULE_STRING_SIZE ) ||
              !(pName->Data) )
    {
        ret = CSSMERR_CSP_INVALID_KEY_LABEL;
    }
    /* When access credentials exist: (pResultKey && !pAccessCred) || */
    else if ( !pResultKey )
    {
        ret = CSSMERR_CSP_INVALID_KEY_POINTER;
    }
    else if ( !phKey )
    {
        ret = KMERR_INVALID_HANDLE_POINTER;
    }
    else if (!pAttachTracker)
    {
        ret = CSSMERR_CSP_INTERNAL_ERROR;
    }
    else
    {
        /* Find the file handle node. */

        if ((ret = MLC_FindItem(&csm_kmKeyFileHandles,
                                csm_kmFindKeyFileHandle,
                                (void *)&hKeyFile,
                                MLC_WRITE_LOCK,
                                &mlcFileHandlesLock,
                                (void **)&pKeyFileHandlesNode)) != CSSM_OK)
        {
            ret = (ret == MLC_ERR_NOT_FOUND) ? KMERR_FILE_DOES_NOT_EXIST :
                                               KMERR_FILE_READ_FAILED;
        }
        else if (!pKeyFileHandlesNode)
        {
            mlcFileHandlesLockIsLocked = CSSM_TRUE;

            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            mlcFileHandlesLockIsLocked = CSSM_TRUE;
        }
    }

    if (ret == CSSM_OK)
    {
        /* Find the key node. */

        if ((ret = MLC_FindItem(&(pKeyFileHandlesNode->keys),
                                csm_kmFindKeyByName,
                                (void *)pName,
                                MLC_READ_LOCK,
                                &mlcKeysLock,
                                (void **)&pKeysNode)) != CSSM_OK)
        {
            ret = (ret == MLC_ERR_NOT_FOUND) ? CSSMERR_CSP_INVALID_KEY_REFERENCE :
                                               CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            mlcKeysLockIsLocked = CSSM_TRUE;

            if (phKey)
            {
                *phKey = pKeysNode->hKey;
            }

            if (!pKeysNode || !(pKeysNode->pRecord))
            {
                ret = CSSMERR_CSP_INTERNAL_ERROR;
            }
        }
    }

    if ((ret == CSSMERR_CSP_PRIVATE_KEY_NOT_FOUND) &&
        (pKeyFileHandlesNode->fd != -1))
    {
        /* The key wasn't in memory.  Check the file. */

        /* Allocate a record. */
        pRecord = (CSM_KM_DB_RECORD *)
                  Addin_calloc(sizeof(CSM_KM_DB_RECORD), 1, NULL);
        if (!pRecord)
        {
            ret = CSSMERR_CSP_MEMORY_ERROR;
        }
        else
        {
            strncpy((char *)keyName, (const char *)pName->Data,
                    CSSM_MODULE_STRING_SIZE);

            query.type = CSM_KM_DB_QUERY_KEY_NAME;
            query.value = (void *)&keyName;

            requests[0].type = CSM_KM_DB_QUERY_RECORD;
            requests[0].value = (void *)pRecord;
            requests[1].type = CSM_KM_DB_QUERY_INDEX;
            requests[1].value = (void *)&index;

            if ((ret = csm_kmQueryDb( pKeyFileHandlesNode->fd,
                                      1,
                                      &query,
                                      2,
                                      requests,
                                      0, /* No lock required */
                                      NULL,
                                      &found ) ) == CSSM_OK)
            {
                if (found == CSSM_FALSE)
                {
                    ret = CSSMERR_CSP_PRIVATE_KEY_NOT_FOUND;
                }
            }
        }

        if (ret == CSSM_OK)
        {
            /* The record was recovered.  Create the node. */
            pNewKeysNode = (CSM_KM_KEYS_NODE *)
                           Addin_calloc(sizeof(CSM_KM_KEYS_NODE), 1, NULL);
            if (!pNewKeysNode)
            {
                ret = CSSMERR_CSP_MEMORY_ERROR;
            }
            else
            {
                if ((ret = csm_kmGetHandle(
                               &(pKeyFileHandlesNode->discardedKeyHandles),
                               &newHandle)) == CSSM_OK)
                {
                    if (newHandle)
                    {
                        pNewKeysNode->hKey = newHandle;
                    }
                    else if (!(pKeyFileHandlesNode->keysNextHandle))
                    {
                        ret = KMERR_HANDLES_EXHAUSTED;
                    }
                    else
                    {
                        pNewKeysNode->hKey =
                            pKeyFileHandlesNode->keysNextHandle++;
                        newHandle = pNewKeysNode->hKey;
                    }
                }

                if (ret == CSSM_OK)
                {
                    pNewKeysNode->dbIndex = index;
                    pNewKeysNode->pRecord = pRecord;

                    if (phKey)
                    {
                        *phKey = pNewKeysNode->hKey;
                    }
                }
            }
        }

        if (ret == CSSM_OK)
        {
            /* The key node has been created, add it to the list. */

            if (MLC_AddItem(&(pKeyFileHandlesNode->keys),
                            (void *)pNewKeysNode,
                            MLC_READ_LOCK,
                            &mlcKeysLock) != CSSM_OK)
            {
                /* Reclaim the key handle. */
                csm_kmInsertHandle(&(pKeyFileHandlesNode->discardedKeyHandles),
                                   pNewKeysNode->hKey);

                ret = (ret == CSSM_ERRCODE_MEMORY_ERROR) ? CSSMERR_CSP_MEMORY_ERROR :
                                                   CSSMERR_CSP_INTERNAL_ERROR;
            }
            else
            {
                mlcKeysLockIsLocked = CSSM_TRUE;

                /* So we can get some info from the node. */
                pKeysNode = pNewKeysNode;

                /* So they won't get freed. */
                pRecord = NULL;
                pNewKeysNode = NULL;
            }
        }
    }

    if ((ret == CSSM_OK) && (!pKeysNode || !(pKeysNode->pRecord)))
    {
        /*
         * pKeysNode should point to either a node already in memory, or
         * a new node created by loading a record from the file.
         */

        ret = CSSMERR_CSP_INTERNAL_ERROR;
    }

    if ((ret == CSSM_OK) && pResultKey)
    {
        /*
         * The key has been retrieved.  Copy the headerFormat the key, and set
         * the file handle.
         */

        memset((void *)&unWrappedKey, 0, sizeof(CSSM_KEY));

        if ( ( pKeysNode->pRecord->acl.aclType != 0 ) &&
             ( pAccessCred == NULL ) )
        {
            ERR( ret = CSSMERR_CSP_INVALID_ACCESS_CREDENTIALS );
        }
        else if ( pKeysNode->pRecord->acl.aclType != 0 )
        {
            CSSM_DATA PbeSalt;
            CSSM_DATA PbeIV;

            memcpy((void *)&(wrappedKey.KeyHeader),
                   (const void *)&(pKeysNode->pRecord->keyHeader),
                   sizeof(CSSM_KEYHEADER));

            wrappedKey.KeyData.Data = pKeysNode->pRecord->key;
            wrappedKey.KeyData.Length = pKeysNode->pRecord->keyLength;

            /* Attempt to unwrap the key. */
            /* Attempt to unwrap the key. */
            PbeSalt.Data = pKeysNode->pRecord->WrapPBESalt;
            PbeSalt.Length = sizeof(pKeysNode->pRecord->WrapPBESalt);
            PbeIV.Data = pKeysNode->pRecord->WrapIV;
            PbeIV.Length = pKeysNode->pRecord->WrapIVLength;
            ret = csm_kmUnwrapKey( MAF_ALLOCATE_LOCAL,
                                   &wrappedKey,
                                   &unWrappedKey,
                                   pAccessCred,
                                   CSSM_SAMPLE_TYPE_PASSWORD,
                                   NULL,
                                   &PbeSalt,
                                   &PbeIV );
            if ((ret != CSSM_OK) && unWrappedKey.KeyData.Data)
            {
                Addin_free((void *)unWrappedKey.KeyData.Data, NULL);
                unWrappedKey.KeyData.Data = NULL;
                unWrappedKey.KeyData.Length = 0;
            }
        }
        else
        {
            memcpy((void *)&(unWrappedKey.KeyHeader),
                   (const void *)&(pKeysNode->pRecord->keyHeader),
                   sizeof(CSSM_KEYHEADER));

            unWrappedKey.KeyData.Data = (uint8 *)
                Addin_calloc(1, pKeysNode->pRecord->keyLength, NULL);
            if (!(unWrappedKey.KeyData.Data))
            {
                ret = CSSMERR_CSP_MEMORY_ERROR;
            }

            unWrappedKey.KeyData.Length = pKeysNode->pRecord->keyLength;
        }

        /* For now, just copy the key. */

        if (ret == CSSM_OK)
        {
            /* Reformat the key to match the requested format */
            if ((ret = CSM_KMReformatKey( MAF_ALLOCATE_LOCAL,
                                          &unWrappedKey,
                                          pFormatContext )) == CSSM_OK )
            {
                memcpy( &pResultKey->KeyHeader,
                        &unWrappedKey.KeyHeader,
                        sizeof(unWrappedKey.KeyHeader) );
                pResultKey->KeyData = unWrappedKey.KeyData;
            }
        }

        if ( ret != CSSM_OK )
        {
            if ( unWrappedKey.KeyData.Data != NULL )
            {
                Addin_free((void *)unWrappedKey.KeyData.Data, NULL);
                unWrappedKey.KeyData.Data = NULL;
                unWrappedKey.KeyData.Length = 0;
            }
        }
    }

    if ((ret != CSSM_OK) && phKey)
    {
        *phKey = 0;
    }

    if (pRecord)
    {
        Addin_free((void *)pRecord, NULL);
    }

    if (pNewKeysNode)
    {
        Addin_free((void *)pNewKeysNode, NULL);
    }

    if (mlcKeysLockIsLocked == CSSM_TRUE)
    {
        if (MLC_ReleaseItem(MLC_READ_LOCK, mlcKeysLock) != CSSM_OK)
        {
            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
    }

    if (mlcFileHandlesLockIsLocked == CSSM_TRUE)
    {
        if (MLC_ReleaseItem(MLC_WRITE_LOCK, mlcFileHandlesLock) != CSSM_OK)
        {
            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
    }

    return ret;
}

/*
 * Name: CSM_KMFindKeyByRefKey
 *
 * Description:
 *     Search the key database for existence of a key by a reference key. If
 *     the requested key exists, it is returned in pResultKey with the format
 *     specified in pFormatContext. A handle for manipulation of the key is
 *     also returned in phKey.
 *
 *     pRefKey must be a public key in raw format. 
 *
 *     Setting pResultKey to NULL will cause only the key handle to be returned.
 *     pAccessCred may be NULL in this situation.
 *
 *     When a private or symmetric key value is returned, it is in an unwrapped
 *     state. pAccessCred must contain the appropriate credentials to authorize
 *     unwrapping the key. The key manager will make any callbacks necessary to
 *     acquire authentication samples.
 *
 *     Key values can be returned in any format supported by CSM_CreateKeyBlob.
 *     pFormatContext is used to indicate the desired format.
 *     pFormatContext->Format must be one of the raw key blob format constants
 *     (CSSM_KEYBLOB_RAW_*). If the format requested is CSSM_KEYBLOB_RAW_SPKI,
 *     the pFormatContext->SigAlg field is used to generate the signature
 *     algorithm tag in the s-expression.  
 *
 *     After successful completion of the function, phKey will contain a unique
 *     reference that can be used to access the key with CSM_KMDeleteKey,
 *     CSM_KMGetKeyAcl, and CSM_KMFindKeyByHandle. The handle becomes invalid
 *     when either CSM_KMDeleteKey or CSM_KMCloseKeyFile is called with the same
 *     hKeyFile value that was used to perform the insertion.
 *
 * Parameters:
 *     CSM_KM_HANDLE                          hKeyFile
 *     const CSSM_KEY *                       pRefKey
 *     const CSM_KM_FORMAT_CONTEXT *          pFormatContext
 *     const CSSM_ACCESS_CREDENTIALS *        pAccessCred
 *     CSSM_KEY *                             pResultKey
 *     CSM_KM_KEY_HANDLE *                    phKey
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     KMERR_NOT_INITIALIZED                 - The key manager wasn't initialized.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 *     CSSM_CSP_OPERATION_UNSUPPORTED       - Could not hash this key.
 *     KMERR_FILE_DOES_NOT_EXIST             - The file referenced by hKeyFile
 *                                            was not found. Most likely it
 *                                            was deleted by another process.
 *                                            Use CSM_KMCreateKeyFile to
 *                                            create the file.
 *     CSSMERR_CSP_INTERNAL_ERROR              - Something failed that shouldn't.
 *     CSSMERR_CSP_PRIVATE_KEY_NOT_FOUND            - The key couldn't be found.
 *     CSSM_CSP_FILE_READ_FAILED            - The file could not be read.
 *     CSSM_CSP_FILE_FORMAT_ERROR           - The file isn't a valid database.
 *     CSSMERR_CSP_MEMORY_ERROR                - Couldn't allocate memory.
 *     KMERR_HANDLES_EXHAUSTED               - No more key handles for hKeyFile.
 */
CSSM_RETURN CSSMAPI CSM_KMFindKeyByRefKey(
                        CSM_KM_HANDLE hKeyFile,
                        const CSSM_KEY *pRefKey,
                        const CSM_KM_FORMAT_CONTEXT *pFormatContext,
                        const CSSM_ACCESS_CREDENTIALS *pAccessCred,
                        CSSM_KEY *pResultKey,
                        CSM_KM_KEY_HANDLE *phKey )
{
    CSSM_RETURN ret = CSSM_OK;
    CSSM_DATA hash;
    uint8 hashValue[CSM_KM_DB_HASH_SIZE];

    if (csm_kmIsInitialized != CSSM_TRUE)
    {
        ret = KMERR_NOT_INITIALIZED;
    }
    else if ( !pRefKey )
    {
        ret = CSSMERR_CSP_INVALID_KEY_POINTER;
    }
    else
    {
        if ((ret = csm_kmHashKey(pRefKey, hashValue)) == CSSM_OK)
        {
            hash.Length = CSM_KM_DB_HASH_SIZE;
            hash.Data = hashValue;

            ret = CSM_KMFindKeyByHash(hKeyFile, (const CSSM_DATA *)&hash,
                                      pFormatContext, pAccessCred,
                                      pResultKey, phKey);
        }
    }

    return ret;
}

/*
 * Name: CSM_KMFindKeyByHandle
 *
 * Description:
 *     Search the key database for existence of a key by a key handle. If
 *     the requested key exists, it is returned in pResultKey with the format
 *     specified in pFormatContext.
 *
 *     Setting pResultKey to NULL will cause only the key handle to be returned.
 *     pAccessCred may be NULL in this situation.
 *
 *     When a private or symmetric key value is returned, it is in an unwrapped
 *     state. pAccessCred must contain the appropriate credentials to authorize
 *     unwrapping the key. The key manager will make any callbacks necessary to
 *     acquire authentication samples.
 *
 *     Key values can be returned in any format supported by CSM_CreateKeyBlob.
 *     pFormatContext is used to indicate the desired format.
 *     pFormatContext->Format must be one of the raw key blob format constants
 *     (CSSM_KEYBLOB_RAW_*). If the format requested is CSSM_KEYBLOB_RAW_SPKI,
 *     the pFormatContext->SigAlg field is used to generate the signature
 *     algorithm tag in the s-expression.  
 *
 * Parameters:
 *     CSM_KM_HANDLE                          hKeyFile
 *     CSM_KM_KEY_HANDLE                      hKey
 *     const CSM_KM_FORMAT_CONTEXT *          pFormatContext
 *     const CSSM_ACCESS_CREDENTIALS *        pAccessCred
 *     CSSM_KEY *                             pResultKey
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     KMERR_NOT_INITIALIZED                 - The key manager wasn't initialized.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 *     KMERR_FILE_DOES_NOT_EXIST             - The file referenced by hKeyFile
 *                                            was not found. Most likely it
 *                                            was deleted by another process.
 *                                            Use CSM_KMCreateKeyFile to
 *                                            create the file.
 *     CSSMERR_CSP_INTERNAL_ERROR              - Something failed that shouldn't.
 *     CSSMERR_CSP_PRIVATE_KEY_NOT_FOUND            - The key couldn't be found.
 */
CSSM_RETURN CSSMAPI CSM_KMFindKeyByHandle(
                        CSM_KM_HANDLE hKeyFile,
                        CSM_KM_KEY_HANDLE hKey,
                        const CSM_KM_FORMAT_CONTEXT *pFormatContext,
                        const CSSM_ACCESS_CREDENTIALS *pAccessCred,
                        CSSM_KEY *pResultKey )
{
    CSSM_RETURN ret = CSSM_OK;
    const MAF_MODULE_ATTACH_TRACKER * pAttachTracker = NULL;
    MLC_LOCK_REF mlcFileHandlesLock = NULL;
    CSSM_BOOL mlcFileHandlesLockIsLocked = CSSM_FALSE;
    CSM_KM_KEY_FILE_HANDLES_NODE * pKeyFileHandlesNode = NULL;
    MLC_LOCK_REF mlcKeysLock = NULL;
    CSSM_BOOL mlcKeysLockIsLocked = CSSM_FALSE;
    CSM_KM_KEYS_NODE * pKeysNode = NULL;
    CSSM_KEY wrappedKey;
    CSSM_KEY unWrappedKey;

    pAttachTracker = Addin_GetAttachTracker();

    if (csm_kmIsInitialized != CSSM_TRUE)
    {
        ret = KMERR_NOT_INITIALIZED;
    }
    else if ( !hKeyFile || !hKey )
    {
        ret = KMERR_INVALID_HANDLE;
    }
    /* When access credentials exist: (pResultKey && !pAccessCred) || */
    else if ( !pResultKey )
    {
        ret = CSSMERR_CSP_INVALID_KEY_POINTER;
    }
    else if (!pAttachTracker)
    {
        ret = CSSMERR_CSP_INTERNAL_ERROR;
    }
    else
    {
        /* Find the file handle node. */

        if ((ret = MLC_FindItem(&csm_kmKeyFileHandles,
                                csm_kmFindKeyFileHandle,
                                (void *)&hKeyFile,
                                MLC_READ_LOCK,
                                &mlcFileHandlesLock,
                                (void **)&pKeyFileHandlesNode)) != CSSM_OK)
        {
            ret = (ret == MLC_ERR_NOT_FOUND) ? KMERR_FILE_DOES_NOT_EXIST :
                                               CSSMERR_CSP_INTERNAL_ERROR;
        }
        else if (!pKeyFileHandlesNode)
        {
            mlcFileHandlesLockIsLocked = CSSM_TRUE;

            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            mlcFileHandlesLockIsLocked = CSSM_TRUE;
        }
    }

    if (ret == CSSM_OK)
    {
        /* Find the key node. */

        if ((ret = MLC_FindItem(&(pKeyFileHandlesNode->keys),
                                csm_kmFindKeyByHandle,
                                (void *)&hKey,
                                MLC_READ_LOCK,
                                &mlcKeysLock,
                                (void **)&pKeysNode)) != CSSM_OK)
        {
            ret = (ret == MLC_ERR_NOT_FOUND) ? CSSMERR_CSP_INVALID_KEY_REFERENCE :
                                               CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            mlcKeysLockIsLocked = CSSM_TRUE;

            if (!pKeysNode || !(pKeysNode->pRecord))
            {
                ret = CSSMERR_CSP_INTERNAL_ERROR;
            }
        }
    }

    if (ret == CSSM_OK)
    {
        /*
         * The key has been retrieved.  Copy the headerFormat the key, and set
         * the file handle.
         */

        memset((void *)&unWrappedKey, 0, sizeof(CSSM_KEY));

        if ( ( pKeysNode->pRecord->acl.aclType != 0 ) &&
             ( pAccessCred == NULL ) )
        {
            ERR( ret = CSSMERR_CSP_INVALID_ACCESS_CREDENTIALS );
        }
        else if ( pKeysNode->pRecord->acl.aclType != 0 )
        {
            CSSM_DATA PbeSalt;
            CSSM_DATA PbeIV;

            memcpy((void *)&(wrappedKey.KeyHeader),
                   (const void *)&(pKeysNode->pRecord->keyHeader),
                   sizeof(CSSM_KEYHEADER));

            wrappedKey.KeyData.Data = pKeysNode->pRecord->key;
            wrappedKey.KeyData.Length = pKeysNode->pRecord->keyLength;

            /* Attempt to unwrap the key. */
            PbeSalt.Data = pKeysNode->pRecord->WrapPBESalt;
            PbeSalt.Length = sizeof(pKeysNode->pRecord->WrapPBESalt);
            PbeIV.Data = pKeysNode->pRecord->WrapIV;
            PbeIV.Length = pKeysNode->pRecord->WrapIVLength;
            ret = csm_kmUnwrapKey( MAF_ALLOCATE_LOCAL,
                                   &wrappedKey,
                                   &unWrappedKey,
                                   pAccessCred,
                                   CSSM_SAMPLE_TYPE_PASSWORD,
                                   NULL,
                                   &PbeSalt,
                                   &PbeIV );
            if ((ret != CSSM_OK) && unWrappedKey.KeyData.Data)
            {
                Addin_free((void *)unWrappedKey.KeyData.Data, NULL);
                unWrappedKey.KeyData.Data = NULL;
                unWrappedKey.KeyData.Length = 0;
            }
        }
        else
        {
            memcpy((void *)&(unWrappedKey.KeyHeader),
                   (const void *)&(pKeysNode->pRecord->keyHeader),
                   sizeof(CSSM_KEYHEADER));

            unWrappedKey.KeyData.Data = (uint8 *)
                Addin_calloc(1, pKeysNode->pRecord->keyLength, NULL);
            if (!(unWrappedKey.KeyData.Data))
            {
                ret = CSSMERR_CSP_MEMORY_ERROR;
            }
            else
            {
                memcpy( unWrappedKey.KeyData.Data,
                        pKeysNode->pRecord->key,
                        pKeysNode->pRecord->keyLength );
                unWrappedKey.KeyData.Length = pKeysNode->pRecord->keyLength;
            }
        }

        if (ret == CSSM_OK)
        {
            /* Reformat the key to match the requested format */
            if ((ret = CSM_KMReformatKey( MAF_ALLOCATE_LOCAL,
                                          &unWrappedKey,
                                          pFormatContext )) == CSSM_OK )
            {
                memcpy( &pResultKey->KeyHeader,
                        &unWrappedKey.KeyHeader,
                        sizeof(unWrappedKey.KeyHeader) );
                pResultKey->KeyData = unWrappedKey.KeyData;
            }
        }

        if ( ret != CSSM_OK )
        {
            if ( unWrappedKey.KeyData.Data != NULL )
            {
                Addin_free((void *)unWrappedKey.KeyData.Data, NULL);
                unWrappedKey.KeyData.Data = NULL;
                unWrappedKey.KeyData.Length = 0;
            }
        }
    }

    if (mlcKeysLockIsLocked == CSSM_TRUE)
    {
        if (MLC_ReleaseItem(MLC_READ_LOCK, mlcKeysLock) != CSSM_OK)
        {
            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
    }

    if (mlcFileHandlesLockIsLocked == CSSM_TRUE)
    {
        if (MLC_ReleaseItem(MLC_READ_LOCK, mlcFileHandlesLock) != CSSM_OK)
        {
            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
    }

    return ret;
}

CSSM_RETURN CSSMAPI CSM_KMGetKeyHeader(
                        CSM_KM_HANDLE hKeyFile,
                        CSM_KM_KEY_HANDLE hKey,
                        CSSM_KEYHEADER *pKeyHeader )
{
    CSSM_RETURN ret = CSSM_OK;
    const MAF_MODULE_ATTACH_TRACKER * pAttachTracker = NULL;
    MLC_LOCK_REF mlcFileHandlesLock = NULL;
    CSSM_BOOL mlcFileHandlesLockIsLocked = CSSM_FALSE;
    CSM_KM_KEY_FILE_HANDLES_NODE * pKeyFileHandlesNode = NULL;
    MLC_LOCK_REF mlcKeysLock = NULL;
    CSSM_BOOL mlcKeysLockIsLocked = CSSM_FALSE;
    CSM_KM_KEYS_NODE * pKeysNode = NULL;

    pAttachTracker = Addin_GetAttachTracker();

    if (csm_kmIsInitialized != CSSM_TRUE)
    {
        ret = KMERR_NOT_INITIALIZED;
    }
    else if ( !hKeyFile || !hKey )
    {
        ret = KMERR_INVALID_HANDLE;
    }
    else if ( !pKeyHeader )
    {
        ret = CSSMERR_CSP_INVALID_KEY_POINTER;
    }
    else if (!pAttachTracker)
    {
        ret = CSSMERR_CSP_INTERNAL_ERROR;
    }
    else
    {
        /* Find the file handle node. */

        if ((ret = MLC_FindItem(&csm_kmKeyFileHandles,
                                csm_kmFindKeyFileHandle,
                                (void *)&hKeyFile,
                                MLC_READ_LOCK,
                                &mlcFileHandlesLock,
                                (void **)&pKeyFileHandlesNode)) != CSSM_OK)
        {
            ret = (ret == MLC_ERR_NOT_FOUND) ? KMERR_FILE_DOES_NOT_EXIST :
                                               CSSMERR_CSP_INTERNAL_ERROR;
        }
        else if (!pKeyFileHandlesNode)
        {
            mlcFileHandlesLockIsLocked = CSSM_TRUE;

            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            mlcFileHandlesLockIsLocked = CSSM_TRUE;
        }
    }

    if (ret == CSSM_OK)
    {
        /* Find the key node. */

        if ((ret = MLC_FindItem(&(pKeyFileHandlesNode->keys),
                                csm_kmFindKeyByHandle,
                                (void *)&hKey,
                                MLC_READ_LOCK,
                                &mlcKeysLock,
                                (void **)&pKeysNode)) != CSSM_OK)
        {
            ret = (ret == MLC_ERR_NOT_FOUND) ? CSSMERR_CSP_INVALID_KEY_REFERENCE :
                                               CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            mlcKeysLockIsLocked = CSSM_TRUE;

            if (!pKeysNode || !(pKeysNode->pRecord))
            {
                ret = CSSMERR_CSP_INTERNAL_ERROR;
            }
        }
    }

    if (ret == CSSM_OK)
    {
        /*
         * The key has been retrieved.  Copy the header.
         */
        memcpy( pKeyHeader,
                &pKeysNode->pRecord->keyHeader,
                sizeof(pKeysNode->pRecord->keyHeader) );
    }

    if (mlcKeysLockIsLocked == CSSM_TRUE)
    {
        if (MLC_ReleaseItem(MLC_READ_LOCK, mlcKeysLock) != CSSM_OK)
        {
            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
    }

    if (mlcFileHandlesLockIsLocked == CSSM_TRUE)
    {
        if (MLC_ReleaseItem(MLC_READ_LOCK, mlcFileHandlesLock) != CSSM_OK)
        {
            ret = CSSMERR_CSP_INTERNAL_ERROR;
        }
    }

    return ret;
}

CSSM_RETURN CSSMAPI CSM_KMGetKeyAcl(
                        CSSM_CSP_HANDLE CSPHandle,
                        CSM_KM_HANDLE hKeyFile,
                        CSM_KM_KEY_HANDLE hKey,
                        const CSSM_STRING *pSelectionTag,
                        CSSM_ACL_ENTRY_INFO_PTR *pAclInfos,
                        uint32 *pNumAclInfos )
{
    CSSM_RETURN ret = CSSM_OK;
    MLC_LOCK_REF mlcFileHandlesLock = NULL;
    CSM_KM_KEY_FILE_HANDLES_NODE *pKeyFileHandlesNode = NULL;
    MLC_LOCK_REF mlcKeysLock = NULL;
    CSM_KM_KEYS_NODE *pKeysNode = NULL;
    CSM_KM_DB_RECORD *pRecord = NULL;
    uint32 NumAclInfos = 0;
    CSSM_ACL_ENTRY_INFO *AclInfos = NULL;
    CSSM_DATA Prompt = { 0, NULL };

    if ( csm_kmIsInitialized != CSSM_TRUE )
    {
        ret = KMERR_NOT_INITIALIZED;
    }
    else if ( !hKeyFile || !hKey )
    {
        ret = KMERR_INVALID_HANDLE;
    }
    else if ( !pNumAclInfos || !pAclInfos )
    {
        ERR( ret = CSSMERR_CSP_INVALID_OUTPUT_POINTER );
    }
    else if ( pSelectionTag )
    {
        /* Selection tags are not supported */
        ERR( ret = CSSMERR_CSP_INVALID_ACL_ENTRY_TAG );
    }
    else
    {
        /* Find the file handle node. */
        if ( (ret = MLC_FindItem(
                            &csm_kmKeyFileHandles,
                            csm_kmFindKeyFileHandle,
                            (void *)&hKeyFile,
                            MLC_READ_LOCK,
                            &mlcFileHandlesLock,
                            (void **)&pKeyFileHandlesNode ) ) != CSSM_OK )
        {
            ret = ( ret == MLC_ERR_NOT_FOUND ) ? KMERR_FILE_DOES_NOT_EXIST :
                                                 CSSMERR_CSP_INTERNAL_ERROR;
        }
        else
        {
            /* Find the key node. */
            if ( (ret = MLC_FindItem( &pKeyFileHandlesNode->keys,
                                      csm_kmFindKeyByHandle,
                                      (void *)&hKey,
                                      MLC_READ_LOCK,
                                      &mlcKeysLock,
                                      (void **)&pKeysNode) ) != CSSM_OK )
            {
                ret = (ret == MLC_ERR_NOT_FOUND) ? CSSMERR_CSP_INVALID_KEY_REFERENCE :
                                                   CSSMERR_CSP_INTERNAL_ERROR;
            }
            else
            {
                assert( pKeysNode->pRecord );
                pRecord = pKeysNode->pRecord;

                /* Construct the ACL info structure */
                if ( pRecord->acl.aclType == 0 )
                { /* Unrestricted key */
                    *pNumAclInfos = 0;
                    *pAclInfos = NULL;
                } /* Unrestricted key */
                else if ( ( pRecord->acl.aclType == 
                                CSSM_ACL_SUBJECT_TYPE_PASSWORD ) ||
                          ( pRecord->acl.aclType == 
                                CSSM_ACL_SUBJECT_TYPE_PROMPTED_PASSWORD ) )
                { /* Password variant */
                    /* There is always just one ACL info for passwords */
                    NumAclInfos = 1;

                    /* Allocate and fill the info structure */
                    AclInfos = MAF_CALLOC( CSPHandle, sizeof(*AclInfos), 1 );
                    if ( AclInfos )
                    {
                        AclInfos->EntryHandle = hKey; /* Make handle matching easy */
                        AclInfos->EntryPublicInfo.Delegate = CSSM_FALSE;
                        /* Convert key permissions to auth list? */
                        /* How do you represent open-ended start and end dates? */

                        /* Build the subject list. Luckily the same function
                         * used to build password challenges will work here
                         * too.
                         */
                        if ( pRecord->acl.aclType ==
                                CSSM_ACL_SUBJECT_TYPE_PROMPTED_PASSWORD )
                        {
                            Prompt.Data = (uint8*)pRecord->acl.prompt;
                            Prompt.Length = strlen(pRecord->acl.prompt);
                        }
                        ret = MAF_BuildPasswordChallenge(
                                    CSPHandle,
                                    &Prompt,
                                    &AclInfos->EntryPublicInfo.TypedSubject );
                        
                        /* Clean up if building the list failed */
                        if ( ret != CSSM_OK )
                        {
                            MAF_FREE( CSPHandle, AclInfos );
                        }
                        else
                        {
                            /* Assign the output values */
                            *pNumAclInfos = NumAclInfos;
                            *pAclInfos = AclInfos;
                        }
                    }
                    else
                    {
                        ERR( ret = CSSMERR_CSP_MEMORY_ERROR );
                    }
                } /* Password variant */
                else
                { /* Unknown/unsupported */
                    ERR( ret = CSSMERR_CSP_ACL_SUBJECT_TYPE_NOT_SUPPORTED );
                } /* Unknown/unsupported */

                /* Unlock the key node */
                MLC_ReleaseItem( MLC_READ_LOCK, mlcKeysLock );
            } /* Find key */

            /* Unlock the file handle node */
            MLC_ReleaseItem( MLC_READ_LOCK, mlcFileHandlesLock );
        } /* Find key file */
    } /* Param check */

    return ret;
}

CSSM_RETURN CSSMAPI CSM_KMReformatKey(
                        CSSM_CSP_HANDLE CSPHandle,
                        CSSM_KEY *pKey,
                        const CSM_KM_FORMAT_CONTEXT *pFormatContext )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_DATA KeyParts[CSM_MAX_KEY_PARTS];
    uint32 NumKeyParts = CSM_MAX_KEY_PARTS;
    CSSM_DATA TempKeyBlob = { 0, NULL };
    CSSM_BOOL bAllocatedBlob = CSSM_FALSE;

    assert( pKey && pFormatContext );

    if ( pKey->KeyHeader.Format != pFormatContext->Format )
    {
        /* Parse the key into it's component parts */
        if ( ( rv = CSM_ParseKeyblob( pKey->KeyHeader.AlgorithmId,
                                      pKey->KeyHeader.KeyClass,
                                      pKey->KeyHeader.Format,
                                      KeyParts,
                                      &NumKeyParts,
                                      &pKey->KeyData ) ) == CSSM_OK )
        {
            if ( ( rv = CSM_CreateKeyblob( CSPHandle,
                                           pKey->KeyHeader.AlgorithmId,
                                           pKey->KeyHeader.KeyClass,
                                           pFormatContext->Format,
                                           KeyParts,
                                           NumKeyParts,
                                           &TempKeyBlob,
                                           &bAllocatedBlob ) ) == CSSM_OK )
            {
                /* Deallocate the existing key blob and set it to the
                 * reformatted blob.
                 */
                MAF_FREE( CSPHandle, pKey->KeyData.Data );
                pKey->KeyData = TempKeyBlob;

                /* Adjust the key header */
                pKey->KeyHeader.Format = pFormatContext->Format;
            } /* Reconstitute key blob */
        } /* Parse key blob */
    } /* Check for need to convert */

    return rv;
}
