/*-----------------------------------------------------------------------
 * File: PKCS5V2.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

/* Implements the PKCS #5 v2.0 PBKD2 key derivation mechanism. */

#include "maf_include.h"

#include "eicl.h"

#include "hmacsha1.h"
#include "pkcs5v2.h"

static void PKCS5_v2_PBKD2_HMAC_SHA1_F(
                        const CSSM_DATA *Password,
                        const CSSM_DATA *Salt,
                        uint32 Iterations,
                        uint32 Index,
                        uint8 *Result )
{
    uint8 IndexBytes[4];
    uint32 i, j;
    uint8 TempF[HMAC_SHA1_HASH_LENGTH];
    uint8 IterationResult[HMAC_SHA1_HASH_LENGTH];
    CSSM_DATA IndexBuf;
    CSSM_DATA HmacResult;
    HMAC_SHA1_CONTEXT HmacContext;

    /* Convert the index value into bytes */
    for ( i = 0; i < sizeof(IndexBytes); i++ )
    {
        IndexBytes[i] = (uint8)( ( Iterations >> ( 8 * i ) ) & 0xFF );
    }

    /* Perform the first iteration */
    HMAC_SHA1_Begin( &HmacContext, Password );
    HMAC_SHA1_Update( &HmacContext, Salt );

    IndexBuf.Data = IndexBytes;
    IndexBuf.Length = sizeof(IndexBytes);
    HMAC_SHA1_Update( &HmacContext, &IndexBuf );

    HmacResult.Data = IterationResult;
    HmacResult.Length = sizeof(IterationResult);
    HMAC_SHA1_End( &HmacContext, &HmacResult );

    memcpy( TempF, IterationResult, sizeof(IterationResult) );

    /* Perform the other iterations */
    for ( i = 1; i < Iterations; i++ )
    {
        /* HMAC the previous iteration with the password */
        HMAC_SHA1_Begin( &HmacContext, Password );
        HMAC_SHA1_Update( &HmacContext, &HmacResult );
        HMAC_SHA1_End( &HmacContext, &HmacResult );

        /* XOR this iteration with the temporary state */
        for ( j = 0; j < sizeof(TempF); j++ )
        {
            TempF[j] ^= IterationResult[j];
        }
    }

    /* Copy the temporary buffer to the result */
    memcpy( Result, TempF, sizeof(TempF) );
}

CSSM_RETURN CSSMAPI PKCS5_v2_PBKD2_HMAC_SHA1(
                        const CSSM_DATA *Password,
                        const CSSM_DATA *Salt,
                        uint32 DataLength,
                        uint32 Iterations,
                        uint8 *DataBuf )
{
    CSSM_RETURN rv = CSSM_OK;
    uint32 TopIndex;
    uint32 Index;
    uint8 BlockBuffer[HMAC_SHA1_HASH_LENGTH];
    uint8 *DataPtr;
    
    assert( Password &&
            Password->Data &&
            ( Password->Length > 0 ) );
    assert( Salt &&
            Salt->Data &&
            ( Salt->Length > 0 ) );
    assert( DataBuf );

    /* Calculate the number of blocks to calculate */
    TopIndex = ( DataLength / HMAC_SHA1_HASH_LENGTH ) +
                ( ( ( DataLength % HMAC_SHA1_HASH_LENGTH ) > 0 ) ? 1 : 0 );

    /* Calculate TopIndex blocks to produce DataLength bytes */
    DataPtr = DataBuf;
    for ( Index = 1; Index <= TopIndex; Index++ )
    {
        PKCS5_v2_PBKD2_HMAC_SHA1_F( Password,
                                    Salt,
                                    Iterations,
                                    Index,
                                    BlockBuffer );
        memcpy( DataPtr,
                BlockBuffer,
                min( HMAC_SHA1_HASH_LENGTH, DataLength ) );
        DataLength -= HMAC_SHA1_HASH_LENGTH;
        DataPtr += HMAC_SHA1_HASH_LENGTH;
    }

    return rv;
}
