#include "cssm.h"
#include "cssmport.h"
#include "cssmMemTrack.h"
#include "misc_spki.h"

#include "spkiutil.h"
#include "mempool.h"

extern MEMORY_POOL_FUNCS memory_pool_funcs;
extern MEMORY_POOL_PTR poolptr;


/*-----------------------------------------------------------------------------
 * Name: spki_IsEquivalentCertGroup
 *
 * Description:
 * This function compares the fields of the input CSSM_CERTGROUP structs 
 * to determine if they are equivalent
 *
 * Parameters: 
 * CertGrp1 (input) : A CSSM_CERTGROUP data structure
 * CertGrp2 (input) : A CSSM_CERTGROUP data structure
 *
 * Return value:
 * CSSM_TRUE if the input CSSM_CERTGROUP structs are equivalent
 * CSSM_FALSE otherwise
 *---------------------------------------------------------------------------*/
CSSM_BOOL spki_IsEquivalentCertGroup (CSSM_CERTGROUP_PTR CertGrp1, CSSM_CERTGROUP_PTR CertGrp2)
{
    uint32 i=0;
    CSSM_ENCODED_CERT Cert1, Cert2;

    if (!CertGrp1 && !CertGrp2) return CSSM_TRUE;
    if (!CertGrp1 || !CertGrp2) return CSSM_FALSE;

    if (CertGrp1->NumCerts != CertGrp2->NumCerts) return CSSM_FALSE;
    for (i=0; i < CertGrp1->NumCerts; i++)
    {
        Cert1 = CertGrp1->GroupList.EncodedCertList[i];
        Cert2 = CertGrp2->GroupList.EncodedCertList[i];

        if (Cert1.CertType != Cert2.CertType ||
            Cert1.CertEncoding != Cert2.CertEncoding)
        {
            return CSSM_FALSE;
        }
        if (!IsEquivalentCssmData(&Cert1.CertBlob, &Cert2.CertBlob))
            return CSSM_FALSE;
    }

    return CSSM_TRUE;
}


/*-----------------------------------------------------------------------------
 * Name: IsEquivalentCssmData
 *
 * Description:
 * This function compares the fields of the input CSSM_DATA structs 
 * to determine if they are equivalent
 *
 * Parameters: 
 * Data1 (input) : A CSSM_DATA data structure
 * Data2 (input) : A CSSM_DATA data structure
 *
 * Return value:
 * CSSM_TRUE if the input CSSM_DATA structs are equivalent
 * CSSM_FALSE otherwise
 *---------------------------------------------------------------------------*/
CSSM_BOOL IsEquivalentCssmData(CSSM_DATA_PTR Data1, CSSM_DATA_PTR Data2)
{
    if (!Data1 && !Data2)   /* both pointers are NULL*/
    {
        return CSSM_TRUE;
    }

    if (!Data1 || !Data2)   /* one pointer is NULL    */
    {
        return CSSM_FALSE;
    }

    if (port_IsBadReadPtr(Data1, sizeof(CSSM_DATA)) ||
        port_IsBadReadPtr(Data2, sizeof(CSSM_DATA)))  /* A CSSM_DATA pointer is bad*/
    {
        return CSSM_FALSE;
    }

    if ((Data1->Length == Data2->Length) && (Data1->Length == 0))  /* data lengths are zero*/
    {
        return CSSM_TRUE;
    }

    if (port_IsBadReadPtr(Data1->Data, Data1->Length) ||
        port_IsBadReadPtr(Data2->Data, Data2->Length) )  /* A data pointer is bad*/
    {
        return CSSM_FALSE;
    }

    if (Data1->Length == Data2->Length &&
        memcmp(Data1->Data, Data2->Data, Data1->Length) == 0)  /* data is equivalent*/
    {
        return CSSM_TRUE;
    }

    return CSSM_FALSE;
}

CSSM_RETURN CreateSPKITBSCert(CSSM_SPKI_TBSCERT_PTR Parts,
                              CSSM_DATA *TBSCert)
{
    CSSM_RETURN ret = CSSM_OK;
    CSSM_LIST Issuer, Subject, Tag, Validity;
    CSSM_LIST TBSCertTempList;
    CSSM_LIST_ELEMENT_PTR w;

    if (!Parts)
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
    }
    if (!TBSCert)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    init_list(&Issuer);
    init_list(&Subject);
    init_list(&Validity);
    init_list(&Tag);

    /* Create Issuer list*/
    ret = AddKeyToPrincipal((uint8 *)"(6:issuer1:#)",
                            &Parts->IssuerPubKey,
							&Issuer);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* Create Subject List */
    ret = AddKeyToPrincipal((uint8 *)"(7:subject1:#)",
                            &Parts->SubjectPubKey,
							&Subject);
    if (CSSM_OK != ret)
    {
        drop_list(&Issuer);
        return ret;
    }


    /* Create Tag List */
    init_list(&Tag);
	ret = cl_lit_sexp(&Tag,
					  Parts->Tag.Data);

    if (CSSM_OK != ret)
    {
        drop_list(&Issuer);
        drop_list(&Subject);
        return ret;
    }

    if (Parts->Validity.Data)
    {
        /* Create Validity List */
        init_list(&Validity);
	    ret = cl_lit_sexp(&Validity,
					      Parts->Validity.Data);

        if (CSSM_OK != ret)
        {
            drop_list(&Issuer);
            drop_list(&Subject);
            drop_list(&Tag);
            return ret;
        }
    }
    init_list(&TBSCertTempList);

	w = new_element();
	if (!w)
    {
        drop_list(&Issuer);
        drop_list(&Subject);
        drop_list(&Tag);
        drop_list(&Validity);    
        return CSSMERR_CL_MEMORY_ERROR;
    }
    w->ElementType = CSSM_LIST_ELEMENT_DATUM;
    w->WordID = CSSM_WORDID_CERT;
    w->Element.Word.Length = 4;
    w->Element.Word.Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 4, poolptr);
    if (!w->Element.Word.Data )
    {
        CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, w, poolptr);
        drop_list(&Issuer);
        drop_list(&Subject);
        drop_list(&Tag);
        drop_list(&Validity);    
        return CSSMERR_CL_MEMORY_ERROR;
    }
    port_memcpy(w->Element.Word.Data, "cert", 4);


    TBSCertTempList.Head = w;
    TBSCertTempList.Tail = w;

    /* Add the issuer */
    w = new_element();
	if (!w)
    {
        drop_list(&Issuer);
        drop_list(&Subject);
        drop_list(&Tag);
        drop_list(&Validity);    
        drop_list(&TBSCertTempList);
        return CSSMERR_CL_MEMORY_ERROR;
    }
    w->ElementType = CSSM_LIST_ELEMENT_SUBLIST;
    w->WordID = CSSM_WORDID_ISSUER;
    w->Element.Sublist = Issuer;
    
    TBSCertTempList.Tail->NextElement = w;
    TBSCertTempList.Tail = w;

    /* Add the subject */
    w = new_element();
	if (!w)
    {
        drop_list(&Subject);
        drop_list(&Tag);
        drop_list(&Validity);    
        drop_list(&TBSCertTempList);
        return CSSMERR_CL_MEMORY_ERROR;
    }
    w->ElementType = CSSM_LIST_ELEMENT_SUBLIST;
    w->WordID = CSSM_WORDID_SUBJECT;
    w->Element.Sublist = Subject;
    
    TBSCertTempList.Tail->NextElement = w;
    TBSCertTempList.Tail = w;

    /* Add the Tag */
    w = new_element();
	if (!w)
    {
        drop_list(&Tag);
        drop_list(&Validity);    
        drop_list(&TBSCertTempList);
        return CSSMERR_CL_MEMORY_ERROR;
    }
    w->ElementType = CSSM_LIST_ELEMENT_SUBLIST;
    w->WordID = CSSM_WORDID_TAG;
    w->Element.Sublist = Tag;
    
    TBSCertTempList.Tail->NextElement = w;
    TBSCertTempList.Tail = w;

    /* Add the Validity if it exists */
    if (Parts->Validity.Data)
    {

        w = new_element();
	    if (!w)
        {
            drop_list(&Validity);    
            drop_list(&TBSCertTempList);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->ElementType = CSSM_LIST_ELEMENT_SUBLIST;
        /* w->WordID = CSSM_WORDID_VALID; WHAT SHD go here ??? */
        w->Element.Sublist = Validity;
    
        TBSCertTempList.Tail->NextElement = w;
        TBSCertTempList.Tail = w;
    }

        /* Lets Pack the Template */
    TBSCert->Length = size_list(&TBSCertTempList);
    TBSCert->Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func,TBSCert->Length, poolptr);
    if (!TBSCert->Data)
    {
        drop_list(&TBSCertTempList);
        return CSSMERR_CL_MEMORY_ERROR;
    }
    pack_list(&(TBSCert->Data), &TBSCertTempList);


    /* Free and return. */
    drop_list(&TBSCertTempList);


	return CSSM_OK;
}

