/***************************************************************************/
/* tagint.c                                                                */
/*                                                                         */
/* Tag intersection code.                                                  */
/***************************************************************************/

#include <stdio.h>
#include <string.h>
#include "tagint.h"
#include "bs_compare.h"
#include "lookup.h"
#include "sutil.h"
#include "free.h"
#include "psexp.h"
#include "mempool.h"           /* define MALLOC, etc. */
#include "make_bs.h"            /* make a literal S_BS from WordID */


extern MEMORY_POOL_FUNCS memory_pool_funcs;
extern MEMORY_POOL_PTR poolptr;

static int elt_compare( pSexp u, pSexp v ) ;
static pSexp ti_manual( pSexp u, pSexp v ) ;
static pSexp ti_element( pSexp u, pSexp v ) ;

/***************************************************************************/
/* tag_compare return values -- defined locally only.                      */
/***************************************************************************/

#define T_EITHER (0)            /* can use either tag, they're equal */
#define T_U (1)                 /* use u */
#define T_V (2)                 /* use v */
#define T_FAIL (3)              /* can't use either -- the match fails */
#define T_MANUAL (4)            /* have to intersect the hard way */

/***************************************************************************/
/* tag_compare                                                             */
/*                                                                         */
/* Compare two tag S-expressions.  Return one of the return values         */
/* defined above.                                                          */
/***************************************************************************/

static int tag_compare( pSexp u, pSexp v )
{
  int res = T_EITHER ;          /* result so far, walking L to R */
  int elt = T_EITHER ;          /* result from one element */

  while (1) {                   /* walk u and v together */
    if (u == v) return ( res ) ; /* they're the same */
    if (u == NULL)              /* v is longer, so it should be used */
      switch (res) {
      case T_EITHER: case T_V: return ( T_V ) ;
      default: return ( T_MANUAL ) ;
      } /* switch */
    if (v == NULL)              /* u is longer, so it should be used */
      switch (res) {
      case T_EITHER: case T_U: return ( T_U ) ;
      default: return ( T_MANUAL ) ;
      } /* switch */

    elt = elt_compare( u, v ) ; /* compare just these two elements */

    switch (elt) {
    case T_EITHER:
      break ;                   /* leave res alone */

    case T_U:
      switch ( res ) {
      case T_EITHER: res = T_U ;
      case T_U: break ;
      case T_V: res = T_MANUAL ;
      default: break ;
      } /* switch */
      break ;

    case T_V:
      switch ( res ) {
      case T_EITHER: res = T_V ;
      case T_V: break ;
      case T_U: res = T_MANUAL ;
      default: break ;
      } /* switch */
      break ;

    case T_MANUAL:
      res = T_MANUAL ;
      break ;

    case T_FAIL:
    default: return ( T_FAIL ) ;
    } /* switch (elt) */

   u = u->term ;                /* walk to the next elements */
   v = v->term ;                /* ... */
  } /* while (1) */
  return ( res ) ;              /* can never reach this statement */
} /* tag_compare */

/***************************************************************************/
/* elt_match                                                               */
/*                                                                         */
/* Test two pSexp elements to see if they match.                           */
/* u is being tested against an element, v, of a SET list.                 */
/***************************************************************************/

static int elt_match( pSexp u, pSexp v )
{
  switch (elt_compare( u, v )) {
  case T_EITHER: case T_U: return ( 1 ) ; /* TRUE -- it is a member */
  default: break ;
  } /* switch */
  return ( 0 ) ;                /* FALSE -- it's not a member here */
} /* elt_match */

/***************************************************************************/
/* set_compare                                                             */
/*                                                                         */
/* Walk the set, s.  For each element in s, do an elt_compare to v.        */
/* Return T_FAIL if no element matches.                                    */
/* Return T_MANUAL if at least one does.                                   */
/***************************************************************************/

static int set_compare( pSexp s, pSexp v )
{
  if ((v == NULL) || (s == NULL)) return ( 0 ) ;
  if (s->typ == S_SEXP) s = s->u.s ; /* get to the list head */
  if (s->typ != S_KW) return ( T_FAIL ) ;
  if (s->u.k != O__STAR_) return ( T_FAIL ) ;
  s = s->term ;                 /* next term */
  if (s == NULL) return ( T_V ) ; /* it's a (*) */
  if (s->typ != S_BS) return ( T_FAIL ) ;
  if (s->u.b->lu != O_SET) return ( T_FAIL ) ;
  /* OK, now we're at the set contents.  Scan it. */
  while ((s = s->term) != NULL) {
    if (elt_match( s, v )) return ( T_MANUAL ) ; /* found one. */
  } /* while s */
  return ( T_FAIL ) ;           /* no element survives */
} /* set_compare */

/***************************************************************************/
/* in_set                                                                  */
/*                                                                         */
/* Return (1) if S-expression element a is listed in set s.                */
/***************************************************************************/

static int in_set( pSexp a, pSexp s )
{
  if ((a == NULL) || (s == NULL)) return ( 0 ) ;
  if (s->typ == S_SEXP) s = s->u.s ; /* get to the list head */
  if (s->typ != S_KW) return ( 0 ) ;
  if (s->u.k != O__STAR_) return ( 0 ) ;
  s = s->term ;                 /* next term */
  if (s == NULL) return ( 1 ) ; /* it's a (*) -- and therefore matches */
  if (s->typ != S_BS) return ( 0 ) ;
  if (s->u.b->lu != O_SET) return ( 0 ) ;
  /* OK, now we're at the set contents.  Scan it. */
  while ((s = s->term) != NULL) {
    if (elt_match( a, s )) return ( 1 ) ; /* found it! */
  } /* while s */
  return ( 0 ) ;                /* not there */
} /* in_set */

/***************************************************************************/
/* ok_numeric                                                              */
/*                                                                         */
/* Return (1) if this is signed decimal (in ASCII).                        */
/***************************************************************************/

static int ok_numeric( pBytes b )
{
  int i ;                       /* loop index */
  int j = 0 ;                   /* initial index */
  UC c ;                        /* temp char */

  if (b == NULL) return ( 0 ) ;
  if (b->d_lth != 0) return ( 0 ) ;
  if ((c = b->val[0]) == '+') j = 1 ;
  if (c == '-') j = 1 ;
  if (j < ((b->v_lth)-1))       /* not the last digit? */
    if (b->val[j] == '0')       /* unnecessary leading 0? */
      return ( 0 ) ;            /* disallow it */
  for (i=j; i<b->v_lth; i++) {
    c = b->val[i] ;
    if ((c < '0') || ('9' < c)) return ( 0 ) ;
  } /* for */
  return ( 1 ) ;
} /* ok_numeric */

/***************************************************************************/
/* ok_fmt                                                                  */
/*                                                                         */
/* Check b against the format in f.  We're not allowed to cut a numeric    */
/* field short (one where f has a '#').  b must not extend beyond f.       */
/*                                                                         */
/* b must not have a display hint.                                         */
/***************************************************************************/

static int ok_fmt( pBytes b, char *f )
{
  int fl = strlen( f ) ;
  int i ;                       /* loop index */

  if (b == NULL) return ( 0 ) ; /* must be there */
  if (b->d_lth != 0) return ( 0 ) ; /* must not have display hint */
  if (b->v_lth > fl) return ( 0 ) ; /* must not be too long */
  if (f[ b->v_lth ] == '#') return ( 0 ) ; /* partial numeric field */
  for (i=0; i<b->v_lth; i++)
    if (f[i] == '#') {
      UC c = b->val[i] ;
      if ((c < '0') || ('9' < c)) return ( 0 ) ;
    } else {
      if (b->val[i] != f[i]) return ( 0 ) ;
    } /* else */
  return ( 1 ) ;                /* succeeded */
} /* ok_fmt */

/***************************************************************************/
/* ok_time                                                                 */
/*                                                                         */
/* Check b against time syntax.  It's allowed to be shorter, but only      */
/* by whole fields.                                                        */
/***************************************************************************/

static int ok_time( pBytes b )
{
  int n, h, m, s ;
  char st[24] ;                 /* for a copy of the string */

  if (!ok_fmt( b, "##:##:##" )) return ( 0 ) ;
  memcpy( st, b->val, 22 ) ;
  st[ b->v_lth ] = '\0' ;       /* terminate the string */
  n = sscanf( st, "%d:%d:%d", &h, &m, &s ) ;
  switch (n) {                  /* how many fields are present? */
  case 3: if (s > 59) return ( 0 ) ;
  case 2: if (m > 59) return ( 0 ) ;
  case 1: if (h > 23) return ( 0 ) ;
    return ( 1 ) ;              /* success */
  default: break ;
  } /* switch (n) */
  return ( 0 ) ;                /* fail */
} /* ok_time */

/***************************************************************************/
/* leap                                                                    */
/*                                                                         */
/* return 1 if this is a leap year.                                        */
/***************************************************************************/

static int leap( int y )
{
  int x ;

  if ((y%4) != 0) return ( 0 ) ; /* must be == 0 mod 4 */
  if ((y%100) != 0) return ( 1 ) ; /* only centuries are special */
  x = y / 100 ;                 /* century number */
  if ((x%4) != 0) return ( 0 ) ; /* century must be == 0 mod 4 */
  return ( 1 ) ;                /* OK, this is a leap year */
} /* leap */

/***************************************************************************/
/* ok_date                                                                 */
/*                                                                         */
/* Check b against date syntax.  It's allowed to be shorter, but only      */
/* by whole fields.                                                        */
/***************************************************************************/

static int ok_date( pBytes b )
{
  int n, y, mo, d, h, m, s ;
  char st[24] ;                 /* for a copy of the string */

  if (!ok_fmt( b, "####-##-##_##:##:##" )) return ( 0 ) ;
  memcpy( st, b->val, 22 ) ;
  st[ b->v_lth ] = '\0' ;       /* terminate the string */
  n = sscanf( st, "%d-%d-%d_%d:%d:%d", &y, &mo, &d, &h, &m, &s ) ;
  switch (n) {                  /* how many fields are present? */
  case 6: if (s > 59) return ( 0 ) ;
  case 5: if (m > 59) return ( 0 ) ;
  case 4: if (h > 23) return ( 0 ) ;
    /* check date only after checking month */
  case 2: if ((mo > 12) || (mo < 1)) return ( 0 ) ;
  case 1: if (n < 3) return ( 1 ) ; /* success -- allow any year */
  case 3:                       /* check the date of the month */
    if (d < 1) return ( 0 ) ;   /* date of 0 not allowed */
    switch (mo) {
    case 1: case 5: case 7: case 8: case 10: case 12:
      if (d <= 31) return ( 1 ) ;
      break ;

    case 2:                     /* check leap year */
      if (d <= (28+leap(y))) return ( 1 ) ;
      break ;

    case 4: case 6: case 9: case 11:
      if (d <= 30) return ( 1 ) ;
      break ;

    default: break ;
    } /* check the month */
    return ( 0 ) ;
  default: break ;
  } /* switch */
  return ( 0 ) ;                /* fail */
} /* ok_date */

/***************************************************************************/
/* fmt_comp                                                                */
/*                                                                         */
/* Compare two byte strings, a and b, according to the format.             */
/***************************************************************************/

static int fmt_comp( pBytes a, pBytes b, int format )
{
  if ((a == NULL)||(b == NULL)) return ( INCOMPARABLE ) ;
  switch ( format ) {           /* now, do the comparison */
  case O_ALPHA: case O_DATE: case O_TIME:
    return ( bs_compare( a, b ) ) ;
  case O_BINARY:
    return ( bin_compare( a, b ) ) ;
  case O_NUMERIC:
    return ( num_compare( a, b ) ) ;
  default: break ;       /* don't know how to do this type */
  } /* switch ( format ) */
  return ( INCOMPARABLE ) ;
} /* fmt_comp */

/***************************************************************************/
/* fmt_force_comp                                                          */
/*                                                                         */
/* If a is NULL, fail.                                                     */
/* If b is NULL, force success (with a -1 return).                         */
/* Otherwise, do fmt_comp.                                                 */
/***************************************************************************/

static int fmt_force_comp( pBytes a, pBytes b, int format )
{
  if (a == NULL) return ( INCOMPARABLE ) ;
  if (b == NULL) return ( -1 ) ; /* force it */
  return ( fmt_comp( a, b, format ) ) ;
} /* fmt_force_comp */

/***************************************************************************/
/* bs_comp                                                                 */
/*                                                                         */
/* Compare the byte string, b, against the range check in s.  This can be  */
/* (xx <byte-string>) where xx is l, le, g or ge.                          */
/***************************************************************************/

static int bs_comp( pBytes b, pSexp s, int format )
{
  int comp ;                    /* desired comparison type */
  int res ;                     /* actual comparison result */

  if ((b == NULL) || (s == NULL)) return ( 0 ) ;
  if (s->typ == S_SEXP) s = s->u.s ;
  if (s == NULL) return ( 0 ) ;
  if (s->typ != S_KW) return ( 0 ) ;
  comp = s->u.k ;               /* comparison type */
  s = s->term ;
  if (s == NULL) return ( 0 ) ;
  if (s->term != NULL) return ( 0 ) ;
  if (s->typ != S_BS) return ( 0 ) ;

  switch (format) {             /* check the format of the range element */
  case O_BINARY:                /* right-justified binary */
  case O_ALPHA:                 /* left-justified arbitrary */
    break ; /* any string is OK */
  case O_NUMERIC:               /* right-justified, signed decimal */
    if (!ok_numeric( s->u.b )) return ( 0 ) ;
    break ;
  case O_DATE:                  /* left-justified fixed date format */
    if (!ok_date( s->u.b )) return ( 0 ) ;
    break ;
  case O_TIME:                  /* left-justified fixed time format */
    if (!ok_time( s->u.b )) return ( 0 ) ;
    break ;
  default: return ( 0 ) ;
  } /* switch (format code) */

  res = fmt_comp( b, s->u.b, format ) ;

  switch (comp) {
  case O_L:                     /* strictly less than */
    return ( res == LESSER ) ;
  case O_G:
    return ( res == GREATER ) ;
  case O_LE:
    return ( (res == LESSER) || (res == EQUAL) ) ;
  case O_GE:
    return ( (res == GREATER) || (res == EQUAL) ) ;
  default: break ;
  } /* switch (comp) */
  return ( 0 ) ;
} /* bs_comp */

/***************************************************************************/
/* in_range                                                                */
/*                                                                         */
/* Return (1) if S-expression element a is a byte string in range s.       */
/***************************************************************************/

static int in_range( pSexp a, pSexp s )
{
  int format = 0 ;              /* format type O_... */

  if ((a == NULL) || (s == NULL)) return ( 0 ) ;
  if (a->typ != S_BS) return ( 0 ) ;
  if (s->typ != S_SEXP) return ( 0 ) ;
  s = s->u.s ;                  /* get to the * */
  if (s == NULL) return ( 0 ) ;
  if (s->typ != S_KW) return ( 0 ) ;
  if (s->u.k != O__STAR_) return ( 0 ) ;
  s = s->term ;                 /* next term */
  if (s == NULL) return ( 1 ) ; /* it's a (*) node */
  if (s->typ != S_BS) return ( 0 ) ;
  if (s->u.b->lu != O_RANGE) return ( 0 ) ;
  s = s->term ;                 /* next term: format of string */
  if (s == NULL) return ( 0 ) ; /* no limits -- always fail */
  if (s->typ != S_BS) return ( 0 ) ;
  switch (format = s->u.b->lu) {
  case O_BINARY:                /* right-justified binary */
  case O_ALPHA:                 /* left-justified arbitrary */
    break ; /* any string is OK */
  case O_NUMERIC:               /* right-justified, signed decimal */
    if (!ok_numeric( a->u.b )) return ( 0 ) ;
    break ;
  case O_DATE:                  /* left-justified fixed date format */
    if (!ok_date( a->u.b )) return ( 0 ) ;
    break ;
  case O_TIME:                  /* left-justified fixed time format */
    if (!ok_time( a->u.b )) return ( 0 ) ;
    break ;
  default: return ( 0 ) ;
  } /* switch (format code) */
  /* passed the format check */
  /* OK, we're at the range contents.  Test a against those limits. */
  while ((s = s->term) != NULL)
    if (!bs_comp( a->u.b, s, format )) return ( 0 ) ;
  return ( 1 ) ;                /* success if passed all comparisons */
} /* in_range */

/***************************************************************************/
/* in_prefix                                                               */
/*                                                                         */
/* Return (1) if byte string a starts with the prefix, s.                  */
/***************************************************************************/

static int in_prefix( pSexp a, pSexp s )
{
  pBytes ab, sb ;               /* (walkers) */

  if ((a == NULL) || (s == NULL)) return ( 0 ) ;
  if (a->typ != S_BS) return ( 0 ) ;
  ab = a->u.b ;
  if (ab == NULL) return ( 0 ) ;
  if (s->typ != S_SEXP) return ( 0 ) ;
  s = s->u.s ;                  /* get to the set */
  if (s == NULL) return ( 0 ) ;
  if (s->typ != S_KW) return ( 0 ) ;
  if (s->u.k != O__STAR_) return ( 0 ) ;
  s = s->term ;                 /* next term */
  if (s == NULL) return ( 0 ) ;
  if (s->typ != S_BS) return ( 0 ) ;
  if (s->u.b->lu != O_PREFIX) return ( 0 ) ;
  s = s->term ;                 /* next term: the prefix */
  if (s == NULL) return ( 0 ) ;
  if (s->typ != S_BS) return ( 0 ) ;
  sb = s->u.b ;
  if (sb == NULL) return ( 0 ) ;
  if (ab->d_lth != sb->d_lth) return ( 0 ) ; /* display hints must be == */
  if (ab->d_lth > 0)
    if (memcmp( ab->disp, sb->disp, ab->d_lth ) != 0)
      return ( 0 ) ;            /* must be == to process */
  if (ab->v_lth < sb->v_lth) return ( 0 ) ; /* must be longer than prefix */
  return (memcmp( ab->val, sb->val, sb->v_lth ) == 0) ;
} /* in_prefix */

/***************************************************************************/
/* pull_ts                                                                 */
/*                                                                         */
/* Pull the type of u, and if it's an S_SEXP, pull_star for it.            */
/***************************************************************************/

static int pull_ts( pSexp u )
{
  int ut = pull_typ( u ) ;
  int uts ;
  switch (ut) {
  case S_SEXP: case S_KW:
    uts = pull_star( u ) ;      /* could be a star node */
    switch (uts) {
    case ST_STAR: case ST_SET: case ST_PREFIX: case ST_RANGE:
      return ( uts ) ;
    default: break ;
    } /* switch (uts) */
  default: break ;
  } /* switch */
  return ( ut ) ;
} /* pull_ts */

/***************************************************************************/
/* prefix_compare                                                          */
/*                                                                         */
/* Compare two (* prefix ) tag elements.                                   */
/***************************************************************************/

static int prefix_compare( pSexp u0, pSexp v0 )
{
  pSexp u = u0 ;                /* walker of u0 */
  pSexp v = v0 ;                /* walker of v0 */

  if (pull_ts( u ) != ST_PREFIX) return ( T_FAIL ) ;
  if (pull_ts( v ) != ST_PREFIX) return ( T_FAIL ) ;
  if (u->typ == S_SEXP) u = u->u.s ;
  if (v->typ == S_SEXP) v = v->u.s ;
  u = u->term->term ;           /* the prefix string */
  v = v->term->term ;           /* .. */
  if ((u == NULL) || (v == NULL)) return ( T_FAIL ) ;
  if (pull_typ( u ) != S_BS) return ( T_FAIL ) ;
  if (pull_typ( v ) != S_BS) return ( T_FAIL ) ;

  if ( in_prefix( u, v0 ) ) return ( T_U ) ;
  if ( in_prefix( v, u0 ) ) return ( T_V ) ;

  return ( T_FAIL ) ;           /* couldn't compare */
} /* prefix_compare */

/***************************************************************************/
/* Rbuild                                                                  */
/*                                                                         */
/* Structure for gathering the min and max of two ranges, for building     */
/* an intersection of the two.                                             */
/***************************************************************************/

typedef struct {                /* for range_compare */
  pBytes minv ;                 /* (walker: no drop) min value */
  int mintyp ;                  /* type (O_G or O_GE) */
  int minloc ;                  /* location (T_U or T_V) */
  pBytes maxv ;                 /* (walker: no drop) max value */
  int maxtyp ;                  /* type (O_L or O_LE) */
  int maxloc ;                  /* location (T_U or T_V) */
  int format ;                  /* format for this range */
} Rbuild, *pRbuild ; 

/***************************************************************************/
/* init_rbuild                                                             */
/*                                                                         */
/* Initialize an Rbuild block.                                             */
/***************************************************************************/

static void init_rbuild( pRbuild r )
{
  if (r == NULL) return ;
  memset( r, 0, sizeof( Rbuild ) ) ;
  r->maxloc = r->minloc = T_EITHER ;
  r->format = -1 ;              /* meaning, not initialized yet */
  return ;
} /* init_rbuild */

/***************************************************************************/
/* clear_rbuild                                                            */
/*                                                                         */
/* Clean up an Rbuild block.                                               */
/***************************************************************************/

static void clear_rbuild( pRbuild r )
{
  if (r == NULL) return ;
  drop_bytes( r->minv ) ;
  drop_bytes( r->maxv ) ;
  init_rbuild( r ) ;
  return ;
} /* clear_rbuild */

/***************************************************************************/
/* elt_of_range                                                            */
/*                                                                         */
/* Return the loc result from the Rbuild, r.                               */
/***************************************************************************/

static int elt_of_range( pRbuild r )
{
  if (r == NULL) return ( T_FAIL ) ;
  /* check first for overlap in the range */
  if ((r->maxv != NULL) && (r->minv != NULL)) /* possible -- check more */
    switch (fmt_comp( r->minv, r->maxv, r->format )) {
    case LESSER: break ;        /* good -- all OK */
    case EQUAL:                 /* OK iff both allow equality */
      if ((r->mintyp == O_GE) && (r->maxtyp == O_LE))
        return ( T_MANUAL ) ;   /* convert to a single S_BS */
      else return ( T_FAIL ) ;  /* failure */
    default: return ( T_FAIL ) ; /* range empty or some error */
    } /* if, switch */
  switch ( r->minloc ) {
  case T_EITHER: return ( r->maxloc ) ;
  case T_U: case T_V:
    switch ( r->maxloc ) {
    case T_U: case T_V:
      if ( r->maxloc == r->minloc ) return ( r->minloc ) ;
      else return ( T_MANUAL ) ;
    case T_EITHER:
      return ( r->minloc ) ;
    default: break ;
    } /* switch (r->maxloc) */
  default: break ;
  } /* switch */
  return ( T_FAIL ) ;           /* couldn't match the range */
} /* elt_of_range */

/***************************************************************************/
/* note_range_limit                                                        */
/*                                                                         */
/* Gather a single range limit into the Rbuild, r.                         */
/***************************************************************************/

static void note_range_limit( pRbuild r,        /* the gathering block */
                             pBytes b, /* the value */
                             int typ, /* the type of limit (e.g., O_G) */
                             int fmt, /* the string format */
                             int loc /* the location (e.g., T_U) */
                           )
{
  if ((r == NULL)||(b == NULL)) return ; /* no work to do */
  if (r->format < 0) r->format = fmt ; /* save the first format */
  if (r->format != fmt) {       /* and it has to match from then on */
    clear_rbuild( r ) ;         /* drop any results so far */
    r->maxloc = T_FAIL ;        /* and fail */
  } /* if */
  /* Now, check b for format */
  switch (fmt) {                /* check format */
  case O_BINARY: case O_ALPHA: break ; /* all strings OK */
  case O_NUMERIC:               /* right-justified, signed decimal */
    if (!ok_numeric( b )) r->maxloc = T_FAIL ;
    break ;
  case O_DATE:                  /* left-justified fixed date_time format */
    if (!ok_date( b )) r->maxloc = T_FAIL ;
    break ;
  case O_TIME:                  /* left-justified fixed time format */
    if (!ok_time( b )) r->maxloc = T_FAIL ;
    break ;
  default: r->maxloc = T_FAIL ;
  } /* switch (f) */

  if (r->maxloc == T_FAIL) return ;

  /* OK.  Let's gather max or min. */
  switch (typ) {
  case O_G: case O_GE:          /* min nodes */
    switch (fmt_force_comp( b, r->minv, fmt )) {
    case -1: case GREATER:      /* replace this one */
      store_bytes( &(r->minv), b ) ;
      r->mintyp = typ ;
      r->minloc = loc ;
      break ;
    case EQUAL:
      if ((r->mintyp == O_GE) && (typ == O_G)) {
        r->mintyp = O_G ;
        r->minloc = loc ;
      } /* if */
      break ;
    case LESSER: break ;        /* no effect */
    default: r->maxloc = T_FAIL ; /* bad comparison, so really fail */
      break ;
    } /* switch (typ) */
    break ;

  case O_L: case O_LE:
    switch (fmt_force_comp( b, r->maxv, fmt )) {
    case -1: case LESSER:       /* replace this one */
      store_bytes( &(r->maxv), b ) ;
      r->maxtyp = typ ;
      r->maxloc = loc ;
      break ;
    case EQUAL:
      if ((r->maxtyp == O_LE) && (typ == O_L)) {
        r->maxtyp = O_L ;
        r->maxloc = loc ;
      } /* if */
      break ;
    case GREATER: break ;       /* no effect */
    default: r->maxloc = T_FAIL ; /* bad comparison, so really fail */
      break ;
    } /* switch (typ) */
    break ;

  default: r->maxloc = T_FAIL ; /* bad type, so fail */
    break ;
  } /* switch (typ) */
  return ;
} /* note_range_limit */

/***************************************************************************/
/* r_gather                                                                */
/*                                                                         */
/* Gather range comparison results into r.                                 */
/* t should be a range test (O_L O_LE O_G O_GE).                           */
/* t came from side loc (T_U or T_V).                                      */
/* The string is in the format f.                                          */
/***************************************************************************/

static void r_gather( pSexp t, pRbuild r, int loc, int f )
{
  int comp = 0 ;                /* comparison type */
  int ok = 1 ;                  /* format OK so far */
  pBytes b = NULL ;             /* (walker) -- the value in t */

  if ((t == NULL)||(r == NULL)) return ; /* can't work without data */
  if (t->typ == S_SEXP) t = t->u.s ; /* walk into the expression */
  if (t == NULL) ok = 0 ;
  else if (t->typ != S_KW) ok = 0 ;
  else comp = t->u.k ;          /* get the keyword */
  if (ok) if ((t = t->term) == NULL) ok = 0 ;
  /* OK -- have the term.  Check its format. */
  if (ok) {
    if (t->typ != S_BS) ok = 0 ;
    else b = t->u.b ;
  }
  if (ok) note_range_limit( r, b, comp, f, loc ) ;
  else r->maxloc = T_FAIL ;
  return ;                      /* and return to the caller */
} /* r_gather */

/***************************************************************************/
/* note_range                                                              */
/*                                                                         */
/* Plant a whole (* range ...) in Rbuild, r.                               */
/***************************************************************************/

static void note_range( pRbuild r, pSexp s, int loc )
{
  int format ;                  /* comparison format */

  if (r == NULL) return ;       /* can't work with no place to put it */
  if (pull_ts( s ) != ST_RANGE) {
    r->maxloc = T_FAIL ;
    return ;
  } /* if */
  if (s->typ == S_SEXP) s = s->u.s ;
  s = s->term->term ;           /* the type of range */
  if (pull_typ( s ) != S_BS) {
    r->maxloc = T_FAIL ;        /* no type, error */
    return ;
  } /* if */
  format = s->u.b->lu ;         /* save the comparison type for later  */
  while( (s = s->term) != NULL ) r_gather( s, r, loc, format ) ;
  return ;
} /* note_range */

/***************************************************************************/
/* put_pref_limits                                                         */
/*                                                                         */
/* Convert a prefix S-expression to a range pair of limits and             */
/* enter those in the Rbuild, r.                                           */
/***************************************************************************/

static void put_pref_limits( pRbuild r, pSexp s, int loc )
{
  if (r == NULL) return ;       /* can't work with no block */
  if (pull_ts( s ) != ST_PREFIX) { /* has to be a prefix block */
    r->maxloc = T_FAIL ;        /* bad format, so fail */
    return ;
  } /* if */
  if (s->typ == S_SEXP) s = s->u.s ;
  s = s->term->term ;           /* the prefix value */
  if (pull_typ( s ) != S_BS) {
    r->maxloc = T_FAIL ;        /* no type, error */
    return ;
  } /* if */
  /* have the prefix value -- use it */
  note_range_limit( r, s->u.b, O_GE, O_ALPHA, loc ) ;
  if (elt_of_range( r ) == T_FAIL) return ; /* don't bother if already fail */
  /* OK -- now we have to do the hard work. */
  /* Build a new pBytes node, representing the end of that range. */
  {
    pBytes b = NULL ;           /* the new bytes node */
    unsigned char *txt = NULL ; /* the bytes inside */
    int i ;                     /* an index into that array */

    i = s->u.b->v_lth ;         /* length of the prefix string */
    while ((i > 0) && (s->u.b->val[i-1] == 0xff)) i-- ;
    /* i is now the length of the upper limit string */
    if (i > 0) {                /* otherwise, don't have to do it */
      txt = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func,
                                i,
                                poolptr) ;
      b = new_bytes() ;         /* and a block to put it in */
      if ((b == NULL)||(txt == NULL)) {
        r->maxloc = T_FAIL ;
        if (txt != NULL) CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, txt, poolptr ) ;
        store_bytes( &b, NULL ) ;
        return ;
      } /* if */
      /* have the blocks -- use them */
      b->val = txt ;
      b->drop_cont = 1 ;        /* perhaps the only instance of this */
      memcpy( txt, s->u.b->val, i ) ; /* move the bytes */
      txt[ i-1 ]++ ;            /* bump that byte */
      b->v_lth = i ;            /* plant the length */
      note_range_limit( r, b, O_L, O_ALPHA, loc ) ;
    } /* have a block to put away */
    store_bytes( &b, NULL ) ;   /* and drop my temp copy */
  } /* build top end of range of this prefix */
  return ;
} /* put_pref_limits */

/***************************************************************************/
/* range_compare                                                           */
/*                                                                         */
/* Compare two (* range ) tag elements.                                    */
/***************************************************************************/

static int range_compare( pSexp u, pSexp v )
{
  Rbuild rb ;                   /* the range to build */
  int ans ;                     /* function result */

  init_rbuild( &rb ) ;          /* init the block gathering limits */
  note_range( &rb, u, T_U ) ;   /* put away u's limits */
  note_range( &rb, v, T_V ) ;   /* put v's limits */
  ans = elt_of_range( &rb ) ;   /* get the answer (U, V, MANUAL, ...) */
  clear_rbuild( &rb ) ;         /* clear the working block */
  return ( ans ) ;              /* and return */
} /* range_compare */

/***************************************************************************/
/* pref_range_compare                                                      */
/*                                                                         */
/* Compare prefix, u, against range, v.                                    */
/***************************************************************************/

static int pref_range_compare( pSexp u, pSexp v )
{
  Rbuild rb ;                   /* working block */
  int ans ;                     /* my answer */

  init_rbuild( &rb ) ;          /* init the block */
  note_range( &rb, v, T_V ) ;   /* plant its range */
  put_pref_limits( &rb, u, T_U ) ; /* plant the prefix block */
  ans = elt_of_range( &rb ) ;   /* get the answer */
  clear_rbuild( &rb ) ;         /* clear the working block */
  return ( ans ) ;              /* and return */
} /* pref_range_compare */

/***************************************************************************/
/* t_reverse                                                               */
/*                                                                         */
/* Return the reverse of the input T_ value.                               */
/* That is, reverse T_U and T_V.  Everything else is the same.             */
/***************************************************************************/

static int t_reverse( int res )
{
  switch ( res ) {
  case T_U: return ( T_V ) ;
  case T_V: return ( T_U ) ;
  default: break ;
  } /* switch */
  return ( res ) ;
} /* t_reverse */

/***************************************************************************/
/* elt_compare                                                             */
/*                                                                         */
/* Compare two elements of a tag -- not whole lists.                       */
/* Return whether to use the first, the second, neither, either or some    */
/* manual mixture of the two.                                              */
/***************************************************************************/

static int elt_compare( pSexp u, pSexp v )
{
  int ut ;                 /* tupe of u, with and without star */
  int vt ;                 /* type of v, with and without star */

  if (u == v) return ( T_EITHER ) ;
  if ((u == NULL) || (v == NULL))
    return ( T_FAIL ) ;         /* bad calling sequence -- should */
                                /* always be non-null */
  ut = pull_ts( u ) ;           /* gather the element types */
  vt = pull_ts( v ) ;           /* ... */

  if (vt == ST_STAR) {
    if (ut == ST_STAR)
      return ( T_EITHER ) ;
    else
      return ( T_U ) ;
  } /* v was (*) */
  /* v is not (*) now */
  if (ut == ST_STAR) return ( T_V ) ;
  /* u is not (*) either, now */
  if (vt == ST_SET) {
    switch (ut) {
    case ST_SET: case ST_RANGE: case ST_PREFIX:
      return ( t_reverse( set_compare( v, u ) ) ) ;
    case S_BS: case S_KW: case S_UKW: case S_SEXP:
      if (in_set( u, v ))
        return ( T_U ) ;
      else return ( T_FAIL ) ;
    default: return ( T_FAIL ) ;
    } /* switch (ut) */
  } /* v is (* set ...) */

  /* v is not (*) or (* set) */
  /*  do the remaining comparisons */
  switch (ut) {
  case S_KW:                    /* known keyword */
    if (vt != S_KW) return ( T_FAIL ) ; /* must match */
    if (u->u.k != v->u.k) return ( T_FAIL ) ;
    else return ( T_EITHER ) ;

  case S_UKW:                   /* unknown keyword */
    if (vt != S_UKW) return ( T_FAIL ) ; /* must match */
    if (bs_compare( u->u.b, v->u.b ) != EQUAL)
      return ( T_FAIL ) ;
    else return ( T_EITHER ) ;

  case S_SEXP:                /* routine S-expression */
    if (vt == S_SEXP)
      return ( tag_compare( u->u.s, v->u.s ) ) ;
        else return ( T_FAIL ) ;

  case S_BS:                    /* byte string */
    switch (vt) {
    case S_BS:
      if (bs_compare( u->u.b, v->u.b ) != EQUAL)
        return ( T_FAIL ) ;
      else return ( T_EITHER ) ;

    case ST_RANGE:
      if (in_range( u, v ))
        return ( T_U ) ;
      else return ( T_FAIL ) ;

    case ST_PREFIX:
      if (in_prefix( u, v ))
        return ( T_U ) ;
      else return ( T_FAIL ) ;

    default: return ( T_FAIL ) ;
    } /* switch (vt) */

  case ST_SET:
    switch (vt) {
    case ST_SET: case ST_RANGE: case ST_PREFIX:
      return ( set_compare( u, v ) ) ;
    case S_KW: case S_UKW: case S_BS: case S_SEXP:
      if (in_set( v, u )) return ( T_V ) ;
      else return ( T_FAIL ) ;
    default: return ( T_FAIL ) ;
    } /* switch (vt) */

  case ST_RANGE:
    switch (vt) {
    case ST_RANGE:
      return ( range_compare( u, v ) ) ;
    case ST_PREFIX:
      return ( t_reverse( pref_range_compare( v, u ) ) ) ;
    case S_BS:
      if (in_range( v, u )) return ( T_V ) ;
      /* otherwise, fall through */
    default:
      return ( T_FAIL ) ;
    } /* switch (vt) */

  case ST_PREFIX:
    switch (vt) {
    case ST_PREFIX:
      return ( prefix_compare( u, v ) ) ;
    case ST_RANGE:
      return ( pref_range_compare( u, v ) ) ;
    case S_BS:
      if (in_prefix( v, u )) return ( T_V ) ;
      /* otherwise, fall through */
    default:
      return ( T_FAIL ) ;
    } /* switch (vt) */

  default: break ;
  } /* switch (ut) */
  return ( T_FAIL ) ;           /* if we get here */
} /* elt_compare */

/***************************************************************************/
/* pull_star                                                               */
/*                                                                         */
/* Grab the type of * node from the S-expression, s.                       */
/***************************************************************************/

int pull_star( pSexp s )
{
  if (s == NULL) return ( ST_NONE ) ;
  if (s->typ == S_SEXP) return ( pull_star( s->u.s ) ) ;
  if (s->typ != S_KW) return ( s->typ ) ;
  if (s->u.k != O__STAR_) return ( s->typ ) ;
  s = s->term ;
  if (s == NULL) return ( ST_STAR ) ;
  if (s->typ != S_BS) return ( ST_NONE ) ;
  switch (s->u.b->lu) {
  case O_SET: return ( ST_SET ) ;
  case O_PREFIX: return ( ST_PREFIX ) ;
  case O_RANGE: return ( ST_RANGE ) ;
  default: break ;
  } /* switch */
  return ( ST_NONE ) ;
} /* pull_star */

/***************************************************************************/
/* bump_ref                                                                */
/*                                                                         */
/* In-line bump of refcnt.                                                 */
/***************************************************************************/

static pSexp bump_ref( pSexp x )
{
  if (x == NULL) return ( NULL ) ;
  x->refcnt++ ;                 /* bump it */
  return ( x ) ;
} /* bump_ref */

/***************************************************************************/
/* decr_ref                                                                */
/*                                                                         */
/* In-line decrement of refcnt.                                            */
/***************************************************************************/

static pSexp decr_ref( pSexp x )
{
  if (x == NULL) return ( NULL ) ;
  x->refcnt-- ;                 /* decrement it */
  return ( x ) ;
} /* decr_ref */

/***************************************************************************/
/* add_to_set                                                              */
/*                                                                         */
/* Add element v to set s.                                                 */
/* pS points to the start of the list of set elements, minus the           */
/* (* set ...) header.                                                     */
/* If v is itself a set, add all of it.                                    */
/* v is available to be added to the set, without additional copy.         */
/***************************************************************************/

static void add_to_set( pSexp *pS, pSexp v )
{
  pSexp s = NULL ;              /* for walking pS */

  if ((pS == NULL)||(v == NULL)) return ; /* no action if NULLs */

  if (pull_ts( v ) == ST_SET) { /* v != NULL, if pass pull_ts */
    s = v->u.s->term ;          /* the O_SET node */
    while ((s = s->term) != NULL) {
      pSexp n = copy_sexp_elt( s ) ; /* new element */
      add_to_set( pS, n ) ;     /* add it */
      drop_sexp( n ) ;          /* and drop our copy */
    } /* while */
    return ;                    /* and we're done */
  } /* exploding that set */

  /* v is a single element */
  /* walk s first, looking to see if v is already there */
  /* if so, just return */
  s = *pS ;                     /* get the set body */
  while (s != NULL) {
    if (elt_compare( s, v ) == T_EITHER) 
      return ;                  /* v is already there */
    s = s->term ;
  } /* while s */
  /* v is not in s yet.  Put it there. */
  store_sexp( &(v->term), *pS ) ; /* link in the new item */
  store_sexp( pS, v ) ;         /* ... */
  return ;                      /* and return */
} /* add_to_set */

/***************************************************************************/
/* pref_pref                                                               */
/*                                                                         */
/* Construct a node intersecting the two prefixes, p1 and p2.              */
/* The longer one needs to be the winner.                                  */
/*                                                                         */
/* This returns one of those two nodes and needs to be stored via          */
/* store_sexp().                                                           */
/***************************************************************************/

static pSexp pref_pref( pSexp p1, pSexp p2 )
{
  pSexp u, v ;                  /* walkers for p1 and p2 */

  if (pull_ts( u = p1 ) != ST_PREFIX) return ( NULL ) ;
  if (pull_ts( v = p2 ) != ST_PREFIX) return ( NULL ) ;
  if (p1->typ == S_SEXP) u = p1->u.s ;
  if (p2->typ == S_SEXP) v = p2->u.s ;
  u = u->term->term ;           /* the prefix string */
  v = v->term->term ;           /* .. */
  if ((u == NULL) || (v == NULL)) return ( NULL ) ;
  if (pull_typ( u ) != S_BS) return ( NULL ) ;
  if (pull_typ( v ) != S_BS) return ( NULL ) ;

  if ( in_prefix( u, p2 ) ) return ( p1 ) ;
  if ( in_prefix( v, p1 ) ) return ( p2 ) ;

  return ( NULL ) ;
} /* pref_pref */

/***************************************************************************/
/* put_range                                                               */
/*                                                                         */
/* Put a range test hanging off u->term and update u.                      */
/***************************************************************************/

static void put_range( pSexp *pU, pBytes b, int typ )
{
  pSexp bs = new_sexp_bs( b ) ;
  pSexp kw = new_sexp_kw( typ ) ;
  pSexp ref = new_sexp_sexp( kw ) ;

  if ((pU == NULL)||(b == NULL)||(bs == NULL)||(kw == NULL)||(ref == NULL)) {
    drop_sexp( bs ) ;
    drop_sexp( kw ) ;
    drop_sexp( ref ) ;
    return ;
  } /* if some NULLs */
  kw->term = bs ;               /* refcnt from the stack */
  ref->u.s = kw ;               /* refcnt from the stack */
  (*pU)->term = ref ;           /* refcnt from the stack */
  *pU = ref ;                   /* move the pointer over to the new tail */
  return ;
} /* put_range */

/***************************************************************************/
/* copy_sexp                                                               */
/*                                                                         */
/* Make a full copy of an S-expression.                                    */
/* This is rarely called and not provided in free.c, to discourage use.    */
/***************************************************************************/

static pSexp copy_sexp( pSexp u )
{
  pSexp res ;                   /* the result */

  switch (pull_typ( u ) ) {
  case S_KW: res = new_sexp_kw( u->u.k ) ;
    break ;
  case S_SEXP: res = copy_sexp( u->u.s ) ;
    break ;
  case S_UKW: case S_BS:
    res = new_sexp_bs( u->u.b ) ;
    if (res != NULL) res->typ = u->typ ;
    break ;
  default: res = NULL ;
    break ;
  } /* switch pull_typ( u ) */
  if (res == NULL) return ( NULL ) ;
  if (u->term != NULL) {
    pSexp term = copy_sexp( u->term ) ; /* rest of the list */

    if (term == NULL) {
      drop_sexp( res ) ;
      return ( NULL ) ;
    } /* failure */
    res->term = term ;
  } /* do the term */
  return ( res ) ;
} /* copy_sexp */

/***************************************************************************/
/* build_range                                                             */
/*                                                                         */
/* Construct a (* range ...) S-expression from the contents of r.          */
/***************************************************************************/

static pSexp build_range( pRbuild r )
{
  pSexp out = NULL ;            /* output expression */
  pSexp oref = NULL ;           /* and a ptr to it */
  pBytes b = NULL ;             /* for the format S_BS node */
  int res ;                     /* return value */

  switch (elt_of_range( r )) {  /* is this a legal range? */
  case T_U: case T_EITHER: case T_V: case T_MANUAL: break ;
  default:                      /* these are the failure cases */
    return ( NULL ) ;
  } /* switch */

  /* build the result */
  res = lit_sexp( &out, "(1:*5:range)" ) ;
  if ((res != 0)||(out == NULL)) {
    drop_sexp( out ) ;
    return ( NULL ) ;
  } /* error */

  /* have the basic output S-expression */
  oref = new_sexp_sexp( out ) ; /* make a reference to it, to return */
  drop_sexp( out ) ;          /* now, out is just a walker */
  if (oref == NULL) {
    drop_sexp( out ) ;
    return ( NULL ) ;
  } /* error */

  out = out->term ;             /* out is before the format type node */
                                /* add the real format */
  b = lit_bs( r->format ) ;     /* the right string for that wordid */
  if (b == NULL) {              /* did it work? */
    drop_sexp( oref ) ;
    return ( NULL ) ;
  } /* error */

  out->term = new_sexp_bs( b ) ; /* get the pSexp node to hold it */
  if (out->term == NULL) {      /* any error? */
    drop_sexp( oref ) ;
    return ( NULL ) ;
  } /* error */

  out = out->term ;             /* now, get ready for the range tests */

  /* OK -- tack nodes onto the end */
  /* have oref to return, and out to build */

  store_sexp( &(out->term), NULL ) ; /* build from here */
  put_range( &out, r->minv, r->mintyp ) ;
  put_range( &out, r->maxv, r->maxtyp ) ;

  return ( oref ) ;
} /* build_range */

/***************************************************************************/
/* range_range                                                             */
/*                                                                         */
/* Intersect two ranges.                                                   */
/***************************************************************************/

static pSexp range_range( pSexp r1, pSexp r2 )
{
  Rbuild rb ;                   /* the range to build */
  pSexp oref = NULL ;           /* and a ptr to it */

  init_rbuild( &rb ) ;          /* init the range result block */
  note_range( &rb, r1, T_U ) ;  /* gather r1's limits */
  note_range( &rb, r2, T_V ) ;  /* and r2's */
  /* now, we have what we need to build an output range */
  oref = build_range( &rb ) ;   /* get the S-expression built */
  clear_rbuild( &rb ) ;         /* clean up the temp storage */
  return ( oref ) ;             /* and return */
} /* range_range */

/***************************************************************************/
/* pref_range                                                              */
/*                                                                         */
/* Intersect a prefix, p, with a range, r.                                 */
/***************************************************************************/

static pSexp pref_range( pSexp p, pSexp r )
{
  Rbuild rb ;                   /* the range to build */
  pSexp oref = NULL ;           /* and a ptr to it */

  init_rbuild( &rb ) ;          /* init the range result block */
  note_range( &rb, r, T_V ) ;   /* gather the range limits */
  put_pref_limits( &rb, p, T_U ) ; /* gather the prefix limits */
  /* now, we have what we need to build an output range */
  oref = build_range( &rb ) ;   /* get the S-expression built */
  clear_rbuild( &rb ) ;         /* clean up the temp storage */
  return ( oref ) ;             /* and return */
} /* pref_range */

/***************************************************************************/
/* set_int                                                                 */
/*                                                                         */
/* Intersect the set, s, with an arbitrary other element, v.               */
/* Walk s and intersect each of its elements against v.                    */
/* If v is also a set, walk both.                                          */
/***************************************************************************/

static pSexp set_int( pSexp s, pSexp v )
{
  pSexp u = NULL ;              /* walker over s */
  pSexp res = NULL ;            /* result set */
  pSexp elt = NULL ;            /* (needs drop) new element */
  pSexp oset = NULL ;           /* output set */
  int vt = 0 ;                  /* type of the element v */

  if ((s == NULL) || (v == NULL)) return ( NULL ) ;
  if (pull_ts( s ) != ST_SET) return ( NULL ) ;
  vt = pull_ts( v ) ;
  if (s->typ == S_SEXP)
    u = s->u.s ;                /* walk to the start of the set */
  else
    u = s ;
  u = u->term ;                 /* O_SET */
  
  /* next element in the list is the first set element */
  while ((u = u->term) != NULL) {
    switch (elt_compare( u, v )) {
    case T_EITHER: case T_U:
      elt = copy_sexp_elt( u ) ;
      break ;
    case T_V:
      elt = copy_sexp_elt( v ) ;
      break ;
    case T_MANUAL:
      elt = ti_element( u, v ) ;
      break ;
    default: case T_FAIL:
      elt = NULL ;
      break ;
    } /* switch elt_compare */
    add_to_set( &res, elt ) ;
    drop_sexp( elt ) ;          /* done with it, if it wasn't used */
  } /* while */
  /* have any result */
  if (res == NULL) return ( NULL ) ; /* no intersection */
  if (res->term == NULL) return ( res ) ; /* return this one term */
  /* the result is a set -- construct a set */
  if (lit_sexp( &oset, "(1:*3:set)" ) != 0) {
    drop_sexp( res ) ;
    drop_sexp( oset ) ;
    return ( NULL ) ;
  } /* error parsing -- out of memory probably */
  oset->term->term = res ;      /* link in the res list */
  return ( oset ) ;
} /* set_int */

/***************************************************************************/
/* ti_element                                                              */
/*                                                                         */
/* Tag intersection, manually.                                             */
/*                                                                         */
/* Build a single element which is the intersection of these two elts.     */
/* Always return copies of elements, rather than references to them.       */
/*                                                                         */
/* ti_element returns those copies with a refcnt of 1.  If these           */
/* are used as parameters to store_sexp (which is almost always the        */
/* case) then they need their refcnt adjusted.                             */
/***************************************************************************/

static pSexp ti_element( pSexp u, pSexp v )
{
  int ut = pull_typ( u ) ;      /* type of u */
  int vt = pull_typ( v ) ;      /* type of v */

  if ((ut != S_SEXP)||(vt != S_SEXP))
    switch( elt_compare( u, v ) ) {
    case T_EITHER: case T_U: return ( copy_sexp_elt( u ) ) ;
    case T_V: return ( copy_sexp_elt( v ) ) ;
    default: return ( NULL ) ;
    } /* switch elt_compare */

  /* OK -- they are both SEXP nodes. */
  /* See if either is a *-form */
  if ((pull_resword( u ) != O__STAR_) &&
      (pull_resword( v ) != O__STAR_) )
    return ( new_sexp_sexp( tag_intersect( u->u.s, v->u.s ) ) ) ;

  /* At least one is a star form */
  ut = pull_star( u ) ;         /* get the *-form ID */
  vt = pull_star( v ) ;         /* ... */

  if (ut == ST_STAR) return ( copy_sexp_elt( v ) ) ;
  if (vt == ST_STAR) return ( copy_sexp_elt( u ) ) ;
  if (ut == ST_SET) return ( set_int( u, v ) ) ; /* intersect v with set u */
  if (vt == ST_SET) return ( set_int( v, u ) ) ; /* intersect u with set v */

  switch (ut) {
  case ST_RANGE:
    switch (vt) {
    case ST_PREFIX: return ( pref_range( v, u ) ) ;
    case ST_RANGE: return ( range_range( u, v ) ) ;
    default: return ( NULL ) ;
    } /* switch vt */

  case ST_PREFIX:
    switch (vt) {
    case ST_PREFIX: return ( bump_ref( pref_pref( u, v ) ) ) ;
    case ST_RANGE: return ( pref_range( u, v ) ) ;
    default: return ( NULL ) ;
    } /* switch vt */

  default: break ;
  } /* switch pull_ts(u) */
  return ( NULL ) ;
} /* ti_element */

/***************************************************************************/
/* ti_manual                                                               */
/*                                                                         */
/* Tag intersection, manually.                                             */
/*                                                                         */
/* Build a tag S-expression, mixing and matching elements from the         */
/* two tags, u and v.                                                      */
/* u and v are lists, possibly in the middle.                              */
/*                                                                         */
/* ti_manual returns a value whose refcnt counts the copy on the stack.    */
/* Therefore, if it is used with store_sexp, then the refcnt must be       */
/* decremented.                                                            */
/***************************************************************************/

static pSexp ti_manual( pSexp u, pSexp v )
{
  pSexp res = NULL ;            /* result list */

  if (u == NULL) return ( v ) ;
  if (v == NULL) return ( u ) ;
  /* both non-null -- carve a new block */
  res = ti_element( u, v ) ;    /* carve and return the intersection elt */
  if (res == NULL) return ( NULL ) ; /* error condition (no mem) */
  u = u->term ;
  v = v->term ;
  if (u == NULL) store_sexp( &(res->term), v ) ;
  else if (v == NULL) store_sexp( &(res->term), u ) ;
  else {
    store_sexp( &(res->term),
                tag_intersect( u, v ) ) ; /*  the rest of it */
    if (res->term == NULL) {
      drop_sexp( res ) ;
      return ( NULL ) ;
    } /* if */
  } /* else */
  return ( res ) ;              /* and return it */
} /* ti_manual */

/***************************************************************************/
/* tag_intersect                                                           */
/*                                                                         */
/* Compute the intersection of two tags and return the intersection.       */
/* This can be either of the two inputs, or a new, constructed Sexp.       */
/*                                                                         */
/* The pSexp returned by tag_intersect needs to have its refcnt            */
/* bumped if it is stored.  That is, you can call                          */
/*   store_sexp( &loc, tag_intersect( ) )                                  */
/* without adjusting the refcnt.                                           */
/*                                                                         */
/* This also means that it is possible for tag_intersect to return         */
/* a pSexp with a refcnt of 0, implying that if it isn't stored            */
/* it needs to be dropped.                                                 */
/*                                                                         */
/***************************************************************************/

pSexp tag_intersect( pSexp u, pSexp v )
{

  switch (tag_compare( u, v )) {
  case T_U: case T_EITHER: return ( u ) ;
  case T_V: return ( v ) ;
  case T_MANUAL: return ( decr_ref( ti_manual( u, v ) ) ) ;
  case T_FAIL: 
  default: break ;
  } /* switch */

  return ( NULL ) ;             /* intersection failed */
} /* tag_intersect */
