/*-----------------------------------------------------------------------
 * File: ACAPI.C
 *
 * Copyright (c) 1999-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

/*
 * This file contains the functions that are contained in the AC portion
 * of the CSSM exported functions.
 */

#include "precomp.h"



/******************************************************************************
 *
 *  Name:  CSSM_AC_AuthCompute
 *
 *  Description:  This function performs an authorization computation
 *      and returns the results as a group of tuple-certificates.  The
 *      computation is based on the five input values.
 *
 *  Parameters:
 *      ACHandle (input) - The handle that describes the  authorization
 *                         computation module used to perform this function.
 *
 *      BaseAuthorizations (input) - A pointer to a CSSM_TUPLEGROUP
 *                         containing at least one ACL, specifying the
 *                         authorization granted to certain root keys, named
 *                         entities or combinations thereof.
 *
 *      Credentials (input/optional) - A pointer to a CSSM_TUPLEGROUP
 *                         containing a group of certificates.
 *
 *      NumberOfRequestors (input) - The number of entries in the Requestors
 *                         array.
 *
 *      Requestors (input) - A pointer to a list of requestors which define
 *                         the "who" portion of the request.
 *
 *      RequestedAuthorizationPeriod (input/optional) - A list defining a
 *                         validity period or NULL (implying "all time").
 *
 *      RequestedAuthorization (input) - A list defining the "what" portion of
 *                         the authorization being requested.
 *
 *      AuthorizationResult (output) - A CSSM_TUPLEGROUP structure, giving
 *                         the result of the authorization computation.
 *
 *  Returns:
 *      CSSM_NOT_INITIALIZED
 *      CSSM_FUNCTION_NOT_IMPLEMENTED
 *      Internal and addin errors
 *
 *****************************************************************************/

CSSM_RETURN CSSMAPI
CSSM_AC_AuthCompute(CSSM_AC_HANDLE ACHandle,
                    const CSSM_TUPLEGROUP *BaseAuthorizations,
                    const CSSM_TUPLEGROUP *Credentials,
                    uint32 NumberOfRequestors,
                    const CSSM_LIST *Requestors,
                    const CSSM_LIST *RequestedAuthorizationPeriod,
                    const CSSM_LIST *RequestedAuthorization,
                    CSSM_TUPLEGROUP_PTR AuthorizationResult)
{
    CSSM_SPI_AC_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_AC_AuthCompute);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        ACHandle,
                                        CSSM_SERVICE_AC,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->AuthCompute))
        {
            ERR(err = CallBack->AuthCompute(ACHandle,
                                            BaseAuthorizations,
                                            Credentials,
                                            NumberOfRequestors,
                                            Requestors,
                                            RequestedAuthorizationPeriod,
                                            RequestedAuthorization,
                                            AuthorizationResult));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}





/******************************************************************************
 *
 *  Name:  CSSM_AC_PassThrough
 *
 *  Description:  This function allows applications to call authorization
 *      computation module-specific operations that have been exported.
 *
 *  Parameters:
 *      ACHandle (input) - The handle that describes the authorization
 *                         computation module used to perform this function.
 *
 *      TPHandle (input/optional) - The handle that describes the trust policy
 *                         module that can be used by the AC service.
 *
 *      CLHandle (input/optional) - The handle that describes the add-in
 *                         certificate library module that can be used to
 *                         manipulate the certificates.
 *
 *      CCHandle (input/optional) - The handle that describes the context of
 *                         the cryptographic operation.
 *
 *      DBList (input/optional) - A list of handle pairs specifying a data
 *                         storage library module and a data store managed by
 *                         that module.
 *
 *      PassThroughId (input) - An identifier assigned by the AC module to
 *                         indicate the exported function to perform.
 *
 *      InputParams (input/optional) - A pointer to a module,
 *                         implementation-specific structure.
 *
 *      OutputParams (output/optional) - A pointer to a module,
 *                         implementation-specific structure.
 *
 *  Returns:
 *      CSSM_NOT_INITIALIZED
 *      CSSM_FUNCTION_NOT_IMPLEMENTED
 *      Internal and addin errors
 *
 *****************************************************************************/

CSSM_RETURN CSSMAPI
CSSM_AC_PassThrough(CSSM_AC_HANDLE ACHandle,
                    CSSM_TP_HANDLE TPHandle,
                    CSSM_CL_HANDLE CLHandle,
                    CSSM_CC_HANDLE CCHandle,
                    const CSSM_DL_DB_LIST *DBList,
                    uint32 PassThroughId,
                    const void *InputParams,
                    void **OutputParams)
{
    CSSM_SPI_AC_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_AC_PassThrough);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        ACHandle,
                                        CSSM_SERVICE_AC,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->PassThrough))
        {
            ERR(err = CallBack->PassThrough(ACHandle,
                                            TPHandle,
                                            CLHandle,
                                            CCHandle,
                                            DBList,
                                            PassThroughId,
                                            InputParams,
                                            OutputParams));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}


