/*-----------------------------------------------------------------------
 * File: CALLER.C
 *
 * Copyright (c) 1999-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

/*
 * Description: This file includes utility functions for manipulating the
 * caller list.
 */

#include "precomp.h"




static CSSM_RETURN
cssm_SearchCallerIndex(CSSM_PROC_ADDR AddressToLocate,
                       uint32 NumEntries,
                       const cssm_INDEX_ENTRY *EntryIndex,
                       CSSM_LOCK_TYPE LockForCaller,
                       cssm_CALLER_LIST_PTR *Caller);

static CSSM_RETURN
cssm_RemoveCallerIndexEntry(CSSM_PROC_ADDR EntryToRemove,
                            CSSM_BOOL TargetEntryLocked,
                            cssm_CALLER_LIST_INDEX_PTR ListIndex);

static CSSM_RETURN
cssm_AddCallerIndexEntry(CSSM_PROC_ADDR StartAddrIndexKey,
                         const cssm_CALLER_LIST *NewCallerListEntry,
                         cssm_CALLER_LIST_INDEX_PTR ListIndex);



CSSM_RETURN
cssm_GetCallerListIndexLock(CSSM_LOCK_TYPE ltLockType)
{
    CSSM_RETURN err = CSSM_OK;

    /* make sure the lock has been allocated */
    if(NULL != hCallerListIndexLock)
    {
        /* see if this is a reader lock we want */
        if(CSSM_READER_LOCK == ltLockType)
        {
            ERR(err = cssm_SWMRLockWaitToRead(hCallerListIndexLock,
                                              CSSM_INFINITE_WAIT));
        }
        else
        {
            /* make sure we're requesting a writer lock */
            if(CSSM_WRITER_LOCK == ltLockType)
            {
                ERR(err = cssm_SWMRLockWaitToWrite(hCallerListIndexLock,
                                                   CSSM_INFINITE_WAIT));
            }
            else
            {
                /* anything other than NO_LOCK is an error */
                if(CSSM_NO_LOCK != ltLockType)
                {
                    ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
                }
            }
        }
    }
    else
    {
        /* the lock didn't exist for some reason */
        ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
    }

    return err;
}



CSSM_RETURN
cssm_GetCallerListAndMultiLock(CSSM_LOCK_TYPE CallerListIndexLock,
                               cssm_CALLER_LIST_INDEX_PTR *CallerListIndex,
                               CSSM_LOCK_TYPE CallerListLock,
                               cssm_CALLER_LIST_PTR *CallerList)
{
    CSSM_RETURN err = CSSM_OK;
    CSSM_LOCK_TYPE tempCallerListIndexLockType;
    cssm_CALLER_LIST_INDEX_PTR tempCallerListIndex = NULL;

    if(NULL == CallerListIndex)
    {
        ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
    }
    else
    {
        if(CSSM_NO_LOCK == CallerListIndexLock)
        {
            tempCallerListIndexLockType = CSSM_READER_LOCK;
        }
        else
        {
            tempCallerListIndexLockType = CallerListIndexLock;
        }

        ERR(err = cssm_GetCallerListIndexLock(tempCallerListIndexLockType));

        if(CSSM_OK == err)
        {
            tempCallerListIndex = hCallerListIndex;

            /* make sure the lock has been allocated */
            if(NULL != tempCallerListIndex->CallerListLock)
            {
                /* see if this is a reader lock we want */
                if(CSSM_READER_LOCK == CallerListLock)
                {
                    ERR(err = cssm_SWMRLockWaitToRead(tempCallerListIndex->CallerListLock,
                                                      CSSM_INFINITE_WAIT));
                }
                else
                {
                    /* make sure we're requesting a writer lock */
                    if(CSSM_WRITER_LOCK == CallerListLock)
                    {
                        ERR(err = cssm_SWMRLockWaitToWrite(tempCallerListIndex->CallerListLock,
                                                           CSSM_INFINITE_WAIT));
                    }
                    else
                    {
                        /* anything other than NO_LOCK is an error */
                        if(CSSM_NO_LOCK != CallerListLock)
                        {
                            ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
                        }
                    }
                }
            }
            else
            {
                /* the lock didn't exist for some reason */
                ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
            }

            /* if the caller wanted the caller list, we set it here */
            if(CSSM_OK == err)
            {
                if(NULL != CallerList)
                {
                    (*CallerList) = tempCallerListIndex->CallerList;
                }
            }

            /* if the caller wanted the caller list index, we set it here */
            if(CSSM_OK == err)
            {
                if(NULL != CallerListIndex)
                {
                    (*CallerListIndex) = tempCallerListIndex;
                }
            }

            /* if the caller didn't want the global lock set, we release it
               here
            */
            if((CSSM_NO_LOCK == CallerListIndexLock) ||
               (CSSM_OK != err))
            {
                cssm_ReleaseCallerListIndex(CSSM_READER_LOCK);
            }
        }
    }

    return err;
}



CSSM_RETURN
cssm_FindCallerListEntryAndMultiLock(CSSM_PROC_ADDR CallerAddress,
                                     CSSM_LOCK_TYPE CallerListIndexLock,
                                     cssm_CALLER_LIST_INDEX_PTR *CallerListIndex,
                                     CSSM_LOCK_TYPE CallerListLock,
                                     cssm_CALLER_LIST_PTR *CallerList,
                                     CSSM_LOCK_TYPE CallerListEntryLock,
                                     cssm_CALLER_LIST_ENTRY_PTR *CallerListEntry)
{
    CSSM_RETURN err = CSSM_OK;
    cssm_CALLER_LIST_INDEX_PTR tempCallerListIndex = NULL;
    CSSM_LOCK_TYPE tempCallerListLock;
    CSSM_LOCK_TYPE tempListIndexLock;
    CSSM_LOCK_TYPE tempCallerListEntryLock;
    cssm_CALLER_LIST_PTR tempCallerList = NULL;


    /* make sure we have a valid caller address to look for - quick check to 
       get rid of obviously incorrect input
    */
    if(NULL == CallerAddress)
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        /* We need a locks on all levels to do our search - if the
           caller doesn't have one set, we'll set one temporarily.
        */
        if(CSSM_NO_LOCK == CallerListIndexLock)
        {
            tempListIndexLock = CSSM_READER_LOCK;
        }
        else
        {
            tempListIndexLock = CallerListIndexLock;
        }

        if(CSSM_NO_LOCK == CallerListLock)
        {
            tempCallerListLock = CSSM_READER_LOCK;
        }
        else
        {
            tempCallerListLock = CallerListLock;
        }

        if(CSSM_NO_LOCK == CallerListEntryLock)
        {
            tempCallerListEntryLock = CSSM_READER_LOCK;
        }
        else
        {
            tempCallerListEntryLock = CallerListEntryLock;
        }

        /* Get the caller list index so we can search it.  We don't get the
           caller list here because we essentially get it from the search
           we do right after this.  In the search, we get the right spot in
           the caller list instead of just the head of the list.
        */
        ERR(err = cssm_GetCallerListAndMultiLock(tempListIndexLock,
                                                 &tempCallerListIndex,
                                                 tempCallerListLock,
                                                 NULL));
        if(CSSM_OK == err)
        {
            ERR(err = cssm_SearchCallerIndex(CallerAddress,
                                             tempCallerListIndex->NumEntries,
                                             tempCallerListIndex->IndexArray,
                                             tempCallerListEntryLock,
                                             &tempCallerList));

            /* We now release locks that weren't desired and assign return
               values for the desired components.
            */
            if(CSSM_OK == err)
            {
                if(NULL != CallerListEntry)
                {
                    (*CallerListEntry) = tempCallerList->ListEntry;
                }

                if(CSSM_NO_LOCK == CallerListEntryLock)
                {
                    cssm_ReleaseCallerListEntry(tempCallerList,
                                                tempCallerListEntryLock);
                }

                if(NULL != CallerList)
                {
                    (*CallerList) = tempCallerList;
                }

                if(NULL != CallerListIndex)
                {
                    (*CallerListIndex) = tempCallerListIndex;
                }
            }

            /* These locks have been held whether the search succeeded or not,
               so we must always free them in fail cases.
            */
            if((CSSM_NO_LOCK == CallerListLock) ||
               (CSSM_OK != err))
            {
                cssm_ReleaseCallerList(tempCallerListIndex,
                                       tempCallerListLock);
            }

            if((CSSM_NO_LOCK == CallerListIndexLock) ||
               (CSSM_OK != err))
            {
                cssm_ReleaseCallerListIndex(tempListIndexLock);
            }
        }
    }

    return err;
}



void
cssm_ReleaseCallerListEntry(cssm_CALLER_LIST_PTR CallerList,
                            CSSM_LOCK_TYPE ltLockType)
{
    if((NULL != CallerList) &&
       (NULL != CallerList->EntryLock))
    {
        if(CSSM_READER_LOCK == ltLockType)
        {
            cssm_SWMRLockDoneReading(CallerList->EntryLock);
        }
        else
        {
            if(CSSM_WRITER_LOCK == ltLockType)
            {
                cssm_SWMRLockDoneWriting(CallerList->EntryLock);
            }
        }
    }

}



void
cssm_ReleaseCallerList(cssm_CALLER_LIST_INDEX_PTR CallerListIndex,
                       CSSM_LOCK_TYPE ltLockType)
{
    if((NULL != CallerListIndex) &&
       (NULL != CallerListIndex->CallerListLock))
    {
        if(CSSM_READER_LOCK == ltLockType)
        {
            cssm_SWMRLockDoneReading(CallerListIndex->CallerListLock);
        }
        else
        {
            if(CSSM_WRITER_LOCK == ltLockType)
            {
                cssm_SWMRLockDoneWriting(CallerListIndex->CallerListLock);
            }
        }
    }
}



void
cssm_ReleaseCallerListIndex(CSSM_LOCK_TYPE ltLockType)
{
    if(NULL != hCallerListIndexLock)
    {
        if(CSSM_READER_LOCK == ltLockType)
        {
            cssm_SWMRLockDoneReading(hCallerListIndexLock);
        }
        else
        {
            if(CSSM_WRITER_LOCK == ltLockType)
            {
                cssm_SWMRLockDoneWriting(hCallerListIndexLock);
            }
        }
    }
}



CSSM_RETURN
cssm_InsertCallerListEntry(cssm_CALLER_LIST_ENTRY_PTR NewEntry)
{
    CSSM_RETURN err = CSSM_OK;
    cssm_CALLER_LIST_PTR NewListNode = NULL;
    cssm_CALLER_LIST_INDEX_PTR LocalListIndex = NULL;

    /* make sure the new entry is valid */
    if(NULL == NewEntry)
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        /* allocate memory for the new list node */
        NewListNode = (cssm_CALLER_LIST_PTR)local_malloc(sizeof(cssm_CALLER_LIST),
                                                         NULL);
        if(NULL == NewListNode)
        {
            ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
        }
        else
        {
            /* allocate a lock for the new entry */
            if(CSSM_OK == ERR(err = cssm_SWMRInit(&NewListNode->EntryLock)))
            {
                /* let the new entry dangle off the new node */
                NewListNode->ListEntry = NewEntry;

                /* get the global caller list so we can insert the new entry */
                if(CSSM_OK == ERR(err = cssm_GetCallerListAndMultiLock(CSSM_READER_LOCK,
                                                                       &LocalListIndex,
                                                                       CSSM_WRITER_LOCK,
                                                                       NULL)))
                {
                    /* In addition to adding a new entry in the index array,
                       this function will update LocalListIndex->NumEntries
                    */
                    ERR(err = cssm_AddCallerIndexEntry(NewEntry->ModuleStartAddr,
                                                       NewListNode,
                                                       LocalListIndex));

                    if(CSSM_OK == err)
                    {
                        /* insert at the head of the list */
                        NewListNode->Next = LocalListIndex->CallerList;
                        LocalListIndex->CallerList = NewListNode;
                    }

                    /* release the locks we got for inserting the new entry */
                    cssm_ReleaseCallerList(LocalListIndex,
                                           CSSM_WRITER_LOCK);
                    cssm_ReleaseCallerListIndex(CSSM_READER_LOCK);
                }

                /* free the allocated lock if there's been a problem */
                if(CSSM_OK != err)
                {
                    ERR(cssm_SWMRDelete(NewListNode->EntryLock));
                }
            }
            else
            {
                /* we mask out whatever error may have returned from the init
                   function - we don't want to reveal this to the application
                */
                ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
            }

            /* free allocated memory if there's been an error */
            if(CSSM_OK != err)
            {
                local_free(NewListNode,
                           NULL);
            }
        }
    }

    return err;
}



CSSM_RETURN
cssm_RemoveCallerListEntry(CSSM_PROC_ADDR CallerAddress,
                           cssm_CALLER_LIST_ENTRY_PTR *RemovedEntry)
{
    CSSM_RETURN err = CSSM_OK;
    cssm_CALLER_LIST_PTR ListNodeToRemove = NULL;
    cssm_CALLER_LIST_PTR PrevListNode = NULL;
    cssm_CALLER_LIST_INDEX_PTR LocalListIndex = NULL;

#ifdef ISL_INCLUDED
    uint32 NumberOfCurrentComponent;
    cssm_MODULE_COMPONENTS_PTR tempComponentPtr = NULL;
#endif

    CSSM_BOOL EntryFound = CSSM_FALSE;

    /* make sure the search address and return entry are valid */
    if((NULL == CallerAddress) ||
       (NULL == RemovedEntry))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        ERR(err = cssm_GetCallerListAndMultiLock(CSSM_READER_LOCK,
                                                 &LocalListIndex,
                                                 CSSM_WRITER_LOCK,
                                                 NULL));
        if(CSSM_OK == err)
        {
            /* Decrements numentries*/
            ERR(err = cssm_RemoveCallerIndexEntry(CallerAddress,
                                                  CSSM_FALSE,
                                                  LocalListIndex));

            if(CSSM_OK == err)
            {
                ListNodeToRemove = LocalListIndex->CallerList;

                /* now we search the list looking for the entry to remove */
                while((CSSM_FALSE == EntryFound) &&
                      (NULL != ListNodeToRemove))
                {
                    /* do the address check */ 
                    /* #ifdef VMS apply cast */
                    if(((ulong)CallerAddress >= (ulong)ListNodeToRemove->ListEntry->ModuleStartAddr) &&
                       ((ulong)CallerAddress <= (ulong)ListNodeToRemove->ListEntry->ModuleEndAddr))
                    {
                        /* We have the right record, if there is one.  Our final
                           check will be to compare the address against the EISL
                           value we have.
                        */
#ifdef ISL_INCLUDED
                        /* start a temp pointer to go looking through the module
                           components
                        */
                        tempComponentPtr = ListNodeToRemove->ListEntry->ModuleComponents;
                        NumberOfCurrentComponent = 0;

                        /* loop through the components - we stop when either there
                           aren't any components left or we've found the correct
                           component
                        */
                        while((NULL != tempComponentPtr) &&
                              (NumberOfCurrentComponent < tempComponentPtr->NumComponents) &&
                              (ISL_OK != EISL_CheckAddressWithinModule(tempComponentPtr->VerifiedModule[NumberOfCurrentComponent],
                                                                       (ISL_FUNCTION_PTR)CallerAddress)))
                        {
                            tempComponentPtr = tempComponentPtr->Next;
                            NumberOfCurrentComponent++;
                        }

                        /* if the component pointer is NULL, then we didn't find
                           the component with the EISL check
                        */
                        if(NULL == tempComponentPtr)
                        {
                            ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                        }
#endif
                        if(CSSM_OK == err)
                        {
                            EntryFound = CSSM_TRUE;
                        }
                    }
                    else
                    {
                        /* bump things up to the next node */
                        PrevListNode = ListNodeToRemove;
                        ListNodeToRemove = ListNodeToRemove->Next;
                    }
                }

                /* see if we found anything - we've found anything if there's
                   anything to find
                */
                if(NULL == ListNodeToRemove)
                {
                    (*RemovedEntry) = NULL;
                    ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
                else
                {
                    /* check if we're removing the first entry */
                    if(NULL == PrevListNode)
                    {
                        LocalListIndex->CallerList = ListNodeToRemove->Next;
                        PrevListNode = LocalListIndex->CallerList;
                    }
                    else
                    {
                        PrevListNode->Next = ListNodeToRemove->Next;
                    }

                    /* write lock the list entry and make sure no one else is
                       using the entry
                    */
                    if(CSSM_OK == cssm_SWMRLockWaitToWrite(ListNodeToRemove->EntryLock,
                                                           CSSM_INFINITE_WAIT))
                    {
                        /* Double check that we're removing the correct node.
                           It's doubtful that this information has changed, but if
                           it has then we need to re-insert this entry.
                        */
                        /* #ifdef VMS apply cast */
                        if(((ulong)CallerAddress < (ulong)ListNodeToRemove->ListEntry->ModuleStartAddr) ||
                           ((ulong)CallerAddress > (ulong)ListNodeToRemove->ListEntry->ModuleEndAddr))
                        {
                            /* now we need to re-add this entry */
                            PrevListNode->Next = ListNodeToRemove;

                            /* release the write lock we're holding on the entry */
                            cssm_SWMRLockDoneWriting(ListNodeToRemove->EntryLock);

                            ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                        }
                        else
                        {
                            /* set the entry for return */
                            (*RemovedEntry) = ListNodeToRemove->ListEntry;

                            /* now we release our node lock and de-allocate the
                               memory used for the list node
                            */
                            cssm_SWMRLockDoneWriting(ListNodeToRemove->EntryLock);
                            cssm_SWMRDelete(ListNodeToRemove->EntryLock);

                            local_free(ListNodeToRemove,
                                       NULL);

                        }
                    }
                    else
                    {
                        ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
                    }
                }
            }

            /* release the caller list and caller list index */
            cssm_ReleaseCallerList(LocalListIndex,
                                   CSSM_WRITER_LOCK);
            cssm_ReleaseCallerListIndex(CSSM_READER_LOCK);
        }
    }

    return err;
}



/* should only be called on "1st" Init */
CSSM_RETURN
cssm_CreateGlobalCallerList()
{
    CSSM_RETURN err = CSSM_OK;

    if(NULL == hCallerListIndexLock)
    {
        ERR(err = cssm_SWMRInit(&hCallerListIndexLock));
    }

    if((NULL == hCallerListIndex) &&
       (CSSM_OK == err))
    {
        hCallerListIndex = (cssm_CALLER_LIST_INDEX_PTR)local_calloc(sizeof(cssm_CALLER_LIST_INDEX),
                                                                    1,
                                                                    NULL);
        if(NULL == hCallerListIndex)
        {
            ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
        }
        else
        {
            ERR(err = cssm_SWMRInit(&hCallerListIndex->CallerListLock));

            if(CSSM_OK != err)
            {
                /* free the global structure and the global lock */
                cssm_SWMRDelete(hCallerListIndexLock);
                local_free(hCallerListIndex,
                           NULL);
                ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
            }
        }
    }

    return err;
}



/* Must be called only after module list is destroyed or a risk of deadlock
   exists.  Should only be called after "last" Term.
*/
CSSM_RETURN
cssm_DestroyGlobalCallerList()
{
    CSSM_RETURN err = CSSM_OK;
    cssm_CALLER_LIST_PTR tempCallerListPtr;
    cssm_CALLER_LIST_PTR tempNextListNode;
    cssm_CALLER_LIST_ENTRY_PTR tempListEntryPtr;


    /* make sure there is a lock to acquire */
    if(NULL != hCallerListIndexLock)
    {
        /* make sure there is a list we need to destroy */
        if(NULL != hCallerListIndex)
        {
            ERR(err = cssm_SWMRLockWaitToWrite(hCallerListIndexLock,
                                               CSSM_INFINITE_WAIT));

            if(CSSM_OK == err)
            {
                ERR(err = cssm_SWMRLockWaitToWrite(hCallerListIndex->CallerListLock,
                                                   CSSM_INFINITE_WAIT));
                if(CSSM_OK == err)
                {
                    tempCallerListPtr = hCallerListIndex->CallerList;

                    while(NULL != tempCallerListPtr)
                    {
                        ERR(cssm_SWMRLockWaitToWrite(tempCallerListPtr->EntryLock,
                                                     CSSM_INFINITE_WAIT));

                        /* move a pointer to the next list entry */
                        tempNextListNode = tempCallerListPtr->Next;

                        /* set the entry ptr so we can delete the entry */
                        tempListEntryPtr = tempCallerListPtr->ListEntry;

                        /* now we delete the contents of the list entry */
                        ERR(cssm_DestroyCallerListEntry(&tempListEntryPtr));

                        /* unlock and delete the list node */
                        cssm_SWMRLockDoneWriting(tempCallerListPtr->EntryLock);
                        cssm_SWMRDelete(tempCallerListPtr->EntryLock);

                        local_free(tempCallerListPtr,
                                   NULL);

                        /* move to next list entry */
                        tempCallerListPtr = tempNextListNode;
                    }

                    /* release the caller list lock - we're about to
                       delete it
                    */
                    cssm_SWMRLockDoneWriting(hCallerListIndex->CallerListLock);
                }

                /* we're ready to destroy the lock - it's in an unlocked state
                   because either the locking failed or the lock was unlocked
                   inside the above "if" statement.
                */
                cssm_SWMRDelete(hCallerListIndex->CallerListLock);

                /* unlock the global lock - we're about to delete it */
                cssm_SWMRLockDoneWriting(hCallerListIndexLock);
            }

            /* we'll free the index search list now */
            if(NULL != hCallerListIndex->IndexArray)
            {
                local_free(hCallerListIndex->IndexArray,
                           NULL);
            }

            /* we're ready to destroy the lock - it's in an unlocked state
               because either the locking failed or the lock was unlocked
               inside the above "if" statement.
            */
            cssm_SWMRDelete(hCallerListIndexLock);
            hCallerListIndexLock = NULL;
            local_free( hCallerListIndex, NULL);
            hCallerListIndex = NULL;
        }

    }

    return err;
}



CSSM_RETURN
cssm_DestroyCallerListEntry(cssm_CALLER_LIST_ENTRY_PTR *Entry)
{
    CSSM_RETURN err = CSSM_OK;
    cssm_VENDOR_PRIVILEGES_PTR tempVendorPrivs;
    cssm_MODULE_COMPONENTS_PTR tempModuleComponents;


    if(NULL == Entry)
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        /* destroy TOG privileges, if there are any */
        if(NULL != (*Entry)->TOGPrivs)
        {
            local_free((*Entry)->TOGPrivs,
                       NULL);
        }

        /* destroy USEE tags, if there are any */
        if(NULL != (*Entry)->USEETags)
        {
            local_free((*Entry)->USEETags,
                       NULL);
        }

        /* now we delete the vendor privileges */
        tempVendorPrivs = (*Entry)->VendorPrivs;
        while(NULL != tempVendorPrivs)
        {
            (*Entry)->VendorPrivs = tempVendorPrivs->Next;

            if(NULL != tempVendorPrivs->VendorPrivs)
            {
                local_free(tempVendorPrivs->VendorPrivs,
                           NULL);
            }

            local_free(tempVendorPrivs,
                       NULL);

            tempVendorPrivs = (*Entry)->VendorPrivs;
        }

        /* now we delete the module components */
        tempModuleComponents = (*Entry)->ModuleComponents;

        while(NULL != tempModuleComponents)
        {
            (*Entry)->ModuleComponents = tempModuleComponents->Next;

#ifdef ISL_INCLUDED
            if(NULL != tempModuleComponents->VerifiedModule)
            {
                EISL_RecycleVerifiedModuleCredentials(tempModuleComponents->VerifiedModule[0]);
                local_free(tempModuleComponents->VerifiedModule,
                           NULL);
            }
#endif
            local_free(tempModuleComponents,
                       NULL);

            tempModuleComponents = (*Entry)->ModuleComponents;
        }

        local_free((*Entry),
                   NULL);
        (*Entry) = NULL;
    }

    return err;
}



CSSM_RETURN
cssm_FindCallerListEntryByGuidAndMultiLock(const CSSM_GUID *GuidToFind,
                                           CSSM_LOCK_TYPE CallerListIndexLock,
                                           cssm_CALLER_LIST_INDEX_PTR *CallerListIndex,
                                           CSSM_LOCK_TYPE CallerListLock,
                                           cssm_CALLER_LIST_PTR *CallerList,
                                           CSSM_LOCK_TYPE CallerListEntryLock,
                                           cssm_CALLER_LIST_ENTRY_PTR *CallerListEntry)
{
    CSSM_RETURN err = CSSM_OK;
    cssm_CALLER_LIST_INDEX_PTR tempCallerListIndex = NULL;
    CSSM_LOCK_TYPE tempCallerListLock;
    cssm_CALLER_LIST_PTR tempCallerList = NULL;

    CSSM_BOOL EntryFound = CSSM_FALSE;


    /* make sure we have a valid caller address to look for - quick check to 
       get rid of obviously incorrect input
    */
    if(NULL == GuidToFind)
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        /* we need a lock on the caller list to search - if the caller doesn't
           have one set, we'll set one temporarily
        */
        if(CSSM_NO_LOCK == CallerListLock)
        {
            tempCallerListLock = CSSM_READER_LOCK;
        }
        else
        {
            tempCallerListLock = CSSM_READER_LOCK;
        }

        /* get the caller list so we can search it */
        ERR(err = cssm_GetCallerListAndMultiLock(CallerListIndexLock,
                                                 &tempCallerListIndex,
                                                 tempCallerListLock,
                                                 &tempCallerList));
        if(CSSM_OK == err)
        {
            /* search the caller list for the correct entry */
            while((CSSM_FALSE == EntryFound) &&
                  (NULL != tempCallerList))
            {
                if(CSSM_TRUE == CSSM_CompareGuids((*GuidToFind),
                                                  tempCallerList->ListEntry->CallerGuid))
                {
                    EntryFound = CSSM_TRUE;
                }
                else
                {
                    tempCallerList = tempCallerList->Next;
                }
            }

            /* if we found the entry, now we need to lock it */
            if(NULL == tempCallerList)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
            else
            {
                /* see if this is a reader lock we want */
                if(CSSM_READER_LOCK == CallerListEntryLock)
                {
                    ERR(err = cssm_SWMRLockWaitToRead(tempCallerList->EntryLock,
                                                      CSSM_INFINITE_WAIT));
                }
                else
                {
                    /* make sure we're requesting a writer lock */
                    if(CSSM_WRITER_LOCK == CallerListEntryLock)
                    {
                        ERR(err = cssm_SWMRLockWaitToWrite(tempCallerList->EntryLock,
                                                           CSSM_INFINITE_WAIT));
                    }
                    else
                    {
                        /* anything other than NO_LOCK is an error */
                        if(CSSM_NO_LOCK != CallerListEntryLock)
                        {
                            ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
                        }
                    }
                }

                /* set the entry we found for return if everything is
                   ready for return to the caller
                */
                if((CSSM_OK == err) &&
                   (NULL != CallerListEntry))
                {
                    (*CallerListEntry) = tempCallerList->ListEntry;
                }
            }

            /* now we release the caller list lock if it wasn't desired */
            if(CSSM_NO_LOCK == CallerListLock)
            {
                cssm_ReleaseCallerList(tempCallerListIndex,
                                       CSSM_READER_LOCK);
            }
        }
    }

    /* if the user asked for the caller list index or the caller list record,
       we set those output values here
    */
    if(CSSM_OK == err)
    {
        if(NULL != CallerListIndex)
        {
            (*CallerListIndex) = tempCallerListIndex;
        }

        if(NULL != CallerList)
        {
            (*CallerList) = tempCallerList;
        }
    }

    return err;
}



/* It is intended that cssm_FindPotentialCaller only be called by other static
   functions.
*/
static CSSM_RETURN
cssm_FindPotentialCaller(CSSM_PROC_ADDR TargetAddress,
                         uint32 NumEntries,
                         const cssm_INDEX_ENTRY *EntryIndex,
                         uint32 *CandidateIndexEntry);



/* cssm_SearchCallerIndex
 *
 * This function searches the caller list index to find a caller.  This is
 * done by a "bucket" binary search.  Once the "bucket" entry is identified,
 * a more complete range check is performed to make sure we have the right
 * caller.
 *
 * Valid values for LockForCaller are CSSM_READER_LOCK and CSSM_WRITER_LOCK.
 * CSSM_NO_LOCK is not allowed because the range check requires a lock.
 *
 * For correct functionality, EntryIndex should already have either a reader
 * or a writer lock applied when this function is called.
 */

static CSSM_RETURN
cssm_SearchCallerIndex(CSSM_PROC_ADDR AddressToLocate,
                       uint32 NumEntries,
                       const cssm_INDEX_ENTRY *EntryIndex,
                       CSSM_LOCK_TYPE LockForCaller,
                       cssm_CALLER_LIST_PTR *Caller)
{
    CSSM_RETURN err = CSSM_OK;
    cssm_CALLER_LIST_ENTRY_PTR PotentialEntry;
    uint32 CandidateIndexEntry;

    /* check parameters for pointer validity */
    if((NULL == AddressToLocate) ||
       (NULL == EntryIndex) ||
       (NULL == Caller))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        /* initialize for failure, update on success */
        *Caller = NULL;

        if(0 == NumEntries)
        {
            ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
        }
        else
        {
            ERR(err = cssm_FindPotentialCaller(AddressToLocate,
                                               NumEntries,
                                               EntryIndex,
                                               &CandidateIndexEntry));

            if(CSSM_OK == err)
            {
                /* make sure the lock type is either reader or writer */
                if((CSSM_READER_LOCK == LockForCaller) ||
                   (CSSM_WRITER_LOCK == LockForCaller))
                {
                    /* acquire a lock and do final checking on our candidate
                       caller
                    */
                    if(CSSM_READER_LOCK == LockForCaller)
                    {
                        ERR(err = cssm_SWMRLockWaitToRead(EntryIndex[CandidateIndexEntry].CallerList->EntryLock,
                                                          CSSM_INFINITE_WAIT));
                    }
                    else
                    {
                        ERR(err = cssm_SWMRLockWaitToWrite(EntryIndex[CandidateIndexEntry].CallerList->EntryLock,
                                                           CSSM_INFINITE_WAIT));
                    }

                    if(CSSM_OK == err)
                    {
                        PotentialEntry = EntryIndex[CandidateIndexEntry].CallerList->ListEntry;
                        /* #ifdef VMS apply cast */
                        if(((ulong)PotentialEntry->ModuleStartAddr <= (ulong)AddressToLocate) &&
                           ((ulong)PotentialEntry->ModuleEndAddr >= (ulong)AddressToLocate))
                        {
                            /* we have the correct entry */
                            *Caller = EntryIndex[CandidateIndexEntry].CallerList;
                        }
                        else
                        {
                            /* this entry isn't the correct one, so we release our
                               lock and return an error
                            */
                            if(CSSM_READER_LOCK == LockForCaller)
                            {
                                ERR(cssm_SWMRLockDoneReading(EntryIndex[CandidateIndexEntry].CallerList->EntryLock));
                            }
                            else
                            {
                                ERR(cssm_SWMRLockDoneWriting(EntryIndex[CandidateIndexEntry].CallerList->EntryLock));
                            }

                            ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                        }
                    }
                }
                else
                {
                    ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
                }
            }
        }
    }

    return err;
}



/* cssm_RemoveCallerIndexEntry
 *
 * This function removes the caller entry identified by EntryToRemove from
 * the caller index.  It does not remove the caller entry from the caller
 * list.  Thus it is still possible to conduct a search of the caller list
 * and find the caller, though current search functions do not support this.
 *
 * TargetEntryLocked should indicate whether the entry we're removing has
 * an intentional lock applied.  If the value is false, a reader lock will
 * be acquired for checking the correct entry before removal.
 *
 * For correct functionality, ListIndex should already have a writer lock
 * applied when this function is called.  
 */

static CSSM_RETURN
cssm_RemoveCallerIndexEntry(CSSM_PROC_ADDR EntryToRemove,
                            CSSM_BOOL TargetEntryLocked,
                            cssm_CALLER_LIST_INDEX_PTR ListIndex)
{
    CSSM_RETURN err = CSSM_OK;

    cssm_CALLER_LIST_ENTRY_PTR PotentialEntry;
    cssm_CALLER_LIST_PTR CallerList;
    uint32 CandidateIndexEntry;
    uint32 i;

    /* check parameters for pointer validity */
    if((NULL == EntryToRemove) ||
       (NULL == ListIndex))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        ERR(err = cssm_FindPotentialCaller(EntryToRemove,
                                           ListIndex->NumEntries,
                                           ListIndex->IndexArray,
                                           &CandidateIndexEntry));

        if(CSSM_OK == err)
        {
            if(CSSM_FALSE == TargetEntryLocked)
            {
                /* acquire a reader lock and do final checking on our candidate
                   caller
                */
                ERR(err = cssm_SWMRLockWaitToRead(ListIndex->IndexArray[CandidateIndexEntry].CallerList->EntryLock,
                                                  CSSM_INFINITE_WAIT));
            }

            if(CSSM_OK == err)
            {
                PotentialEntry = ListIndex->IndexArray[CandidateIndexEntry].CallerList->ListEntry;

                /* We have the correct entry, so we need to identify the
                   place in memory of the list entry lock.  This is done
                   in case we have to release the lock before returning.
                */
                CallerList = ListIndex->IndexArray[CandidateIndexEntry].CallerList;

                /* #ifdef VMS apply cast */
                if(( (ulong)PotentialEntry->ModuleStartAddr <= (ulong)EntryToRemove) &&
                   ( (ulong)PotentialEntry->ModuleEndAddr >= (ulong)EntryToRemove))
                {
                    /* We have the correct entry, so now we adjust the other
                       index entries.
                    */
                    for(i = CandidateIndexEntry; i < (ListIndex->NumEntries-1); i++)
                    {
                        ListIndex->IndexArray[i] = ListIndex->IndexArray[i+1];
                    }

                    /* Now we fix up the number of entries and the memory used
                       for the index
                    */
                    ListIndex->NumEntries--;

                    if(0 == ListIndex->NumEntries)
                    {
                        local_free(ListIndex->IndexArray,
                                   NULL);
                        ListIndex->IndexArray = NULL;
                    }
                    else
                    {
                        ListIndex->IndexArray = local_realloc(ListIndex->IndexArray,
                                                              (ListIndex->NumEntries * sizeof(cssm_INDEX_ENTRY)),
                                                              NULL);
                        if(NULL == ListIndex->IndexArray)
                        {
                            ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                        }
                    }
                }
                else
                {
                    /* We didn't find the correct entry to remove, so it's not
                       present.
                    */
                    ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }

                if(CSSM_FALSE == TargetEntryLocked)
                {
                    ERR(cssm_SWMRLockDoneReading(CallerList->EntryLock));
                }
            }
        }
    }

    return err;
}



/* cssm_AddCallerIndexEntry
 *
 * This function adds the caller entry identified by EntryToAdd to
 * the caller index.  It does not add the caller entry from the caller
 * list.
 *
 * The StartAddrIndexKey should be the start address of the caller list entry.
 *
 * For correct functionality, ListIndex should already have a writer lock
 * applied when this function is called.  NewCallerListEntry should have at
 * least a reader lock applied.
 */

static CSSM_RETURN
cssm_AddCallerIndexEntry(CSSM_PROC_ADDR StartAddrIndexKey,
                         const cssm_CALLER_LIST *NewCallerListEntry,
                         cssm_CALLER_LIST_INDEX_PTR ListIndex)
{
    CSSM_RETURN err = CSSM_OK;

    uint32 CandidateIndexEntry;
    uint32 i;
    uint32 EndCopy;

    /* check parameters for pointer validity */
    if((NULL == StartAddrIndexKey) ||
       (NULL == NewCallerListEntry) ||
       (NULL == ListIndex))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        /* see if the entry is already in the index */
        ERR(err = cssm_FindPotentialCaller(StartAddrIndexKey,
                                           ListIndex->NumEntries,
                                           ListIndex->IndexArray,
                                           &CandidateIndexEntry));

        if((CSSM_OK == err) ||
           (CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND == err))
        {
            if(0 != ListIndex->NumEntries)
            {
                /* now we check if we actually need to add this entry or if it's
                   already present
                */
                if((CSSM_OK != err) ||
                   /* #ifdef VMS apply cast */
                   ((ulong)StartAddrIndexKey < (ulong)ListIndex->IndexArray[CandidateIndexEntry].CallerList->ListEntry->ModuleStartAddr) ||
                   ((ulong)StartAddrIndexKey > (ulong)ListIndex->IndexArray[CandidateIndexEntry].CallerList->ListEntry->ModuleEndAddr))
                {
                    /* clear the error code since we know we're not in an error
                       condition
                    */
                    err = CSSM_OK;

                    /* Since we hadn't found the entry, we know this key hasn't been
                       used.  So we'll make a new index entry.
                    */

                    /* allocate an array with one more entry */
                    ListIndex->IndexArray = local_realloc(ListIndex->IndexArray,
                                                          ((ListIndex->NumEntries+1) * sizeof(cssm_INDEX_ENTRY)),
                                                          NULL);
                    if(NULL != ListIndex->IndexArray)
                    {
                        /* If the start address is less than the start address
                           for the candidate entry, we know that the new entry
                           should come before the candidate, so we copy one 
                           extra entry.  This should only occur if the new
                           entry comes before the currently first entry.
                        */
                        /* #ifdef VMS apply cast */
                        if((ulong)StartAddrIndexKey < (ulong)ListIndex->IndexArray[CandidateIndexEntry].CallerList->ListEntry->ModuleStartAddr)
                        {
                            EndCopy = CandidateIndexEntry;
                        }
                        else
                        {
                            EndCopy = CandidateIndexEntry +1;
                        }

                        for(i = ListIndex->NumEntries; i > EndCopy; i--)
                        {
                            ListIndex->IndexArray[i] = ListIndex->IndexArray[i-1];
                        }

                        ListIndex->IndexArray[i].IndexKey = StartAddrIndexKey;
                        ListIndex->IndexArray[i].CallerList = (cssm_CALLER_LIST_PTR)NewCallerListEntry;
                        ListIndex->NumEntries++;
                    }
                    else
                    {
                        ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                    }
                }
            }
            else
            {
                err = CSSM_OK;

                ListIndex->IndexArray = local_calloc(1,
                                                     sizeof(cssm_INDEX_ENTRY),
                                                     NULL);
                if(NULL != ListIndex->IndexArray)
                {
                    ListIndex->IndexArray[0].IndexKey = StartAddrIndexKey;
                    ListIndex->IndexArray[0].CallerList = (cssm_CALLER_LIST_PTR)NewCallerListEntry;
                    ListIndex->NumEntries++;
                }
                else
                {
                    ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                }
            }
        }
    }

    return err;
}



/* cssm_FindPotentialCaller
 *
 * This function searches the caller list index to find a caller.  This is
 * done by a "bucket" binary search.  Once the "bucket" entry is identified,
 * the correct bucket is returned to the caller.  This does NOT guarrantee
 * that the bucket contains the caller entry, only that if the caller is known
 * then this is the caller.  This is done to avoid the necessity of acquiring
 * a lock on the target bucket, thus avoiding deadlock scenarios.
 *
 * For correct functionality, EntryIndex should already have either a reader
 * or a writer lock applied when this function is called.
 */

static CSSM_RETURN
cssm_FindPotentialCaller(CSSM_PROC_ADDR TargetAddress,
                         uint32 NumEntries,
                         const cssm_INDEX_ENTRY *EntryIndex,
                         uint32 *CandidateIndexEntry)
{
    CSSM_RETURN err = CSSM_OK;
    CSSM_BOOL CandidateFound = CSSM_FALSE;
    uint32 CurrentMiddlePoint;
    uint32 LastLowPoint;
    uint32 LastHighPoint;
    uint32 PointDiff;

    /* check parameters for consistency and validity */
    if((NULL == TargetAddress) ||
       ((0 != NumEntries) && (NULL == EntryIndex)) ||
       (NULL == CandidateIndexEntry))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        /* initialize the return position */
        *CandidateIndexEntry = 0;

        if(0 == NumEntries)
        {
            ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
        }
    }

    if(CSSM_OK == err)
    {
        /* Use binary search to identify a possible caller.
        */
        CurrentMiddlePoint = NumEntries / 2;

        /* We do one extra check to initialize LastHighPoint and LastLowPoint
           to the extreme values.  This is for the future updates to
           CurrentMiddlePoint.
        */
        LastHighPoint = NumEntries - 1;
        LastLowPoint = 0;


        /* loop until we either find a candidate or know there is no match */
        while((CSSM_FALSE == CandidateFound) &&
              (CSSM_OK == err))
        {
            /* #ifdef VMS apply cast */
            if((ulong)EntryIndex[CurrentMiddlePoint].IndexKey < (ulong)TargetAddress)
            {
                if(CurrentMiddlePoint < (NumEntries-1))
                {
                    /* check the next starting point to see if this is
                       our candidate
                    */ 
                    /* #ifdef VMS apply cast */
                    if((ulong)EntryIndex[CurrentMiddlePoint+1].IndexKey > (ulong)TargetAddress)
                    {
                        CandidateFound = CSSM_TRUE;
                    }
                    else
                    {
                        /* calculate the new middle point */
                        PointDiff = LastHighPoint - CurrentMiddlePoint;

                        /* We're right next to the extreme point.  This means
                           that either we've already checked the point (in
                           case execution won't have gotten this far) or that
                           this is the initial extreme point.  If it's the
                           initial extreme point, we have to be certain we can
                           move to check it, so we bump the difference by one.
                        */
                        if(1 == PointDiff)
                        {
                            PointDiff = 2;
                        }

                        LastLowPoint = CurrentMiddlePoint;
                        CurrentMiddlePoint = CurrentMiddlePoint + (PointDiff / 2);
                    }
                }
                else
                {
                    /* This extreme edge is our candidate since we can't
                       check in more detail.  We know that if the caller is
                       present, this is the caller.
                    */
                    CandidateFound = CSSM_TRUE;
                }
            }
            else
            {
                /* #ifdef VMS apply cast */
                if((ulong)EntryIndex[CurrentMiddlePoint].IndexKey > (ulong)TargetAddress)
                {
                    if(0 != CurrentMiddlePoint)
                    {
                        PointDiff = CurrentMiddlePoint - LastLowPoint;

                        /* We're right next to the extreme point.  This means
                           that either we've already checked the point (in
                           case execution won't have gotten this far) or that
                           this is the initial extreme point.  If it's the
                           initial extreme point, we have to be certain we can
                           move to check it, so we bump the difference by one.
                        */
                        if(1 == PointDiff)
                        {
                            PointDiff = 2;
                        }

                        LastHighPoint = CurrentMiddlePoint;
                        CurrentMiddlePoint = CurrentMiddlePoint - (PointDiff / 2);
                    }
                    else
                    {
                        /* our search key is less than the lowest value in
                           our list, so we know the caller isn't present.
                        */
                        ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                    }
                }
                else
                {
                    /* we got an exact match for our search key */
                    CandidateFound = CSSM_TRUE;
                }
            }
        }

        /* We set the return in all cases, even if there's an error.  This is
           because the CandidateIndexEntry identifies where the entry would be
           if it existed.
        */
        *CandidateIndexEntry = CurrentMiddlePoint;
    }

    return err;
}
