/*-----------------------------------------------------------------------
 * File: CONTEXT.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */
 
#include "precomp.h"
#ifdef VMS
#include <string.h>
#endif

CSSM_RETURN
cssm_CallChallengeCallback(const CSSM_LIST *Challenge,
                           CSSM_SAMPLEGROUP_PTR Response,
                           CSSM_CHALLENGE_CALLBACK_PROXY_PTR ChallengeProxy,
                           const CSSM_MEMORY_FUNCS *MemFuncs)
{
    CSSM_RETURN err = CSSM_OK;

    if((NULL != ChallengeProxy) &&
       (CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)ChallengeProxy->Callback)))
    {

#ifdef ISL_INCLUDED
        /*
         * Check that callback function is in the app's verified space.
         */
        if(PvcMode & CSSM_PVC_APP)
        {
            if(CSSM_OK != ERR(cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)ChallengeProxy->Callback,
                                                                   CSSM_NO_LOCK,
                                                                   NULL,
                                                                   CSSM_NO_LOCK,
                                                                   NULL,
                                                                   CSSM_NO_LOCK,
                                                                   NULL)))
            {
                ERR(err = CSSMERR_CSSM_FUNCTION_INTEGRITY_FAIL);
            }
        }

#endif

    }
    else
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }

    if( CSSM_OK == err )
    {
    /* now we actually call the callback */
        ERR(err = ChallengeProxy->Callback(Challenge,Response,ChallengeProxy->CallbackCtx,MemFuncs));
    }
    else
    {
        if(CSSM_OK == err)
        {
            /* we should've gotten a callback if it was a complete and
                   valid structure pointed to
                */
            ERR(err = CSSMERR_CSSM_INVALID_POINTER);
        }
    }

    RETURN (err)
}

/* CSSM proxy for seed callback*/
CSSM_RETURN CSSMAPI
CSSM_CHALLENGE_CALLBACK_PROXY_CCHANDLE(const CSSM_LIST *Challenge,
                                  CSSM_SAMPLEGROUP_PTR Response,
                                  void *CallerCtx,
                                  const CSSM_MEMORY_FUNCS *MemFuncs)
{
    CSSM_CHALLENGE_CALLBACK_PROXY_PTR ProxyCtx;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_CONTEXT_PTR ContextRecord = NULL;
    cssm_INTERNAL_SERVICE_LIST_PTR pService = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR pAttach = NULL;

#ifdef ISL_INCLUDED
    uint32 i;
    ISL_STATUS callValid = ISL_FAIL;
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress( retAddress);
#endif


    if(CSSM_TRUE == port_IsBadReadPtr(CallerCtx,
                                      sizeof(CSSM_CHALLENGE_CALLBACK_PROXY)))
    {
        ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
    } 
    else 
    {
        ProxyCtx = (CSSM_CHALLENGE_CALLBACK_PROXY_PTR) CallerCtx;
        /* we need to verify that the caller is the valid service provider */
        ERR(err = cssm_FindContextAndMultiLock(ProxyCtx->CCHandle,
                                               CSSM_NO_LOCK,
                                               NULL,
                                               CSSM_NO_LOCK,
                                               &pService,
                                               CSSM_READER_LOCK,
                                               &pAttach,
                                               CSSM_READER_LOCK,
                                               &ContextRecord,
                                               CSSM_READER_LOCK));
        if(CSSM_OK == err)
        {

#ifdef ISL_INCLUDED
            if( cssm_GetPVCPolicy() & CSSM_PVC_SP )
            {
                if(NULL != pService->SPCredentialPtr)
                {
                    i = 0;

                    /* Loop through checking all components in the manifest -
                       it's likely that there's only one, but the spec allows
                       for more.
                    */
                    while((ISL_FAIL == callValid) &&
                          (i < pService->NumSPCreds))
                    {
                        callValid = EISL_CheckAddressWithinModule(pService->SPCredentialPtr[i],
                                                                  retAddress);
                        if(ISL_FAIL == callValid)
                        {
                            i++;
                        }
                    }
                }


                /* if we couldn't validate the caller, we'll be failing the call */
                if(ISL_FAIL == callValid)
                {
                    ERR(err = CSSMERR_CSSM_FUNCTION_INTEGRITY_FAIL);
                }
            }
#endif
            /* release the service record */
            cssm_ReleaseServiceLock (pService, CSSM_READER_LOCK);

            /* if everything's OK, we'll call the callback now */
            if(CSSM_OK == err)
            {
                ERR(err = cssm_CallChallengeCallback(Challenge,
                                                     Response,
                                                     ProxyCtx,
                                                     MemFuncs));
            }

            /* now we can release the attach and context records */
            cssm_ReleaseAttachLock(pAttach,
                                   CSSM_READER_LOCK);
            cssm_ReleaseContextLock(ContextRecord,
                                    CSSM_READER_LOCK);
        }
        else
        {
            ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
        }
    }

    RETURN (err)
}

/*---------------------------------------------------------------
 *
 *Name: cssm_CreateContext
 *
 *Description:
 *   Create the context structure
 *
 *Parameters: 
 *   CSPHandle(input) - CSPHandle
 *   Class (input) - class type to initialize the context info
 *   AlgorithmID (input) - algorithm to initialize context info
 *   Context (output) - the newly created context
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_CreateContext(CSSM_CSP_HANDLE CSPHandle,
                   CSSM_CONTEXT_TYPE Class, 
                   CSSM_ALGORITHMS AlgorithmID,
                   CSSM_CONTEXT_PTR *Context)
{
    CSSM_RETURN err = CSSM_OK;
    /* Parameter checking is not done here since this is internal function.
       The caller needs to make sure that the parameter is valid*/


    *Context = (CSSM_CONTEXT_PTR)local_calloc(1,
                                              (uint32)sizeof(CSSM_CONTEXT),
                                              0);

    if(NULL != *Context)
    {
        /* initialize the context structure */
        (*Context)->ContextType = Class;
        (*Context)->AlgorithmType = AlgorithmID;
        (*Context)->CSPHandle = CSPHandle;
    }
    else
    {
        ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: cssm_CopyDataStruct
 *
 *Description:
 *   Create a pointer for a CSSM_DATA structure and duplicate the
 *   data
 *
 *Parameters: 
 *
 *Returns:
 *   NULL - unable to complete function
 *   non NULL - pointer to the CSSM_DATA structure
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_CopyDataStruct(CSSM_DATA_PTR Data,
                    uint32 *Len,
                    CSSM_MALLOC malloc_func,
                    CSSM_FREE free_func,
                    void * local_ref,
                    CSSM_DATA_PTR *newData)
{
    CSSM_DATA_PTR Dest = NULL;
    CSSM_RETURN err = CSSM_OK;

    /* allocating the CSSM_DATA struct */
    TRACE(cssm_trace_info, FNCODE_cssm_CopyDataStruct);

    if((NULL == Data) ||
       (NULL == Data->Data) ||
       (0 == Data->Length) ||
       (NULL == Len) ||
       (NULL == malloc_func) ||
       (NULL == free_func) ||
       (NULL == newData))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        *newData = NULL;
        Dest = (CSSM_DATA_PTR)CSSM_MALLOC_WRAPPER(malloc_func,
                                                  (uint32)sizeof (CSSM_DATA),
                                                  local_ref);
        if(NULL == Dest)
        {
            ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
        }
        else
        {
            /* allocating the actual data pointer */
            *Len = Data->Length;

            if(0 < *Len)
            {
                Dest->Data = (uint8*)CSSM_MALLOC_WRAPPER(malloc_func,
                                                         *Len,
                                                         local_ref);
                if(NULL == Dest->Data)
                {
                    ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                }
                else
                {
                    /* copying the data */
                    Dest->Length = Data->Length;
                    port_memcpy(Dest->Data,
                                Data->Data,
                                Data->Length);
                }
            }
            else
            {
                Dest->Data = NULL;
                Dest->Length = 0;
            }

            if(CSSM_OK == err)
            {
                *newData = Dest;
            }
            else
            {
                CSSM_FREE_WRAPPER(free_func,
                                  Dest,
                                  local_ref);
            }
        }
    }

    RETURN(err);
}



/*---------------------------------------------------------------
 *
 *Name: cssm_CopyCryptoDataStruct
 *
 *Description:
 *   Duplicate a CSSM_CRYPTO_DATA structure
 *
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_CopyCryptoDataStruct(CSSM_CRYPTO_DATA_PTR Data, 
                          uint32 *Length,
                          CSSM_MALLOC malloc_func,
                          CSSM_FREE free_func,
                          void * local_ref,
                          CSSM_CRYPTO_DATA_PTR *newData)
{
    CSSM_RETURN err = CSSM_OK;
    CSSM_CRYPTO_DATA_PTR CryptoDest = NULL;
    uint32 tempLength;

    /* allocate the top level structure */
    TRACE(cssm_trace_info, FNCODE_cssm_CopyCryptoDataStruct);

    /* check parameters, remembering that either Data->Callback OR
       Data->Param.Data (but not both) may be NULL.
    */
    if((NULL == Data) ||
       ((NULL == Data->Callback) && (NULL == Data->Param.Data)) ||
       (NULL == Length) ||
       (NULL == malloc_func) ||
       (NULL == free_func) ||
       (NULL == newData))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        *newData = NULL;
        tempLength = (uint32)sizeof (CSSM_CRYPTO_DATA);
        CryptoDest = (CSSM_CRYPTO_DATA_PTR)CSSM_MALLOC_WRAPPER(malloc_func,
                                                               tempLength,
                                                               local_ref);
        if(NULL != CryptoDest)
        {
            /* initialize the new structure */
            CryptoDest->Callback = Data->Callback;
            CryptoDest->CallerCtx = Data->CallerCtx;

            /* duplicate the CSSM_DATA structure */
            CryptoDest->Param.Length = Data->Param.Length;
            if(0 < Data->Param.Length)
            {
                /* allocating the actual data pointer */
                CryptoDest->Param.Data = (uint8*)CSSM_MALLOC_WRAPPER(malloc_func,
                                                                     Data->Param.Length,
                                                                     local_ref);
                if(NULL != CryptoDest->Param.Data)
                {
                    /* copying the data */
                    port_memcpy(CryptoDest->Param.Data,
                                Data->Param.Data,
                                Data->Param.Length);

                }
                else
                {
                    ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                }
            }
            else
            {
                CryptoDest->Param.Data = NULL;
            }

            if(CSSM_OK == err)
            {
                *newData = CryptoDest;
                *Length = tempLength;
            }
            else
            {
                CSSM_FREE_WRAPPER(free_func,
                                  CryptoDest,
                                  local_ref);
            }
        }
        else
        {
            ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: cssm_CopyDateStruct
 *
 *Description:
 *   Duplicate a CSSM_DATE structure
 *
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_CopyDateStruct(CSSM_DATE_PTR Data,
                    CSSM_MALLOC malloc_func,
                    CSSM_FREE free_func,
                    void * local_ref,
                    CSSM_DATE_PTR *newDate)
{
    CSSM_DATE_PTR DateDest = NULL;
    uint32 Len1;
    CSSM_RETURN err = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_cssm_CopyDateStruct);

    if((NULL == Data) ||
       (NULL == malloc_func) ||
       (NULL == free_func) ||
       (NULL == newDate))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        *newDate = NULL;
        Len1 = (uint32)sizeof (CSSM_DATE);
        DateDest = (CSSM_DATE_PTR) CSSM_MALLOC_WRAPPER(malloc_func,
                                                       Len1,
                                                       local_ref);

        if(NULL != DateDest)
        {
            /* copying the data */
            port_memcpy(DateDest,
                        Data,
                        Len1);
        }
        else
        {
            ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
        }

        if(CSSM_OK == err)
        {
            *newDate = DateDest;
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: cssm_CopyRangeStruct
 *
 *Description:
 *   Duplicate a CSSM_CRYPTO_DATA structure
 *
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_CopyRangeStruct(CSSM_RANGE_PTR Data,
                     CSSM_MALLOC malloc_func,
                     void * local_ref,
                     CSSM_RANGE_PTR *newRange)
{
    CSSM_RETURN err = CSSM_OK;
    CSSM_RANGE_PTR Dest = NULL;
    uint32 Len1;

    TRACE(cssm_trace_info, FNCODE_cssm_CopyRangeStruct);

    if((NULL == Data) ||
       (NULL == malloc_func) ||
       (NULL == newRange))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        *newRange = NULL;
        Len1 = (uint32)sizeof (CSSM_RANGE);
        if(CSSM_TRUE == port_IsBadReadPtr(Data,
                                          Len1))
        {
            ERR(err = CSSMERR_CSSM_INVALID_POINTER);
        }
        else
        {
            /* allocate the top level structure */
            if((Dest = (CSSM_RANGE_PTR)CSSM_MALLOC_WRAPPER(malloc_func,
                                                           Len1,
                                                           local_ref)) == NULL)
            {
                ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
            }
            else
            {
                Dest->Min = Data->Min;
                Dest->Max = Data->Max;
            }
        }

        *newRange = Dest;
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: cssm_CopyKeyStruct
 *
 *Description:
 *   Duplicate a CSSM_KEY structure
 *
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_CopyKeyStruct(CSSM_KEY_PTR Data,
                   CSSM_MALLOC malloc_func,
                   CSSM_FREE free_func,
                   void * local_ref,
                   CSSM_KEY_PTR *newKey)
{
    CSSM_KEY_PTR Dest = NULL;
    uint32 Len1;
    CSSM_RETURN err = CSSM_OK;

    /* allocate the top level structure */
    TRACE(cssm_trace_info, FNCODE_cssm_CopyKeyStruct);

    if((NULL == Data) ||
       (NULL == malloc_func) ||
       (NULL == free_func) ||
       (NULL == newKey))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        *newKey = NULL;
        Len1 = (uint32)sizeof (CSSM_KEY);
        if((CSSM_TRUE == port_IsBadReadPtr(Data,
                                           Len1)) ||
           (CSSM_TRUE == port_IsBadReadPtr(Data->KeyData.Data,
                                           Data->KeyData.Length)))
        {
            ERR(err = CSSMERR_CSSM_INVALID_POINTER);
        }
        else
        {
            if((Dest = (CSSM_KEY_PTR)CSSM_MALLOC_WRAPPER(malloc_func,
                                                         Len1,
                                                         local_ref)) == NULL)
            {
                ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
            }
            else
            {
                /* copying the keyheader */
                port_memcpy(Dest,
                            Data,
                            sizeof (CSSM_KEYHEADER));

                /* duplicate the KeyData structure */
                Dest->KeyData.Length = Data->KeyData.Length;
                if( 0 < Data->KeyData.Length )
                {
                    if((Dest->KeyData.Data = (uint8*)CSSM_MALLOC_WRAPPER(malloc_func,
                                                                         Data->KeyData.Length,
                                                                         local_ref)) == NULL)
                    {
                        ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                    }
                    else
                    {
                        /* copying the key data */
                        port_memcpy(Dest->KeyData.Data,
                                    Data->KeyData.Data,
                                    Data->KeyData.Length);
                    }
                }
                else
                {
                    Dest->KeyData.Data = NULL;
                }

                if(CSSM_OK == err)
                {
                    /* assign the key for return */
                    *newKey = Dest;
                }
                else
                {
                    CSSM_FREE_WRAPPER(free_func,
                                      Dest,
                                      local_ref);
                }
            }
        }
    }

    RETURN(err)
}

/*---------------------------------------------------------------
 *
 *Name: cssm_CallCryptoSeedCallback
 *
 *Description:
 *   cssm callback for seed proxy. Check that callback function is in the app's verified space.
 * And calls the app's seed callback
 *
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *-------------------------------------------------------------*/


CSSM_RETURN
cssm_CallCryptoSeedCallback(CSSM_SEED_CALLBACK_PROXY_PTR SeedProxy,
                            CSSM_DATA_PTR SeedData)
{
    CSSM_RETURN err = CSSM_OK;

    if((NULL != SeedProxy) &&
       (CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)SeedProxy->Callback)))
    {

#ifdef ISL_INCLUDED
        /*
         * Check that callback function is in the app's verified space.
         */
        if(PvcMode & CSSM_PVC_APP)
        {
            if(CSSM_OK != ERR(cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)SeedProxy->Callback,
                                                                   CSSM_NO_LOCK,
                                                                   NULL,
                                                                   CSSM_NO_LOCK,
                                                                   NULL,
                                                                   CSSM_NO_LOCK,
                                                                   NULL)))
            {
                ERR(err = CSSMERR_CSSM_FUNCTION_INTEGRITY_FAIL);
            }
        }

#endif

    }
    else
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }

    if( CSSM_OK == err )
    {
    /* now we actually call the callback */
        ERR(err = SeedProxy->Callback(SeedData,
                              SeedProxy->CallbackCtx));
    }

    RETURN (err)
}


/*---------------------------------------------------------------
 *
 *Name: CSSM_SEED_CALLBACK
 *
 *Description:
 *   cssm proxy  for seed callback. Check that the return address is in the service provider's
     address space.And calls the app's seed callback
 *
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *-------------------------------------------------------------*/


/* CSSM proxy for  seed callback*/
CSSM_RETURN CSSMAPI
CSSM_SEED_CALLBACK(CSSM_DATA_PTR OutData,
                   void *CallbackCtx)
{
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_CONTEXT_PTR ContextRecord = NULL;
    cssm_INTERNAL_SERVICE_LIST_PTR pService;
    cssm_INTERNAL_ATTACH_LIST_PTR pAttach;
    CSSM_SEED_CALLBACK_PROXY_PTR SeedProxy = NULL;
#ifdef ISL_INCLUDED
    ISL_FUNCTION_PTR retAddress;
    EISL_GetReturnAddress( retAddress);
#endif

    if(CSSM_TRUE == port_IsBadReadPtr(CallbackCtx,
                                      sizeof(CSSM_SEED_CALLBACK_PROXY)))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        SeedProxy = (CSSM_SEED_CALLBACK_PROXY_PTR)CallbackCtx;
        /* we get a reader lock on the attach record to guarrantee that our
           call to cssm_FindCryptoSeedCallback won't deadlock.
        */
        ERR(err = cssm_FindContextAndMultiLock((SeedProxy->CCHandle),
                                               CSSM_NO_LOCK,
                                               NULL,
                                               CSSM_NO_LOCK,
                                               &pService,
                                               CSSM_READER_LOCK,
                                               &pAttach,
                                               CSSM_READER_LOCK,
                                               &ContextRecord,
                                               CSSM_READER_LOCK));
        if(CSSM_OK == err)
        {

#ifdef ISL_INCLUDED
            if( cssm_GetPVCPolicy() & CSSM_PVC_SP )
            {
                /*
                 * Check that return address falls in the calling module's range.
                 */
                if(NULL != pService->SPCredentialPtr)
                {
                    uint32 i = 0;
                    err = CSSMERR_CSSM_FUNCTION_INTEGRITY_FAIL;

                    /* Loop through the service provider's credentials to see if
                       the caller is trusted.  If it is, then we reset the error
                       code to CSSM_OK.
                    */
                    while((i < pService->NumSPCreds) &&
                          (CSSMERR_CSSM_FUNCTION_INTEGRITY_FAIL == err))
                    {
                        if(ISL_OK == EISL_CheckAddressWithinModule(pService->SPCredentialPtr[i],
                                                                   retAddress))
                        {
                            err = CSSM_OK;
                        }
                        else
                        {
                            i++;
                        }
                    }

                    /* this will notify the ERR macro that there was a func. integ.
                       fail in this section if there was one
                    */
                    ERR(err);
                }
                else
                {
                    /* we don't know anything about this module - we shouldn't have
                       gotten to this state by any trusted path
                    */
                    ERR(err = CSSMERR_CSSM_FUNCTION_INTEGRITY_FAIL);
                }
            }
#endif


            /* release the service record - if we ever needed it,
               we're done now
            */
            cssm_ReleaseServiceLock(pService,
                                    CSSM_READER_LOCK);

            if(CSSM_OK == err)
            {
                ERR( err = cssm_CallCryptoSeedCallback( SeedProxy, OutData));
            }

            /* now we can release the attach and context records */
            cssm_ReleaseAttachLock(pAttach,
                                   CSSM_READER_LOCK);
            cssm_ReleaseContextLock(ContextRecord,
                                    CSSM_READER_LOCK);
        }
    }

    RETURN (err)
}

/*---------------------------------------------------------------
 *
 *Name: cssm_CallCryptoPassphraseCallback
 *
 *Description:
 *   cssm callback for passphrase proxy. Check that callback function is in the app's verified space.
 * And calls the app's seed callback
 *
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *-------------------------------------------------------------*/


CSSM_RETURN
cssm_CallCryptoPassphraseCallback(CSSM_PASSPHRASE_PROXY_PTR PassphraseProxy,
                            CSSM_DATA_PTR Passphrase)
{
    CSSM_RETURN err = CSSM_OK;

    if((NULL != PassphraseProxy) &&
       (CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)PassphraseProxy->Callback)))
    {

#ifdef ISL_INCLUDED
        /*
         * Check that callback function is in the app's verified space.
         */
        if(PvcMode & CSSM_PVC_APP)
        {
            if(CSSM_OK != ERR(cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)PassphraseProxy->Callback,
                                                                   CSSM_NO_LOCK,
                                                                   NULL,
                                                                   CSSM_NO_LOCK,
                                                                   NULL,
                                                                   CSSM_NO_LOCK,
                                                                   NULL)))
            {
                ERR(err = CSSMERR_CSSM_FUNCTION_INTEGRITY_FAIL);
            }
        }

#endif

    }
    else
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }

    if( CSSM_OK == err )
    {
    /* now we actually call the callback */
        ERR(err = PassphraseProxy->Callback(Passphrase,
                              PassphraseProxy->CallbackCtx));
    }
    else
    {
        if(CSSM_OK == err)
        {
            /* we should've gotten a callback if it was a complete and
                   valid structure pointed to
                */
            ERR(err = CSSMERR_CSSM_INVALID_POINTER);
        }
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_PASSPHRASE_CALLBACK
 *
 *Description:
 *   cssm proxy  for passphrase callback. Check that the return address is in the service provider's
     address space.And calls the app's seed callback
 *
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *-------------------------------------------------------------*/


CSSM_RETURN CSSMAPI
CSSM_PASSPHRASE_CALLBACK(CSSM_DATA_PTR OutData,
                   void *CallbackCtx)
{
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_CONTEXT_PTR ContextRecord = NULL;
    cssm_INTERNAL_SERVICE_LIST_PTR pService;
    cssm_INTERNAL_ATTACH_LIST_PTR pAttach;
    CSSM_PASSPHRASE_PROXY_PTR PassphraseProxy = NULL;
#ifdef ISL_INCLUDED
    ISL_FUNCTION_PTR retAddress;
    EISL_GetReturnAddress( retAddress);
#endif

    if(CSSM_TRUE == port_IsBadReadPtr(CallbackCtx,
                                      sizeof(CSSM_PASSPHRASE_PROXY)))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        PassphraseProxy = (CSSM_PASSPHRASE_PROXY_PTR)CallbackCtx;
        /* we get a reader lock on the attach record to guarrantee that our
           call to cssm_FindCryptoSeedCallback won't deadlock.
        */
        ERR(err = cssm_FindContextAndMultiLock((PassphraseProxy->CCHandle),
                                               CSSM_NO_LOCK,
                                               NULL,
                                               CSSM_NO_LOCK,
                                               &pService,
                                               CSSM_READER_LOCK,
                                               &pAttach,
                                               CSSM_READER_LOCK,
                                               &ContextRecord,
                                               CSSM_READER_LOCK));
        if(CSSM_OK == err)
        {

#ifdef ISL_INCLUDED
            if( cssm_GetPVCPolicy() & CSSM_PVC_SP )
            {
                /*
                 * Check that return address falls in the calling addin module's range.
                 */
                if(NULL != pService->SPCredentialPtr)
                {
                    uint32 i = 0;
                    err = CSSMERR_CSSM_FUNCTION_INTEGRITY_FAIL;

                    /* Loop through the service provider's credentials to see if
                       the caller is trusted.  If it is, then we reset the error
                       code to CSSM_OK.
                    */
                    while((i < pService->NumSPCreds) &&
                          (CSSMERR_CSSM_FUNCTION_INTEGRITY_FAIL == err))
                    {
                        if(ISL_OK == EISL_CheckAddressWithinModule(pService->SPCredentialPtr[i],
                                                                   retAddress))
                        {
                            err = CSSM_OK;
                        }
                        else
                        {
                            i++;
                        }
                    }

                    /* this will notify the ERR macro that there was a func. integ.
                       fail in this section if there was one
                    */
                    ERR(err);
                }
                else
                {
                    /* we don't know anything about this module - we shouldn't have
                       gotten to this state by any trusted path
                    */
                    ERR(err = CSSMERR_CSSM_FUNCTION_INTEGRITY_FAIL);
                }
            }
#endif


            /* release the service record - if we ever needed it,
               we're done now
            */
            cssm_ReleaseServiceLock(pService,
                                    CSSM_READER_LOCK);

            if(CSSM_OK == err)
            {
                ERR( err = cssm_CallCryptoPassphraseCallback( PassphraseProxy, OutData));
            }

            /* now we can release the attach and context records */
            cssm_ReleaseAttachLock(pAttach,
                                   CSSM_READER_LOCK);
            cssm_ReleaseContextLock(ContextRecord,
                                    CSSM_READER_LOCK);
        }
    }

    RETURN (err)
}


/*---------------------------------------------------------------
 *
 *Name: cssm_DuplicateAttributes
 *
 *Description:
 *   Given a set of attributes, duplicate it and return a pointer
 *   to it.
 *
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_DuplicateAttributes(uint32 Number,
                         CSSM_CONTEXT_ATTRIBUTE_PTR Attributes,
                         CSSM_MALLOC malloc_func,
                         CSSM_FREE free_func,
                         void * local_ref,
                         CSSM_CONTEXT_ATTRIBUTE_PTR *newAttr)
{
    CSSM_RETURN err = CSSM_OK;
    uint16 i;
    CSSM_CONTEXT_ATTRIBUTE_PTR AttributesTemp;


    /* allocate memory for the attributes structure */
    TRACE(cssm_trace_info, FNCODE_cssm_DuplicateAttributes);

    if((NULL == Attributes) ||
       (NULL == malloc_func) ||
       (NULL == free_func) ||
       (NULL == newAttr))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        if((AttributesTemp = (CSSM_CONTEXT_ATTRIBUTE_PTR)CSSM_MALLOC_WRAPPER(malloc_func,
                                                                             Number * sizeof (CSSM_CONTEXT_ATTRIBUTE),
                                                                             local_ref)) == NULL)
        {
            ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
        }
        else
        {
            /* set memory of AttributesTemp to 0 - we would've used calloc but
               we don't want to mess with the interface to get a calloc func
               in here
            */
            memset(AttributesTemp,
                   0,
                   (Number * sizeof(CSSM_CONTEXT_ATTRIBUTE)));

            i = 0;

            /* for every attribute */
            while((i < Number) &&
                  (CSSM_OK == err))
            {
                AttributesTemp[i].AttributeType = Attributes[i].AttributeType;
                AttributesTemp[i].AttributeLength = Attributes[i].AttributeLength;

                /* see if there is data to copy */
                if((NULL != Attributes[i].Attribute.Data) ||
                   (CSSM_ATTRIBUTE_DATA_UINT32 == (Attributes[i].AttributeType & CSSM_ATTRIBUTE_TYPE_MASK)))
                {
                    uint32 Length;

                    /* Get actual length of data */
                    Length = Attributes[i].AttributeLength;

                    /* check the validity of the data pointer */
                    if((Attributes[i].AttributeType & CSSM_ATTRIBUTE_TYPE_MASK) != CSSM_ATTRIBUTE_DATA_UINT32)
                    {
                        if(port_IsBadReadPtr(Attributes[i].Attribute.Data, Length)) 
                        {
                            ERR(err = CSSMERR_CSSM_INVALID_POINTER);
                        }
                    }

                    /* every attribute type is handled differently */
                    switch (Attributes[i].AttributeType & CSSM_ATTRIBUTE_TYPE_MASK)
                    {

                        case CSSM_ATTRIBUTE_DATA_NONE:
                        {
                            ERR(CSSMERR_CSSM_INVALID_ATTRIBUTE);
                        }
                        break;

                        case CSSM_ATTRIBUTE_DATA_UINT32:
                        {
                            AttributesTemp[i].Attribute.Uint32 = 
                                                Attributes[i].Attribute.Uint32;
                        }
                        break;

                        case CSSM_ATTRIBUTE_DATA_CSSM_DATA:
                        {
                            CSSM_DATA_PTR Data;
                            uint32 attLength = Attributes[i].Attribute.Data->Length;

                            if((Data = (CSSM_DATA_PTR)CSSM_MALLOC_WRAPPER(malloc_func,
                                                                          (uint32)sizeof (CSSM_DATA),
                                                                          local_ref)) == NULL)
                            {
                                ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                            }
                            else
                            {
                                if(0 < attLength)
                                {
                                    if((Data->Data = (uint8 *)CSSM_MALLOC_WRAPPER(malloc_func,
                                                                                  attLength,
                                                                                  local_ref)) == NULL)
                                    {
                                        ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                                    }
                                    else
                                    {
                                        /* copy the data */
                                        if(ERR(err = cssm_CheckCopy(Data->Data,
                                                                    Attributes[i].Attribute.Data->Data,
                                                                    attLength)) != CSSM_OK)
                                        {
                                            ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                                        }
                                        else
                                        {
                                            /* we can actually set the data now */
                                            Data->Length = attLength;
                                        }

                                        if(CSSM_OK != err)
                                        {
                                            /* free memory */
                                            CSSM_FREE_WRAPPER(free_func,
                                                              Data->Data,
                                                              local_ref);
                                        }
                                    }
                                }
                                else
                                {
                                    Data->Length = 0;
                                    Data->Data = NULL;
                                }

                                if(CSSM_OK == err)
                                {
                                    AttributesTemp[i].Attribute.Data = Data;
                                }
                                else
                                {
                                    CSSM_FREE_WRAPPER(free_func,
                                                      Data,
                                                      local_ref);
                                }
                            }

                            if(CSSM_OK != err)
                            {
                                /* free memory */
                                cssm_FreeAttributes(i,  /* so far there are i elements */
                                                    AttributesTemp,
                                                    free_func,
                                                    local_ref,
                                                    0); /* we don't free "AttributesTemp" here */
                            }
                        }
                        break;

                        case CSSM_ATTRIBUTE_DATA_CRYPTO_DATA:
                        {
                            CSSM_CRYPTO_DATA_PTR p = NULL;
                            CSSM_CRYPTO_DATA_PTR q = Attributes[i].Attribute.CryptoData;

                            if(CSSM_FALSE == cssm_IsBadCryptoDataPtr(q))
                            {
                                if((p = (CSSM_CRYPTO_DATA_PTR)CSSM_MALLOC_WRAPPER(malloc_func,
                                                                                  (uint32)sizeof (CSSM_CRYPTO_DATA),
                                                                                  local_ref)) == NULL)
                                {
                                    /* free memory */
                                    cssm_FreeAttributes(i,  /* so far there are i elements */
                                                        AttributesTemp,
                                                        free_func,
                                                        local_ref,
                                                        0); /* we don't free "AttributesTemp" here */

                                    ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                                }
                                else
                                {
                                    p->Callback = q->Callback;
                                    p->CallerCtx = q->CallerCtx;

                                    if(q->Param.Data == NULL)
                                    {
                                        p->Param.Data = q->Param.Data;
                                        p->Param.Length = q->Param.Length;
                                    }
                                    else
                                    {
                                        if ((p->Param.Data = (uint8*)CSSM_MALLOC_WRAPPER(malloc_func,
                                                                                         q->Param.Length,
                                                                                         local_ref)) == NULL)
                                        {
                                            CSSM_FREE_WRAPPER(free_func,
                                                               p,
                                                               local_ref);
                                            p = NULL;
                                            /* free memory */
                                            cssm_FreeAttributes(i,  /* so far there are i elements */
                                                                AttributesTemp,
                                                                free_func,
                                                                local_ref,
                                                                0); /* we don't free "AttributesTemp" here */

                                            ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                                        }
                                        else
                                        {
                                            p->Param.Length = q->Param.Length;
                                            if(ERR(err = cssm_CheckCopy(p->Param.Data,
                                                                        q->Param.Data,
                                                                        q->Param.Length)) != CSSM_OK)
                                            {
                                                CSSM_FREE_WRAPPER(free_func,
                                                                  p->Param.Data,
                                                                  local_ref);
                                                CSSM_FREE_WRAPPER(free_func,
                                                                  p,
                                                                  local_ref);
                                                p = NULL;
                                                /* free memory */
                                                cssm_FreeAttributes(i,  /* so far there are i elements */
                                                                    AttributesTemp,
                                                                    free_func,
                                                                    local_ref,
                                                                    0); /* we don't free "AttributesTemp" here */

                                                ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                                            } /* end if */
                                        } /* end else */
                                    }
                                } /* end else */
                            }   /* end if */
                            else
                            {
                                ERR(err = CSSMERR_CSSM_INVALID_POINTER);

                                /* free memory */
                                cssm_FreeAttributes(i,  /* so far there are i elements */
                                                    AttributesTemp,
                                                    free_func,
                                                    local_ref,
                                                    0); /* we don't free "AttributesTemp" here */
                            }

                            /* assign the attribute - in fail cases it's NULL */
                            AttributesTemp[i].Attribute.CryptoData = p;
                        }
                        break;

                        case CSSM_ATTRIBUTE_DATA_KEY:
                        {
                            CSSM_KEY_PTR q = Attributes[i].Attribute.Key;

                            /* copy the key */
                            if(ERR(err = cssm_CopyKeyStruct(q,
                                                            malloc_func,
                                                            free_func,
                                                            local_ref,
                                                            &AttributesTemp[i].Attribute.Key)) != CSSM_OK)
                            {
                                /* free memory */
                                cssm_FreeAttributes(i,  /* so far there are i elements */
                                                    AttributesTemp,
                                                    free_func,
                                                    local_ref,
                                                    0); /* we don't free "AttributesTemp" here */
                            }
                     
                        }
                        break;

                        case CSSM_ATTRIBUTE_DATA_STRING:
                        {
                            char *str;


                            if((str = (char *)CSSM_MALLOC_WRAPPER(malloc_func,
                                                                  strlen(Attributes[i].Attribute.String)+1,
                                                                  local_ref)) == NULL)
                            {
                                /* free memory */
                                cssm_FreeAttributes(i,  /* so far there are i elements */
                                                    AttributesTemp,
                                                    free_func,
                                                    local_ref,
                                                    0); /* we don't free "AttributesTemp" here */

                                ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                            }
                            else
                            {
                                strcpy (str, Attributes[i].Attribute.String);
                                AttributesTemp[i].Attribute.String = str;
                            }

                        }
                        break;

                        case CSSM_ATTRIBUTE_DATA_DATE:
                        {
                            if(ERR(err = cssm_CopyDateStruct(Attributes[i].Attribute.Date,
                                                             malloc_func,
                                                             free_func,
                                                             local_ref,
                                                             &AttributesTemp[i].Attribute.Date)) != CSSM_OK)
                            {
                                /* free memory */
                                cssm_FreeAttributes(i,  /* so far there are i elements */
                                                    AttributesTemp,
                                                    free_func,
                                                    local_ref,
                                                    0); /* we don't free "AttributesTemp" here */
                            }

                        }
                        break;

                        case CSSM_ATTRIBUTE_DATA_RANGE:
                        {
                            if(ERR(err = cssm_CopyRangeStruct(Attributes[i].Attribute.Range,
                                                              malloc_func,
                                                              local_ref,
                                                              &AttributesTemp[i].Attribute.Range)) != CSSM_OK)
                            {
                                /* free memory */
                                cssm_FreeAttributes(i,  /* so far there are i elements */
                                                    AttributesTemp,
                                                    free_func,
                                                    local_ref,
                                                    0); /* we don't free "AttributesTemp" here */
                            }

                        }
                        break;

                        case CSSM_ATTRIBUTE_DATA_ACCESS_CREDENTIALS:
                        {
                            CSSM_ACCESS_CREDENTIALS_PTR AccessCredentials;

                            AccessCredentials = (CSSM_ACCESS_CREDENTIALS_PTR)CSSM_MALLOC_WRAPPER(malloc_func,
                                                                                          sizeof(CSSM_ACCESS_CREDENTIALS),
                                                                                          local_ref);
                            if(NULL == AccessCredentials)
                            {
                                ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                            }
                            else
                            {
                                port_memcpy(AccessCredentials->EntryTag,
                                            Attributes[i].Attribute.AccessCredentials->EntryTag,
                                            sizeof(CSSM_STRING));
                                AccessCredentials->Callback = Attributes[i].Attribute.AccessCredentials->Callback;
                                AccessCredentials->CallerCtx = Attributes[i].Attribute.AccessCredentials->CallerCtx;

                                ERR(err = cssmutil_DuplicateBaseCerts(&Attributes[i].Attribute.AccessCredentials->BaseCerts,
                                                                      malloc_func,
                                                                      free_func,
                                                                      local_ref,
                                                                      &AccessCredentials->BaseCerts));
                                if(CSSM_OK == err)
                                {
                                    ERR(cssmutil_DuplicateSampleGroup(&Attributes[i].Attribute.AccessCredentials->Samples,
                                                                      malloc_func,
                                                                      free_func,
                                                                      local_ref,
                                                                      &AccessCredentials->Samples));
                                }

                                if(CSSM_OK != err)
                                {
                                    CSSM_FREE_WRAPPER(free_func,
                                                      AccessCredentials,
                                                      local_ref);
                                }
                                else
                                {
                                    AttributesTemp[i].Attribute.AccessCredentials = AccessCredentials;
                                }
                            }
                        }
                        break;

                        case CSSM_ATTRIBUTE_DATA_VERSION:
                        {
                            CSSM_VERSION_PTR Version;

                            Version = (CSSM_VERSION_PTR)CSSM_MALLOC_WRAPPER(malloc_func,
                                                                            sizeof(CSSM_VERSION),
                                                                            local_ref);
                            if(NULL == Version)
                            {
                                /* free memory */
                                cssm_FreeAttributes(i,  /* so far there are i elements */
                                                    AttributesTemp,
                                                    free_func,
                                                    local_ref,
                                                    0); /* we don't free "AttributesTemp" here */

                                ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                            }
                            else
                            {
                                Version->Major = Attributes[i].Attribute.Version->Major;
                                Version->Minor = Attributes[i].Attribute.Version->Minor;
                                AttributesTemp[i].Attribute.Version = Version;
                            }

                        }
                        break;

                        case CSSM_ATTRIBUTE_DATA_DL_DB_HANDLE:
                        {
                            CSSM_DL_DB_HANDLE_PTR DLDBHandle;

                            DLDBHandle = (CSSM_DL_DB_HANDLE_PTR)CSSM_MALLOC_WRAPPER(malloc_func,
                                                                                     sizeof(CSSM_DL_DB_HANDLE),
                                                                                     local_ref);
                            if(NULL == DLDBHandle)
                            {
                                /* free memory */
                                cssm_FreeAttributes(i,  /* so far there are i elements */
                                                    AttributesTemp,
                                                    free_func,
                                                    local_ref,
                                                    0); /* we don't free "AttributesTemp" here */

                                ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                            }
                            else
                            {
                                DLDBHandle->DLHandle = Attributes[i].Attribute.DLDBHandle->DLHandle;
                                DLDBHandle->DBHandle = Attributes[i].Attribute.DLDBHandle->DBHandle;
                                AttributesTemp[i].Attribute.DLDBHandle = DLDBHandle;
                            }
                        }
                        break;


                        default:
                        {
                            ERR(err = CSSMERR_CSSM_INVALID_ATTRIBUTE);
                        }
                        break;

                    }   /* end of switch */
                }
                else
                {
                    /* the attribute isn't a legal attribute */
                    ERR(err = CSSMERR_CSSM_INVALID_ATTRIBUTE);
                }

                i++;    /* move to the next attribute */
            }

            if(CSSM_OK == err)
            {
                *newAttr = AttributesTemp;
            }
            else
            {
                /* there was an error, so we'll free AttributesTemp */
                CSSM_FREE_WRAPPER(free_func,
                                  AttributesTemp,
                                  local_ref);
            }
        }
    }

    RETURN(err)
}
/*---------------------------------------------------------------
 *
 *Name: cssm_ReplaceAppCallbacksWithCssmProxy
 *
 *Description:
 *   
 *   Before calling the CSP, replace all app callbacks in the context with cssm proxy callbacks.
 *
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *-------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
cssm_ReplaceAppCallbacksWithCssmProxy(cssm_INTERNAL_CONTEXT_PTR TempContext,
                               uint32 Number,
                               CSSM_CONTEXT_ATTRIBUTE_PTR Attributes)
{
    CSSM_RETURN err = CSSM_OK;
    uint16 i;
    i = 0;

    /* for every attribute */
    while((i < Number) &&
          (CSSM_OK == err))
    {
        /* see if there is data to copy */
        if( NULL != Attributes[i].Attribute.Data )
        {
            uint32 Length;

            /* Get actual length of data */
            Length = Attributes[i].AttributeLength;
            /* every attribute type is handled differently */
            switch (Attributes[i].AttributeType & CSSM_ATTRIBUTE_TYPE_MASK)
            {

                case CSSM_ATTRIBUTE_DATA_CRYPTO_DATA:
                {
                     /* before calling the spi, cssm proxy is
                        to be used. use CSSM_SEED or
                        CSSM_PASSPHRASE callback,
                      */
                    if(CSSM_FALSE == cssm_IsBadCryptoDataPtr(Attributes[i].Attribute.CryptoData))
                    {

                        if( Attributes[i].AttributeType == (uint32)CSSM_ATTRIBUTE_SEED)
                        {
                            if( Attributes[i].Attribute.CryptoData->Callback != NULL)
                            {
                                CSSM_SEED_CALLBACK_PROXY_PTR SeedProxy = NULL;
                                if(NULL == (SeedProxy = (CSSM_SEED_CALLBACK_PROXY *)local_malloc((uint32)sizeof(CSSM_SEED_CALLBACK_PROXY),
                                                                                               NULL)))
                                {
                                    ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                                }
                                else
                                {
                                    TempContext->SeedProxy = SeedProxy;
                                    SeedProxy->CCHandle = TempContext->ContextHandle;
                                    SeedProxy->Callback = Attributes[i].Attribute.CryptoData->Callback;
                                    SeedProxy->CallbackCtx = Attributes[i].Attribute.CryptoData->CallerCtx;
                                    Attributes[i].Attribute.CryptoData->Callback = CSSM_SEED_CALLBACK;
                                    Attributes[i].Attribute.CryptoData->CallerCtx = (void*)SeedProxy;
                                }
                            }
                        }
                        else if(Attributes[i].AttributeType == (uint32)CSSM_ATTRIBUTE_PASSPHRASE)
                        {
                            if( Attributes[i].Attribute.CryptoData->Callback != NULL)
                            {
                                CSSM_PASSPHRASE_PROXY_PTR PassphraseProxy = NULL;
                                if(NULL == (PassphraseProxy = (CSSM_PASSPHRASE_PROXY *)local_malloc((uint32)sizeof(CSSM_PASSPHRASE_PROXY),
                                                                                                NULL)))
                                {
                                    ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                                }
                                else
                                {
                                    TempContext->PassphraseProxy = PassphraseProxy;
                                    PassphraseProxy->CCHandle = TempContext->ContextHandle;
                                    PassphraseProxy->Callback = Attributes[i].Attribute.CryptoData->Callback;
                                    PassphraseProxy->CallbackCtx = Attributes[i].Attribute.CryptoData->CallerCtx;
                                    Attributes[i].Attribute.CryptoData->Callback = CSSM_PASSPHRASE_CALLBACK;
                                    Attributes[i].Attribute.CryptoData->CallerCtx = (void*)PassphraseProxy;
                                }
                            }

                        }
                    }

                    break;

                case CSSM_ATTRIBUTE_DATA_ACCESS_CREDENTIALS:
                {
                    CSSM_CHALLENGE_CALLBACK_PROXY_PTR ChallengeCbProxy = NULL;
                    if( Attributes[i].Attribute.AccessCredentials != NULL) 
                    {
                        if(NULL == (ChallengeCbProxy = (CSSM_CHALLENGE_CALLBACK_PROXY *)local_malloc((uint32)sizeof(CSSM_CHALLENGE_CALLBACK_PROXY),
                                                                                             NULL)))
                        {
                            ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                        }
                        else
                        {
                            TempContext->ChallengeCbProxy = ChallengeCbProxy;
                            ChallengeCbProxy->CCHandle = TempContext->ContextHandle;
                            ChallengeCbProxy->Callback = Attributes[i].Attribute.AccessCredentials->Callback;
                            ChallengeCbProxy->CallbackCtx = Attributes[i].Attribute.AccessCredentials->CallerCtx;
                            Attributes[i].Attribute.AccessCredentials->Callback = CSSM_CHALLENGE_CALLBACK_PROXY_CCHANDLE;
                            Attributes[i].Attribute.AccessCredentials->CallerCtx = (void*)ChallengeCbProxy;
                        }

                    }

                }
                break;

                default:
                    break;

                    }   /* end of switch */
            }
        }
                i++;    /* move to the next attribute */
    }
    if( err != CSSM_OK )
    {
        if( TempContext->ChallengeCbProxy != NULL )
        {
            local_free( TempContext->ChallengeCbProxy, NULL );
            TempContext->ChallengeCbProxy = NULL;
        }
        if( TempContext->SeedProxy != NULL )
        {
            local_free( TempContext->SeedProxy, NULL );
            TempContext->SeedProxy = NULL;
        }

        if( TempContext->PassphraseProxy != NULL )
        {
            local_free( TempContext->PassphraseProxy, NULL );
            TempContext->PassphraseProxy = NULL;
        }

    }

    RETURN(err)
}

/*---------------------------------------------------------------
 *
 *Name: cssm_ReplaceCssmProxyWithAppCallback
 *
 *Description:
 *   Before returning a context structure to the application, replace all cssm proxy callbacks with 
 *   app's callback
 *
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_ReplaceCssmProxyWithAppCallback(cssm_INTERNAL_CONTEXT_PTR TempContext,
                               uint32 Number,
                               CSSM_CONTEXT_ATTRIBUTE_PTR Attributes)
{
    CSSM_RETURN err = CSSM_OK;
    uint16 i;
    i = 0;

    /* for every attribute */
    while((i < Number) &&
          (CSSM_OK == err))
    {
        /* see if there is data to copy */
        if( NULL != Attributes[i].Attribute.Data )
        {
            uint32 Length;

            /* Get actual length of data */
            Length = Attributes[i].AttributeLength;
            /* every attribute type is handled differently */
            switch (Attributes[i].AttributeType & CSSM_ATTRIBUTE_TYPE_MASK)
            {

                case CSSM_ATTRIBUTE_DATA_CRYPTO_DATA:
                {
                     /* before calling the spi, cssm proxy is
                        to be used. use CSSM_SEED or
                        CSSM_PASSPHRASE callback,
                      */
                    if(CSSM_FALSE == cssm_IsBadCryptoDataPtr(Attributes[i].Attribute.CryptoData))
                    {

                        if( Attributes[i].AttributeType == (uint32)CSSM_ATTRIBUTE_SEED)
                        {
                            if( TempContext->SeedProxy != NULL)
                            {
                                Attributes[i].Attribute.CryptoData->Callback = TempContext->SeedProxy->Callback;
                                Attributes[i].Attribute.CryptoData->CallerCtx = TempContext->SeedProxy->CallbackCtx;
                            }
                        }
                        else if(Attributes[i].AttributeType == (uint32)CSSM_ATTRIBUTE_PASSPHRASE)
                        {
                            if(  TempContext->PassphraseProxy != NULL )
                            {
                                Attributes[i].Attribute.CryptoData->Callback = TempContext->PassphraseProxy->Callback;
                                Attributes[i].Attribute.CryptoData->CallerCtx = TempContext->PassphraseProxy->CallbackCtx;
                            }
                        }
                    }

                    break;

                case CSSM_ATTRIBUTE_DATA_ACCESS_CREDENTIALS:
                {
                    if( Attributes[i].Attribute.AccessCredentials != NULL) 
                    {
                        if( TempContext->ChallengeCbProxy != NULL)
                        {
                            Attributes[i].Attribute.AccessCredentials->Callback = TempContext->ChallengeCbProxy->Callback;
                            Attributes[i].Attribute.AccessCredentials->CallerCtx = TempContext->ChallengeCbProxy->CallbackCtx;
                        }
                    }

                }
                break;

                default:
                    break;

                    }   /* end of switch */
            }
        }
                i++;    /* move to the next attribute */
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: cssm_DuplicateContext
 *
 *Description:
 *   Given a context, duplicate it.
 *
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_DuplicateContext(const CSSM_CONTEXT *Context,                   
                      CSSM_MALLOC malloc_func, 
                      CSSM_FREE free_func,
                      void *mem_ref, 
                      cssm_INTERNAL_CONTEXT_PTR TempInternalContext,
                      CSSM_CONTEXT_PTR *DuplicateContext)
{
    CSSM_CONTEXT_PTR ContextTemp = NULL;
    CSSM_RETURN err = CSSM_OK;

    /* allocating memory for context */
    TRACE(cssm_trace_info, FNCODE_cssm_DuplicateContext);

    if((NULL == Context) ||
       (NULL == malloc_func) ||
       (NULL == free_func) ||
       (NULL == DuplicateContext))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        *DuplicateContext = NULL;
        if((ContextTemp = (CSSM_CONTEXT_PTR)CSSM_MALLOC_WRAPPER(malloc_func,
                                                                (uint32)sizeof (CSSM_CONTEXT),
                                                                mem_ref)) == NULL)
        {
            ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
        }
        else
        {
            /* assign fields in context */
            memcpy( ContextTemp, Context, sizeof(*ContextTemp) );
            ContextTemp->ContextAttributes = NULL;

            if(ContextTemp->NumberOfAttributes > 0)
            {
               /* duplicate every attribute in context */
               ERR(err = cssm_DuplicateAttributes(Context->NumberOfAttributes,
                                                  Context->ContextAttributes,
                                                  malloc_func,
                                                  free_func,
                                                  mem_ref,
                                                  &(ContextTemp->ContextAttributes)));
            }

            if(err != CSSM_OK)
            {
                CSSM_FREE_WRAPPER(free_func,
                                  ContextTemp,
                                  mem_ref);
                ContextTemp = NULL;
            }
        }

        *DuplicateContext = ContextTemp;
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: cssm_InsertContext
 *
 *Description:
 *   Insert the context into an internal list of context
 *
 *Parameters: 
 *   Context (input) - pointer to the context
 *   CSPHandle (input) - CSP handle associated with context
 *   pCCHandle( output) - the new CCHandle
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_InsertContext(CSSM_CONTEXT_PTR Context, 
                   CSSM_CSP_HANDLE CSPHandle, 
                   CSSM_CC_HANDLE *pCCHandle)
{
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR TempContext = NULL;
    cssm_INTERNAL_CONTEXT_PTR ListIndexPtr = NULL;
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CC_HANDLE tempHandle = 0;
    uint32 RandContextHandleBits = 0;

    /* allocate structure for context */
    TRACE(cssm_trace_info, FNCODE_cssm_InsertContext);

    if(NULL == (TempContext = (cssm_INTERNAL_CONTEXT_PTR)local_malloc((uint32)sizeof(cssm_INTERNAL_CONTEXT),
                                                                      0)))
    {
        ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
    }
    else
    {
        /* lock the attach record - which does the context list */
        if(ERR(err = cssm_FindAttachAndMultiLock(CSPHandle,
                                                 CSSM_NO_LOCK,
                                                 NULL,
                                                 CSSM_NO_LOCK,
                                                 NULL,
                                                 CSSM_NO_LOCK,
                                                 &AttachRecord,
                                                 CSSM_WRITER_LOCK)) == CSSM_OK)
        {
            /* if this isn't a CSP handle */
            if(0 == (AttachRecord->ServiceType & CSSM_SERVICE_CSP))
            {
                ERR(err = CSSMERR_CSSM_INVALID_HANDLE_USAGE);
            }
            else
            {
                /* Generate random handle for context */
                while(0 == RandContextHandleBits)
                {
                    RandContextHandleBits = cssm_Rand();
                    ListIndexPtr = AttachRecord->ContextList;

                    while((NULL != ListIndexPtr) &&
                          (0 != RandContextHandleBits))
                    {
                        if(RandContextHandleBits == CSSM_UINT64_LOW(ListIndexPtr->ContextHandle))
                            RandContextHandleBits = 0;

                        ListIndexPtr = ListIndexPtr->Next;
                    }
                }

                /* set the CSPHandle in the context */
                Context->CSPHandle = CSPHandle;

                /* set the Context in the internal context */
                TempContext->Context = Context;

                /* construct the context handle */
                TempContext->ContextHandle = CSPHandle;
                TempContext->ContextHandle = (TempContext->ContextHandle  << 32);
                TempContext->ContextHandle = (TempContext->ContextHandle |
                                              (0x00000000FFFFFFFF & RandContextHandleBits));

                TempContext->CSPHandle = CSPHandle;
                TempContext->Next = NULL;
                TempContext->hContextSWMRLock = NULL;
                TempContext->PassphraseProxy = NULL;
                TempContext->ChallengeCbProxy = NULL;
                TempContext->SeedProxy = NULL;

                /*
                 * Create the multiple-reader-single-write synchronization object.
                 */
                ERR(err = cssm_SWMRInit(&TempContext->hContextSWMRLock));

                if(CSSM_OK == err)
                {
                        /* duplicate the context for giving to the CSP */
                        if(ERR(err = cssm_ReplaceAppCallbacksWithCssmProxy( TempContext,
                                                        Context->NumberOfAttributes,
                                                        Context->ContextAttributes)) == CSSM_OK)
                        {
                            /* we now only need a read lock on the attach list entry */
                            tempHandle = TempContext->ContextHandle;

                            /* 
                             * callback to csp to notify context creation 
                             */
                            if(NULL != AttachRecord->AddInFunctions)
                            {
                                /* get the CSP callback functions */
                                ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                                      CSSM_SERVICE_CSP,
                                                                      (void *)&CallBack)); /* #ifdef VMS apply cast */
                                if(CSSM_OK == err)
                                {
                                    /* this function will lock the mutex if necessary */
                                    ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                                    if(CSSM_OK == err)
                                    {
                                        if(NULL != CallBack->EventNotify)
                                        {
                                            ERR(err = CallBack->EventNotify(CSPHandle,
                                                                            CSSM_CONTEXT_EVENT_CREATE,
                                                                            tempHandle,         /* this is a context handle */
                                                                            Context));
                                        }

                                        cssm_UnlockMultiTaskMutex(AttachRecord);
                                    }
                                }
                            }
                            else
                            {
                                /* there really should've been a function
                                   table in the attach record
                                */
                                ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
                            }


                            if(CSSM_OK == err)
                            {
                                /* insert context at the head of the list*/
                                TempContext->Next = AttachRecord->ContextList;
                                AttachRecord->ContextList = TempContext;
                            }
                        }

                    }

                    if(CSSM_OK != err)
                    {
                        /* clean up the lock since we're erroring out */
                        ERR(cssm_SWMRDelete(TempContext->hContextSWMRLock));
                    }
                }
            /*
             * Release the attach record.
             */
            cssm_ReleaseAttachLock(AttachRecord,
                                   CSSM_WRITER_LOCK);

            }

        }

        if(CSSM_OK != err)
        {
            /* Free the temp context since there was a problem - note that
               we're not screwing up the context list because there was no
               place for an error between where we add this to the list and
               here.  Be sure this assumption is maintained if this code is
               modified.
            */
            local_free (TempContext, 0);
        }

    /* set the return handle value - it's 0 if there's been an error */
    *pCCHandle = tempHandle;

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: cssm_FreeIntContext
 *
 *Description:
 *   Free memory that was allocated for internal context information.
 *   Writer lock on the context needs to be set before calling.
 *
 *Parameters: 
 *   Context (input) - pointer to internal context structure
 *
 *Returns:
 *   none
 *
 *-------------------------------------------------------------*/
void CSSMAPI
cssm_FreeIntContext(cssm_INTERNAL_CONTEXT_PTR *IntContext)
{
    TRACE(cssm_trace_info, FNCODE_cssm_FreeIntContext);

    if((NULL != IntContext) &&
       (NULL != *IntContext))
    {
        /* free the CSSM_CONTEXT */
        if(NULL != (*IntContext)->Context)
        {
            cssm_FreeContext(&(*IntContext)->Context, internal_free , NULL);
        }

        /*
         * Destroy the multiple-reader-single-write lock.
         */
        if(NULL != (*IntContext)->hContextSWMRLock)
        {
            cssm_SWMRLockDoneWriting((*IntContext)->hContextSWMRLock);
            cssm_SWMRDelete((*IntContext)->hContextSWMRLock);
            (*IntContext)->hContextSWMRLock = NULL;
        }
        
        if(NULL != (*IntContext)->SeedProxy)
        {
            local_free ((*IntContext)->SeedProxy, NULL);
        }
        if(NULL != (*IntContext)->PassphraseProxy)
        {
            local_free ((*IntContext)->PassphraseProxy, NULL);
        }
        if(NULL != (*IntContext)->ChallengeCbProxy)
        {
            local_free ((*IntContext)->ChallengeCbProxy, NULL);
        }

        local_free ((*IntContext), 0);

        (*IntContext) = NULL;
    }

    VOID_RETURN
}

/*---------------------------------------------------------------
 *
 *Name: cssm_FreeContext
 *
 *Description:
 *   Free memory that was allocated for user context information
 *
 *Parameters: 
 *   Context (input) - pointer to user context structure
 *
 *Returns:
 *   none
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_FreeContext(CSSM_CONTEXT_PTR *Context, CSSM_FREE free_func, void *mem_ref)
{
    CSSM_RETURN err = CSSM_OK;

    /* free the attributes and context structure */
    TRACE(cssm_trace_info, FNCODE_cssm_FreeContext);

    if((NULL == Context) ||
       (NULL == *Context))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        cssm_FreeAttributes((*Context)->NumberOfAttributes,
                            (*Context)->ContextAttributes,
                            free_func,
                            mem_ref,
                            1); /* delete the ContextAttributes list as well
                                   as the data held in the list
                                */
        CSSM_FREE_WRAPPER(free_func,*Context, mem_ref);
        *Context = NULL;
    }

    RETURN(err);
}

/*---------------------------------------------------------------
 *
 *Name: cssm_CreateAttribute
 *
 *Description:
 *   Create a pointer for the attribute structure
 *
 *Parameters: 
 *   Attributes (output) - pointer to receive the attribute structure
 *   NumberOfAttributes (input) - number of attributes memory to hold
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_CreateAttribute(CSSM_CONTEXT_ATTRIBUTE_PTR *Attributes,
                     uint32 NumberAttributes)
{
    CSSM_RETURN err = CSSM_OK;

    /* allocating the structure */
    TRACE(cssm_trace_info, FNCODE_cssm_CreateAttribute);

    if(NULL == Attributes)
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else if ( NumberAttributes == 0 )
    {
        *Attributes = NULL;
    }
    else
    {
        *Attributes = (CSSM_CONTEXT_ATTRIBUTE_PTR)local_calloc(NumberAttributes,
                                                               sizeof(CSSM_CONTEXT_ATTRIBUTE),
                                                               0);
        if (*Attributes == NULL)
        {
            ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
        }
    }

    RETURN(err)
}

/*---------------------------------------------------------------
 *
 *Name: cssm_AddAttribute
 *
 *Description:
 *   Add the attribute data to the pointer
 *
 *Parameters: 
 *   Attribute (output) - pointer to attribute info
 *   type (input) - type of attribute data
 *   Data (input) - pointer to attribute data
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI cssm_AddAttribute (CSSM_CONTEXT_ATTRIBUTE_PTR Attributes,
                                       CSSM_ATTRIBUTE_TYPE type,
                                       void *Data)
{
    CSSM_RETURN err = CSSM_OK;
    TRACE(cssm_trace_info, FNCODE_cssm_AddAttribute);

    switch ((int)type & CSSM_ATTRIBUTE_TYPE_MASK)
    {
        case CSSM_ATTRIBUTE_DATA_UINT32:
        {
            /* just assign the pointer to be a number - note that NULL
               is essentially a number and will be zero
            */
            Attributes->AttributeType = type;
            Attributes->Attribute.Uint32 = *(uint32*)Data;
            Attributes->AttributeLength = (uint32)sizeof (uint32);
        }
        break;

        case CSSM_ATTRIBUTE_DATA_CSSM_DATA:
        {
            CSSM_DATA_PTR Source = (CSSM_DATA_PTR)Data;
            uint32 i;

            Attributes->AttributeType = type;
            Attributes->AttributeLength = (uint32)sizeof (CSSM_DATA);
            ERR(err = cssm_CopyDataStruct(Source,
                                          &i,
                                          internal_malloc,
                                          internal_free,
                                          0,
                                          &Attributes->Attribute.Data));
        }
        break;

        case CSSM_ATTRIBUTE_DATA_CRYPTO_DATA:
        {
            CSSM_CRYPTO_DATA_PTR Source = (CSSM_CRYPTO_DATA_PTR)Data;

            Attributes->AttributeType = type;
            ERR( err = cssm_CopyCryptoDataStruct(Source,
                                                 &Attributes->AttributeLength,
                                                 internal_malloc,
                                                 internal_free,
                                                 0,
                                                 &Attributes->Attribute.CryptoData));
        }
        break;

        case CSSM_ATTRIBUTE_DATA_KEY:
        {
            CSSM_KEY_PTR Source = (CSSM_KEY_PTR)Data;

            Attributes->AttributeType = type;
            Attributes->AttributeLength = (uint32)sizeof (CSSM_KEY);
            ERR(err = cssm_CopyKeyStruct(Source,
                                         internal_malloc,
                                         internal_free,
                                         0,
                                         &Attributes->Attribute.Key));
        }
        break;

        case CSSM_ATTRIBUTE_DATA_STRING:
        {
            char *Description = (char*)Data;
            uint32 size = (uint32)strlen (Description);

            Attributes->Attribute.String = (char*)local_malloc(size+1,
                                                               0);

            if(NULL != Attributes->Attribute.String)
            {
                memcpy(Attributes->Attribute.String,
                       Description,
                       size);
                Attributes->Attribute.String[size] = 0;
                Attributes->AttributeType = type;
                Attributes->AttributeLength = size;
            }
            else
            {
                ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
            }
        }
        break;

        case CSSM_ATTRIBUTE_DATA_DATE:
        {
            CSSM_DATE_PTR Source = (CSSM_DATE_PTR)Data;

            Attributes->AttributeType = type;
            Attributes->AttributeLength = (uint32)sizeof (CSSM_DATE);
            ERR(err = cssm_CopyDateStruct(Source,
                                          internal_malloc,
                                          internal_free,
                                          0,
                                          &Attributes->Attribute.Date));
        }
        break;

        case CSSM_ATTRIBUTE_DATA_RANGE:
        {
            CSSM_RANGE_PTR Source = (CSSM_RANGE_PTR)Data;

            Attributes->AttributeType = type;
            Attributes->AttributeLength = (uint32)sizeof (CSSM_RANGE);
            ERR(err = cssm_CopyRangeStruct(Source,
                                           internal_malloc,
                                           0,
                                           &Attributes->Attribute.Range));
        }
        break;

        case CSSM_ATTRIBUTE_DATA_ACCESS_CREDENTIALS:
        {
            CSSM_ACCESS_CREDENTIALS_PTR SourceCred = (CSSM_ACCESS_CREDENTIALS_PTR)Data;
            CSSM_ACCESS_CREDENTIALS_PTR AccessCredentials; /* this is the new credential */

            /* assign the attribute information */
            Attributes->AttributeType = type;
            Attributes->AttributeLength = (uint32)sizeof(CSSM_ACCESS_CREDENTIALS);


            AccessCredentials = (CSSM_ACCESS_CREDENTIALS_PTR)local_malloc(sizeof(CSSM_ACCESS_CREDENTIALS),
                                                                   NULL);
            if(NULL == AccessCredentials)
            {
                ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
            }
            else
            {
                port_memcpy(AccessCredentials->EntryTag,
                            SourceCred->EntryTag,
                            sizeof(CSSM_STRING));
                AccessCredentials->Callback = SourceCred->Callback;
                AccessCredentials->CallerCtx = SourceCred->CallerCtx;

                ERR(err = cssmutil_DuplicateBaseCerts(&SourceCred->BaseCerts,
                                                      internal_malloc,
                                                      internal_free,
                                                      NULL,
                                                      &AccessCredentials->BaseCerts));
                if(CSSM_OK == err)
                {
                    ERR(cssmutil_DuplicateSampleGroup(&SourceCred->Samples,
                                                      internal_malloc,
                                                      internal_free,
                                                      NULL,
                                                      &AccessCredentials->Samples));
                }

                if(CSSM_OK != err)
                {
                    local_free(AccessCredentials,
                               NULL);
                    AccessCredentials = NULL;
                }
            }

            Attributes->Attribute.AccessCredentials = AccessCredentials;
        }
        break;

        case CSSM_ATTRIBUTE_DATA_VERSION:
        {
            Attributes->Attribute.Version = local_calloc(sizeof(CSSM_VERSION),
                                                         1,
                                                         NULL);
            if(NULL != Attributes->Attribute.Version)
            {
                Attributes->AttributeType = type;
                Attributes->AttributeLength = (uint32)sizeof(CSSM_VERSION);
                port_memcpy(Attributes->Attribute.Version,
                            Data,
                            sizeof(CSSM_VERSION));
            }
            else
            {
                ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
            }
        }
        break;

        case CSSM_ATTRIBUTE_DATA_DL_DB_HANDLE:
        {
            Attributes->Attribute.DLDBHandle = local_calloc(sizeof(CSSM_DL_DB_HANDLE),
                                                            1,
                                                            NULL);
            if(NULL != Attributes->Attribute.DLDBHandle)
            {
                Attributes->AttributeType = type;
                Attributes->AttributeLength = (uint32)sizeof(CSSM_DL_DB_HANDLE);
                port_memcpy(Attributes->Attribute.DLDBHandle,
                            Data,
                            sizeof(CSSM_DL_DB_HANDLE));
            }
            else
            {
                ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
            }
        }
        break;

        default:
        {
            ERR(err = CSSMERR_CSSM_INVALID_ATTRIBUTE);
        }
    }   /* end switch */

    RETURN(err)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_UpdateAlgParamContextAttribute
 *
 *Description:
 *     updates an attribute in the context handle.
 *
 *Parameters: 
 *   CCHandle - handle to context data
 *   NumberOfAttributes - number of attributes to update
 *   ContextAttributes - pointer to data that describes new attributes
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_UpdateAlgParamContextAttributes(CSSM_CONTEXT_PTR Context,
                                     uint32 NumberAttributes,
                                     const CSSM_CONTEXT_ATTRIBUTE *ContextAttributes)
{
    uint32 i=0;
    CSSM_RETURN retValue = CSSM_OK;

    if((CSSM_TRUE == port_IsBadWritePtr(Context,
                                        (uint32)sizeof(CSSM_CONTEXT))) ||
       (CSSM_TRUE == port_IsBadReadPtr(ContextAttributes,
                                       (uint32)sizeof(CSSM_CONTEXT_ATTRIBUTE) * NumberAttributes)))
    {
        ERR(retValue = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        /* Found the context, update the data */
        i = 0;
        while((CSSM_OK == retValue) &&
              (i < NumberAttributes))
        {
            ERR(retValue = cssm_UpdateContext (Context, &ContextAttributes[i]));
            i++;
        }
    }

    RETURN (retValue)
}

/*---------------------------------------------------------------
 *
 *Name: cssm_AssignContext
 *
 *Description:
 *     Free existing context associated with the CCHandle and 
       Adds new context to the internal list of contexts.  
 *
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_AssignContext(CSSM_CC_HANDLE CCHandle, 
                   const CSSM_CONTEXT *Context)
{
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_CONTEXT_PTR TempInternalContext;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_MODULE_HANDLE hAttach = CSSM_UINT64_HIGH( CCHandle );
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR tempContext = NULL;

    TRACE(cssm_trace_info, FNCODE_cssm_AssignContext);


    if(ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                              CSSM_NO_LOCK,
                                              NULL,
                                              CSSM_NO_LOCK,
                                              NULL,
                                              CSSM_NO_LOCK,
                                              &AttachRecord,
                                              CSSM_READER_LOCK,
                                              &TempInternalContext,
                                              CSSM_WRITER_LOCK)) == CSSM_OK)
    {
        /* 
         * duplicate context into internal structure 
         */
        ERR(err = cssm_DuplicateContext(Context,
                                        internal_malloc,
                                        internal_free,
                                        0,
                                        TempInternalContext,
                                        &tempContext));
        if(CSSM_OK == err)
        {
            ERR(err = cssm_ReplaceAppCallbacksWithCssmProxy(TempInternalContext,
                                                            tempContext->NumberOfAttributes,
                                                            tempContext->ContextAttributes));
            if(CSSM_OK == err)
            {
                /* 
                 * callback to csp to notify context update
                 */
                ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                      CSSM_SERVICE_CSP,
                                                      (void **)&CallBack));    /* #ifdef VMS apply cast */
                if(CSSM_OK == err)
                {
                    /* this function will lock the mutex if necessary */
                    ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                    if(CSSM_OK == err)
                    {
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->EventNotify))
                        {
                            ERR(err = CallBack->EventNotify(hAttach,
                                                            CSSM_CONTEXT_EVENT_UPDATE,
                                                            CCHandle,
                                                            tempContext));
                            if(CSSM_OK == err)
                            {
                                /* now we free up the old info and insert the new info
                                   since the CSP said it was acceptable
                                */
                                cssm_FreeAttributes(TempInternalContext->Context->NumberOfAttributes, 
                                                    TempInternalContext->Context->ContextAttributes,
                                                    internal_free,
                                                    0,
                                                    1);
                                local_free(TempInternalContext->Context,
                                           NULL);
                                TempInternalContext->Context = tempContext;
                            }
                        }

                        cssm_UnlockMultiTaskMutex(AttachRecord);
                    }
                }

                if(CSSM_OK != err)
                {
                    /* the update didn't work, so we revert to the old
                       values
                    */
                    ERR(cssm_ReplaceAppCallbacksWithCssmProxy(TempInternalContext,
                                                              TempInternalContext->Context->NumberOfAttributes,
                                                              TempInternalContext->Context->ContextAttributes));
                }
            }

            if(CSSM_OK != err)
            {
                ERR(cssm_FreeContext(&tempContext, internal_free, NULL));
            }
        }

        /*
         * Release the attach and context records.
         */
        cssm_ReleaseAttachLock(AttachRecord,
                               CSSM_READER_LOCK);
        cssm_ReleaseContextLock(TempInternalContext,
                                CSSM_WRITER_LOCK);
    }

    return err;
}

/*---------------------------------------------------------------
 *
 *Name: cssm_IsBadCryptoKeyPtr
 *
 *Description:
 *   Checks if the Key pointers are valid

 *Parameters: 
 *
 *Returns:
 *   CSSM_FALSE - Invalid key pointers
 *   CSSM_TRUE - Valid key pointer
 *
 *-------------------------------------------------------------*/
CSSM_BOOL CSSMAPI
cssm_IsBadCryptoKeyPtr(const CSSM_KEY *Key)
{
    CSSM_BOOL ret = CSSM_TRUE;

    TRACE(cssm_trace_info, FNCODE_cssm_IsBadCryptoKeyPtr);

    if(CSSM_FALSE == port_IsBadReadPtr(Key,
                                       (uint32)sizeof(CSSM_KEY)))
    {
        if(CSSM_FALSE == port_IsBadReadPtr(Key->KeyData.Data,
                                           Key->KeyData.Length))
        {
            ret = CSSM_FALSE;
        }
    }

    RETURN(ret);
}



/*---------------------------------------------------------------
 *
 *Name: cssm_IsBadCryptoDataPtr
 *
 *Description:
 *   Checks if the cryptodata pointers are valid

 *Parameters: 
 *
 *Returns:
 *   CSSM_FALSE - Invalid crypto data pointers
 *   CSSM_TRUE - Valid cryptodata pointer
 *
 *-------------------------------------------------------------*/
CSSM_BOOL CSSMAPI
cssm_IsBadCryptoDataPtr(const CSSM_CRYPTO_DATA *Data)
{
    CSSM_BOOL ret = CSSM_TRUE;

    TRACE(cssm_trace_info, FNCODE_cssm_IsBadCryptoDataPtr);

    if(CSSM_FALSE == port_IsBadReadPtr(Data,
                                       (uint32)sizeof (CSSM_CRYPTO_DATA)))
    {
        /* If there's not a callback, we must have a good Param value.
           Otherwise we'll just check the callback function for validity.
        */
        if(NULL == Data->Callback)
        {
            if(CSSM_FALSE == port_IsBadReadPtr(Data->Param.Data,
                                               Data->Param.Length))
            {
                ret = CSSM_FALSE;
            }
        }
        else
        {
            if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)Data->Callback))
            {
                ret = CSSM_FALSE;
            }
        }
    }

    RETURN(ret)
}



/*---------------------------------------------------------------
 *
 *Name: cssm_UpdateAttribute
 *
 *Description:
 *  Updates the attributes within the Context structure
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_UpdateAttribute(CSSM_CONTEXT_ATTRIBUTE_PTR Des,
                     const CSSM_CONTEXT_ATTRIBUTE *Src)
{
    CSSM_RETURN err = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_cssm_UpdateAttribute);

    switch(Src->AttributeType & CSSM_ATTRIBUTE_TYPE_MASK)
    {
        case CSSM_ATTRIBUTE_DATA_STRING:
        {
            if(0 != Src->AttributeLength)
            {
                if((Des->Attribute.String = (char*)local_realloc(Des->Attribute.String,
                                                                Src->AttributeLength+1,
                                                                0)) == NULL)
                {
                    ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                }
                else
                {
                    memcpy(Des->Attribute.String,
                           Src->Attribute.String,
                           Src->AttributeLength);
                    Des->Attribute.String[Src->AttributeLength] = 0;
                    Des->AttributeLength = Src->AttributeLength;
                }
            }
            else
            {
                if(NULL != Des->Attribute.String)
                {
                    local_free(Des->Attribute.String,
                               NULL);
                    Des->Attribute.String = NULL;
                }
                Des->AttributeLength = 0;
            }
        }
        break;

        case CSSM_ATTRIBUTE_DATA_KEY:
        {
            CSSM_KEY_PTR Key = Src->Attribute.Key;
            
            if( Des->Attribute.Key )
            {
                port_memcpy(&(Des->Attribute.Key->KeyHeader),
                            &(Key->KeyHeader),
                            sizeof(CSSM_KEYHEADER));

                if(0 != Key->KeyData.Length)
                {
                    if((Des->Attribute.Key->KeyData.Data = (uint8*)local_realloc(Des->Attribute.Key->KeyData.Data,
                                                  Key->KeyData.Length,
                                                  0)) == NULL)
                    {
                        ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                    }
                    else
                    {
                        port_memcpy(Des->Attribute.Key->KeyData.Data,
                                    Key->KeyData.Data,
                                    Key->KeyData.Length);
                        Des->Attribute.Key->KeyData.Length = Key->KeyData.Length;
                    }
                }
                else
                {
                    if(NULL != Des->Attribute.Key->KeyData.Data)
                    {
                        local_free(Des->Attribute.Key->KeyData.Data,
                                   NULL);
                        Des->Attribute.Key->KeyData.Data = NULL;
                    }
                    Des->Attribute.Key->KeyData.Length = 0;
                }
            }
        }
        break;

        case CSSM_ATTRIBUTE_DATA_UINT32:
        {
            Des->Attribute.Uint32 = Src->Attribute.Uint32;
        }
        break;

        case CSSM_ATTRIBUTE_DATA_CSSM_DATA:
        {
            CSSM_DATA_PTR Source = Src->Attribute.Data, 
                                Dest = Des->Attribute.Data;

            if(0 < Source->Length)
            {
                if((Dest->Data = (uint8*)local_realloc(Dest->Data,
                                                      Source->Length,
                                                      0)) == NULL)
                {
                    ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                }
                else
                {
                    Dest->Length = Source->Length;
                    ERR(err = cssm_CheckCopy(Dest->Data,
                                             Source->Data, 
                                             Source->Length));
                }
            }
            else
            {
                if(NULL != Dest->Data)
                {
                    local_free(Dest->Data,
                               NULL);
                    Dest->Data = NULL;
                }
                Dest->Length = 0;
            }
        }
        break;

        case CSSM_ATTRIBUTE_DATA_CRYPTO_DATA:
        {
            CSSM_CRYPTO_DATA_PTR Source = Src->Attribute.CryptoData, 
                                 Dest = Des->Attribute.CryptoData;

            Dest->Callback = Source->Callback;
            Dest->CallerCtx = Source->CallerCtx;

            if((0 == Source->Param.Length) ||
               (CSSM_TRUE == port_IsBadReadPtr(Source->Param.Data,
                                               Source->Param.Length)))
            {
                if(NULL != Dest->Param.Data)
                {
                    local_free(Dest->Param.Data,
                               NULL);
                    Dest->Param.Data = NULL;
                }
                Dest->Param.Length = 0;
            }
            else
            {
                if((Dest->Param.Data = (uint8*)local_realloc(Dest->Param.Data,
                                                            Source->Param.Length,
                                                            0)) == NULL)
                {
                    ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                }
                else
                {
                    Dest->Param.Length = Source->Param.Length;
                    ERR(err = cssm_CheckCopy(Dest->Param.Data,
                                             Source->Param.Data,
                                             Source->Param.Length));
                    if(CSSM_OK != err)
                    {
                        local_free(Dest->Param.Data, NULL);
                        Dest->Param.Data = NULL;
                        Dest->Param.Length = 0;
                    }
                }
            }
        }
        break;

        case CSSM_ATTRIBUTE_DATA_DATE:
        {
            memcpy(Des->Attribute.Date,
                   Src->Attribute.Date,
                   sizeof(CSSM_DATE));
        }
        break;

        case CSSM_ATTRIBUTE_DATA_RANGE:
        {
            memcpy(Des->Attribute.Range,
                   Src->Attribute.Range,
                   sizeof(CSSM_RANGE));
        }
        break;

        case CSSM_ATTRIBUTE_DATA_ACCESS_CREDENTIALS:
        {
            CSSM_CONTEXT_ATTRIBUTE_PTR tempAccessCred = NULL;

            /* free the old access credentials */
            cssm_FreeAttributes(1, /* NumberOfAttributes */
                                Des, /* Attributes */
                                internal_free,
                                NULL,
                                0); /* don't free top structure */

            /* now we copy the access credential, but this function
               over-allocates for what we need, so we free the top level of
               the return value
            */
            ERR(err = cssm_DuplicateAttributes(1, /* Number */
                                               (CSSM_CONTEXT_ATTRIBUTE_PTR)Src, /* Attributes */
                                               internal_malloc,
                                               internal_free,
                                               NULL,
                                               &tempAccessCred));
            if(CSSM_OK == err)
            {
                /* now we copy the top level data and then free the top
                   level
                */
                memcpy(Des,
                       tempAccessCred,
                       sizeof(CSSM_CONTEXT_ATTRIBUTE));
                local_free(tempAccessCred,
                           NULL);
            }
        }
        break;

        case CSSM_ATTRIBUTE_DATA_VERSION:
        {
            memcpy(Des->Attribute.Version,
                   Src->Attribute.Version,
                   sizeof(CSSM_VERSION));
        }
        break;

        case CSSM_ATTRIBUTE_DATA_DL_DB_HANDLE:
        {
            memcpy(Des->Attribute.DLDBHandle,
                   Src->Attribute.DLDBHandle,
                   sizeof(CSSM_DL_DB_HANDLE));
        }
        break;

        default:
            ERR(err = CSSMERR_CSSM_INVALID_ATTRIBUTE);
    }

    RETURN(err)
}

/*---------------------------------------------------------------
 *
 *Name: cssm_UpdateContext
 *
 *Description:
 *
 *Parameters: 
 *
  *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_UpdateContext(CSSM_CONTEXT_PTR Context, 
                   const CSSM_CONTEXT_ATTRIBUTE *Attribute)
{
    uint32 i;
    CSSM_RETURN err = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_cssm_UpdateContext);

    for (i = 0; i<Context->NumberOfAttributes; i++)
    {
        if (Context->ContextAttributes[i].AttributeType == Attribute->AttributeType)
        {
            RETURN_EXPRESSION(cssm_UpdateAttribute (&Context->ContextAttributes[i], Attribute))
        }
    }

    Context->NumberOfAttributes++;
    if((Context->ContextAttributes = (CSSM_CONTEXT_ATTRIBUTE_PTR)local_realloc(Context->ContextAttributes,
                                                                              (uint32)sizeof (CSSM_CONTEXT_ATTRIBUTE) * Context->NumberOfAttributes, 0)) == NULL)
    {
        ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
    }
    else
    {
        if(!port_IsBadReadPtr(Attribute,
                              (uint32)sizeof (CSSM_CONTEXT_ATTRIBUTE)))
        {
            if(CSSM_ATTRIBUTE_DATA_UINT32 != (Attribute->AttributeType & CSSM_ATTRIBUTE_TYPE_MASK))
            {
                ERR(err = cssm_AddAttribute(&Context->ContextAttributes[Context->NumberOfAttributes - 1],
                                            (CSSM_ATTRIBUTE_TYPE)Attribute->AttributeType,
                                            (void *)Attribute->Attribute.Data));
            }
            else
            {
                ERR(err = cssm_AddAttribute(&Context->ContextAttributes[Context->NumberOfAttributes - 1],
                                            (CSSM_ATTRIBUTE_TYPE)Attribute->AttributeType,
                                            (void *)&Attribute->Attribute.Uint32));
            }

        }
        else
        {
            ERR(err = CSSMERR_CSSM_INVALID_POINTER);
        }
    }
    if( CSSM_OK != err)
    {
        Context->NumberOfAttributes--;
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: cssm_DelContextAttr
 *
 *Description:  This function removes the attribute identified by
 *  Attribute from the Context parameter.
 *
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_DelContextAttr(CSSM_CONTEXT_PTR Context,
                    const CSSM_CONTEXT_ATTRIBUTE *Attribute)
{
    uint32 i;
    CSSM_RETURN err = CSSM_OK;
    CSSM_BOOL AttributePresent = CSSM_FALSE;

    TRACE(cssm_trace_info, FNCODE_cssm_DelContextAttr);

    if((NULL == Context) ||
       (NULL == Attribute))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        i = 0;

        /* prepare to loop through all of the attributes */
        while((i<Context->NumberOfAttributes) &&
              (CSSM_OK == err))
        {
            /* see if this is the correct attribute type and length */
            if(Context->ContextAttributes[i].AttributeType == Attribute->AttributeType)
            {
                AttributePresent = CSSM_TRUE;

                /* free the attribute - remove it from the context */
                cssm_FreeAttributes(1,
                                    &Context->ContextAttributes[i],
                                    internal_free,
                                    0,
                                    0);

                /* decrease the number of attributes */
                Context->NumberOfAttributes--;


                /* Now we shift all the attributes "down" one.  We increment
                   i inside this loop so we'll get to the end of the array and
                   end the larger while loop when we leave this.
                */
                while((i<Context->NumberOfAttributes) &&
                      (CSSM_OK == err))
                {
                    ERR(err = cssm_CheckCopy(&Context->ContextAttributes[i],
                                             &Context->ContextAttributes[i+1],
                                             (uint32)sizeof (CSSM_CONTEXT_ATTRIBUTE)));
                    i++;
                }
            }

            i++;
        }

        if((CSSM_FALSE == AttributePresent) &&
           (CSSM_OK == err))
        {
            ERR(err = CSSMERR_CSSM_ATTRIBUTE_NOT_IN_CONTEXT);
        }
    }

    return err;
}



/*---------------------------------------------------------------
 *
 *Name: cssm_SaveContextClone
 *
 *Description:
 *  Clone the context and save it in the internal context list
 *Parameters: 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_SaveContextClone(CSSM_CSP_HANDLE CSPHandle,
                      cssm_INTERNAL_CONTEXT_PTR InternalCtx,
                      CSSM_CC_HANDLE CCHandle,
                      cssm_INTERNAL_CONTEXT_PTR *ContextList)
{
    CSSM_RETURN err = CSSM_OK;
    CSSM_CONTEXT_PTR Context; 
    cssm_INTERNAL_CONTEXT_PTR TempContext;

    TRACE(cssm_trace_info, FNCODE_cssm_SaveContextClone);

    if((TempContext = (cssm_INTERNAL_CONTEXT_PTR)local_malloc((uint32)sizeof (cssm_INTERNAL_CONTEXT),
                                                              0)) == NULL)
    {
        ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
    }
    else
    {
        if(ERR(err = cssm_DuplicateContext(InternalCtx->Context,
                                           internal_malloc,
                                           internal_free,
                                           0,
                                           InternalCtx,
                                           &Context)) == CSSM_OK)
        {
            TempContext->Context = Context;
            TempContext->CSPHandle = CSPHandle;
            TempContext->Next = NULL;
            TempContext->hContextSWMRLock = NULL;

            /* assign handle for context */
            TempContext->ContextHandle = CCHandle;
            TempContext->SeedProxy = NULL;
            TempContext->PassphraseProxy = NULL;
            TempContext->ChallengeCbProxy = NULL;

            if(NULL != InternalCtx->SeedProxy)
            {
                if(NULL == (TempContext->SeedProxy = (CSSM_SEED_CALLBACK_PROXY *)local_malloc((uint32)sizeof(CSSM_SEED_CALLBACK_PROXY),
                                                                                  NULL)))
                {
                    ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                }
                else
                {
                    TempContext->SeedProxy->CCHandle = TempContext->ContextHandle;
                    TempContext->SeedProxy->Callback = InternalCtx->SeedProxy->Callback;
                    TempContext->SeedProxy->CallbackCtx = InternalCtx->SeedProxy->CallbackCtx;
                }
            }

            if((CSSM_OK == err) &&
               (NULL != InternalCtx->PassphraseProxy))
            {
                if(NULL == (TempContext->PassphraseProxy = (CSSM_PASSPHRASE_PROXY *)local_malloc((uint32)sizeof(CSSM_PASSPHRASE_PROXY),
                                                                                  NULL)))
                {
                    ERR(err = CSSMERR_CSSM_MEMORY_ERROR);

                    if(NULL != TempContext->SeedProxy)
                    {
                        local_free(TempContext->SeedProxy,
                                   NULL);
                        TempContext->SeedProxy = NULL;
                    }
                }
                else
                {
                    TempContext->PassphraseProxy->CCHandle = TempContext->ContextHandle;
                    TempContext->PassphraseProxy->Callback = InternalCtx->PassphraseProxy->Callback;
                    TempContext->PassphraseProxy->CallbackCtx = InternalCtx->PassphraseProxy->CallbackCtx;
                }
            }

            if((CSSM_OK == err) &&
               (NULL != InternalCtx->ChallengeCbProxy))
            {
                if(NULL == (TempContext->ChallengeCbProxy = (CSSM_CHALLENGE_CALLBACK_PROXY *)local_malloc((uint32)sizeof(CSSM_CHALLENGE_CALLBACK_PROXY),
                                                                                  NULL)))
                {
                    ERR(err = CSSMERR_CSSM_MEMORY_ERROR);

                    if(NULL != TempContext->SeedProxy)
                    {
                        local_free(TempContext->SeedProxy,
                                   NULL);
                        TempContext->SeedProxy = NULL;
                    }

                    if(NULL != TempContext->PassphraseProxy)
                    {
                        local_free(TempContext->PassphraseProxy,
                                   NULL);
                        TempContext->PassphraseProxy = NULL;
                    }
                }
                else
                {
                    TempContext->ChallengeCbProxy->CCHandle = TempContext->ContextHandle;
                    TempContext->ChallengeCbProxy->Callback = InternalCtx->ChallengeCbProxy->Callback;
                    TempContext->ChallengeCbProxy->CallbackCtx = InternalCtx->ChallengeCbProxy->CallbackCtx;
                }
            }

            if(CSSM_OK == err)
            {
                /*
                 * Create the multiple-reader-single-write
                 * synchronization object.
                 */
                if(ERR(err = cssm_SWMRInit(&TempContext->hContextSWMRLock)) == CSSM_OK)
                {
                    /* set the new context at the head of the context list */
                    TempContext->Next = *ContextList;
                    *ContextList = TempContext;
                }
            }

            if(CSSM_OK != err)
            {
                /* clean up memory */
                if(NULL != TempContext->SeedProxy)
                {
                    local_free(TempContext->SeedProxy,
                               NULL);
                    TempContext->SeedProxy = NULL;
                }

                if(NULL != TempContext->PassphraseProxy)
                {
                    local_free(TempContext->PassphraseProxy,
                               NULL);
                    TempContext->PassphraseProxy = NULL;
                }
                if(NULL != TempContext->ChallengeCbProxy)
                {
                    local_free(TempContext->ChallengeCbProxy,
                               NULL);
                    TempContext->ChallengeCbProxy = NULL;
                }
                cssm_FreeContext(&Context, internal_free, NULL);
            }
        }

        if(CSSM_OK != err)
        {
            local_free(TempContext, NULL);
        }
    }

    return err;
}



/*---------------------------------------------------------------
 *
 *Name: cssm_DeleteContext
 *
 *Description:
 *   Internal function for deleting the internal respresentation of
 *   the context
 *
 *Parameters: 
 *   CCHandle - handle to a created context
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_DeleteContext(cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord,
                   cssm_INTERNAL_CONTEXT_PTR *ContextRecord)
{
    cssm_INTERNAL_CONTEXT_PTR IntContextLead = NULL;
    cssm_INTERNAL_CONTEXT_PTR IntContextTrailer = NULL;
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_RETURN Ret = CSSM_OK;
    CSSM_CC_HANDLE CCHandle;

    /* get the callback table for this attach */
    if(ERR(Ret = cssm_GetAddInCallBackByType(AttachRecord,
                                             CSSM_SERVICE_CSP,
                                             (void**)&CallBack)) == CSSM_OK)
    {
        CCHandle = (*ContextRecord)->ContextHandle;

        IntContextLead = IntContextTrailer = AttachRecord->ContextList;

        while((NULL != IntContextLead) &&
              (CCHandle != IntContextLead->ContextHandle))
        {
            IntContextTrailer = IntContextLead;
            IntContextLead = IntContextLead->Next;
        }

        if(NULL == IntContextLead)
        {
            /* We couldn't find the list entry */
            ERR(Ret = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
        }
        else
        {
            /* if we're deleting the first node */
            if(IntContextLead == IntContextTrailer)
            {
                AttachRecord->ContextList = IntContextLead->Next;
            }
            else
            {
                IntContextTrailer->Next = IntContextLead->Next;
            }

            /* this function will lock the mutex if necessary */
            ERR(Ret = cssm_LockMultiTaskMutex(AttachRecord));

            if(CSSM_OK == Ret)
            {
                /* We've removed the entry from the list, now we're going
                   to free all the memory in the context record - including
                   deleting the context locks.
                */
                if(NULL != CallBack->EventNotify)
                {
                
                    ERR(Ret= CallBack->EventNotify(AttachRecord->AttachHandle,
                                                   CSSM_CONTEXT_EVENT_DELETE,
                                                   CCHandle,
                                                   (*ContextRecord)->Context));
                }

                cssm_FreeIntContext(ContextRecord);

                cssm_UnlockMultiTaskMutex(AttachRecord);
            }
        }

    }
    return Ret;
}
