/*-----------------------------------------------------------------------
 * File: EMMSPI.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
 * This file contains the functions that are contained in the SPI portion
 * of the EMM exported functions.
 */

#include "precomp.h"



/*---------------------------------------------------------------
 *
 *Name: CSSM_DeliverModuleManagerEvent
 *
 *Description:
 *  This function requests that CSSM core services deliver an event
 *  notification to a module manager
 *
 *Parameters: 
 *  EventDescription - Pointer describing the event and sender/receiver of the event.
 *
 *Returns:
 *   CSSM_OK - event delivered.
 *   All other returns indicate errors.
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_DeliverModuleManagerEvent(
                      const CSSM_MANAGER_EVENT_NOTIFICATION *EventDescription)
{
    CSSM_RETURN err = CSSM_OK;

    uint32 NumPotentialSrcEMMs;
    cssm_INTERNALMANAGERLIST_PTR PossibleSourceEMMRecords = NULL;

    uint32 NumDestEMMs = 0;
    cssm_INTERNALMANAGERLIST_PTR ppDestEMMRecords = NULL;

    uint32 i;

#ifdef ISL_INCLUDED
    CSSM_BOOL SrcModuleFound;
    uint32 j;

    ISL_FUNCTION_PTR retAddress;
    EISL_GetReturnAddress(retAddress);
#endif

    if(CSSM_TRUE == (port_IsBadReadPtr(EventDescription,
                                       (uint32)sizeof(CSSM_MANAGER_EVENT_NOTIFICATION))))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        /*
         * Get the global reader lock.
         */
        if(cssm_ReaderLock (hEMMListSWMRLock) != CSSM_OK)
        {
            ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
        }
        else
        {

            /* We get a list of all known (and thus possible) destinations
               for this event.  We do this by finding out how many EMMs respond
               to this service type.  This results in a shallow copy of manager
               records.
            */
            ERR(err = cssm_GetEMMManagerRecordsByType(EventDescription->DestinationModuleManagerType,
                                                      &NumDestEMMs,
                                                      &ppDestEMMRecords));
            if(CSSM_OK == err)
            {
                if(0 == NumDestEMMs)
                {
                    ERR(err = CSSMERR_CSSM_INVALID_SERVICE_MASK);
                }
                else
                {
                    /* We're going to see if we actually know this caller.
                       We'll check the type against known modules and do a
                       return address check.  All we care about here is the
                       fact that we do know this module, so we'll release the
                       module and associated memory as soon as we know of it -
                       we won't use it again.
                    */

                    ERR(err = cssm_GetEMMManagerRecordsByType(EventDescription->SourceModuleManagerType,
                                                              &NumPotentialSrcEMMs,
                                                              &PossibleSourceEMMRecords));
                    if(CSSM_OK == err)
                    {
                        if(0 == NumPotentialSrcEMMs)
                        {
                            ERR(err = CSSMERR_CSSM_INVALID_SERVICE_MASK);
                        }
                        else
                        {

#ifdef ISL_INCLUDED
                            if( cssm_GetPVCPolicy() & CSSM_PVC_SP )
                            {

                                /* now we determine which potential EMM is our
                                   actual EMM
                                */
                                i = 0;
                                SrcModuleFound = CSSM_FALSE;

                                /* check every EMM to find the right one */
                                while((i < NumPotentialSrcEMMs) &&
                                      (CSSM_FALSE == SrcModuleFound) &&
                                      (CSSM_OK == err))
                                {
                                    ERR(err = cssm_ReaderLock(PossibleSourceEMMRecords[i].hEMMRecordSWMRLock));

                                    if(CSSM_OK == err)
                                    {
                                        j = 0;

                                        /* check every section in this EMM */
                                        while((j < PossibleSourceEMMRecords[i].NumEmmSections) &&
                                              (CSSM_FALSE == SrcModuleFound))
                                        {
                                            if(EISL_CheckAddressWithinModule(PossibleSourceEMMRecords[i].EMMCredentialPtr[j],
                                                                             (ISL_FUNCTION_PTR)retAddress) == ISL_OK)
                                            {
                                                SrcModuleFound = CSSM_TRUE;
                                            }
                                            else
                                            {
                                                j++;
                                            }
                                        }

                                        cssm_ReaderUnlock(PossibleSourceEMMRecords[i].hEMMRecordSWMRLock);
                                    }

                                    i++;
                                }

                                if(CSSM_FALSE == SrcModuleFound)
                                {
                                    /* we checked all the EMMs and they weren't
                                       in the list, so we free the shallow copy we
                                       made for the records
                                    */
                                    ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                                }
                            }
#endif

                            local_free(PossibleSourceEMMRecords,
                                       NULL);
                        }
                    }

                    /* there's been an error, so we release the shallow copy
                       we made for these records
                    */
                    if(CSSM_OK != err)
                    {
                        local_free(ppDestEMMRecords,
                                   NULL);
                    }
                }
            }

            /* there's been an error, so we release the global lock */
            if(CSSM_OK != err)
            {
                cssm_ReaderUnlock(hEMMListSWMRLock);
            }
        }
    }


    if(CSSM_OK == err)
    {
        /* At this point, we have established that this event is authorized
           to be forwarded, and we have a list of modules to receive this
           event.  We'll now go through the destinations locking and sending
           the event.
        */

        i = 0;

        while((i < NumDestEMMs) &&
              (CSSM_OK == err))
        {
            ERR(err = cssm_ReaderLock(ppDestEMMRecords[i].hEMMRecordSWMRLock));

            if(CSSM_OK == err)
            {
                if(NULL != ppDestEMMRecords[i].EMMJT->EventNotifyManager)
                {
                    ERR(err = ppDestEMMRecords[i].EMMJT->EventNotifyManager(EventDescription));
                }

                cssm_ReaderUnlock(ppDestEMMRecords[i].hEMMRecordSWMRLock);
            }

            i++;
        }

        /* Now we're finished and can finally release the global lock.  We
           unlocked it earlier if there was an error, but it's still around if
           we had success.  We can also free up the shallow destination copy
           we made earlier.
        */
        cssm_ReaderUnlock(hEMMListSWMRLock);

        local_free(ppDestEMMRecords,
                   NULL);
    }

    return err;
}
