/*-----------------------------------------------------------------------
 * File: SHA1.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
**	SHA-1 algorithm glue
*/
#include "cssm.h"
#include "integapi.h"
#include "guids.h"
#include "isl_cssm.h"
#include <isl_memtrack.h>

/* 
**	The SHA-1 OID is defined in the OIW/SECSIG stable agreements Mar, 1995.
**	URL: http://nemo.ncsl.nist.gov/oiw/agreements/stable/OSI/12s_9506.txt
**	iso(1) identified-organization(3) oiw(14) secsig(3) algorithm(2) sha1(26)
*/
extern CSSM_CSP_HANDLE gCSPHandle;

static unsigned char sha1oid[] = { 6, 5, 1 * 40 + 3, 14, 3, 2, 26};

static
void pkcs7identify(
	ISL_SERVICE_CLASS *algClass,				
	ISL_CONST_DATA *algID,				
	ISL_CONST_DATA *serviceName)
{
	*algClass = ISL_ServiceDigest;
	algID->Data = sha1oid;
	algID->Length = sizeof(sha1oid);
	serviceName->Data = (const unsigned char *)"SHA-1";             /* ifdef VMS apply cast */
	serviceName->Length = strlen((const char *)serviceName->Data);  /* ifdef VMS apply cast */
}
static
void jaridentify(
	ISL_SERVICE_CLASS *algClass,				
	ISL_CONST_DATA *algID,				
	ISL_CONST_DATA *serviceName)
{
	*algClass = ISL_ServiceDigest;
	algID->Data = (const unsigned char *)"SHA-1";                   /* ifdef VMS apply cast */
	algID->Length = strlen((const char *)algID->Data);              /* ifdef VMS apply cast */
	serviceName->Data = (const unsigned char *)"SHA-1";             /* ifdef VMS apply cast */
	serviceName->Length = strlen((const char *)serviceName->Data);  /* ifdef VMS apply cast */
}

struct shacontext {
	CSSM_CSP_HANDLE hCSP;
	CSSM_CC_HANDLE hCssmContext;
	ISL_CSSM_METHODS *pCssmMethods;
};

static ISL_SIZE contextsize()
{
	return sizeof(struct shacontext);
}

static
ISL_STATUS init(
	ISL_DIGEST_SERVICE_CONTEXT *memory,
	ISL_CONST_DATA parameters,
	ISL_SIZE *sizeofDigest)
{
	return ISL_FAIL;
}

static
ISL_ERROR_CODE update(
	ISL_DIGEST_SERVICE_CONTEXT *context,
	ISL_CONST_DATA data)
{
	int digestLength = 20;
	CSSM_RETURN status;
	CSSM_DATA ptr;
	struct shacontext *sha = (struct shacontext*) context;
	ptr.Data = (uint8 *)data.Data;
	ptr.Length = data.Length;
	status = sha->pCssmMethods->CSSM_DigestDataUpdate(sha->hCssmContext, &ptr, 1); 
	switch (status) {
		case CSSM_OK:
			break;
		case CSSMERR_CSP_INVALID_ALGORITHM:
			return ISL_UNSUPPORTED_ALGORITHM;
		default:
			return ISL_INVALID_INPUT;
	}
	return ISL_OK;
}

static
ISL_ERROR_CODE recycle(
	ISL_DIGEST_SERVICE_CONTEXT *context,
	ISL_DATA_PTR digest)
{

	struct shacontext *sha = (struct shacontext*) context;
	CSSM_DATA ptr;
	CSSM_RETURN status;
	ptr.Data = digest->Data;
	ptr.Length = digest->Length;
	status = sha->pCssmMethods->CSSM_DigestDataFinal(sha->hCssmContext, &ptr);
	if (status != CSSM_OK) {
		return ISL_INVALID_INPUT;
	}
	if (sha->pCssmMethods->CSSM_DeleteContext(sha->hCssmContext) != CSSM_OK) 
		return ISL_INVALID_INPUT;
	sha->hCssmContext = 0;
	sha->hCSP = 0;
	return ISL_OK;

}

static
ISL_STATUS initializeWithClass(
		ISL_DIGEST_SERVICE_CONTEXT *Memory,
		ISL_CONST_DATA Parameters,
		ISL_SIZE *SizeofDigest,
		ISL_CLASS_PTR Class) {
		
	CSSM_RETURN status;
	struct shacontext *context = (struct shacontext*)Memory;
	ISL_CSSM_METHODS *pCssmMethods;
	ISL_CSSM_MODULE_HANDLE *pCspHandle;

	if (!Class)
		return ISL_FAIL;
	else if (!Class->ClassContext)
		return ISL_FAIL;
	else if(!((ISL_DIGEST_CLASS_CONTEXT *)Class->ClassContext)->CSPList.ModuleHandles)
		return ISL_FAIL;
	else if(!((ISL_DIGEST_CLASS_CONTEXT *)Class->ClassContext)->CSPList.NumberOfHandles)
		return ISL_FAIL;
	else if(!((ISL_DIGEST_CLASS_CONTEXT *)Class->ClassContext)->CssmMethods)
		return ISL_FAIL;

	*SizeofDigest = 20;
	
	pCssmMethods = ((ISL_DIGEST_CLASS_CONTEXT *)Class->ClassContext)->CssmMethods;


	//use the first csp handle in the list to perform the operation
	pCspHandle = ((ISL_DIGEST_CLASS_CONTEXT *)Class->ClassContext)->CSPList.ModuleHandles;
	context->hCSP = pCspHandle->Handle;
	if (context->hCSP == 0)
		return ISL_FAIL;

	context->pCssmMethods = pCssmMethods; 

	//create digest context handle
	status = context->pCssmMethods->CSSM_CSP_CreateDigestContext(
		context->hCSP, CSSM_ALGID_SHA1, &context->hCssmContext);
	if( (status != CSSM_OK ) || (context->hCssmContext == 0) ) {
		return ISL_FAIL;
	}

	status = context->pCssmMethods->CSSM_DigestDataInit(context->hCssmContext);
	if (status != CSSM_OK)
		return ISL_FAIL;
	return ISL_OK;
}	

ISL_DIGEST_METHODS PKCS7SHA1Methods = { 
	{ pkcs7identify },	
	contextsize,
	init,
	update,
	recycle,
	initializeWithClass };

ISL_DIGEST_METHODS JarSHA1Methods = { 
	{ jaridentify },	
	contextsize,
	init,
	update,
	recycle,
	initializeWithClass };

