/*-----------------------------------------------------------------------
* File: KEYFUNCS.C
*
* Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
*-----------------------------------------------------------------------
*/
/*
* INTEL CONFIDENTIAL 
* This file, software, or program is supplied under the terms of a 
* license agreement or nondisclosure agreement with Intel Corporation 
* and may not be copied or disclosed except in accordance with the 
* terms of that agreement. This file, software, or program contains 
* copyrighted material and/or trade secret information of Intel 
* Corporation, and must be treated as such. Intel reserves all rights 
* in this material, except as the license agreement or nondisclosure 
* agreement specifically indicate. 
*/
/* 
* WARNING: EXPORT RESTRICTED. 
* This software listing contains cryptographic methods and technology. 
* It is export restricted by the Office of Defense Trade Controls, United 
* States Department of State and cannot be downloaded or otherwise 
* exported or re-exported (i) into (or to a national or resident of) Cuba, 
* Iraq, Libya, Yugoslavia, North Korea, Iran, Syria or any other country 
* to which the US has embargoed goods; or (ii) to anyone on the US 
* Treasury Department's list of Specially Designated Nationals or the US 
* Commerce Department's Table of Denial Orders. By downloading or using 
* this product, you are agreeing to the foregoing and you are representing 
* and warranting that you are not located in, under the control of, or a 
* national or resident of any such country or on any such list. 
*/

#include <stdlib.h>
#ifdef VMS
#include <string.h>
#endif
#include "cssm.h"
#include "cssmMemTrack.h"
#include "x509defs.h"
#include "oidsalg.h"
#include "ber_der.h"
#include "memfuncs.h"
#include "algoid.h"

/* Wrappers to be used since the BER_DER lib uses different order of args for memfuncs */
#ifndef CSSM_MEMTRACK_ON
static void * CSSMAPI 
berder_malloc (BER_HANDLE AddInHandle, uint32 Size)
#else
static void * CSSMAPI 
berder_malloc (BER_HANDLE AddInHandle, uint32 Size, 
	const char * szFilename, uint32 u32LineNumber)
#endif
{
	return CSSM_MALLOC_WRAPPER(_app_malloc, Size, NULL);
}

#ifndef CSSM_MEMTRACK_ON
static  void CSSMAPI 
berder_free (BER_HANDLE AddInHandle, void *MemPtr)
#else
static  void CSSMAPI 
berder_free (BER_HANDLE AddInHandle, void *MemPtr, 
	const char * szFilename, uint32 u32LineNumber)
#endif
{
	CSSM_FREE_WRAPPER(_app_free, MemPtr, NULL);
}

#ifndef CSSM_MEMTRACK_ON
static void * CSSMAPI 
berder_realloc(BER_HANDLE AddInHandle, void *MemPtr, uint32 Size)
#else
static void * CSSMAPI 
berder_realloc(BER_HANDLE AddInHandle, void *MemPtr, uint32 Size,
	const char * szFilename, uint32 u32LineNumber)
#endif
{
	return CSSM_REALLOC_WRAPPER(_app_realloc, MemPtr, Size, NULL);
}

#ifndef CSSM_MEMTRACK_ON
static void * CSSMAPI
berder_calloc(BER_HANDLE AddInHandle, uint32 Num, uint32 Size)
#else
static void * CSSMAPI
berder_calloc(BER_HANDLE AddInHandle, uint32 Num, uint32 Size,
	const char * szFilename, uint32 u32LineNumber)
#endif
{
	return CSSM_CALLOC_WRAPPER(_app_calloc, Num, Size, NULL);
}

static BER_MEMORY_FUNCS berder_funcs_wrapper = {
	berder_malloc, 
	berder_free, 
	berder_realloc, 
	berder_calloc
};



CSSM_RETURN CSSMKeyToSpki(CSSM_KEY_PTR CSSMKey, 
                          CSSM_X509_SUBJECT_PUBLIC_KEY_INFO_PTR spki_ptr)
{
    
    CSSM_OID_PTR algoid_ptr = NULL;
    BER_PARSED_ITEM key_item;
    BER_PARSED_ITEM ber_item;
    DER_NODE_PTR key_node_ptr;
    uint32 i;



    if (!spki_ptr) 
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    memset(spki_ptr, 0, sizeof(CSSM_X509_SUBJECT_PUBLIC_KEY_INFO));

    if ( !CSSMKey->KeyData.Data ) 
    {
        return CSSM_OK;
    }

    
    /* Obtain the algorithm OID */
    algoid_ptr = X509_AlgIdToAlgorithmOid(CSSMKey->KeyHeader.AlgorithmId);

    if (algoid_ptr == NULL)
    {
        return CSSMERR_CL_FUNCTION_FAILED;
    }

	spki_ptr->algorithm.algorithm.Data = CSSM_CALLOC_WRAPPER(_app_calloc, 
                                            1, algoid_ptr->Length, NULL);

	if (spki_ptr->algorithm.algorithm.Data == NULL)
	{
    
        
        return CSSMERR_CL_MEMORY_ERROR;	
	}
    
    spki_ptr->algorithm.algorithm.Length = algoid_ptr->Length;
	memcpy(spki_ptr->algorithm.algorithm.Data, 
                algoid_ptr->Data, 
                algoid_ptr->Length);

    if (CSSM_ALGID_DSA == CSSMKey->KeyHeader.AlgorithmId)
    {
    
        BER_ExpandItem(CSSMKey->KeyData.Data, CSSMKey->KeyData.Length, &key_item);
        key_node_ptr = BER_ParseComplexObject((BER_HANDLE)NULL, &berder_funcs_wrapper, NULL, &key_item);
        /* Check that the parse tree we got was the correct one */
        if (!key_node_ptr || !key_node_ptr->Child || key_node_ptr->Count != 4)
        {

            CSSM_FREE_WRAPPER(_app_free, spki_ptr->algorithm.algorithm.Data, NULL);
            memset(spki_ptr, 0, sizeof(CSSM_X509_SUBJECT_PUBLIC_KEY_INFO));        
            DER_RecycleTree((BER_HANDLE)NULL, &berder_funcs_wrapper, key_node_ptr);
            return CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT;
        }

        for(i=0;i<(uint32)key_node_ptr->Count;i++)
        {
            if (!key_node_ptr->Child[i].IsLeaf || 
                key_node_ptr->Child[i].X.Input.Tag[0] != BER_INTEGER)
            {
                CSSM_FREE_WRAPPER(_app_free, spki_ptr->algorithm.algorithm.Data, NULL);
                memset(spki_ptr, 0, sizeof(CSSM_X509_SUBJECT_PUBLIC_KEY_INFO));
                DER_RecycleTree((BER_HANDLE)NULL, &berder_funcs_wrapper, key_node_ptr);
                return CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT;
            }
        }
        
        /* Encode the first 3 children into the params */
        key_node_ptr->Count--;
        spki_ptr->algorithm.parameters.Data = DER_Generate((BER_HANDLE)NULL, &berder_funcs_wrapper, key_node_ptr);
        if (!spki_ptr->algorithm.parameters.Data)
        {
            CSSM_FREE_WRAPPER(_app_free, spki_ptr->algorithm.algorithm.Data, NULL);
            memset(spki_ptr, 0, sizeof(CSSM_X509_SUBJECT_PUBLIC_KEY_INFO));
            key_node_ptr->Count++;
            DER_RecycleTree((BER_HANDLE)NULL, &berder_funcs_wrapper, key_node_ptr);
            return CSSMERR_CL_FUNCTION_FAILED;
        }
        spki_ptr->algorithm.parameters.Length = BER_SizeofObject(spki_ptr->algorithm.parameters.Data);

        key_node_ptr->Count++;

        /* The BER_PARSED_ITEM as a CSSM_DATA */
        ber_item = key_node_ptr->Child[3].X.Input;
        spki_ptr->subjectPublicKey.Length = BER_LengthOfTag(ber_item.Tag)
                                            + BER_LengthOfLength(ber_item.ContentLength)
                                            + ber_item.ContentLength;
        spki_ptr->subjectPublicKey.Data = CSSM_MALLOC_WRAPPER(_app_malloc,
                                                              spki_ptr->subjectPublicKey.Length,
                                                              NULL);
        if (!spki_ptr->subjectPublicKey.Data)
        {
            CSSM_FREE_WRAPPER(_app_free, spki_ptr->algorithm.algorithm.Data, NULL);
            memset(spki_ptr, 0, sizeof(CSSM_X509_SUBJECT_PUBLIC_KEY_INFO));
            DER_RecycleTree((BER_HANDLE)NULL, &berder_funcs_wrapper, key_node_ptr);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        BER_OutputTag(ber_item.Tag, &spki_ptr->subjectPublicKey.Data[0]);
        BER_OutputLength(ber_item.ContentLength, 
            &spki_ptr->subjectPublicKey.Data[BER_LengthOfTag(ber_item.Tag)]);
        memcpy(&spki_ptr->subjectPublicKey.Data[BER_LengthOfTag(ber_item.Tag) + BER_LengthOfLength(ber_item.ContentLength)],
            ber_item.Content, ber_item.ContentLength);

        DER_RecycleTree((BER_HANDLE)NULL, &berder_funcs_wrapper, key_node_ptr);

        

    }
    else
    {
        spki_ptr->subjectPublicKey.Length = CSSMKey->KeyData.Length;
        spki_ptr->subjectPublicKey.Data = CSSM_MALLOC_WRAPPER(_app_malloc, 
                                    spki_ptr->subjectPublicKey.Length, NULL);
        if (!spki_ptr->subjectPublicKey.Data)
        {
            CSSM_FREE_WRAPPER(_app_free, spki_ptr->algorithm.algorithm.Data, NULL);
            memset(spki_ptr, 0, sizeof(CSSM_X509_SUBJECT_PUBLIC_KEY_INFO));
            return CSSMERR_CL_MEMORY_ERROR;
        }
        memcpy(spki_ptr->subjectPublicKey.Data, CSSMKey->KeyData.Data, spki_ptr->subjectPublicKey.Length);
        spki_ptr->algorithm.parameters.Length = 0;
        spki_ptr->algorithm.parameters.Data = NULL;
    }

    
    return CSSM_OK;
}
