/* User EXPAT to convert an XML file into a CSSM X509 name structure */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <ctype.h>
#include <string.h>

#include "cssm.h"
#include "x509defs.h"
#include "oidscert.h"
#include "oidconvert.h"

#include "xmlCertTags.h"
#include "xmlparse.h"


/* 
 * Here are some constants and a struct that we use
 * to keep track of where we are in the parsing process
 */
enum xmlCertParseState {STATE_NO_STATE,
						STATE_IN_SUBJECTNAME,
						STATE_IN_NUMRDNS,
						STATE_IN_RDN,
						STATE_IN_NUMTVPAIRS,
						STATE_IN_TVPAIR,
						STATE_IN_BERTYPE,
						STATE_IN_OID,
						STATE_IN_VALUE};

typedef struct CertNameParseTracker
{
	CSSM_X509_NAME_PTR pSubjectName;
	enum xmlCertParseState state;
	int currentRDNindex;
	int currentTVPindex;
#ifdef VMS
}CertNameParseTracker;
#else
};
typedef struct CertNameParseTracker CertNameParseTracker;
#endif
typedef struct CertNameParseTracker* CertNameParseTrackerPtr;

/* 
 * Call back functions for EXPAT parser
 */
void startElement(void *userData, const char *name, const char **atts)
{
	CertNameParseTrackerPtr pTracker = (CertNameParseTrackerPtr)userData;

	if(strcmp(name,XMLCERT_SUBJECTNAME) == 0)
	{
		// at start of subject name, allocate subject name structure,
		pTracker->pSubjectName = malloc(sizeof(CSSM_X509_NAME));
		memset(pTracker->pSubjectName,0,sizeof(CSSM_X509_NAME));
		pTracker->state = STATE_IN_SUBJECTNAME;

		return;
	}
	if(strcmp(name,XMLCERT_X509NAME) == 0)
	{
		// we can ignore this since we have just one name, always the subject name
		// but if that changes, then we would have to respond to this in some way
		return;
	}
	if(strcmp(name,XMLCERT_NUMRDNS) == 0)
	{
		// about to get num of rdns, set state machine	
		pTracker->state = STATE_IN_NUMRDNS;
		return;
	}
	if(strcmp(name,XMLCERT_RDN) == 0)
	{
		pTracker->state = STATE_IN_RDN;

		// each time we start a new RDN,
		// we have to reset the TV Pair index
		// back to zero
		pTracker->currentTVPindex = 0;
		
		return;
	}
	if(strcmp(name,XMLCERT_NUMTVPAIRS) == 0)
	{
		pTracker->state = STATE_IN_NUMTVPAIRS;
		
		return;
	}
	if(strcmp(name,XMLCERT_TVPAIR) == 0)
	{
		pTracker->state = STATE_IN_TVPAIR;
		
		return;
	}
	if(strcmp(name,XMLCERT_BERTYPE) == 0)
	{
		pTracker->state = STATE_IN_BERTYPE;
		
		return;
	}
	if(strcmp(name,XMLCERT_OID) == 0)
	{
		pTracker->state = STATE_IN_OID;
		
		return;
	}

	if(strcmp(name,XMLCERT_VALUE) == 0)
	{
		pTracker->state = STATE_IN_VALUE;
		
		return;
	}

	
}

void endElement(void *userData, const char *name)
{
	CertNameParseTrackerPtr pTracker = (CertNameParseTrackerPtr)userData;

	if(strcmp(name,XMLCERT_RDN) == 0)
	{
		// leaving the current RDN
		// bump index to next RDN
		pTracker->currentRDNindex++;

		return;
	}
	if(strcmp(name,XMLCERT_TVPAIR) == 0)
	{
		// leaving the current TVP
		// bump index to next TVP
		pTracker->currentTVPindex++;
		return;
	}

	pTracker->state = STATE_NO_STATE;
	return;

	
}

void CharacterDataHandler(void *userData,
					 const XML_Char *s,
					 int len)
{
	// use this buffer to unpack numbers
	char buffer[32];
	CertNameParseTrackerPtr pTracker = (CertNameParseTrackerPtr)userData;
	
	if(pTracker->state == STATE_NO_STATE)
	{
		// just ignore the incoming data
		return;
	}
	switch(pTracker->state)
	{
	case STATE_IN_NUMRDNS:
		// convert string to number
		{
			int numRDNs;
			CSSM_X509_RDN_PTR pRDNs;
			strncpy(buffer,s,len);
			buffer[len] = 0;

			numRDNs = atoi(buffer);

			// allocate array of RDNs
			pRDNs = malloc(sizeof(CSSM_X509_RDN)*numRDNs);
			memset(pRDNs,0,sizeof(CSSM_X509_RDN)*numRDNs);

			// update data structure of current X509 name
			pTracker->pSubjectName->numberOfRDNs = numRDNs;
			pTracker->pSubjectName->RelativeDistinguishedName = pRDNs;

		}
		break;
	case STATE_IN_NUMTVPAIRS:
		{
			int numTVPs;
			CSSM_X509_TYPE_VALUE_PAIR_PTR pTVPs;
			strncpy(buffer,s,len);
			buffer[len] = 0;

			numTVPs = atoi(buffer);

			// allocate array of RDNs
			pTVPs = malloc(sizeof(CSSM_X509_TYPE_VALUE_PAIR)*numTVPs);
			memset(pTVPs,0,sizeof(CSSM_X509_TYPE_VALUE_PAIR)*numTVPs);

			// update data structure of current RDN
			pTracker->pSubjectName->
				RelativeDistinguishedName[pTracker->currentRDNindex].numberOfPairs = numTVPs;
			pTracker->pSubjectName->
				RelativeDistinguishedName[pTracker->currentRDNindex].AttributeTypeAndValue = pTVPs;
		}
		break;
	case STATE_IN_BERTYPE:
		{
			CSSM_BER_TAG* pTagType;
			// get a null terminated copy of string
			strncpy(buffer,s,len);
			buffer[len] = 0;

			
			// figure out where we are going to put the tag type
			pTagType = &(pTracker->pSubjectName->
				RelativeDistinguishedName[pTracker->currentRDNindex].
				AttributeTypeAndValue[pTracker->currentTVPindex].valueType);
			
			*pTagType = atoi(buffer);

		}
		break;
	case STATE_IN_OID:
		{
			// get a null terminated copy of OID string
			CSSM_DATA* pOID;
			strncpy(buffer,s,len);
			buffer[len] = 0;

			// find the oid slot in current TVP
			pOID = &(pTracker->pSubjectName->
				RelativeDistinguishedName[pTracker->currentRDNindex].
				AttributeTypeAndValue[pTracker->currentTVPindex].type);

			// use utility to convert dot notation to oid
			DotOidToCSSMData(buffer,pOID);
			
		}
		break;
	case STATE_IN_VALUE:
		{
			// find the current value slot
			CSSM_DATA* pValue;
			pValue = &(pTracker->pSubjectName->
				RelativeDistinguishedName[pTracker->currentRDNindex].
				AttributeTypeAndValue[pTracker->currentTVPindex].value);

			// allocate and populate
			pValue->Length = len+1;
			pValue->Data = malloc(len);
			memcpy(pValue->Data,s,len);
			pValue->Data[len] = 0;
		}
		break;


	}
	
	/* ignore leading white space 
	if((s[0] == '\t') || (s[0] == '\n') || (s[0] == '\r'))
		return;
	*/

}

/*
 * Main entry point for library
 * Takes a buffer full of an XML encoded X509 subject name
 * and returns a CSSM_X509_NAME structure wrapped up in a field
 */
CSSM_RETURN ParseTemplate(const CSSM_DATA_PTR pCertTemplate,
						  CSSM_FIELD_PTR* ppFields,
						  uint32* numFields)
{
	// parser variable and its 'done' flag
	XML_Parser parser;
	int done = 0;

	// this is our data structure  to maintain the state of the parse
	CertNameParseTracker parseTracker;
	memset(&parseTracker,0,sizeof(CertNameParseTracker));
		
	// create a new parser
	parser = XML_ParserCreate(NULL);
	
	// pass our tracking structure to parser
	XML_SetUserData(parser, &parseTracker);

	// set callback functions for parser
	XML_SetElementHandler(parser, startElement, endElement);
	XML_SetCharacterDataHandler(parser,CharacterDataHandler);

	// parse it.
	// (const char *) added to pCertTemplate->Data to resolve warning
	if (!XML_Parse(	parser, 
					(const char *) pCertTemplate->Data, 
					pCertTemplate->Length, 
					done)) 
	{
	  fprintf(stderr,
		  "%s at line %d\n",
		  XML_ErrorString(XML_GetErrorCode(parser)),
		  XML_GetCurrentLineNumber(parser));
	  return CSSMERR_CL_UNKNOWN_FORMAT;
	}
	XML_ParserFree(parser);


	// pack up the X509Name in a field and return to caller
	*ppFields = malloc(sizeof(CSSM_FIELD));
	(*ppFields)->FieldOid = CSSMOID_X509V1SubjectNameCStruct;
	(*ppFields)->FieldValue.Data = (unsigned char*)parseTracker.pSubjectName;
	(*ppFields)->FieldValue.Length = CSSM_FIELDVALUE_COMPLEX_DATA_TYPE;
	*numFields = 1;
	

	return CSSM_OK;
}

