/*-----------------------------------------------------------------------
 * File: MANIFEST.H
 *
 * Copyright (c) 1998-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

// Manifest.h
//
// This file contains the class definition for the CManifest class, a
//  class which implements an object based on the CDSA 2.0 Signed 
//  Manifest Specification (Open Group standard).
//  This class implments signed manifests using the Intel Integrity 
//  Services API (version 1.2.1).
//
//////////////////////////////////////////////////////////////////////
#ifndef MANIFEST_H
#define MANIFEST_H

// Mrm includes
#include "wf_type.h"            // typedefs for all wf-related objects
#include "SignatureGraph.h"     // signature graph class definition
#include "ObjectList.h"         // object list class definition

#ifdef VMS
typedef struct memlist {
        struct memlist *next;
        void *buffer;
} memlist;
#endif
//////////////////////////////////////////////////////////////////////
// Class description: CManifest
//  Members:
//      mContext(public)        
//          a context handle which relates to the current active 
//          manifest archive
//      mExternalFilename(private)
//          fully-qualified pathname for the external file 
//          representation of the active manifest archive
//      mArchCfg(private)
//          a handle to the current configuration of the active 
//          manifest archive
//      mObjectList(private)
//          a list of all objects covered by the active manifest 
//          archive(equivalent to the manifest sections)
//      mCDSAProviderHandle(private)
//          handle to the instatiation of the CDSA provider for 
//          use with the integrity provider
//      mCDSAProviderFuncTable(private)
//          pointer to the functions provided by the CDSA
//          provider
//      mSigCfg
//          handle to the active manifest's signature context
//  Methods:
//      Init(public)
//          method used to initialize an constructed CManifest object
//          (all CManifest objects MUST be initialized by on of the 
//           Init() methods)
//      SignAllObjects(public)
//          method used to apply a signature over all sections in the 
//          active manifest
//      Commit(public)
//          writes the current operational manifest to its persitent
//          location
//      VerifyAllSections(public)
//          verifies the integrity of all sections included in the 
//          active manifest (rechecks hashes over content covered)
//      VerifyAllSigners(public)
//          verifies the signatures applied by each signer included
//          in the active manifest (implicitly trusts content hashes)
//      GetObjectList(public)
//          returns the list of all objects included in the active 
//          manifest
//      GetAttributes(public)
//          returns a list of all attributes for the active manifest
//      InitManifestFromList(private)
//          initializes a manifest with the provided object list
//      InitCDSAProviders(private)
//          initializes the CDSA providers for use by this class
//          (initializes CSSM and integrity services provider)
/////////////////////////////////////////////////////////////////////
class DllExport CManifest  
{
public:
    // Public Members
	WF_ARCHIVE_CONTEXT *mContext;
	
    // Public Methods
    // constructors
	CManifest();
	~CManifest();

    // initializers
    WF_RETURN Init(CObjectList *List, 
                   char *ManifestPath);
    WF_RETURN Init(CObjectList *List, 
                   char *ManifestPath,
                   WF_GUID *CSPGuid,
                   WF_CSSM_METHODS *CSSMMethods);
    WF_RETURN Init(char* FilePath);
    WF_RETURN Init(char* FilePath,
                   WF_GUID *CSPGuid,
                   WF_CSSM_METHODS *CSSMMethods);

    // misc methods
    WF_RETURN AddSections(CObjectList *ObjectList);
    WF_RETURN SignAllSections(WF_CONST_DATA *Certificate, 
                              WF_CONST_DATA *Password,
                              WF_CONST_DATA *SignerName,
                              WF_CERTGROUP_PTR CertGroup,
                              CAttributeList *SignerAttributes);
    WF_RETURN Commit();
    
    // verify methods
    WF_RETURN VerifyAllSigners();
    WF_RETURN VerifyAllSections();

    // get methods
    WF_RETURN GetSignatureGraph(CSignatureGraph* SigGraph);
    CObjectList* GetSectionList();
    WF_DATA_PTR* GetAttributes();

private:
    // Private Members
	char* mExternalFilename;
	WF_ARCHIVE_CONFIG *mArchCfg;
    CObjectList *mSectionList;
    WF_CSSM_HANDLE mCSSMHandle;
    WF_CSSM_FUNC_TABLE *mCSSMFuncTable;
    WF_SIGNATURE_CONFIG *mSigCfg;
    struct memlist *mIntegBuffers;
    WF_BOOL mIsSigned;

    // Private Methods
    WF_RETURN InitManifestFromList(CObjectList *List, 
                                   char *ManifestPath);
    WF_RETURN InitManifestFromList(CObjectList *List, 
                                   char *ManifestPath,
                                   WF_GUID *CSPGuid,
                                   WF_CSSM_METHODS *CSSMMethods);
    WF_RETURN InitCDSAProviders();
    WF_RETURN InitCDSAProviders(WF_GUID *CSPGuid,
                                WF_CSSM_METHODS *CSSMMethods);
    void UninitCDSAProviders();
    void Recycle();
};

#endif // !def MANIFEST_H
