/*-----------------------------------------------------------------------
 * File: ATTRIBUTELIST.CPP
 *
 * Copyright (c) 1998-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

// AttributeList.cpp
//
// This file contains the class implementation for the 
//  CAttributeListNode and CattributeList classes.  These classes 
//  implement a attribute/value pair list.
//
//////////////////////////////////////////////////////////////////////

#ifdef CSSM_MEMTRACK
#undef new
#include <crtdbg.h>
static char SZ_FILE_NAME[] = __FILE__;
#define new new(_NORMAL_BLOCK, SZ_FILE_NAME, __LINE__)  /*need space*/
#endif /* ifdef VMS comment this out - CSSM_MEMTRACK */

#include "AttributeList.h"
#ifndef VMS
#pragma warning (disable:4201 4514 4214 4115)
#endif
#include <memory.h>
#include <stdlib.h>
#include <scMemTrack.h>
#ifndef VMS
#pragma warning (default:4201 4214 4115)
#endif

//======================//
//                      //
//  CAttributeListNode  //
//                      //
//======================//

//////////////////////////////////////////////////////////////////////
// CAttributeListNode::CAttributeListNode()
//  constructs an empty CAttributeListNode object
//
// Parameters:
//  none
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CAttributeListNode::CAttributeListNode()
{
    mAttribute.Length = 0;
    mAttribute.Data = 0;
    mValue.Length = 0;
    mValue.Data = 0;
    mNext = 0;
}

//////////////////////////////////////////////////////////////////////
// CAttributeListNode::~CAttributeListNode()
//  destructs a CAttributeListNode object
//
// Parameters:
//  none
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CAttributeListNode::~CAttributeListNode()
{
    if(mAttribute.Data)
        free((void*)mAttribute.Data);
    if(mValue.Data)
        free((void*)mValue.Data);
}

//////////////////////////////////////////////////////////////////////
// CAttributeListNode::CAttributeListNode(const CAttributeListNode *node)
//  constructs a CAttributeListNode object which is a copy of the one 
//  passed 
//
// Parameters:
//  node(input)         pointer to the CAttributeListNode object to copy
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CAttributeListNode::CAttributeListNode(const CAttributeListNode *node)
{
    mAttribute.Length = node->mAttribute.Length;
    mAttribute.Data = (unsigned char*)malloc(node->mAttribute.Length);
    memcpy((unsigned char*)mAttribute.Data, node->mAttribute.Data, node->mAttribute.Length);
    mValue.Length = node->mValue.Length;
    mValue.Data = (unsigned char*)malloc(node->mValue.Length);
    memcpy((unsigned char*)mValue.Data, node->mValue.Data, node->mValue.Length);
    mNext = 0;
}

//////////////////////////////////////////////////////////////////////
// CAttributeListNode::CAttributeListNode(WF_CONST_DATA_PTR Attribute, 
//                                        WF_CONST_DATA_PTR Value)
//  constructs a CAttributeListNode object with the attribute and value
//  fields filled with the passed Attribute and Value parameters
//
// Parameters:
//  Attribute(input)
//      name (tag) to be associated with the value of this node
//  Value(input)
//      value for this node
// Returns:
//  none
// Side effects:
//  Attribute and Value data is copied before construction
//////////////////////////////////////////////////////////////////////
CAttributeListNode::CAttributeListNode(WF_CONST_DATA_PTR Attribute, 
                                       WF_CONST_DATA_PTR Value)
{
    mAttribute.Length = Attribute->Length;
    mAttribute.Data = (unsigned char*)malloc(Attribute->Length);
    memcpy((unsigned char*)mAttribute.Data, Attribute->Data, Attribute->Length);
    mValue.Length = Value->Length;
    mValue.Data = (unsigned char*)malloc(Value->Length);
    memcpy((unsigned char*)mValue.Data, Value->Data, Value->Length);
    mNext = 0;
}
    

//==================//
//                  //
//  CAttributeList  //
//                  //
//==================//

//////////////////////////////////////////////////////////////////////
// CAttributeList::CAttributeList()
//  constructs an empty CAttributeList object
//
// Parameters:
//  none
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CAttributeList::CAttributeList()
{
    mHead = 0;
	mTail = 0;
	mNumberOfNodes = 0;
}

//////////////////////////////////////////////////////////////////////
// CAttributeList::~CAttributeList()
//  destructs a CAttributeList object
//
// Parameters:
//  none
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CAttributeList::~CAttributeList()
{
    CAttributeListNode *current_node = mHead, *next_node;

	// delete all nodes
	for( unsigned int i=0; i<mNumberOfNodes; i++ )
	{
		next_node = current_node->mNext;
		delete current_node;
		current_node = next_node;
	}
}


//////////////////////////////////////////////////////////////////////
// CAttributeList::CAttributeList(const CAttributeList *list)
//  constructs a CAttributeList object which is a copy of the one 
//  passed 
//
// Parameters:
//  list(input)         
//      pointer to the CAttributeList object to copy
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CAttributeList::CAttributeList(const CAttributeList *list)
{
	CAttributeListNode *new_node, *old_node;
	
	// setup head ptr
	if(list)
	{
		if( list->mHead )
		{
			// copy head
            mHead = new CAttributeListNode(list->mHead);
			mTail = mHead;
			new_node = mHead;
			old_node = list->mHead;
			// copy all remaining nodes
			for( unsigned int i=1; i<list->mNumberOfNodes; i++ )
			{
				new_node->mNext = new CAttributeListNode( old_node->mNext );
				new_node = new_node->mNext;
				mTail = new_node;
				old_node = old_node->mNext;
			}
	        // intialize number of nodes
	        mNumberOfNodes = list->mNumberOfNodes;
        }
		else
		{
			mHead = 0;
			mTail = 0;
    	    mNumberOfNodes = 0;
		}
	}
    else
    {
        mHead = 0;
	    mTail = 0;
	    mNumberOfNodes = 0;
    }

}

//////////////////////////////////////////////////////////////////////
// CAttributeList::CAttributeList(CAttributeListNode *node)
//  constructs a CAttributeList object containing the node passed
//
// Parameters:
//  node(input)         
//      pointer to the CAttributeListNode object to embed in the list
// Returns:
//  none
// Side effects:
//  node will be copied prior to insertion
//////////////////////////////////////////////////////////////////////
CAttributeList::CAttributeList(CAttributeListNode *node)
{
	// construct a list of one
    mHead = new CAttributeListNode( node );
	mTail = mHead;
	mNumberOfNodes = 1;
}

//////////////////////////////////////////////////////////////////////
// CAttributeList::AddNode( CAttributeListNode *NodeToAdd )
//  appends NodeToAdd to the end of the list
//
// Parameters:
//  NodeToAdd(input)         
//      pointer to the CAttributeListNode object to add to the list
// Returns:
//  none
// Side effects:
//  node will be copied prior to addition
//////////////////////////////////////////////////////////////////////
void CAttributeList::AddNode( CAttributeListNode *NodeToAdd )
{
    if( !NodeToAdd )
    {
        // nothing to add, just return
        return;
    }
    if( !mHead )
    {
        // no head, so make this node the head
        mHead = new CAttributeListNode( NodeToAdd );
        mTail = mHead;
    }
    else
    {
        // add to end of the list
        mTail->mNext = new CAttributeListNode( NodeToAdd );
	    mTail = mTail->mNext;
	    mTail->mNext = 0;
    }
	mNumberOfNodes++;
	return;
}

//////////////////////////////////////////////////////////////////////
// CAttributeList::GetNumberOfNodes()
//  gets the total number of elements on the list
//
// Parameters:
//  none
// Returns:
//  number of elements in the list
// Side effects:
//  node will be copied prior to addition
//////////////////////////////////////////////////////////////////////
unsigned int CAttributeList::GetNumberOfNodes()
{
	return mNumberOfNodes;
}


