/*-----------------------------------------------------------------------
 * File: CSPAPI.C
 *
 * Copyright (c) 1995-1999 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */ 


#ifdef VMS
#include "precomp.h"
#else
#include "../precomp.h"
#endif

/*---------------------------------------------------------------
*
*Name: CSSM_QuerySize
*
*Description:
*   Exported API for querying the input and output sizes for
*   purpose of allocating output buffers.
*
*Parameters: 
*  CCHandle (input) - the context handle
*  ReqSizeInBlock (output) - size of input block for context
*  ReqSizeOutBlock (output) - size of output block for context
*
*Returns:
*   CSSM_FAIL - CSP wasn't able to complete function
*   CSSM_OK - Call successful executed by CSP add-in
*
*------------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_QuerySize(CSSM_CC_HANDLE CCHandle,
               CSSM_BOOL Encrypt,
               uint32 QuerySizeCount,
               CSSM_QUERY_SIZE_DATA_PTR DataBlock)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_QuerySize);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;

                       /*
                        * Call the SPI function if it's valid.
                       */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->QuerySize))
                        {
                            ERR(err = CallBack->QuerySize(hCSP,
                                                          CCHandle,
                                                          Context,
                                                          Encrypt,
                                                          QuerySizeCount,
                                                          DataBlock));
                        }
                        else
                        {
                            ERR(err =CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}


/*---------------------------------------------------------------
*
*Name: CSSM_SignData
*
*Description:
*   Exported API for signing the data specified in DataBufs
*
*Parameters: 
*  CCHandle (input) - the context handle
*  DataBufs (input) - array of data buffers that holds
*                     information that needs to be sign
*  DataBufCount (input) - number of data buffers pointed to
*                         by parameter DataBufs
*  Signature (output) - signature of DataBufs
*
*Returns:
*   CSSM_FAIL - CSP wasn't able to complete function
*   CSSM_OK - Call successful executed by CSP add-in
*
*-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_SignData(CSSM_CC_HANDLE CCHandle,
              const CSSM_DATA *DataBufs,
              uint32 DataBufCount,
              CSSM_ALGORITHMS DigestAlgorithm,
              CSSM_DATA_PTR Signature)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_SignData);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;

                        /*
                         * Call the SPI function if it's valid.
                        */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->SignData))
                        {
                            ERR(err = CallBack->SignData(hCSP,
                                                         CCHandle,
                                                         Context,
                                                         DataBufs,
                                                         DataBufCount,
                                                         DigestAlgorithm,
                                                         Signature));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }

                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_SignDataInit
 *
 *Description:
 *   Exported API for initializing the stage version of
 *   sign data.
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_SignDataInit(CSSM_CC_HANDLE CCHandle)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_SignDataInit);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;
                        /*
                         * Call the SPI function if it's valid.
                        */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->SignDataInit))
                        {
                            ERR(err = CallBack->SignDataInit(hCSP,
                                                             CCHandle,
                                                             Context));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_SignDataUpdate
 *
 *Description:
 *   Exported API for updating the sign data staged operation.
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  DataBufs (input) - array of data buffers that holds
 *                     information that needs to be sign
 *  DataBufCount (input) - number of data buffers pointed to
 *                         by parameter DataBufs
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_SignDataUpdate(CSSM_CC_HANDLE CCHandle,
                    const CSSM_DATA *DataBufs,
                    uint32 DataBufCount)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_SignDataUpdate);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_NO_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                hCSP = AttachRecord->AttachHandle;

                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->SignDataUpdate))
                        {
                            ERR(err = CallBack->SignDataUpdate(hCSP,
                                                               CCHandle,
                                                               DataBufs,
                                                               DataBufCount));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_SignDataFinal
 *
 *Description:
 *   Exported API for completing the staged sign operation
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  Signature (output) - signature of stage function
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_SignDataFinal(CSSM_CC_HANDLE CCHandle,
                   CSSM_DATA_PTR Signature)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_SignDataFinal);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_NO_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                hCSP = AttachRecord->AttachHandle;

                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->SignDataFinal))
                        {
                            ERR(err = CallBack->SignDataFinal(hCSP,
                                                              CCHandle,
                                                              Signature));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_VerifyData
 *
 *Description:
 *   Exported API for verifying the signature
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  DataBufs (input) - array of data buffers that holds
 *                     information that needs to be sign
 *  DataBufCount (input) - number of data buffers pointed to
 *                         by parameter DataBufs
 *  Signature (input) - signature to verify
 *
 *Returns:
 *   CSSM_OK - the signature matches the data
 *   or an error code.
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_VerifyData(CSSM_CC_HANDLE CCHandle,
                const CSSM_DATA *DataBufs,
                uint32 DataBufCount,
                CSSM_ALGORITHMS DigestAlgorithm,
                const CSSM_DATA *Signature)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_VerifyData);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;

                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->VerifyData))
                        {
                            ERR(err = CallBack->VerifyData(hCSP,
                                                           CCHandle,
                                                           Context,
                                                           DataBufs,
                                                           DataBufCount,
                                                           DigestAlgorithm,
                                                           Signature));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }

                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_VerifyDataInit
 *
 *Description:
 *   Exported API for initializing the staged verify function
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  Signature (input) - signature to verify
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_VerifyDataInit(CSSM_CC_HANDLE CCHandle)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_VerifyDataInit);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;

                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->VerifyDataInit))
                        {
                            ERR(err = CallBack->VerifyDataInit(hCSP,
                                                               CCHandle,
                                                               Context));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_VerifyDataUpdate
 *
 *Description:
 *   Exported API for updating the stage verify operation
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  DataBufs (input) - array of data buffers that holds
 *                     information that needs to be sign
 *  DataBufCount (input) - number of data buffers pointed to
 *                         by parameter DataBufs
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_VerifyDataUpdate(CSSM_CC_HANDLE CCHandle,
                      const CSSM_DATA *DataBufs,
                      uint32 DataBufCount)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_VerifyDataUpdate);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_NO_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                hCSP = AttachRecord->AttachHandle;

                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->VerifyDataUpdate))
                        {
                            ERR(err = CallBack->VerifyDataUpdate(hCSP,
                                                                 CCHandle,
                                                                 DataBufs,
                                                                 DataBufCount));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_VerifyDataFinal
 *
 *Description:
 *   Exported API for completing the staged verify operation
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *
 *Returns:
 *   CSSM_OK - the signature matches the data
 *   or an error code.
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_VerifyDataFinal(CSSM_CC_HANDLE CCHandle,
                     const CSSM_DATA *Signature)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_VerifyDataFinal);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_NO_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                hCSP = AttachRecord->AttachHandle;

                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->VerifyDataFinal))
                        {
                            ERR(err = CallBack->VerifyDataFinal(hCSP,
                                                                CCHandle,
                                                                Signature));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_DigestData
 *
 *Description:
 *   Exported API for performing a digest on data
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  DataBufs (input) - array of data buffers that holds
 *                     information that needs to be hashed
 *  DataBufCount (input) - number of data buffers pointed to
 *                         by parameter DataBufs
 *  Digest (output) - pointer to hash result
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_DigestData(CSSM_CC_HANDLE CCHandle,
                const CSSM_DATA *DataBufs,
                uint32 DataBufCount,
                CSSM_DATA_PTR Digest)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_DigestData);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;

                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->DigestData))
                        {
                            ERR(err = CallBack->DigestData(hCSP,
                                                           CCHandle,
                                                           Context,
                                                           DataBufs,
                                                           DataBufCount,
                                                           Digest));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }

                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_DigestDataInit
 *
 *Description:
 *   Exported API for initializing the staged digest operation
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_DigestDataInit(CSSM_CC_HANDLE CCHandle)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_DigestDataInit);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;

                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->DigestDataInit))
                        {
                            ERR(err = CallBack->DigestDataInit(hCSP,
                                                               CCHandle,
                                                               Context));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }

                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_DigestDataUpdate
 *
 *Description:
 *   Exported API for updating the data to the stage digest operation
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  DataBufs (input) - array of data buffers that holds
 *                     information that needs to be hashed
 *  DataBufCount (input) - number of data buffers pointed to
 *                         by parameter DataBufs
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_DigestDataUpdate(CSSM_CC_HANDLE CCHandle,
                      const CSSM_DATA *DataBufs,
                      uint32 DataBufCount)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_DigestDataUpdate);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_NO_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                hCSP = AttachRecord->AttachHandle;

                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->DigestDataUpdate))
                        {
                            ERR(err = CallBack->DigestDataUpdate(hCSP,
                                                                 CCHandle,
                                                                 DataBufs,
                                                                 DataBufCount));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_DigestDataClone
 *
 *Description:
 *   Exported API to clone a stage digest operation
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *
 *Returns:
 *   zero - CSP wasn't able to complete function
 *   none zero - handle to clone cryptographic context
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_DigestDataClone(CSSM_CC_HANDLE CCHandle,
                     CSSM_CC_HANDLE *ClonedCCHandle)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;

    cssm_INTERNAL_CONTEXT_PTR ListIndexPtr = NULL;
    CSSM_CC_HANDLE tempClonedCCHandle = 0;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;
    CSSM_CONTEXT_PTR Context = NULL;

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_DigestDataUpdate);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {
        /* Check for valid parameters */
        if((port_IsBadWritePtr(ClonedCCHandle,
                               sizeof( CSSM_CC_HANDLE ))) )
        {
            ERR(err = CSSMERR_CSSM_INVALID_POINTER);
        }


#ifdef ISL_INCLUDED
        if( CSSM_OK == err)
        {
            ISL_FUNCTION_PTR retAddress = NULL;
            EISL_GetReturnAddress( retAddress);

            /* if we have to, we check the application here */
            if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
            {
                ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL));
                if(CSSM_OK != err)
                {
                    ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
            }
        }
#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_WRITER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;


                        /* Generate random handle for context */
                        while(0 == tempClonedCCHandle)
                        {
                            /* init newHandle to the high bits of the old handle */
                            CSSM_UINT64_SETHIGH(tempClonedCCHandle,
                                                hCSP);

                            /* set the low bits of the new handle to random bits */
                            tempClonedCCHandle = tempClonedCCHandle | cssm_Rand();
                            ListIndexPtr = AttachRecord->ContextList;

                            while((NULL != ListIndexPtr) &&
                                  (0 != tempClonedCCHandle))
                            {
                                if((tempClonedCCHandle == ListIndexPtr->ContextHandle) ||
                                   (CSSM_UINT64_LOW(tempClonedCCHandle) == 0))
                                {
                                    tempClonedCCHandle = 0;
                                }

                                ListIndexPtr = ListIndexPtr->Next;
                            }
                        }

                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->DigestDataClone))
                        {
                            ERR(err = CallBack->DigestDataClone(hCSP,
                                                                CCHandle,
                                                                tempClonedCCHandle));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }

                        if(CSSM_OK == err)
                        {
                            ERR(err = cssm_SaveContextClone(hCSP,
                                                            InternalCtx,
                                                            tempClonedCCHandle,
                                                            &(AttachRecord->ContextList)));
                            if(CSSM_OK != err)
                            {
                                if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->EventNotify))
                                {

                                    ERR(err =  CallBack->EventNotify(hCSP,
                                                                     CSSM_CONTEXT_EVENT_DELETE,
                                                                     tempClonedCCHandle,
                                                                     AttachRecord->ContextList->Context));
                                }
                            }
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_WRITER_LOCK);
            }
        }
    }

    /* if everything worked, we set the cloned CC handle for return */
    if(CSSM_OK == err)
    {
        *ClonedCCHandle = tempClonedCCHandle;
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_DigestFinal
 *
 *Description:
 *   Exported API for completing he stage digest operation
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  Digest (output) - pointer to hash result
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_DigestDataFinal(CSSM_CC_HANDLE CCHandle,
                     CSSM_DATA_PTR Digest)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_DigestDataFinal);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_NO_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                hCSP = AttachRecord->AttachHandle;

                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->DigestDataFinal))
                        {
                            ERR(err = CallBack->DigestDataFinal(hCSP,
                                                                CCHandle,
                                                                Digest));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_GenerateMac
 *
 *Description:
 *   Exported API for generating a MAC
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  DataBufs (input) - array of data buffers that holds
 *                     information that needs to be hashed
 *  DataBufCount (input) - number of data buffers pointed to
 *                         by parameter DataBufs
 *  Mac (output) - pointer to mac result
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_GenerateMac(CSSM_CC_HANDLE CCHandle,
                 const CSSM_DATA *DataBufs,
                 uint32 DataBufCount,
                 CSSM_DATA_PTR Mac)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_GenerateMac);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;

                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GenerateMac))
                        {
                            ERR(err = CallBack->GenerateMac(hCSP,
                                                            CCHandle,
                                                            Context,
                                                            DataBufs,
                                                            DataBufCount,
                                                            Mac));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_GenerateMacInit
 *
 *Description:
 *   Exported API for stage generation of a mac
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_GenerateMacInit(CSSM_CC_HANDLE CCHandle)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_GenerateMacInit);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack));  /* #ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;

                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GenerateMacInit))
                        {
                            ERR(err = CallBack->GenerateMacInit(hCSP,
                                                                CCHandle,
                                                                Context));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_GenerateMacUpdate
 *
 *Description:
 *   Exported API for updating the data to generate the mac
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  DataBufs (input) - array of data buffers that holds
 *                     information that needs to be hashed
 *  DataBufCount (input) - number of data buffers pointed to
 *                         by parameter DataBufs
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_GenerateMacUpdate(CSSM_CC_HANDLE CCHandle,
                       const CSSM_DATA *DataBufs,
                       uint32 DataBufCount)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_GenerateMacUpdate);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_NO_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                hCSP = AttachRecord->AttachHandle;

                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GenerateMacUpdate))
                        {
                            ERR(err = CallBack->GenerateMacUpdate(hCSP,
                                                                  CCHandle,
                                                                  DataBufs,
                                                                  DataBufCount));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_GenerateMacFinal
 *
 *Description:
 *   Exported API complete the stage mac generate operation
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  Mac (output) - pointer to mac result
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_GenerateMacFinal(CSSM_CC_HANDLE CCHandle,
                      CSSM_DATA_PTR Mac)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_GenerateMacFinal);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_NO_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                hCSP = AttachRecord->AttachHandle;

                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GenerateMacFinal))
                        {
                            ERR(err = CallBack->GenerateMacFinal(hCSP,
                                                                 CCHandle,
                                                                 Mac));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_VerifyMac
 *
 *Description:
 *   Exported API for verifying a MAC
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  DataBufs (input) - array of data buffers that holds
 *                     information that needs to be hashed
 *  DataBufCount (input) - number of data buffers pointed to
 *                         by parameter DataBufs
 *  Mac (output) - pointer to mac to verify
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_VerifyMac(CSSM_CC_HANDLE CCHandle,
               const CSSM_DATA *DataBufs,
               uint32 DataBufCount,
               const CSSM_DATA *Mac)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_VerifyMac);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;

                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->VerifyMac))
                        {
                            ERR(err = CallBack->VerifyMac(hCSP,
                                                          CCHandle,
                                                          Context,
                                                          DataBufs,
                                                          DataBufCount,
                                                          Mac));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_VerifyMacInit
 *
 *Description:
 *   Exported API for stage verification of a mac
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_VerifyMacInit(CSSM_CC_HANDLE CCHandle)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_VerifyMacInit);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;

                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->VerifyMacInit))
                        {
                            ERR(err = CallBack->VerifyMacInit(hCSP,
                                                              CCHandle,
                                                              Context));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }

                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_VerifyMacUpdate
 *
 *Description:
 *   Exported API for updating the data to verify the mac
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  DataBufs (input) - array of data buffers that holds
 *                     information that needs to be hashed
 *  DataBufCount (input) - number of data buffers pointed to
 *                         by parameter DataBufs
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_VerifyMacUpdate(CSSM_CC_HANDLE CCHandle,
                     const CSSM_DATA *DataBufs,
                     uint32 DataBufCount)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_VerifyMacUpdate);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_NO_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                hCSP = AttachRecord->AttachHandle;

                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->VerifyMacUpdate))
                        {
                            ERR(err = CallBack->VerifyMacUpdate(hCSP,
                                                                CCHandle,
                                                                DataBufs,
                                                                DataBufCount));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_VerifyMacFinal
 *
 *Description:
 *   Exported API complete the stage mac verify operation
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  Mac (output) - pointer to mac result
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_VerifyMacFinal(CSSM_CC_HANDLE CCHandle,
                    const CSSM_DATA *Mac)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_VerifyMacFinal);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_NO_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                hCSP = AttachRecord->AttachHandle;

                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->VerifyMacFinal))
                        {
                            ERR(err = CallBack->VerifyMacFinal(hCSP,
                                                               CCHandle,
                                                               Mac));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
            }
        }
    }

    RETURN(err)
}



CSSM_PRIVILEGE
cssm_GetUSEETag( )
{
    /* Get privilege and return it here! */
    CSSM_PRIVILEGE privilege = CSSM_USEE_NONE;

    if( PrivilegeScope == CSSM_PRIVILEGE_SCOPE_PROCESS )
        cssm_GetGlobalPrivilege( &privilege );
    
    else if( PrivilegeScope == CSSM_PRIVILEGE_SCOPE_THREAD )
        cssm_GetThreadPrivilege ( &privilege );

    return privilege;
}

CSSM_RETURN cssm_GetAlgorithmId(cssm_INTERNAL_CONTEXT_PTR ContextInfo, CSSM_ALGORITHMS *AlgorithmId)
{
    CSSM_RETURN err = CSSM_OK;

    if (ContextInfo == NULL || ContextInfo->Context == NULL || AlgorithmId == NULL)
    {
        err = CSSMERR_CSSM_INVALID_POINTER;
    }
    else
    {
        *AlgorithmId = ContextInfo->Context->AlgorithmType;
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_EncryptData
 *
 *Description:
 *   Exported API for encrypting the data. Privilege that is passed down to the SP is the 
 *   the privilege that is set by CSSM_SetPrivilege
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  ClearBufs (input) - array of clear buffers that holds
 *                     information that needs to be encrypted
 *  ClearBufCount (input) - number of clear buffers pointed to
 *                         by parameter DataBufs
 *  CipherBufs (output) - array of cipher buffers that holds
 *                     the encrypted data
 *  CipherBufCount (input) - number of cipher buffers pointed to
 *                         by parameter DataBufs
 *  bytesEncrypted (output) - count of bytes encrypted
 *  RemData (output) - buffers that holds encrypted data that
 *                     did not fit in CipherBufs
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
/* Returns the encrypted data */
CSSM_RETURN CSSMAPI CSSM_EncryptData (CSSM_CC_HANDLE CCHandle,
                                      const CSSM_DATA *ClearBufs,
                                      uint32 ClearBufCount,
                                      CSSM_DATA_PTR CipherBufs,
                                      uint32 CipherBufCount,
                                      uint32 *bytesEncrypted,
                                      CSSM_DATA_PTR RemData)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;
    CSSM_PRIVILEGE Privilege;

    TRACE(cssm_trace_info, FNCODE_CSSM_EncryptData);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif
        if(CSSM_OK == err)
        {
            /* get CSP handle and actual context information */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        &AttachRecord,
                        CSSM_READER_LOCK,
                        &InternalCtx,
                        CSSM_READER_LOCK)) ;
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {

                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;
                            

                        Privilege = cssm_GetUSEETag();
                        /*
                         * Call the callback function. 
                         * Make sure that function structure and function pointer is valid 
                         */
                        if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->EncryptData)) 
                        {
                            err = CallBack->EncryptData (hCSP, CCHandle, Context, ClearBufs,
                                                        ClearBufCount, CipherBufs, CipherBufCount,
                                                        bytesEncrypted, RemData, Privilege);
                        }
                        else 
                        {
                            err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                        }
                                        
                    }
                    
                    /*
                    * cssm_ReleaseModuleRecord and multitask mutex.
                    */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                cssm_ReleaseAttachLock(AttachRecord,
                                    CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                    CSSM_READER_LOCK);
            }
        }
    }
    RETURN(err)
}
/*---------------------------------------------------------------
 *
 *Name: CSSM_EncryptDataP
 *
 *Description:
 *   Exported API for encrypting the data. CSSM checks the app's manifest 
 *  for th eprivilege if it is signed or CSSM's manifest and matches with the SP's manifest
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  ClearBufs (input) - array of clear buffers that holds
 *                     information that needs to be encrypted
 *  ClearBufCount (input) - number of clear buffers pointed to
 *                         by parameter DataBufs
 *  CipherBufs (output) - array of cipher buffers that holds
 *                     the encrypted data
 *  CipherBufCount (input) - number of cipher buffers pointed to
 *                         by parameter DataBufs
 *  bytesEncrypted (output) - count of bytes encrypted
 *  RemData (output) - buffers that holds encrypted data that
 *                     did not fit in CipherBufs
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_EncryptDataP(
                        CSSM_CC_HANDLE CCHandle,
                        const CSSM_DATA *ClearBufs,
                        uint32 ClearBufCount,
                        CSSM_DATA_PTR CipherBufs,
                        uint32 CipherBufCount,
                        uint32 *bytesEncrypted,
                        CSSM_DATA_PTR RemData,
                        CSSM_PRIVILEGE Privilege )
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;
    cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord = NULL;

    TRACE(cssm_trace_info, FNCODE_CSSM_EncryptDataP);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED
        cssm_CALLER_LIST_ENTRY_PTR CallerListEntry = NULL;
        cssm_CALLER_LIST_PTR CallerList = NULL;

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if((0 != Privilege) ||
           (cssm_GetPVCPolicy() & CSSM_PVC_APP))
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           &CallerList,
                                                           CSSM_READER_LOCK,
                                                           &CallerListEntry));
            if(CSSM_OK != err)
            {
                if((cssm_GetPVCPolicy() & CSSM_PVC_APP))
                {
                    ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
                else
                {
                    err = CSSM_OK;
                }
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* get CSP handle and actual context information */

                ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                            CSSM_NO_LOCK,
                            NULL,
                            CSSM_NO_LOCK,
                            &ServiceRecord,
                            CSSM_READER_LOCK,
                            &AttachRecord,
                            CSSM_READER_LOCK,
                            &InternalCtx,
                            CSSM_READER_LOCK)) ;
                if(CSSM_OK != err)
                {
                    /* we couldn't find the context */
                    ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
                }
                else
                {
#ifdef ISL_INCLUDED
                    if( Privilege != 0 )
                    {
                        if(!cssm_PrivilegeMatchForCallerAndSP(Privilege,
                                                              CallerListEntry,
                                                              AttachRecord,
                                                              ServiceRecord) )
                        {
                            ERR(err = CSSMERR_CSSM_PRIVILEGE_NOT_GRANTED);
                        }
                    }
#endif
                    cssm_ReleaseServiceLock( ServiceRecord, CSSM_READER_LOCK );

                    if( err == CSSM_OK )
                    {   
                        /* this function will lock the mutex if necessary */
                        ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                        if(CSSM_OK == err)
                        {

                            /* get the callback functions */
                            ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                                  CSSM_SERVICE_CSP,
                                                                  (void *)&CallBack)); /* ifdef VMS apply cast */
                            if(CSSM_OK == err)
                            {
                                Context = InternalCtx->Context;
                                hCSP = InternalCtx->CSPHandle;
                                        

                                /*
                                 * Call the callback function. 
                                 * Make sure that function structure and function pointer is valid 
                                 */
                                if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->EncryptData)) 
                                {
                                    err = CallBack->EncryptData (hCSP, CCHandle, Context, ClearBufs,
                                                                ClearBufCount, CipherBufs, CipherBufCount,
                                                                bytesEncrypted, RemData, Privilege);
                                }
                                else 
                                {
                                    err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                                }
                            }
                            /*
                            * cssm_ReleaseModuleRecord and multitask mutex.
                            */
                            cssm_UnlockMultiTaskMutex(AttachRecord);
                        }
                    }

                    cssm_ReleaseAttachLock(AttachRecord,
                                        CSSM_READER_LOCK);
                    cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
                }

#ifdef ISL_INCLUDED
            /* if we locked the caller record, we need to release it */
            if(NULL != CallerList)
            {
                cssm_ReleaseCallerListEntry(CallerList,
                                            CSSM_READER_LOCK);
            }
#endif

        }
    }
    RETURN(err)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_EncryptDataInit
 *
 *Description:
 *   Exported API for initializing the staged encrypt operation
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
 CSSM_RETURN CSSMAPI CSSM_EncryptDataInit (CSSM_CC_HANDLE CCHandle)
 {
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;
    CSSM_PRIVILEGE Privilege;

    TRACE(cssm_trace_info, FNCODE_CSSM_EncryptDataInit);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PRIVILEGE_NOT_GRANTED);
            }
        }

#endif
        if(CSSM_OK == err)
        {

            /* get CSP handle and actual context information */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        &AttachRecord,
                        CSSM_READER_LOCK,
                        &InternalCtx,
                        CSSM_READER_LOCK)) ;
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {

                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;
                            
                        Privilege = cssm_GetUSEETag();
                        /*
                         * Call the callback function. 
                         * Make sure that function structure and function pointer is valid 
                         */
                        if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->EncryptDataInit)) 
                        {
                            err = CallBack->EncryptDataInit (hCSP, CCHandle, Context, Privilege);
                        }
                        else 
                        {
                            err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                        }
                    }
                        
                    /*
                    * cssm_ReleaseModuleRecord and multitask mutex.
                    */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }
                cssm_ReleaseAttachLock(AttachRecord,
                                    CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                    CSSM_READER_LOCK);
            }
        }
    }
    RETURN(err)
}

CSSM_RETURN CSSMAPI CSSM_EncryptDataInitP(
                        CSSM_CC_HANDLE CCHandle,
                        CSSM_PRIVILEGE Privilege )
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;
    cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord = NULL;

    TRACE(cssm_trace_info, FNCODE_CSSM_EncryptDataP);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED
        cssm_CALLER_LIST_PTR CallerList = NULL;
        cssm_CALLER_LIST_ENTRY_PTR CallerListEntry = NULL;
        
        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if((0 != Privilege) ||
           (cssm_GetPVCPolicy() & CSSM_PVC_APP))
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           &CallerList,
                                                           CSSM_READER_LOCK,
                                                           &CallerListEntry));
            if(CSSM_OK != err)
            {
                if((cssm_GetPVCPolicy() & CSSM_PVC_APP))
                {
                    ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
                else
                {
                    err = CSSM_OK;
                }
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* get CSP handle and actual context information */

                ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                            CSSM_NO_LOCK,
                            NULL,
                            CSSM_NO_LOCK,
                            &ServiceRecord,
                            CSSM_READER_LOCK,
                            &AttachRecord,
                            CSSM_READER_LOCK,
                            &InternalCtx,
                            CSSM_READER_LOCK)) ;
                if(CSSM_OK != err)
                {
                    /* we couldn't find the context */
                    ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
                }
                else
                {
#ifdef ISL_INCLUDED
                    if( Privilege != 0)
                    {
                        if(!cssm_PrivilegeMatchForCallerAndSP( Privilege,
                                                           CallerListEntry,
                                                           AttachRecord,
                                                           ServiceRecord) )
                        {
                            err = CSSMERR_CSSM_PRIVILEGE_NOT_GRANTED;
                        }
                    }
#endif
                    cssm_ReleaseServiceLock( ServiceRecord, CSSM_READER_LOCK );
                    if( err == CSSM_OK )
                    {   
                        /* this function will lock the mutex if necessary */
                        ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                        if(CSSM_OK == err)
                        {

                            /* get the callback functions */
                            ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                                  CSSM_SERVICE_CSP,
                                                                  (void *)&CallBack)); /* ifdef VMS apply cast */
                            if(CSSM_OK == err)
                            {
                                Context = InternalCtx->Context;
                                hCSP = InternalCtx->CSPHandle;
                                        
                                /*
                                 * Call the callback function. 
                                 * Make sure that function structure and function pointer is valid 
                                 */
                                if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->EncryptDataInit)) 
                                {
                                    err = CallBack->EncryptDataInit (hCSP, CCHandle, Context, Privilege);
                                }
                                else 
                                {
                                    err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                                }
                                                    
                            }

                            /*
                            * cssm_ReleaseModuleRecord and multitask mutex.
                            */
                            cssm_UnlockMultiTaskMutex(AttachRecord);
                        }
                    }

                    cssm_ReleaseAttachLock(AttachRecord,
                                        CSSM_READER_LOCK);
                    cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
                }

#ifdef ISL_INCLUDED
            /* if we locked the caller record, we need to release it */
            if(NULL != CallerList)
            {
                cssm_ReleaseCallerListEntry(CallerList,
                                            CSSM_READER_LOCK);
            }
#endif

        }
    }
    RETURN(err)
}


/*---------------------------------------------------------------
 *
 *Name: CSSM_EncryptDataUpdate
 *
 *Description:
 *   Exported API for updating the encrypt staged operation
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  ClearBufs (input) - array of clear buffers that holds
 *                     information that needs to be encrypted
 *  ClearBufCount (input) - number of clear buffers pointed to
 *                         by parameter DataBufs
 *  CipherBufs (output) - array of cipher buffers that holds
 *                     the encrypted data
 *  CipherBufCount (input) - number of cipher buffers pointed to
 *                         by parameter DataBufs
 *  bytesEncrypted (output) - count of bytes encrypted
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
 CSSM_RETURN CSSMAPI CSSM_EncryptDataUpdate (CSSM_CC_HANDLE CCHandle,
     const CSSM_DATA *ClearBufs,
     uint32 ClearBufCount,
     CSSM_DATA_PTR CipherBufs,
     uint32 CipherBufCount,
     uint32 *bytesEncrypted)
 {
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;

    TRACE(cssm_trace_info, FNCODE_CSSM_EncryptData);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif
        if(CSSM_OK == err)
        {

            /* get CSP handle and actual context information */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        &AttachRecord,
                        CSSM_READER_LOCK,
                        &InternalCtx,
                        CSSM_READER_LOCK)) ;
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {

                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;
                            

                        /*
                         * Call the callback function. 
                         * Make sure that function structure and function pointer is valid 
                         */
                        if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->EncryptDataUpdate)) 
                        {
                            ERR( err = CallBack->EncryptDataUpdate (hCSP, CCHandle, ClearBufs,
                                                                    ClearBufCount, CipherBufs,
                                                                    CipherBufCount, bytesEncrypted));
                        }
                        else 
                        {
                            err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                        }
                    }
                        
                    /*
                    * cssm_ReleaseModuleRecord and multitask mutex.
                    */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }
                cssm_ReleaseAttachLock(AttachRecord,
                                    CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                    CSSM_READER_LOCK);
            }
        }
    }
    RETURN(err)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_EncryptDataFinal
 *
 *Description:
 *   Exported API complete the stage encrypt operation
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  RemData (output) - buffers that holds encrypted data that
 *                     did not fit in CipherBufs
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
/* Returns the encrypted data */
CSSM_RETURN CSSMAPI CSSM_EncryptDataFinal (CSSM_CC_HANDLE CCHandle,
                                           CSSM_DATA_PTR RemData)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;

    TRACE(cssm_trace_info, FNCODE_CSSM_EncryptDataFinal);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif
        if(CSSM_OK == err)
        {

            /* get CSP handle and actual context information */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        &AttachRecord,
                        CSSM_READER_LOCK,
                        &InternalCtx,
                        CSSM_READER_LOCK)) ;
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {

                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;
                            
                        /*
                         * Call the callback function. 
                         * Make sure that function structure and function pointer is valid 
                         */
                        if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->EncryptDataFinal)) 
                        {
                            ERR( err = CallBack->EncryptDataFinal (hCSP, CCHandle, RemData));
                        }
                        else 
                        {
                            err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                        }
                                        
                    }
                        
                    /*
                    * cssm_ReleaseModuleRecord and multitask mutex.
                    */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }
                cssm_ReleaseAttachLock(AttachRecord,
                                    CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }
    RETURN(err)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_DecryptData
 *
 *Description:
 *   Exported API for decrypting the data
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  CipherBufs (input) - array of cipher buffers that holds
 *                     the encrypted data
 *  CipherBufCount (input) - number of cipher buffers pointed to
 *                         by parameter DataBufs
 *  ClearBufs (output) - array of clear buffers that holds
 *                     information that needs to be decrypted
 *  ClearBufCount (input) - number of clear buffers pointed to
 *                         by parameter DataBufs
 *  bytesEncrypted (output) - count of bytes decrypted
 *  RemData (output) - buffers that holds decrypted data that
 *                     did not fit in ClearBufs
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
/* Returns the decrypted data */
CSSM_RETURN CSSMAPI CSSM_DecryptData (CSSM_CC_HANDLE CCHandle,
                                      const CSSM_DATA *CipherBufs,
                                      uint32 CipherBufCount,
                                      CSSM_DATA_PTR ClearBufs,
                                      uint32 ClearBufCount,
                                      uint32 *bytesDecrypted,
                                      CSSM_DATA_PTR RemData)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;
    CSSM_PRIVILEGE Privilege;

    TRACE(cssm_trace_info, FNCODE_CSSM_DecryptData);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif
        if(CSSM_OK == err)
        {

            /* get CSP handle and actual context information */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        &AttachRecord,
                        CSSM_READER_LOCK,
                        &InternalCtx,
                        CSSM_READER_LOCK)) ;
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {

                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;
                            
                            

                        Privilege = cssm_GetUSEETag();
                        /*
                         * Call the callback function. 
                         * Make sure that function structure and function pointer is valid 
                         */
                        if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->DecryptData)) 
                        {
                            err = CallBack->DecryptData (hCSP, CCHandle, Context, 
                                                         CipherBufs, CipherBufCount, 
                                                         ClearBufs, ClearBufCount,
                                                         bytesDecrypted,RemData, Privilege);
                        }
                        else 
                        {
                            err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                        }
                    }
                        
                    /*
                    * cssm_ReleaseModuleRecord and multitask mutex.
                    */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                cssm_ReleaseAttachLock(AttachRecord,
                                    CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                    CSSM_READER_LOCK);
            }
        }
    }
    RETURN(err)
}

CSSM_RETURN CSSMAPI CSSM_DecryptDataP(
                        CSSM_CC_HANDLE CCHandle,
                        const CSSM_DATA *CipherBufs,
                        uint32 CipherBufCount,
                        CSSM_DATA_PTR ClearBufs,
                        uint32 ClearBufCount,
                        uint32 *bytesDecrypted,
                        CSSM_DATA_PTR RemData,
                        CSSM_PRIVILEGE Privilege )
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;
    cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord = NULL;

    TRACE(cssm_trace_info, FNCODE_CSSM_DecryptDataP);
    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED
        cssm_CALLER_LIST_PTR CallerList = NULL;
        cssm_CALLER_LIST_ENTRY_PTR CallerListEntry = NULL;
        
        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if((0 != Privilege) ||
           (cssm_GetPVCPolicy() & CSSM_PVC_APP))
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           &CallerList,
                                                           CSSM_READER_LOCK,
                                                           &CallerListEntry));
            if(CSSM_OK != err)
            {
                if((cssm_GetPVCPolicy() & CSSM_PVC_APP))
                {
                    ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
                else
                {
                    err = CSSM_OK;
                }
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* get CSP handle and actual context information */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        &ServiceRecord,
                        CSSM_READER_LOCK,
                        &AttachRecord,
                        CSSM_READER_LOCK,
                        &InternalCtx,
                        CSSM_READER_LOCK)) ;
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
#ifdef ISL_INCLUDED
                if( Privilege != 0)
                {
                    if(!cssm_PrivilegeMatchForCallerAndSP( Privilege,
                                                       CallerListEntry,
                                                       AttachRecord,
                                                       ServiceRecord) )
                    {
                        err = CSSMERR_CSSM_PRIVILEGE_NOT_GRANTED;
                    }
                }
#endif
                cssm_ReleaseServiceLock( ServiceRecord, CSSM_READER_LOCK );
                if( err == CSSM_OK )
                {   
                    /* this function will lock the mutex if necessary */
                    ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                    if(CSSM_OK == err)
                    {

                        /* get the callback functions */
                        ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                              CSSM_SERVICE_CSP,
                                                              (void *)&CallBack)); /* ifdef VMS apply cast */
                        if(CSSM_OK == err)
                        {
                            Context = InternalCtx->Context;
                            hCSP = InternalCtx->CSPHandle;
                                    

                            /*
                             * Call the callback function. 
                             * Make sure that function structure and function pointer is valid 
                             */
                            if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->DecryptData)) 
                            {
                                err = CallBack->DecryptData (hCSP, CCHandle, Context, CipherBufs,
                                                            CipherBufCount, ClearBufs, ClearBufCount,
                                                            bytesDecrypted, RemData, Privilege);
                            }
                            else 
                            {
                                err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                            }
                        }

                        /*
                        * cssm_ReleaseModuleRecord and multitask mutex.
                        */
                        cssm_UnlockMultiTaskMutex(AttachRecord);
                    }
                }

                cssm_ReleaseAttachLock(AttachRecord,
                                    CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                    CSSM_READER_LOCK);
            }

#ifdef ISL_INCLUDED
            /* if we locked the caller record, we need to release it */
            if(NULL != CallerList)
            {
                cssm_ReleaseCallerListEntry(CallerList,
                                            CSSM_READER_LOCK);
            }
#endif

        }
    }
    RETURN(err)

}

/*---------------------------------------------------------------
 *
 *Name: CSSM_DecryptDataInit
 *
 *Description:
 *   Exported API for initializing the stage decrypt operation.
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
 CSSM_RETURN CSSMAPI CSSM_DecryptDataInit (CSSM_CC_HANDLE CCHandle)
 {
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;
    CSSM_PRIVILEGE Privilege;
    TRACE(cssm_trace_info, FNCODE_CSSM_DecryptDataInit);
    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif
        if(CSSM_OK == err)
        {

            /* get CSP handle and actual context information */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        &AttachRecord,
                        CSSM_READER_LOCK,
                        &InternalCtx,
                        CSSM_READER_LOCK)) ;
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {

                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;
                            
                        Privilege = cssm_GetUSEETag();
                        /*
                         * Call the callback function. 
                         * Make sure that function structure and function pointer is valid 
                         */
                        if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->DecryptDataInit)) 
                        {
                             err = CallBack->DecryptDataInit (hCSP, CCHandle, Context, Privilege);
                        }
                        else 
                        {
                            err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                        }
                    }
                        
                    /*
                    * cssm_ReleaseModuleRecord and multitask mutex.
                    */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                cssm_ReleaseAttachLock(AttachRecord,
                                    CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                    CSSM_READER_LOCK);
            }
        }
    }
    RETURN(err)

}

CSSM_RETURN CSSMAPI CSSM_DecryptDataInitP(
                        CSSM_CC_HANDLE CCHandle,
                        CSSM_PRIVILEGE Privilege )
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;
    cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord = NULL;

    TRACE(cssm_trace_info, CSSM_DecryptDataInitP);
    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED
        cssm_CALLER_LIST_PTR CallerList = NULL;
        cssm_CALLER_LIST_ENTRY_PTR CallerListEntry = NULL;
        
        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);
        /* if we have to, we check the application here */
        if((0 != Privilege) ||
           (cssm_GetPVCPolicy() & CSSM_PVC_APP))
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           &CallerList,
                                                           CSSM_READER_LOCK,
                                                           &CallerListEntry));
            if(CSSM_OK != err)
            {
                if((cssm_GetPVCPolicy() & CSSM_PVC_APP))
                {
                    ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
                else
                {
                    err = CSSM_OK;
                }
            }
        }
#endif

        if(CSSM_OK == err)
        {
            /* get CSP handle and actual context information */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        &ServiceRecord,
                        CSSM_READER_LOCK,
                        &AttachRecord,
                        CSSM_READER_LOCK,
                        &InternalCtx,
                        CSSM_READER_LOCK)) ;
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
#ifdef ISL_INCLUDED
                if( Privilege != 0)
                {
                    if(!cssm_PrivilegeMatchForCallerAndSP( Privilege,
                                                       CallerListEntry,
                                                       AttachRecord,
                                                       ServiceRecord) )
                    {
                        err = CSSMERR_CSSM_PRIVILEGE_NOT_GRANTED;
                    }
                }
#endif
                cssm_ReleaseServiceLock( ServiceRecord, CSSM_READER_LOCK );
                if( err == CSSM_OK )
                {   
                    /* this function will lock the mutex if necessary */
                    ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                    if(CSSM_OK == err)
                    {

                        /* get the callback functions */
                        ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                              CSSM_SERVICE_CSP,
                                                              (void *)&CallBack)); /* ifdef VMS apply cast */
                        if(CSSM_OK == err)
                        {
                            Context = InternalCtx->Context;
                            hCSP = InternalCtx->CSPHandle;
                                    

                            /*
                             * Call the callback function. 
                             * Make sure that function structure and function pointer is valid 
                             */
                            if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->DecryptDataInit)) 
                            {
                                err = CallBack->DecryptDataInit (hCSP, CCHandle, Context, Privilege);
                            }
                            else 
                            {
                                err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                            }
                        }

                        /*
                        * cssm_ReleaseModuleRecord and multitask mutex.
                        */
                        cssm_UnlockMultiTaskMutex(AttachRecord);
                    }
                }

                cssm_ReleaseAttachLock(AttachRecord,
                                    CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                    CSSM_READER_LOCK);
            }

#ifdef ISL_INCLUDED
            /* if we locked the caller record, we need to release it */
            if(NULL != CallerList)
            {
                cssm_ReleaseCallerListEntry(CallerList,
                                            CSSM_READER_LOCK);
            }
#endif

        }
    }
    RETURN(err)

}

/*---------------------------------------------------------------
 *
 *Name: CSSM_DecryptUpdate
 *
 *Description:
 *   Exported API updating the staged decrypt operation with data
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  CipherBufs (input) - array of cipher buffers that holds
 *                     the encrypted data
 *  CipherBufCount (input) - number of cipher buffers pointed to
 *                         by parameter DataBufs
 *  ClearBufs (output) - array of clear buffers that holds
 *                     information that needs to be decrypted
 *  ClearBufCount (input) - number of clear buffers pointed to
 *                         by parameter DataBufs
 *  bytesEncrypted (output) - count of bytes decrypted
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
 CSSM_RETURN CSSMAPI CSSM_DecryptDataUpdate (CSSM_CC_HANDLE CCHandle,
     const CSSM_DATA *CipherBufs,
     uint32 CipherBufCount,
     CSSM_DATA_PTR ClearBufs,
     uint32 ClearBufCount,
     uint32 *bytesDecrypted)
 {
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;
    CSSM_PRIVILEGE Privilege;

    TRACE(cssm_trace_info, FNCODE_CSSM_DecryptDataUpdate);
    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif
        if(CSSM_OK == err)
        {

            /* get CSP handle and actual context information */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        &AttachRecord,
                        CSSM_READER_LOCK,
                        &InternalCtx,
                        CSSM_READER_LOCK)) ;
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {

                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;
                            

                        Privilege = cssm_GetUSEETag();
                        /*
                         * Call the callback function. 
                         * Make sure that function structure and function pointer is valid 
                         */
                        if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->DecryptDataUpdate)) 
                        {
                            ERR( err = CallBack->DecryptDataUpdate (hCSP, CCHandle, CipherBufs,
                                                                    CipherBufCount, ClearBufs, ClearBufCount,
                                                                    bytesDecrypted));
                        }
                        else 
                        {
                            err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                        }
                    }
                        
                    /*
                    * cssm_ReleaseModuleRecord and multitask mutex.
                    */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                cssm_ReleaseAttachLock(AttachRecord,
                                    CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                    CSSM_READER_LOCK);
            }
        }
    }
    RETURN(err)

}

/*---------------------------------------------------------------
*
*Name: CSSM_DecryptDataFinal
*
*Description:
*   Exported API for completing the stage decrypt operation
*
*Parameters: 
*  CCHandle (input) - the context handle
*  RemData (output) - buffers that holds decrypted data that
*                     did not fit in ClearBufs during the stage
*                     decrypt operation
*
*Returns:
*   CSSM_FAIL - CSP wasn't able to complete function
*   CSSM_OK - Call successful executed by CSP add-in
*
*-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_DecryptDataFinal (CSSM_CC_HANDLE CCHandle,
                                           CSSM_DATA_PTR RemData)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;
    CSSM_PRIVILEGE Privilege;

    TRACE(cssm_trace_info, FNCODE_CSSM_DecryptDataFinal);
    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif
        if(CSSM_OK == err)
        {

            /* get CSP handle and actual context information */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        &AttachRecord,
                        CSSM_READER_LOCK,
                        &InternalCtx,
                        CSSM_READER_LOCK)) ;
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {

                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;
                            
                        Privilege = cssm_GetUSEETag();
                        /*
                         * Call the callback function. 
                         * Make sure that function structure and function pointer is valid 
                         */
                        if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->DecryptDataFinal)) 
                        {
                            err = CallBack->DecryptDataFinal (hCSP, CCHandle, RemData);
                        }
                        else 
                        {
                            err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                        }
                                        
                    }
                        
                    /*
                    * cssm_ReleaseModuleRecord and multitask mutex.
                    */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }
                cssm_ReleaseAttachLock(AttachRecord,
                                    CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                    CSSM_READER_LOCK);
            }
        }
    }
    RETURN(err)

}

/*---------------------------------------------------------------
 *
 *Name: CSSM_QueryKeySizeInBits
 *
 *Description:
 *   Exported API for querying key size
 *
 *Parameters: 
 *  CSPHandle (input) - the CSP handle
 *  CCHandle (input) - context handle containing the key
 *  Key (input/optional) - key for which size is to be determined
 *  KeySize (output) -  pointer to keysize
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_QueryKeySizeInBits(CSSM_CSP_HANDLE CSPHandle,
                        CSSM_CC_HANDLE CCHandle,
                        const CSSM_KEY *Key,
                        CSSM_KEY_SIZE_PTR KeySize)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CSP_HANDLE tempCSPHandle = CSSM_INVALID_HANDLE;
    CSSM_CONTEXT_PTR tempContext = NULL;
    CSSM_CC_HANDLE tempCCHandle = CSSM_INVALID_HANDLE;

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* We couldn't find the context, so we'll try the attach
                   record.
                */
                ERR(err = cssm_FindAttachAndMultiLock(CSPHandle,
                                                      CSSM_NO_LOCK,
                                                      NULL,
                                                      CSSM_NO_LOCK,
                                                      NULL,
                                                      CSSM_NO_LOCK,
                                                      &AttachRecord,
                                                      CSSM_READER_LOCK));

                /* If the "find" failed, this is a meaningless assignment, but
                   if it didn't (the expected case), it's necessary
                */
                tempCSPHandle = CSPHandle;
            }
            else
            {
                tempCCHandle = CCHandle;
                tempContext = InternalCtx->Context;
                tempCSPHandle = InternalCtx->CSPHandle;
            }

            if(CSSM_OK == err)
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the callback function.
                         * Make sure that function structure and function
                         * pointer is valid
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->QueryKeySizeInBits))
                        {
                            ERR( err = CallBack->QueryKeySizeInBits(tempCSPHandle,
                                                                    tempCCHandle,
                                                                    tempContext,
                                                                    Key,
                                                                    KeySize));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * Release module record and multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);

                if(NULL != InternalCtx)
                {
                    cssm_ReleaseContextLock(InternalCtx,
                                            CSSM_READER_LOCK);
                }
            }
        }
    }

    return err;
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_GenerateKey
 *
 *Description:
 *   Exported API for generating a symmetric.
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  StorageMask (input) - how to store the key.
 *  Key (output) - pointer to the generate key
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
 CSSM_RETURN CSSMAPI CSSM_GenerateKey (CSSM_CC_HANDLE CCHandle,
     uint32 KeyUsage,
     uint32 KeyAttr,
     const CSSM_DATA *KeyLabel,
     const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
     CSSM_KEY_PTR Key)
 {
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    CSSM_PRIVILEGE Privilege = 0;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RESOURCE_CONTROL_CONTEXT ProxyCallback;
    CSSM_RESOURCE_CONTROL_CONTEXT_PTR pProxyCallback= NULL;
    CSSM_CHALLENGE_CALLBACK_PROXY_MODULEHANDLE ChallengeProxyCtx;
    CSSM_SUBJECT_CALLBACK_PROXY_MODULEHANDLE SubjectProxyCtx;

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_GenerateKey);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;
                        if( CredAndAclEntry != NULL) 
                        {
                            if( !port_IsBadReadPtr( CredAndAclEntry, sizeof(CSSM_RESOURCE_CONTROL_CONTEXT) ) )
                            {
                                cssm_SetResourceControlProxyModuleHandle(hCSP, CredAndAclEntry, &ChallengeProxyCtx,
                                                           &SubjectProxyCtx,&ProxyCallback);
                                 
                                pProxyCallback = &ProxyCallback;
                            }
                            else
                                err = CSSMERR_CSSM_INVALID_POINTER;
                        }
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GenerateKey))
                        {
                            Privilege = cssm_GetUSEETag();

                            ERR(err = CallBack->GenerateKey(hCSP,
                                                            CCHandle,
                                                            Context,
                                                            KeyUsage,
                                                            KeyAttr,
                                                            KeyLabel,
                                                            pProxyCallback,
                                                            Key,
                                                            Privilege));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }

                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }
    RETURN(err)
}



CSSM_RETURN CSSMAPI
CSSM_GenerateKeyP(CSSM_CC_HANDLE CCHandle,
                  uint32 KeyUsage,
                  uint32 KeyAttr,
                  const CSSM_DATA *KeyLabel,
                  const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
                  CSSM_KEY_PTR Key,
                  CSSM_PRIVILEGE Privilege)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;
    cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord = NULL;
    CSSM_RESOURCE_CONTROL_CONTEXT ProxyCallback;
    CSSM_RESOURCE_CONTROL_CONTEXT_PTR pProxyCallback= NULL;
    CSSM_CHALLENGE_CALLBACK_PROXY_MODULEHANDLE ChallengeProxyCtx;
    CSSM_SUBJECT_CALLBACK_PROXY_MODULEHANDLE SubjectProxyCtx;


    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {
#ifdef ISL_INCLUDED
        cssm_CALLER_LIST_PTR CallerList = NULL;
        cssm_CALLER_LIST_ENTRY_PTR CallerListEntry = NULL;
        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if((0 != Privilege) ||
           (cssm_GetPVCPolicy() & CSSM_PVC_APP))
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           &CallerList,
                                                           CSSM_READER_LOCK,
                                                           &CallerListEntry));
            if(CSSM_OK != err)
            {
                if((cssm_GetPVCPolicy() & CSSM_PVC_APP))
                {
                    ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
                else
                {
                    err = CSSM_OK;
                }
            }
        }
#endif

        if(CSSM_OK == err)
        {
            /* get CSP handle and actual context information */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        &ServiceRecord,
                        CSSM_READER_LOCK,
                        &AttachRecord,
                        CSSM_READER_LOCK,
                        &InternalCtx,
                        CSSM_READER_LOCK)) ;
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
#ifdef ISL_INCLUDED
                if( Privilege != 0)
                {
                    if(!cssm_PrivilegeMatchForCallerAndSP( Privilege,
                                                       CallerListEntry,
                                                       AttachRecord,
                                                       ServiceRecord) )
                    {
                        err = CSSMERR_CSSM_PRIVILEGE_NOT_GRANTED;
                    }
                }
#endif
                cssm_ReleaseServiceLock( ServiceRecord, CSSM_READER_LOCK );
                if( err == CSSM_OK )
                {   
                    /* this function will lock the mutex if necessary */
                    ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                    if(CSSM_OK == err)
                    {

                        /* get the callback functions */
                        ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                              CSSM_SERVICE_CSP,
                                                              (void *)&CallBack)); /* ifdef VMS apply cast */
                        if(CSSM_OK == err)
                        {
                            Context = InternalCtx->Context;
                            hCSP = InternalCtx->CSPHandle;

                            /* Do we have a callback to proxy? */
                            if( CredAndAclEntry != NULL) 
                            {
                                if( !port_IsBadReadPtr( CredAndAclEntry, sizeof(CSSM_RESOURCE_CONTROL_CONTEXT) ) )
                                {
                                    cssm_SetResourceControlProxyModuleHandle(hCSP, CredAndAclEntry, &ChallengeProxyCtx,
                                                               &SubjectProxyCtx,&ProxyCallback);
                                 
                                    pProxyCallback = &ProxyCallback;
                                }
                                else
                                    err = CSSMERR_CSSM_INVALID_POINTER;
                            }
                                /*
                                 * Call the callback function. 
                                 * Make sure that function structure and function pointer is valid 
                                 */
                            if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GenerateKey)) 
                            {
                                ERR(err = CallBack->GenerateKey(hCSP,
                                                                CCHandle,
                                                                Context,
                                                                KeyUsage,
                                                                KeyAttr,
                                                                KeyLabel,
                                                                pProxyCallback,
                                                                Key,
                                                                Privilege));
                            }
                            else 
                            {
                                err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                            }
                                                
                        }

                        /*
                        * cssm_ReleaseModuleRecord and multitask mutex.
                        */
                        cssm_UnlockMultiTaskMutex(AttachRecord);
                    }
                }
                cssm_ReleaseAttachLock(AttachRecord,
                                    CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                    CSSM_READER_LOCK);
            }

#ifdef ISL_INCLUDED
            /* if we locked the caller record, we need to release it */
            if(NULL != CallerList)
            {
                cssm_ReleaseCallerListEntry(CallerList,
                                            CSSM_READER_LOCK);
            }
#endif

        }
    }
    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_GenerateKeyPair
 *
 *Description:
 *   Exported API for generating an asymmetric key pair.
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  PublicStorageMask (input) - how to store the public key
 *  PublicKey (output) - pointer to the generated public key
 *  PrivateStorageMask (input) - how to store the private key
 *  PrivateKey (output) - pointer to the generated private key
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
 CSSM_RETURN CSSMAPI CSSM_GenerateKeyPair (CSSM_CC_HANDLE CCHandle,
     uint32 PublicKeyUsage,
     uint32 PublicKeyAttr,
     const CSSM_DATA *PublicKeyLabel,
     CSSM_KEY_PTR PublicKey,
     uint32 PrivateKeyUsage,
     uint32 PrivateKeyAttr,
     const CSSM_DATA *PrivateKeyLabel,
     const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
     CSSM_KEY_PTR PrivateKey)
 {
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    CSSM_RESOURCE_CONTROL_CONTEXT ProxyCallback;
    CSSM_RESOURCE_CONTROL_CONTEXT_PTR pProxyCallback= NULL;
    CSSM_CHALLENGE_CALLBACK_PROXY_MODULEHANDLE ChallengeProxyCtx;
    CSSM_SUBJECT_CALLBACK_PROXY_MODULEHANDLE SubjectProxyCtx;
    
    CSSM_CSP_HANDLE hCSP;
    CSSM_PRIVILEGE Privilege;

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_GenerateKey);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;

                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GenerateKeyPair))
                        {
                            Privilege = cssm_GetUSEETag();
                            if( CredAndAclEntry != NULL)
                            {
                                if( !port_IsBadReadPtr( CredAndAclEntry, sizeof( CSSM_RESOURCE_CONTROL_CONTEXT)) )
                                {
                                    cssm_SetResourceControlProxyModuleHandle(hCSP,CredAndAclEntry,&ChallengeProxyCtx,
                                                       &SubjectProxyCtx,&ProxyCallback);
                                                 
                                    pProxyCallback = &ProxyCallback;
                                }
                                else
                                {
                                    err = CSSMERR_CSSM_INVALID_POINTER;
                                }
                            }

                            ERR( err = CallBack->GenerateKeyPair (hCSP, CCHandle, Context,
                                                                PublicKeyUsage, PublicKeyAttr, 
                                                                PublicKeyLabel, PublicKey,
                                                                PrivateKeyUsage, PrivateKeyAttr, 
                                                                PrivateKeyLabel, pProxyCallback, 
                                                                PrivateKey, Privilege));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }
    RETURN(err)

}



CSSM_RETURN CSSMAPI
CSSM_GenerateKeyPairP(CSSM_CC_HANDLE CCHandle,
                      uint32 PublicKeyUsage,
                      uint32 PublicKeyAttr,
                      const CSSM_DATA *PublicKeyLabel,
                      CSSM_KEY_PTR PublicKey,
                      uint32 PrivateKeyUsage,
                      uint32 PrivateKeyAttr,
                      const CSSM_DATA *PrivateKeyLabel,
                      const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
                      CSSM_KEY_PTR PrivateKey,
                      CSSM_PRIVILEGE Privilege)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;
    cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord = NULL;
    CSSM_RESOURCE_CONTROL_CONTEXT ProxyCallback;
    CSSM_RESOURCE_CONTROL_CONTEXT_PTR pProxyCallback= NULL;
    CSSM_CHALLENGE_CALLBACK_PROXY_MODULEHANDLE ChallengeProxyCtx;
    CSSM_SUBJECT_CALLBACK_PROXY_MODULEHANDLE SubjectProxyCtx;


    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {
#ifdef ISL_INCLUDED
        cssm_CALLER_LIST_PTR CallerList = NULL;
        cssm_CALLER_LIST_ENTRY_PTR CallerListEntry = NULL;
        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if((0 != Privilege) ||
           (cssm_GetPVCPolicy() & CSSM_PVC_APP))
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           &CallerList,
                                                           CSSM_READER_LOCK,
                                                           &CallerListEntry));
            if(CSSM_OK != err)
            {
                if((cssm_GetPVCPolicy() & CSSM_PVC_APP))
                {
                    ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
                else
                {
                    err = CSSM_OK;
                }
            }
        }
#endif

        if(CSSM_OK == err)
        {
            /* get CSP handle and actual context information */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        &ServiceRecord,
                        CSSM_READER_LOCK,
                        &AttachRecord,
                        CSSM_READER_LOCK,
                        &InternalCtx,
                        CSSM_READER_LOCK)) ;
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
#ifdef ISL_INCLUDED
                if( Privilege != 0)
                {
                    if(!cssm_PrivilegeMatchForCallerAndSP( Privilege,
                                                       CallerListEntry,
                                                       AttachRecord,
                                                       ServiceRecord) )
                    {
                        err = CSSMERR_CSSM_PRIVILEGE_NOT_GRANTED;
                    }
                }
#endif
                cssm_ReleaseServiceLock( ServiceRecord, CSSM_READER_LOCK );
                if( err == CSSM_OK )
                {   
                    /* this function will lock the mutex if necessary */
                    ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                    if(CSSM_OK == err)
                    {

                        /* get the callback functions */
                        ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                              CSSM_SERVICE_CSP,
                                                              (void *)&CallBack)); /* ifdef VMS apply cast */
                        if(CSSM_OK == err)
                        {
                            Context = InternalCtx->Context;
                            hCSP = InternalCtx->CSPHandle;

                            /* Do we have a callback to proxy? */
                            if( CredAndAclEntry != NULL) 
                            {
                                if( !port_IsBadReadPtr( CredAndAclEntry, sizeof(CSSM_RESOURCE_CONTROL_CONTEXT) ) )
                                {
                                    cssm_SetResourceControlProxyModuleHandle(hCSP, CredAndAclEntry, &ChallengeProxyCtx,
                                                               &SubjectProxyCtx,&ProxyCallback);
                                 
                                    pProxyCallback = &ProxyCallback;
                                }
                                else
                                    err = CSSMERR_CSSM_INVALID_POINTER;
                            }
                                /*
                                 * Call the callback function. 
                                 * Make sure that function structure and function pointer is valid 
                                 */
                            if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GenerateKeyPair)) 
                            {
                                ERR(err = CallBack->GenerateKeyPair(hCSP,
                                                                    CCHandle,
                                                                    Context,
                                                                    PublicKeyUsage,
                                                                    PublicKeyAttr,
                                                                    PublicKeyLabel,
                                                                    PublicKey,
                                                                    PrivateKeyUsage,
                                                                    PrivateKeyAttr,
                                                                    PrivateKeyLabel,
                                                                    pProxyCallback,
                                                                    PrivateKey,
                                                                    Privilege));
                            }
                            else 
                            {
                                err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                            }
                                                
                        }

                        /*
                        * cssm_ReleaseModuleRecord and multitask mutex.
                        */
                        cssm_UnlockMultiTaskMutex(AttachRecord);
                    }
                }
                cssm_ReleaseAttachLock(AttachRecord,
                                    CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                    CSSM_READER_LOCK);
            }

#ifdef ISL_INCLUDED
            /* if we locked the caller record, we need to release it */
            if(NULL != CallerList)
            {
                cssm_ReleaseCallerListEntry(CallerList,
                                            CSSM_READER_LOCK);
            }
#endif

        }
    }
    RETURN(err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_GenerateRandom
 *
 *Description:
 *   Exported API for generating random data
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  RandomNumber (output) - pointer to the random data
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
 CSSM_RETURN CSSMAPI CSSM_GenerateRandom (CSSM_CC_HANDLE CCHandle,
     CSSM_DATA_PTR RandomNumber)
 {
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_GenerateKey);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;

                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GenerateRandom))
                        {
                            ERR( err = CallBack->GenerateRandom (hCSP, CCHandle, Context, RandomNumber));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }
    RETURN(err)

}

/*---------------------------------------------------------------
 *
 *Name: CSSM_GenerateAlgorithmParams
 *
 *Description:
 *   Exported API for generating algorithm parameters
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  ParamBits (input) - number of bits in key exchange parameter
 *  Param (output) - pointer to the generated key exchange
 *                   parameter
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
 CSSM_RETURN CSSMAPI CSSM_GenerateAlgorithmParams (CSSM_CC_HANDLE CCHandle,
     uint32 ParamBits,
     CSSM_DATA_PTR Param)
 {
     CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
     CSSM_CONTEXT_PTR Context = NULL;
     CSSM_CSP_HANDLE hCSP;
     CSSM_RETURN err = CSSM_OK;
     CSSM_RETURN tempErr;
     cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
     cssm_INTERNAL_CONTEXT_PTR InternalCtx;
     CSSM_API_MEMORY_FUNCS_PTR memfuncs = NULL;
     CSSM_CONTEXT_ATTRIBUTE_PTR ContextAttributes = NULL;
     uint32 NumAttributes = 0;
    TRACE(cssm_trace_info, FNCODE_CSSM_GenerateAlgorithmParams);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_WRITER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;

                       /*
                        * Call the SPI function if it's valid.
                        */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GenerateAlgorithmParams))
                        {
                            ERR(err = CallBack->GenerateAlgorithmParams(hCSP,
                                                                        CCHandle,
                                                                        Context,
                                                                        ParamBits,
                                                                        Param,
                                                                        &NumAttributes,
                                                                        &ContextAttributes));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                        if ( err == CSSM_OK )
                        {
                             /* Are the return values correct? */
                             if ( ( NumAttributes != 0 ) && ( ContextAttributes == NULL ) )
                             {
                                 ERR(err = CSSMERR_CSSM_INVALID_ATTRIBUTE);
                             }
                             else if ( NumAttributes == 0 )
                             {
                                 err = CSSM_OK;
                             }
                             else
                             {
                                 /* Update the internal context and free the returned attributes */
                                ERR(err = cssm_UpdateAlgParamContextAttributes(Context,
                                                                               NumAttributes,
                                                                               ContextAttributes));
                                if(CSSM_OK == err)
                                {
                                    /* update any affected proxies */
                                    ERR(err = cssm_ReplaceAppCallbacksWithCssmProxy(InternalCtx,
                                                                                    InternalCtx->Context->NumberOfAttributes,
                                                                                    InternalCtx->Context->ContextAttributes));
                                }

                                ERR(tempErr = cssm_GetAppMemoryFunctions(AttachRecord,
                                                                         &memfuncs));
                                if((CSSM_OK == tempErr) &&
                                   (CSSM_OK == err))
                                {
                                    cssm_FreeAttributes(NumAttributes,
                                                        ContextAttributes,
                                                        memfuncs->free_func,
                                                        memfuncs->AllocRef,
                                                        1); 
                                }

                                if(CSSM_OK == err)
                                {
                                    err = tempErr;
                                }
                             }
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_WRITER_LOCK);
            }
        }
    }
    RETURN(err)

}

/*---------------------------------------------------------------
 *
 *Name: CSSM_WrapKey
 *
 *Description:
 *   Exported API
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  PassPhrase (input) - passphrase to unlock the wrapping key
 *  Key (input) - Key used to wrap
 *  DescriptiveData (input/optional) - additional descriptive data to be 
 *          associated and included with the key during the wrapping operation
 *  WrappedKey (putput) - pointer to the wrapped key
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
 /* Returns the key */
 CSSM_RETURN CSSMAPI CSSM_WrapKey (CSSM_CC_HANDLE CCHandle,
     const CSSM_ACCESS_CREDENTIALS *AccessCred,
     const CSSM_KEY *Key,
     const CSSM_DATA *DescriptiveData,
     CSSM_WRAP_KEY_PTR WrappedKey)
 {
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;

     CSSM_ACCESS_CREDENTIALS ProxyCallback;
     CSSM_ACCESS_CREDENTIALS_PTR pProxyCallback = NULL;
     CSSM_CHALLENGE_CALLBACK_PROXY_MODULEHANDLE ProxyCtx;

     CSSM_PRIVILEGE Privilege = 0;

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_GenerateKey);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;

						if (AccessCred != NULL)
						{
                            if( !port_IsBadReadPtr( AccessCred, sizeof(CSSM_ACCESS_CREDENTIALS) ) )
                            {
								 cssm_SetAccessCredProxyModuleHandle(  CSSM_UINT64_HIGH( CCHandle ), AccessCred, &ProxyCallback,&ProxyCtx);                               
								 pProxyCallback = &ProxyCallback;
							}
						}

                         Privilege = cssm_GetUSEETag();
                         if(CSSM_OK == err)
                         {
                            /*
                             * Call the SPI function if it's valid.
                             */
                            if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->WrapKey))
                            {
                                 ERR( err = CallBack->WrapKey (hCSP, 
                                                             CCHandle,
                                                             Context,
                                                             pProxyCallback,
                                                             Key,
                                                             DescriptiveData,
                                                             WrappedKey,
                                                             Privilege));
                            }
                            else
                            {
                                ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                            }

                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }
                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }
    RETURN(err)
}

CSSM_RETURN CSSMAPI CSSM_WrapKeyP(
                        CSSM_CC_HANDLE CCHandle,
                        const CSSM_ACCESS_CREDENTIALS *AccessCred,
                        const CSSM_KEY *Key,
                        const CSSM_DATA *DescriptiveData,
                        CSSM_WRAP_KEY_PTR WrappedKey,
                        CSSM_PRIVILEGE Privilege )
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;
    cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord = NULL;
    CSSM_ACCESS_CREDENTIALS ProxyCallback;
    CSSM_ACCESS_CREDENTIALS_PTR pProxyCallback = NULL;
    CSSM_CHALLENGE_CALLBACK_PROXY_MODULEHANDLE ProxyCtx;

    TRACE(cssm_trace_info, FNCODE_CSSM_WrapKeyP);
    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {
#ifdef ISL_INCLUDED
        cssm_CALLER_LIST_PTR CallerList = NULL;
        cssm_CALLER_LIST_ENTRY_PTR CallerListEntry = NULL;
        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if((0 != Privilege) ||
           (cssm_GetPVCPolicy() & CSSM_PVC_APP))
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           &CallerList,
                                                           CSSM_READER_LOCK,
                                                           &CallerListEntry));
            if(CSSM_OK != err)
            {
                if((cssm_GetPVCPolicy() & CSSM_PVC_APP))
                {
                    ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
                else
                {
                    err = CSSM_OK;
                }
            }
        }
#endif

        if(CSSM_OK == err)
        {
            /* get CSP handle and actual context information */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        &ServiceRecord,
                        CSSM_READER_LOCK,
                        &AttachRecord,
                        CSSM_READER_LOCK,
                        &InternalCtx,
                        CSSM_READER_LOCK)) ;
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
#ifdef ISL_INCLUDED
                if( Privilege != 0)
                {
                    if(!cssm_PrivilegeMatchForCallerAndSP( Privilege,
                                                       CallerListEntry,
                                                       AttachRecord,
                                                       ServiceRecord) )
                    {
                        err = CSSMERR_CSSM_PRIVILEGE_NOT_GRANTED;
                    }
                }
#endif
                cssm_ReleaseServiceLock( ServiceRecord, CSSM_READER_LOCK );
                if( err == CSSM_OK )
                {   
                    /* this function will lock the mutex if necessary */
                    ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                    if(CSSM_OK == err)
                    {

                        /* get the callback functions */
                        ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                              CSSM_SERVICE_CSP,
                                                              (void *)&CallBack)); /* ifdef VMS apply cast */
                        if(CSSM_OK == err)
                        {
                            Context = InternalCtx->Context;
                            hCSP = InternalCtx->CSPHandle;
                            /* Do we have a callback to proxy? */
							if (AccessCred != NULL)
							{
								if( !port_IsBadReadPtr( AccessCred, sizeof(CSSM_ACCESS_CREDENTIALS) ) )
								{
									cssm_SetAccessCredProxyModuleHandle(  CSSM_UINT64_HIGH( CCHandle ), AccessCred, &ProxyCallback, &ProxyCtx);                                
									pProxyCallback = &ProxyCallback;
								}
							}
                                /*
                                 * Call the callback function. 
                                 * Make sure that function structure and function pointer is valid 
                                 */
                            if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->WrapKey)) 
                            {
                                 ERR( err = CallBack->WrapKey (hCSP, 
                                                             CCHandle,
                                                             Context,
                                                             pProxyCallback,
                                                             Key,
                                                             DescriptiveData,
                                                             WrappedKey,
                                                             Privilege));
                            }
                            else 
                            {
                                err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                            }
                                                
                        }

                        /*
                        * cssm_ReleaseModuleRecord and multitask mutex.
                        */
                        cssm_UnlockMultiTaskMutex(AttachRecord);
                    }
                }
                cssm_ReleaseAttachLock(AttachRecord,
                                    CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                    CSSM_READER_LOCK);
            }

#ifdef ISL_INCLUDED
            /* if we locked the caller record, we need to release it */
            if(NULL != CallerList)
            {
                cssm_ReleaseCallerListEntry(CallerList,
                                            CSSM_READER_LOCK);
            }
#endif

        }
    }
    RETURN(err)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_UnwrapKey
 *
 *Description:
 *   Exported API
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  NewPassPhrase (input) - passphrase to unlock the wrapping key
 *  PublicKey (input/optional) - public key corresponding to the private key 
 *                      being unwrapped
 *  WrappedKey (input) - Key used to wrap
 *  KeyUsage (input/optional) - A bit mask indicating all permitted uses for 
 *                      the imported key
 *  KeyAttr (input) - A bit mask defining attribute values to be associated with 
 *                      the unwrapped key
 *  KeyLabel (input) - Label for key
 *  UnwrappedKey (output) - pointer to unwrapped key
 *  DescriptiveData (output) - any additional descriptive data that was associated
 *                          with the key during the wrapping operation
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
/* Returns the key */
CSSM_RETURN CSSMAPI CSSM_UnwrapKey (CSSM_CC_HANDLE CCHandle,
                                    const CSSM_KEY *PublicKey,
                                    const CSSM_WRAP_KEY *WrappedKey,
                                    uint32 KeyUsage,
                                    uint32 KeyAttr,
                                    const CSSM_DATA *KeyLabel,
                                    const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
                                    CSSM_KEY_PTR UnwrappedKey,
                                    CSSM_DATA_PTR DescriptiveData)
{

    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;
    CSSM_PRIVILEGE Privilege = 0;
    CSSM_RESOURCE_CONTROL_CONTEXT ProxyCallback;
    CSSM_RESOURCE_CONTROL_CONTEXT_PTR pProxyCallback= NULL;
    CSSM_CHALLENGE_CALLBACK_PROXY_MODULEHANDLE ChallengeProxyCtx;
    CSSM_SUBJECT_CALLBACK_PROXY_MODULEHANDLE SubjectProxyCtx;

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_UnwrapKey);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;
						if (CredAndAclEntry != NULL)
						{
                            if( !port_IsBadReadPtr( CredAndAclEntry, sizeof(CSSM_RESOURCE_CONTROL_CONTEXT) ) )
                            {
								cssm_SetResourceControlProxyModuleHandle(hCSP,CredAndAclEntry,&ChallengeProxyCtx,
													   &SubjectProxyCtx,&ProxyCallback);
                                 
								pProxyCallback = &ProxyCallback;
							}
						}
                            
                         Privilege = cssm_GetUSEETag();
                         if(CSSM_OK == err)
                         {
                            /*
                             * Call the SPI function if it's valid.
                             */
                            if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->UnwrapKey))
                            {
                                ERR( err = CallBack->UnwrapKey (hCSP, 
                                                                CCHandle,
                                                                Context,
                                                                PublicKey,
                                                                WrappedKey,
                                                                KeyUsage,
                                                                KeyAttr,
                                                                KeyLabel,
                                                                pProxyCallback,
                                                                UnwrappedKey,
                                                                DescriptiveData,
                                                                Privilege));
                            }
                            else
                            {
                                ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                            }

                        }

                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }
                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }
    RETURN(err)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_UnwrapKeyP
 *
 *Description:
 *   Exported API
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  NewPassPhrase (input) - passphrase to unlock the wrapping key
 *  PublicKey (input/optional) - public key corresponding to the private key 
 *                      being unwrapped
 *  WrappedKey (input) - Key used to wrap
 *  KeyUsage (input/optional) - A bit mask indicating all permitted uses for 
 *                      the imported key
 *  KeyAttr (input) - A bit mask defining attribute values to be associated with 
 *                      the unwrapped key
 *  KeyLabel (input) - Label for key
 *  UnwrappedKey (output) - pointer to unwrapped key
 *  DescriptiveData (output) - any additional descriptive data that was associated
 *                          with the key during the wrapping operation
 *  Privilege (input) - The privilege to be applied during cryptographic operation.
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
/* Returns the key */
CSSM_RETURN CSSMAPI CSSM_UnwrapKeyP (CSSM_CC_HANDLE CCHandle,
                                     const CSSM_KEY *PublicKey,
                                     const CSSM_WRAP_KEY *WrappedKey,
                                     uint32 KeyUsage,
                                     uint32 KeyAttr,
                                     const CSSM_DATA *KeyLabel,
                                     const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
                                     CSSM_KEY_PTR UnwrappedKey,
                                     CSSM_DATA_PTR DescriptiveData,
                                     CSSM_PRIVILEGE Privilege)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CSP_HANDLE hCSP;
    CSSM_RETURN err = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx;
    cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord = NULL;
    CSSM_RESOURCE_CONTROL_CONTEXT ProxyCallback;
    CSSM_RESOURCE_CONTROL_CONTEXT_PTR pProxyCallback= NULL;
    CSSM_CHALLENGE_CALLBACK_PROXY_MODULEHANDLE ChallengeProxyCtx;
    CSSM_SUBJECT_CALLBACK_PROXY_MODULEHANDLE SubjectProxyCtx;

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED
        cssm_CALLER_LIST_PTR CallerList = NULL;
        cssm_CALLER_LIST_ENTRY_PTR CallerListEntry = NULL;
        
        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if((0 != Privilege) ||
           (cssm_GetPVCPolicy() & CSSM_PVC_APP))
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           &CallerList,
                                                           CSSM_READER_LOCK,
                                                           &CallerListEntry));
            if(CSSM_OK != err)
            {
                if((cssm_GetPVCPolicy() & CSSM_PVC_APP))
                {
                    ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
                else
                {
                    err = CSSM_OK;
                }
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* get CSP handle and actual context information */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                        CSSM_NO_LOCK,
                        NULL,
                        CSSM_NO_LOCK,
                        &ServiceRecord,
                        CSSM_READER_LOCK,
                        &AttachRecord,
                        CSSM_READER_LOCK,
                        &InternalCtx,
                        CSSM_READER_LOCK)) ;
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
#ifdef ISL_INCLUDED
                if( Privilege != 0)
                {
                    if(!cssm_PrivilegeMatchForCallerAndSP( Privilege,
                                                       CallerListEntry,
                                                       AttachRecord,
                                                       ServiceRecord) )
                    {
                        err = CSSMERR_CSSM_PRIVILEGE_NOT_GRANTED;
                    }
                }
#endif
                cssm_ReleaseServiceLock( ServiceRecord, CSSM_READER_LOCK );
                if( err == CSSM_OK )
                {   
                    /* this function will lock the mutex if necessary */
                    ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                    if(CSSM_OK == err)
                    {

                        /* get the callback functions */
                        ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                              CSSM_SERVICE_CSP,
                                                              (void *)&CallBack)); /* ifdef VMS apply cast */
                        if(CSSM_OK == err)
                        {
                            Context = InternalCtx->Context;
                            hCSP = InternalCtx->CSPHandle;
							if (CredAndAclEntry != NULL)
							{
								if( !port_IsBadReadPtr( CredAndAclEntry, sizeof(CSSM_RESOURCE_CONTROL_CONTEXT) ) )
								{
									cssm_SetResourceControlProxyModuleHandle(hCSP,CredAndAclEntry,&ChallengeProxyCtx,
													   &SubjectProxyCtx,&ProxyCallback);
                                 
									pProxyCallback = &ProxyCallback;
								}
							}
                            
                            /*
                             * Call the callback function. 
                             * Make sure that function structure and function pointer is valid 
                             */
                            if (!port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->UnwrapKey)) 
                            {
                                ERR( err = CallBack->UnwrapKey (hCSP, 
                                                                CCHandle,
                                                                Context,
                                                                PublicKey,
                                                                WrappedKey,
                                                                KeyUsage,
                                                                KeyAttr,
                                                                KeyLabel,
                                                                pProxyCallback,
                                                                UnwrappedKey,
                                                                DescriptiveData,
                                                                Privilege));
                            }
                            else 
                            {
                                err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
                            }
                        }

                        /*
                        * cssm_ReleaseModuleRecord and multitask mutex.
                        */
                        cssm_UnlockMultiTaskMutex(AttachRecord);
                    }
                }

                cssm_ReleaseAttachLock(AttachRecord,
                                    CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                    CSSM_READER_LOCK);
            }

#ifdef ISL_INCLUDED
            /* if we locked the caller record, we need to release it */
            if(NULL != CallerList)
            {
                cssm_ReleaseCallerListEntry(CallerList,
                                            CSSM_READER_LOCK);
            }
#endif

        }
    }
    RETURN(err)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_DeriveKey
 *
 *Description:
 *   Exported API
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  BaseKey (input) - source key
 *  Param (input) - inputs to derive from
 *  StorageMask (input) - indicate how to store key
 *  KeyLabel (input) - label for key
 *  DerivedKey (output) - pointer to derived key
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
/* Returns the key */
CSSM_RETURN CSSMAPI CSSM_DeriveKey  (CSSM_CC_HANDLE CCHandle,
                                     CSSM_DATA_PTR Param,
                                     uint32 KeyUsage,
                                     uint32 KeyAttr,
                                     const CSSM_DATA *KeyLabel,
                                     const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
                                     CSSM_KEY_PTR DerivedKey )
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    CSSM_RESOURCE_CONTROL_CONTEXT ProxyCallback;
    CSSM_RESOURCE_CONTROL_CONTEXT_PTR pProxyCallback= NULL;
    CSSM_CHALLENGE_CALLBACK_PROXY_MODULEHANDLE ChallengeProxyCtx;
    CSSM_SUBJECT_CALLBACK_PROXY_MODULEHANDLE SubjectProxyCtx;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_DeriveKey);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;
						if (CredAndAclEntry != NULL)
						{
                            if( !port_IsBadReadPtr( CredAndAclEntry, sizeof(CSSM_RESOURCE_CONTROL_CONTEXT) ) )
                            {
								cssm_SetResourceControlProxyModuleHandle(hCSP,CredAndAclEntry,&ChallengeProxyCtx,
													   &SubjectProxyCtx,&ProxyCallback);
                                     
								pProxyCallback = &ProxyCallback;
							}
						}
                           
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->DeriveKey))
                        {
                                ERR( err = CallBack->DeriveKey (hCSP, 
                                                                CCHandle, 
                                                                Context, 
                                                                Param,
                                                                KeyUsage, 
                                                                KeyAttr, 
                                                                KeyLabel, 
                                                                pProxyCallback, 
                                                                DerivedKey));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }
    RETURN(err)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_FreeKey
 *
 *Description:
 *  Exported API to request the CSP to clean up any key material associated 
 *  with the key.  
 *
 *Parameters: 
 *  CSPHandle (input) - handle associated with the CSP.
 *  KeyPtr (input) - key to be freed.
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
 CSSM_RETURN CSSMAPI CSSM_FreeKey (
     CSSM_CSP_HANDLE CSPHandle,
     const CSSM_ACCESS_CREDENTIALS *AccessCred,
     CSSM_KEY_PTR KeyPtr,
     CSSM_BOOL Delete )
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;

    CSSM_RETURN err = CSSM_OK;
    

    CSSM_ACCESS_CREDENTIALS ProxyCallback;
    CSSM_ACCESS_CREDENTIALS_PTR pProxyCallback = NULL;
    CSSM_CHALLENGE_CALLBACK_PROXY_MODULEHANDLE ProxyCtx;

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_FreeKey);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR( err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                 CSSM_SERVICE_CSP,
                                                 (void **)&CallBack,
                                                 &AttachRecord));
            if(CSSM_OK == err)
            {
                /* get the callback functions */
                ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                      CSSM_SERVICE_CSP,
                                                      (void *)&CallBack)); /* ifdef VMS apply cast */
                if(CSSM_OK == err)
                {
                    /* This is the only function that does not have a CCHandle*/
                    /* So create a special proxy just for this. */
                    if( AccessCred != NULL) 
					{
                        if( !port_IsBadReadPtr( AccessCred, sizeof(CSSM_ACCESS_CREDENTIALS) ) )
                        {
							cssm_SetAccessCredProxyModuleHandle(  CSPHandle, AccessCred,&ProxyCallback,&ProxyCtx);                              
							
							pProxyCallback = &ProxyCallback;
						}
					}

                    /*
                    * Call the SPI function if it's valid.
                    */
                    if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->FreeKey))
                    {
                        ERR( err = CallBack->FreeKey (CSPHandle, pProxyCallback, KeyPtr, Delete));
                    }
                    else
                    {
                        ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                    }

                }

                /*
                * release multitask mutex.
                */
                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);                    
            }
        }
    }

    RETURN(err)
}


/*---------------------------------------------------------------
 *
 *Name: CSSM_CSP_PassThrough
 *
 *Description:
 *   Exported API for handling CSP specific operations not covered
 *   in the crypto API
 *
 *Parameters: 
 *  CCHandle (input) - the context handle
 *  PassThroughId (input) - CSP specific ID for identifying operation
 *  InData (input) - pointer to input date for the operation
 *
 *Returns:
 *   not NULL - pointer to output results from operation
 *   NULL - CSP wasn't able to complete function
 *
 *-------------------------------------------------------------*/
 CSSM_RETURN CSSMAPI CSSM_CSP_PassThrough (CSSM_CC_HANDLE CCHandle, 
     uint32 PassThroughId,
     const void * InData,
     void **OutData)
 {

    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR InternalCtx = NULL;
    CSSM_CONTEXT_PTR Context = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_CSP_HANDLE hCSP;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_GenerateKey);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /* lock the attach and context records */
            ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_READER_LOCK,
                                                   &InternalCtx,
                                                   CSSM_READER_LOCK));
            if(CSSM_OK != err)
            {
                /* we couldn't find the context */
                ERR(err = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE);
            }
            else
            {
                /* this function will lock the mutex if necessary */
                ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        Context = InternalCtx->Context;
                        hCSP = InternalCtx->CSPHandle;

                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->PassThrough))
                        {
                            ERR( err = CallBack->PassThrough (hCSP, 
                                                              CCHandle,
                                                              Context,
                                                              PassThroughId,
                                                              InData,
                                                              OutData));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                    /*
                     * release multitask mutex.
                     */
                    cssm_UnlockMultiTaskMutex(AttachRecord);
                }

                /* release the context and attach records */
                cssm_ReleaseAttachLock(AttachRecord,
                                       CSSM_READER_LOCK);
                cssm_ReleaseContextLock(InternalCtx,
                                        CSSM_READER_LOCK);
            }
        }
    }
    RETURN(err)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_CSP_Login
 *
 *Description:
 *   Exported API for Loging into CSPs.
 *
 *Parameters: 
 *  CSPHandle (input) - the target CSP handle
 *  Password (input) - passphrase for login
 *  pReserved (input) - pointer to input date for the operation
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
 CSSM_RETURN CSSMAPI CSSM_CSP_Login (CSSM_CSP_HANDLE CSPHandle,
     const CSSM_ACCESS_CREDENTIALS *AccessCred,
     const CSSM_DATA *LoginName,
     const void *Reserved)
 {
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_ACCESS_CREDENTIALS ProxyCallback;
    CSSM_ACCESS_CREDENTIALS_PTR pProxyCallback = NULL;
    CSSM_CHALLENGE_CALLBACK_PROXY_MODULEHANDLE ProxyCtx;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_CSP_Login);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
				if (AccessCred != NULL)
				{
                    if( !port_IsBadReadPtr( AccessCred, sizeof(CSSM_ACCESS_CREDENTIALS) ) )
                    {
						cssm_SetAccessCredProxyModuleHandle(  CSPHandle, AccessCred,&ProxyCallback,&ProxyCtx);                              
						pProxyCallback = &ProxyCallback;
					}
				}
                  
                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->Login))
                        {
                            ERR( err = CallBack->Login (CSPHandle, pProxyCallback, LoginName, Reserved));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)
 }

/*---------------------------------------------------------------
 *
 *Name: CSSM_CSP_Logout
 *
 *Description:
 *   Exported API for Loging of out a CSP.
 *
 *Parameters: 
 *  CSPHandle (input) - the target CSP handle
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
 CSSM_RETURN CSSMAPI CSSM_CSP_Logout  (CSSM_CSP_HANDLE CSPHandle)
 {
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;

    CSSM_RETURN err = CSSM_OK;
    

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_CSP_Logout);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->Logout))
                        {
                            ERR( err = CallBack->Logout (CSPHandle) );
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)
 }

/*---------------------------------------------------------------
 *
 *Name: CSSM_CSP_ChangeLoginAcl
 *
 *Description:
 *   Exported API for Changing the Login ACL.
 *
 *Parameters: 
 *  CSPHandle (input) - the target CSP handle
 *  OldPassword (input) - old passphrase for login
 *  NewPassword (input) - new passphrase for login
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
 CSSM_RETURN CSSMAPI CSSM_CSP_ChangeLoginAcl (CSSM_CSP_HANDLE CSPHandle,
     const CSSM_ACCESS_CREDENTIALS *AccessCred,
     const CSSM_ACL_EDIT *AclEdit)
 {
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;

    CSSM_RETURN err = CSSM_OK;
    
    CSSM_ACCESS_CREDENTIALS ProxyCallback;
    CSSM_ACCESS_CREDENTIALS_PTR pProxyCallback = NULL;
    CSSM_CHALLENGE_CALLBACK_PROXY_MODULEHANDLE ProxyCtx;


    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_CSP_ChangeLoginAcl);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
				if (AccessCred != NULL)
				{
                    if( !port_IsBadReadPtr( AccessCred, sizeof(CSSM_ACCESS_CREDENTIALS) ) )
                    {
						cssm_SetAccessCredProxyModuleHandle(  CSPHandle, AccessCred,&ProxyCallback,&ProxyCtx);                              
						pProxyCallback = &ProxyCallback;
					}
				}
                 
                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->ChangeLoginAcl))
                        {
                            ERR( err = CallBack->ChangeLoginAcl (CSPHandle, pProxyCallback, AclEdit));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_CSP_ObtainPrivateKeyFromPublicKey
 *
 *Description:
 *   Exported API for associating the keypair held by CSP.
 *
 *Parameters: 
 *  CSPHandle (input) - the target CSP handle
 *  PublicKey (input) - the public key
 *  PrivateKey (input) - the private key
 *
 *Returns:
 *   CSSM_FAIL - CSP wasn't able to complete function
 *   CSSM_OK - Call successful executed by CSP add-in
 *
 *-------------------------------------------------------------*/
 CSSM_RETURN CSSMAPI CSSM_CSP_ObtainPrivateKeyFromPublicKey (
     CSSM_CSP_HANDLE CSPHandle,
     const CSSM_KEY *PublicKey,
     CSSM_KEY_PTR PrivateKey)
 {
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;
    

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_CSP_ObtainPrivateKeyFromPublicKey);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->ObtainPrivateKeyFromPublicKey))
                        {
                            ERR( err = CallBack->ObtainPrivateKeyFromPublicKey (CSPHandle, 
                                                                                PublicKey,
                                                                                PrivateKey));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)
 }

/*---------------------------------------------------------------
 *
 *Name: CSSM_RetrieveUniqueId
 *
 *Description:
 *   This function returns an identifier that could be used to uniquely
 *   differentiate the CSP from all other devices from the same vendor or 
 *   different vendors.
 *
 *Parameters: 
 *   CSPHandle - handle of the CSP
 *   UniqueID - pointer to CSSM_DATA buffer in which Unique id will be returned.
 *
 *Returns:
 *   CSSM_OK - In case ID is sucesfully retrieved.
 *   CSSM_FAIL - In case of error.
 *
 *-------------------------------------------------------------*/

 CSSM_RETURN CSSMAPI CSSM_RetrieveUniqueId (CSSM_CSP_HANDLE CSPHandle,
     CSSM_DATA_PTR UniqueID)
 {
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;
    

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_RetrieveUniqueId);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->RetrieveUniqueId))
                        {
                            ERR( err = CallBack->RetrieveUniqueId (CSPHandle, UniqueID));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)
 }

/*---------------------------------------------------------------
 *
 *Name: CSSM_RetrieveCounter
 *
 *Description:
 *      This function returns the value of a tamper resistant clock/counter of 
 *      the CSP.  
 *
 *Parameters: 
 *   CSPHandle - handle of the CSP
 *   Counter - pointer to CSSM_DATA buffer in which counter will be returned.
 *
 *Returns:
 *   CSSM_OK - In case ID is sucesfully retrieved.
 *   CSSM_FAIL - In case of error.
 *
 *-------------------------------------------------------------*/

 CSSM_RETURN CSSMAPI CSSM_RetrieveCounter (CSSM_CSP_HANDLE CSPHandle,
                                           CSSM_DATA_PTR Counter)
 {
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;
    

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_RetrieveCounter);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->RetrieveCounter))
                        {
                             ERR( err = CallBack->RetrieveCounter (CSPHandle, Counter));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)
 }

/*---------------------------------------------------------------
 *
 *Name: CSSM_VerifyDevice
 *
 *Description:
 *   This function triggers the CSP module to perform self verification 
 *   and integrity checking.
 *
 *Parameters: 
 *   CSPHandle - handle of the CSP
 *   DeviceCert - pointer to CSSM_DATA buffer containing the data identifying
 *                the CSP.
 *
 *Returns:
 *   CSSM_OK - In case verification was successful.
 *   CSSM_FAIL - In case of error.
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_VerifyDevice (
     CSSM_CSP_HANDLE CSPHandle,
     const CSSM_DATA *DeviceCert)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;
    

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_VerifyDevice);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->VerifyDevice))
                        {
                            ERR( err = CallBack->VerifyDevice (CSPHandle, DeviceCert));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)
}



CSSM_RETURN CSSMAPI CSSM_GetTimeValue(
     CSSM_CSP_HANDLE CSPHandle,
     CSSM_ALGORITHMS TimeAlgorithm,
     CSSM_DATA *TimeData)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;
    

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_GetTimeValue);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GetTimeValue))
                        {
                            ERR( err = CallBack->GetTimeValue (CSPHandle, TimeAlgorithm, TimeData));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)
}



CSSM_RETURN CSSMAPI CSSM_CSP_GetOperationalStatistics(
     CSSM_CSP_HANDLE CSPHandle,
     CSSM_CSP_OPERATIONAL_STATISTICS *Statistics )
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;
    

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_GetOperationalStatistics);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
                if(CSSM_OK == err)
                {
                    /* get the callback functions */
                    ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                          CSSM_SERVICE_CSP,
                                                          (void *)&CallBack)); /* ifdef VMS apply cast */
                    if(CSSM_OK == err)
                    {
                        /*
                         * Call the SPI function if it's valid.
                         */
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GetOperationalStatistics))
                        {
                            ERR( err = CallBack->GetOperationalStatistics (CSPHandle, Statistics));
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }

                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)
}



CSSM_RETURN CSSMAPI CSSM_CSP_GetLoginAcl(
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_STRING *SelectionTag,
                        uint32 *NumberOfAclInfos,
                        CSSM_ACL_ENTRY_INFO_PTR *AclInfos )
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;
    

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_GetLoginAcl);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
                /* get the callback functions */
                ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                      CSSM_SERVICE_CSP,
                                                      (void *)&CallBack)); /* ifdef VMS apply cast */
                if(CSSM_OK == err)
                {
                    /*
                     * Call the SPI function if it's valid.
                     */
                    if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GetLoginAcl))
                    {
                        ERR( err = CallBack->GetLoginAcl(CSPHandle, SelectionTag, NumberOfAclInfos , AclInfos));
                    }
                    else
                    {
                        ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                    }
                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)
}



CSSM_RETURN CSSMAPI CSSM_GetKeyAcl(
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_KEY *Key,
                        const CSSM_STRING *SelectionTag,
                        uint32 *NumberOfAclInfos,
                        CSSM_ACL_ENTRY_INFO_PTR *AclInfos)
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;
    

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_GetKeyAcl);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
                /* get the callback functions */
                ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                      CSSM_SERVICE_CSP,
                                                      (void *)&CallBack)); /* ifdef VMS apply cast */
                if(CSSM_OK == err)
                {
                    /*
                     * Call the SPI function if it's valid.
                     */
                    if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GetKeyAcl))
                    {
                        ERR( err = CallBack->GetKeyAcl(CSPHandle, Key, SelectionTag, NumberOfAclInfos, AclInfos));
                    }
                    else
                    {
                        ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                    }
                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)
}



CSSM_RETURN CSSMAPI CSSM_ChangeKeyAcl(
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_ACCESS_CREDENTIALS *AccessCred,
                        const CSSM_ACL_EDIT *AclEdit,
                        const CSSM_KEY *Key )
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;
    
     CSSM_ACCESS_CREDENTIALS ProxyCallback;
     CSSM_ACCESS_CREDENTIALS_PTR pProxyCallback = NULL;
     CSSM_CHALLENGE_CALLBACK_PROXY_MODULEHANDLE ProxyCtx;

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_ChangeKeyAcl);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
                /* get the callback functions */
                ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                      CSSM_SERVICE_CSP,
                                                      (void *)&CallBack)); /* ifdef VMS apply cast */
                if(CSSM_OK == err)
                {
					if (AccessCred != NULL)
					{
                        if( !port_IsBadReadPtr( AccessCred, sizeof(CSSM_ACCESS_CREDENTIALS) ) )
                        {
							 cssm_SetAccessCredProxyModuleHandle(  CSPHandle, AccessCred,&ProxyCallback,&ProxyCtx);                                 
							 pProxyCallback = &ProxyCallback;
						}
					}

                    /*
                     * Call the SPI function if it's valid.
                     */
                    if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->ChangeKeyAcl))
                    {
                        ERR( err = CallBack->ChangeKeyAcl(CSPHandle, pProxyCallback, AclEdit, Key));
                    }
                    else
                    {
                        ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                    }
                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)
}



CSSM_RETURN CSSMAPI CSSM_GetKeyOwner(
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_KEY *Key,
                        CSSM_ACL_OWNER_PROTOTYPE_PTR Owner )
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;
    

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_GetKeyOwner);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
                /* get the callback functions */
                ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                      CSSM_SERVICE_CSP,
                                                      (void *)&CallBack)); /* ifdef VMS apply cast */
                if(CSSM_OK == err)
                {
                    /*
                     * Call the SPI function if it's valid.
                     */
                    if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GetKeyOwner))
                    {
                        ERR( err = CallBack->GetKeyOwner(CSPHandle, Key, Owner));
                    }
                    else
                    {
                        ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                    }
                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)

}



CSSM_RETURN CSSMAPI CSSM_ChangeKeyOwner(
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_ACCESS_CREDENTIALS *AccessCred,
                        const CSSM_KEY *Key,
                        const CSSM_ACL_OWNER_PROTOTYPE *NewOwner )
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;
     CSSM_ACCESS_CREDENTIALS ProxyCallback;
     CSSM_ACCESS_CREDENTIALS_PTR pProxyCallback = NULL;
     CSSM_CHALLENGE_CALLBACK_PROXY_MODULEHANDLE ProxyCtx;
    

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_ChangeKeyOwner);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
                /* get the callback functions */
                ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                      CSSM_SERVICE_CSP,
                                                      (void *)&CallBack)); /* ifdef VMS apply cast */
                if(CSSM_OK == err)
                {
					if (AccessCred != NULL)
					{
                        if( !port_IsBadReadPtr( AccessCred, sizeof(CSSM_ACCESS_CREDENTIALS) ) )
                        {
							 cssm_SetAccessCredProxyModuleHandle(  CSPHandle, AccessCred,&ProxyCallback,&ProxyCtx);                                 
							 pProxyCallback = &ProxyCallback;
						}
					}

                    /*
                     * Call the SPI function if it's valid.
                     */
                    if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->ChangeKeyOwner))
                    {
                        ERR( err = CallBack->ChangeKeyOwner(CSPHandle, pProxyCallback, Key, NewOwner));
                    }
                    else
                    {
                        ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                    }
                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)
}



CSSM_RETURN CSSMAPI CSSM_CSP_GetLoginOwner(
                        CSSM_CSP_HANDLE CSPHandle,
                        CSSM_ACL_OWNER_PROTOTYPE_PTR Owner )
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;
    

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_GetLoginOwner);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
                /* get the callback functions */
                ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                      CSSM_SERVICE_CSP,
                                                      (void *)&CallBack)); /* ifdef VMS apply cast */
                if(CSSM_OK == err)
                {
                    /*
                     * Call the SPI function if it's valid.
                     */
                    if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->GetLoginOwner))
                    {
                        ERR(err = CallBack->GetLoginOwner(CSPHandle, Owner)); 
                    }
                    else
                    {
                        ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                    }
                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)
}



CSSM_RETURN CSSMAPI CSSM_CSP_ChangeLoginOwner(
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_ACCESS_CREDENTIALS *AccessCred,
                        const CSSM_ACL_OWNER_PROTOTYPE *NewOwner )
{
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;

    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;
     CSSM_ACCESS_CREDENTIALS ProxyCallback;
     CSSM_ACCESS_CREDENTIALS_PTR pProxyCallback = NULL;
     CSSM_CHALLENGE_CALLBACK_PROXY_MODULEHANDLE ProxyCtx;
    

    /* get CSP handle and actual context information */
    TRACE(cssm_trace_info, FNCODE_CSSM_ChangeLoginOwner);

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            if(ERR(err = cssm_GetAttachRecordAndLockMultiTask(CSPHandle,
                                                              CSSM_SERVICE_CSP,
                                                              (void **)&CallBack,
                                                              &AttachRecord)) == CSSM_OK)
            {
                /* get the callback functions */
                ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                      CSSM_SERVICE_CSP,
                                                      (void *)&CallBack)); /* ifdef VMS apply cast */
                                                      
                if(CSSM_OK == err)
                {
                    /* Do we have a callback to proxy? */
					if (AccessCred != NULL)
					{
                        if( !port_IsBadReadPtr( AccessCred, sizeof(CSSM_ACCESS_CREDENTIALS) ) )
                        {
							 cssm_SetAccessCredProxyModuleHandle(  CSPHandle, AccessCred,&ProxyCallback,&ProxyCtx);                                 
							 pProxyCallback = &ProxyCallback;
						}
					}

                    /*
                     * Call the SPI function if it's valid.
                     */
                    if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->ChangeLoginOwner))
                    {
                        ERR( err = CallBack->ChangeLoginOwner(CSPHandle, pProxyCallback, NewOwner));
                    }
                    else
                    {
                        ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                    }
                }

                cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord);
            }
        }
    }
    RETURN(err)
}
